#!/usr/bin/perl

# This small program acts as an API service to view logs and perform certain functions.
#
# Endpoints:
#  errlog -- View the contents of the 5 most recent OUF error log files
#  trace  -- View the contents of the 5 most recent OUF trace log files
#  start  -- Start all OUF and Perl scripts (same as all_start.sh)
#  stop   -- Stop all OUF and Perl scripts (same as all_stop.sh)
#
# As a thin layer of security an API key is required and is passed in "key" parameter
# Ex. http://localhost:3030/errlog?key=123456789

use Config::Tiny;
use JSON;
use strict;
use warnings;

my $confdir = 'confdir';
my $conf = Config::Tiny->read("$confdir/other.ini");
my $ouf_conf = Config::Tiny->read("$confdir/ouf.ini");

my $json = JSON->new;
$json->canonical(1);
$json->pretty(1);

{
  package OUF::API::Server;
  
  use HTTP::Server::Simple::CGI;
  use base qw(HTTP::Server::Simple::CGI);
  use File::Copy;
  
  my %dispatch = (
    '/' => \&resp_menu,
    '/errlog' => \&resp_errlog,
    '/start' => \&resp_start,
    '/stop' => \&resp_stop,
    '/status' => \&resp_status,
    '/trace' => \&resp_trace,
    '/settings' => \&resp_settings,
    '/picapull_log' => \&resp_pp,
    '/decode_log' => \&resp_dc,
    '/save_ini' => \&resp_save_ini,
    '/view_env' => \&resp_view_env,
    '/files' => \&resp_files,
    '/harvester' => \&resp_harvester,
    '/prune' => \&resp_prune
  );
  
  sub handle_request {
    my $self = shift;
    my $cgi  = shift;
  
    my $path = $cgi->path_info();
    my $handler = $dispatch{$path};
    my $apikey;
    if ($cgi->{param}->{key}) {
      $apikey = $cgi->{param}->{key}[0];
    } elsif ($cgi->url_param('key')) {
      $apikey = $cgi->url_param('key');
    } else {
      $apikey = '';
    }

    if ($apikey ne $conf->{api}->{key}) {
      my $msg;
      print "HTTP/1.0 403 Forbidden\r\n";
      print $cgi->header( -type => 'application/json', -access_control_allow_origin => '*');
      if ($apikey) {
        $msg = 'Wrong API key'
      } else {
        $msg = 'The parameter "key" is required'
      }
      my $out = { status => 'error', message => $msg };
      print $json->encode($out);
    } elsif (ref($handler) eq "CODE") {
      print "HTTP/1.0 200 OK\r\n";
      print $cgi->header( -type => 'application/json; charset=utf-8', -access_control_allow_origin => '*', -access_control_allow_headers => '*', -access_control_allow_methods => '*');
      $handler->($cgi);
    } else {
      print "HTTP/1.0 404 Not found\r\n";
      print $cgi->header( -type => 'application/json', -access_control_allow_origin => '*');
      my $out = { status => 'error', message => "Not found" };
      print $json->encode($out);
    }
  }

  sub resp_menu {
    my $cgi  = shift;
    return if !ref $cgi;
    
    my $out = { endpoints => [] };

    my @ends = qw(start stop status errlog trace picapull_log decode_log settings view_env save_ini files harvester prune);
    foreach (@ends) {
      push @{ $out->{endpoints} }, $_;
    }
    print $json->encode($out);
  }

  sub resp_view_env {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out = {};

    for (sort keys %ENV) {
      $out->{$_} = $ENV{$_};
    }
    
    print $json->encode($out);
  }

  sub resp_files {
    my $cgi  = shift;
    return if !ref $cgi;
    my $type = $cgi->{param}->{type}[0] || '';
    my $name = $cgi->{param}->{name}[0] || '';
    my $subdir = $cgi->{param}->{sub}[0] || '';
    my $method = $cgi->request_method();

    my $out = {};
    my $dir;
    
    if ($type eq 'fetch') {
      $dir = $ouf_conf->{Directories}->{'Fetch Request Dir'};
    } elsif ($type eq 'titles') {
      $dir = $ouf_conf->{Directories}->{'Title Dir'};
    } elsif ($type eq 'updsig') {
      $dir = $ouf_conf->{Directories}->{'Update Signal Dir'};
    } elsif ($type eq 'updsig-done') {
      $dir = $conf->{dir}->{processed};
    } elsif ($type eq 'titles-done') {
      $dir = $conf->{dir}->{done};
    } else {
      print '{"status": "error", "message": "The \"type\" parameter is required or is not valid"}';
      return;
    }

    $out->{directory} = $dir;

    if ($name && !$subdir) {
      my $contents = '';
      my $err = 0;
      my $path = "$dir/$name";
      open RFILE, $path or $err = 1;
      if ($err) {
        print '{"status":"error", "Can\'t open input file"}'; 
        return; 
      }
      while (<RFILE>) {
        $contents .= $_;
      }
      $out->{contents} = $contents;

    } else {
      $out->{files} = [];
      my $c = 0;
      if ($subdir) {
        $dir .= "/$name";
      }
      foreach (<$dir/*>) {
        $c++;
        if ($method eq 'DELETE') {
          unlink $_;
          $out->{message} = "$c files deleted";
        } else {
          s/^.+\///;
          push @{ $out->{files} }, $_;
        }
      }
      if ($method eq 'DELETE' && $subdir) {
        rmdir $dir;
      }
      $out->{count} = $c;
      $out->{type} = $type;
      if ($type =~ /-done/ && !$subdir) {
        $out->{isSub} = 1;
      } else {
        $out->{isSub} = 0;
      }
    }
    print $json->encode($out);
  }

  sub resp_harvester {
    use LWP::Simple;
    use XML::LibXML;

    my $cgi  = shift;
    return if !ref $cgi;

    my $com = $cgi->param('command') || '';
    my $from = $cgi->param('from') || '';
    my $jid = $cgi->param('jid') || '';

    if (!$com) {
      print '{"status":"error", "message","The command parameter is required.  Options: log, failed-records"}';
      return;
    }

    my $url = $ENV{HARVESTER_URL} || $conf->{harvester}->{url};

    my $out = { harvesterUrl => $url };
    my $ttl = 0;

    if (!$jid && $com ne 'list') {
      my $lib = $ouf_conf->{UpdateSignal}->{Library};
      my $xml = get("$url/harvester/records/harvestables/?query=(usedBy=$lib)");
      if ($xml) {
        my $dom = XML::LibXML->load_xml(string => $xml);
        $jid = $dom->findvalue('//id');
      } else {
        print '{"status":"error", "message":"There was a problem connecting to the harvester"}';
        return;
      }
    }

    if (!$jid && $com ne 'list') {
      print '{"status":"error", "message":"Can\'t find default harvest job"}';
      return; 
    }
    if ($com eq 'list') {
      my $xlist = get("$url/harvester/records/harvestables");
      my $ldom;
      eval { $ldom = XML::LibXML->load_xml(string => $xlist) };
      if ($@) { 
        print '{"status":"error", "message":"Can\'t connect to harvester at ' . $url . '" }';
        return;
      }
      $out->{jobs} = [];
      foreach ($ldom->findnodes('//harvestableBrief')) {
        $ttl++;
        my $job = {};
        $job->{id} = $_->findvalue('./id');
        $job->{name} = $_->findvalue('./name');
        $job->{status} = $_->findvalue('./currentStatus');
        push @{ $out->{jobs} }, $job;
      }
    } elsif ($com eq 'log') {
      $out->{jobId} = $jid;
      $out->{lines} = [];
      my $log_url = "$url/harvester/records/harvestables/$jid/log?from=$from";
      my $rawlog = get($log_url) || '';
      foreach (split /\n/, $rawlog) {
        $ttl++;
        push @{ $out->{lines} }, $_;
      }
    } elsif ($com eq 'failed-records' || $com eq 'failed-records-fetch') {
      $out->{jobId} = $jid;
      $out->{'failedRecords'} = [];
      my $furl = "$url/harvester/records/harvestables/$jid/failed-records";
      my $fxml = get($furl);
      my $dom = XML::LibXML::Document->new();
      if ($fxml) {
        $dom = XML::LibXML->load_xml(string => $fxml);
      }
      foreach ($dom->findnodes('/failed-records/record')) {
        $ttl++;
        my $rec = {};
        $rec->{file} = {};
        my $ppn = $_->findvalue('./file/name');
        $rec->{file}->{name} = $ppn;
        $rec->{file}->{date} = $_->findvalue('./file/date');
        $rec->{url} = $_->findvalue('./url');
        if ($com eq 'failed-records-fetch') {
          $ppn =~ s/\..+$//;
          my $fn = "fetch_$ppn";
          my $dir = $ouf_conf->{Directories}->{'Fetch Request Dir'};
          open FFILE, ">$dir/$fn";
          print FFILE $ppn;
        }
        push @{ $out->{'failedRecords'} }, $rec;
      }
    } else {
      print '{"status":"error", "message","Not a valid command value.  Valid entries are: log, failed-records, failed-records-fetch"}';
      return;
    }
    $out->{totalRecords} = $ttl;
    print $json->encode($out);
  }

  sub resp_prune {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out = { directories => [] };
    
    my @dirs = qw(xml done processed);
    foreach my $dir (@dirs) {
      my $dcount = 0;
      my $dobj = { name => $dir, deleted => [] };
      foreach my $f (<"rundir/$dir/*">) {
        my @s = stat $f;
        my $atime = $s[8];
        my $mtime = $s[9];
        my $seen = $atime - $mtime;
        my $time = time();
        my $day = $time - $mtime;
        if ($day > 86400) {
          my $confirm = 1;
          if ($dir eq 'xml' && $seen < 4) {
            $confirm = 0;
          }
          if ($confirm) {
            unlink $f;
            my ($fn) = $f =~ m/^.+\/(.+)/;
            push @{ $dobj->{deleted} }, $fn;
            $dcount++;
          }
        }
      }
      $dobj->{count} = $dcount;
      push @{ $out->{directories} }, $dobj;
    }
    print $json->encode($out);
  }
  
  sub resp_errlog {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out = { lines => [] };

    my @list = <"rundir/err/*">;
    open ER, "<:encoding(utf8)", $list[-1];
    my @lines = reverse <ER>;
    foreach (@lines) {
      chomp;
      push @{ $out->{lines} }, $_;
    }
    close ER;
    print $json->encode($out);
  }

  sub resp_trace {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out = { lines => [] };
    
    my $tradir = $ouf_conf->{Trace}->{Dir};
    my @list = <"$tradir/*">;
    open TRC, "<:encoding(utf8)", $list[-1];
    my @lines = reverse <TRC>;
    foreach (@lines) {
      chomp;
      push @{ $out->{lines} }, $_;
    }
    close TRC;
    print $json->encode($out);
  }

  sub resp_start {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out = { messages => [] };

    my $newts = $cgi->param('ts');
    if ($newts) {
      if ($newts =~ /^(\d\d)-(\d\d)-(\d{4}) (\d\d):(\d\d):(\d\d)$/) {
        if ($6 > 59 || $5 > 59 || $4 > 23 || $2 > 12 || $1 > 31) {
          push @{ $out->{messages} }, "ERROR: $newts contains bad date/time value!";
          print $json->encode($out);
          return;
        } else {
          $newts .= '.000000';
          push @{ $out->{messages} }, "Setting \"Last Time\" to $newts\n\n";
          open INI, "<:encoding(utf8)", "$confdir/ouf.ini" or print "ERROR Can't open ouf.ini for reading\n\n";
          my @out;
          while (<INI>) {
            chomp;
            s/^\s*(Last Time\s*=\s*).+/$1$newts/;
            push @out, $_;
          };
          close INI;
          my $newini = join "\n", @out;
          open INI, ">:encoding(utf8)", "$confdir/ouf.ini" or print "ERROR Can't save to ouf.ini\n\n";
          print INI $newini;
          close INI,
        }
      } else {
        push @{ $out->{messages} }, "ERROR: Invalid timestamp format! Use DD-MM-YYYY hh:mm:ss";
        print $json->encode($out);
        return;
      }
    }
    my $new_conf = Config::Tiny->read("$confdir/ouf.ini");
    my $curts = $new_conf->{UpdateSignal}->{'Last Time'};
    $out->{last_time} = $curts;
    my $res = `./all_start.sh`;
    foreach (split /\n/, $res) {
      push @{ $out->{messages} }, $_;
    }
    print $json->encode($out);
  }

  sub resp_stop {
    my $cgi  = shift;
    return if !ref $cgi;

    my $sig = $cgi->param('sig') || '15';

    my $out = { messages => [] };
    
    my $res = `./all_stop.sh $sig`;
    foreach (split /\n/, $res) {
      push @{ $out->{messages} }, $_;
    }
    print $json->encode($out) ;
  }

  sub resp_status {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out = { scripts => [] };
    
    my @progs = qw(ouf_update ouf_fetch picapull.pl decode.pl);
    foreach (@progs) {
      my $ps = `ps ax | grep $_ | grep -v grep`;
      my $pid = 0;
      if ($ps =~ s/^\s*(\d+)//) {
        $pid = $1;
      }
      my $script = {};
      $script->{name} = $_;
      if ($pid) {
        $script->{pid} = $pid;
        $script->{status} = 'running';
      } else {
        $script->{status} = 'stopped';
      }
      push @{ $out->{scripts} }, $script;
    }
    print $json->encode($out);
  }

  sub resp_settings {
    my $cgi  = shift;
    return if !ref $cgi;

    my $type = $cgi->param('type') || 'ouf';
    
    my $out = { lines => [] };

    open INI, "confdir/$type.ini" or $out->{error} = { status => 'error', message => "Can't open $type.ini" };
    my $raw = '';
    while (<INI>) {
      $raw .= $_;
      chomp;
      push @{ $out->{lines} }, $_;
    }
    $out->{raw} = $raw;
    print $json->encode($out);
  }

  sub resp_pp {
    my $cgi  = shift;
    return if !ref $cgi;

    my $out->{lines} = [];
    
    open PPL, "<:encoding(utf8)", $conf->{log}->{picapull} or push @{ $out->{lines} }, "WARN Can't open picapull log.\n";
    my @lines = reverse <PPL>;
    foreach (@lines) {
      chomp;
      push @{ $out->{lines} }, $_;
    }
    close PPL; 
    print $json->encode($out);
  }

  sub resp_dc {
    my $cgi  = shift;
    return if !ref $cgi;
    
    my $out->{lines} = [];

    open DC, "<:encoding(utf8)", $conf->{log}->{decode} or push @{ $out->{lines} }, "WARN Can't open decode log.\n";
    my @lines = reverse <DC>;
    foreach (@lines) {
      chomp;
      push @{ $out->{lines} }, $_;
    }
    close DC;
    print $json->encode($out); 
  }

  sub resp_save_ini {
    my $cgi  = shift;
    return if !ref $cgi;

    my $method = $cgi->request_method();
    my $ct = $cgi->content_type();
    my $out = {};
    my $err = 0;

    if ($method ne 'POST') {
      print '{"status": "error", "message": "Only POST method is allowed"}';
    } elsif ($ct !~ /json/) {
      print '{"status": "error", "message": "Header must be application/json"}';
    } else {
      my $pd = $cgi->param('POSTDATA');
      my $ini_json = $json->decode($pd);
      my $type = $ini_json->{type};
      my $cur_file = "confdir/$type.ini";
      my $t = time();
      my $back_file = "/tmp/${type}_${t}.ini";
      copy($cur_file, $back_file);
      open SAVE, ">$cur_file" or $err = $!;
      if ($err) {
        $out->{status} = $err;
      } else {
        print SAVE $ini_json->{text};
        close SAVE;
        $out->{status} = 'Success! Please restart all scripts.';
        $conf = Config::Tiny->read("$confdir/other.ini");
        $ouf_conf = Config::Tiny->read("$confdir/ouf.ini");
      }
      print $json->encode($out);

    }
  }
} 

my $port = $conf->{api}->{port} || 8080;
my $pid = OUF::API::Server->new($port)->background();
print "Use 'kill $pid' to stop server.\n";