/*=============================================================================
  * Copyright (c) 2010 by OCLC
  *
  * File        : dl_debugtrace.h
  * Description : Main file for Debug/Trace utility.
  * Author(s)   : Onder Karpat
  * Creation    : 29-Jun-2010
  * 
  * Notes       : Contains all macros, configuration file processing,
  *               and the logic. 
  *               
  *               Implementation follows C style programming and language 
  *               constructs whenever possible.
  *
  *============================================================================
  */
#ifndef DL_DEBUGTRACE_H_INCLUDED
#define DL_DEBUGTRACE_H_INCLUDED

#pragma ident "$Id: dl_debugtrace.h 335828 2016-06-10 11:33:48Z heesakkm $"

/* Automatically set DL_PRESENT to 0 if not compiling for C++. */
#ifndef __cplusplus
    #undef DL_PRESENT
    #define DL_PRESENT 0
#else
    /* If C++ compilation, then enable DL_PRESENT by default. */
    #ifndef DL_PRESENT
    #define DL_PRESENT 1
    #endif
#endif

#if DL_PRESENT == 1
/* Include these codes only when Debug/Trace is present. */
extern "C++" {

#ifdef bool
#if bool != Cbool
#error When defined bool must be Cbool.
#endif
#define __bool_define_push
#undef bool
#endif
    
#include <time.h>
#include <stdarg.h>
#include <string.h>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <typeinfo>
#include <assert.h>

#ifdef __bool_define_push
#undef __bool_define_push
#define bool Cbool
#endif
    
/* Call dl_finalize when program exists unexpectedly, e.g. SIGSEGV.
 * When application normally exits, atexit calls dl_finalize, so
 * there is no need to add the dl_finalize call at the end of main. */
extern void dl_finalize();

#define DL_CFGMAXLINELENGTH 2048
#define DL_LOGFILEMAXLENGTH (DL_CFGMAXLINELENGTH + 0) /* Must be greater than DL_CFGMAXLINELENGTH */
#define DL_INDENTSIZE 3 /* Must be greater than 0 */
#define DL_MAXPARAMCOUNT 128

#define DL_SELFDEBUG 0 /* Set to 0 or 1. When 1 emits printfs for debugging dl_debugtrace.h */

#if DL_SELFDEBUG == 1
#define DL_DEBUG printf
#else
#define DL_DEBUG dl_ignore
/* dl_ignore is a catch all parameters function, that expands to
 * nothing when dl_debugtrace.h is not debugged. */
inline void dl_ignore (...) {}
#endif

typedef enum {
    DL_SS_INVALID   = -1,
    DL_SS_SHOW      = 1, 
    DL_SS_SKIP      = 2, 
    DL_SS_SKIPSUBS  = 3, 
    DL_SS_DEFAULT   = 4,
    DL_SS_LOGAT     = 5,
    DL_SS_END
} DL_STACKSTATE;

/* Pretty print DL_DTACKSTATE values. */
inline const char* 
toStr (DL_STACKSTATE ss) {
    switch (ss) {
    case DL_SS_SHOW: return "ss_Show";
    case DL_SS_SKIP: return "ss_Skip";
    case DL_SS_SKIPSUBS: return "ss_SkipSubs";
    case DL_SS_DEFAULT: return "ss_Default";
    case DL_SS_INVALID: return "ss_Invalid";
    case DL_SS_LOGAT: return "ss_LogAt";
    default: 
        break;
    } /* switch */
    return "<error>";
} /* toStr */

/* DebugTraceData holds all the information, including configuration
 * values, memory for log-lines.
 * 
 * Note: Access to this data with dl_getdata ().
 */
typedef struct tagDebugTraceData {
    /* Configuration variables */
    int maxLines;         /* Maximum total number of lines. */
    int maxLineLength;    /* Long lines are truncated to this length. */
    char** logLines;      /* Actual trace text is stored here. */
    char* outputFileName; /* This is where the outputs are stored. */
    char** skip;          /* skip, skipCount are used to skip logging of functions. */
    int skipCount;        /* Number of skip patterns specified in configuration file. */
    char** skipSubs;      /* skipSubs, and skipSubsCount are used to skip logging of */
    int skipSubsCount;    /*  calls within the specified functions. */
    char** show;          /* Show functions even if they happen to be in skip cases. */
    int showCount;        /* Number of show patterns specified in configuration file. */
    char** logAt;         /* Causes a log file write at the end of given function names. */
    int logAtCount;       /* Count of logAt entries. */
    int isDisabled;       /* Disables Debug/Trace, default is 0. */
    int maxCallDepth;     /* If call depth exceeds this amount Debug/Trace is disabled. */
    int consoleOutput;    /* If 1, lines are also output immediately to STDOUT. */
    int logFreqInSec;     /* If -1, disabled, otherwise causes log dump at given intervals. */
    clock_t logFreqLatest;/* Latest log time. */
    int changed;          /* When new log line comes in 'changed' is set to 1. */
    /* Variables used for book keeping */
    int isInitialized;
    int currLogLineIndex; /* Index to current log-line to be written. */
    DL_STACKSTATE* showSkipStateStack; /* Keep track of show state in this array, e.g. DL_SS_SHOW. */
    int showSkipStateStackIndex;
    int showSkipStateStackSize; /* Keep track of size, expands if necessary. */
    int indentLevel;      /* Keep track of function nesting levels. */
    char* paramTexts[DL_MAXPARAMCOUNT];   /* Temporary buffer to keep parameter formattings. */
    int showParamInfo;    /* When set to 1, parameter type and id is included in the logs. */
} DebugTraceData;

/*
   Configuration file syntax:
   The syntax is strict as it reduces parsing logic (it is easier to adapt to syntax than
   to implement flexible parsing logic).

   The file is line oriented (maximum line length can be DL_CFGMAXLINELENGTH). Lines are 
   independent from each other. Followings are case sensitive first tokens that are recognized. 
   Everything else is reported as warnings and the line is ignored:
   - LOGFILEPATH <full path and file name> (Do not use whitespace in <*>)
   - MAXLINES <10000 or more, maximum number of lines>
   - MAXLINELENGTH <1024 or more, maximum length of lines>
   - SKIP <weak wildcard expression> (Do not use whitespace in <*>)
   - SKIPSUBS <weak wildcard expression> (Do not use whitespace in <*>)
   - SHOW <weak wildcard expression> (Do not use whitespace in <*>)
   - # <any non empty string, interpret as comment and ignored without warning>
   - ENABLE <0 or 1> 
   - CONSOLEOUT <0 or 1>, default is 0.
   - MAXCALLDEPTH <1024 or more, when reached Debug/Trace disables itself>
   - LOGAT <funcname> 

   Note 1:
   <weak wildcard expression> can be one of these three:
   - *pattern, match only the right of function name
   - pattern*, match only the left of function name
   - *pattern*, match the center of function name

   Note 2:
   You can skip any configuration variable (then defaults are used):
   - LOGFILEPATH /tmp/log_dl_pid_xxx
   - MAXLINES 30000
   - MAXLINELENGTH 2048
   If you use them more than once, the last one is effective - no warning is displayed. 
   
   Note 3:
   You can use as many as SKIP, SHOW, SKIPSUBS lines.
   
   Note 4:
   When SKIP matches, the function and all of its sub calls are not
   included in log outputs.

   When SKIPSUBS matches, the function is logged (if it is not already SKIP-ed),
   but not the subcalls. SKIP supersedes SKIPSUBS.

   When SHOW matches, no matter if the current function is SKIP-ed it is 
   included in the log (but not the sub calls).
   
   Order of SHOW, SKIP, SKIPSUBS do not matter. In general there are more
   SKIPs than SHOWs, and SHOW is a more useful function, therefore SHOW 
   have precedence over SKIP, and SKIPSUBS.
   
   Consider this example:
    func0 {
	func1 {
	    func2 {
		func5 {
		}
	    }
	}
	func3 {
	    func4 {
	    }
	}
    }
    - SHOW func2     : this line would force func2 to be included.
    - SKIP func1     : this line would inhibit logs of func1, func2, and func5.
    - SKIPSUBS func3 : this line would inhibit logs of func4.
    The final result of this configuration is that func0, func2, and func3 is 
    included in the logs.
    
    Note 5:
    LOGAT <function name>, causes log dump of the trace at the end of given 
    function. The file name being created is the log file appended with 
    function name and hit count of the function. 
    For example, /tmp/log_dl_pid_###_fname_###.
 */

/*-----------------------------------------------------------------------------
  * Function      : dl_getdata
  * Purpose       : Returns the data associated with Debug/Trace.
  * Method        : This function has a static local variable which is in turn
  *                 used as global variable. This helps debugging the Debug
  *                 Trace itself, and makes it unnecessary to define the data
  *                 in a source file.
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : 
  *
  * Returns       : The data structure containing all the Debug/Trace data.
  *
  *-----------------------------------------------------------------------------
  */
inline DebugTraceData* dl_getdata () 
{
    static DebugTraceData debugTraceData;
    return &debugTraceData;
} /* dl_getdata */

/*-----------------------------------------------------------------------------
  * Function      : dl_initialize
  * Purpose       : Initializes the Debug/Trace data by reading the 
  *                 configuration file.
  * Method        : This function is called when dl_getdata()->isInitialized
  *                 is 0. dl_begin, dl_end checks for this condition, so no
  *                 need to call dl_initialize in the main of an application.
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : Sets default values if configuration file is not set in 
  *                 the environment.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void 
dl_initialize () 
{
    DL_DEBUG ("dl_initialize pid=%d\n", getpid ());
    atexit (dl_finalize);
    dl_getdata ()->isInitialized = 1;
    /* Assume default values */
    dl_getdata ()->maxLines = 30000;
    dl_getdata ()->maxLineLength = 1024;
    dl_getdata ()->isDisabled = 0;
    dl_getdata ()->logFreqInSec = -1;
    dl_getdata ()->logFreqLatest = 0;
    dl_getdata ()->changed = 1;
    dl_getdata ()->showParamInfo = 0;
    dl_getdata ()->consoleOutput = 0;
    dl_getdata ()->maxCallDepth = 1024;
    { /* Read configuration file from DL_CONFIG_FILE environment variable. */
	char* configFileName = NULL;
	configFileName = getenv ("DL_CONFIG_FILE");
	if (configFileName != NULL) {
	    printf ("Info(%d): Debug/Trace configuration file is $DL_CONFIG_FILE=%s\n", getpid (), configFileName);
	    FILE* configFile = NULL;
	    configFile = fopen (configFileName, "r");
	    if (configFile == NULL) {
                printf ("Warning(%d): Debug/Trace configuration file %s "
                        "is not found, using defaults.\n", getpid (), configFileName);
            } else {
		char buf[DL_CFGMAXLINELENGTH];
		while (fgets (buf, sizeof buf, configFile)) {
                    if (buf[0] == '#') {
                        /* Comment line just ignore */
                    } else if (!strchr(buf, ' ')) {
                        printf ("Warning(%d): Debug/Trace configuration file line ignored: %s", getpid (), buf);
                    } else {
			char key[32] = {0};
			char value[DL_CFGMAXLINELENGTH] = {0};
			char unexpected[16] = {0};
			char formatStr[128];
			sprintf (formatStr, "%%32s %%%ds %%16s", DL_CFGMAXLINELENGTH);
			sscanf (buf, formatStr, key, value, unexpected);
			if (strlen (value) == 0 || strlen(unexpected) != 0) {
			    printf ("Warning(%d): Debug/Trace configuration file line ignored: %s", getpid (), buf);
			} else {
			    if (strcmp (key, "LOGFILEPATH") == 0) {
				dl_getdata ()->outputFileName = strcpy ((char*)malloc(strlen(value)+1), value);
			    } else if (strcmp (key, "MAXLINES") == 0) {
				/* Set a reasonable minimum amount */
				dl_getdata ()->maxLines = (atoi (value) < 10000 ? 10000 : atoi (value));
			    } else if (strcmp (key, "MAXLINELENGTH") == 0) {
				/* Set a reasonable minimum amount */
				dl_getdata ()->maxLineLength = (atoi (value) < 1024 ? 1024 : atoi (value));
			    } else if (strcmp (key, "SKIP") == 0) {
				dl_getdata ()->skipCount ++;
				dl_getdata ()->skip = (char**)realloc (dl_getdata ()->skip, dl_getdata ()->skipCount * sizeof (char*));
				dl_getdata ()->skip [dl_getdata ()->skipCount-1] = strcpy ((char*)malloc (strlen (value)+1), value);
                                DL_DEBUG ("SKIP %s\n", dl_getdata ()->skip [dl_getdata()->skipCount-1]);
			    } else if (strcmp (key, "LOGAT") == 0) {
				dl_getdata ()->logAtCount ++;
				dl_getdata ()->logAt = (char**)realloc (dl_getdata ()->logAt, dl_getdata ()->logAtCount * sizeof (char*));
				dl_getdata ()->logAt [dl_getdata()->logAtCount-1] = strcpy ((char*)malloc (strlen (value)+1), value);
			    } else if (strcmp (key, "SKIPSUBS") == 0) {
				dl_getdata ()->skipSubsCount ++;
				dl_getdata ()->skipSubs = (char**)realloc (dl_getdata ()->skipSubs, dl_getdata ()->skipSubsCount * sizeof (char*));
				dl_getdata ()->skipSubs [dl_getdata()->skipSubsCount-1] = strcpy ((char*)malloc (strlen (value)+1), value);
			    } else if (strcmp (key, "SHOW") == 0) {
				dl_getdata ()->showCount ++;
				dl_getdata ()->show = (char**)realloc (dl_getdata()->show, dl_getdata()->showCount * sizeof (char*));
				dl_getdata ()->show [dl_getdata()->showCount-1] = strcpy ((char*)malloc(strlen(value)+1), value);
			    } else if (strcmp (key, "LOGFREQINSEC") == 0) {
				dl_getdata ()->logFreqInSec = atoi (value);
			    } else if (strcmp (key, "SHOWPARAMINFO") == 0) {
				dl_getdata ()->showParamInfo = atoi (value);
			    } else if (strcmp (key, "CONSOLEOUT") == 0) {
				dl_getdata ()->consoleOutput = atoi (value) == 1;
			    } else if (strcmp (key, "ENABLE") == 0) {
				dl_getdata ()->isDisabled = atoi (value) == 0;
			    } else if (strcmp (key, "MAXCALLDEPTH") == 0) {
				dl_getdata ()->maxCallDepth = atoi (value) < 1024 ? 1024 : atoi (value);
			    } else if (strcmp (key, "#") == 0) {
				/* This is a comment line, just ignore. */
			    } else {
				printf ("Warning(%d): Debug/Trace configuration file line ignored: %s", getpid (), buf);
			    } /* if */
			} /* if */
		    } /* if */
		} /* while */
		fclose (configFile);
	    } /* if */
	} /* if */
        if (dl_getdata()->isDisabled) {
            printf ("Warning(%d): Debug/Trace disabled by configuration.\n", getpid ());
        } /* if */
	dl_getdata ()->logLines = (char**)calloc (dl_getdata ()->maxLines, sizeof (char*));
        for (int c = 0; c < dl_getdata ()->maxLines; c ++) {
            dl_getdata ()->logLines [c] = (char*)calloc (dl_getdata ()->maxLineLength, sizeof (char));
        } /* for */
        for (int d = 0; d < DL_MAXPARAMCOUNT; d ++) {
            dl_getdata ()->paramTexts [d] = (char*)calloc (dl_getdata ()->maxLineLength, sizeof (char));
        } /* for */
    } /* block */
    return;
} /* dl_initialize */

/*-----------------------------------------------------------------------------
  * Function      : dl_incLogLineIndex
  * Purpose       : Increments and wraps if necessary, the log line index.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 17-Nov-2010  : Creation
  *
  * Notes         : None.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void
dl_incLogLineIndex () {
    dl_getdata ()->currLogLineIndex ++;
    if (dl_getdata ()->currLogLineIndex >= dl_getdata ()->maxLines) {
	dl_getdata ()->currLogLineIndex = 0;
    } /* if */
} /* dl_incLogLineIndex */

/*-----------------------------------------------------------------------------
  * Function      : dl_logAt
  * Purpose       : Writes out the logs to output file. 
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : If output file name is not specified, logs are written to 
  *                 /tmp/log_dl_pid_#pid. If funcName is not NULL it is 
  *                 appended to the log output file name.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void
dl_logAt (const char* funcName) 
{
    FILE* logFile = NULL;
    char logOutputFileName[DL_LOGFILEMAXLENGTH] = {0};
    if (!dl_getdata ()->isInitialized) {
	return;
    } /* if */
    if (dl_getdata()->isDisabled) {
	return;
    } /* if */
    if (dl_getdata()->outputFileName == NULL) {
	sprintf (logOutputFileName, "/tmp/log_dl_pid_%d", getpid ());
    } else {
	sprintf (logOutputFileName, "%s.%d", dl_getdata()->outputFileName, getpid ());
    } /* if */
    if (funcName != NULL) {
	strncat (logOutputFileName, "_", DL_LOGFILEMAXLENGTH-strlen (logOutputFileName));
	strncat (logOutputFileName, funcName, DL_LOGFILEMAXLENGTH-strlen (logOutputFileName));
    } /* if */
    logFile = fopen (logOutputFileName, "w");
    if (logFile != NULL) {
	fprintf (logFile, "Debug/Trace dump started for pid=%d, logFile=%s\n", getpid (), logOutputFileName);
	if (strlen(dl_getdata ()->logLines[dl_getdata ()->currLogLineIndex]) != 0) {
	    /* Buffer limit has been exceeded first write oldest entries. */
	    for (int c = dl_getdata ()->currLogLineIndex; c < dl_getdata ()->maxLines; c ++) {
		fprintf (logFile, "%s\n", dl_getdata ()->logLines [c]);
	    } /* for */
	} /* if */
	for (int c = 0; c < dl_getdata ()->currLogLineIndex; c ++) {
	    fprintf (logFile, "%s\n", dl_getdata ()->logLines [c]);
	} /* for */
	fprintf (logFile, "Debug/Trace(%d) dump finished.\n", getpid ());
	printf ("Debug/Trace(%d) is stored in %s.\n", getpid (), logOutputFileName);
	fclose (logFile);
    } else {
	printf ("Error(%d): could not open Debug/Trace file %s for writing.\n", getpid (), logOutputFileName);
    } /* if */
    return;
} /* dl_logAt */

/*-----------------------------------------------------------------------------
  * Function      : dl_finalize
  * Purpose       : Finalizes the Debug/Trace, writes the logs to output file.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : This function must be called within the containing 
  *		    application logic if atexit is bypassed, e.g. with SIGABRT.
  *		
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void
dl_finalize() 
{
    if (!dl_getdata ()->isInitialized) {
	return;
    } /* if */
    if (dl_getdata()->isDisabled) {
	return;
    } /* if */
    dl_logAt (NULL);
    printf ("Debug/Trace(%d) finalized and disabled.\n", getpid ());
 //   dl_getdata()->isDisabled = 1;
    return;
} /* dl_finalize */

typedef struct {
    int result; /* Can be 0 or 1. */
    const char* funcName;
} DL_FUNCMATCHRESULT;

/*-----------------------------------------------------------------------------
  * Function      : dl_compareMatchResults
  * Purpose       : Compare function that is used in bsearch and qsort of 
  *                 DL_FUNCMATCHRESULT.
  * Method        : Just compare the function name.
  * Author        : Onder Karpat
  *
  * History       :
  * 30-June-2010  : Creation
  *
  * Notes         : Nothing
  *
  * Returns       : -1, 0, or 1.
  *
  *-----------------------------------------------------------------------------
  */
inline int
dl_compareMatchResults (DL_FUNCMATCHRESULT* rec1, DL_FUNCMATCHRESULT* rec2) 
{
    assert (rec1->funcName);
    assert (rec2->funcName);
    return strcmp (rec1->funcName, rec2->funcName);
} /* dl_compareMatchResults */

/*-----------------------------------------------------------------------------
  * Function      : dl_match
  * Purpose       : Checks the function name against configuration file show/
  *                 skip lists. If function name matches to the given list
  *                 it returns 1.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 30-June-2010  : Creation
  *
  * Notes         : Nothing
  *
  * Returns       : 0 or 1.
  *
  *-----------------------------------------------------------------------------
  */
inline int
dl_match (const char* funcName, DL_STACKSTATE ss, char** patternList, int patternListCount) 
{
    DL_DEBUG ("dl_match (funcName=%s, state=%s, patListCount=%d)\n", funcName, toStr (ss), patternListCount);
    static DL_FUNCMATCHRESULT* prevResults [DL_SS_END];
    static int prevResultsCount [DL_SS_END];
    static int prevResultsSize [DL_SS_END];
    DL_FUNCMATCHRESULT search = {0, funcName};
    DL_FUNCMATCHRESULT* found = NULL;
    /* TODO: OKar- disabled due to a bug in caching logic. */
    if (0) {
	if ((found = (DL_FUNCMATCHRESULT*)bsearch ((void*)&search, (void*)prevResults [ss], 
			prevResultsCount [ss], sizeof (DL_FUNCMATCHRESULT),
			(int (*)(const void*, const void*))dl_compareMatchResults))) {
	    DL_DEBUG ("   dl_match: %s in history with result %d\n", funcName, found->result);
	    return found->result; // There is already a recorded result, so use it.
	}
    }
    {
	int match = 0;
	for (int c = 0; c < patternListCount; c ++) {
	    char* currPattern = patternList[c];
	    int starBeg = 0, starEnd = 0;
	    starBeg = currPattern[0] == '*';
	    starEnd = currPattern[strlen(currPattern)-1] == '*';
	    DL_DEBUG ("    dl_match: func=%s, pat=%s, *x=%d, x*=%d\n", funcName, currPattern, starBeg, starEnd); 
	    if (strcmp(currPattern, "*") == 0) {
		match = 1; /* Special case '*' matches all. */
		break;
	    } /* if */
	    if (!starBeg && !starEnd) {
		if (strcmp (currPattern, funcName) == 0) {
		    match = 1; /* Exact match without wildcard. */
		    break;
		} /* if */
	    } else if (starBeg && !starEnd) {
		char* found = strstr (funcName, currPattern+1);
		if (found && (strlen (found) == strlen(currPattern+1))) {
		    match = 1;
		    break;
		} /* if */
	    } else if (starEnd && !starBeg) {
		char buf[DL_CFGMAXLINELENGTH];
		strncpy (buf, currPattern, strlen (currPattern) - 1);
		if (strstr (funcName, buf) == funcName) {
		    match = 1;
		    break;
		} /* if */
	    } else { /* starEnd && starBeg */
		char buf[DL_CFGMAXLINELENGTH];
		strncpy (buf, currPattern + 1, strlen (currPattern) - 2);
		if (strstr (funcName, buf) != NULL) {
		    match = 1;
		    break;
		} /* if */
	    } /* if */
	} /* for */
	/* TODO: OKar- There is a bug in the caching logic, disabled for now. */
	if (0) {
	    /* Store the result in the history. */
	    if (prevResultsSize [ss] <= prevResultsCount [ss]) {
		prevResultsSize [ss] += 1024;
		prevResults [ss] = (DL_FUNCMATCHRESULT*) realloc (prevResults [ss], prevResultsSize [ss] * sizeof (DL_FUNCMATCHRESULT));
	    } /* if */
	    prevResults [ss][prevResultsCount [ss]].result = match;
	    prevResults [ss][prevResultsCount [ss]].funcName = strcpy ((char*)malloc (strlen(funcName)+1), funcName);
	    prevResultsCount [ss] ++;
	    DL_DEBUG ("    dl_match: added func=%s, res=%d, new history size=%d\n", funcName, match, prevResultsCount [ss]);
	    /* We need to sort the result for bsearch. */
	    qsort (prevResults [ss], prevResultsCount [ss], sizeof (DL_FUNCMATCHRESULT), 
		    (int (*)(const void*, const void*))dl_compareMatchResults);
	}
	return match;
    } /* if */
    return 0; /* Program execution should actually never come here. */
} /* dl_match */

/*-----------------------------------------------------------------------------
  * Function      : dl_calcCurrSSState
  * Purpose       : By looking at the show skip stack function states, a
  *                 final show/skip decision is made.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 30-June-2010  : Creation
  *
  * Notes         : Nothing
  *
  * Returns       : DL_SS_SHOW, or DL_SS_SKIP
  *
  *-----------------------------------------------------------------------------
  */
inline DL_STACKSTATE 
dl_calcCurrSSState () 
{
    DL_DEBUG ("dl_calcCurrSSState ()\n"); 
    if (dl_getdata ()->showSkipStateStackIndex == 0) {
	return DL_SS_SHOW;
    } /* if */
    DL_STACKSTATE currVal = dl_getdata ()->showSkipStateStack [dl_getdata ()->showSkipStateStackIndex-1];
    if (dl_getdata ()->showSkipStateStackIndex == 1) {
	return currVal;
    } /* if */
    DL_STACKSTATE prevVal = dl_getdata ()->showSkipStateStack [dl_getdata ()->showSkipStateStackIndex-2];
    if (currVal == DL_SS_SHOW || currVal == DL_SS_SKIP) {
	return currVal;
    } /* if */
    return prevVal == DL_SS_SHOW ? DL_SS_SHOW : DL_SS_SKIP;
} /* dl_calcCurrSSState */

/*-----------------------------------------------------------------------------
  * Function      : dl_begin
  * Purpose       : Main entry for DL_BEGIN macro implementation. Logs a
  *                 function begin, with parameters specified in variable
  *                 arguments.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : None.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void 
dl_begin (const char* funcName, const char* fileName, int lineNumber, int numArgs, ...) 
{
    DL_DEBUG ("dl_begin (func=%s, file=%s, line=%d, numargs=%d) \n", funcName, fileName, lineNumber, numArgs);
    if (!dl_getdata ()->isInitialized) { 
	dl_initialize();
    } /* if */
    if (dl_getdata ()->isDisabled) {
	return;
    } /* if */
    { /* Adjust stack size if needed */
	if (dl_getdata ()->showSkipStateStackSize <= dl_getdata ()->showSkipStateStackIndex) {
	    dl_getdata ()->showSkipStateStackSize += 2048;
	    dl_getdata ()->showSkipStateStack = (DL_STACKSTATE*)realloc(dl_getdata ()->showSkipStateStack, 
		    dl_getdata ()->showSkipStateStackSize);
	} /* if */
	if (dl_getdata ()->showSkipStateStackIndex >= dl_getdata ()->maxCallDepth) {
	    dl_finalize ();
	} /* if */
    } /* block */
    { /* Check the show/skip states */
	DL_STACKSTATE currFuncState = DL_SS_INVALID; 
	if (dl_match (funcName, DL_SS_SKIP, dl_getdata ()->skip, dl_getdata ()->skipCount)) {
	    currFuncState = DL_SS_SKIP; 
	} else if (dl_match (funcName, DL_SS_SKIPSUBS, dl_getdata ()->skipSubs, dl_getdata ()->skipSubsCount)) {
	    currFuncState = DL_SS_SKIPSUBS; 
	} else {
	    currFuncState = DL_SS_SHOW;
	} /* if */
	if (dl_match (funcName, DL_SS_SHOW, dl_getdata ()->show, dl_getdata ()->showCount)) {
	    currFuncState = DL_SS_SHOW;
	} /* if */
        DL_DEBUG ("    dl_begin: %s func state %s\n", funcName, toStr (currFuncState));
        DL_DEBUG ("    dl_begin: %s curr state %s\n", funcName, toStr (dl_calcCurrSSState ()));
	dl_getdata ()->showSkipStateStack [dl_getdata ()->showSkipStateStackIndex] = currFuncState;
	dl_getdata ()->showSkipStateStackIndex ++;
	if (dl_calcCurrSSState () != DL_SS_SHOW) {
            DL_DEBUG ("    dl_begin skipped %s\n", funcName);
        } else {
	    static char* lineBuf;
	    if (!lineBuf) {
		lineBuf = (char*)malloc (dl_getdata ()->maxLineLength);
	    } /* if */
	    snprintf (lineBuf, dl_getdata()->maxLineLength, "%s (", funcName);
	    { /* process arguments */
		va_list ap;
		va_start (ap, numArgs);
		for (int c = 0; c < numArgs; c ++) {
		    char* paramString = va_arg (ap, char*);
		    if (c != 0) {
			strncat (lineBuf, ", ", dl_getdata ()->maxLineLength - strlen (lineBuf));
		    } /* if */
		    strncat (lineBuf, paramString, dl_getdata ()->maxLineLength - strlen (lineBuf));
		} /* for */
		strncat (lineBuf, ") {", dl_getdata ()->maxLineLength - strlen (lineBuf));
		va_end (ap);
	    } /* block */
	    memset (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex], ' ',
		    dl_getdata()->maxLineLength);
	    strncpy (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] + dl_getdata ()->indentLevel *
		    DL_INDENTSIZE, lineBuf, dl_getdata ()->maxLineLength);
	    if (dl_getdata ()->consoleOutput == 1) {
		printf ("%s\n", dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] +
			dl_getdata ()->indentLevel * DL_INDENTSIZE);
	    } /* if */
	    dl_getdata ()->indentLevel ++;
	    dl_getdata ()->changed = 1;
	    dl_incLogLineIndex ();
	} /* if */
    } /* block */
    if (dl_match (funcName, DL_SS_LOGAT, dl_getdata()->logAt, dl_getdata()->logAtCount)) {
	dl_logAt (funcName);
    } /* if */
    return;
} /* dl_begin */

/*-----------------------------------------------------------------------------
  * Function      : dl_print
  * Purpose       : Main entry for DL_PRINT macro implementation. Logs 
  *                 given arguments. 
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 23-Nov-2010   : Creation
  *
  * Notes         : Current function must be in show state.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void 
dl_print (const char* funcName, const char* fileName, int lineNumber, int numArgs, ...) 
{
    DL_DEBUG ("dl_print (func=%s, file=%s, line=%d, numargs=%d) \n", funcName, fileName, lineNumber, numArgs);
    if (!dl_getdata ()->isInitialized) { 
	dl_initialize();
    } /* if */
    if (dl_getdata ()->isDisabled) {
	return;
    } /* if */
    if (dl_calcCurrSSState () != DL_SS_SHOW) {
	DL_DEBUG ("    dl_print skipped %s\n", funcName);
    } else {
	static char* lineBuf;
	if (!lineBuf) {
	    lineBuf = (char*)malloc (dl_getdata ()->maxLineLength);
	} /* if */
	{ /* process arguments */
	    va_list ap;
	    va_start (ap, numArgs);
	    for (int c = 0; c < numArgs; c ++) {
		char* paramString = va_arg (ap, char*);
		strncat (lineBuf, paramString, dl_getdata ()->maxLineLength - strlen (lineBuf));
	    } /* for */
	    va_end (ap);
	} /* block */
	memset (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex], ' ',
		dl_getdata()->maxLineLength);
	strncpy (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] + dl_getdata ()->indentLevel *
		DL_INDENTSIZE, lineBuf, dl_getdata ()->maxLineLength);
	if (dl_getdata ()->consoleOutput == 1) {
	    printf ("%s\n", dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] +
		    dl_getdata ()->indentLevel * DL_INDENTSIZE);
	} /* if */
	dl_getdata ()->changed = 1;
	dl_incLogLineIndex ();
    } /* if */
    if (dl_match (funcName, DL_SS_LOGAT, dl_getdata()->logAt, dl_getdata()->logAtCount)) {
	dl_logAt (funcName);
    } /* if */
    return;
} /* dl_print */

/*-----------------------------------------------------------------------------
  * Function      : dl_end
  * Purpose       : Emits log output for function end.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : None.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void 
dl_end (const char* funcName, const char* fileName, int lineNumber, int numArgs, 
	const char* param0, va_list ap) 
{
    if (!dl_getdata ()->isInitialized) { 
	dl_initialize();
    } /* if */
    if (dl_getdata ()->isDisabled) {
	return;
    } /* if */
    dl_getdata ()->changed = 1;
    if (dl_calcCurrSSState () == DL_SS_SHOW) {
	static char* lineBuf;
	if (!lineBuf) {
	    lineBuf = (char*)malloc (dl_getdata ()->maxLineLength);
	} /* if */
	snprintf (lineBuf, dl_getdata ()->maxLineLength, "return %s", param0);
	{ /* process arguments */
	    for (int c = 0; c < numArgs-1; c ++) {
		char* paramString = va_arg (ap, char*);
                strncat (lineBuf, ", ", dl_getdata ()->maxLineLength - strlen (lineBuf));
		strncat (lineBuf, paramString, dl_getdata ()->maxLineLength - strlen (lineBuf));
	    } /* for */
	} /* block */
	strncat (lineBuf, ";", dl_getdata ()->maxLineLength - strlen (lineBuf));
	memset (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex], ' ',
		dl_getdata()->maxLineLength);
	strncpy (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] + dl_getdata ()->indentLevel *
		DL_INDENTSIZE, lineBuf, dl_getdata ()->maxLineLength);
	if (dl_getdata ()->consoleOutput == 1) {
	    printf ("%s\n", dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] +
			dl_getdata ()->indentLevel * DL_INDENTSIZE);
	} /* if */
	dl_incLogLineIndex ();

	memset (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex], ' ',
		dl_getdata()->maxLineLength);
	dl_getdata ()->indentLevel --;
	snprintf (lineBuf, dl_getdata ()->maxLineLength, "} // %s", funcName);
	strncpy (dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] + dl_getdata ()->indentLevel *
		DL_INDENTSIZE, lineBuf, dl_getdata ()->maxLineLength);
	if (dl_getdata ()->consoleOutput == 1) {
	    printf ("%s\n", dl_getdata ()->logLines [dl_getdata ()->currLogLineIndex] +
			dl_getdata ()->indentLevel * DL_INDENTSIZE);
	} /* if */
	dl_incLogLineIndex ();
    } /* if */
    if (dl_getdata ()->logFreqInSec != -1) {
	if (dl_getdata ()->changed != 0) {
	    if (clock () - dl_getdata ()->logFreqLatest > dl_getdata ()->logFreqInSec * CLOCKS_PER_SEC) {
		dl_logAt (NULL);
		dl_getdata ()->logFreqLatest = clock ();
		dl_getdata ()->changed = 0;
	    } /* if */
	} /* if */
    } /* if */ 
    if (dl_match (funcName, DL_SS_LOGAT, dl_getdata()->logAt, dl_getdata()->logAtCount)) {
	dl_logAt (funcName);
    } /* if */
    if (dl_getdata ()->showSkipStateStackIndex == 0) {
	printf ("DL_END stack underflow(%d), Debug/Trace disabled.", getpid ());
	dl_finalize ();
    } else {
	dl_getdata ()->showSkipStateStackIndex --;
    } /* if */
    return;
} /* dl_end */

/*-----------------------------------------------------------------------------
  * Function      : dl_end_t
  * Purpose       : Helper function that returns the supplied argument
  *                 with original type.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : None.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
template<typename T>
inline T 
dl_end_t (const char* retText, T retVal, const char* funcName, 
	  const char* fileName, int line, int numArgs, ...) 
{
    va_list ap;
    va_start (ap, numArgs);
    dl_end (funcName, fileName, line, numArgs, dl_paramToText (127, retText, retVal), ap);
    va_end (ap);
    return retVal;
} /* dl_end_t */

/*-----------------------------------------------------------------------------
  * Function      : dl_end_v
  * Purpose       : Helper function that returns void, and records the
  *                 function end in the Debug/Trace log.
  * Method        : 
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : None.
  *
  * Returns       : Nothing.
  *
  *-----------------------------------------------------------------------------
  */
inline void
dl_end_v (const char* funcName, const char* fileName, int line, int numArgs, ...) 
{
    va_list ap;
    va_start (ap, numArgs);
    dl_end (funcName, fileName, line, numArgs, "", ap);
    va_end (ap);
} /* dl_end_v */

/*-----------------------------------------------------------------------------
  * Function      : dl_paramToText
  * Purpose       : Template function that converts given variables in
  *                 to strings.
  * Method        : Template specializations helps knowing the exact
  *                 type, thus variable can be printed with type information.
  * Author        : Onder Karpat
  *
  * History       :
  * 29-June-2010  : Creation
  *
  * Notes         : It is possible to add your own struct types in to
  *                 this mechanism, so you get more information than just
  *                 pointer values.
  *
  * Returns       : Value converted in to string.
  *
  *-----------------------------------------------------------------------------
  */
template<typename T>
inline char* 
dl_paramToText (int paramIdx, const char* pText, T aParam) 
{
    if (!dl_getdata ()->isInitialized) {
	dl_initialize ();
    } /* if */
    int maxLineLen = dl_getdata()->maxLineLength;
    if (dl_getdata()->showParamInfo) {
        snprintf (dl_getdata()->paramTexts[paramIdx], maxLineLen, "%s %s=%p", typeid (aParam).name (), pText, &aParam); 
    } else {
        snprintf (dl_getdata()->paramTexts[paramIdx], maxLineLen, "%p", &aParam); 
    } /* if */
    return dl_getdata()->paramTexts[paramIdx];
} /* dl_paramToText */

#define DL_DEFINE_PARAMTOTEXT_FUNC(type, formatStr)                                     \
template<>                                                                              \
inline char* dl_paramToText <type>(int paramIdx, const char* pText, type aParam) {      \
    if (!dl_getdata ()->isInitialized) {						\
	dl_initialize ();								\
    } /* if */										\
    int maxLineLen = dl_getdata()->maxLineLength;					\
    if (dl_getdata()->showParamInfo) {							\
      snprintf (dl_getdata ()->paramTexts [paramIdx], maxLineLen, "%s %s=" formatStr,	\
  		typeid (aParam).name (), pText, aParam);				\
    } else {										\
      snprintf (dl_getdata ()->paramTexts [paramIdx], maxLineLen, formatStr, aParam);	\
    } /* if */										\
    return dl_getdata ()->paramTexts [paramIdx];                                        \
} /* dl_paramToText */                                                                  \
template<>                                                                              \
inline char* dl_paramToText <const type>(int paramIdx, const char* pText,		\
	const type aParam) {								\
    if (!dl_getdata ()->isInitialized) {						\
	dl_initialize ();								\
    } /* if */										\
    int maxLineLen = dl_getdata()->maxLineLength;					\
    if (dl_getdata()->showParamInfo) {							\
        snprintf (dl_getdata ()->paramTexts [paramIdx], maxLineLen, "%s %s=" formatStr,	\
	    typeid (aParam).name (), pText, aParam);					\
    } else {										\
      snprintf (dl_getdata ()->paramTexts [paramIdx], maxLineLen, formatStr, aParam);	\
    } /* if */										\
    return dl_getdata ()->paramTexts [paramIdx];                                        \
} /* dl_paramToText */

DL_DEFINE_PARAMTOTEXT_FUNC(int, "%d")
DL_DEFINE_PARAMTOTEXT_FUNC(char, "%c")
DL_DEFINE_PARAMTOTEXT_FUNC(double, "%f")
DL_DEFINE_PARAMTOTEXT_FUNC(float, "%f")
DL_DEFINE_PARAMTOTEXT_FUNC(size_t, "%d")
DL_DEFINE_PARAMTOTEXT_FUNC(char*, "\"%s\"")

#undef DL_CFGMAXLINELENGTH
#undef DL_LOGFILEMAXLENGTH
#undef DL_DEFINE_PARAMTOTEXT_FUNC
#undef DL_INDENTSIZE
#undef DL_MAXPARAMCOUNT
} /* extern "C++" */
#endif

/* When DL_PRESENT is 0, all macros removes Debug/Trace codes. */
#if DL_PRESENT == 1

#define DL_BEGIN() \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 0)
#define DL_BEGIN1(p0) \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 1, dl_paramToText (0, #p0, p0))
#define DL_BEGIN2(p0, p1) \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 2, dl_paramToText (0, #p0, p0), \
						  dl_paramToText (1, #p1, p1))
#define DL_BEGIN3(p0, p1, p2) \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 3, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2))
#define DL_BEGIN4(p0, p1, p2, p3) \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 4, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3))
#define DL_BEGIN5(p0, p1, p2, p3, p4)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 5, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4))
#define DL_BEGIN6(p0, p1, p2, p3, p4, p5)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 6, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5))
#define DL_BEGIN7(p0, p1, p2, p3, p4, p5, p6)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 7, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6))
#define DL_BEGIN8(p0, p1, p2, p3, p4, p5, p6, p7)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 8, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7))
#define DL_BEGIN9(p0, p1, p2, p3, p4, p5, p6, p7, p8)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 9, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8))
#define DL_BEGIN10(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 10, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9))
#define DL_BEGIN11(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 11, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10))
#define DL_BEGIN12(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 12, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11))
#define DL_BEGIN13(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 13, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12))
#define DL_BEGIN14(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 14, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12), \
	                                          dl_paramToText (13, #p13, p13))
#define DL_BEGIN15(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 15, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12), \
	                                          dl_paramToText (13, #p13, p13), \
	                                          dl_paramToText (14, #p14, p14))
#define DL_BEGIN16(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15)   \
    dl_begin(__FUNCTION__, __FILE__, __LINE__, 16, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12), \
	                                          dl_paramToText (13, #p13, p13), \
	                                          dl_paramToText (14, #p14, p14), \
	                                          dl_paramToText (15, #p15, p15))

#define DL_PRINT1(p0) \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 1, dl_paramToText (0, #p0, p0))
#define DL_PRINT2(p0, p1) \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 2, dl_paramToText (0, #p0, p0), \
						  dl_paramToText (1, #p1, p1))
#define DL_PRINT3(p0, p1, p2) \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 3, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2))
#define DL_PRINT4(p0, p1, p2, p3) \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 4, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3))
#define DL_PRINT5(p0, p1, p2, p3, p4)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 5, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4))
#define DL_PRINT6(p0, p1, p2, p3, p4, p5)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 6, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5))
#define DL_PRINT7(p0, p1, p2, p3, p4, p5, p6)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 7, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6))
#define DL_PRINT8(p0, p1, p2, p3, p4, p5, p6, p7)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 8, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7))
#define DL_PRINT9(p0, p1, p2, p3, p4, p5, p6, p7, p8)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 9, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8))
#define DL_PRINT10(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 10, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9))
#define DL_PRINT11(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 11, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10))
#define DL_PRINT12(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 12, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11))
#define DL_PRINT13(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 13, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12))
#define DL_PRINT14(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 14, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12), \
	                                          dl_paramToText (13, #p13, p13))
#define DL_PRINT15(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 15, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12), \
	                                          dl_paramToText (13, #p13, p13), \
	                                          dl_paramToText (14, #p14, p14))
#define DL_PRINT16(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15)   \
    dl_print(__FUNCTION__, __FILE__, __LINE__, 16, dl_paramToText (0, #p0, p0), \
	                                          dl_paramToText (1, #p1, p1), \
	                                          dl_paramToText (2, #p2, p2), \
	                                          dl_paramToText (3, #p3, p3), \
	                                          dl_paramToText (4, #p4, p4), \
	                                          dl_paramToText (5, #p5, p5), \
	                                          dl_paramToText (6, #p6, p6), \
	                                          dl_paramToText (7, #p7, p7), \
	                                          dl_paramToText (8, #p8, p8), \
	                                          dl_paramToText (9, #p9, p9), \
	                                          dl_paramToText (10, #p10, p10), \
	                                          dl_paramToText (11, #p11, p11), \
	                                          dl_paramToText (12, #p12, p12), \
	                                          dl_paramToText (13, #p13, p13), \
	                                          dl_paramToText (14, #p14, p14), \
	                                          dl_paramToText (15, #p15, p15))

#define DL_END() \
    dl_end_v( __FUNCTION__, __FILE__, __LINE__, 0)
#define DL_END1(p0) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 1)
#define DL_END2(p0, p1) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 2, \
    dl_paramToText (0, #p1, p1))
#define DL_END3(p0, p1, p2) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 3, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2))
#define DL_END4(p0, p1, p2, p3) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 4, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3))
#define DL_END5(p0, p1, p2, p3, p4) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 5, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4))
#define DL_END6(p0, p1, p2, p3, p4, p5) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 6, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5))
#define DL_END7(p0, p1, p2, p3, p4, p5, p6) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 7, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6))
#define DL_END8(p0, p1, p2, p3, p4, p5, p6, p7) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 8, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6),\
    dl_paramToText (6, #p7, p7))
#define DL_END9(p0, p1, p2, p3, p4, p5, p6, p7, p8) \
    dl_end_t(#p0, p0, __FUNCTION__, __FILE__, __LINE__, 9, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6),\
    dl_paramToText (6, #p7, p7),\
    dl_paramToText (7, #p8, p8))


#define DL_ENDV1(p1) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 2, \
    dl_paramToText (0, #p1, p1))
#define DL_ENDV2(p1, p2) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 3, \
    dl_paramToText (0, #p1, p1), \
    dl_paramToText (1, #p2, p2))
#define DL_ENDV3(p1, p2, p3) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 4, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3))
#define DL_ENDV4(p1, p2, p3, p4) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 5, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4))
#define DL_ENDV5(p1, p2, p3, p4, p5) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 6, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5))
#define DL_ENDV6(p1, p2, p3, p4, p5, p6) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 7, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6))
#define DL_ENDV7(p1, p2, p3, p4, p5, p6, p7) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 8, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6),\
    dl_paramToText (6, #p7, p7))
#define DL_ENDV8(p1, p2, p3, p4, p5, p6, p7, p8) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 9, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6),\
    dl_paramToText (6, #p7, p7),\
    dl_paramToText (7, #p8, p8))
#define DL_ENDV9(p1, p2, p3, p4, p5, p6, p7, p8, p9) \
    dl_end_v(__FUNCTION__, __FILE__, __LINE__, 10, \
    dl_paramToText (0, #p1, p1),\
    dl_paramToText (1, #p2, p2),\
    dl_paramToText (2, #p3, p3),\
    dl_paramToText (3, #p4, p4),\
    dl_paramToText (4, #p5, p5),\
    dl_paramToText (5, #p6, p6),\
    dl_paramToText (6, #p7, p7),\
    dl_paramToText (7, #p8, p8),\
    dl_paramToText (8, #p9, p9))

#else /* DL_PRESENT == 0 */
#define DL_BEGIN()
#define DL_BEGIN1( p0)
#define DL_BEGIN2( p0, p1)
#define DL_BEGIN3( p0, p1, p2)
#define DL_BEGIN4( p0, p1, p2, p3)
#define DL_BEGIN5( p0, p1, p2, p3, p4)
#define DL_BEGIN6( p0, p1, p2, p3, p4, p5)
#define DL_BEGIN7( p0, p1, p2, p3, p4, p5, p6)
#define DL_BEGIN8( p0, p1, p2, p3, p4, p5, p6, p7)
#define DL_BEGIN9( p0, p1, p2, p3, p4, p5, p6, p7, p8)
#define DL_BEGIN10(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9)
#define DL_BEGIN11(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10)
#define DL_BEGIN12(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11)
#define DL_BEGIN13(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12)
#define DL_BEGIN14(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13)
#define DL_BEGIN15(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14)
#define DL_BEGIN16(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15)

#define DL_PRINT1( p0)
#define DL_PRINT2( p0, p1)
#define DL_PRINT3( p0, p1, p2)
#define DL_PRINT4( p0, p1, p2, p3)
#define DL_PRINT5( p0, p1, p2, p3, p4)
#define DL_PRINT6( p0, p1, p2, p3, p4, p5)
#define DL_PRINT7( p0, p1, p2, p3, p4, p5, p6)
#define DL_PRINT8( p0, p1, p2, p3, p4, p5, p6, p7)
#define DL_PRINT9( p0, p1, p2, p3, p4, p5, p6, p7, p8)
#define DL_PRINT10(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9)
#define DL_PRINT11(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10)
#define DL_PRINT12(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11)
#define DL_PRINT13(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12)
#define DL_PRINT14(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13)
#define DL_PRINT15(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14)
#define DL_PRINT16(p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15)

#define DL_END()
#define DL_END1(p0)                         p0
#define DL_END2(p0,p1)                      p0
#define DL_END3(p0,p1,p2)                   p0
#define DL_END4(p0,p1,p2,p3)                p0
#define DL_END5(p0,p1,p2,p3,p4)             p0
#define DL_END6(p0,p1,p2,p3,p4,p5)          p0
#define DL_END7(p0,p1,p2,p3,p4,p5,p6)       p0
#define DL_END8(p0,p1,p2,p3,p4,p5,p6,p7)    p0
#define DL_END9(p0,p1,p2,p3,p4,p5,p6,p7,p8) p0

#define DL_ENDV1(p0)                 
#define DL_ENDV2(p0,p1)               
#define DL_ENDV3(p0,p1,p2)             
#define DL_ENDV4(p0,p1,p2,p3)           
#define DL_ENDV5(p0,p1,p2,p3,p4)         
#define DL_ENDV6(p0,p1,p2,p3,p4,p5)       
#define DL_ENDV7(p0,p1,p2,p3,p4,p5,p6)     
#define DL_ENDV8(p0,p1,p2,p3,p4,p5,p6,p7)   
#define DL_ENDV9(p0,p1,p2,p3,p4,p5,p6,p7,p8) 
#endif

#endif /* DL_DEBUGTRACE_H_INCLUDED */
