/*========================================================================
 *  Copyright (c) 1995 by Pica
 *
 *  File        : val.h
 *  Description : The external interface for the VAL-library
 *  Author(s)   : Paul Smit
 *  Creation    : 20-09-1995
 *
 *  History:
 *  20-09-1995  : paul     : creation
 *  28-11-1995  : paul     : new global parameter for function validation:
 *                           second number is copy number in case of copy val.
 *  15-07-1996  : paul     : valdir is new input parameter of VALValidate:
 *                           VALDIR_NONE, VALDIR_SIZE and valdir_def added
 *  04-07-1997  : paul     : enum VALErrno_t
 *  07-12-1997  : gert     : moved VALerrfmt() from VALLOCAH into VALH
 *  03-08-1998  : paul     : VALStrip() added
 *                           VAL_DIAG_NRLINES+VAL_DIAG_LINELEN replaced by
 *                           VAL_DIAG_SIZE
 *                           VALErrnoStr() added
 *  30-09-1998  : paul     : valerrs_def replaced by VALErrors_t
 *                           valdir_def  replaced by VALDirectory_t
 *                           VALErrors_t, VALError_t copied from csrc
 *                           interface of VALerrfmt() generalized
 *                           VALErrorFormat() added
 *                           ERROR_TYPE(VALErrno_t) instead of enum
 *                           VALS_NONE introduced
 *                           VALDenormalize() introduced
 *  12-10-1998  : paul     : Moved SYNPTRMAX, VALExprType_t from vallocal.h
 *                           VAL_SYNBUF_LEN() and VAL_MESSBUF_LEN() added
 *  13-11-1998  : paul     : TK_PICAPLUS and VAL_MT_PICAPLUS removed
 *  26-01-1999  : paul     : VALErrorSubj_t introduced
 *                           VALFormatTable() introduced
 *                           VALRecordFormat_t added
 *                           interface of VALErrorFormat()+VALerrfmt() changed
 *                           VALE_ERROR added
 *  08-03-2002  : paul     : Added VALSyntax_t
 *                           Added VALComputeNumber(), VALDenormalizeCreate()
 *                           and VALDenormalizeDestroy()
 *  09-04-2002  : paul     : Added VALDenormalizeCreateCS()
 *  22-12-2003  : paul     : Added new member sb_charset to VALSynBuf_t
 *                         : Added VALCharset()
 *                         : Added VALCompileCreate() + VALCompileDestroy()
 *  29-12-2003  : paul     : Added VALCharset_t
 *  13-01-2004  : paul     : Changed type of VALErrorSubj_t members p1 and p2
 *                           from unsigned short to size_t
 *  16-11-2004  : paul     : added VALErrorFormatFull()
 *  01-12-2004  : paul     : replaced in interface of VALErrFmt()
 *                           cvoidp *cnvsyn by const FCVSynBuf_t *cnvsyn
 *                           since CVKMPP syntax is no longer used
 *  07-01-2005  : paul     : renamed VALFormatTable() to VALFiltering()
 *                         : added flttags param to VALFiltering()
 *  17-01-2005  : paul     : Added VAL_ET_TAGSUBFCODE
 *  24-01-2005  : paul     : VAL_TAGSUBFCODE_SIZE extended to fit 209A/999 $x00
 *  23-03-2005  : paul     : Error subjects pointing to FCV syntax represented
 *                           as strings, rather than offsets
 *                         : Removed VALerrfmt()
 *  20-09-2005  : paul     : Added #define VAL_ET_STRPTRMASK
 *  06-07-2006  : paul     : removed VALDenormalize()
 *  30-01-2007  : paul     : Added VALValidateCat()
 *  30-01-2007  : paul     : Added VALValidateCat()
 *  06-02-2007  : paul     : Removed VAL_ET_TAGSUBFCODE
 *  09-02-2007  : paul     : Added sb_tagname (syntax of function 'tagname')
 *  04-08-2008  : paul     : Maximum number increased from 32K to 2G
 *  18-03-2009	: paul     : Added support for level 3 (social) tags
 *  24-02-2010	: paul     : Moved VALSynBuf_t to vallocal.h
 *                           Removed VALCompile() + VALStrip()
 *  16-03-2010	: paul     : Made VALMessages_t more flexible
 *
 *  Notes       :
 *
 *========================================================================
 */

#if ! defined(VAL_H_INCLUDED)
#define VAL_H_INCLUDED

#pragma ident "$Id: val.h 368722 2019-07-19 14:10:23Z smitp $"

#include <limits.h>
#include <stddef.h>

#include <pica.h>
#include <picatypes.h>
#include <idxtypes.h>
#include <cs.h>
#include <fcv.h>
#include <spp.h>
#include <pp.h>
#include <pa.h>

ERROR_TYPE(VALErrno_t);

ERROR_DECLARE(VALE_NOERROR);
ERROR_DECLARE(VALE_NOMEM);
ERROR_DECLARE(VALE_NOHEAP);
ERROR_DECLARE(VALE_SYNTAX);
ERROR_DECLARE(VALE_UNEXPECTED_GLOBAL_PARAMETERS);
ERROR_DECLARE(VALE_OBJECT_TABLE_CORRUPT);
ERROR_DECLARE(VALE_VARIABLE_STACK_CORRUPT);
ERROR_DECLARE(VALE_INVALID_EXPRESSION);
ERROR_DECLARE(VALE_STACK_OVERFLOW);
ERROR_DECLARE(VALE_STACK_UNDERFLOW);
ERROR_DECLARE(VALE_BUFFER_OVERFLOW);
ERROR_DECLARE(VALE_INVALID_TITLE);
ERROR_DECLARE(VALE_ERROR);

extern VALErrno_t VALerrno;

/*
 * During the transition from charset Pica to charset Utf8, it's necessary to
 * keep track of the charset. CHARSET_ASCII means that none of the string
 * constant in the syntax contains either a Pica-diacritic, or a utf8-grapheme.
 */
typedef enum VALCharset_t {
    VAL_CHARSET_ASCII,
    VAL_CHARSET_PICA,
    VAL_CHARSET_UTF8
} VALCharset_t;
    
#define VAL_DIAG_SIZE           800
extern char     VALDiagnostic[VAL_DIAG_SIZE];

/* The messages are stored in a struct of type VALMessages_t. For each
 * validation type there is one such struct. The actual size of the object
 * containing the messages may vary: the array messidx will only partly be
 * filled with indexes of strings. The strings these indexes point to stored
 * at the end of messidx.
 */
#define VAL_MT_MAX VAL_MT_SYNTACTIC
typedef enum VALMessType_t {
    VAL_MT_NONE, VAL_MT_SEMANTIC,
    VAL_MT_MAX
} VALMessType_t;
#define VAL_LANG_MIN VAL_LANG_NED
#define VAL_LANG_MAX VAL_LANG_ENG
typedef enum VALLanguage_t {
    VAL_LANG_MIN, VAL_LANG_FRA, VAL_LANG_DEU,
    VAL_LANG_MAX
} VALLanguage_t;

#define VAL_MAX_MESSNR   999
#define VAL_MESSIDX_NONE -1
typedef struct VALMessages_t {
   unsigned short	mb_seqnr[VAL_LANG_MAX + 1];
   long			*mb_idx;
   long			mb_idx_end;
   long			mb_mess_size;
   long			mb_mess_size_max;
   char			*mb_mess;
} VALMessages_t, *VALMessages_p;
typedef const VALMessages_t	*VALMessages_cp;

typedef enum VALStatus_t {
    VALS_NONE, VALS_OK, VALS_WARNING, VALS_ERROR, VALS_FATAL
} VALStatus_t;

typedef struct VALSynBuf_t VALSynBuf_t, *VALSynBuf_p;
typedef const struct VALSynBuf_t *VALSynBuf_cp;
typedef struct VALSyntax_t {
    VALSynBuf_t     *synbuf;
    VALMessages_p   *messages;
} VALSyntax_t, *VALSyntax_p;

// Pointer to an element of a validation syntax
typedef const struct VALSynbufElt_t	*VALSynbufElt_cp;

/*
 * VALErrors_t is a list of internal errors. These errors will be
 * formatted by VALerrfmt()
 */
#define VAL_ET_SYNPTRMASK          0X8000
#define VAL_ET_STRPTRMASK          0X4000
#define VAL_ET_PPELTMASK           0X2000
#define VAL_ET_TYPEOF(typ)         ((typ) & ~VAL_ET_SYNPTRMASK \
                                          & ~VAL_ET_PPELTMASK \
                                          & ~VAL_ET_STRPTRMASK)
typedef enum VALExprType_t {
    VAL_ET_NONE = ' ', VAL_ET_FUNC = 'f',
    VAL_ET_NUMBER = 'n', VAL_ET_STRING = 's', VAL_ET_BOOL = 'b',
    VAL_ET_SUBF = '$', VAL_ET_TAG = '#', VAL_ET_TABLE = 't'
} VALExprType_t;
#define VAL_VARSTACK_MAX       32
#define VAL_ERROR_MAX          8

/* A tag-subf-code consists of 1 position for the terminating 0-byte + either:
 * 4 positions tagcode + 2 positions subfieldcode
 * or 4 positions tagcode + 4 positions occurrence + 4 positions $x-occurrence
 * for example:
 * "028B$d" or "209A/999 $x00"
 */
#define	VAL_TAGCODE_LEN		4

typedef struct VALErrorSubj_t {
    unsigned			typ; /* can contain VALExprType_t 
					or-ed with VAL_ET_PPELTMASK
					or-ed with VAL_ET_SYNPTRMASK */
    union {
	struct {
	    size_t				p1;
	    size_t				p2;
	}				offset;
	struct {
	    cvoidp				s1;
	    cvoidp				s2;
	}				strptr;
    }				val;
} VALErrorSubj_t;
typedef struct VALError_t {
    short                           messnr;
    short                           len;
    VALErrorSubj_t                  subj[VAL_VARSTACK_MAX];
} VALError_t;
typedef struct VALErrors_t {
    short                           len;
    VALError_t                      err[VAL_ERROR_MAX];
} VALErrors_t, *VALErrors_p;
typedef const VALErrors_t	*VALErrors_cp;

/*
 * VALDirectory_t is an array of indexes of the record to validate.
 * Each entry either points to a tag, or is equal to VALDIR_NONE.
 * Entry 128 contains the index of the first tag with code 128*.
 */
#define VAL_MAX_LEV  3
#define VAL_MAX_SEQ  99
#define VALDIR_SIZE  ((VAL_MAX_LEV + 1) * (VAL_MAX_SEQ + 1))
#if defined(__XMEM)
#define VALDIR_NONE  ULONG_MAX
#else
#define VALDIR_NONE  USHRT_MAX
#endif /* #if defined(__XMEM) */
typedef size_t VALDirectory_t[VALDIR_SIZE];
typedef const VALDirectory_t	VALDirectory_ct;

/*
 * VALRecordFormat_t enumerates the record formats as they are recognized
 * by VALErrorFormat() and VALerrfmt()
 */
typedef enum VALRecordFormat_t {
    VAL_RECFMT_NONE, VAL_RECFMT_CAT,
    VAL_RECFMT_PICA, VAL_RECFMT_VAL
} VALRecordFormat_t;

VALSyntax_p
VALCompileCreate(
	const char	*options,	/* charset=pica or charset=utf8 */
        const char      *text_in);	/* input text */
void
VALCompileDestroy(
	VALSyntax_p		syntax);

VALStatus_t VALValidate(               /* EX error, warning or ok            */
    const char            *options,    /* IN options                         */
    const char            *recbuf,     /* IN the buffer containing the record*/
    size_t                 reclen,     /* IN the length of the record        */
    const VALDirectory_t   val_dir,    /* IN directory of record             */
    const VALSynBuf_t     *syn_buf,    /* IN the syntax                      */
    short                  iln,        /* IN the ILN                         */
    short                  expnum,     /* IN copy number or zero             */
    VALErrors_t           *val_err);   /* EX error or warning messages       */

VALStatus_t
VALValidateCat(
  const char           *optionsp,
  const char           *recbuf,        /* the buffer containing the record */
  size_t                reclen,        /* the length of the record */
  const VALDirectory_t  val_dir,       /* directory of record */
  const VALSynBuf_t     *syn_buf,      /* the validation syntax */
  FCVSynBuf_t		*cnvsyn,       /* Pica+ -> editfmt conv. syntax */
  short                 iln,           /* the ILN */
  short                 expnum,        /* copy number or zero */
  VALErrors_t          *val_err);      /* error or warning messages */

int32 VALComputeNumber(
  const char           *options,       /* IN options                         */
  const char           *recbuf,        /* the buffer containing the record */
  size_t                reclen,        /* the length of the record */
  const VALDirectory_t  val_dir,       /* directory of record */
  const VALSynBuf_t     *syn_buf,      /* IN the syntax */
  short                 iln,           /* the ILN */
  VALErrors_t          *val_err);      /* error or warning messages */

void VALErrorFormat(
    const char            *options,    /* IN options                       */
    const VALErrors_t     *valerrs,    /* IN validation error(s)           */
    const char            *recbuf,     /* IN record that caused the error  */
    size_t                 reclen,     /* IN length of record              */
    VALRecordFormat_t      recfmt,     /* IN format of the record          */
    bool                   isttl,      /* IN whether record type is title  */
    const VALSynBuf_t     *val_syn,    /* IN validation syntax table       */
    const FCVSynBuf_t     *cnvsyn,     /* IN Pica+ -> editfmt conv. syntax */
    const VALMessages_t   *messbuf,    /* IN error message format(s)       */
    char                  *outbuf,     /* EX buffer for formatted error(s) */
    size_t                 outsize);   /* IN size of error buffer          */

const char *
VALErrorFormatFull(
    const char		*options,	/* IN  options                       */
    const VALErrors_t	*valerrs,	/* IN  validation error(s)           */
    int			errornr,	/* IN  seq nr of the error or 0==all */
    const char		*recbuf,	/* IN  record that caused the error  */
    size_t		reclen,		/* IN  length of record              */
    VALRecordFormat_t	recfmt,		/* IN  format of the record          */
    bool		isttl,		/* IN  whether record type is title  */
    const VALSynBuf_t	*val_syn,	/* IN  validation syntax table       */
    const FCVSynBuf_t	*cnvsyn,	/* IN  Pica+ -> editfmt conv. syntax */
    const VALMessages_t	*messbuf);	/* IN  error message format(s)       */

// The id used by VAL-library to check if filter-validation needs to be traced
#define VAL_TRACE_ELTID_FILTER_VALIDATION "VAL_FILTER_VALIDATION"

// Declaration of callback function that writes info to trace file:
typedef void			(*VALTrace_f)(const char *text,
					      int text_len,
					      voidp userdata,
					      const char *options);
#define VAL_TRACE_NOFUNC	(VALTrace_f)0

VALStatus_t
VALFilteringTrace(
	const char	*options,	/* IN options                       */
	CSBuf_p 	recbuf,		/* IN record to be validated
					 * EX record with restored protected
					 *    tags
					 */
	PPTtl_p		flttags,	/* IN invisible tags                */
	PPTtl_p		old_editbuf,	/* IN visible tags in edit screen
					 *    or NULL for record insert     */
	FCVSynBuf_p	flttab,		/* IN Format table                  */
	CSBuf_p 	valfmtbuf,	/* EX record for VALErrorFormat()   */
	VALErrors_t	*valerrs,	/* EX validation error(s)           */
	VALTrace_f	valtrace,	// Tracing callback function
	voidp		userdata);	// Tracing callback data
VALStatus_t
VALFiltering(				/* EX error, warning or ok          */
	const char	*options,	/* IN options                       */
	CSBuf_p 	recbuf,	/* IN record to be checked          */
	PPTtl_p		flttags,	/* IN invisible tags                */
	PPTtl_p		old_editbuf,	/* IN PP-tags in original editbuf   */
	FCVSynBuf_p	flttab,		/* IN Format table                  */
	CSBuf_p 	tmpbuf,	/* EX record for VALErrorFormat()   */
	VALErrors_t	*valerrs);	/* EX validation error(s)           */
typedef struct {
    VALTrace_f		valtrace;		// Tracing callback function
    voidp		valtrace_userdata;	// Tracing callback data
    FCVConvertCB_t	fcvcb;			// Callbacks for FCV conversion
} VALCB_t, *VALCB_p;
VALStatus_t
VALFilteringCB(
	const char	*options,
	CSBuf_p		recbuf,
	PPTtl_p		flttags,
	PPTtl_p		old_editbuf,
	FCVSynBuf_p	flttab,
	CSBuf_p 	valfmtbuf,
	VALErrors_t	*valerrs,
	VALCB_p		cb);

void VALCBClear(VALCB_p cb);

void VALDenormalizeDestroy(char *tarbuf);

char *VALDenormalizeCreate(            /* EX denormalized title             */
    const char            *options,    /* IN options                        */
    PPTtl_p                srcttl,     /* IN normalized pica record         */
                                       /* EX sorted pica record             */
    size_t                *tarlen,     /* EX length of output title         */
    VALDirectory_t         valdir);    /* EX directory of tag codes present */

CSBuf_p
VALDenormalizeCreateCS(			/* EX denormalized title             */
    const char		*options,
    PPTtl_p		srcttl,		/* IN normalized pica record         */
					/* EX sorted pica record             */
    VALDirectory_t	valdir);	/* EX directory of tag codes present */

VALCharset_t VALCharset(const VALSyntax_t *syntax);

const char *
VALStringHasTemplate(
	const char	*str);
const char *
VALStringPrint(
        const char	*str);

// Interface of valcatinfo.c

// CatInfo function id's
typedef enum {
    VAL_CI_TAG_ACCESS_ALLOW,
    VAL_CI_TAG_OCC_RANGE,
    VAL_CI_TAG_XOCC_RANGE,
    VAL_CI_TAG_REQUIRED,
    VAL_CI_TAG_NON_REPEATABLE,
    VAL_CI_SBF_REQUIRED,
    VAL_CI_SBF_NON_REPEATABLE,
    VAL_CI_SBF_ACCESS_FORBID,
    VAL_CI_SBF_LENGTH,
    VAL_CI_SBF_VALUE,
    VAL_CI_SBF_TYPE,
    VAL_CI_SBF_CODELIST,
    VAL_CI_SBF_CODELIST_VALUES,
    VAL_CI_FUNC_NONE
} VAL_CI_Func_t;

const char *
VALCatInfoFuncCallArgToString(
	VALSynbufElt_cp		actArg);

typedef bool	(*VALCatInfoFuncCall_CB_f)(VAL_CI_Func_t, PAHandle_p, voidp);
bool
VALCatInfoFuncCallWalk(
	VALSyntax_p		valTab,
	VALCatInfoFuncCall_CB_f	funcCall_CB,
	voidp			userdata);

typedef bool	(*VALCatInfoCodeListWalk_CB_f)(const char *, voidp);
bool
VALCatInfoCodeListWalk(
	VALSynbufElt_cp			tableElt,
	VALCatInfoCodeListWalk_CB_f	processCode,
	voidp				userdata);

typedef struct VAL_tagspecset_t	*VAL_tagspecset_p;
void
VAL_TagspecSet_Destroy(
	VAL_tagspecset_p	tagset);
VAL_tagspecset_p
VAL_TagspecSet_Create(
	const char	*options,
	const char	*speclist);
SPP_tagspecset_p
VAL_TagspecSet_Get(
	VAL_tagspecset_p tagset);
bool
VAL_TagspecSet_Exclude(
	VAL_tagspecset_p tagset);

/*-----------------------------------------------------
 * PP-handle interface
 */

// Handle for valdidating and formatting messages
typedef struct VALHandle_t	*VALHandle_p;

// Return value of validation
typedef struct VALResult_t	*VALResult_p;

void
VALHandleDestroy(
	VALHandle_p	handle);

VALHandle_p
VALHandleCreate(
	VALSynBuf_cp	valSyn,
	VALMessages_p   *valMes,
	const char	*valKey,
	FCVSynBuf_p	fcvSyn,
	const char	*fcvKey,
	const char	*catFmt,
	IDXRecType_t	recType,
	const char	*sbfInd,
	const char	*dollar);

void
VALResultDestroy(VALResult_p	handle);

VALStatus_t
VALResultStatusGet(
	VALResult_p	result);
long
VALResultNumberGet(
	VALResult_p	result);
const char *
VALResultBriefMessGet(
	VALResult_p	result);
size_t
VALResultFullMessCountGet(
	VALResult_p	result);
const char *
VALResultFullMessGetN(
	VALResult_p	result,
	int		idx);
VALErrors_cp
VALResultErrorsGet(
	VALResult_p	result);

VALResult_p
VALValidatePP(
	VALHandle_p	handle,
	PPTtl_p		ppTtl,
	PICIln_t	iln,		// iln of local/copy blocks in ppTtl
	PICCopyNum_t	copyNum,	/* window on a single copy-record
					   or PIC_COPYNUM_NONE: only all copies
					   or PIC_COPYNUM_ANY: all levels */
	const char	*lang,		// language of error messages
	const char	*options);

VALResult_p
VALComputeNumberPP(
	VALHandle_p	handle,
	PPTtl_p		ppTtl,
	PICIln_t	iln,
	const char	*lang,
	const char	*options);

const char *
VALErrorFormatPPFull(
	VALResult_p	result,
	VALHandle_p	handle,
	PPTtl_p		ppTtl,
	int		errornr,
	const char	*lang,
	const char	*options);

#endif /* ! defined(VAL_H_INCLUDED) */
