/*========================================================================
 *  Copyright (c) 1997 by Pica
 *
 *  File	: xpn.h
 *  Description	: Declarations for expansion library
 *  Author(s)	: Bas de Bakker
 *  Creation	: 05-11-1997
 *
 *  History:
 *  05-11-1997	: BdB      : creation
 *  24-05-1999	: eric     : additions for dollar expansion (xpndollar.c)
 *  26-11-2001	: eck, dongens  :  created XPNExpand2,  which can have a
 *                           PCCTtlRead/charset-aware callback. XPNExpand now
 *                           wraps around XPNExpand2.
 *
 *  Notes	:
 *  Needs pp.h, pccttl.h and ttl.h included first for XPNExpand() interface.
 *  If we would admit that types like TTLHandle_p are pointers to structs,
 *  instead of hiding them behind such typedefs, we could simply declare the
 *  (incomplete) struct in this header and save a lot of dependencies and
 *  therefore compile time.
 *
 *========================================================================
 */

#ifndef XPN_H_INCLUDED
#define XPN_H_INCLUDED

#pragma ident "$Id: xpn.h 350691 2017-11-30 07:53:31Z smitp $"


#include <pp.h>
#include <ttl.h>
#include <picamisc.h>


ERROR_TYPE(XPNErrno_t);

ERROR_DECLARE(XPNE_NOERROR);
ERROR_DECLARE(XPNE_NOMEM);	/* Out of memory. */
ERROR_DECLARE(XPNE_SYNTAX);	/* Syntax error in expansion table. */
ERROR_DECLARE(XPNE_TITLE);	/* Problems getting new title. */
ERROR_DECLARE(XPNE_INTERNAL);	/* Internal error in xpn library. */
ERROR_DECLARE(XPNE_NMT);	/* Unexpected NMT error. */
ERROR_DECLARE(XPNE_PARAM);	/* Incorrect parameter/argument. */

extern XPNErrno_t XPNerrno;

#define XPN_DIAG_SIZE   (800)
extern char XPNDiagnostic[ XPN_DIAG_SIZE ];

/** Number of positions of the material type used in the expansion table */
#define XPN_MATERIAL_LEN 2

typedef const struct XPNHandle_t *XPNHandle_p;

typedef struct XPNDollarTable *XPNDollarTable_p;

/* Compile an expansion table to a handle.  Returns handle or NULL on error.
   If a %include is used, will call 'include' (if not NULL) and expects it to
   return the text of the included table (or NULL which will be ignored). */
XPNHandle_p
XPNCompile(const char *table, const char *(*include)(const char *table));

/* Destroy a handle created by XPNCompile (i.e., free associated memory). */
void
XPNDestroy(XPNHandle_p handle);

// Callback function for reading a record
typedef PPTtl_p (*XPNGetTitle_f)(TTLHandle_p, PICIpn_t, cvoidp);
#define XPN_GET_TITLE_NOFUNC	((XPNGetTitle_f)0)

// Callback function for computing IPN from PPN
typedef PICIpn_t (*XPNIpnFromEdp_f)(PICFileSet_t titleset, const char *, long);
#define XPN_IPN_FROM_EDP_NOFUNC ((XPNIpnFromEdp_f)0)
						 
// Callback function for hyphenating ISBN's
typedef size_t (*XPNHyphenateIsbn_f)(const char *, char *, size_t);
#define XPN_HYPHENATE_ISBN_NOFUNC ((XPNHyphenateIsbn_f)0)

// Callbacks+data needed for expansion
typedef struct {
    PMPpnTab_p		ppn2ipn;
    XPNIpnFromEdp_f	ipn_from_edp;
    XPNGetTitle_f	gettitle;
    cvoidp		gettitle_userdata;
    XPNHyphenateIsbn_f	hyphenate_isbn;
} XPNExpandCB_t, *XPNExpandCB_p;

/* Expand a title according to a table handle.  Get titles used for expansion
   with 'ttlhandle'.  Convert PPNs to IPNs according to 'ppn2ipn'.  Use
   'language' for language dependent expansions.  Return FALSE on error.
   Current options are "mode=online" (put expansion with interpunctions in
   subfield '8') or "mode=offline". */
bool
XPNExpand(const char *options, XPNHandle_p handle, PPTtl_p title,
	TTLHandle_p ttlhandle, PMPpnTab_p ppn2ipn, const char *language);
bool
XPNExpand2(const char *options, XPNHandle_p handle, PPTtl_p title,
	TTLHandle_p ttlhandle, PMPpnTab_p ppn2ipn, const char *language,
	XPNGetTitle_f gettitle,
	cvoidp gettitle_userdata); 

bool
XPNExpand3(const char *options, XPNHandle_p handle, PPTtl_p title,
	TTLHandle_p ttlhandle, const char *language,
	XPNIpnFromEdp_f ipn_from_edp,
	XPNGetTitle_f gettitle,
	cvoidp gettitle_userdata);

void
XPNExpandCBClear(
	XPNExpandCB_p	cb);

bool
XPNExpandCB(
	const char	*options,
	XPNHandle_p	handle,
	PPTtl_p		title,
	TTLHandle_p	ttlhandle,
	const char	*language,
	XPNExpandCB_p	cb);

/* Deexpand title: remove all $8 subfields. */
bool
XPNDeexpand(PPTtl_p title);

/* Prototypes xpndollar.c */

XPNDollarTable_p XPNDollarTableCompile(
	CBBuf_t *cbbuf,
	size_t cbsize);

void XPNDollarTableFree(
	XPNDollarTable_p table);

bool XPNDollarExpand(
	PPTtl_p ppttl,
	int16 level,
	XPNDollarTable_p table);

#endif /* XPN_H_INCLUDED */
