/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbcopy.c
 *  Description	: Copy an adb to another one
 *  Author(s)	: Hans van Eck
 *  Creation	: 29-11-1993
 *
 *  History:
 *  29-11-1993	: eck      : creation
 *
 *========================================================================
 */
#pragma ident "$Id: adbcopy.c 295831 2013-01-29 11:28:09Z helspere $"

#include <pica.h>
#include <adb.h>
#include "adblocal.h"

static void CopyEltList(adb_eltlist_p *list_to_pp, adb_eltlist_p list_from_p);

/*------------------------------------------------------------------------
 *  Function	: CopyElt
 *  Purpose	: Copy a list of elements with the same name
 *  Method	: Walk down link lists
 *
 *  History:
 *  29-11-1993	: eck      : creation
 *
 *------------------------------------------------------------------------
 */
static void
CopyElt(adb_elt_p	*elt_to_pp,	/* address of ptr to elt. to copy to */
	adb_elt_p	elt_from_p)	/* pointer to element to copy from */
{
    while (elt_from_p != NULL) {
	switch (elt_from_p->ae_type) {
	case ADBELT_VALUE :
		(*elt_to_pp) = ADBLNewElt(elt_from_p->ae_value);
		break;
	case ADBELT_RECORD :
		(*elt_to_pp) = ADBLNewElt(NULL);
		CopyEltList(&(*elt_to_pp)->ae_record, elt_from_p->ae_record);
	}
	elt_to_pp = &(*elt_to_pp)->ae_next;
	elt_from_p = elt_from_p->ae_next;
    }
} /* CopyElt() */

/*------------------------------------------------------------------------
 *  Function	: CopyEltList
 *  Purpose	: Copy all the elements of a list
 *  Method	: - Walk down linked list of element-list
 *		  - For every variable in the form list:
 *			- create the name in the to list (if necessary)
 *			- append contents of all elements with this name
 *
 *  History:
 *  29-11-1993	: eck      : creation
 *
 *------------------------------------------------------------------------
 */
static void
CopyEltList(
	adb_eltlist_p	*list_to_pp,	/* addr of ptr to elt-list to copy to */
	adb_eltlist_p	list_from_p)	/* pointer to elt-list to copy from */
{
    adb_eltlist_p	*tmp_list_pp = NULL;
    adb_elt_p		*tmp_elt_pp = NULL;
    const char		*fieldname;

    while ((ADBerrno == ADBE_NOERROR) && (list_from_p != NULL)) {
	tmp_list_pp = ADBLGetEltList(list_to_pp, list_from_p->ael_name,
					&fieldname);
	if (tmp_list_pp != NULL) {
	    if (*tmp_list_pp == NULL) {
		*tmp_list_pp = ADBLNewEltList(fieldname);
	    }
	} else {
	    ADBerrno = ADBE_INTEGRITY;
	}
	if (ADBerrno == ADBE_NOERROR) {
	    tmp_elt_pp = &(*tmp_list_pp)->ael_eltp;

	    /* Find the end of the list in order to append the element */
	    while (*tmp_elt_pp != NULL) {
		tmp_elt_pp = &(*tmp_elt_pp)->ae_next;
	    }
	    CopyElt(tmp_elt_pp, list_from_p->ael_eltp);
	    list_from_p = list_from_p->ael_next;
	}
    } /* while ((ADBerrno == ADBE_NOERROR) && (list_from_p != NULL)) */
} /* CopyEltList() */

/*------------------------------------------------------------------------
 *  Function	: ADBCopy
 *  Purpose	: Copy an ADB to another ADB.
 *  Method	: Recursive copy of the elements
 *
 *  History:
 *  29-11-1993	: eck      : creation
 *
 *  Notes	:
 *  - By using ADBWithBegin()'s, parts of ADB's can be copied.
 *  - The copy does not delete old data.
 *  - It is not possible to copy an ADB to itself.
 *
 *  Description :
 *
 *  Both ADBto and ADBfrom must be valid full-memory ADB-handles
 *  (returned by ADBCreate(ADBT_FULLMEM)).  The handles may not be the
 *  same.
 *
 *  This routines copies the contents of (the current with-context of)
 *  ADBfrom to (the current with-context of) ADBto.
 *
 *  The function returns TRUE if the ADB could be fully copied.  If
 *  not, FALSE is returned and ADBerrno will contain the error.  In
 *  this case, ADBto may be left in an inconsistent state.
 *
 *  Example     :
 *
 *	#include <adb.h>
 *	ADBHandle_p	adb1, adb2;
 *	.
 *	adb1 = ADBCreate(ADBT_FULLMEM);
 *	adb2 = ADBCreate(ADBT_FULLMEM);
 *	.
 *	.
 *	if (!ADBCopy(adb2, adb1)) {
 *		ERWrite(ADBerrno, jobname, "copying adb");
 *	}
 *	.
 *
 *  Returns	: indication of success.
 *
 *------------------------------------------------------------------------
 */
bool
ADBCopy(ADBHandle_p	ADBto,		/* handle to ADB to copy to */
	ADBHandle_p	ADBfrom)	/* handle to ADB to copy from */
{
    if (ADBto == ADBfrom) {
	ADBerrno = ADBE_SELFCOPY;
    } else {
	ADBerrno = ADBE_NOERROR;

	CopyEltList(ADBto->sh_last->se_eltlist_pp,
		    *(ADBfrom->sh_last->se_eltlist_pp));
    }
    return (ADBerrno == ADBE_NOERROR);
} /* ADBCopy() */
