/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbdelvar.c
 *  Description	: Delete a variable (recursively)
 *  Author(s)	: Hans van Eck
 *  Creation	: 12-11-1993
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *  09-12-1993	: eck      : changed test for deletion of element list
 *  13-01-1994	: eck      : added initialisation of ADBerrno
 *
 *========================================================================
 */

#pragma ident "$Id: adbdelvar.c 295831 2013-01-29 11:28:09Z helspere $"

#include <stdlib.h>
#include <pica.h>
#include <adb.h>
#include "adblocal.h"

/*------------------------------------------------------------------------
 *  Function	: ADBDelVar
 *  Purpose	: Delete a variable from an ADB
 *  Method	: - find (address of) element-list struct containing
 *			pointer to the element
 *		: - find the element in the list && delete
 *		: - if this was the last element
 *		:	remove element-list struct
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *  09-12-1993	: eck      : Test for deleting eltlist was too greedy.
 *			   : Deleting indexed variables went wrong
 *  13-01-1994	: eck      : added initialisation of ADBerrno
 *  07-04-1994	: eck      : When the last variable with a name was deleted,
 *			     its eltlist could be deleted.  The eltlists
 *			     following this one, however, could not.  This
 *			     is prevented by setting the ael_next pointer
 *			     to NULL.
 *
 *  Description :
 *
 *  ADBHandle must be a valid handle (returned by ADBCreate(ADBT_FULLMEM)).
 *
 *  ADBVar must be a valid variable reference.
 *
 *  This function will delete a variable or field (within the current
 *  with-context if possible) and its subfields.  When the variable is
 *  an array-element, the indexes of the higher array elements will
 *  decrease.
 *
 *  The function returns TRUE if the variable and all its subfields
 *  could be deleted.  If not, FALSE is returned and ADBerrno will
 *  contain the error.  The resulting ADB will then still contain the
 *  variable, but some of its subfields may have been deleted.
 *
 *  Deleting elements for which a with-context is active is not
 *  allowed and will yield undefined results.
 *
 *  Example     :
 *
 *	#include <adb.h>
 *
 *	ADBHandle_p		cur_adb;
 *
 *	cur_adb = ADBCreate(ADBT_FULLMEM);
 *	ADBNewVar(cur_adb, "set", NULL);
 *	ADBWithBegin(cur_adb, "set[:last]");
 *	.
 *	.
 *	ADBWithEnd(cur_adb);
 *	if (!ADBDelVar(cur_adb, "set[:last]")) {
 *		ERWrite(ADBerrno, jobname, "deleting ADB-var");
 *	}
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
ADBDelVar(
	ADBHandle_p	ADBhandle,	/* The ADB to delete a variable in */
	const char	*ADBvar)	/* Variable to delete */
{
    adb_elt_p		*elt_pp = NULL;
    adb_eltlist_p	*eltlist_pp;
    const char		*field_name;

    ADBerrno = ADBE_NOERROR;
    /*
     * Go to the variable in two steps.  This makes it easier to delete
     * the last element.
     */
    eltlist_pp = ADBhandle->sh_last->se_eltlist_pp;
    eltlist_pp = ADBLGetEltList(eltlist_pp, ADBvar, &field_name);

    if (eltlist_pp != NULL)
	elt_pp = ADBLGetElt(eltlist_pp, field_name);

    if ((elt_pp != NULL) && (*elt_pp != NULL)) {
	adb_elt_p elt_p = *elt_pp;

	(*elt_pp) = elt_p->ae_next;
	elt_p->ae_next = NULL;
	ADBLFreeElt(elt_p);
	if ((*eltlist_pp)->ael_eltp == NULL) {
	    /* the entry can now be deleted from the list */
	    adb_eltlist_p	eltlist_p = *eltlist_pp;

	    (*eltlist_pp) = (*eltlist_pp)->ael_next;

	    eltlist_p->ael_next = NULL;		/* Don't delete too much! */
	    ADBLFreeEltList(eltlist_p);
	}
    }
    return ADBerrno == ADBE_NOERROR;
} /* ADBDelVar() */
