/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbelts.c
 *  Description	: Element and element-list allocation & deallocation
 *  Author(s)	: Hans van Eck
 *  Creation	: 12-11-1993
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *
 *========================================================================
 */

#pragma ident "$Id: adbelts.c 295831 2013-01-29 11:28:09Z helspere $"

#include <stdlib.h>
#include <string.h>
#include <pica.h>
#include <adb.h>
#include "adblocal.h"

/*------------------------------------------------------------------------
 *  Function	: ADBLNewElt
 *  Purpose	: Create a new initialised struct adb_elt_t
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Notes	: NULL as value indicates the creation of a record
 *
 *  Returns	: pointer to the new element or NULL
 *------------------------------------------------------------------------
 */
adb_elt_p
ADBLNewElt(
	const char *ADBvalue)		/* The value to initialise with */
{
    struct adb_elt_t *eltp;

    eltp = malloc(sizeof(*eltp));

    if (eltp != NULL) {
	eltp->ae_next = NULL;
	if (ADBvalue == NULL) {
	    eltp->ae_type = ADBELT_RECORD;
	    eltp->ae_record = NULL;
	} else {
	    eltp->ae_type = ADBELT_VALUE;
	    eltp->ae_value = malloc(strlen(ADBvalue) + 1);
	    if (eltp->ae_value != NULL) {
		(void) strcpy((char *)eltp->ae_value, ADBvalue);
	    } else {
		/* malloc of the value failed. clean up the element */
		free(eltp);
		eltp = NULL;
	    }
	}
    }
    if (eltp == NULL) ADBerrno = ADBE_NOMEM;

    return eltp;
} /* ADBLNewElt() */

/*------------------------------------------------------------------------
 *  Function	: ADBLNewEltList
 *  Purpose	: Allocate a new struct abd_eltlist_t && initialise
 *
 *  History:
 *  09-11-1993	: eck      : creation
 *
 *  Returns	: pointer to the element or NULL
 *------------------------------------------------------------------------
 */
adb_eltlist_p
ADBLNewEltList(
	const char 	*ADBvar)	/* The field to create */
{
    adb_eltlist_p	listp ;

    listp = malloc(sizeof(*listp));
    if (listp != NULL) {
	listp->ael_next = NULL;
	listp->ael_eltp = NULL;
	listp->ael_name = ADBNTEnter(ADBvar);
	if (listp->ael_name == NULL) {
	    free(listp);
	    listp = NULL;
	}
    } else {
	ADBerrno = ADBE_NOMEM;
    }
    return listp;
} /* ADBLNewEltList() */

/*------------------------------------------------------------------------
 *  Function	: ADBLFreeEltList
 *  Purpose	: Free an element-list and all it's elements
 *  Method	: Obvious
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *  18-11-1993	: eck      : clear pointers to catch illegal use
 *------------------------------------------------------------------------
 */
void
ADBLFreeEltList(
	adb_eltlist_p	eltlist_p)	/* The element-list to free */
{
    adb_eltlist_p	tmp_p;

    while (eltlist_p != NULL) {
	if (eltlist_p->ael_eltp != NULL) {
	    ADBLFreeElt(eltlist_p->ael_eltp);
	}
	tmp_p = eltlist_p;
	eltlist_p = eltlist_p->ael_next;
	tmp_p->ael_next = NULL;
	tmp_p->ael_eltp = NULL;
	tmp_p->ael_name = NULL;
	free(tmp_p);
    }
} /* ADBLFreeEltList() */

/*------------------------------------------------------------------------
 *  Function	: ADBLFreeElt
 *  Purpose	: Free an element (and possibly sublists)
 *  Method	: Obvious
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *  18-11-1993	: eck      : clear pointers to catch illegal use
 *------------------------------------------------------------------------
 */
void
ADBLFreeElt(
	adb_elt_p	elt_p)		/* The element to free */
{
    adb_elt_p	tmp_elt_p;

    while (elt_p != NULL) {
	switch (elt_p->ae_type) {
	case ADBELT_VALUE:
		if (elt_p->ae_value != NULL) {
		    /* Cast to get rid of const */
		    free((void *)elt_p->ae_value);
		    elt_p->ae_value= NULL;
		}
		break;
	case ADBELT_RECORD:
		if (elt_p->ae_record != NULL) {
		    ADBLFreeEltList(elt_p->ae_record);
		    elt_p->ae_record = NULL;
		}
		break;
	}
	tmp_elt_p = elt_p;
	elt_p = elt_p->ae_next;
	tmp_elt_p->ae_next = NULL;
	free(tmp_elt_p);
    } /* while (elt_p != NULL) */
} /* ADBLFreeElt() */
