/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbid.c
 *  Description	: Identifier handling for ADB's
 *  Author(s)	: Hans van Eck
 *  Creation	: 08-11-1993
 *
 *  History:
 *  08-11-1993	: eck      : creation
 *  09-12-1993	: eck      : fixed ADBLGetEltList
 *  12-01-1993	: eck      : CS_DYNAMIC() replaced by CS_LOCAL() declarations.
 *			   : removed testing on fields with '?'
 *  26-09-1994	: eck      : fixed ADBLGetEltList (again...)
 *  21-03-1995	: eck      : Removed heavily used Counted Strings: speed
 *
 *========================================================================
 */

#pragma ident "$Id: adbid.c 295831 2013-01-29 11:28:09Z helspere $"

#include <ctype.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <pica.h>
#include <cs.h>
#include <adb.h>
#include "adblocal.h"

#define IDX_LAST	1
#define IDX_UNKNOWN	2

#define SkipSpaces(var)	while (**var == ' ') (*var)++

/*------------------------------------------------------------------------
 *  Function	: GetVar
 *  Purpose	: Get a variable and save it
 *  Method	: Save the legal characters until an illegal one found.
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *  21-03-1995	: eck      : Don't use CSAppendC(): speed
 *
 *  Notes	:
 *  - A field starts with a letter followed by letters, digits, underscores
 *  and minus-signs.
 *  - Extra characters are ingnored.
 *
 *  Returns	: indication of success (#chars > 0)
 *------------------------------------------------------------------------
 */
static bool
GetVar(
	char		*name,
	size_t		name_size,
	const char	**var)
{
    size_t	size_left = name_size - 1;	/* room for '\0'-byte */

    if (isalpha(**var)) {
	while (isalnum(**var) || (**var == '_') || (**var == '-')) {
	    if (size_left != 0) {
		*name++ = *(*var)++;
		size_left--;
	    }
	}
    }
    *name = '\0';
    return (size_left != (name_size - 1));
} /* GetVar() */

/*------------------------------------------------------------------------
 *  Function	: GetInt
 *  Purpose	: Get the integer part of an index
 *  Method	: strtoul
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *
 *  Notes	: It is just a wrapper
 *
 *  Returns	: the value read
 *------------------------------------------------------------------------
 */
static unsigned long
GetInt(
	const char **var_pp)
{
    /* The 10 is the base. */
    return strtoul(*var_pp, (char **) var_pp, 10);
} /* GetInt() */

/*------------------------------------------------------------------------
 *  Function	: GetContents
 *  Purpose	: Get the contents to match
 *  Method	: Save characters until ']' or end of string
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *------------------------------------------------------------------------
 */
static void
GetContents(
	char		*buf,
	size_t		buf_size,
	const char	**var_pp)
{
    size_t	size_left = buf_size - 1;	/* room for '\0'-byte */

    SkipSpaces(var_pp);

    while ((**var_pp != '\0') && (**var_pp != ']')) {
	if (size_left != 0) {
	    *buf++ = *(*var_pp)++;
	    size_left--;
	}
    }
    *buf = '\0';
} /* GetContents() */

/*------------------------------------------------------------------------
 *  Function	: GetSpecialIndex
 *  Purpose	: Get a special index (starting with ':')
 *  Method	: Obvious.  Only ":last" is recognised.
 *
 *  History:
 *  12-11-1993	: eck      : creation
 *
 *  Returns	: IDX_LAST or IDX_UNKNOWN
 *------------------------------------------------------------------------
 */
static int
GetSpecialIndex(
	const char **var_pp)
{
    if (**var_pp == ':') { (*var_pp)++; } else { return IDX_UNKNOWN; }
    if (**var_pp == 'l') { (*var_pp)++; } else { return IDX_UNKNOWN; }
    if (**var_pp == 'a') { (*var_pp)++; } else { return IDX_UNKNOWN; }
    if (**var_pp == 's') { (*var_pp)++; } else { return IDX_UNKNOWN; }
    if (**var_pp == 't') { (*var_pp)++; } else { return IDX_UNKNOWN; }
    return IDX_LAST;
} /* GetSpecialIndex() */

static adb_elt_p *ParseVar(adb_eltlist_p *eltlist_pp, const char **var_pp);

/*------------------------------------------------------------------------
 *  Function	: ParseIndex
 *  Purpose	: parse an (associative or number) index
 *  Method	: - check for specials (number, :last) and handle them
 *		: - loop over the elements:
 *		:	- look for field (if one given)
 *		:	- match the value (if necessary)
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *  12-01-1994	: eck      : removed presence testing
 *  06-12-1994	: eck      : Fixed bug which caused an ADBE_NOTPRESENT for
 *			   : absent fields when matching array-elements
 *
 *  Returns	:
 *  (address of) pointer to the first element that matches the index
 *  or NULL on failure.
 *------------------------------------------------------------------------
 */
static adb_elt_p *
ParseIndex(
	adb_elt_p	*elt_pp,
	const char	**var_pp)
{
    /* Test first char if index, might be digit, ':' or in [A-Za-z] */
    if (isdigit(**var_pp)) {
	unsigned long idx;

	idx = GetInt(var_pp);

	/* Skip fields until idx == 0 */
	while ((*elt_pp != NULL) && (idx != 0)) {
	    elt_pp = &(*elt_pp)->ae_next;
	    idx--;
	}
	if (*elt_pp == NULL) {
	    ADBerrno = ADBE_NOTPRESENT;
	    elt_pp = NULL;
	}
	return elt_pp;
    } else if (**var_pp == ':') {
	int	idxtyp = GetSpecialIndex(var_pp);

	if (idxtyp == IDX_LAST) {
	    while ((*elt_pp)->ae_next != NULL) {
		elt_pp = &(*elt_pp)->ae_next;
	    }
	} else {
	    ADBerrno = ADBE_ILLEGALKEY;
	    elt_pp = NULL;
	}
	return elt_pp;
    }

    /* We only come here when a variable is expected! */
    while (*elt_pp != NULL) {
	const char	*idx_start = *var_pp;
	char		value[ADB_KEYLENGTH_MAX + 1];
	adb_elt_p	*idx_elt_pp;

	/*
	 * We have something like var[f=a] or var[=a]
	 */
	if ((*elt_pp)->ae_type == ADBELT_RECORD) {
	    idx_elt_pp = ParseVar(&(*elt_pp)->ae_record, var_pp);
	} else {
	    idx_elt_pp = elt_pp;
	}
	if (idx_elt_pp != NULL) {
	    /* Found the right field, see if syntax is ok. */
	    if ((**var_pp) == '=') {
		(*var_pp)++;
		if ((*idx_elt_pp)->ae_type == ADBELT_VALUE) {
		    GetContents(value, sizeof(value), var_pp);
		    if (STREQ(value, (*idx_elt_pp)->ae_value)) {
			return elt_pp;
		    }
		}
	    } else {
		ADBerrno = ADBE_ILLEGALKEY;
		return NULL;
	    }
	} else {
	    if (ADBerrno == ADBE_NOTPRESENT) ADBerrno = ADBE_NOERROR;
	    if (ADBerrno != ADBE_NOERROR) return NULL;
	}
	/* Reset the start of the index, so we can parse it again, but now
	 * with the following element.
	 */
	*var_pp = idx_start;
	elt_pp = &(*elt_pp)->ae_next;
    } /* while (*elt_pp != NULL) */

    if (ADBerrno == ADBE_NOERROR) {
	ADBerrno = ADBE_NOTPRESENT;
    }
    return NULL;
} /* ParseIndex() */

/*------------------------------------------------------------------------
 *  Function	: ParseVar
 *  Purpose	: Parse a variable and return address of pointer to element
 *  Method	: - get parts of a variable until there are no more
 *		: - find the element with name in the current element list
 *		: - parse following parts (arrays && fields) when possible
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *  Notes	:
 *
 *  Returns	: address of pointer to the element or NULL
 *------------------------------------------------------------------------
 */
static adb_elt_p *
ParseVar(
	adb_eltlist_p	*eltlist_pp,
	const char	**var_pp)
{
    char		name[ADB_KEYLENGTH_MAX + 1];
    const char		*name_p;
    adb_elt_p		*elt_pp = NULL;
 
    SkipSpaces(var_pp);

    /* Loop until the variable has no more parts */
    while (GetVar(name, sizeof(name), var_pp)) {
	SkipSpaces(var_pp);

	name_p = ADBNTLookup(name);
	while (*eltlist_pp != NULL) {
	    if ((*eltlist_pp)->ael_name == name_p) {
		break;
	    }
	    eltlist_pp = &(*eltlist_pp)->ael_next;
	}
	if (*eltlist_pp == NULL) {
	    ADBerrno = ADBE_NOTPRESENT;
	    return NULL;
	}
	/*
	 * We found the name in the element list.  First parse an optional
	 * index.
	 */
	elt_pp = &(*eltlist_pp)->ael_eltp;
	if (**var_pp == '[') {
	    (*var_pp)++;
	    if ((elt_pp = ParseIndex(elt_pp, var_pp)) == NULL) {
		break;			/* from GetVar() loop */
	    }
	    if (**var_pp == ']') (*var_pp)++;
	    else ADBerrno = ADBE_ILLEGALKEY;
	}
	/*
	 * Now, be ready to parse a field (starting with '.')
	 */
	if (**var_pp != '.') break;	/* from GetVar() loop */

	(*var_pp)++;
	if ((*elt_pp)->ae_type != ADBELT_RECORD) {
	    ADBerrno = ADBE_TYPE;
	    return NULL;
	}
	/*
	 * This could also be a recursive call, but just assigning
	 * eltlist_pp is sufficient.
	 */
	eltlist_pp = &(*elt_pp)->ae_record;
    } /* while GetVar() */

    /* Return the element found */
    return elt_pp;
} /* ParseVar() */

/*------------------------------------------------------------------------
 *  Function	: ADBLGetElt
 *  Purpose	: Get (the address of) the pointer to the element indicated
 *		: by a variable.
 *  Method	: Call ParseVar && check if the end if reached
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *  Returns	: address of the pointer or NULL
 *------------------------------------------------------------------------
 */
adb_elt_p *
ADBLGetElt(
	adb_eltlist_p	*eltlist_pp,
	const char	*i_variable)
{
    const char	*var_p;
    adb_elt_p	*elt_pp;

    var_p = i_variable;
    elt_pp = ParseVar(eltlist_pp, &var_p);

    if (*var_p != '\0') {
	/* only last field unknown will yield ok pointers but ADBerrno set */
	/* so this test should be sufficient */
	if (ADBerrno == ADBE_NOERROR) ADBerrno = ADBE_ILLEGALKEY;
	return NULL;
    }
    return elt_pp;
} /* ADBLGetElt() */

/*------------------------------------------------------------------------
 *  Function	: ADBLGetList
 *  Purpose	: Get (the address of) a pointer to an element-list for
 *		: creating a variable.
 *  Method	: - find the last field
 *		: - ADBLGetElt without the last field
 *		: - Look for the field name in linked list
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *  09-12-1993	: eck      : Determination of last field did not take
 *			   : indexes into account, so the ADBNTLookup
 *			   : failed.
 *  26-09-1994	: eck      : Determination of last field failed when it was
 *			     indexed and there was a preceding path.
 *
 *  Notes	:
 *  fieldname will point to the name of the last field on return
 *
 *  Returns	:
 *  Address of pointer to where a new variable can be created or an old one
 *  deleted (or NULL).
 *------------------------------------------------------------------------
 */
adb_eltlist_p *
ADBLGetEltList(
	adb_eltlist_p	*eltlist_pp,
	const char	*i_variable,
	const char	**fieldname)
{
    const char	*var_p;
    const char	*end_p = NULL;
    adb_elt_p	*elt_pp;
    size_t	head_len;
    int		brack_level = 0;

    head_len = 0;
    var_p = i_variable;

    /*
     * Find the last field (preceded by a '.').  Ignore fields inside
     * indexes, and maintain a pointer (end_p) to the opening '[' of the
     * last field (if present).
     */
    while (*var_p != '\0') {
	if ((*var_p) == '.') {
	    head_len = var_p - i_variable + 1;
	    if (brack_level == 0) end_p = NULL;
	} else if ((*var_p) == '[') {
	    if (brack_level == 0) end_p = var_p;
	    brack_level++;
	} else if ((*var_p) == ']') {
	    brack_level--;
	}
	var_p++;
    }

    *fieldname = i_variable + head_len;
    if (head_len != 0) {
	CS_LOCAL(	tmpvar, ADB_KEYLENGTH_MAX);

	CSSetString(tmpvar, i_variable);
	CSDeletePart(tmpvar, head_len - 1, CSLength(tmpvar) - head_len + 1);

	elt_pp = ADBLGetElt(eltlist_pp, CSContents(tmpvar));

	if (elt_pp == NULL) {
	    return NULL;
	}
	if ((*elt_pp)->ae_type != ADBELT_RECORD) {
	    ADBerrno = ADBE_TYPE;
	    return NULL;
	}
	eltlist_pp = &((*elt_pp)->ae_record);
    } /* if (head_len != 0) */

    /* Now try to find the field */
    while (*eltlist_pp != NULL) {
	const char *name_p = NULL;
	
	/* When endp != NULL, the last field is indexed. */
	if (end_p == NULL) {
	    name_p = ADBNTLookup(*fieldname);
	} else {
	    CS_LOCAL(	tmpvar, ADB_KEYLENGTH_MAX);

	    CSPrintf(tmpvar, "%.*s",
		    (int) (end_p - i_variable - head_len), *fieldname);
	    name_p = ADBNTLookup(CSContents(tmpvar));
	}

	/* This code will also work when name_p == NULL */
	if ((*eltlist_pp)->ael_name == name_p) {
	    return eltlist_pp;
	}
	eltlist_pp = &(*eltlist_pp)->ael_next;
    }

    /* Name not found: return address of pointer where it should come */
    return eltlist_pp;
} /* ADBLGetEltList() */
