/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbnewvar.c
 *  Description	: Create a new variable
 *  Author(s)	: Hans van Eck
 *  Creation	: 08-11-1993
 *
 *  History:
 *  08-11-1993	: eck      : creation
 *  12-11-1993	: eck      : moved New.. routines to adbelts.c
 *
 *========================================================================
 */

#pragma ident "$Id: adbnewvar.c 295831 2013-01-29 11:28:09Z helspere $"

#include <pica.h>
#include <adb.h>
#include "adblocal.h"

/*------------------------------------------------------------------------
 *  Function	: ADBNewVar
 *  Purpose	: Create a new field
 *  Method	: - Get pointer to the list into which to insert the field
 *		: - If the field name is not present in the list then add
 *		: - Append the new value to the value-list
 *
 *  History:
 *  09-11-1993	: eck      : creation
 *
 *  Notes	:
 *  - Passing NULL as value will set the type to ADBELT_RECORD.  To create
 *  an empty field, an empty string should be passed.
 *
 *  Description :
 *
 *  ADBHandle must be a valid handle (returned by ADBCreate(ADBT_FULLMEM)).
 *
 *  ADBVar must be a valid variable reference, but needs not exist.
 *  It may not be indexed in the last part.
 *
 *  ADBValue must be either a '\0'-terminated string or a NULL-pointer.
 *
 *  This function will create a variable (in the current with-context
 *  if possible) that can then later be retrieved with (an indexed
 *  variant of) ADBVar.  When the variable is already present, it is
 *  seen as an array, and a new element is created.
 *
 *  The newly created value can be referenced by using :last as index.
 *
 *  The function returns TRUE if the variable could be set.  If not,
 *  FALSE is returned and ADBerrno will contain the error.
 *
 *  Example     :
 *
 *  See example at ADBGet().
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
ADBNewVar(
	ADBHandle_p	ADBhandle,	/* The ADB to create a var in */
	const char	*ADBvar,	/* The variable to create */
	const char	*ADBvalue)	/* The value to initialise with */
{
    adb_eltlist_p	*eltlist_pp = NULL;
    adb_elt_p		*elt_pp = NULL;
    const char		*field_name;

    ADBerrno = ADBE_NOERROR;

    /*
     * First get a pointer to the place where the field should be inserted
     * if there isn't already one.
     */
    eltlist_pp = ADBLGetEltList(ADBhandle->sh_last->se_eltlist_pp,
			    ADBvar, &field_name);

    if (eltlist_pp != NULL) {
	/*
	 * Now check if the proper eltlist is present, and allocate if
	 * necessary.  Initialise with the field-name.
	 */
	if (*eltlist_pp == NULL) {
	    (*eltlist_pp) = ADBLNewEltList(field_name);
	}
	/* Now a pointer to the element contents can be searched */
	if (*eltlist_pp != NULL) {
	    elt_pp = ADBLGetElt(eltlist_pp, field_name);
	}
    }
    if (elt_pp != NULL) {
	/* Follow linked list */
	while (*elt_pp != NULL) {
	    elt_pp = &((*elt_pp)->ae_next);
	}
	/* Create the element with contents */
	(*elt_pp) = ADBLNewElt(ADBvalue);
    }
    return ADBerrno == ADBE_NOERROR;
} /* ADBNewVar() */
