/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbprint.c
 *  Description	: Routine to print an ADB to file
 *  Author(s)	: Hans van Eck
 *  Creation	: 10-11-1993
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *========================================================================
 */
#pragma ident "$Id: adbprint.c 295831 2013-01-29 11:28:09Z helspere $"

#include <stdio.h>
#include <pica.h>
#include <adb.h>
#include "adblocal.h"

/*------------------------------------------------------------------------
 *  Function	: PrtIndentation
 *  Purpose	: Print a number of blanks for indentation purposes
 *  Method	: print an empty string with width depending on level
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
PrtIndentation(
	FILE	*file,
	int	level)
{
    return (fprintf(file, "%*s", 4 * level, "") >= 0);
} /* PrtIndentation() */

static bool
PrtEltList(FILE *file, adb_eltlist_p eltlist_p, bool compiled, int level);

/*------------------------------------------------------------------------
 *  Function	: PrtValue
 *  Purpose	: print a single value
 *  Method	: obvious (depending on compiled)
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
PrtValue(
	FILE		*file,		/* Handle to stream to print to */
	const char	*name,		/* The name of the element */
	adb_elt_p	elt_p,		/* The element to print */
	bool		compiled,	/* flag indicating compilation */
	int		level)		/* level of the element-list */
{
    bool	ok;

    if (compiled) {
	ok = (fprintf(file,"V%s=", name) >= 0);
	if (ok && elt_p->ae_value != NULL) {
	    ok = (fprintf(file,"%s", elt_p->ae_value) >= 0);
	}
	if (ok) {
	    ok = (fputc('\0', file) != EOF);
	}
    } else {
	ok = PrtIndentation(file, level);
	if (ok) {
	    ok = (fprintf(file,"%s = \"%s\";\n", name, elt_p->ae_value) >= 0);
	}
    }
    if (!ok) ADBerrno = ADBE_IO;
    return ok;
} /* PrtValue() */

/*------------------------------------------------------------------------
 *  Function	: PrtRecord
 *  Purpose	: Print a single record
 *  Method	  - print head (depending on compiled)
 *		  - print element list (with increased level)
 *		  - print tail (depending on compiled)
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
PrtRecord(
	FILE		*file,		/* Handle to stream to print to */
	const char	*name,		/* The name of the element */
	adb_elt_p	elt_p,		/* The element to print */
	bool		compiled,	/* flag indicating compilation */
	int		level)		/* level of the element-list */
{
    bool	ok;

    if (compiled) {
	ok = (fprintf(file,"R%s", name) >= 0);
	if (ok) {
	    ok = (fputc('\0', file) != EOF);
	}
    } else {
	ok = PrtIndentation(file, level);
	if (ok) {
	    ok = (fprintf(file, "%s {\n", name) >= 0);	/* } for vi */
	}
    }
    if (ok) {
	ok = PrtEltList(file, elt_p->ae_record, compiled, level + 1);
    }
    if (compiled) {
	if (ok) {
	    ok = (fputc('E', file) != EOF);
	}
    } else {
	if (ok) {
	    ok = PrtIndentation(file, level);		/* { for vi */
	}
	if (ok) {
	    ok = (fprintf(file, "}\n") >= 0);
	}
    }
    if (!ok) ADBerrno = ADBE_IO;
    return ok;
} /* PrtRecord() */

/*------------------------------------------------------------------------
 *  Function	: PrtElement
 *  Purpose	: Print elements with the same name
 *  Method	: walk down linked list  && call PrtValue() or PrtRecord()
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
PrtElement(
	FILE		*file,		/* Handle to stream to print to */
	const char	*name,		/* The name of the element */
	adb_elt_p	elt_p,		/* The element to print */
	bool		compiled,	/* flag indicating compilation */
	int		level)		/* level of the element-list */
{
    bool	ok = TRUE;

    while (ok && (elt_p != NULL)) {
	switch (elt_p->ae_type) {
	case ADBELT_RECORD:
	    ok = PrtRecord(file, name, elt_p, compiled, level);
	    break;
	case ADBELT_VALUE:
	    ok = PrtValue(file, name, elt_p, compiled, level);
	    break;
	}
	elt_p = elt_p->ae_next;
    } /* while (elt_p != NULL) */
    return ok;
} /* PrtElement() */

/*------------------------------------------------------------------------
 *  Function	: PrtEltList
 *  Purpose	: Print a whole element list
 *  Method	: walk down linked list && print the elements
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
PrtEltList(
	FILE		*file,		/* Handle to stream to print to */
	adb_eltlist_p	eltlist_p,	/* The element-list to print */
	bool		compiled,	/* flag indicating compilation */
	int		level)		/* level of the element-list */
{
    bool	ok = TRUE;

    while (ok && (eltlist_p != NULL)) {
	ok = PrtElement(file, eltlist_p->ael_name, eltlist_p->ael_eltp,
		    compiled, level);
	eltlist_p = eltlist_p->ael_next;
    }
    return ok;
} /* PrtEltList() */

/*------------------------------------------------------------------------
 *  Function	: ADBPrint
 *  Purpose	: Print a part of an ADB.
 *  Method	: call PrtEltList && fflush output
 *
 *  History:
 *  10-11-1993	: eck      : creation
 *
 *  Description :
 *
 *  ADBHandle must be a valid handle (returned by ADBCreate()).
 *
 *  file must be a valid stream handle (returned by fopen() or be stdout or stderr).
 *
 *  compiled indicates if the ADB should be printed in compiled form.
 *  The compiled form is suitable for reading with ADBReadFile(), in
 *  uncompiled form, is is suitable for human readers.
 *
 *  This function will print the contents of (the current with-context of) the ADB.
 *
 *  The function returns TRUE if the ADB could be fully printed. If
 *  not, FALSE is returned and ADBerrno will contain the error.
 *
 *  Example     :
 *
 * 	#include <adb.h>
 *	ADBHandle_p	adb;
 *	.
 *	adb = ADBCreate(ADBT_FULLMEM);
 *	.
 *	.
 *	if (!ADBPrint(stdout, adb, TRUE)) {
 *		ERWrite(ADBerrno, jobname, "printing failed");
 *	}
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
ADBPrint(
	FILE		*file,		/* Handle to stream to print to */
	ADBHandle_p	ADBhandle,	/* The ADB to print */
	bool		compiled)	/* flag indicating compilation */
{
    adb_eltlist_p	eltlist_p = *(ADBhandle->sh_last->se_eltlist_pp);
    bool		ok;
    
    ADBerrno = ADBE_NOERROR;

    ok = PrtEltList(file, eltlist_p, compiled, 0);

    if (ok) {
	ok = (fflush(file) != EOF);
	if (!ok) ADBerrno = ADBE_IO;
    }
    return ok;
} /* ADBDump() */
