/*=========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbreadfile.c
 *  Description	: Read a file into an ADB
 *  Author(s)	: Hans van Eck
 *  Creation	: 15-11-1993
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *  21-03-1995	: eck      : Removed Counted Strings for speed reasons
 *  12-11-1996	: eck      : Changed GetVar() into ReadVar() (Tandem problem).
 *
 *=========================================================================
 */
#pragma ident "$Id: adbreadfile.c 362304 2018-11-27 14:08:45Z heesakkm $"

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <pica.h>
#include <adb.h>
#include <cs.h>

/*-------------------------------------------------------------------------
 *  Function	: ReadVar
 *  Purpose	: Get a variable from the stream and store it.
 *  Method	: Obvious
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *  14-01-1994	: eck	   : changed CS_DYNAMIC()'s to CS_LOCAL()'s
 *  21-03-1995	: eck      : Removed CS-s: speed
 *
 *  Returns	: length of the variable read
 *-------------------------------------------------------------------------
 */
static size_t
ReadVar(
	char		*var_p,		/* buffer for name of the variable */
	size_t		var_size,	/* size of the buffer */
	FILE		*file)
{
    int		c;
    size_t	size_left = var_size - 1;

    c = getc(file);
    if (isalpha(c)) {
	while (isalnum(c) || (c == '_') || (c == '-')) {
	    if (size_left != 0) {
		*var_p ++ = (char) c;
		size_left--;
	    }
	    c = getc(file);
	}
    }
    (void) ungetc(c, file);
    *var_p = '\0';
    return (var_size - size_left - 1);
} /* ReadVar() */

/*------------------------------------------------------------------------
 *  Function	: GetValue
 *  Purpose	: Get a value from the stream and store it.
 *  Method	: Obvious
 *
 *  History:
 *  21-03-1995	: eck      : creation
 *
 *  Notes	: returns nothing, since the error(s) are detected in
 *		  DoVariable().
 *------------------------------------------------------------------------
 */
static void
GetValue(
	char		*value,		/* buffer for the value */
	size_t		val_size,	/* size of the buffer */
	FILE		*file)
{
    int		c;
    size_t	size_left = val_size - 1;

    while (((c = getc(file)) != EOF) && (c != '\0')) {
	if (size_left != 0) {
	    *value ++ = (char) c;
	    size_left--;
	} else {
	    /* Oops, value too large */
	    break;
	}
    }
    (void) ungetc(c, file);
    *value = '\0';
} /* GetValue() */

/*-------------------------------------------------------------------------
 *  Function	: DoRecord
 *  Purpose	: Read a variable and start a with
 *  Method	: read name, and call ADBWithBegin()
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *-------------------------------------------------------------------------
 */
static void
DoRecord(
	ADBHandle_p	ADBhandle,
	FILE		*file)
{
    char	var_buffer[ADB_KEYLENGTH_MAX + 1];
    size_t	var_len;

    if (((var_len = ReadVar(var_buffer, sizeof(var_buffer), file)) != 0)
	    && (getc(file) == '\0')) {
	(void) ADBNewVar(ADBhandle, var_buffer, NULL);
    } else {
	ADBerrno = ADBE_FILE;
    }

    if (ADBerrno == ADBE_NOERROR) {
	if ((var_len + 1) < sizeof(var_buffer) + sizeof("[:last]")) {
	    (void) strcpy(var_buffer + var_len, "[:last]");
	    (void) ADBWithBegin(ADBhandle, var_buffer);
	} else {
	    ADBerrno = ADBE_FILE;
	}
    }
} /* DoRecord() */

/*-------------------------------------------------------------------------
 *  Function	: DoVariable
 *  Purpose	: Read a variable + value and store it in an ADB
 *  Method	: read name, '=' and value and store
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *-------------------------------------------------------------------------
 */
static void
DoVariable(
	ADBHandle_p	ADBhandle,
	FILE		*file)
{
    char	var_buffer[ADB_KEYLENGTH_MAX + 1];
    char	value[ADB_KEYLENGTH_MAX + 1];	/* not right, but it will do */
    
    if ((ReadVar(var_buffer, sizeof(var_buffer), file) == 0)
	    || (getc(file) != '=')) {
	ADBerrno = ADBE_FILE;
    } else {
	GetValue(value, sizeof(value), file);
	if (getc(file) != '\0') {
	    ADBerrno = ADBE_FILE;
	}
    }

    if (ADBerrno == ADBE_NOERROR) {
	(void) ADBNewVar(ADBhandle, var_buffer, value);
    }
} /* DoVariable() */

/*-------------------------------------------------------------------------
 *  Function	: ADBRead
 *  Purpose	: Read a compiled ADB from a stream
 *  Method	: Read && parse
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *
 *  Returns	: indication of success
 *-------------------------------------------------------------------------
 */
static bool
ADBRead(
	ADBHandle_p	ADBhandle,
	FILE		*file)
{
    int	withcount = 0;
    int	c;

    do {
	switch (c = getc(file)) {
	case 'R':   DoRecord(ADBhandle, file);
		    withcount++;
		    break;
	case 'V':   DoVariable(ADBhandle, file);
		    break;
	case 'E':   (void) ADBWithEnd(ADBhandle);
		    withcount--;
		    break;
	case EOF:   break;
	default:    ADBerrno = ADBE_FILE;
	}
    } while ((c != EOF) && (ADBerrno == ADBE_NOERROR));

    if ((ADBerrno == ADBE_NOERROR) && (withcount != 0)) {
	ADBerrno = ADBE_FILE;
    }
    return (ADBerrno == ADBE_NOERROR);
} /* ADBRead() */

/*-------------------------------------------------------------------------
 *  Function	: ADBReadFile
 *  Purpose	: Read a compiled ADB from file.
 *  Method	: Obvious
 *  
 *  History:
 *  15-11-1993	: eck      : creation
 *
 *  Description :
 *
 *  ADBHandle must be a valid handle (returned by
 *  ADBCreate(ADBT_FULLMEM)).  The ADB need not be empty.
 *
 *  A call to this function between an ADBWithBegin() and an
 *  ADBWithEnd() may yield unexpected results, since this will insert
 *  the contents of a file in a single variable (though this may be
 *  what is wanted).
 *
 *  ADBfile must be the name of a compiled ADB.
 *
 *  The function returns TRUE if all went well.  When the function
 *  returns FALSE, ADBerrno will contain the error.
 *
 *  Example     :
 * 
 * 	#include <adb.h>
 *	.
 *	ADBHandle_p	conn_adb;
 *	
 *	conn_adb = ADBCreate(ADBT_FULLMEM);
 *	if (conn_adb == NULL) {
 *		ERWrite(ADBerrno, jobname, "creating adb");
 *		return;
 *	}
 *	if (!ADBReadFile(conn_adb, "adb")) {
 *		ERWrite(ADBerrno, jobname, "reading adb-file");
 *		return;
 *	}
 *	.
 *
 *  Returns	: indication of success
 *-------------------------------------------------------------------------
 */
bool
ADBReadFile(
	ADBHandle_p	ADBhandle,	/* The ADB to fill */
	const char	*ADBfile)	/* The file to read from */
{
    FILE	*file;

    ADBerrno = ADBE_NOERROR;

    file = fopen(ADBfile, "r");
    if (file == NULL) {
	ADBerrno = ADBE_IO;
	return FALSE;
    }
    (void) ADBRead(ADBhandle, file);
    if (fclose(file) == EOF) {
	if (ADBerrno == ADBE_NOERROR) ADBerrno = ADBE_IO;
    }
    return (ADBerrno == ADBE_NOERROR);
} /* ADBReadFile() */
