/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbstack.c
 *  Description	: Handling of the with-stack
 *  Author(s)	: Hans van Eck
 *  Creation	: 11-11-1993
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *========================================================================
 */

#pragma ident "$Id: adbstack.c 295831 2013-01-29 11:28:09Z helspere $"

#include <stdlib.h>
#include <string.h>
#include <pica.h>
#include <adb.h>
#include "adblocal.h"

/*------------------------------------------------------------------------
 *  Function	: ADBLNewHandle
 *  Purpose	: Create a new ADB-handle with a specific type
 *  Method	: - allocate & initialise
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *  Notes	:
 *  An ADB-handle is in reality a pointer to a stackhead.
 *
 *  Returns	: pointer to the allocated structure or NULL
 *------------------------------------------------------------------------
 */
ADBHandle_p
ADBLNewHandle(
	ADBType_t	ADBtype)	/* type of the ADB */
{
    ADBHandle_p		adb_p;
    
    adb_p = malloc(sizeof(* adb_p));

    if (adb_p != NULL) {
	adb_p->sh_first = NULL;
	adb_p->sh_last = NULL;
	adb_p->sh_type = ADBtype;

	if (ADBtype == ADBT_FULLMEM) {
	    adb_p->sh_adb = NULL;
	} else {
	    /* don't know yet what to do */
	    ADBerrno = ADBE_INTEGRITY;	/* What else ? */
	    free(adb_p);
	    adb_p = NULL;
	}
    } else {
	ADBerrno = ADBE_NOMEM;
    }
    return adb_p;
} /* ADBLNewHandle() */

/*------------------------------------------------------------------------
 *  Function	: ADBLFreeHandle
 *  Purpose	: Deallocate an ADB handle (actually a stack-head)
 *  Method	: obvious
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *  18-11-1993	: eck      : clear pointers to catch illegal use
 *------------------------------------------------------------------------
 */
void
ADBLFreeHandle(
	ADBHandle_p	ADBhandle)	/* The ADB to deallocate */
{
    ADBhandle->sh_adb = NULL;	
    ADBhandle->sh_first = NULL;
    ADBhandle->sh_last = NULL;
    free(ADBhandle);
} /* ADBLFreeHandle() */

/*------------------------------------------------------------------------
 *  Function	: ADBLNewStackelt
 *  Purpose	: Allocate a new stack-element
 *  Method	: malloc & initialise
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *  Returns	: pointer to the stack-element or NULL
 *------------------------------------------------------------------------
 */
struct adb_stackelt_t *
ADBLNewStackelt(
	const char	*path,		/* The argument of the with */
	adb_eltlist_p	*adb_pp)	/* The address of the list pointer */
{
    struct adb_stackelt_t	*eltp;

    eltp = malloc(sizeof(*eltp));

    if (eltp != NULL) {
	eltp->se_string = malloc(strlen(path) + 1);
	if (eltp->se_string == NULL) {
	    free(eltp);
	    eltp = NULL;
	}
    }
    if (eltp != NULL) {
	eltp->se_next = NULL;
	eltp->se_prev = NULL;
	/* cast to get temporarily of const */
	(void) strcpy((char *)eltp->se_string, path);
	eltp->se_eltlist_pp = adb_pp;
    }
    if (eltp == NULL) {
	ADBerrno = ADBE_NOMEM;
    }

    return eltp;
} /* ADBLNewStackelt() */

/*------------------------------------------------------------------------
 *  Function	: ADBLFreeStackelt
 *  Purpose	: Free an element of a with-stack
 *  Method	: obvious
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *  18-11-1993	: eck      : clear pointers to catch illegal use
 *------------------------------------------------------------------------
 */
void
ADBLFreeStackelt(
	struct adb_stackelt_t *selt_p)	/* Stack element to deallocate */
{
    free((void *)selt_p->se_string);
    selt_p->se_string = NULL;
    selt_p->se_next = NULL;
    selt_p->se_prev = NULL;
    free(selt_p);
} /* FreeStackElt() */
