/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: adbwith.c
 *  Description	: Handling of ABDWith* statements.
 *  Author(s)	: Hans van Eck
 *  Creation	: 11-11-1993
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *========================================================================
 */
#pragma ident "$Id: adbwith.c 295831 2013-01-29 11:28:09Z helspere $"

#include <pica.h>
#include <adb.h>
#include "adblocal.h"

/*------------------------------------------------------------------------
 *  Function	: ADBWithBegin
 *  Purpose	: Begin a with-statement
 *  Method	: - Find the variable for the with
 *		: - Add an element to the with-stack
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *  Notes	:
 *
 *  Description :
 *
 *  ADBHandle must be a valid handle (returned by ADBCreate()).
 *
 *  ADBVar must be a valid variable reference.
 *
 *  This function will create a with-context (within the current one,
 *  if possible).  Later variable references can then omit the part
 *  specified in ADBVar when specifying fields.
 *
 *  The function returns TRUE if the context could be set.  If not,
 *  FALSE is returned and ADBerrno will contain the error.
 *
 *  A with-context can be deleted with a call to ADBWithEnd().
 *  With-contexts can be nested.  When referencing a variable, they
 *  will be tried in reverse order (the last with-context created is
 *  tried first).
 *
 *  Example     :
 *
 *	#include <adb.h>
 *
 *	ADBHandle_p		cur_adb;
 *
 *	cur_adb = ADBCreate(ADBT_FULLMEM);
 *	ADBNewVar(cur_adb, "set", NULL);
 *	ADBWithBegin(cur_adb, "set[:last]");
 *	.
 *	.
 *	ADBWithEnd(cur_adb);
 *	if (!ADBDelVar(cur_adb, "set[:last]")) {
 *	ERWrite(ADBerrno, jobname, "deleting ADB-var");
 *	}
 * 
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
ADBWithBegin(
	ADBHandle_p	ADBhandle,	/* The handle to the ADB */
	const char	*ADBvar)	/* The variable for the with */
{
    adb_elt_p	*elt_pp;

    ADBerrno = ADBE_NOERROR;

    /* Find address of pointer to variable */
    elt_pp = ADBLGetElt(ADBhandle->sh_last->se_eltlist_pp, ADBvar);

    if ((elt_pp != NULL) && (*elt_pp != NULL)) {
	if ((*elt_pp)->ae_type != ADBELT_RECORD) {
	    ADBerrno = ADBE_TYPE;
	} else {
	    struct adb_stackelt_t *selt_p;

	    /* Allocate & initialise stack-element */
	    selt_p = ADBLNewStackelt(ADBvar, &((*elt_pp)->ae_record));

	    if (selt_p != NULL) {
		/* append the element */
		ADBhandle->sh_last->se_next = selt_p;
		selt_p->se_prev = ADBhandle->sh_last;
		ADBhandle->sh_last = selt_p;
	    }
	}
    }
    return ADBerrno == ADBE_NOERROR;
} /* ADBWithBegin() */

/*------------------------------------------------------------------------
 *  Function	: ADBWithEnd
 *  Purpose	: End a with statement
 *  Method	: delete the last stack element (if possible)
 *
 *  History:
 *  11-11-1993	: eck      : creation
 *
 *  Description :
 *
 *  ADBHandle must be a valid handle (returned by ADBCreate()).
 *
 *  This function will end the last with-context created.  It must
 *  match a ADBWithBegin().
 *
 *  The function returns TRUE if the context could be set.  If not,
 *  FALSE is returned and ADBerrno will contain the error.
 *
 *  Example     :
 *  See the example at ADBWithBegin().
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
ADBWithEnd(
	ADBHandle_p	ADBhandle)	/* The handle to the ADB */
{
    ADBerrno = ADBE_NOERROR;

    /* Test if there is there a with */
    if (ADBhandle->sh_first == ADBhandle->sh_last) {
	ADBerrno = ADBE_WITHEND;
    } else {
	struct adb_stackelt_t *selt_p;

	selt_p = ADBhandle->sh_last;
	ADBhandle->sh_last = selt_p->se_prev;
	ADBhandle->sh_last->se_next = NULL;

	ADBLFreeStackelt(selt_p);
    }
    return ADBerrno == ADBE_NOERROR;
} /* ADBWithEnd() */
