<!--
!
!  Copyright (c) 2013 by OCLC
!
!  File		: lib_adb.md
!  Description	: General description of the ADB library.
!  Author(s)	: Eric Helsper
!  Creation	: 23-01-2013
!
!  History:
!  23-01-2013	: eric     : creation
!
!  SVN ident	: $Id: lib_adb.md 298022 2013-04-05 14:11:47Z sieburgw $
-->
The ADB library
===============

Attribute Database Structure
----------------------------

The logical structure of an attribute database is:

    ATTRIBUTE_DATABASE	::= <ATTRIBUTE>*

    ATTRIBUTE		::= <VARIABLE> '=' <CONTENTS>

    VARIABLE		::= <NAME> [ <INDEX> ]? [ '.' <FIELD> [ <INDEX> ]? ]*

    INDEX		::= '['	<FIELD> '=' <SIMPLE_CONTENTS>
					    | <INTEGER_STRING>
					    | <LAST_INDEX>
			    ']'

    NAME		::=  <letter> [ <letter> | <digit> | '_' | '-' ]*

    FIELD		::= <NAME>

    INTEGER_STRING	::= <digit> [ <digit> ]*

    CONTENTS		::= <zero terminated string>

    SIMPLE_CONTENTS	::= <string not containing '\0' or ']'>

    LAST_INDEX		::= ":last"

In other words:

- An attribute database (ADB for short) consists of variables and values.
- Variables are structured, consisting of a name and (possibly) fields.  Both
  name and fields may be indexed.
- Conceptually, an ADB contains arrays of either records or values.  Each
  record contains fields that are arrays of either subrecords or values.
- An index can be an integer(-string), a field/value combination or last-index
  indication.
- Values are zero-terminated strings, though there is an extra restriction
  when a value is to be used in an index.

### Notes

- The physical implementation of an ADB is not described in this document.
  The ADB interface will hide the details.
- All names and field are indexed, a missing index meaning "[0]" which is the
  first element.
- An ADB is specified in user readable format, and must be preprocessed and
  compiled into a low-level ADB-format.
- When specifying :last as index, a newly created variable or field can easily
  be referenced without additional bookkeeping in a program.

### Example:

A possible variable specification is:

    server[id=pica.nl].connection[1].protocol

The server-array is indexed with a field named "id", which must be "pica.nl".
From this server the second connection protocol (indexes start at 0) is
referenced.

When there is only one field "x" in a variable "var", then by definition the
value of "var[x=y].x" is "y".
 
An example ADB
--------------

This is an example of an attribute database.  It tries to show how the
different routines and components should interact.  The names in this example
are fictional, but there should be some convention for filename extensions.

File servers.spec:

    server {
	    id = "pica.nl";
	    connection {
		    protocol	= "tcp/ip"
		    address	= "zoef@pica.nl"
		    port	= 1027
		    port	= 1028	/* alternative port */
	    }
    }

The result after 'compilation' (file servers.adb):

    Rserver\0Vid=pica.nl\0Rconnection\0Vprotocol=tcp/ip\0Vaddress=zoef@pica.nl\0Vport=1027\0Vport=1028\0EE

For reading such a compiled adb, the following algorithm is used:

    withcount = 0;
    do {
	    switch(c = getchar()) {
	    case 'R':	getstring(string);
			    ADBNewVar(adb, string, NULL);
			    strcat(strcat, "[:last]");
			    ADBWithBegin(adb, string);
			    withcount++;
			    break;
	    case 'V':	getstring(string);
			    value = strchr(string, '=');
			    if (value != NULL) {
			    crash("expected '=' in %s\n", string);
			    }
			    value++;
			    ADBNewVar(adb, variable, value);
			    break;
	    case 'E':	ADBWithEnd(adb);
			    withcount--;
			    break;
	    case EOF:	break;
	    default:	crash("read char %d (decimal)", c);
	    }
    } while ((c != EOF) && (ADBerrno == ADBE_NOERROR));

    assert(withcount == 0);		/* ADB-compiler wrong ??? */

The calls from ADBReadFile(adb, "server.adb") will then be:

    ADBNewVar(adb, "server", NULL);
    ADBWithBegin(adb, "server[:last]");
	    ADBNewVar(adb, "id", "pica.nl");
	    ADBNewVar(adb, "connection", NULL);
	    ADBWithBegin(adb, "connection[:last]");
		    ADBNewVar(adb, "protocol", "tcp/ip");
		    ADBNewVar(adb, "address", "zoef@pica.nl");
		    ADBNewVar(adb, "port", "1027");
		    ADBNewVar(adb, "port", "1028");
	    ADBWithEnd(adb);	/* "connection" */
    ADBWithEnd(adb);	/* "server" */


Port-numbers can then be retrieved with (note that these examples will not
always return the same port):

    port_string = ADBGet(adb, "server.connection.port");

or

    port_string = ADBGet(adb, "server[id=pica.nl].connection[0].port[1]");

or

    port_string = ADBGet(adb,"server[0].connection[protocol=tcp/ip].port");

It is also possible to use the with:

    if (ADBWithBegin(adb,"server[id=pica.nl]") {
	    port_string = ADBGet(adb, "connection.port");
	    ADBWithEnd(adb);
    }

or

    if (ADBWithBegin(adb,"server[id=pica.nl]") {
	    if (ADBWithBegin(adb, "connection[protocol=tcp/ip]") {
		    port_string0 = ADBGet(adb, "port[0]");
		    port_string1 = ADBGet(adb, "port[1]");
		    ADBWithEnd(adb);
	    }
	    ADBWithEnd(adb);
    }

or any other method that the application/library requires.
 
Implementation details
----------------------

This chapter contains some details of the internal data-structure of the ADB-library.

An ADB-library contains 4 basic data-structures. Two of these data-structures
are necessary for implementing the 'with'-statements, and two are used for
describing the data.

An ADBHandle_p is simply a pointer to a data-structure (called the stack-head) containing:
- pointers to the first and last stack-element
- the type of the ADB (currently only FULLMEM)
- a pointer to the element-list (described later)

Each stack-element contains:
- pointers to previous and next element
- the string of the 'with'-statement
- a pointer to the pointer to the element-list

When a 'with'-statement is encountered, a new stack-element is pushed on the
stack, and the pointer to the last stack-element in the stack-head is
changed. Every reference to the ADB will use this 'last' pointer, thus
implementing a scope. The pointer to the next stack-element and the string of
the 'with'-statement are only present for easier debugging.

The implementation uses the 'pointer-to-pointer' concept. If you are not
familiar with this, don't change the code.

Data is described using the element-list and the element data-structures. The
element-list is a linked-list. Each element-list structure contains:
- (a pointer to) the name of the element
- a pointer to the element-structure
- a pointer to the next element-list structure

Each element structure contains:
- the type of the element (RECORD or VALUE)
- the value of the element, which is:
	* a string when the element-type is VALUE
	* a pointer to an element-list when the element-type is RECORD
- a pointer to the next element structure with the same name

Thus, array indexing is done by looking up the name in the element-list, and
following the elements until the appropriate one is found.

In the image, you can see how the 'Example ABD' above looks
internally. (The strings are actually implemented as pointers).

*image still to be copied from Word file*
