/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: cball.c
 *  Description	: All the Counted-Buffer routines
 *  Author(s)	: Hans van Eck
 *  Creation	: 13-09-1993
 *
 *  History:
 *  13-09-1993 : eck      : creation
 *  03-11-1993 : eck      : conformed interface to specification
 *  21-12-1993 : clemens  : CBLength() added
 *  14-10-1994 : eck      : Added CB*Attribute() routines (were in TB-libary)
 *  17-10-1994 : eck      : Removed CBE_* errors (except CBE_NOERROR)
 *  02-12-1994 : HSi      : Modified interface of CB*Attribute routines and
 *			  : added CBDelAttribute and CBUpdAttribute.
 *  16-12-1994 : eck      : Modified CB*Attribute() interfaces: The length
 *			  : is implicit now, since they all expect strings.
 *  06-06-1995 : eck      : Initialise value to NULL in CB[Get|Next]Attribute()
 *  25-11-1996 : lolke    : CBAppToAttribute added.
 *  29-11-1996 : lolke    : UpdateLength added and used in CBUpdElement,
 *			  : CBUpdAttribute and CBAppToAttribute.
 *  05-08-1998 : paul     : bugfix in OkSize()
 *  06-08-1998 : paul     : bugfix in GetLength()
 *  28-04-2000	: eck      : Added CBIsValid(), removed some typecasts.
 *  01-05-2000	: eck      : Added CBGetOptAttribute().
 *  11-04-2011	: paul     : CBS-2850 preparation: no functional change
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: cball.c 339156 2016-11-18 18:26:14Z heesakkm $"

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

#include <pica.h>
#include <cb.h>

/* The #define of CB_LENGTH_SIZE can not be changed! */
#define CB_LENGTH_SIZE	2	/* 2 bytes for length */

/* This may be removed later, when there is no reference anymore to
 * CBerrno.
 */
ERROR_DEFINE(CBE_NOERROR);

/* Here we define CBerrno. There is a declaration in <cb.h> */
CBErrno_t CBerrno = CBE_NOERROR;

/*------------------------------------------------------------------------
 *  Function	: OkSize
 *  Purpose	: Check if a given size is all right
 *  Method	: trivial
 *
 *  History:
 *  03-11-1993	: eck      : creation
 *  05-08-1998	: paul     : (CBsize / (UCHAR_MAX + 1)) <= UCHAR_MAX instead of
 *                           (CBsize % (UCHAR_MAX + 1)) <= UCHAR_MAX
 *
 *  Notes	:
 *  CBsize should be at least larger than CB_LENGTH_SIZE, and it must fit in
 *  CB_LENGTH_SIZE bytes.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
OkSize(
	const size_t     CBsize)
{
    if (CBsize >= CB_LENGTH_SIZE) {
        return (CBsize / (UCHAR_MAX + 1)) <= UCHAR_MAX;
    } else {
        return FALSE;
    }
} /* OkSize() */

/*------------------------------------------------------------------------
 *  Function    : OkLength
 *  Purpose     : Check whether a length is ok given the size of the buffer
 *  Method      :
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *
 *  Notes       :
 *  The CBsize should be ok, and the length should be larger than
 *  CB_LENGTH_SIZE, and it smaller than CBsize.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
static bool
OkLength(
        const size_t    CBsize, /* the size of a CB */
        const size_t    length) /* length of one or all elt(s) in the CB */
{
    if (OkSize(CBsize)) {
        return (length >= CB_LENGTH_SIZE) && (length <= CBsize);
    } else {
        return FALSE;
    }
} /* OkLength() */

/*------------------------------------------------------------------------
 *  Function    : PutLength
 *  Purpose     : Set the langth part of a buffer
 *  Method      :
 *  Author(s)   : Hans van Eck
 *  Creation    : 03-11-1993
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *
 *  Notes       :
 *  This routine is declared static, so programmers can not use it.
 *------------------------------------------------------------------------
 */
static void
PutLength(
        byte            *bytep, /* ptr to where length should be set */
        const size_t    length) /* the lenghth to set */
{
    bytep[0] = length % (UCHAR_MAX + 1);
    bytep[1] = length / (UCHAR_MAX + 1);
} /* PutLength() */

/*------------------------------------------------------------------------
 *  Function    : GetLength
 *  Purpose     : Return the current length of a counted buffer
 *  Method      : Add the first byte to 256 * second byte
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  02-11-1993  : eck      : creation
 *  06-08-1998  : paul     : bytep[0] + ((UCHAR_MAX + 1) * bytep[1]) changed to
 *                   (size_t)bytep[0] + ((UCHAR_MAX + 1) * (size_t)bytep[1])
 *
 *  Returns     : Length of counted buffer
 *------------------------------------------------------------------------
 */
static size_t
GetLength(
	const byte	*bytep)
{
    return (size_t)bytep[0] + ((UCHAR_MAX + 1) * (size_t)bytep[1]);
} /* GetLength() */

/*------------------------------------------------------------------------
 *  Function    : CBLength
 *  Purpose     : Retrieve current length of a counted buffer
 *  Method      : GetLength()
 *  Author(s)   : Clemens Buijs
 *
 *  History:
 *  21-12-1993  : clemens    : creation
 *
 *  Notes       : CBbuf must point to a counted buffer.
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t	CBbuf[80];
 *	size_t	length;
 *	.
 *	length = CBLength(CBbuf);
 *	.
 *
 *  Returns     :
 *  The current length of the counted buffer (the value of BUFFER_LENGTH)
 *------------------------------------------------------------------------
 */
size_t
CBLength(
	const CBBuf_t  *CBbuf)         /* the counted buffer */
{
    return GetLength(CBbuf);
} /* CBLength() */

/*------------------------------------------------------------------------
 *  Function    : CBSizeLeft
 *  Purpose     : Retrieve the available space left in a counted buffer
 *  Method      : return size - current length
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  02-11-1993  : eck      : creation
 *
 *  Notes       :
 *  - CBbuf must point to a counted buffer.
 *  - CBsize is the size of CBbuf.
 *  - The size left is not the amount of space left for CBAppElement()!
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t	CBbuf[80];
 *	size_t	value;
 *	.
 *	value = CBSizeLeft(CBbuf, sizeof(CBbuf));
 *	.
 *
 *  Returns     : Size left in the counted buffer
 *------------------------------------------------------------------------
 */
size_t
CBSizeLeft(
        const CBBuf_t   *CBbuf,         /* the CB */
        size_t          CBsize)         /* the size of the CB */
{
    return CBsize - GetLength(CBbuf);
} /* CBSizeLeft() */

/*------------------------------------------------------------------------
 *  Function    : CBElementLength
 *  Purpose     : Determine the length of an element in a counted buffer
 *  Method      : Pointer juggling
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  02-11-1993  : eck      : creation
 *
 *  Notes       : CBelm must point to an element value in the counted buffer.
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBbuf_t	CBbuf[1024];
 *	CBElm_p	CBelm;
 *	size_t	value;
 *
 *	.
 *	CBelm = CBGetElement(CBbuf, sizeof(CBbuf), 0);
 *	if (CBelm == NULL) {
 *	    ERWrite(CBerrno, fname, "CBGetElement() failed");
 *	    return FALSE;
 *	}
 *	value = CBGetElementLength(CBbuf, sizeof(CBbuf), CBelm);
 *	.
 *
 *  Returns     :
 *  The calculated length of the pointed element (the length parameter itself
 *  EXCLUDED)
 *------------------------------------------------------------------------
 */
size_t
CBElementLength(
        CBElm_p         CBelm)  /* the element to determine the length of */
{
    return GetLength(((byte *) CBelm) - CB_LENGTH_SIZE) - CB_LENGTH_SIZE;
} /* CBElementLength() */

/*------------------------------------------------------------------------
 *  Function    : CBGetElement
 *  Purpose     : Retrieve the pointer to the n-th element in a counted buffer
 *  Method      : Skip element until right one found
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *  28-04-2000	: eck      : Added check on zero element length for robustness
 *
 *  Notes       :
 *  - CBelm must point to an element value in the counted buffer.
 *  - CBsize is the size of CBbuf.
 *  - seqnum is the sequence number of the element to be retrieved.
 *  - The first element has sequence number zero (0).
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBbuf_t	CBbuf[1024];
 *	CBElm_p	CBelm;
 *	size_t	value;
 *
 *	.
 *	CBelm = CBGetElement(CBbuf, sizeof(CBbuf), 0);
 *	if (CBelm == NULL) {
 *	    ERWrite(CBerrno, fname, "CBGetElement() failed");
 *	    return FALSE;
 *	}
 *	value = CBGetElementLength(CBbuf, sizeof(CBbuf), CBelm);
 *	.
 *
 *  Returns     :
 *  Pointer to the selected element value or NULL if the element was not
 *  located (and at erroneous situations)
 *------------------------------------------------------------------------
 */
CBElm_p
CBGetElement(
        CBBuf_t         *CBbuf,         /* the CB to get the element from */
        size_t          CBsize,         /* the size of the CB */
        unsigned        seqnum)         /* sequence number of the element */
{
    byte        *freep =  CBbuf + GetLength(CBbuf);
    byte        *elmheadp = CBbuf + CB_LENGTH_SIZE;

    (void) CBsize;	// Keep lint happy

    while ((seqnum != 0) && (elmheadp < freep)) {
	size_t elmsiz = GetLength(elmheadp);

	if (elmsiz == 0) return NULL;
        elmheadp += elmsiz;
        seqnum--;
    }
    if (elmheadp >= freep) {            /* the length is exhausted */
        return NULL;
    }
    return elmheadp + CB_LENGTH_SIZE;
} /* CBGetElement() */

/*------------------------------------------------------------------------
 *  Function    : CBNextElement
 *  Purpose     : Retrieve the pointer to the next element
 *  Method      : Pointer juggling
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *
 *  Notes       :
 *  - CBbuf must point to a counted buffer.
 *  - CBsize is the size of CBbuf.
 *  - CBelm must point to the previous element value.
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBbuf_t	CBbuf[80];
 *	CBElm_p	CBelm;
 *
 *	.
 *	CBelm = CBNextElement(CBbuf, sizeof(CBbuf), CBelm);
 *	if (CBelm == NULL) {
 *	    ERWrite(CBerrno, fname, "CBNextElement() failed");
 *	    return FALSE;
 *	}
 *	.
 *
 *  Returns     :
 *  Pointer to the selected element value or NULL if the next element was not
 *  located (and at erroneous situations)
 *------------------------------------------------------------------------
 */
CBElm_p
CBNextElement(
        CBBuf_t         *CBbuf,         /* the CB to get the element from */
        size_t          CBsize,         /* the size of the CB */
        CBElm_p         CBelm)          /* the current element value */
{
    byte        *freep = CBbuf + GetLength(CBbuf);
    byte        *elmheadp = ((byte *) CBelm) - CB_LENGTH_SIZE;

    (void) CBsize;	// Keep lint happy

    elmheadp += GetLength(elmheadp);
    if (elmheadp >= freep) {            /* the length is exhausted */
        return NULL;
    }
    return elmheadp + CB_LENGTH_SIZE;
} /* CBNextElement() */

/*------------------------------------------------------------------------
 *  Function    : CBClear
 *  Purpose     : Initialize a buffer with an empty length
 *  Method      :
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *
 *  Notes       :
 *  - CBbuf must point to a counted buffer.
 *  - CBsize is the size of CBbuf.
 *
 *  Description :
 *  This function should always be called first when elements are written into
 *  the counted buffer.
 *
 *  If this function returns FALSE, something terrible has happened and the
 *  caller better takes some harsh measurements.
 *
 *  This function calls OkSize(), so the size must not exceed 65535.
 *
 *  This function will set the length of the counted buffer to CB_LENGTH_SIZE
 *  and returns FALSE in erroneous situations (in which case the counted buffer
 *  is left unchanged).
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t	CBbuf[256];
 *	.
 *	if (!CBClear(CBbuf)) {
 *	    ERWrite(CBerrno, fname, "CBClear() failed");
 *	    return FALSE;
 *	}
 *	.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBClear(
        CBBuf_t         *CBbuf,         /* the CB to clear */
        size_t          CBsize)         /* the size of the CB */
{
    CBerrno = CBE_NOERROR;
    if (OkSize(CBsize)) {
        PutLength(CBbuf, CB_LENGTH_SIZE);
        return TRUE;
    } else {
        /* This is really bad! */
        return FALSE;
    }
} /* CBClear() */

/*------------------------------------------------------------------------
 *  Function    : UpdateLength
 *  Purpose     : Set length of buffer and element
 *  Method      :
 *  Author(s)   : Lolke Boonstra
 *  Creation    : 29-11-1996
 *
 *  History:
 *  29-11-1996  : lolke    : creation
 *
 *  Notes       :
 *  Also makes room for new element value.
 *  Used in CBUpdElement(), CBUpdAttribute() and CBAppToAttribute()
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
static bool
UpdateLength(
        CBBuf_t         *CBbuf,         /* CB to add an element to */
        size_t          CBsize,         /* the size of the CB */
        CBElm_p         CBelm,          /* the element to replace */
        size_t          elmsiz)         /* the size of new element */
{
    size_t      buflen = GetLength(CBbuf);
    size_t      oldelmsiz = CBElementLength(CBelm);
    byte        *elmheadp = ((byte *) CBelm) - CB_LENGTH_SIZE;

    CBerrno = CBE_NOERROR;
    if (! OkSize(CBsize)) {
        /* crash() */
        return FALSE;
    } else if (! OkLength(CBsize, elmsiz + CB_LENGTH_SIZE)) {
        /* crash() */
        return FALSE;
    } else if ((oldelmsiz < elmsiz)
                && ((elmsiz - oldelmsiz) > CBSizeLeft(CBbuf, CBsize))) {
        return FALSE;
    } else {
        CBElm_p nextelmp = CBNextElement(CBbuf, CBsize, CBelm);

        /* Move the element after the one to update (if there is one). */
        if (nextelmp != NULL) {
            byte        *from = ((byte *) nextelmp) - CB_LENGTH_SIZE;
            byte        *to = from + elmsiz - oldelmsiz;
            size_t      len = buflen - (from - CBbuf);

            (void) memmove(to, from, len);
        }
        PutLength(CBbuf, buflen + elmsiz - oldelmsiz);
        PutLength(elmheadp, elmsiz + CB_LENGTH_SIZE);

        return TRUE;
    }
} /* UpdateLength() */

/*------------------------------------------------------------------------
 *  Function    : CBAppElement
 *  Purpose     : Append an element to a counted buffer
 *  Method      : trivial
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  02-11-1993  : eck      : creation
 *
 *  Notes       :
 *  - CBbuf must point to a counted buffer.
 *  - CBsize is the size of CBbuf.
 *  - elmval must point to the object to be appended as new element.
 *  - elmlen is the length of the object to be appended.
 *
 *  The counted buffer where the elements will be appended to, should first
 *  be cleared, using CBClear().
 *
 *  Description :
 *  This function appends the object to the counted buffer (and updates the
 *  buffer length) and returns FALSE in erroneous situations (like buffer
 *  overflow, in which case the counted buffer is left unchanged).
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t	CBbuf[256];
 *	char	newval[30];
 *	.
 *	(void) PM_strncpy(newval, "new text", sizeof(newval));
 *
 *	if (!CBClear(CBbuf)) {
 *	    ERWrite(CBerrno, fname, "CBClear() failed");
 *	    return FALSE;
 *	}
 *	if (!CBAppElement(CBbuf, sizeof(CBbuf), newval, strlen(newval)+1)) {
 *	    ERWrite(CBerrno, fname, "CBAppElement() failed");
 *	    return FALSE;
 *	}
 *	.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBAppElement(
        CBBuf_t         *CBbuf,         /* CB to add an element to */
        size_t          CBsize,         /* the size of the CB */
        voidp           elmval,         /* the element value to add */
        size_t          elmsiz)         /* the length of the element */
{
    size_t      curlen = GetLength(CBbuf);
    byte        *freep = CBbuf + curlen;
    size_t      newlen;

    CBerrno = CBE_NOERROR;
    newlen = curlen + elmsiz + CB_LENGTH_SIZE;

    if (! OkSize(CBsize)) {
        /* crash(); */
        return FALSE;
    } else if (! OkLength(CBsize, curlen)) {
        /* crash(); */
        return FALSE;
    } else if (! OkSize(elmsiz + CB_LENGTH_SIZE)) {
        return FALSE;
    } else if (! OkLength(CBsize, newlen)) {
        return FALSE;
    } else {
        PutLength(CBbuf, newlen);
        PutLength(freep, elmsiz + CB_LENGTH_SIZE);
        (void) memmove(freep + CB_LENGTH_SIZE, elmval, elmsiz);
        return TRUE;
    }
} /* CBAppElement() */

/*------------------------------------------------------------------------
 *  Function    : CBDelElement
 *  Purpose     : Delete an element from a counted buffer
 *  Method      : trivial
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *
 *  Notes       :
 *  - CBbuf must point to a counted buffer.
 *  - CBsize is the size of CBbuf.
 *  - CBelm must point to a valid element in the counted buffer.
 *
 *  Description :
 *  This function deletes the element from the counted buffer (and updates the
 *  buffer length) and returns FALSE in erroneous situations (like 'element not
 *  found', in which case the counted buffer is left unchanged).
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t	CBbuf[256];
 *	CBElm_p	CBelm;
 *	.
 *	if (!CBClear(CBbuf)) {
 *	    ERWrite(CBerrno, fname, "CBClear() failed");
 *	    return FALSE;
 *	}
 *	.
 *	if (!CBDelElement(CBbuf, sizeof(CBbuf), CBelm)) {
 *	    ERWrite(CBerrno, fname, "CBDelElement() failed");
 *	    return FALSE;
 *	}
 *	.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBDelElement(
        CBBuf_t         *CBbuf,         /* CB to delete an element from */
        size_t          CBsize,         /* the size of the CB */
        CBElm_p         CBelm)          /* the element to delete */
{
    size_t      buflen = GetLength(CBbuf);
    byte        *elmheadp = ((byte *) CBelm) - CB_LENGTH_SIZE;
    size_t      dellen = GetLength(elmheadp);

    CBerrno = CBE_NOERROR;

    if (! OkSize(CBsize)) {
        /* crash(); */
        return FALSE;
    } else if (! OkLength(CBsize, dellen)) {
        /* crash(); */
        return FALSE;
    } else if (dellen > buflen) {
        /* crash(); */
        return FALSE;
    } else {
        PutLength(CBbuf, buflen - dellen);
        (void) memmove(elmheadp, elmheadp + dellen,
                        buflen - (elmheadp + dellen - CBbuf));
        return TRUE;
    }
} /* CBDelElement() */

/*------------------------------------------------------------------------
 *  Function    : CBUpdElement
 *  Purpose     : Update an element in a counted buffer
 *  Method      : trivial
 *  Author(s)   : Hans van Eck
 *
 *  History:
 *  03-11-1993  : eck      : creation
 *
 *  Notes       :
 *  - CBbuf must point to a counted buffer.
 *  - CBsize is the size of CBbuf.
 *  - CBelm must point to a valid element in the counted buffer.
 *  - elmval must point to the object to be used as replacement for the element.
 *  - elmsiz is the size of the object.
 *
 *  Description :
 *  This function replaces the element with the object in the counted buffer
 *  (and updates the buffer length) and returns FALSE in erroneous situations
 *  (like 'element not found', in which case the counted buffer is left
 *  unchanged).
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t	CBbuf[256];
 *	CBElm_p	CBelm;
 *	char	newval[30];
 *	.
 *	(void) PM_strncpy(newval, "new text", sizeof(newval));
 *
 *	CBelm = CBGetElement(CBbuf, sizeof(CBbuf), 0);
 *	if (CBelm == NULL) {
 *	    ERWrite(CBerrno, fname, "CBGetElement() failed");
 *	    return FALSE;
 *	}
 *	if (!CBUpdElement(CBbuf, sizeof(CBbuf), CBelm, newval, strlen(newval)+1)) {
 *	    ERWrite(CBerrno, fname, "CBUpdElement() failed");
 *	    return FALSE;
 *	}
 *	.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBUpdElement(
        CBBuf_t         *CBbuf,         /* CB to add an element to */
        size_t          CBsize,         /* the size of the CB */
        CBElm_p         CBelm,          /* the element to replace */
        voidp           elmval,         /* the new element value */
        size_t          elmsiz)         /* the size of new element */
{

    CBerrno = CBE_NOERROR;
    if (UpdateLength(CBbuf, CBsize, CBelm, elmsiz)) {
        (void) memmove(CBelm, elmval, elmsiz);
        return TRUE;
    }
    return FALSE;
} /* CBUpdElement() */

/*------------------------------------------------------------------------
 *  Function    : CBGetAttribute
 *  Purpose     : Retrieve a specific attribute from a (table) record
 *  Method      : Scan counted buffer and check attribute name
 *  Author(s)   : Clemens Buys
 *  Creation    : 23-12-1993
 *
 *  Notes       :
 *  - CBbuf must be a pointer to a counted buffer containing the table record
 *    to be scanned.
 *  - CBsize is the maximum length of CBbuf.
 *  - CBfndnam is a pointer to a zero-terminated string containing the name of
 *    the attribute to be retrieved or NULL ('first attribute').
 *  - CBattvalue is a (returned) pointer to the attribute value.
 *
 *
 *  History:
 *  23-12-1993  : clemens  : creation
 *  04-01-1994  : clemens  : changed interface
 *  14-10-1994  : eck      : copied from tb-library and changed names
 *  02-12-1994  : HSi      : modified interface. A CBfndnam of NULL returns
 *                           the first attribute (if present).
 *
 *  Returns     :
 *  Pointer to attribute name or NULL (no valid attribute was detected)
 *------------------------------------------------------------------------
 */
CBElm_p
CBGetAttribute(
        CBBuf_t         *CBbuf,         /* the CB to get the attribute from */
        size_t          CBsize,         /* the size of the CB */
        const char      *CBfndnam,      /* the attribute name to look for */
        const char      **CBattvalue)   /* the return pointer to value */
{
    CBElm_p     p  = NULL;
    byte        *attnam;
    size_t      namelen;

    CBerrno    = CBE_NOERROR;
    *CBattvalue = NULL;
    p = CBGetElement(CBbuf, CBsize, 0);
    while (p != NULL) {
        attnam  = (byte *) p;

        if ( CBfndnam == NULL || STREQ(CBfndnam, (char *)attnam)) {
            namelen     = strlen ( (char *) attnam );
            *CBattvalue = (voidp) (attnam + namelen + 1);
            return p;
        }
        p = CBNextElement(CBbuf, CBsize, p);
    } /* while */
    return NULL;
} /* CBGetAttribute() */

/*------------------------------------------------------------------------
 *  Function    : CBNextAttribute
 *  Purpose     : Retrieve next attribute from a (table) record
 *  Method      : Get attribute next to the one passed as argument
 *  Author(s)   : Clemens Buys
 *  Creation    : 23-12-1993
 *
 *  History:
 *  23-12-1993  : clemens  : creation
 *  04-01-1994  : clemens  : changed interface
 *  14-10-1994  : eck      : copied from tb-library and changed names
 *  02-12-1994  : HSi      : modified interface. Funtion now returns next
 *                           attribute no matter what the name of this
 *                           attribute is.
 *
 *  Notes       :
 *  - CBbuf must be a pointer to a counted buffer containing the table record
 *    to be scanned.
 *  - CBsize is the maximum length of CBbuf.
 *  - CBprvname is a pointer to a zero-terminated string containing the name of
 *    the previous attribute (offset), as returned by CBGetAttribute() or
 *    CBNextAttribute itself.
 *  - CBattvalue is a (returned) pointer to the attribute value.
 *
 *  Description :
 *  This function will search for the next attribute in the counted buffer,
 *  using prvattname as offset.
 *
 *  Returns     :
 *  Pointer to attribute name of next element or NULL (no next attribute was
 *  detected)
 *------------------------------------------------------------------------
 */
CBElm_p
CBNextAttribute(
        CBBuf_t         *CBbuf,         /* counted buffer to search in */
        size_t          CBsize,         /* the size of the CB */
        CBElm_p         CBprvname,      /* previous name, returned by
                                         * CB[Get|Next]Attribute() */
        const char      **CBattvalue)   /* the pointer to attribute value */
{
    CBElm_p     p = NULL;
    const char  *attnam;

    CBerrno    = CBE_NOERROR;
    *CBattvalue = NULL;
    p = CBNextElement ( CBbuf, CBsize, CBprvname );
    if (p != NULL) {
        attnam       = (const char *) p;
        *CBattvalue  = attnam + strlen(attnam) + 1;
    }
    return p;
} /* CBNextAttribute() */

/*------------------------------------------------------------------------
 *  Function    : CBAppAttribute
 *  Purpose     : Append an attribute to a (table) record
 *  Method      : Construct and append counted buffer element
 *  Author(s)   : Clemens Buys
 *  Creation    : 23-12-1993
 *
 *  History:
 *  23-12-1993  : clemens  : creation
 *  14-10-1994  : eck      : copied from tb-library and changed names
 *
 *  Notes       :
 *  - CBbuf must be a pointer to a counted buffer containing the table record
 *    to be appended.
 *  - CBsize is the maximum length of CBbuf.
 *  - CBattnam is a pointer to a zero-terminated string containing the name of
 *    the attribute to be appended.
 *  - CBattval is a pointer to a zero-terminated string containing the
 *    corresponding attribute value.
 *
 *  Description :
 *  This function will append the attribute to the counted buffer and return
 *  FALSE when the append was not possible (overflow e.g.), otherwise TRUE.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBAppAttribute(
        CBBuf_t         *CBbuf,         /* the CB to append the attribute to */
        size_t          CBsize,         /* the size of the buffer */
        const char      *CBattnam,      /* name of the attribute to append */
        const char      *CBattval)      /* value of the attribute */
{
    size_t      CBattlen = strlen(CBattval) + 1;
    size_t      buf_len = GetLength(CBbuf);
    byte        *freep = CBbuf + buf_len;
    size_t      att_namlen = strlen(CBattnam) + 1;
    size_t      att_totlen = att_namlen + CBattlen;

    if ((CBsize - buf_len) < (att_totlen + CB_LENGTH_SIZE)) {
        return FALSE;
    }
    PutLength(freep, att_totlen + CB_LENGTH_SIZE);
    PutLength(CBbuf, buf_len + att_totlen + CB_LENGTH_SIZE);

    (void) memcpy(freep + CB_LENGTH_SIZE, CBattnam, att_namlen);
    (void) memcpy(freep + CB_LENGTH_SIZE + att_namlen, CBattval, CBattlen);

    return TRUE;
} /* CBAppAttribute() */

/*------------------------------------------------------------------------
 *  Function    : CBDelAttribute
 *  Purpose     : Delete an attribute from a (table) record.
 *  Method      : CBDelElement()
 *  Author(s)   : HSi
 *  Creation    : 02-12-1994
 *
 *  History:
 *  02-12-1994  : HSi   : creation
 *
 *  Notes       :
 *  - CBbuf must be a pointer to a counted buffer containing the table record
 *    to be appended.
 *  - CBsize is the maximum length of CBbuf.
 *  - CBattname is a pointer (as returned by CBGetAttribute() or
 *    CBNextAttribute()) to a zero-terminated string in the counted buffer
 *    containing the name of the attribute to be deleted.
 *
 *  Description :
 *  This function deletes the attribute from the counted buffer (and updates
 *  the buffer length) and returns FALSE in erroneous situations (like
 *  'attribute not found', in which case the counted buffer is left unchanged).
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBDelAttribute (
        CBBuf_t         *CBbuf,         /* CB to remove attribute from       */
        size_t          CBsize,         /* the size of the CB                */
        CBElm_p         CBattname)      /* the attribute (name) to delete    */
{
    return CBDelElement(CBbuf, CBsize, CBattname);
} /* CBDelAttribute() */

/*------------------------------------------------------------------------
 *  Function    : CBUpdAttribute
 *  Purpose     : Update an attribute from a (table) record
 *  Method      : Check the attribute's size
 *  Author(s)   : HSi
 *  Creation    : 02-12-1994
 *
 *  History:
 *  02-12-1994  : HSi   : creation
 *
 *  Notes       :
 *  - CBbuf must be a pointer to a counted buffer containing the table record
 *    to be appended.
 *  - CBsize is the maximum length of CBbuf.
 *  - CBattname is a pointer (as returned by CBGetAttribute() or
 *    CBNextAttribute()) to a zero-terminated string in the counted buffer
 *    containing the name of the attribute to be updated.
 *  - CBattValue must point to the new attribute value.
 *
 *  Description :
 *  This function replaces the value of the selected attribute with the new
 *  value (and updates the buffer length) and returns FALSE in erroneous
 *  situations (like 'attribute not found', in which case the counted buffer is
 *  left unchanged).
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBUpdAttribute(
        CBBuf_t         *CBbuf,         /* CB to update attribute in         */
        size_t          CBsize,         /* the size of the CB                */
        CBElm_p         CBattName,      /* name of attribute to update       */
        const char      *CBattValue)    /* the new attribute value           */
{
    /* N.B.: bufLength includes the length of the bufferLengthField.
     *       oldElmLength doesn't include the length of the elementLengthField.
     */
    size_t      CBattValueLength = strlen(CBattValue) + 1;
    size_t      oldElmLength = CBElementLength ( CBattName );
    size_t      oldValLength = oldElmLength - strlen ( (char *)CBattName ) - 1;

    CBerrno = CBE_NOERROR;
    if (UpdateLength(CBbuf, CBsize, CBattName,
		    oldElmLength + CBattValueLength - oldValLength)) {
        /* Put new value in place. */
        (void) memmove ( (byte *)CBattName + strlen ( (char *)CBattName ) + 1,
                         CBattValue,
                         CBattValueLength );
        return TRUE;
    }
    return FALSE;
} /* CBUpdAttribute() */

/*------------------------------------------------------------------------
 *  Function    : CBAppToAttribute
 *  Purpose     : Append a value to an attribute name
 *  Method      :
 *  Author(s)   : Lolke Boonstra
 *  Creation    : 26-11-1996
 *
 *  History:
 *  26-11-1996  : lolke    : creation
 *
 *  Notes	:
 *  - CBbuf must be a pointer to a counted buffer containing the table record
 *    to be appended.
 *  - CBsize is the maximum length of CBbuf.
 *  - CBattName is a pointer to a zero-terminated string containing the name of
 *    the attribute to be appended.
 *  - CBattValue is a pointer to a zero-terminated string containing the
 *    corresponding attribute value.
 *
 *  Description :
 *  This function will replace the value of the attribute if it already exists
 *  within the counted buffer or append the attribute to the counted buffer.
 *  This function returns FALSE when the required action was not possible
 *  (overflow e.g.), otherwise it returns TRUE.
 *
 *  Returns     : Indication of success
 *------------------------------------------------------------------------
 */
bool
CBAppToAttribute(
        CBBuf_t         *CBbuf,         /* the CB to append the attribute to */
        size_t          CBsize,         /* the size of the buffer */
        const char      *CBattName,     /* name of the attribute to append to */
        const char      *CBattValue)    /* value of the attribute */
{
    const char *cp;
    CBElm_p CBelm;
    size_t oldElmLength;
    size_t oldlen;
    size_t len;

    if ((CBelm = CBGetAttribute(CBbuf, CBsize, CBattName, &cp)) != NULL) {
        len = strlen(CBattValue);
        oldlen = strlen(cp);
        oldElmLength = CBElementLength ( CBelm );
        if (UpdateLength(CBbuf, CBsize, CBelm, oldElmLength + len)) {
            /* Put new value in place. */
            (void) memmove((byte *)CBelm + strlen(CBattName) + 1 + oldlen,
			    CBattValue, len + 1);
            return TRUE;
        }
        return FALSE;
    } else {
        return CBAppAttribute(CBbuf, CBsize, CBattName, CBattValue);
    }
} /* CBAppToAttribute() */

/*------------------------------------------------------------------------
 *  Function	: CBIsValid
 *  Purpose	: Check validity of a counted buffer
 *  Method	:
 *  1. Check whether the length (NOT size) of the buffer conforms to what is
 *     stored as size.
 *  2. Check whether the buffer is the sum of its' elements.
 *
 *  History:
 *  28-04-2000	: eck      : creation
 *
 *  Notes	:
 *  - CBbuf must be a pointer to the buffer to be checked.
 *  - CBlength is the actual length of CBbuf (not its size).
 *
 *  This function is intended to be used in a transitional situation where the
 *  database can have records with and without counted buffers.
 *
 *  This routine has become necessary for situations in which there might be a
 *  non-CB read from a database. This is necessary for very large records (like
 *  the WinIBW PARAM-records).
 *
 *  The CBlength in the interface is should be determined from an external
 *  source, (like the length returned by TBReadN() or TBNextN()).
 *
 *  Description :
 *  This function checks whether CBbuf actually points to a counted buffer.
 *  This is done by comparing the length of the buffer and its elements with
 *  Cblength (which should NOT be the size of the buffer, on contrast to the
 *  other interfaces).
 *
 *  Returns	: Indication of validity of the buffer
 *------------------------------------------------------------------------
 */
bool
CBIsValid(
	const CBBuf_t	*CBbuf,		/* possible counted buffer */
	size_t		CBlength)	/* the length of CBbuf */
{
    size_t	poss_len = GetLength(CBbuf);
    const byte	*elmheadp;
    const byte	*endp;

    /* If the length is wrong, it is not a valid CB. */
    if (poss_len != CBlength) return FALSE;

    /* In rare cases the first characters of a text-element miraculously are
     * the same as would be expected when it really was a CB.  For this case,
     * we check if all elements also look like CB-elements (only their lengths
     * are checked). This is still not perfect, but should be good enough.
     */
    elmheadp = CBbuf + CB_LENGTH_SIZE;
    endp = CBbuf + poss_len;

    while (elmheadp < endp) {
	size_t elmsiz = GetLength(elmheadp);

	if (elmsiz == 0) return FALSE;
        elmheadp += elmsiz;
    }
    return (elmheadp == endp);
} /* CBIsValid() */

/*------------------------------------------------------------------------
 *  Function	: CBGetOptAttribute
 *  Purpose	: Get an attribute (or buffer if not a CB)
 *  Method	:
 *  Check if the buffer is a valid CB. If not, Assume it is just text and
 *  return a pointer to the begin. Otherwise, get the attribute (if present).
 *
 *  History:
 *  01-05-2000	: eck      : creation
 *
 *  Notes	:
 *  - CBbuf must be a pointer to a buffer containing a table record just read.
 *  - CBlength is the actual size of the record.
 *  - CBattnam is a pointer to the attribute name to fetch.
 *
 *  Description :
 *  This function will check whether the buffer is actually a counted buffer.
 *  If it is, CBGetAttribute() is used to determine where the content of the
 *  attribute starts.
 *
 *  If the buffer is not a counted buffer, it is assumed to be a pure-text
 *  buffer, and a pointer to it will be returned.
 *
 *  This function can be used in a transitional situation where records with
 *  counted-buffers are replace by text-only records.
 *
 *  Example	:
 *
 *	#include <cb.h>
 *
 *	CBBuf_t		CBbuf[256];
 *	size_t		reclen;
 *	const char	*val;
 *	.
 *	if (TBReadN("mode=exact", "balbla", NULL, 0, CBbuf, sizeof(CBbuf), &reclen)) {
 *	    val = CBGetOptAttribute(CBbuf, reclen, "RULES");
 *	}
 *	.
 *
 *  Returns	: Pointer to attribute, complete buffer or NULL.
 *------------------------------------------------------------------------
 */
const char *
CBGetOptAttribute(
	CBBuf_t		*CBbuf,		/* possible counted buffer */
	size_t		CBlength,	/* the length of CBbuf */
	const char	*CBattnam)	/* the attribute name */
{
    const char	*attval;

    if (! CBIsValid(CBbuf, CBlength)) return (const char *) CBbuf;
    if ( CBGetAttribute(CBbuf, CBlength, CBattnam, &attval)) return attval;
    return NULL;
} /* CBGetOptAttribute() */

/*------------------------------------------------------------------------
 *  Function	: CBGetToggleValue
 *  Purpose	: Parse boolean value for option toggles
 *  Method	:
 *
 *  History:
 *  18-10-2002	: jaap     : creation
 *  11-04-2011	: paul     : renamed from HMSNamesGetToggleValue()
 *
 *  Notes	:
 *  - Values are case-independent.
 *  - Positive values: 'Y', 'T', '1', 'J', "TRUE", "YES", "ON".
 *  - Negative values: 'N', 'F', '0', "FALSE", "NO", "OF".
 *
 *  Returns	: TRUE if value was parsed successfully, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
CBGetToggleValue(
	const char	*value,
	bool		*result)
{
    if (strlen(value) == 1) {
	const char c = toupper((unsigned char)*value);

	switch (c) {
	case 'Y': /* FALLTHROUGH */
	case 'T':
	case '1':
	case 'J':
	    *result = TRUE;
	    return TRUE;
	case 'N': /* FALLTHROUGH */
	case 'F':
	case '0':
	    *result = FALSE;
	    return TRUE;
	}
    } else {
	const char	*valptr = value;
	char		toggle[128];
	char		*t;

	for (t = toggle; t - toggle < sizeof(toggle) - 1; t++) {
	    if (*valptr == '\0') {
		break;
	    }
	    *t = toupper((unsigned char)*valptr);
	    valptr++;
	}
	*t = '\0';

	if (STREQ(toggle, "TRUE")
	    || STREQ(toggle, "YES")
	    || STREQ(toggle, "ON")) {
	    *result = TRUE;
	    return TRUE;
	} else if (STREQ(toggle, "FALSE")
		   || STREQ(toggle, "NO")
		   || STREQ(toggle, "OFF")) {
	    *result = FALSE;
	    return TRUE;
	}
    }
    return FALSE;
} /* CBGetToggleValue() */
