/*========================================================================
 *  Copyright (c) 2000 by Pica
 *
 *  File	: conn.h
 *  Description	: The connection interface
 *  Author(s)	: Hans van Eck
 *  Creation	: 22-05-2000
 *
 *  History:
 *  22-05-2000	: eck      : creation
 *  20-05-2005	: michael  : Added CONN_nodelay()
 *  01-04-2011	: heesakkm : Added CONN_nonblock(), CONN_bufsize() and
 *			     CONN_keepalive()
 *  05-04-2011	: heesakkm : Added CONN_passive_tcp_opts() and
 *			     CONN_connect_*_opts().
 *  18-04-2011	: heesakkm : Added CONN_connect_IPv4() and CONN_connect_IPv6().
 *  13-05-2011	: heesakkm : Added CONN_address_equal().
 *  15-06-2011	: heesakkm : Added CONN_my_ipaddr_opts().
 *  22-06-2011	: heesakkm : Added CONN_passive_tcp4 and CONN_passive_tcp6().
 *  27-06-2011	: heesakkm : Added INET6_ADDRSTRLEN.
 *  28-06-2011	: heesakkm : Moved INET6_ADDRSTRLEN back to connect.c
 *			     Added CONN_ADDRESS_UNKNOWN
 *  01-07-2011	: heesakkm : Added CONN_valid_ipv4() and CONN_valid_ipv6().
 *  04-11-2013	: martinz  : Added CONN_accept_host_port().
 *
 *  Notes	:
 *  Some simple interfaces for creating connections in order to avoid recoding
 *  lots of stuff.
 *
 *========================================================================
 */

#ifndef CONN_H_INCLUDED
#define CONN_H_INCLUDED

#pragma ident "$Id: conn.h 335412 2016-05-27 11:01:38Z heesakkm $"

#include <pica.h>
#include <cs.h>


#define CONN_ADDRESS_UNKNOWN	'?'
#define CONN_ADDRESS_IPV4	'T'
#define CONN_ADDRESS_IPV6	'S'
#define CONN_ADDRESS_LOCAL	'L'

#define CONN_IP_VERSION4	'4'
#define CONN_IP_VERSION6	'6'

/* Test if an address string is a valid IPv4/IPv6 address. */
bool CONN_valid_ipv4(const char *addr);
bool CONN_valid_ipv6(const char *addr);

/* Determine if two address specifications refer to the same address. */
bool CONN_address_equal(const char *addrspec1, const char *addrspec2,
			const char *options);

/* Retrieve the configured IP order. */
const char * CONN_ip_order(void);

/* Wrapper around uname(2). */
const char *CONN_my_name(void);

/* Determine own IP-address. Returns dotted decimal string for IPv4 or an IPv6
 * address.
 */
const char *CONN_my_ipaddr(void);
const char *CONN_my_ipaddr_opts(const char *options);

/* CONN_parse_addrspec(): determine the address, service (port) and address
 * type.
 */
bool CONN_parse_addrspec(const char *addrspec, const char *options,
			    CSBuf_p address, CSBuf_p service,
			    char *address_type);

/* CONN_passive_tcp(): create socket to connect to. Returns file-descriptor.
 * Initial value of fullname is of form [<ip-address>][:<port>]
 * On success, fullname will contain <ip-address>:<port>
 */
int CONN_passive_tcp(CSBuf_p fullname, int backlog);
int CONN_passive_tcp4(CSBuf_p fullname, const char *options, int backlog);
int CONN_passive_tcp6(CSBuf_p fullname, const char *options, int backlog);
int CONN_passive_tcp_opts(CSBuf_p fullname, const char *options, int backlog);

/* CONN_connect_tcp(): Connect to a server using TCP/IP.
 * Returns file-descriptor. The addrspec is of the form <ip-address>:<port>
 */
int CONN_connect_tcp(const char *addrspec);
int CONN_connect_tcp4(const char *addrspec, const char *options);
int CONN_connect_tcp6(const char *addrspec, const char *options);
int CONN_connect_tcp_opts(const char *addrspec, const char *options);

/* CONN_nonblock(): set or clear the non-blocking flag of a connection.
 */
bool CONN_nonblock(int sockfd, bool nodelay_flag);

/* CONN_close_on_exec(): set or clear the close-on-exec flag of a connection.
 */
bool CONN_close_on_exec(int sockfd, bool close_flag);

/* CONN_bufsize(): set the size of the send/receive buffer of a connection.
 */
bool CONN_bufsize(int sockfd, int buftype, int bufsize);

/* CONN_nodelay(): set the TCP_NODELAY option of a connection.
 * It can be called after CONN_connect_tcp().
 */
bool CONN_nodelay(int sockfd, bool nodelay_flag);

/* CONN_keepalive(): set the SO_KEEPALIVE option of a connection.
 */
bool CONN_keepalive(int sockfd, bool keepalive_flag);

/*
 * Local connections are Unix-domain sockets.
 */

/* CONN_passive_local(): create socket to connect to. Returns file-descriptor.
 * fullname should be a full path of a file or empty.
 * On success, fullname will contain filename.
 */
int CONN_passive_local(CSBuf_p fullname, int backlog);

/* CONN_connect_local(): create connection to server. Returns file-descriptor.
 * addr should contain filename.
 */
int CONN_connect_local(const char *addr);

/* CONN_connect(): Connect to either TCP/IP or local server, depending on
 * whether string starts with T= or L= respectively. The rest of the string is
 * use as parameter for CONN_connect_tcp() and CONN_connect_local().
 */
int CONN_connect(const char *addr);
int CONN_connect_opts(const char *addr, const char *options);

/* CONN_connect_oneof(): Connect to one of a number of servers (tried in order
 * of specification). The specs are separated by '|'.
 */
int CONN_connect_oneof(const char *addrspecs);
int CONN_connect_oneof_opts(const char *addrspecs, const char *options);

/* CONN_accept_host_port(): Accept a connection (either TCP/IP or local). When
 * TCP/IP, clt_host, clt_port, sin_family will contain the <ip-address>, <port>
 * and <address family> of the client. clt_host, clt_port will be empty for
 * local connections.
 */
int
CONN_accept_host_port(
	int sockfd, CSBuf_p clt_host, CSBuf_p clt_port, short *sin_family);

/* CONN_accept(): Accept a connection (either TCP/IP or local). When TCP/IP,
 * clt_address will contain the <ip-address>:<port> of the client. The string
 * will be empty for local connections.
 */
int CONN_accept(int fildes, CSBuf_p clt_address);

/* CONN_close(): close a connection. This is a plain wrapper around close(),
 * but avoids having to include <unistd.h> for the close() only.
 */
int CONN_close(int fildes);

#endif /* CONN_H_INCLUDED */
