/*========================================================================
 *  Copyright (c) 2000 by Pica
 *
 *  File	: connect.c
 *  Description	: "Easy" interface for handling connections
 *  Author(s)	: Hans van Eck
 *  Creation	: 11-05-2000
 *
 *  History:
 *  11-05-2000	: eck      : creation
 *  24-05-2000	: eck      : Changed CONN_my_ipaddr() and CONN_passive_local()
 *			   : semantics.
 *  08-06-2000	: eck      : New implementation of CONN_my_ipaddr() using ioctl()
 *  20-05-2005	: michael  : Added CONN_nodelay()
 *  31-05-2007	: heesakkm : Added len-fix for Linux in CONN_my_ipaddr().
 *  14-02-2011	: heesakkm : Implemented IPv6 support.
 *  24-05-2011	: martinz  : Fixed CONN_accept().
 *  14-06-2011	: heesakkm : Fixed passive sockets not listening on INADDR_ANY.
 *  17-06-2011	: heesakkm : Reworked get_my_ipaddr() for IPv6 support.
 *  20-06-2011	: heesakkm : Apply preferred IP-order when creating sockets.
 *  22-06-2011	: heesakkm : Changed CONN_DEFAULT_IP_ORDER to "4" and added IP
 *			     dependent versions of CONN_passive_tcp().
 *  24-06-2011	: heesakkm : Allow both IPv4 and IPv6 sockets on the same port.
 *  27-06-2011	: heesakkm : Moved INET6_ADDRSTRLEN to conn.h,
 *			     removed unused INADDR_NONE.
 *  27-06-2011	: heesakkm : Support link-local IPv6 addresses.
 *  28-06-2011	: heesakkm : Moved INET6_ADDRSTRLEN back from conn.h
 *			     Changed CONN_parse_addrspec() interface.
 *  29-06-2011	: heesakkm : Initialize socket flags and options on new sockets.
 *  30-06-2011	: heesakkm : Only set socket options relevant to address family.
 *  01-07-2011	: heesakkm : Fixed ADDR_UNSPECIFIED() macro.
 *  01-07-2011	: heesakkm : Added CONN_valid_ipv4() and CONN_valid_ipv6().
 *  04-11-2013	: martinz  : Added CONN_accept_host_port().
 *
 *  Notes	:
 *  The IPC-interfaces are not easily usable in programs. Apart from the
 *  problems with #define's and #include's, they require detailed knowledge
 *  about underlying problems (like byte-order issues, data-structures, etc.).
 *  In order to alleviate these problems, a number of functions have been
 *  specified here which interface to the socket stuff. They require the
 *  inclusion of only one header (picamisc.h), and accept/return somewhat
 *  easier types (like counted-strings, file-descriptors). Strings returned by
 *  one part of these interfaces (like CONN_passive_local() / CONN_passive_tcp())
 *  can be used as input for another one (like CONN_connect_local() /
 *  CONN_connect_tcp()).
 *
 *  The file-descriptors returned here will have the FD_CLOEXEC flags set,
 *  meaning that programs that want to pass such file-descriptors to children
 *  will have to clear the bit. A dup2() will also do this.
 *
 *  The syntax for the creation of connections is:
 *
 *	addrspecs      ::= <spec> [ '|' <spec> ]*
 *	spec           ::= <tcp-spec> | <local-spec>
 *	local-spec     ::= 'L' <path-to-file>
 *	tcp-spec       ::= 'T'|'S' <ip-addr> ':' <tcp-port>
 *	ip-addr        ::= <hostname> | <IPv4-address> | <IPv6-reference>
 *	IPv6-reference ::= '[' <IPv6-address> ']'
 *
 *  Use of gethostbyname() requires -D_POSIX_C_SOURCE=200112L in case the C99
 *  implementation is enforced with -std=c99.
 *
 *  In support of IPV6:
 *  - calls to gethostbyname() have been replaced by getaddrinfo()
 *  - depreciated inet_addr() has been replaced by inet_pton().
 *========================================================================
 */

#pragma ident "$Id: connect.c 368670 2019-07-17 13:40:58Z vaneckh $"

#if defined(__sun)
#undef _XOPEN_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#if ! defined(NO_LOCAL)
#include <sys/un.h>
#include <sys/stat.h>
#endif

#include <unistd.h>
#include <fcntl.h>

#if defined(__sun)
#include <sys/sockio.h>
#endif
#include <sys/ioctl.h>
#include <net/if.h>

/* If both <net/if.h> and <ifaddrs.h> are being included, <net/if.h> must be
 * included before <ifaddrs.h>.
 */
#if defined(__linux)
#include <ifaddrs.h>	/* getifaddrs() is used to determine the IP address */
#endif

#include <ctype.h>
#include <errno.h>

#include <pica.h>
#include <picamisc.h>
#include <conn.h>
#include <cs.h>
#include <env.h>
#include <er.h>
#include <ihl.h>

/* According to POSIX.1g, the parameter to accept() and friends that determines
 * the length of the socket structure is a socklen_t. However, at this moment
 * (07-04-1998) of the systems in use here only Linux defines this type. I
 * have purposely used a typedef here so that when there is a system definition
 * of socklen_t you will get a compile error and can update this ifdef.
 * Hopefully some day all systems will do this the POSIX way.
 */
#if defined(__sun)
#	if defined(__SunOS_5_4) || defined(__SunOS_5_5) \
	    || defined(__SunOS_5_6)
	/* Solaris 2.4 through 2.6. Assume later versions have
	 * the socklen_t. At least 2.7 (== Solaris 7) does.
	 */
	typedef int	socklen_t;
#	endif
#endif

/* Some systems do not define AF_LOCAL. */
#ifndef AF_LOCAL
#define AF_LOCAL	AF_UNIX
#endif

/* Local defines. */

/* Solaris does not define INET6_ADDRSTRLEN. */
#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN	46	/* max len of IPv6 addr in ascii */
#endif

/* CONN_DEFAULT_IP_ORDER is the preferred protocol order that is used for
 * creating a socket in case no address is specified (e.g. "4,6").
 */
#define CONN_DEFAULT_IP_ORDER	"4"

#define	IPVSTR(family)		((family == AF_INET) ? "IPv4" : \
				    ((family == AF_INET6) ? "IPv6" : "IPv?"))

#define ADDR_UNSPECIFIED(f, s)	\
	    (f == AF_INET6)	\
			? (IN6_IS_ADDR_UNSPECIFIED(((struct in6_addr *)s))) \
			: (((struct in_addr *)s)->s_addr == INADDR_ANY)

/* CONN_DEBUG: Use ERWrite() instead of ERWarning() in some cases. */
#define CONN_DEBUG	0

/* CONN_addrdata_t is used to keep track of the network interface name of an IP
 * address.
 */
typedef struct CONN_addrdata_t {
    char	ca_ipnum[INET6_ADDRSTRLEN];	/* IP address */
    char	ca_ifname[25];			/* Network interface name */
} CONN_addrdata_t;


/*------------------------------------------------------------------------
 *  Function	: CreateCSbuf
 *  Purpose	: Create and initialize a CS
 *  Method	: CSCreate()
 *
 *  History:
 *  24-06-2011	: heesakkm : creation
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
CreateCSbuf(
	CSBuf_p		*csbufp,
	const char	*fname)
{
    if (*csbufp == NULL) {
	*csbufp = CSCreate();
	if (*csbufp == NULL) {
	    ERWrite(CSerrno, fname, "Could not create CS buffer");
	    return FALSE;
	}
    }
    return CSClear(*csbufp);
} /* CreateCSbuf() */

/*------------------------------------------------------------------------
 *  Function	: sockoptname_to_string
 *  Purpose	: Convert a user-settable setsockopt option to string
 *  Method	:
 *
 *  History:
 *  16-03-2011	: heesakkm : creation
 *  24-06-2011	: heesakkm : Added IPPROTO_IPV6-IPV6_V6ONLY
 *
 *  Returns	: String with option name or "???" for unknown option
 *------------------------------------------------------------------------
 */
static const char *
sockoptname_to_string(
	int	level,
	int	optname)
{
    if (level == IPPROTO_TCP) {
	switch (optname) {
	case TCP_NODELAY:	return "IPPROTO_TCP-TCP_NODELAY";
	case TCP_MAXSEG:	return "IPPROTO_TCP-TCP_MAXSEG";
	default:		return "IPPROTO_TCP-???";
	}
    }
    if (level == SOL_SOCKET) {
	switch (optname) {
	case SO_KEEPALIVE:	return "SOL_SOCKET-SO_KEEPALIVE";
	case SO_REUSEADDR:	return "SOL_SOCKET-SO_REUSEADDR";
	case SO_RCVBUF:		return "SOL_SOCKET-SO_RCVBUF";
	case SO_SNDBUF:		return "SOL_SOCKET-SO_SNDBUF";
	case SO_DEBUG:		return "SOL_SOCKET-SO_DEBUG";
	default:		return "SOL_SOCKET-???";
	}
    }
    if (level == IPPROTO_IPV6) {
	switch (optname) {
	case IPV6_V6ONLY:	return "IPPROTO_IPV6-IPV6_V6ONLY";
	default:		return "IPPROTO_IPV6-???";
	}
    }
    return "???";
} /* sockoptname_to_string() */

/*------------------------------------------------------------------------
 *  Function	: conn_setsockopt
 *  Purpose	: Wrapper for setting socket options
 *  Method	:
 *  setsockopt() i.c.w. ERWrite() if verbose_err is raised, otherwise
 *  ERWarning().
 *
 *  History:
 *  16-03-2011	: heesakkm : creation
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
conn_setsockopt(
	int		sockfd,		/* File descriptor */
	int		level,		/* Protocol level */
	int		optname,
	const int	optval,
	const char	*caller,	/* Name of calling function */
	bool		verbose_err)	/* TRUE=ERWrite(), FALSE=ERWarning() */
{
    static const char	fname[] = "conn_setsockopt";

    /* A successful setsockopt() returns 0. */
    if (setsockopt(sockfd, level, optname,
		    (char *)&optval, sizeof(optval)) != 0) {
	CS_LOCAL(	csfname, 256);

	CSPrintf(csfname, "%s.%s", caller, fname);

	if (verbose_err) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "setsockopt(%s) failed: %s (errno %d)",
		    sockoptname_to_string(level, optname),
		    strerror(errno), errno);
	} else {
	    ERWarning(APPLE_PARAM, CSContents(csfname),
		    "setsockopt(%s) failed: %s (errno %d)",
		    sockoptname_to_string(level, optname),
		    strerror(errno), errno);
	}
	return FALSE;
    }
    return TRUE;
} /* conn_setsockopt() */

/*------------------------------------------------------------------------
 *  Function	: conn_setsockoptions
 *  Purpose	: Perform the requested setsockopt() operation(s)
 *  Method	:
 *  Process the options in the supplied options string. The option names are of
 *  the form <protocol level>-<optionname>. Call conn_setsockopt() for each
 *  supported option.
 *
 *  History:
 *  04-04-2011	: heesakkm : creation
 *  24-06-2011	: heesakkm : Added IPPROTO_IPV6-IPV6_V6ONLY
 *  30-06-2011	: heesakkm : Only set IPV6_V6ONLY for AF_INET6.
 *
 *  Notes	:
 *  Supported options:
 *
 *	SOL_SOCKET-SO_KEEPALIVE
 *	SOL_SOCKET-SO_REUSEADDR
 *	SOL_SOCKET-SO_RCVBUF
 *	SOL_SOCKET-SO_SNDBUF
 *	SOL_SOCKET-SO_DEBUG
 *	IPPROTO_TCP-TCP_NODELAY
 *	IPPROTO_TCP-TCP_MAXSEG
 *	IPPROTO_IPV6-IPV6_V6ONLY
 *
 *  Unknown values in the options string are ignored.
 *  The return value will be FALSE if any of the calls to conn_setsockopt()
 *  fails. So it is up to the caller to deal with the outcome.
 *
 *  Important: When adding support for new socket options, also update
 *  sockoptname_to_string() and possibly initialize_socket().
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
conn_setsockoptions(
	int		sockfd,		/* File descriptor */
	int		family,		/* Address family */
	bool		bind_call,	/* Is socket used to bind() to? */
	const char	*options,
	const char	*caller,	/* Name of calling function */
	bool		verbose_err)	/* TRUE=ERWrite(), FALSE=ERWarning() */
{
    static const char	fname[] = "conn_setsockoptions";
    CS_LOCAL(		csfname, 256);
    CS_LOCAL(		csoption, 100);		/* <protocol level>-<optionname> */
    CS_LOCAL(		csvalue, 100);		/* option value */
    CS_LOCAL(		cslevel, 25);		/* <protocol level> part in csoption */
    CS_LOCAL(		csoptname, 75);		/* <optionname> part in csoption */
    const char		*ptr, *sep;
    const char		*name_start, *name_end;
    const char		*value_start, *value_end;
    int			sockopt_level;
    int			sockopt_name;
    int			sockopt_val;
    bool		ok = TRUE;

    CSPrintf(csfname, "%s.%s", caller, fname);

    ptr = options;
    while ((ptr != NULL) && (*ptr != '\0')) {
	name_start = PM_opt_findname(ptr);
	name_end = PM_opt_findnameend(ptr);
	value_start = PM_opt_findval(ptr);
	value_end = PM_opt_findvalend(ptr);

	CSSetStringN(csoption, name_start, name_end - name_start);
	CSSetStringN(csvalue, value_start, value_end - value_start);

	ptr = PM_opt_findnext(ptr);

	sep = strchr(CSContents(csoption), '-');
	if (sep == NULL) continue;

	CSSetStringN(cslevel, CSContents(csoption),
			sep - CSContents(csoption));
	CSSetString(csoptname, sep+1);

	if (PM_strcmpi(CSContents(cslevel), "SOL_SOCKET") == 0) {
	    sockopt_level = SOL_SOCKET;

	    if (PM_strcmpi(CSContents(csoptname), "SO_KEEPALIVE") == 0) {
		/* Periodically test if connection still alive */
		sockopt_name = SO_KEEPALIVE;
	    } else if (PM_strcmpi(CSContents(csoptname), "SO_REUSEADDR") == 0) {
		if (!bind_call) continue;

		/* Allow local address reuse */
		sockopt_name = SO_REUSEADDR;
	    } else if (PM_strcmpi(CSContents(csoptname), "SO_RCVBUF") == 0) {
		/* Receive buffer size */
		sockopt_name = SO_RCVBUF;
	    } else if (PM_strcmpi(CSContents(csoptname), "SO_SNDBUF") == 0) {
		/* Send buffer size */
		sockopt_name = SO_SNDBUF;
	    } else if (PM_strcmpi(CSContents(csoptname), "SO_DEBUG") == 0) {
		/* Enable debug tracing */
		sockopt_name = SO_DEBUG;
	    } else {
		continue;
	    }
	} else if (PM_strcmpi(CSContents(cslevel), "IPPROTO_TCP") == 0) {
	    sockopt_level = IPPROTO_TCP;

	    if (PM_strcmpi(CSContents(csoptname), "TCP_NODELAY") == 0) {
		/* Disable Nagle algorithm */
		sockopt_name = TCP_NODELAY;
	    } else if (PM_strcmpi(CSContents(csoptname), "TCP_MAXSEG") == 0) {
		/* TCP maximum segment size */
		sockopt_name = TCP_MAXSEG;
	    } else {
		continue;
	    }
#ifdef IPV6_V6ONLY
	} else if (PM_strcmpi(CSContents(cslevel), "IPPROTO_IPV6") == 0) {
	    sockopt_level = IPPROTO_IPV6;

	    if (PM_strcmpi(CSContents(csoptname), "IPV6_V6ONLY") == 0) {
		if (family != AF_INET6) continue;
		/* Disable v4 compatibility */
		sockopt_name = IPV6_V6ONLY;
	    } else {
		continue;
	    }
#endif
	} else {
	    continue;
	}
	sockopt_val = atoi(CSContents(csvalue));

	ok = conn_setsockopt(sockfd, sockopt_level, sockopt_name,
				(const int) sockopt_val,
				CSContents(csfname), verbose_err)
		&& ok;
    }
    return ok;
} /* conn_setsockoptions() */

/*------------------------------------------------------------------------
 *  Function	: initialize_socket
 *  Purpose	: Set some default socket flags and options
 *  Method	: Guess...
 *
 *  History:
 *  29-06-2011	: heesakkm : creation
 *  30-06-2011	: heesakkm : Initialize options relevant to the address family.
 *
 *  Notes	:
 *  File descriptor flags:
 *	FD_CLOEXEC			-> CONN_close_on_exec()
 *	O_NONBLOCK			-> CONN_nonblock()
 *
 *  File descriptor options:		-> conn_setsockopt()
 *	SOL_SOCKET-SO_REUSEADDR
 *	SOL_SOCKET-SO_KEEPALIVE
 *	SOL_SOCKET-SO_DEBUG
 *	IPPROTO_TCP-TCP_NODELAY
 *	IPPROTO_IPV6-IPV6_V6ONLY
 *
 *  Leave initialization of the following options to the operating-system:
 *	SOL_SOCKET-SO_RCVBUF
 *	SOL_SOCKET-SO_SNDBUF
 *	IPPROTO_TCP-TCP_MAXSEG
 *------------------------------------------------------------------------
 */
static void
initialize_socket(
	int	sockfd,		/* File descriptor */
	int	family,		/* Address family */
	bool	bind_call)	/* Is socket used to bind() to? */
{
    static const char	fname[] = "initialize_socket";

    /* Set socket flags */
    (void) CONN_close_on_exec(sockfd, FALSE);
    (void) CONN_nonblock(sockfd, FALSE);

    /* Set socket options */
    if (bind_call) {
	conn_setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, 0, fname, FALSE);
    }
    if ((family == AF_INET) || (family == AF_INET6)) {
	conn_setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE, 0, fname, FALSE);
	conn_setsockopt(sockfd, SOL_SOCKET, SO_DEBUG, 0, fname, FALSE);
	conn_setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, 0, fname, FALSE);
#ifdef IPV6_V6ONLY
	if (family == AF_INET6) {
	    conn_setsockopt(sockfd, IPPROTO_IPV6, IPV6_V6ONLY, 0, fname, FALSE);
	}
#endif
    }
} /* initialize_socket() */

/*------------------------------------------------------------------------
 *  Function	: get_sin_addr
 *  Purpose	: Get the network address structure
 *  Method	:
 *
 *  History:
 *  18-03-2011	: heesakkm : creation
 *
 *  Notes	:
 *  AF_INET (IPv4) is assumed if the address family is not AF_INET6 (IPv6).
 *
 *  Returns	: Pointer to struct in_addr (IPv4) or struct in_addr6 (IPv6)
 *------------------------------------------------------------------------
 */
static void *
get_sin_addr(
	struct sockaddr	*sa)
{
    if (sa->sa_family == AF_INET6) {
	// LINTED: E_BAD_PTR_CAST_ALIGN
	return &(((struct sockaddr_in6*)sa)->sin6_addr);
    }
    // LINTED: E_BAD_PTR_CAST_ALIGN
    return &(((struct sockaddr_in*)sa)->sin_addr);
} /* get_sin_addr() */

/*------------------------------------------------------------------------
 *  Function	: get_sockaddr_address
 *  Purpose	: Get the address string of a sockaddr
 *  Method	:
 *  Convert an address into a string with inet_ntop() i.c.w. get_sin_addr()
 *
 *  History:
 *  21-04-2011	: heesakkm : creation
 *
 *  Returns	: Pointer to the result buffer or NULL on error
 *------------------------------------------------------------------------
 */
static const char *
get_sockaddr_address(
	struct sockaddr	*sa,
	char		*ipbuf,		/* OUT */
	size_t		ipbuf_len)
{
    static const char	fname[] = "get_sockaddr_address";
    const char		*addr;

    *ipbuf = '\0';

    addr = inet_ntop(sa->sa_family, get_sin_addr(sa),
			ipbuf, (socklen_t) ipbuf_len);
    if (addr == NULL) {
	ERWrite(APPLE_PARAM, fname, "inet_ntop() failed: %s (errno %d)",
		strerror(errno), errno);
    }
    return addr;
} /* get_sockaddr_address() */

/*------------------------------------------------------------------------
 *  Function	: get_sockaddr_port
 *  Purpose	: Get the port number of a sockaddr
 *  Method	: ntohs()
 *
 *  History:
 *  18-03-2011	: heesakkm : creation
 *
 *  Notes	: IPv4 is assumed if the address family is not AF_INET6 (IPv6).
 *
 *  Returns	: The port number
 *------------------------------------------------------------------------
 */
static int
get_sockaddr_port(
	struct sockaddr	*sa)
{
    int	port;

    if (sa->sa_family == AF_INET6) {
	// LINTED: E_BAD_PTR_CAST_ALIGN
	port = ((struct sockaddr_in6*)sa)->sin6_port;
    } else {
	// LINTED: E_BAD_PTR_CAST_ALIGN
	port = ((struct sockaddr_in*)sa)->sin_port;
    }
    return ntohs(port);
} /* get_sockaddr_port() */

/*------------------------------------------------------------------------
 *  Function	: CONN_ip_order
 *  Purpose	: Determine the preferred protocol order
 *  Method	: Call ENVGetVar() for "CONN_IP_ORDER".
 *
 *  History:
 *  22-03-2011	: heesakkm : creation
 *
 *  Notes	: The default order is IPv4, IPv6.
 *
 *  Description	:
 *  Retrieve the preferred IP order that is set with the environment variable
 *  CONN_IP_ORDER. If the environment variable is not configured, the default
 *  value defined by CONN_DEFAULT_IP_ORDER is returned. This default is
 *  currently configured as "4".
 *
 *  Returns	: Pointer to string with protocol order (never NULL)
 *------------------------------------------------------------------------
 */
const char *
CONN_ip_order(void)
{
    const char	*ip_order = ENVGetVar("CONN_IP_ORDER");

    if ((ip_order == NULL) || (*ip_order == '\0')) {
	ip_order = CONN_DEFAULT_IP_ORDER;
    }
    return ip_order;
} /* CONN_ip_order() */

/*------------------------------------------------------------------------
 *  Function	: get_ip_order
 *  Purpose	: Determine the preferred protocol order
 *  Method	:
 *  Retrieve the "ip-order" option value from the options string.
 *  Otherwise call CONN_ip_order().
 *
 *  History:
 *  21-06-2011	: heesakkm : creation
 *
 *  Returns	: Pointer to buffer with protocol order (never NULL)
 *------------------------------------------------------------------------
 */
static const char *
get_ip_order(
	const char	*options,
	char		*buf,		/* OUT */
	size_t		buflen)
{
    if (PM_opt_fetch(options, "ip-order", buf, buflen) > 0) {
	return buf;
    }
    return PM_strncpy(buf, CONN_ip_order(), buflen);
} /* get_ip_order() */

/*------------------------------------------------------------------------
 *  Function	: address_equal
 *  Purpose	: Simple test to see if two IP addresses are equal
 *  Method	:
 *  If an exact string compare fails, then create a simple host list with
 *  IHL_Add() that contains one address. Compare the other address against the
 *  list with IHL_Check().
 *
 *  History:
 *  24-06-2011	: heesakkm : creation
 *  31-10-2013	: heesakkm : CBS-5174: Don't feed hostname to IHL_Add().
 *
 *  Returns	: Whether the addresses match
 *------------------------------------------------------------------------
 */
static bool
address_equal(
	const char	*addr1,
	const char	*addr2)
{
    IHL_Handle_p	ihl = NULL;
    bool		ok;

    if (   (addr1 == NULL) || (*addr1 == '\0')
	|| (addr2 == NULL) || (*addr2 == '\0')) return FALSE;

    if (STREQ(addr1, addr2)) return TRUE;

    if (!IHL_IsValidAddress(addr2)) return FALSE;

    ihl = IHL_Add(NULL, addr2, NULL);
    if (ihl == NULL) return FALSE;

    ok = IHL_Check(ihl, addr1) != NULL;
    IHL_Destroy(ihl);
    return ok;
} /* address_equal() */

/*------------------------------------------------------------------------
 *  Function	: get_ifconfsize
 *  Purpose	: Determine the buffer size for storing the interface entries.
 *  Method	:
 *  Call ioctl() if SIOCGLIFNUM is defined, otherwise use the fixed size of
 *  (NR_INTERFACES * sizeof interface element).
 *
 *  History:
 *  20-06-2011	: heesakkm : creation
 *
 *  Notes	:
 *  If SIOCGLIFNUM is not available or ioctl(SIOCGLIFNUM) fails, then the size
 *  of at most 10 interfaces is returned. So the first usable interface should
 *  be in there.
 *
 *  Returns	:
 *  The size of the buffer that will hold the interface entries
 *------------------------------------------------------------------------
 */
#ifndef _IFADDRS_H
static int
get_ifconfsize(
	int		sockfd,
	int		family)
{
#define NR_INTERFACES	10
#ifdef SIOCGLIFNUM
    static const char	fname[] = "get_ifconfsize";
    struct lifnum	ifn;
    /*
     * Determine the buffer size for all interface.
     */
    memset(&ifn, 0, sizeof(ifn));
    ifn.lifn_family = (sa_family_t)family;
    ifn.lifn_flags = 0;
    ifn.lifn_count = 0;
    if (ioctl(sockfd, SIOCGLIFNUM, (char *)&ifn) < 0) {
	ERWarning(APPLE_PARAM, fname,
		    "ioctl(SIOCGLIFNUM) failed: %s (errno %d)",
		    strerror(errno), errno);
	ifn.lifn_count = NR_INTERFACES;
    }
    return (int) (ifn.lifn_count * sizeof(struct lifreq));
#else
    /*
     * Determine the buffer size for at most 10 interfaces.
     * If SIOCGLIFCONF is not available, then fall back on SIOCGIFCONF.
     */
#ifdef SIOCGLIFCONF
    /* The struct lifreq can be used for both IPv4 and IPv6. */
    return (int) (NR_INTERFACES * sizeof(struct lifreq));
#else
    /* The old (obsolete) struct ifreq only supports IPv4. */
    return (int) (NR_INTERFACES * sizeof(struct ifreq));
#endif	/* #ifdef SIOCGLIFCONF */
#endif	/* #ifdef SIOCGLIFNUM */
#undef NR_INTERFACES
} /* get_ifconfsize() */
#endif	/* #ifndef _IFADDRS_H */

/*------------------------------------------------------------------------
 *  Function	: get_my_address_details
 *  Purpose	: Determine the (or 'a') IP address details of the server
 *  Method	:
 *  Linux platform:
 *  1. Call getifaddrs() to get interfaces
 *  2. Loop over the interfaces ignoring non-INET-interfaces
 *  3. If the flags indicate that the interface is UP and not LOOPBACK convert
 *     the IP address to a string and return.
 *
 *  Sun platform:
 *  1. Create a socket for the requested address family
 *  2. Call get_ifconfsize() to determine the storage size for the interfaces
 *  2. Call ioctl(... SIOCGLIFCONF ...) to get interfaces
 *  3. Loop over the interfaces ignoring non-INET-interfaces
 *  4. Call ioctl(... SIOCGLIFFLAGS ...) to get the flags.
 *  5. If the flags indicate that the interface is UP and not LOOPBACK convert
 *     the IP address to a string with get_sockaddr_address() and return.
 *
 *  History:
 *  22-03-2011	: heesakkm : creation
 *  17-06-2011	: heesakkm : Rewrite for IPv6 support using getifaddrs() on
 *			     Linux and ioctl() on other platforms.
 *  24-06-2011	: heesakkm : Renamed from get_my_ipaddr().
 *  28-06-2011	: heesakkm : Fixed use of ifname pointer to freed memory (Linux)
 *
 *  Notes	:
 *  Use SIOCGLIFCONF instead of SIOCGIFCONF to get the interfaces, because the
 *  latter doesn't return IPv6 information on certain platforms (such as modern
 *  Solaris or older Linux).
 *
 *  If requested_ip is supplied in the interface, then we're only interested in
 *  the interface name for this IP address.
 *
 *  The return value is an IP address string. This circumvents the need for
 *  subsequent name-lookups.
 *
 *  Returns	:
 *  Pointer to the address in addrdata->ca_ipnum or NULL.
 *  Additional details are stored in the address struct.
 *------------------------------------------------------------------------
 */
static const char *
get_my_address_details(
	int		family,		/* Address family */
	const char	*requested_ip,	/* IP number to look for */
	CONN_addrdata_t	*addrdata)	/* IN/OUT */
{
    static const char	fname[] = "get_my_address_details";
    const char		*retval = NULL;

    memset(addrdata, 0, sizeof(*addrdata));

#ifdef _IFADDRS_H
    /* On Linux, getifaddrs() returns the interface information for both IPv4
     * and IPv6 addresses (SIOCGLIFCONF is not available and SIOCGIFCONF only
     * returns IPv4 addresses).
     */
    struct ifaddrs	*addrs = NULL;

    if (getifaddrs(&addrs) != 0) {
	ERWrite(APPLE_PARAM, fname,
		"My IP address unknown: getifaddrs() failed: %s (errno %d)",
		strerror(errno), errno);
    } else {
	struct ifaddrs	*ifa = NULL;

	for (ifa = addrs; ifa != NULL; ifa = ifa->ifa_next) {
	    if (ifa->ifa_addr == NULL) continue;
	    if (ifa->ifa_addr->sa_family != family) continue;
	    if ((ifa->ifa_flags & IFF_UP) == 0) continue;
	    if (ifa->ifa_flags & IFF_LOOPBACK) continue;

	    retval = get_sockaddr_address((struct sockaddr *)ifa->ifa_addr,
					    addrdata->ca_ipnum,
					    sizeof(addrdata->ca_ipnum));
	    if (retval != NULL) {
		/* Check if we have a match on the requested IP number. */
		if ((requested_ip != NULL)
		    && !address_equal(retval, requested_ip)) {
		    continue;
		}
		/* Store the interface name */
		PM_strncpy(addrdata->ca_ifname, ifa->ifa_name,
			    sizeof(addrdata->ca_ifname));
		break;
	    }
	}
	freeifaddrs(addrs);
    }
#else
    /* If "getifaddrs()" is not available, then SIOCGLIFCONF is preferred
     * rather than SIOCGIFCONF in order to get the IPv4 and IPv6 addresses.
     */
    int		ioctl_conf_request;
    int		ioctl_flag_request;
    int		sockfd;

    /* Create a socket from which to fetch IPv4 or IPv6 information. */
    sockfd = socket(family, SOCK_STREAM, IPPROTO_IP);
    if (sockfd < 0) {
	ERWrite(APPLE_PARAM, fname,
		"My IP address unknown: Could not create %s socket: "
		"%s (errno %d)",
		IPVSTR(family), strerror(errno), errno);
    } else {
#ifdef SIOCGLIFCONF			/* IPv4 and IPv6 */
	struct lifconf	ifc;
	struct lifreq	*ifrp;
	struct lifreq	ifrflags;
#define xifreq		lifreq
#define xifc_len	lifc_len
#define xifr_flags	lifr_flags
#define xifr_name	lifr_name
#define xifr_addr	lifr_addr
#else	/* obsolete SIOCGIFCONF */	/* IPv4 only */
	struct ifconf	ifc;
	struct ifreq	*ifrp;
	struct ifreq	ifrflags;
#define xifreq		ifreq
#define xifc_len	ifc_len
#define xifr_flags	ifr_flags
#define xifr_name	ifr_name
#define xifr_addr	ifr_addr
#endif
	struct sockaddr	*sa;
	char		*buf;
	int		buf_size;

	buf_size = get_ifconfsize(sockfd, family);
	if (buf_size <= 0) {
	    close(sockfd);
	    return NULL;
	}

	/*
	 * Allocate a buffer for the interface entries.
	 */
	buf = malloc(buf_size);
	if (buf == NULL) {
	    ERWrite(APPLE_NOMEM, fname,
		    "My IP address unknown: malloc() failed");
	    close(sockfd);
	    return NULL;
	}
	memset(buf, 0, buf_size);

	/*
	 * Fetch the interface entries into the buffer.
	 */
	memset(&ifc, 0, sizeof(ifc));
#ifdef SIOCGLIFCONF
	/* SIOCGLIFCONF returns configuration information for IPv4 and IPv6. */
	ioctl_conf_request = SIOCGLIFCONF;
	ioctl_flag_request = SIOCGLIFFLAGS;
	ifc.lifc_len = buf_size;
	ifc.lifc_buf = buf;
	ifc.lifc_family = (sa_family_t)family;
	ifc.lifc_flags = 0;
#else
	/* SIOCGIFCONF doesn't return IPv6 information. */
	ioctl_conf_request = SIOCGIFCONF;
	ioctl_flag_request = SIOCGIFFLAGS;
	ifc.ifc_len = buf_size;
	ifc.ifc_buf = buf;
#endif
	if (ioctl(sockfd, ioctl_conf_request, &ifc) < 0) {
	    ERWrite(APPLE_PARAM, fname,
		    "My IP address unknown: ioctl() failed: %s (errno %d)",
		    strerror(errno), errno);
	    close(sockfd);
	    free(buf);
	    return NULL;
	}

	/*
	 * Iterate through the list of interfaces.
	 */
	// LINTED: E_BAD_PTR_CAST_ALIGN
	for (ifrp = (struct xifreq *)buf; ifrp < (struct xifreq *)(buf + ifc.xifc_len); ifrp++) {
	    sa = (struct sockaddr *)&ifrp->xifr_addr;
	    if (sa->sa_family != family) continue;

	    /*
	     * We have an IP address. Now check some flags.
	     */
	    memset(&ifrflags, 0, sizeof(ifrflags));
	    PM_strncpy(ifrflags.xifr_name, ifrp->xifr_name,
			sizeof(ifrflags.xifr_name));
	    if (ioctl(sockfd, ioctl_flag_request, (char *)&ifrflags) < 0) {
		continue;
	    }
	    if ((ifrflags.xifr_flags & IFF_UP) == 0) continue;
	    if (ifrflags.xifr_flags & IFF_LOOPBACK) continue;

	    retval = get_sockaddr_address(sa, addrdata->ca_ipnum,
					    sizeof(addrdata->ca_ipnum));
	    if (retval != NULL) {
		/* Check if we have a match on the requested IP number. */
		if ((requested_ip != NULL)
		    && !address_equal(retval, requested_ip)) {
		    continue;
		}
		/* Store the interface name */
		PM_strncpy(addrdata->ca_ifname, ifrp->xifr_name,
			    sizeof(addrdata->ca_ifname));
		break;
	    }
	}
	close(sockfd);
	free(buf);
    }
#endif
    if (retval == NULL) {
	addrdata->ca_ipnum[0] = '\0';
	ERWrite(APPLE_PARAM, fname, "My IP address unknown");
    }
    return retval;
} /* get_my_address_details() */

/*------------------------------------------------------------------------
 *  Function	: test_ip_address
 *  Purpose	: Attempt to read a literal IPv4 or IPv6 address
 *  Method	:
 *  Call inet_pton() for AF_INET or AF_INET6. If the address family is unknown,
 *  then check if it's an AF_INET or AF_INET6 address and set the address
 *  family value accordingly.
 *
 *  History:
 *  06-05-2011	: heesakkm : creation
 *
 *  Notes	:
 *  In case of AF_INET6, the address must *not* be enclosed in square brackets.
 *
 *  Returns	:
 *  The inet_pton() return value:
 *  - a negative value case of an invalid address family;
 *  - 0 if address does not contain a character string representing a valid
 *      network address in the specified address family;
 *  - a positive value if the network address was successfully converted.
 *------------------------------------------------------------------------
 */
static int
test_ip_address(
	const char	*address,
	int		*family)		/* IN/OUT: address family */
{
    static const char	fname[] = "test_ip_address";
    struct sockaddr_in	sa4;
    struct sockaddr_in6	sa6;
    int			rc = -1;
    int			af = AF_UNSPEC;

    if ((address == NULL) || (*address == '\0')) {
	ERWrite(APPLE_PARAM, fname, "Missing or empty IP address in input");
	return -1;
    }
    if (family != NULL) {
	af = *family;
    }
    if (af == AF_INET) {
	rc = inet_pton(af, address, &(sa4.sin_addr));
    } else if (af == AF_INET6) {
	rc = inet_pton(af, address, &(sa6.sin6_addr));
    } else {
	rc = inet_pton(AF_INET, address, &(sa4.sin_addr));
	if (rc > 0) {
	    af = AF_INET;
	} else {
	    rc = inet_pton(AF_INET6, address, &(sa6.sin6_addr));
	    if (rc > 0) {
		af = AF_INET6;
	    } else {
		af = AF_UNSPEC;
	    }
	}
    }
    if (family != NULL) {
	*family = af;
    }
    return rc;
} /* test_ip_address() */

/*------------------------------------------------------------------------
 *  Function	: CONN_valid_ipv4
 *  Purpose	: Test if an address string is an IPv4 address
 *  Method	: test_ip_address()
 *
 *  History:
 *  01-07-2011	: heesakkm : creation
 *
 *  Returns	:
 *  Whether the supplied string contains a dotted-decimal IPv4 address
 *------------------------------------------------------------------------
 */
bool
CONN_valid_ipv4(
	const char	*addr)
{
    int	family = AF_INET;

    return (addr != NULL) && (*addr != '\0')
	    && (test_ip_address(addr, &family) > 0);
} /* CONN_valid_ipv4() */

/*------------------------------------------------------------------------
 *  Function	: CONN_valid_ipv6
 *  Purpose	: Test if an address string is an IPv6 address
 *  Method	: test_ip_address()
 *
 *  History:
 *  01-07-2011	: heesakkm : creation
 *
 *  Notes	:
 *  Fails if the supplied address is surrounded by square brackets.
 *
 *  Returns	:
 *  Whether the supplied string contains a colon-hexadecimal IPv6 address
 *------------------------------------------------------------------------
 */
bool
CONN_valid_ipv6(
	const char	*addr)
{
    int	family = AF_INET6;

    return (addr != NULL) && (*addr != '\0')
	    && (test_ip_address(addr, &family) > 0);
} /* CONN_valid_ipv6() */

/*------------------------------------------------------------------------
 *  Function	: CONN_address_equal
 *  Purpose	: Test whether two address specs refer to the same address
 *  Method	:
 *  Parse both address specifications with CONN_parse_addrspec().
 *  Perform a string comparison on the address and port. If there's no match,
 *  then compare the addresses with address_equal().
 *  Finally, if both addresses are the same, then compare the ports unless
 *  the option "port-required=no" is set.
 *
 *  History:
 *  13-05-2011	: heesakkm : creation
 *  27-05-2011	: heesakkm : Use IHL library.
 *  23-06-2011	: heesakkm : Changed option "ignore-port" -> "port-required".
 *  24-06-2011	: heesakkm : Use address_equal().
 *
 *  Notes	:
 *  Supported options:
 *
 *	"port-required=no|yes"	(no: ignore unequal ports)
 *
 *  Description	:
 *  Determine if two address specifications refer to the same address. Each
 *  address specification can contain a hostname or an IP number.
 *
 *  By default, this function also compares the port numbers that may be part
 *  of a specification. If port numbers should be ignored, then the options
 *  string should contain "port-required=no".
 *
 *  This function uses CONN_parse_addrspec() for parsing the address
 *  specifications. The supplied options string is passed on to
 *  CONN_parse_addrspec().
 *
 *  Returns	: Whether the supplied addresses are equal
 *------------------------------------------------------------------------
 */
bool
CONN_address_equal(
	const char	*addrspec1,
	const char	*addrspec2,
	const char	*options)
{
    static const char	fname[] = "CONN_address_equal";
    static CSBuf_p	csopts = NULL;
    CS_LOCAL(		addr1, 100);
    CS_LOCAL(		addr2, 100);
    CS_LOCAL(		port1, 25);
    CS_LOCAL(		port2, 25);
    char		optval[100] = "";
    bool		ignoreport = FALSE;
    bool		ok;

    if (   (addrspec1 == NULL) || (*addrspec1 == '\0')
	|| (addrspec2 == NULL) || (*addrspec2 == '\0')) {
	return FALSE;
    }
    if (! CreateCSbuf(&csopts, fname)) return FALSE;

    CSSetString(csopts, "ipv6-brackets=no");
    if ((options != NULL) && (*options != '\0')) {
	CSAppendC(csopts, ';');
	CSAppend(csopts, options);
    }
    if (! CONN_parse_addrspec(addrspec1, CSContents(csopts),
				addr1, port1, NULL)) {
	ERWrite(APPLE_PARAM, fname,
		"Invalid address specification: \"%s\"", addrspec1);
	return FALSE;
    }
    if (! CONN_parse_addrspec(addrspec2, CSContents(csopts),
				addr2, port2, NULL)) {
	ERWrite(APPLE_PARAM, fname,
		"Invalid address specification: \"%s\"", addrspec2);
	return FALSE;
    }

    if (   STREQ(CSContents(addr1), CSContents(addr2))
	&& STREQ(CSContents(port1), CSContents(port2))) {
	return TRUE;
    }
    ok = address_equal(CSContents(addr1), CSContents(addr2));

    if (ok && PM_opt_fetch(options, "port-required",
			    optval, sizeof(optval)) > 0) {
	ignoreport = (PM_strcmpi(optval, "no") == 0);
    }
    return ok && (ignoreport
		|| (atoi(CSContents(port1)) == atoi(CSContents(port2))));
} /* CONN_address_equal() */

/*------------------------------------------------------------------------
 *  Function	: get_preferred_family
 *  Purpose	: Determine the preferred address family
 *  Method	: get_ip_order()
 *
 *  History:
 *  06-05-2011	: heesakkm : creation
 *
 *  Returns	: The preferred address family or AF_UNSPEC
 *------------------------------------------------------------------------
 */
static int
get_preferred_family(
	const char	*options)
{
    char	ip_order[20];
    const char	*ptr = get_ip_order(options, ip_order, sizeof(ip_order));

    for (/*void*/; *ptr != '\0'; ptr++) {
	if (*ptr == CONN_IP_VERSION4) return AF_INET;
	if (*ptr == CONN_IP_VERSION6) return AF_INET6;
    }
    return AF_UNSPEC;
} /* get_preferred_family() */

/*------------------------------------------------------------------------
 *  Function	: parse_address_options
 *  Purpose	: Overrule the supplied values based on the options string
 *  Method	:
 *
 *  History:
 *  08-06-2011	: heesakkm : creation
 *
 *  Notes	:
 *  Supported options:
 *
 *	"host-required=yes|no"	a hostname of IP must be present
 *	"port-required=yes|no"	a port number must be present
 *	"ipv6-brackets=yes|no"	keep enclosing brackets around IPv6 address
 *------------------------------------------------------------------------
 */
static void
parse_address_options(
	const char	*options,
	bool		*host_is_required,
	bool		*port_is_required,
	bool		*keep_ipv6_brackets)
{
    char	optval[100] = "";

    if (PM_opt_fetch(options, "host-required", optval, sizeof(optval)) > 0) {
	*host_is_required = (PM_strcmpi(optval, "yes") == 0);
    }
    if (PM_opt_fetch(options, "port-required", optval, sizeof(optval)) > 0) {
	*port_is_required = (PM_strcmpi(optval, "yes") == 0);
    }
    if (PM_opt_fetch(options, "ipv6-brackets", optval, sizeof(optval)) > 0) {
	*keep_ipv6_brackets = (PM_strcmpi(optval, "yes") == 0);
    }
} /* parse_address_options() */

/*------------------------------------------------------------------------
 *  Function	: parse_addrspec
 *  Purpose	: Parse an address specification string
 *  Method	:
 *  1. Determine the address family:
 *     If the addrspec starts with a colon, then only the port number is
 *     present and we'll use the preferred IP order ("CONN_IP_ORDER") to set
 *     the address family. Otherwise if the address is surrounded by square
 *     brackets, then the address family is AF_INET6 else it could be either an
 *     AF_INET6 or an AF_INET address.
 *
 *  2. Determine the address string:
 *     If an address is specified, then an IPv6 address should be surrounded by
 *     square brackets. If the brackets are missing, an attempt is first made
 *     to process it as an IPv6 address before testing for an IPv4 address.
 *     If the address is neither an IPv6 nor an IPv4 IP number, then a hostname
 *     is assumed and the preferred IP order ("CONN_IP_ORDER") is used to set
 *     the address family.
 *
 *     If the preferred address family supplied in the interface is either
 *     AF_INET or AF_INET6, then the supplied address must be an IPv4 address
 *     in case of AF_INET and an IPv6 address in case of AF_INET6.
 *
 *  3. Validate the address string with test_ip_address().
 *
 *  4. Determine the port number:
 *     In case of an IPv6 address that's enclosed in square brackets, the port
 *     is specified after the colon following the closing bracket.
 *     In all other cases, the port number is located after the rightmost colon.
 *     If the port is not optional but missing from the addrspec, then an
 *     attempt is made to retrieve a default port value from the options string.
 *
 *  History:
 *  14-02-2011	: heesakkm : creation
 *  08-06-2011	: heesakkm : Introduced parse_address_options().
 *  09-06-2011	: eck      : Fixed test on empty string if both host and port optional
 *  23-06-2011	: heesakkm : Tweaked test_ip_address() error handling.
 *
 *  Notes	:
 *  The address specification should be of the form : [<address>][:<port>]
 *  where at least one of the two must be present.
 *  The address can be given as a hostname of as an IPv4 or IPv6-address.
 *  IPv6-addresses should be surrounded by square brackets in order to supply a
 *  port number.
 *
 *  If a port is not specified, an attempt is made to retrieve a default value
 *  from the options string.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
parse_addrspec(
	const char	*addrspec,		/* address specification */
	const char	*options,		/* options string */
	const char	*caller,		/* name of calling function */
	bool		host_is_required,
	bool		port_is_required,
	bool		keep_ipv6_brackets,
	CSBuf_p 	address,		/* OUT (optional) */
	CSBuf_p 	service,		/* OUT (optional) */
	int		*family)		/* IN/OUT: (optional) */
{
    static const char	fname[] = "parse_addrspec";
    CS_LOCAL(		csfname, 256);
    char		optval[100] = "";
    const char		*orig_addrspec;		/* only for error logging */
    const char		*port = NULL;
    size_t		addrlen = 0;
    int			address_family = AF_UNSPEC;
    int			preferred_family = AF_UNSPEC;
    bool		literal_ipv6 = FALSE;
    bool		allow_hostname = TRUE;
    int			iptest;

    if (addrspec == NULL) addrspec = "";
    orig_addrspec = addrspec;

    CSPrintf(csfname, "%s.%s", caller, fname);

    /* Overrule the default supplied option values if requested by the options
     * string.
     */
    parse_address_options(options, &host_is_required, &port_is_required,
			    &keep_ipv6_brackets);

    /* Init the output values. */
    if (family != NULL) preferred_family = *family;
    if ((preferred_family != AF_INET) && (preferred_family != AF_INET6)) {
	preferred_family = AF_UNSPEC;
    }
    if (family != NULL) *family = AF_UNSPEC;
    if (address != NULL) CSClear(address);
    if (service != NULL) CSClear(service);

    if (*addrspec == '\0') {
	if (host_is_required || port_is_required) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "Missing or empty address specification");
	    return FALSE;
	}
	return TRUE;
    }

    while (PM_is_ascii_space(*addrspec)) addrspec++;

    if ((addrspec[0] == ':') && (addrspec[1] != ':')) {	/* ":<port>" -> port only */
	/* Since we don't have an address from which to determine the protocol
	 * family, we'll leave it to the caller to decide what family to use.
	 */
	address_family = preferred_family;
	port = addrspec;
	addrlen = 0;
    } else if (addrspec[0] == '[') {	/* Literal IPv6 address */
	const char	*endp = strchr(addrspec, ']');
	const char	*p;

	if (endp == NULL) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "Invalid address: \"%s\", expected format: "
		    "<hostname> | <IPv4 address> | '[' <IPv6 address> ']'",
		    addrspec);
	    return FALSE;
	}
	address_family = AF_INET6;
	port = strrchr(endp, ':');
	if ((port != NULL) && (port != endp+1)) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "Invalid address: \"%s\", expected format: <host>[:<port>]",
		    addrspec);
	    return FALSE;
	}
	addrspec++;
	addrlen = endp - addrspec;
	literal_ipv6 = TRUE;
	/* If the address contains one or more colons, then it's an IP number
	 * instead of an IPv6 hostname.
	 */
	allow_hostname = !(((p = strchr(addrspec, ':')) != NULL) && (p < endp));

    } else {
	/* Test for IPv6 address without surrounding brackets.
	 *
	 * If an IPv6 address without brackets has ":<port>" appended to the
	 * end, then test_ip_address() fails and the address will be
	 * interpreted as a hostname with a port number.
	 */
	address_family = AF_INET6;
	if (test_ip_address(addrspec, &address_family) > 0) {
	    port = NULL;
	    addrlen = strlen(addrspec);
	    allow_hostname = FALSE;
	} else {		/* IPv4 address or hostname or IPv6:<port> */
	    CS_DYNAMIC(	csaddr);

	    port = strrchr(addrspec, ':');
	    addrlen = (port == NULL) ? strlen(addrspec) : (size_t)(port - addrspec);

	    (void) CSSetStringN(csaddr, addrspec, addrlen);

	    /* It's unclear what we have now: use AF_UNSPEC and let
	     * test_ip_address() decide.
	     */
	    address_family = AF_UNSPEC;
	    iptest = test_ip_address(CSContents(csaddr), &address_family);
	    if (iptest < 0) {
		ERWrite(APPLE_PARAM, CSContents(csfname),
			"inet_pton() failed for %s address \"%s\": %s (errno %d)",
			IPVSTR(address_family), CSContents(csaddr),
			strerror(errno), errno);
		CSDestroy(csaddr);
		return FALSE;
	    }
	    if (iptest == 0) {
		/* We don't have a valid IP address, so at this point we assume
		 * that a hostname was supplied. We'll leave it to the caller
		 * to decide what address family to use.
		 */
		if (address_family == AF_UNSPEC) {
		    address_family = preferred_family;
		}
	    }
	    CSDestroy(csaddr);
	}
    }

    /* Test whether the address matches the preferred address family. */
    if (preferred_family != AF_UNSPEC) {
	/* If we don't have an address, then use the preferred family. */
	if (addrlen == 0) address_family = preferred_family;

	if (address_family != preferred_family) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "%s address expected but %s address received: %s",
		    IPVSTR(preferred_family), IPVSTR(address_family),
		    orig_addrspec);
	    return FALSE;
	}
    }

    /* Set the actual address family value. */
    if (family != NULL) {
	*family = address_family;
    }

    /* Validate the IP address. */
    if (addrlen > 0) {
	CS_DYNAMIC(	addrbuf);

	if ((address != NULL) && (addrlen > CSSize(address))) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "Address too long (%lu > %lu): \"%.*s\"",
		    (unsigned long) addrlen, (unsigned long) CSSize(address),
		    (int) addrlen, addrspec);
	    CSDestroy(addrbuf);
	    return FALSE;
	}
	(void) CSSetStringN(addrbuf, addrspec, addrlen);

	/* If iptest()==0, then most likely a network hostname was provided.
	 * In that case, use of getaddrinfo() via set_addrinfo() by the
	 * caller should resolve the network address.
	 */
	iptest = test_ip_address(CSContents(addrbuf), &address_family);
	if (iptest < 0) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "inet_pton() failed for %s address \"%s\": %s (errno %d)",
		    IPVSTR(address_family), CSContents(addrbuf),
		    strerror(errno), errno);
	    CSDestroy(addrbuf);
	    return FALSE;
	} else if ((iptest == 0) && !allow_hostname) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "inet_pton() failed for %s address \"%s\"",
		    IPVSTR(address_family), CSContents(addrbuf));
	    CSDestroy(addrbuf);
	    return FALSE;
	}

	if (address != NULL) {
	    CSSetStringN(address, CSContents(addrbuf), CSLength(addrbuf));

	    /* Enclose the literal IPV6 address in square brackets if
	     * requested.
	     */
	    if (literal_ipv6 && keep_ipv6_brackets) {
		CSInsertN(address, 0, "[", 1);
		CSAppendC(address, ']');
	    }
	}
	CSDestroy(addrbuf);
    } else {
	if (host_is_required) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "No host specified in %s address \"%s\"",
		    IPVSTR(address_family), orig_addrspec);
	    return FALSE;
	}
    }

    /* Determine the port number. */
    if ((port == NULL) || (*port == '\0')) {
	/* Use a default port value, if provided. */
	if (port_is_required) {
	    if (PM_opt_fetch(options, "default-port",
				optval, sizeof(optval)) > 0) {
		port = optval;
	    }
	}
    } else {
	port++;
    }
    /* Validate the port number. */
    if ((port == NULL) || (*port == '\0')) {
	if (port_is_required) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "No port specified in %s address \"%s\"",
		    IPVSTR(address_family), orig_addrspec);
	    return FALSE;
	}
    } else {
	size_t	portlen = strlen(port);

	if (strspn(port, "0123456789") != portlen) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "Port not all digits in %s address: \"%s\"",
		    IPVSTR(address_family), orig_addrspec);
	    return FALSE;
	}
	if (service != NULL) {
	    if (strlen(port) > CSSize(service)) {
		ERWrite(APPLE_PARAM, CSContents(csfname),
			"Port too long (%lu > %lu) in %s address: \"%s\"",
			(unsigned long) portlen, (unsigned long) CSSize(service),
			IPVSTR(address_family), orig_addrspec);
		return FALSE;
	    }
	    (void) CSSetString(service, port);
	}
    }

    return TRUE;
} /* parse_addrspec() */

/*------------------------------------------------------------------------
 *  Function	: CONN_parse_addrspec
 *  Purpose	: Parse address spec to determine address, service and family
 *  Method	: parse_addrspec()
 *
 *  History:
 *  19-04-2011	: heesakkm : creation
 *  20-06-2011	: heesakkm : Use get_preferred_family() for unspecified family.
 *  28-06-2011	: heesakkm : Return the address type instead of the family.
 *
 *  Notes	:
 *  Caller should only use the first position of the returned address type.
 *
 *  The default syntax of the address specification is:
 *
 *	<ipv4-addr> | <ipv6-addr> ':' <tcp-port>
 *
 *  Description	:
 *  Parse an address specification to determine the address, service (port
 *  number) and address type.
 *
 *  A NULL pointer value can be supplied for the address, service and/or
 *  address_type to indicate that no value has to be returned for that specific
 *  variable. However, supplying a NULL pointer does not mean that the address,
 *  service and/or address_type will be ignored when parsing the address
 *  specification. It only means that the parsed result will not be returned to
 *  the caller.
 *
 *  If the address (host) is optional, then "host-required=no" should be set in
 *  the options string. If the service (port) is optional, then
 *  "port-required=no" should be set in the options string.
 *
 *  The option "default-port=<number>" can be used to supply a port number that
 *  will be used if a port number is missing from the address specification.
 *
 *  The address type is set to CONN_ADDRESS_IPV4 if the address family is
 *  AF_INET and set to CONN_ADDRESS_IPV6 if the address family is AF_INET6.
 *  Otherwise the type is set to CONN_ADDRESS_UNKNOWN.
 *
 *  If an IPv6 address in the address specification is enclosed in square
 *  brackets, then the result in the address buffer will contain square
 *  brackets too, unless option "ipv6-brackets=no" is used.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
CONN_parse_addrspec(
	const char	*addrspec,		/* address specification */
	const char	*options,		/* options string */
	CSBuf_p 	address,		/* OUT (optional) */
	CSBuf_p 	service,		/* OUT (optional) */
	char		*address_type)		/* OUT (optional) */
{
    static const char	fname[] = "CONN_parse_addrspec";
    int			family = AF_UNSPEC;
    bool		ok;

    ok = parse_addrspec(addrspec, options, fname,
			TRUE,		/* host_is_required */
			TRUE,		/* port_is_required */
			TRUE,		/* keep_ipv6_brackets */
			address, service, &family);

    if (address_type != NULL) {
	/* Only translate the internal family type if the caller is interested
	 * in the result.
	 */
	if (ok && (family == AF_UNSPEC)) {
	    family = get_preferred_family(options);
	}
	if (family == AF_INET6) {
	    *address_type = CONN_ADDRESS_IPV6;
	}  else if (family == AF_INET) {
	    *address_type = CONN_ADDRESS_IPV4;
	} else {
	    *address_type = CONN_ADDRESS_UNKNOWN;
	}
    }
    return ok;
} /* CONN_parse_addrspec() */

/*------------------------------------------------------------------------
 *  Function	: set_conn_nonblock
 *  Purpose	: Set or clear the non-blocking flag of a file-descriptor
 *  Method	: fcntl()
 *
 *  History:
 *  01-04-2011	: heesakkm : creation
 *
 *  Returns	: FALSE if fcntl() fails
 *------------------------------------------------------------------------
 */
static bool
set_conn_nonblock(
	int		sockfd,		/* File descriptor */
	bool		nonblock_flag,	/* TRUE to set, FALSE to clear bit */
	const char	*caller,	/* Name of calling function */
	bool		verbose_err)	/* TRUE=ERWrite(), FALSE=ERWarning() */
{
    static const char	fname[] = "set_conn_nonblock";
    CS_LOCAL(		csfname, 256);
    int			flags;	/* fcntl() flags */

    CSPrintf(csfname, "%s.%s", caller, fname);

    flags = fcntl(sockfd, F_GETFL);
    if (flags == -1) {
	if (verbose_err) {
	    ERWrite(APPLE_FILEWRITE, CSContents(csfname),
		    "fcntl(F_GETFL) failed: %s (errno %d)",
		    strerror(errno), errno);
	} else {
	    ERWarning(APPLE_FILEWRITE, CSContents(csfname),
		    "fcntl(F_GETFL) failed: %s (errno %d)",
		    strerror(errno), errno);
	}
	return FALSE;
    }
    if (nonblock_flag) {
	flags |= O_NONBLOCK;
    } else {
	flags &= ~O_NONBLOCK;
    }
    if (fcntl(sockfd, F_SETFL, flags) == -1) {
	if (verbose_err) {
	    ERWrite(APPLE_FILEWRITE, CSContents(csfname),
		    "fcntl(F_SETFL, %sO_NONBLOCK) failed: %s (errno %d)",
		    nonblock_flag ? "" : "~", strerror(errno), errno);
	} else {
	    ERWarning(APPLE_FILEWRITE, CSContents(csfname),
		    "fcntl(F_SETFL, %sO_NONBLOCK) failed: %s (errno %d)",
		    nonblock_flag ? "" : "~", strerror(errno), errno);
	}
	return FALSE;
    }
    return TRUE;
} /* set_conn_nonblock() */

/*------------------------------------------------------------------------
 *  Function	: CONN_nonblock
 *  Purpose	: Set or clear the non-blocking flag of a file-descriptor
 *  Method	: set_conn_nonblock()
 *
 *  History:
 *  01-04-2011	: heesakkm : creation
 *
 *  Description	:
 *  Set or clear the non-blocking flag O_NONBLOCK on a descriptor. A second
 *  parameter value of TRUE will set O_NONBLOCK on the descriptor, and a value
 *  of FALSE will clear it. This function returns TRUE if the option was
 *  successfully set or cleared, or FALSE otherwise.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
CONN_nonblock(
	int	sockfd,			/* File descriptor */
	bool	nonblock_flag)		/* TRUE to set, FALSE to clear bit */
{
    static const char	fname[] = "CONN_nonblock";

    return set_conn_nonblock(sockfd, nonblock_flag, fname, FALSE);
} /* CONN_nonblock() */

/*------------------------------------------------------------------------
 *  Function	: CONN_close_on_exec
 *  Purpose	: Set or clear the close-on-exec flag of a file-descriptor
 *  Method	: fcntl()
 *
 *  History:
 *  17-05-2000	: eck      : creation
 *
 *  Notes	: Errors are (silently) ignored.
 *
 *  Description	:
 *  Set or clear the close-on-exec flag FD_CLOEXEC on a descriptor. A second
 *  parameter value of TRUE will set FD_CLOEXEC on the descriptor, and a value
 *  of FALSE will clear it. This function returns TRUE if the option was
 *  successfully set or cleared, or FALSE otherwise.
 *
 *  Returns	: FALSE if fcntl() fails
 *------------------------------------------------------------------------
 */
bool
CONN_close_on_exec(
	int	sockfd,			/* File descriptor */
	bool	close_flag)		/* TRUE to set, FALSE to clear bit */
{
    int	flags;				/* fcntl() flags */

    flags = fcntl(sockfd, F_GETFD, 0);
    if (flags != -1) {
	if (close_flag) {
	    flags |= FD_CLOEXEC;
	} else {
	    flags &= ~FD_CLOEXEC;
	}
	if (fcntl(sockfd, F_SETFD, flags) != -1) {
	    return TRUE;
	}
    }
    return FALSE;
} /* CONN_close_on_exec() */

/*------------------------------------------------------------------------
 *  Function	: CONN_bufsize
 *  Purpose	: Set the send or receive buffer size on a file-descriptor
 *  Method	: setsockopt()
 *
 *  History:
 *  01-04-2011	: heesakkm : creation
 *
 *  Description	:
 *  Set the size of the send/receive buffer on a descriptor. If the second
 *  parameter is set to SO_SNDBUF, then the bufsize is used to set the size of
 *  the send buffer. If the second parameter is set to SO_RCVBUF, then the
 *  bufsize is used to set the size of the receive buffer. This function returns
 *  TRUE if the buffer size was successfully set, or FALSE otherwise.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
CONN_bufsize(
	int	sockfd,			/* File descriptor */
	int	buftype,		/* SO_SNDBUF or SO_RCVBUF */
	int	bufsize)		/* Size of the buffer */
{
    static const char	fname[] = "CONN_bufsize";

    if (bufsize > 0) {
	if ((buftype == SO_SNDBUF) || (buftype == SO_RCVBUF)) {
	    return conn_setsockopt(sockfd, SOL_SOCKET, buftype,
				    bufsize, fname, FALSE);
	} else {
	    ERWrite(APPLE_PARAM, fname,
		    "Invalid buffer type: %d; "
		    "expected %d (SO_SNDBUF) or %d (SO_RCVBUF)",
		    buftype, SO_SNDBUF, SO_RCVBUF);
	    return FALSE;
	}
    }
    return TRUE;
} /* CONN_bufsize() */

/*------------------------------------------------------------------------
 *  Function	: CONN_nodelay
 *  Purpose	: Set the TCP_NODELAY option on a file-descriptor
 *  Method	: setsockopt()
 *
 *  History:
 *  20-05-2005	: michael  : creation
 *  16-03-2011	: heesakkm : Use conn_setsockopt().
 *
 *  Description	:
 *  Set or clear the TCP_NODELAY socket option on a descriptor. The descriptor
 *  should be for an opened TCP/IP socket. A second parameter value of TRUE
 *  will set TCP_NODELAY on the descriptor, and a value of FALSE will clear it.
 *  This function returns TRUE if the option was successfully set or cleared,
 *  or FALSE otherwise.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
CONN_nodelay(
	int	sockfd,			/* File descriptor */
	bool	nodelay_flag)		/* TRUE to enable, FALSE to disable */
{
    static const char	fname[] = "CONN_nodelay";
    const int		optval = nodelay_flag ? 1 : 0;

    return conn_setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY,
			    optval, fname, FALSE);
} /* CONN_nodelay() */

/*------------------------------------------------------------------------
 *  Function	: CONN_keepalive
 *  Purpose	: Set the SO_KEEPALIVE option on a file-descriptor
 *  Method	: setsockopt()
 *
 *  History:
 *  01-04-2011	: heesakkm : creation
 *
 *  Description	:
 *  Set or clear the SO_KEEPALIVE socket option on a descriptor. A second
 *  parameter value of TRUE will set TCP_NODELAY on the descriptor, and a value
 *  of FALSE will clear it. This function returns TRUE if the option was
 *  successfully set or cleared, or FALSE otherwise.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
CONN_keepalive(
	int	sockfd,			/* File descriptor */
	bool	keepalive_flag)		/* TRUE to enable, FALSE to disable */
{
    static const char	fname[] = "CONN_keepalive";
    const int		optval = keepalive_flag ? 1 : 0;

    return conn_setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE,
			    optval, fname, FALSE);
} /* CONN_keepalive() */

/*------------------------------------------------------------------------
 *  Function	: CONN_my_name
 *  Purpose	: Determine nodename of the current machine.
 *  Method	: Just use uname()
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *
 *  Returns	: The nodename field of the utsname structure.
 *------------------------------------------------------------------------
 */
const char *
CONN_my_name(void)
{
    static const char		fname[] = "CONN_my_name";
    static struct utsname	name;
    static bool			been_here = FALSE;

    if (! been_here) {
	been_here = TRUE;
	if (uname(&name) < 0) {
	    /* Bummer. This is almost impossible, though, since the only reason
	     * for uname() to fail would be illegal address of 'name', in which
	     * case errno is set to EFAULT.
	     */
	    name.nodename[0] = '\0';
	    ERWrite(APPLE_PARAM, fname, "uname() failed: %s (errno %d)",
		    strerror(errno), errno);
	}
    }
    return name.nodename;
} /* CONN_my_name() */

/*------------------------------------------------------------------------
 *  Function	: get_my_ipaddr
 *  Purpose	: Get the IP address of the server for the specified family
 *  Method	: get_my_address_details()
 *
 *  History:
 *  15-06-2011	: heesakkm : creation
 *
 *  Returns	: String with the address or NULL.
 *------------------------------------------------------------------------
 */
static const char *
get_my_ipaddr(
	int		family,		/* Address family */
	char		*ipbuf,		/* OUT */
	size_t		ipbuf_len)
{
    static CONN_addrdata_t	addrdata;
    static const char		*retval = NULL;
    static int			last_family = AF_UNSPEC;

    /* Only update addrdata if needed. */
    if (family != last_family) {
	last_family = family;
	retval = get_my_address_details(family, NULL, &addrdata);
    }
    ipbuf[0] = '\0';
    if (retval != NULL) {
	return PM_strncpy(ipbuf, addrdata.ca_ipnum, ipbuf_len);
    }
    return NULL;
} /* get_my_ipaddr() */

/*------------------------------------------------------------------------
 *  Function	: CONN_my_ipaddr_opts
 *  Purpose	: Determine the (or 'a') IP address of the server
 *  Method	:
 *  Determine the IP address with get_my_address_details(), using the preferred
 *  protocol order from the environment variable CONN_IP_ORDER or from the
 *  options string.
 *
 *  History:
 *  15-06-2011	: heesakkm : creation
 *
 *  Notes	:
 *  The return value is an IP address string. This circumvents the need for
 *  subsequent name-lookups.
 *
 *  The address is determined every time the preferred IP order changes, so in
 *  most cases only once during the lifetime of a process.
 *
 *	IP order=6,4	->	try IPv4 if IPv6 fails
 *	IP order=6	->	use IPv6 only
 *
 *  Supported options:
 *
 *	"ip-order=<num>[,<num>]"  (preferred IP order)
 *	"ipv6-brackets=yes|no"    (add enclosing brackets around IPv6 address)
 *
 *  Description	:
 *  This function will return the IP address of the current machine.
 *
 *  The option "ip-order=<num>[,<num>]" can be used to override the configured
 *  environment variable CONN_IP_ORDER.
 *
 *  Use option "ipv6-brackets=yes" to force the use of square brackets around
 *  the returned IPv6 address (the default option is "ipv6-brackets=no").
 *
 *  Returns	: String with the address or NULL.
 *------------------------------------------------------------------------
 */
const char *
CONN_my_ipaddr_opts(
	const char	*options)
{
    static CONN_addrdata_t	addrdata;
    static char			ip6buf[INET6_ADDRSTRLEN + 2]; /* room for brackets */
    static char			last_ip_order[20] = "";
    static const char		*retval = NULL;
    static int			last_family = AF_UNSPEC;
    int				family = AF_UNSPEC;
    char			optval[100];
    const char			*ip_order = NULL;

    ip_order = get_ip_order(options, optval, sizeof(optval));

    if (!STREQ(last_ip_order, ip_order)) {
	const char	*ptr;

	PM_strncpy(last_ip_order, ip_order, sizeof(last_ip_order));

	for (ptr = ip_order; *ptr != '\0'; ptr++) {
	    if (*ptr == CONN_IP_VERSION4) {
		family = AF_INET;
	    } else if (*ptr == CONN_IP_VERSION6) {
		family = AF_INET6;
	    } else {
		continue;
	    }
	    /* Only update addrdata if needed. */
	    if (family != last_family) {
		last_family = family;
		retval = get_my_address_details(family, NULL, &addrdata);
	    }
	    if (retval != NULL) break;
	}
    }
    if ((retval != NULL) && (family == AF_INET6)) {
	bool	add_brackets = FALSE;

	if (PM_opt_fetch(options, "ipv6-brackets", optval, sizeof(optval)) > 0) {
	    add_brackets = (PM_strcmpi(optval, "yes") == 0);
	}
	sprintf(ip6buf, "%s%s%s",
		    (add_brackets) ? "[" : "", retval,
		    (add_brackets) ? "]" : "");
	retval = ip6buf;
    }
    return retval;
} /* CONN_my_ipaddr_opts() */

/*------------------------------------------------------------------------
 *  Function	: CONN_my_ipaddr
 *  Purpose	: Determine the (or 'a') IP address of the server
 *  Method	: CONN_my_ipaddr_opts()
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *  24-05-2000	: eck      : return NULL on failure
 *  08-06-2000	: eck      : Use ioctl calls instead of uname/gethostbyname
 *			   : This avoids DNS/DHCP problems.
 *  10-03-2004	: heesakkm : Clear malloc()-ed memory (avoid dbx rtc messages).
 *  31-05-2007	: heesakkm : Added len-fix for Linux (by Jaap Lutz/GBV).
 *  22-03-2011	: heesakkm : Use get_my_ipaddr().
 *  15-06-2011	: heesakkm : Use CONN_my_ipaddr_opts().
 *
 *  Notes	: The IPv6 address is not surrounded by square brackets.
 *
 *  Description	:
 *  This function will return the IP address of the current machine. It does
 *  this by first determining the name of the machine, and then using DNS to
 *  resolve it. If this fails, the returned string will be NULL.
 *
 *  A machine can have multiple IP addresses. The first one found will be
 *  returned.
 *  Problems can occur when the machine uses DHCP to determine its IP address
 *  and there is no coupling between the DHCP- and DNS-server.
 *
 *  The function CONN_ip_order() is used to determine for which IP version an
 *  attempt to retrieve the IP address will be made first.
 *
 *  An IPv6 address will not be enclosed in square brackets due to the
 *  hardcoded option "ipv6-brackets=no". Use CONN_my_ipaddr_opts() in
 *  combination with the option "ipv6-brackets=yes" if square brackets are
 *  required around the returned IPv6 address.
 *
 *  Returns	: String with the address or NULL.
 *------------------------------------------------------------------------
 */
const char *
CONN_my_ipaddr(void)
{
    return CONN_my_ipaddr_opts("ipv6-brackets=no");
} /* CONN_my_ipaddr() */

#if ! defined(NO_LOCAL)
/*------------------------------------------------------------------------
 *  Function	: CONN_connect_local
 *  Purpose	: Connect to a Local-server (aka Unix-domain)
 *  Method	: socket(AF_LOCAL), connect()
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *
 *  Notes	: Truncation of the filename is more-or-less silent.
 *
 *  Description	:
 *  Open a connection to a local domain server. The supplied addr should be the
 *  absolute path of the file used for rendezvous.
 *
 *  If successful, the file descriptor will have the close-on-exec flag set.
 *
 *  Returns	: File-descriptor for the connection or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_local(
	const char	*addr)		/* name of the file for rendez-vous */
{
    static const char	fname[] = "CONN_connect_local";
    struct sockaddr_un	server;
    size_t		addrlen = 0;
    int			sockfd;

    if (addr != NULL) {
	addrlen = strlen(addr);
    }
    if (addrlen == 0) {
	ERWrite(APPLE_FILEOPEN, fname,
		"No filename for Local-domain socket specified");
	return -1;
    }
    sockfd = socket(AF_LOCAL, SOCK_STREAM, IPPROTO_IP);
    if (sockfd < 0) {
	ERWrite(APPLE_FILEOPEN, fname, "socket() failed: %s (errno %d)",
		strerror(errno), errno);
	return -1;
    }
    initialize_socket(sockfd, AF_LOCAL, FALSE);

    server.sun_family = AF_LOCAL;
    if (addrlen >= sizeof(server.sun_path)) {
	addrlen = sizeof(server.sun_path) - 1;
	ERWarning(APPLE_FILEOPEN, fname,
		"Filename \"%s\" truncated to %lu positions", addr, (unsigned long)addrlen);
    }
    strncpy(server.sun_path, addr, addrlen);
    server.sun_path[addrlen] = '\0';

    if (connect(sockfd, (struct sockaddr *) &server,
		sizeof(server.sun_family) + addrlen) != 0) {
	close(sockfd);
	ERWrite(APPLE_FILEOPEN, fname, "Could not connect to %s: %s (errno %d)",
		server.sun_path, strerror(errno), errno);
	return -1;
    }
    (void) CONN_close_on_exec(sockfd, TRUE);
    return sockfd;
} /* CONN_connect_local() */
#endif /* ! defined(NO_LOCAL) */

/*------------------------------------------------------------------------
 *  Function	: sock_connect
 *  Purpose	: Create and connect a socket
 *  Method	:
 *  1. Allocate a socket descriptor with socket().
 *  2. Set the options with setsockopt() via conn_setsockoptions().
 *  3. Connect the socket to a server with connect().
 *
 *  History:
 *  14-02-2011	: heesakkm : creation
 *  23-06-2011	: heesakkm : Set default setsockopt flag SO_KEEPALIVE=1
 *
 *  Notes	:
 *  The option "SOL_SOCKET-SO_KEEPALIVE" is set by default if missing.
 *  The socket descriptor is set to non-blocking if requested via options
 *  string.
 *
 *  Returns	: File descriptor or -1
 *------------------------------------------------------------------------
 */
static int
sock_connect(
	struct addrinfo	*addr,
	const char	*options,
	const char	*caller)	/* Name of calling function */
{
    static const char	fname[] = "sock_connect";
    static CSBuf_p	csopts = NULL;
    CS_LOCAL(		csfname, 256);
    char		optval[100];
    const char		*opt;
    int			sockfd;
    bool		nonblock;

    if (! CreateCSbuf(&csopts, fname)) return -1;

    if ((addr->ai_family != AF_INET) && (addr->ai_family != AF_INET6)) {
	return -1;
    }
    CSPrintf(csfname, "%s.%s", caller, fname);

    CSSetString(csopts, options);

    /* Default flag: periodically test if connection is still alive. */
    opt = "SOL_SOCKET-SO_KEEPALIVE=";
    if (strstr(CSContents(csopts), opt) == NULL) {
	CSInsertN(csopts, 0, "1;", 2);
	CSInsertN(csopts, 0, opt, strlen(opt));
    }
    nonblock = (PM_opt_fetch(CSContents(csopts), "non-blocking",
					optval, sizeof(optval)) > 0)
		&& (PM_strcmpi(optval, "yes") == 0);

    sockfd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
    if (sockfd < 0) {
#if CONN_DEBUG
	ERWrite
#else
	ERWarning
#endif
	(APPLE_FILEOPEN, CSContents(csfname),
	 "socket() failed for family %d, socktype %d, protocol %d: %s (errno %d)",
	 addr->ai_family, addr->ai_socktype, addr->ai_protocol,
	 strerror(errno), errno);
	return -1;
    }
    initialize_socket(sockfd, addr->ai_family, FALSE);

    (void) conn_setsockoptions(sockfd, addr->ai_family, FALSE,
			       CSContents(csopts),
			       CSContents(csfname), FALSE);

#if 0
    This code may be too problematic for simple clients. In order to use it,
    EINPROGRESS should be checked when connect() fails, and the calling
    function should set an event for writability of the file-descriptor.

    if (setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY,
		    (char *)&on, sizeof(on)) < 0) {
        ERWarning(APPLE_FILEOPEN, fname,
                  "setsockopt(TCP_NODELAY): %s", strerror(errno));
    }
#endif

    /* Set socket descriptor to non-blocking, if requested. */
    if (nonblock) {
	if (! set_conn_nonblock(sockfd, TRUE, CSContents(csfname), TRUE)) {
	    close(sockfd);
	    return -1;
	}
    }

    /* Connect does the bind for us */
    if ((connect(sockfd, addr->ai_addr, addr->ai_addrlen) < 0)
	&& ((nonblock == FALSE) || (errno != EINPROGRESS))) {
#if CONN_DEBUG
	ERWrite
#else
	ERWarning
#endif
	(APPLE_FILEOPEN, fname,
	 "connect() failed: %s (errno %d)", strerror(errno), errno);
	close(sockfd);
	return -1;
    }
    return sockfd;
} /* sock_connect() */

/*------------------------------------------------------------------------
 *  Function	: sock_bind
 *  Purpose	: Create and bind a socket
 *  Method	:
 *  1. Allocate a socket descriptor with socket().
 *  2. Set the options with setsockopt() via conn_setsockoptions().
 *  3. Associate the socket with an IP address and port number with bind().
 *
 *  History:
 *  14-02-2011	: heesakkm : creation
 *  23-06-2011	: heesakkm : Set default setsockopt flag SO_REUSEADDR=1
 *  24-06-2011	: heesakkm : Set default setsockopt flag IPV6_V6ONLY=1
 *
 *  Notes	:
 *  The options "SOL_SOCKET-SO_REUSEADDR" and "IPPROTO_IPV6-IPV6_V6ONLY" are
 *  set by default if missing.
 *  The socket descriptor is set to close-on-exec if requested via options
 *  string.
 *
 *  Returns	: File descriptor or -1
 *------------------------------------------------------------------------
 */
static int
sock_bind(
	struct addrinfo	*addr,
	const char	*options,
	const char	*caller)	/* Name of calling function */
{
    static const char	fname[] = "sock_bind";
    static CSBuf_p	csopts = NULL;
    CS_LOCAL(		csfname, 256);
    char		optval[100];
    const char		*opt;
    int			sockfd;
    bool		cloexec;

    if (! CreateCSbuf(&csopts, fname)) return -1;

    if ((addr->ai_family != AF_INET) && (addr->ai_family != AF_INET6)) {
	return -1;
    }
    CSPrintf(csfname, "%s.%s", caller, fname);

    CSSetString(csopts, options);

    /* Default flag: allow local addresses to be reused. */
    opt = "SOL_SOCKET-SO_REUSEADDR=";
    if (strstr(CSContents(csopts), opt) == NULL) {
	CSInsertN(csopts, 0, "1;", 2);
	CSInsertN(csopts, 0, opt, strlen(opt));
    }
    /* Default flag: allow IPv4 and IPv6 sockets listening on the same port. */
    if (addr->ai_family == AF_INET6) {
	opt = "IPPROTO_IPV6-IPV6_V6ONLY=";
	if (strstr(CSContents(csopts), opt) == NULL) {
	    CSInsertN(csopts, 0, "1;", 2);
	    CSInsertN(csopts, 0, opt, strlen(opt));
	}
    }
    cloexec = (PM_opt_fetch(CSContents(csopts), "close-on-exec",
					optval, sizeof(optval)) > 0)
		&& (PM_strcmpi(optval, "yes") == 0);

    sockfd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
    if (sockfd < 0) {
#if CONN_DEBUG
	ERWrite
#else
	ERWarning
#endif
	(APPLE_FILEOPEN, CSContents(csfname),
	 "socket() failed for family %d, socktype %d, protocol %d: %s (errno %d)",
	 addr->ai_family, addr->ai_socktype, addr->ai_protocol,
	 strerror(errno), errno);
	return -1;
    }
    initialize_socket(sockfd, addr->ai_family, TRUE);

    /* Set socket descriptor to close-on-exec, if requested. */
    if (cloexec) {
	(void) CONN_close_on_exec(sockfd, TRUE);
    }
    if (! conn_setsockoptions(sockfd, addr->ai_family, TRUE,
				CSContents(csopts),
				CSContents(csfname), TRUE)) {
	close(sockfd);
	return -1;
    }

    if (bind(sockfd, addr->ai_addr, addr->ai_addrlen) != 0) {
#if CONN_DEBUG
	ERWrite
#else
	ERWarning
#endif
	(APPLE_FILEOPEN, CSContents(csfname),
	 "bind() failed: %s (errno %d)", strerror(errno), errno);
	close(sockfd);
	return -1;
    }
    return sockfd;
} /* sock_bind() */

/*------------------------------------------------------------------------
 *  Function	: get_address_description
 *  Purpose	: Print the address family, address and port into a CS
 *  Method	:
 *
 *  History:
 *  10-05-2011	: heesakkm : creation
 *
 *  Returns	: Pointer to CS
 *------------------------------------------------------------------------
 */
static const char *
get_address_description(
	int		family,		/* address family */
	const char	*address,
	const char	*port)
{
    static CS_LOCAL(	msg, 512);
    CS_LOCAL(		csbuf, 100);

    if (family != AF_UNSPEC) {
	CSPrintf(msg, "%s ", IPVSTR(family));
    }
    CSAppend(msg, "address");
    if ((address != NULL) && (*address != '\0')) {
	CSPrintf(csbuf, " \"%s\"", address);
	CSAppendN(msg, CSContents(csbuf), CSLength(csbuf));
    }
    if ((port != NULL) && (*port != '\0')) {
	CSPrintf(csbuf, " on port %s", port);
	CSAppendN(msg, CSContents(csbuf), CSLength(csbuf));
    }
    return CSContents(msg);
} /* get_address_description() */

/*------------------------------------------------------------------------
 *  Function	: set_addrinfo
 *  Purpose	: Load the socket address structures for the supplied address
 *  Method	: getaddrinfo()
 *
 *  History:
 *  13-04-2011	: heesakkm : creation
 *  07-06-2011	: heesakkm : Retrieve own IP if address and port are missing.
 *  14-06-2011	: heesakkm : Use unspecified address for bind() if none supplied
 *  27-06-2011	: heesakkm : Support link-local IPv6 addresses.
 *
 *  Notes	:
 *  In the call to getaddrinfo(), the address or service parameter, but not
 *  both, may be NULL.
 *
 *  Returns	: TRUE if the operation succeeded, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
set_addrinfo(
	const char	*address,
	const char	*service,
	int		family,		/* Protocol family for socket. */
	bool		bind_call,	/* Is result used in call to bind()? */
	const char	*caller,	/* Name of calling function */
	struct addrinfo	**res)		/* OUT */
{
    static const char	fname[] = "set_addrinfo";
    CS_LOCAL(		csfname, 256);
    CS_LOCAL(		csaddr, INET6_ADDRSTRLEN + 25);
    struct addrinfo	hints;
    int			error;

    if ((address != NULL) && (*address == '\0')) address = NULL;
    if ((service != NULL) && (*service == '\0')) service = NULL;

    CSPrintf(csfname, "%s.%s", caller, fname);

    /* Set the unspecified address to bind to if no address was supplied.
     * Note: for non-bind() calls, the absence of the AI_PASSIVE flag will
     * cause getaddrinfo() to use the loopback address.
     */
    if (bind_call && (address == NULL)) {
	address = (family == AF_INET6) ? "::" : "0.0.0.0";
    }
    /* For IPv6, the interface name must be added to a link-local address
     * so that getaddrinfo() can fill in the sin6_scope_id field correctly.
     * That is, it translates the "%<id>" part to a proper ID and sets the
     * sin6_scope_id field to the ID value.
     * A link-local address starts with "fe80::".
     */
    if ((family == AF_INET6) && (address != NULL)
	    && (PM_strncmpi(address, "fe80::", 6) == 0)) {
	static CONN_addrdata_t	addrdata;
	static const char	*ptr = NULL;

	/* Only update addrdata if needed. */
	if ((ptr == NULL) || !address_equal(addrdata.ca_ipnum, address)) {
	    ptr = get_my_address_details(AF_INET6, address, &addrdata);
	}
	if (ptr != NULL) {
	    CSSetString(csaddr, address);
	    CSAppendC(csaddr, '%');
	    CSAppend(csaddr, addrdata.ca_ifname);
	    address = CSContents(csaddr);
	}
    }

    /* getaddrinfo() loads up the "sockaddr" struct so we don't need to do that
     * manually. The hints parameter specifies the preferred address family,
     * socket type and additional options.
     */
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = family;
    hints.ai_socktype = SOCK_STREAM;
    /* AI_NUMERICSERV: If specified, a non-NULL service must point to a string
     * containing a numeric port number.
     */
    hints.ai_flags = AI_NUMERICSERV;
    /* AI_PASSIVE: When set, this flag specifies that the returned addresses
     * are suitable for use when binding a socket to accept incoming
     * connections for the specified service. The IP address portion of the
     * returned address structure is set to INADDR_ANY or IN6ADDR_ANY_INIT for
     * IPv4 or IPv6 addresses, respectively.
     * When not set, the returned addresses are suitable for use with connect().
     * The IP address portion of the socket address structure is set to the
     * loopback address.
     * This flag is ignored when address is not null.
     */
    if (bind_call) {
	hints.ai_flags |= AI_PASSIVE;
    }

    error = getaddrinfo(address, service, &hints, res);
    if (error != 0) {
	ERWrite(APPLE_PARAM, CSContents(csfname),
		"getaddrinfo() failed for %s: %s (errno %d)",
		get_address_description(family, address, service),
		gai_strerror(error), error);
	return FALSE;
    }
    return TRUE;
} /* set_addrinfo() */

/*------------------------------------------------------------------------
 *  Function	: connect_tcp_family
 *  Purpose	: Create a TCP/IP-connection for the supplied address family
 *  Method	:
 *  - Load the sockaddr struct with getaddrinfo().
 *  - Call sock_connect() for each returned address in the sockaddr struct
 *    until we have a successful connection.
 *
 *  History:
 *  20-06-2011	: heesakkm : creation
 *
 *  Notes	:
 *
 *  Returns	: File descriptor or -1
 *------------------------------------------------------------------------
 */
static int
connect_tcp_family(
	int		family,			/* Address family */
	const char	*address,
	const char	*service,
	const char	*options,
	const char	*caller)		/* Name of calling function */
{
    static const char	fname[] = "connect_tcp_family";
    CS_LOCAL(		csfname, 256);
    struct addrinfo	*res;
    struct addrinfo	*aip;
    int			sockfd = -1;

    CSPrintf(csfname, "%s.%s", caller, fname);

    /* Set the socket address structure. */
    if (! set_addrinfo(address, service, family, FALSE,
			CSContents(csfname), &res)) {
	return -1;
    }

    /* Loop through the list of returned addresses until a connection can be
     * made to one of them.
     */
    for (aip = res; aip != NULL; aip = aip->ai_next) {
	if ((sockfd = sock_connect(aip, options, CSContents(csfname))) < 0) {
	    continue;
	}
	break;		/* Connected successfully */
    }
    freeaddrinfo(res);	/* No longer needed */

    if (sockfd < 0) {
	/* Looped off the end of the list with no connection. */
	ERWarning(APPLE_FILEOPEN, CSContents(csfname),
		    "Could not connect to %s: %s (errno %d)",
		    get_address_description(family, address, service),
		    strerror(errno), errno);
	sockfd = -1;
    }
    return sockfd;
} /* connect_tcp_family() */

/*------------------------------------------------------------------------
 *  Function	: connect_tcp
 *  Purpose	: Create a TCP/IP-connection
 *  Method	:
 *  Retrieve the remote address (optional) and port (required) with
 *  parse_addrspec() and call connect_tcp_family() if the address family
 *  matches the preferred family. Otherwise call connect_tcp_family() in the
 *  preferred IP order until a socket is created successfully.
 *
 *  History:
 *  14-02-2011	: heesakkm : creation
 *  18-04-2011	: heesakkm : Introduced preferred_family.
 *  20-06-2011	: heesakkm : Call connect_tcp_family() using preferred IP-order.
 *  27-06-2011	: heesakkm : Fixed matching on preferred family.
 *
 *  Returns	: File descriptor or -1
 *------------------------------------------------------------------------
 */
static int
connect_tcp(
	const char	*addrspec,		/* address specification */
	int		preferred_family,	/* preferred address family */
	const char	*options,		/* options string */
	const char	*caller)		/* name of calling function */
{
    static const char	fname[] = "connect_tcp";
    CS_LOCAL(		csfname, 256);
    CS_LOCAL(		address, 100);
    CS_LOCAL(		service, 25);
    char		ip_order[20];
    const char		*ptr;
    int			family = preferred_family;
    int			sockfd = -1;

    CSPrintf(csfname, "%s.%s", caller, fname);

    /* Determine the address & port (aka service). */
    if (! parse_addrspec(addrspec, options, CSContents(csfname),
			    FALSE,	/* host_is_required */
			    TRUE,	/* port_is_required */
			    FALSE,	/* keep_ipv6_brackets */
			    address, service, &family)) {
	return -1;
    }

    /* If the address specification contains an address of the preferred
     * family, then create a socket for that address family.
     */
    if ((family != AF_UNSPEC)
	&& ((preferred_family == AF_UNSPEC) || (preferred_family == family))) {
	sockfd = connect_tcp_family(family, CSContents(address),
				    CSContents(service),
				    options, CSContents(csfname));
	if (sockfd < 0) {
	    ERWrite(APPLE_FILEOPEN, CSContents(csfname),
		    "Could not connect to %s: %s (errno %d)",
		    get_address_description(family, CSContents(address),
					    CSContents(service)),
					    strerror(errno), errno);
	    sockfd = -1;
	}
    } else {
	/* If the address specification only contains a port number, then we
	 * don't know for which IP version to create a socket. So iterate
	 * through the IP order string until a socket was created successfully.
	 */
	ptr = get_ip_order(options, ip_order, sizeof(ip_order));

	for (/*void*/; *ptr != '\0'; ptr++) {
	    if (*ptr == CONN_IP_VERSION4) {
		family = AF_INET;
	    } else if (*ptr == CONN_IP_VERSION6) {
		family = AF_INET6;
	    } else {
		continue;
	    }
	    sockfd = connect_tcp_family(family, CSContents(address),
					CSContents(service),
					options, CSContents(csfname));
	    if (sockfd < 0) continue;
	    break;		/* Connected successfully */
	}
	if (sockfd < 0) {
	    ERWrite(APPLE_FILEOPEN, CSContents(csfname),
		    "Could not connect to %s: %s (errno %d)",
		    get_address_description(AF_UNSPEC, CSContents(address),
					    CSContents(service)),
					    strerror(errno), errno);
	    sockfd = -1;
	}
    }
    return sockfd;
} /* connect_tcp() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_tcp4
 *  Purpose	: Create a TCP/IPv4-connection
 *  Method	: connect_tcp(AF_INET)
 *
 *  History:
 *  18-04-2011	: heesakkm : creation
 *  23-06-2011	: heesakkm : Renamed from CONN_connect_IPv4().
 *
 *  Description	:
 *  Open a TCP/IPv4 connection to a server.
 *  See CONN_connect_tcp_opts() for supported options.
 *
 *  Returns	: File descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_tcp4(
	const char	*addrspec,		/* address specification */
	const char	*options)		/* options string */
{
    static const char	fname[] = "CONN_connect_tcp4";

    return connect_tcp(addrspec, AF_INET, options, fname);
} /* CONN_connect_tcp4() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_tcp6
 *  Purpose	: Create a TCP/IPv6-connection
 *  Method	: connect_tcp(AF_INET6)
 *
 *  History:
 *  18-04-2011	: heesakkm : creation
 *  23-06-2011	: heesakkm : Renamed from CONN_connect_IPv6().
 *
 *  Description	:
 *  Open a TCP/IPv6 connection to a server.
 *  See CONN_connect_tcp_opts() for supported options.
 *
 *  Returns	: File descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_tcp6(
	const char	*addrspec,		/* address specification */
	const char	*options)		/* options string */
{
    static const char	fname[] = "CONN_connect_tcp6";

    return connect_tcp(addrspec, AF_INET6, options, fname);
} /* CONN_connect_tcp6() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_tcp_opts
 *  Purpose	: Create a TCP/IP-connection
 *  Method	: connect_tcp(AF_UNSPEC)
 *
 *  History:
 *  05-04-2011	: heesakkm : creation
 *
 *  Notes	:
 *  - The option "host-required=yes|no" is used to determine whether or not the
 *    presence of a host address is required (the default value "no").
 *  - The option "port-required=yes|no" is used to determine whether or not the
 *    presence of a host address is required (the default value "yes").
 *  - The option "ip-order=<num>[,<num>]" can be used to override the
 *    configured environment variable CONN_IP_ORDER.
 *  - The option "non-blocking=yes|no" is used to determine whether or not to
 *    make the socket non-blocking (the default value "no").
 *
 *  Description	:
 *  Open a TCP/IP connection to a server.
 *  See CONN_connect_tcp() for details.
 *
 *  Returns	: File descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_tcp_opts(
	const char	*addrspec,		/* address specification */
	const char	*options)		/* options string */
{
    static const char	fname[] = "CONN_connect_tcp_opts";

    return connect_tcp(addrspec, AF_UNSPEC, options, fname);
} /* CONN_connect_tcp_opts() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_tcp
 *  Purpose	: Create a TCP/IP-connection
 *  Method	: connect_tcp(AF_UNSPEC)
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *  14-02-2011	: heesakkm : Use connect_tcp().
 *
 *  Notes	:
 *  The tcp_spec should be a string of the form:
 *
 *	[<ipv4-addr> | <ipv6-addr>] ':' <tcp-port>
 *
 *  Description	:
 *  Open a TCP/IP connection to a server.
 *
 *  This function uses CONN_parse_addrspec() for parsing the supplied address
 *  specification. By default, the IP address is optional. A port number is
 *  mandatory. Use CONN_connect_tcp_opts() in combination with the option
 *  "host-required=yes|no" and/or "port-required=yes|no" if different values
 *  are required.
 *
 *  If the address family cannot be determined from the parsed address string,
 *  then CONN_ip_order() is used to determine for which IP version a connection
 *  attempt should be made first.
 *
 *  If successful, the file descriptor will have the close-on-exec flag set.
 *
 *  The underlying connect() will be blocking, which may take some time (as may
 *  name resolution).
 *
 *  Returns	: File descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_tcp(
	const char	*addrspec)		/* address specification */
{
    static const char	fname[] = "CONN_connect_tcp";

    return connect_tcp(addrspec, AF_UNSPEC, NULL, fname);
} /* CONN_connect_tcp() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_opts
 *  Purpose	: Connect to either a TCP-IP server or a Local-domain server
 *  Method	: Determine type of connection and call appropriate function.
 *
 *  History:
 *  05-04-2011	: heesakkm : creation
 *
 *  Notes	:
 *  The address is supposed to be one of the following:
 *
 *	T=<hostname>|<IPv4-address>:<port>
 *	S=<hostname>|<IPv6-address>:<port>
 *	L=/path/to/file
 *
 *  Description	:
 *  Connect to either a TCP/IP or a UNIX domain server.
 *  See CONN_connect_tcp_opts() for supported options for connecting to a
 *  TCP/IP domain server.
 *
 *  Returns	: File descriptor or -1
 *------------------------------------------------------------------------
 */
int
CONN_connect_opts(
	const char	*addrspec,		/* address specification */
	const char	*options)		/* options string */
{
    static const char	fname[] = "CONN_connect_opts";
    int			sockfd = -1;

    if ((addrspec == NULL) || (*addrspec == '\0')) {
	ERWrite(APPLE_PARAM, fname, "Missing or empty address specification");
	return -1;
    }
    if ((  (addrspec[0] != CONN_ADDRESS_IPV4)
	&& (addrspec[0] != CONN_ADDRESS_IPV6)
	&& (addrspec[0] != CONN_ADDRESS_LOCAL))
	    || (addrspec[1] != '=')) {
	ERWrite(APPLE_PARAM, fname,
		"Address not of form [TSL]=...: \"%s\"", addrspec);
	return -1;
    }
    if (addrspec[0] == CONN_ADDRESS_IPV4) {
	sockfd = connect_tcp(addrspec + 2, AF_INET, options, fname);
    } else if (addrspec[0] == CONN_ADDRESS_IPV6) {
	sockfd = connect_tcp(addrspec + 2, AF_INET6, options, fname);
#if ! defined(NO_LOCAL)
    } else {
	return CONN_connect_local(addrspec + 2);
#else
    } else {
	return -1;
#endif
    }
    if (sockfd == -1) {
	ERWrite(APPLE_FILEOPEN, fname,
		"connect_tcp() failed for \"%s\"", addrspec);
    }
    return sockfd;
} /* CONN_connect_opts() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect
 *  Purpose	: Connect to either a TCP-IP server or a Local-domain server
 *  Method	: CONN_connect_opts()
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *  05-04-2011	: heesakkm : Moved to CONN_connect_opts().
 *
 *  Notes	: See CONN_connect_opts() for details.
 *
 *  Description	:
 *  Connect to either a TCP/IP or a UNIX domain server (depending on whether
 *  addr starts with "T=" or "S=" for a TCP/IP server or "L=" for a UNIX domain
 *  server) using CONN_connect_tcp4(), CONN_connect_tcp6() or
 *  CONN_connect_local().
 *
 *  Returns	: File descriptor or -1
 *------------------------------------------------------------------------
 */
int
CONN_connect(
	const char	*addrspec)		/* address specification */
{
    return CONN_connect_opts(addrspec, NULL);
} /* CONN_connect() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_oneof_opts
 *  Purpose	: Connect to one of a list of servers
 *  Method	: Loop over addresses and return when a connect succeeds.
 *
 *  History:
 *  05-04-2011	: heesakkm : creation
 *
 *  Notes	:
 *  The addrspecs is a '|'-separated list of addresses, like:
 *
 *	L=/path/to/file|T=198.87.44.123:54321|S=[::FFFF:198.87.44.123]:54321
 *
 *  The order in which the addresses are listed in the address spec should
 *  adhere to the CONN_IP_ORDER environment variable.
 *
 *  Description	:
 *  Connect to one of a number of servers. The addresses are separated by '|'.
 *  They will be tried in order using CONN_connect_opts(). This function will
 *  return when a CONN_connect_opts() succeeds.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_oneof_opts(
	const char	*addrspecs,	/* '|' separated address list */
	const char	*options)	/* options string */
{
    CS_LOCAL(	addr, 100);

    if (addrspecs == NULL) return -1;

    while (*addrspecs != '\0') {
	const char	*end;
	int		sockfd;

	end = strchr(addrspecs, '|');
	if (end == NULL) end = addrspecs + strlen(addrspecs);
	CSSetStringN(addr, addrspecs, end - addrspecs);

	sockfd = CONN_connect_opts(CSContents(addr), options);
	if (sockfd >= 0) {
	    return sockfd;
	}
	if (*end == '\0') break;
	addrspecs = end + 1;
    }
    return -1;
} /* CONN_connect_oneof_opts() */

/*------------------------------------------------------------------------
 *  Function	: CONN_connect_oneof
 *  Purpose	: Connect to one of a list of servers
 *  Method	: CONN_connect_oneof_opts()
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *  05-04-2011	: heesakkm : Moved to CONN_connect_oneof_opts().
 *
 *  Notes	: See CONN_connect_oneof_opts() for details.
 *
 *  Description	:
 *  Connect to one of a number of servers. The addresses are separated by '|'.
 *  They will be tried in order using CONN_connect_opts() with options string
 *  NULL.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_connect_oneof(
	const char	*addrspecs)	/* '|' separated list */
{
    return CONN_connect_oneof_opts(addrspecs, NULL);
} /* CONN_connect_oneof() */

/*------------------------------------------------------------------------
 *  Function	: passive_tcp_family
 *  Purpose	: Open a TCP/IP socket for listening for the supplied family
 *  Method	: socket(), setsockopt(), bind(), listen()
 *
 *  History:
 *  20-06-2011	: heesakkm : creation
 *  22-06-2011	: heesakkm : Renamed from passive_tcp_opts().
 *  27-06-2011	: heesakkm : Remamed from passive_tcp().
 *  05-10-2011	: heesakkm : Added non-blocking support via options string.
 *
 *  Notes	:
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
static int
passive_tcp_family(
	CSBuf_p 	fullname,		/* value/result address:port */
	int		family,			/* Address family */
	const char	*address,
	const char	*service,
	const char	*options,
	int		backlog,
	const char	*caller)		/* Name of calling function */
{
    static const char		fname[] = "passive_tcp_family";
    CS_LOCAL(			csfname, 256);
    struct sockaddr_storage	server;
    struct addrinfo		*res = NULL;
    struct addrinfo		*aip;
    char			optval[100];
    char			ipbuf[INET6_ADDRSTRLEN];
    int				sockfd = -1;
    socklen_t			len;
    bool			nonblock;

    CSPrintf(csfname, "%s.%s", caller, fname);

    if (! set_addrinfo(address, service, family, TRUE,
			CSContents(csfname), &res)) {
	return -1;
    }

    /* Loop through the returned addresses and bind to the first we can. */
    for (aip = res; aip != NULL; aip = aip->ai_next) {
	if ((sockfd = sock_bind(aip, options, CSContents(csfname))) < 0) {
	    continue;
	}
	break;		/* Successful bind */
    }
    freeaddrinfo(res);	/* No longer needed */

    if (sockfd < 0) {
	/* Looped off the end of the list with no connection. */
	ERWrite(APPLE_FILEOPEN, CSContents(csfname),
		"Could not bind to %s address \"%s\": %s (errno %d)",
		IPVSTR(family), CSContents(fullname), strerror(errno), errno);
	return -1;
    }

    /* Find out what local address and port the system picked for us. */
    memset(&server, 0, sizeof(server));
    len = sizeof(server);
    if (getsockname(sockfd, (struct sockaddr *) &server, &len)) {
	ERWrite(APPLE_FILEOPEN, CSContents(csfname),
		"getsockname() failed for %s address \"%s\": %s (errno %d)",
		IPVSTR(family), CSContents(fullname), strerror(errno), errno);
	close(sockfd);
	return -1;
    }

    /* Get the actual family of the address that was selected. */
    family = ((struct sockaddr *)&server)->sa_family;

    /* We have a usable-socket. If the address if unspecified, then determine a
     * 'real' address, that is, an IP address that can be used to connect to
     * this socket. Note that there may be multiple IP addresses on a single
     * computer (and even on a single interface). get_my_ipaddr() just returns
     * the first one it finds.
     */
    if (ADDR_UNSPECIFIED(family, get_sin_addr((struct sockaddr *)&server))) {

	/* Retrieval of an IP address may fail if the address was unspecified
	 * and IPv6 is the preferred version in the CONN_IP_ORDER environment
	 * variable, but IPv6 is not activated.
	 */
	if (get_my_ipaddr(family, ipbuf, sizeof(ipbuf)) == NULL) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "Could not determine %s address", IPVSTR(family));
	    close(sockfd);
	    return -1;
	}
    } else {
	if (get_sockaddr_address(((struct sockaddr *)&server),
				    ipbuf, sizeof(ipbuf)) == NULL) {
	    ERWrite(APPLE_PARAM, CSContents(csfname),
		    "inet_ntop() failed for %s address \"%s\": %s (errno %d)",
		    IPVSTR(family), CSContents(fullname), strerror(errno), errno);
	    close(sockfd);
	    return -1;
	}
    }

    /* Unless non-blocking is disabled via the options string, we set the
     * listen socket to non-blocking to avoid the following scenario: A
     * connection is established, select() returns but before the accept() is
     * called the connection is aborted by the client and removed from the
     * kernel queue. A blocking accept() could then have to wait for the next
     * connection. Note that only some operating systems do this. Others will
     * return an error on the accept() call.
     */
    nonblock = (PM_opt_fetch(options, "non-blocking",
				optval, sizeof(optval)) == 0)
		|| (PM_strcmpi(optval, "no") != 0);
    if (! set_conn_nonblock(sockfd, nonblock, fname, TRUE)) {
	close(sockfd);
	return -1;
    }

    /* Specify a backlog for incoming connection requests. */
    if (listen(sockfd, backlog) != 0) {
        ERWrite(APPLE_FATAL, CSContents(csfname),
		"listen() failed for %s: %s (errno %d)",
		get_address_description(family, ipbuf, service),
					strerror(errno), errno);
        close(sockfd);
        return -1;
    }

    /* Modify fullname so it contains <ip-address>:<port> */
    CSPrintf(fullname, "%s%s%s:%s",
	    (family == AF_INET6) ? "[" : "", ipbuf,
	    (family == AF_INET6) ? "]" : "",
	    PM_itoa(NULL, get_sockaddr_port((struct sockaddr *)&server)));

    (void) CONN_close_on_exec(sockfd, TRUE);
    return sockfd;
} /* passive_tcp_family() */

/*------------------------------------------------------------------------
 *  Function	: passive_tcp
 *  Purpose	: Open a TCP/IP socket for listening
 *  Method	:
 *  Retrieve the address and port (both optional) from the supplied fullname
 *  buffer and call passive_tcp_family(). If fullname is empty, then call
 *  passive_tcp_family() in the preferred IP order until a passive socket is
 *  created successfully.
 *
 *  History:
 *  22-06-2011	: heesakkm : creation
 *  27-06-2011	: heesakkm : Renamed from passive_tcp_opts().
 *			     Fixed matching on preferred family.
 *
 *  Notes	:
 *  The fullname should be of the form: [<address>][:<port>]
 *  The address can be given as a name or an IP address. It should specify a
 *  valid IP address for the current machine.
 *  The port should be a number.
 *
 *  On success, the fullname will contain <ip-address>:<port>. The IP address
 *  returned will be valid for connecting to from other machines (except when
 *  dealing with masquerading firewalls in between).
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
static int
passive_tcp(
	CSBuf_p 	fullname,		/* value/result address:port */
	int		preferred_family,	/* Preferred address family */
	const char	*options,
	int		backlog,
	const char	*caller)		/* Name of calling function */
{
    static const char	fname[] = "passive_tcp";
    CS_LOCAL(		csfname, 256);
    CS_LOCAL(		address, 100);
    CS_LOCAL(		service, 25);
    char		ip_order[20];
    const char		*ptr;
    int			family = preferred_family;
    int			sockfd = -1;

    CSPrintf(csfname, "%s.%s", caller, fname);

    /* If fullname contains a host address, then create a passive socket for
     * the retrieved address family.
     */
    if (CSLength(fullname) > 0) {
	if (!parse_addrspec(CSContents(fullname), options, CSContents(csfname),
			    FALSE,	/* host_is_required */
			    FALSE,	/* port_is_required */
			    FALSE,	/* keep_ipv6_brackets */
			    address, service, &family)) {
	    return -1;
	}
	if ((CSLength(address) > 0) && (family != AF_UNSPEC)
	    && ((preferred_family == AF_UNSPEC) || (preferred_family == family))) {
	    return passive_tcp_family(fullname, family, CSContents(address),
					CSContents(service), options, backlog,
					CSContents(csfname));
	}
    }

    /* If fullname is empty or only contains a port number, then iterate
     * through the IP order string until a passive socket was created
     * successfully.
     */
    ptr = get_ip_order(options, ip_order, sizeof(ip_order));

    for (/*void*/; *ptr != '\0'; ptr++) {
	if (*ptr == CONN_IP_VERSION4) {
	    family = AF_INET;
	} else if (*ptr == CONN_IP_VERSION6) {
	    family = AF_INET6;
	} else {
	    continue;
	}
	sockfd = passive_tcp_family(fullname, family,
				    CSContents(address), CSContents(service),
				    options, backlog, CSContents(csfname));
	if (sockfd < 0) continue;
	break;		/* Connected successfully */
    }
    return sockfd;
} /* passive_tcp() */

/*------------------------------------------------------------------------
 *  Function	: CONN_passive_tcp_opts
 *  Purpose	: Open a TCP/IP socket for listening
 *  Method	: passive_tcp(AF_UNSPEC)
 *
 *  History:
 *  05-04-2011	: heesakkm : creation
 *  13-04-2011	: heesakkm : Let set_addrinfo() determine the IP-address instead
 *			     of CONN_my_ipaddr() in case no address was provided
 *  14-06-2011	: heesakkm : Handle unspecified address.
 *  16-06-2011	: heesakkm : Add brackets around IPv6 address in result buffer.
 *  20-06-2011	: heesakkm : Call passive_tcp_opts() using preferred IP-order.
 *  22-06-2011	: heesakkm : Use passive_tcp_opts().
 *
 *  Notes	:
 *  If fullname does not contain an address, then the the IP version of the
 *  socket is determined by the "ip-order" in the options string or
 *  alternatively the environment variable "CONN_IP_ORDER".
 *
 *  - The option "host-required=yes|no" is used to determine whether or not the
 *    presence of a host address is required (the default value "no").
 *  - The option "port-required=yes|no" is used to determine whether or not the
 *    presence of a host address is required (the default value "no").
 *  - The option "ip-order=<num>[,<num>]" can be used to override the
 *    configured environment variable CONN_IP_ORDER.
 *  - The option "close-on-exec=yes|no" is used to determine whether or not the
 *    socket should remain open across exec-calls (the default value "no").
 *  - The option "non-blocking=yes|no" is used to determine whether or not to
 *    make the socket non-blocking (the default value "yes").
 *
 *  Description	:
 *  Open a socket for accepting TCP/IP-connections.
 *  See CONN_passive_tcp() for details.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_passive_tcp_opts(
	CSBuf_p 	fullname,	/* value/result address:port */
	const char	*options,	/* options string */
	int		backlog)	/* size of internal queue */
{
    static const char	fname[] = "CONN_passive_tcp_opts";

    return passive_tcp(fullname, AF_UNSPEC, options, backlog, fname);
} /* CONN_passive_tcp_opts() */

/*------------------------------------------------------------------------
 *  Function	: CONN_passive_tcp
 *  Purpose	: Open a TCP/IP socket for listening
 *  Method	: passive_tcp(AF_UNSPEC)
 *
 *  History:
 *  16-05-2000	: eck      : creation
 *  24-05-2000	: eck      : Fixed stupid close-on-exec bug
 *  05-04-2011	: heesakkm : Moved to CONN_passive_tcp_opts().
 *
 *  Notes	:
 *  If fullname does not contain an address, then then the IP version of the
 *  socket is determined by the environment variable "CONN_IP_ORDER".
 *
 *  Description	:
 *  Open a socket for accepting TCP/IP-connections.
 *
 *  On entry, fullname should have the syntax:
 *
 *	[<ipv4-addr> | <ipv6-addr>] [':' <tcp-port>]
 *
 *  If the supplied address is a name, it will be resolved. If it is missing,
 *  all IP addresses of the machine (for the preferred protocol) will be
 *  listened to.
 *  If the port is missing, an ephemeral port number will be issued by the system.
 *
 *  This function uses CONN_parse_addrspec() for parsing the supplied address
 *  specification in fullname. By default, the IP address is optional as is the
 *  port number. Use CONN_passive_tcp_opts() in combination with the option
 *  "host-required=yes" and/or "port-required=yes" if the presence of a host
 *  and/or port value should be mandatory.
 *
 *  If the address family cannot be determined from the parsed address string,
 *  then CONN_ip_order() is used to determine for which IP version a connection
 *  attempt should be made first.
 *
 *  The backlog is the size of an internal queue containing the number of
 *  clients that can connect to the server without the server accepting the
 *  connections.
 *
 *  When this function returns successfully, the fullname will be filled with
 *  "<ip-address>:<port>", where the IP address is either a dotted-decimal
 *  string (IPv4) or a colon-hexadecimal string enclosed in square brackets
 *  (IPv6).
 *
 *  The returned file descriptor will be non-blocking (which means that it will
 *  not be directly usable in an accept() loop) and will have the close-on-exec
 *  flag set.
 *
 *  Activity on this socket indicates a new connection, and CONN_accept() can
 *  be called to accept this connection.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_passive_tcp(
	CSBuf_p 	fullname,	/* [value/]result address:port */
	int		backlog)	/* size of internal queue */
{
    static const char	fname[] = "CONN_passive_tcp";

    return passive_tcp(fullname, AF_UNSPEC,
			"host-required=no;port-required=no",
			backlog, fname);
} /* CONN_passive_tcp() */

/*------------------------------------------------------------------------
 *  Function	: CONN_passive_tcp4
 *  Purpose	: Open an IPv4 socket for listening
 *  Method	: passive_tcp(AF_INET)
 *
 *  History:
 *  22-06-2011	: heesakkm : creation
 *  24-06-2011	: heesakkm : Added options support.
 *
 *  Notes	:
 *  - If fullname does not contain an address, then "ip-order=4" creates an
 *    IPv4 socket. Otherwise IPv4 if enforced by AF_INET when parsing the
 *    address string.
 *  - The ip-order options is mandatory, all other options may be overruled by
 *    the caller.
 *
 *  Description	:
 *  Open a socket for accepting TCP/IPv4 connections.
 *  See CONN_passive_tcp_opts() for supported options.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_passive_tcp4(
	CSBuf_p 	fullname,	/* [value/]result address:port */
	const char	*options,	/* options string */
	int		backlog)	/* size of internal queue */
{
    static const char	fname[] = "CONN_passive_tcp4";
    static CSBuf_p	csopts = NULL;

    if (! CreateCSbuf(&csopts, fname)) return -1;

    CSPrintf(csopts, "ip-order=4;%s;host-required=no;port-required=no",
		    (options != NULL) ? options : "");

    return passive_tcp(fullname, AF_INET, CSContents(csopts), backlog, fname);
} /* CONN_passive_tcp4() */

/*------------------------------------------------------------------------
 *  Function	: CONN_passive_tcp6
 *  Purpose	: Open an IPv6 socket for listening
 *  Method	: passive_tcp(AF_INET6)
 *
 *  History:
 *  22-06-2011	: heesakkm : creation
 *  24-06-2011	: heesakkm : Added options support.
 *
 *  Notes	:
 *  - If fullname does not contain an address, then "ip-order=6" creates an
 *    IPv6 socket. Otherwise IPv6 if enforced by AF_INET6 when parsing the
 *    address string.
 *  - The ip-order options is mandatory, all other options may be overruled by
 *    the caller.
 *
 *  Description	:
 *  Open a socket for accepting TCP/IPv6 connections.
 *  See CONN_passive_tcp_opts() for supported options.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_passive_tcp6(
	CSBuf_p 	fullname,	/* [value/]result address:port */
	const char	*options,	/* options string */
	int		backlog)	/* size of internal queue */
{
    static const char	fname[] = "CONN_passive_tcp6";
    static CSBuf_p	csopts = NULL;

    if (! CreateCSbuf(&csopts, fname)) return -1;

    CSPrintf(csopts, "ip-order=6;%s;host-required=no;port-required=no;",
		    (options != NULL) ? options : "");

    return passive_tcp(fullname, AF_INET6, CSContents(csopts), backlog, fname);
} /* CONN_passive_tcp6() */

#if ! defined(NO_LOCAL)
/*------------------------------------------------------------------------
 *  Function	: CONN_passive_local
 *  Purpose	: Create passive Local-domain socket
 *  Method	:
 *  Generate filename fo rendez-vous if none specified, using socket(AF_LOCAL),
 *  setsockopt(SO_REUSEADDR), bind(), listen().
 *
 *  History:
 *  17-05-2000	: eck      : creation
 *
 *  Notes	:
 *  The generated filename contains nodename, application and pid. This way:
 *  1. Sending the name to other machines will make the connection-creation
 *     fail (instead of possibly binding to a wrong process).
 *  2. Run multiple instantiations of the same program on a single machine
 *     (like context-servers)
 *  3. Run multiple systems on a single machine.
 *  4. Perform easy checks on 'liveness' of servers (instead of using lsof)
 *
 *  If the file exists already, it will be removed if:
 *  1. It is a socket-file
 *  2. It does not accept a connection
 *
 *  Description	:
 *  Open a socket for accepting local domain connections.
 *
 *  On entry, fullname should either contain the absolute path of a file or be
 *  empty. If it is empty, a name will be generated containing the nodename,
 *  application-name, process-id and sequence-number. This way, the name should
 *  be unique in a running system.
 *
 *  CONN_passive_local() will remove the file if it exists, is a socket-file,
 *  and a connection to it is not accepted.
 *
 *  The backlog is the size of an internal queue, containing the number of
 *  clients that can connect to the server without the server accepting the
 *  connections.
 *
 *  Note that the file is only used for rendezvous between client and server.
 *  It will be created by the underlying bind() call. This may fail is the file
 *  exists already, even if there is no server running.
 *
 *  The returned file descriptor will be non-blocking and will have the
 *  close-on-exec flag set.
 *
 *  Activity on this socket indicates a new connection, and CONN_accept() can
 *  be called to accept this connection.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_passive_local(
	CSBuf_p 	fullname,	/* [value/]result filename */
	int		backlog)	/* size of internal queue */
{
    static const char	fname[] = "CONN_passive_local";
    struct sockaddr_un	server;
    int			sockfd;
    const int		optval = 1;
    struct stat		statbuf;

    if (CSLength(fullname) == 0) {
	const char	*appl = ENVGetVar("APPLNAME");
	const char	*pid = ENVGetVar("APPLIDENT");
	static int	count = 0;

	if (appl == NULL) appl ="";
	if (pid == NULL) pid ="";
	count++;

	CSPrintf(fullname, "/tmp/%s_%s_%s_%03d",
		CONN_my_name(), appl, pid, count);
    }
    /* The socket-file is now known. Unlink it if it exists and there is
     * no server for it.
     * The tests are not perfect, since the socket may exist but we can not
     * connect, which leads to the conclusion that there is no server when
     * only the listen queue is full.
     * If the socket-file exists, we just close it and let the bind() fail.
     * The setsockopt(... SO_REUSEADDR) should do stuff like this, but does not
     * (at least on Solaris 2.6).
     */
    if ((stat(CSContents(fullname), &statbuf) == 0)
	    && S_ISSOCK(statbuf.st_mode)) {
	int	cfd = CONN_connect_local(CSContents(fullname));

	if (cfd < 0) {
	    unlink(CSContents(fullname));
	} else {
	    close(cfd);
	}
    }

    sockfd = socket(AF_LOCAL, SOCK_STREAM, IPPROTO_IP);
    if (sockfd < 0) {
	ERWrite(APPLE_FILEOPEN, fname, "socket() failed: %s (errno %d)",
		strerror(errno), errno);
	return -1;
    }
    initialize_socket(sockfd, AF_LOCAL, TRUE);

    if (! conn_setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
				optval, fname, TRUE)) {
        close(sockfd);
        return -1;
    }

    server.sun_family = AF_LOCAL;
    if (CSLength(fullname) >= sizeof(server.sun_path)) {
	ERWarning(APPLE_FILEOPEN, fname,
		    "Filename \"%s\" truncated to %lu positions",
		    CSContents(fullname), (unsigned long)sizeof(server.sun_path) - 1);
	CSDeletePart(fullname, sizeof(server.sun_path) - 1,
			CSLength(fullname) - sizeof(server.sun_path) + 1);
    }
    strcpy(server.sun_path, CSContents(fullname));

    if (bind(sockfd, (struct sockaddr *)&server,
		sizeof(server.sun_family) + CSLength(fullname)) != 0) {
        close(sockfd);
        ERWrite(APPLE_FILEOPEN, fname, "bind() failed (file %s): %s (errno %d)",
		CSContents(fullname), strerror(errno), errno);
        return -1;
    }

    if (listen(sockfd, backlog) != 0) {
        close(sockfd);
        ERWrite(APPLE_FATAL, fname, "listen() failed: %s (errno %d)",
		strerror(errno), errno);
        return -1;
    }
    (void) CONN_close_on_exec(sockfd, TRUE);
    return sockfd;
} /* CONN_passive_local() */
#endif /* ! defined(NO_LOCAL) */

/* The accept() system call needs to know the size of the struct to fill. Since
 * the different sockaddr_* structs may have different sizes, sockaddr_all is
 * introduced, so we can create a variable which is never too small. After a
 * successful accept(), the actual sockaddr-type can be found in
 * remote_sock.sa_family.
 */
typedef union {
    struct sockaddr	remote_sock;
    struct sockaddr_in	remote_in;
    struct sockaddr_in6	remote_in6;
#if ! defined(NO_LOCAL)
    struct sockaddr_un	remote_un;
#endif
} sockaddr_all;


/*------------------------------------------------------------------------
 *  Function	: CONN_accept_host_port
 *  Purpose	: Accept a connection on a socket.
 *  Method	: Call of accept().
 *
 *  History:
 *  06-11-2013	: martinz  : creation, moved code from CONN_accept().
 *
 *  Notes	:
 *  Description : This function	was the old CONN_accept() but now it returns
 *  		  host, port and sin_family instead of a composed address.
 *
 *  Example     :
 *
 *  Returns	: File-descriptor or -1.
 *
 *------------------------------------------------------------------------
 */
int
CONN_accept_host_port(
	int		sockfd,		/* File descriptor */
	CSBuf_p 	clt_host,	/* OUT (optional) hostname */
	CSBuf_p 	clt_port,	/* OUT (optional) port as string */
	short *		sin_family)	/* OUT (optional) address family */
{
    int			newfd;
    sockaddr_all	remote;
    socklen_t		remotelen = sizeof(remote);
    const int		optval = 1;

    memset(&remote, '\0', sizeof(remote));

    newfd = accept(sockfd, (struct sockaddr *) &remote, &remotelen);
    if (newfd < 0) return -1;
    initialize_socket(newfd, remote.remote_sock.sa_family, FALSE);

    if (sin_family != NULL) {
	*sin_family = remote.remote_sock.sa_family;
    }
    if ((remote.remote_sock.sa_family == AF_INET)
	|| (remote.remote_sock.sa_family == AF_INET6)) {
	char		ipbuf[INET6_ADDRSTRLEN];

	(void) conn_setsockopt(newfd, SOL_SOCKET, SO_KEEPALIVE,
			       optval, __func__, TRUE);
	(void) conn_setsockopt(newfd, IPPROTO_TCP, TCP_NODELAY,
			       optval, __func__, FALSE);

	if (get_sockaddr_address(&remote.remote_sock,
				    ipbuf, sizeof(ipbuf)) == NULL) {
	    ERWrite(APPLE_PARAM, __func__, "inet_ntop() failed: %s (errno %d)",
		    strerror(errno), errno);
	    close(newfd);
	    return -1;
	}
	if (clt_host != NULL && !CSSetString(clt_host, ipbuf)) {
	    ERWrite(CSerrno, __func__, "Failed CSSetString()");
	    close(newfd);
	    return -1;
	}
	if (clt_port != NULL
	    && !CSSetString(clt_port,
			    PM_itoa(NULL,
				    get_sockaddr_port(&remote.remote_sock)))) {
	    ERWrite(CSerrno, __func__, "Failed CSSetString()");
	    close(newfd);
	    return -1;
	}
    } else {
	if (clt_host != NULL) {
	    CSClear(clt_host);
	}
	if (clt_port != NULL) {
	    CSClear(clt_port);
	}
    }
    (void) CONN_close_on_exec(newfd, TRUE);
    return newfd;
} /* CONN_accept_host_port() */


/*------------------------------------------------------------------------
 *  Function	: CONN_accept
 *  Purpose	: Accept a connection on a socket.
 *  Method	: Call CONN_accept_host_port() and compose clt_address.
 *
 *  History:
 *  17-05-2000	: eck      : creation
 *  10-03-2004	: heesakkm : Clear variable "remote" (avoid dbx rtc messages).
 *  16-03-2011	: heesakkm : Replaced inet_ntoa() with inet_ntop().
 *  24-05-2011	: martinz  : Made clt_address optional.
 *  06-11-2013	: martinz  : Moved code to and use CONN_accept_host_port().
 *  14-11-2013	: martinz  : Fixed uninitialized result.
 *
 *  Notes	:
 *  For the AF_INET/AF_INET6 address-family, the remote "ip-address:port" is
 *  determined and returned in clt_address.
 *  It is up to the calling function to do something with errno.
 *
 *  Description	:
 *  Accept connections on a file descriptor. The file descriptor can be either
 *  for a TCP/IP or a local domain socket. For TCP/IP-sockets, the SO_KEEPALIVE
 *  and TCP_NODELAY options will be set and clt_address will be filled with the
 *  remote "<address>:<port>" of the client. For local domain connections,
 *  clt_address will be empty.
 *
 *  Returns	: File-descriptor or -1.
 *------------------------------------------------------------------------
 */
int
CONN_accept(
	int		sockfd,		/* File descriptor */
	CSBuf_p 	clt_address)	/* OUT (optional) */
{
    CSBuf_p 	clt_host = NULL;	/* hostname */
    CSBuf_p 	clt_port = NULL;	/* port as string */
    short 	sin_family;		/* address family */
    int		result = 0;		/* return value */

    if (clt_address != NULL
	&& ((clt_host = CSCreate()) == NULL
  	    || (clt_port = CSCreate()) == NULL)) {
	ERWrite(CSerrno, __func__, "Failed CSCreate()");
	result = -1;
    }
    if (result != -1) {
	result = CONN_accept_host_port(sockfd,
				       clt_host,
				       clt_port,
				       &sin_family);
    }
    if (result != -1 && clt_address != NULL) {
	if (sin_family == AF_INET || sin_family == AF_INET6) {
	    if (!CSPrintf(clt_address,
	    		  sin_family == AF_INET6 ? "[%s]:%s" : "%s:%s",
			  CSContents(clt_host),
			  CSContents(clt_port))) {
		ERWrite(CSerrno, __func__, "Failed CSPrintf()");
		result = -1;
	    }
	} else {
	    if (!CSClear(clt_address)) {
		ERWrite(CSerrno, __func__, "Failed CSClear()");
		result = -1;
	    }
	}
    }
    CSDestroy(clt_host);
    CSDestroy(clt_port);
    return result;
} /* CONN_accept() */


/*------------------------------------------------------------------------
 *  Function	: CONN_close
 *  Purpose	: Close a connection
 *  Method	: close()
 *
 *  History:
 *  27-06-2000	: eck      : creation
 *
 *  Notes	:
 *  The main reason for this function is that it avoids having to include
 *  <unistd.h> for the close() only.
 *
 *  Description	: Close a connection. This is a plain wrapper around close().
 *
 *  Returns	: value of close(): 0 on successful completion, -1 otherwise
 *------------------------------------------------------------------------
 */
int
CONN_close(
	int	sockfd)			/* File descriptor */
{
    static const char	fname[] = "CONN_close";

    if (close(sockfd) != 0) {
	ERWrite(APPLE_FILECLOSE, fname,
		"close() failed: %s (errno %d)", strerror(errno), errno);
	return -1;
    }
    return 0;
} /* CONN_close() */
