/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: cs.c
 *  Description	: Counted String-routines
 *  Author(s)	: Hans van Eck
 *  Creation	: 30-09-1993
 *
 *  History:
 *  30-09-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  17-11-1993	: eck      : Added CSAppendC()
 *  21-03-1995	: eck      : Use macros in the definition of CSSize(),
 *			   : CSLength() and CSContents().
 *  06-10-1995	: eck      : Changed length test in CSInsert().
 *  05-06-1996	: eck      : Removed function definitions for CSSize(),
 *			   : CSLength() and CSContents().
 *  06-04-1998	: BdB	   : Added CSAppendN().
 *  25-08-1998	: BdB	   : Added CSEmptyString.
 *  08-04-1999	: BdB	   : Use empty string on NULL string arguments.
 *  13-06-2002  : sutherlr : Added fully-dynamic strings and CSCreate().
 *  02-04-2003  : paul     : CS_DEFAULT_CHUNK_SIZE extended from 20 to 128
 *                         : '\0'-bytes allowed in counted strings
 *                         : Added CSInsertN() and CSSetSize()
 *  03-04-2003	: paul     : Fixed bug in CSAppendC()
 *  08-04-2003	: paul     : Added CSSetLength()
 *  15-10-2003  : paul     : Moved CS_DEFAULT_CHUNK_SIZE to cs.h
 *                         : Added CSCreateOpt()
 *                         : Added CSChunkSize()
 *  20-10-2006	: heesakkm : Call CSSetSize() in CSAppend*() only when needed.
 *  06-11-2006	: paul     : Fixed bug: CSAppend() crashed when appending empty
 *                           string to NULL buffer
 *  31-05-2007	: heesakkm : CSDeletePart(): Replaced memcpy() with memmove().
 *  19-07-2012	: martinz  : Added CSCreateStringNOpt().
 *  01-10-2013	: martinz  : Added CSE_PARAM.
 *
 *  Notes	:
 *  A CSBuf_t is a structure, containing (at least) a pointer to
 *  the buffer containing the data and the size of this buffer.
 *  The room for the '\0'-byte at the end is not visible to an application.
 *  This means that writing the string "ABCD" to a Counted String with
 *  CSSize() == 4 will succeed.
 *
 *  A counted string may contain '\0'-bytes: the caller decides whether the
 *  counted string contains a string or binary information.
 *========================================================================
 */

#pragma ident "$Id: cs.c 339156 2016-11-18 18:26:14Z heesakkm $"

#include <limits.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pica.h>
#include <picamisc.h>
#include <cs.h>


/* Deallocate more than 1 chunk */
#define NR_CHUNKS	1

ERROR_DEFINE(CSE_NOERROR);
ERROR_DEFINE(CSE_NOMEM);
ERROR_DEFINE(CSE_LENGTH);
ERROR_DEFINE(CSE_OVFL);
ERROR_DEFINE(CSE_SPRINTF);
ERROR_DEFINE(CSE_PARAM);

/* This is the definition of the error variable of the CS-library */
CSErrno_t CSerrno = CSE_NOERROR;

/** An empty string to return in CSContents if the buffer is NULL. */
const char *const CSEmptyString = "";

/*------------------------------------------------------------------------
 *  Function	: CSRealloc
 *  Purpose	: (re-) allocate a piece of memory for a dynamic CS
 *  Method	: Allocate and deallocate in chunks
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *  15-10-2003	: paul     : replaced CS_DEFAULT_CHUNK_SIZE
 *                           by strp->cs_chunk_size
 *  12-10-2004  : guy      : replaced % by & operator (assumes cs_chunk_size
 *                           is a power of 2).
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
CSRealloc(
	CSBuf_p 	strp,
	size_t		size)	/* size to allocate (includes '\0') */
{
    size_t	act_size;
    void	*tmpp;
    uint32      mask;

    /* Converts the size into a multiple of the chunk size. */
    act_size = size;
    mask = strp->cs_chunk_size - 1;
    if ((act_size & mask) != 0) {
	act_size += (strp->cs_chunk_size - (act_size & mask));
    }

    /* No change? All is ok! */
    if (strp->cs_size == act_size) { return TRUE; }

    /*
     * Test if we are shrinking by too little. If this is the case,
     * ignore. There is an exception when act_size == 0, since this
     * implies deallocation.
     */
    if ((act_size != 0)
	&& (strp->cs_size > act_size)
	&& (strp->cs_size - act_size <= (NR_CHUNKS * strp->cs_chunk_size))) {
	return TRUE;
    }

    /*
     * Test if we have to free the memory. Though realloc() can also free
     * memory when the size is 0, the return value need not be NULL.
     */
    if (act_size == 0) {
	if (strp->cs_bufp != NULL) {
	    free((void *)strp->cs_bufp);
	}
	tmpp = NULL;
    } else {
	/* The realloc() can also act as malloc(), depending on cs_bufp. */
	tmpp = realloc((void *) strp->cs_bufp, act_size);
    }

    if ((tmpp == NULL) && (act_size != 0)) {
	CSerrno = CSE_NOMEM;
	return FALSE;
    }
    /* The length will be set by the calling routines! */
    strp->cs_bufp = tmpp;
    strp->cs_size = act_size;
    return TRUE;
} /* CSRealloc() */

/*------------------------------------------------------------------------
 *  Function	: CS_LOCALINIT
 *  Purpose	: Initialization for local variables
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  07-04-1995	: eck      : Initalise the buffer here.  Doing it in the
 *			   : CS_LOCAL() macro caused .rodata sections.
 *			   : This increased the size of the executable (on SUN)
 *  15-10-2003	: paul     : Added cs_chunk_size (even though it is not
 *                           relevant in CS_LOCAL's)
 *
 *  Notes	:
 *  This function is called by the CS_LOCAL() macro to initialize a counted
 *  string.
 *
 *  Returns	: An initialized structure
 *------------------------------------------------------------------------
 */
CSBuf_t
CS_LOCALINIT(
	size_t	size,		/* Size to initialize to */
	char	*buf)		/* Buffer to use */
{
    CSBuf_t	retval;

    *buf = '\0';		/* Clear the buffer */

    retval.cs_size = size;
    retval.cs_length = 1;
    retval.cs_chunk_size = 0;	/* not relevant */
    retval.cs_type = CST_LOCAL;
    retval.cs_bufp = buf;

    return retval;
} /* CS_LOCALINIT() */

/*------------------------------------------------------------------------
 *  Function	: CSCreate
 *  Purpose	: Allocate a fully-dynamic counted string
 *  Method      : CSCreateOpt("")
 *
 *  History:
 *  28-05-2002	: sutherlr : creation
 *  15-10-2003	: paul     : moved code to CSCreateOpt()
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CSBuf_p cs;
 *
 *	if ((cs = CSCreate()) == NULL) {
 *	    ERWrite(CSerrno, fname, "CSCreate() failed");
 *	}
 *
 *  Returns	: A valid CSBuf_p if successful, NULL otherwise
 *------------------------------------------------------------------------
 */
CSBuf_p
CSCreate(void)
{
    return CSCreateOpt("");
} /* CSCreate() */

/*---------------------------------------------------------------------
 *  Function    : round_binary
 *  Purpose     : Rounds number to a power of 2
 *  Method      :
 *
 *  History:
 *  12-10-2004  : Guy     : creation
 *
 *  Returns     : the rounded number
 *---------------------------------------------------------------------
 */
static int
round_binary(
	int	num_input)
{
    int result = 1;

    while (result < num_input) {
	result *= 2;
    }
    return result;
}  /* round_binary */

/*------------------------------------------------------------------------
 *  Function	: CSCreateOpt
 *  Purpose	: Allocate a fully-dynamic counted string
 *
 *  History:
 *  28-05-2002	: sutherlr : creation
 *  15-10-2003	: paul     : Added cs_chunk_size 
 *  12-10-2004  : guy      : make sure the chunk_size is a power of 2
 *
 *  Description	:
 *  This function creates a fully-dynamic counted string. The user is
 *  responsible for destroying the returned CSBuf_p with CSDestroy().
 *
 *  Notes	:
 *  Supported options:
 *
 *	"chunk-size=<value>"
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *  Returns	: A valid CSBuf_p if successful, NULL otherwise
 *------------------------------------------------------------------------
 */
CSBuf_p
CSCreateOpt(
	const char	*options)	/* options string */
{
    CSBuf_p	retval = malloc(sizeof(*retval));
    char	optval[100];

    CSerrno = CSE_NOERROR;

    if (retval == NULL) {
	CSerrno = CSE_NOMEM;
	return NULL;
    }

    retval->cs_size = 0;
    retval->cs_length = 0;
    retval->cs_type = CST_FULLY_DYNAMIC;
    retval->cs_bufp = NULL;

    retval->cs_chunk_size = CS_DEFAULT_CHUNK_SIZE;
    if (PM_opt_fetch(options, "chunk-size", optval, sizeof optval) != 0) {
	unsigned long	opt_chunk_size = strtoul(optval, NULL, 10);

	if (opt_chunk_size != 0) {
	    retval->cs_chunk_size = round_binary(opt_chunk_size);
	}
    }

    return retval;
} /* CSCreateOpt() */

/*------------------------------------------------------------------------
 *  Function	: CSSetChunkSize
 *  Purpose	: Change the chunk size of a counted string
 *
 *  History:
 *  15-10-2003	: paul     : creation
 *  12-10-2004  : guy      : make sure the chunk_size is a power of 2
 *
 *  Notes	:
 *  newsize 0 is interpreted as a request to set the chunk size to its
 *  default value
 *------------------------------------------------------------------------
 */
void
CSSetChunkSize(
	CSBuf_p strp,		/* CS to change chunk size of */
	size_t	newsize)	/* new chunk size */
{
    strp->cs_chunk_size = newsize != 0 ? round_binary(newsize) : CS_DEFAULT_CHUNK_SIZE;
} /* CSSetChunkSize() */

/*------------------------------------------------------------------------
 *  Function	: CSSetStringN
 *  Purpose	: Set the contents of a counted string with a certain length
 *  Method	:
 *  1. check parameters
 *  2. check target length (resize if possible)
 *  3. memcpy to target (or memset if source is NULL)
 *  4. add '\0' terminator and set proper length
 *
 *  History:
 *  07-03-2003	: eck      : creation
 *  02-04-2003	: paul     : setting to NULL no longer results in padding with
 *                           '\0' but instead to setting to empty string
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  string must point to a character buffer. It may contain '\0' bytes. If str
 *  is NULL, strp will set to the empty string, even if len > 0.
 *
 *  len should be the length to copy. If len<0, str is assumed to be a
 *  '\0'-terminated string.
 *
 *  CSerrno is set to CSE_LENGTH when the size of the CS is insufficient, or to
 *  CSE_NOMEM if not enough memory could be allocated (in case of a
 *  (fully-)dynamic CS).
 *
 *  This function will return TRUE when the string is set.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	if (!CSSetStringN(string, "hello \0byte", 11)) {
 *	    ERWrite(CSerrno, fname, "CSSetStringN() failed");
 *	}
 *
 *  Returns	: FALSE if size is not enough, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSSetStringN(
	CSBuf_p 	strp,		/* CS to assign to */
	const char	*string,	/* array to assign */
	int		len)		/* length of the array */
{
    CSerrno = CSE_NOERROR;

    if (string == NULL) {
	string = CSEmptyString;
	len = 0;
    } else if (len < 0) {
	len = strlen(string);
    }

    if (len > CSSize(strp)) {
	CSerrno = CSE_LENGTH;
    } else if (CS_IS_DYNAMIC(strp)) {
	if (! CSRealloc(strp, len + 1)) {
	    CSerrno = CSE_NOMEM;
	}
    }
    if (CSerrno != CSE_NOERROR) return FALSE;

    (void) memcpy((char *)strp->cs_bufp, string, len);
    ((char *) strp->cs_bufp)[len] = '\0';
    strp->cs_length = len + 1;

    return TRUE;
} /* CSSetStringN() */

/*------------------------------------------------------------------------
 *  Function	: CSSetString
 *  Purpose	: Set the contents of a counted string to a null-terminated-string
 *  Method	: CSSetStringN() with len -1
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  07-03-2003	: eck      : Code moved to CSSetStringN()
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  string must point to a '\0'-terminated string.
 *
 *  CSerrno is set to CSE_LENGTH when the size of the CS is insufficient, or to
 *  CSE_NOMEM if not enough memory could be allocated (in case of a
 *  (fully-)dynamic CS).
 *
 *  This function will return TRUE when the string is set.
 *
 *  Notes	: This function is a wrapper around CSSetStringN().
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	if (!CSSetString(string, "hello")) {
 *	    ERWrite(CSerrno, fname, "CSSetString() failed");
 *	}
 *
 *  Returns	: FALSE if size is not enough, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSSetString(
	CSBuf_p		strp,		/* CS to assign to */
	const char	*string)	/* string to assign */
{
    return CSSetStringN(strp, string, -1);
} /* CSSetString() */

/*------------------------------------------------------------------------
 *  Function	: CSSetSize
 *  Purpose	: Set the maximum length of a counted string
 *  Method	:
 *  1. Skip non-dynamic counted strings without error
 *  2. Allocate buffer with CSRealloc()
 *  3. Correct the length if necessary
 *  4. Terminate with '\0' byte if length changed
 *
 *  History:
 *  02-04-2003	: paul     : creation
 *  03-04-2003	: paul     : only write '\0' if length changes
 *  01-10-2013	: martinz  : Set CSerrno on invalid strp parameter.
 *  04-10-2013	: martinz  : Revert previous change.
 *
 *  Description	:
 *  strp must be a valid counted string. It must be either a CS_DYNAMIC or
 *  created with CSCreate().
 *
 *  This function will return TRUE if the maximum length is set. The actual
 *  size after a successful call to CSSetSize() may be slightly greater than
 *  the requested size, since the CS-library prefers sizes to be at block-size
 *  boundaries.
 *
 *  Notes	:
 *  - If the size is increased, the extra contents is not initialized.
 *  - The terminating '\0' is guaranteed to be put if length field is changed.
 *  - It is not an error to try to set the size of a non-dynamic counted
 *    string: it will just fail.
 *  - Results might be unexpected, since CSRealloc() converts the size into
 *    a multiple of the chunk size.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_DYNAMIC(	string);
 *	CSBuf_p	created_string = CSCreate();
 *
 *	/ * Since we already know that string needs to contain at least 100000
 *	  * bytes, it is wise to set the size directly. Otherwise the
 *	  * CS-library has a hell of a job to constantly reallocate a slightly
 *	  * bigger buffer.
 *	  * /
 *
 *	if (!CSSetSize(string, 20000)) {
 *	    ERWrite(CSerrno, fname, "CSSetSize() failed");
 *	}
 *	if (created_string != NULL) {
 *	    if (!CSSetSize(created_string, 100000)) {
 *	        ERWrite(CSerrno, fname, "CSSetSize() failed");
 *	    }
 *	}
 *
 *  Returns	: FALSE if size could not be changed, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSSetSize(
	CSBuf_p 	strp,		/* CS to change size of */
	size_t		newsize)	/* new size of CS */
{
    CSerrno = CSE_NOERROR;

    if (! CS_IS_DYNAMIC(strp)) return FALSE;

    if (! CSRealloc(strp, newsize + 1)) {
	CSerrno = CSE_NOMEM;
	return FALSE;
    }
    if (strp->cs_length == 0) {
	/* This counted string has just been created by CSCreate() */
	/* Now that its buffer is created too, give it the minimum length */
	strp->cs_length = 1;
	*(char *)(CSContents(strp) + CSLength(strp)) = '\0';	
    } else if (strp->cs_length > strp->cs_size) {
	strp->cs_length = strp->cs_size;
	*(char *)(CSContents(strp) + CSLength(strp)) = '\0';	
    }

    return TRUE;
} /* CSSetSize() */

/*------------------------------------------------------------------------
 *  Function	: CSSetLength
 *  Purpose	: Set the actual length of a counted string
 *  Method	:
 *  - If size is too small, try to increase it.
 *  - Change length and terminate with '\0'-byte.
 *
 *  History:
 *  08-04-2003	: paul     : creation
 *  01-10-2013	: martinz  : Prevent crash on NULL strp.
 *  04-10-2013	: martinz  : Set CSerrno on CSSetSize() failure.
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  If the requested newlength is greater than the actual size, the buffer will
 *  be extended. Otherwise the buffer will NOT be reduced in size.
 *
 *  This function will return FALSE if the buffer could not be extended (for
 *  example because the counted string was not a dynamic one). CSE_NOMEM will
 *  only be set if the counted string was dynamic, but could not be extended.
 *
 *  If the length is increased, the extra contents consists of garbage. The
 *  caller is responsible for initializing it.
 *  The terminating '\0' will always be put.
 *
 *  `CSSetLength(strp, 0)` is almost equivalent to `CSClear(strp)`, except that
 *  CSClear() may shrink the counted buffer and CSSetLength() will not.
 *
 *  `CSSetLength(strp, 1)` is almost equivalent to `CSDeletePart(strp, 1,
 *  CSLength(strp) - 1)`, except that CSDeletePart() may shrink the counted
 *  buffer and CSSetLength() will not.
 *
 *  Notes	:
 *  - Increasing the length will result in a counted string that contains
 *    garbage in its tail.
 *  - The terminating '\0' is guaranteed to be put.
 *  - CSRealloc() will only be called to increase the buffer size, never to
 *    decrease it, as is usual in CSClear() and CSDeletePart().
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_PARAM	(missing or incorrect parameter)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_DYNAMIC(	source);
 *	CS_DYNAMIC(	target);
 *
 *	while (next_source_title(source) {
 *	    / * Don't use CSClear() because this causes unnecessary
 *	      * deallocation of the counted buffer.
 *	      * /
 *	    if (! CSSetLength(target, 0)) {
 *	        ERWrite(CSerrno, fname, "CSSetLength() failed");
 *	    	break;
 *	    }
 *	    fcverr = FCVConvertCS(optionsFCV,
 *	    			CSContents(source),
 *	    			CSLength(source),
 *	    			target, 2 * CSLength(source),
 *	    			syntax, HTML_NOFUNC);
 *	    ...
 *	}
 *
 *  Returns	:
 *  FALSE if specified length was greater than size of counted string,
 *  TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSSetLength(
	CSBuf_p		strp,		/* CS to change length of */
	size_t		newlength)	/* new length of CS */
{
    CSerrno = CSE_NOERROR;

    if (strp == NULL) {
	CSerrno = CSE_PARAM;
	return FALSE;
    }
    if (newlength >= strp->cs_size) {
	if (! CSSetSize(strp, newlength)) {
	    if (CSerrno == CSE_NOERROR) {
		CSerrno = CSE_PARAM;
	    }
	    return FALSE;
	}
    }

    strp->cs_length = newlength + 1;
    *(char *)(CSContents(strp) + CSLength(strp)) = '\0';	

    return TRUE;
} /* CSSetLength() */

/*------------------------------------------------------------------------
 *  Function	: CSClear
 *  Purpose	: Clear a counted string
 *  Method	: Set length to 1 (the '\0' byte)
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  15-11-1993	: eck      : added dynamic strings
 *			   : function type changed to bool
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  21-04-2009	: sutherlr : Avoid reallocating memory for empty
 *                           dynamic cs
 *  23-04-2009	: sutherlr : Rolled back previous change
 *
 *  Description	:
 *  strp must be a valid counted string. This means that it must have
 *  been declared with CS_GLOBAL(...), CS_LOCAL(..), CSBuf_p ... or a
 *  CSCreate()-ed value.
 *
 *  This function will clear the counted string. This means that
 *  CSLength(...) will be 0. When strp is a (fully-)dynamic counted
 *  string, space will be allocated (if necessary) to hold the string.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *  Example:
 *
 *	#include <cs.h>
 *
 *	CSBuf_p string;
 *
 *	(void) CSClear(string);
 *
 *  Returns	: TRUE if the string was cleared
 *------------------------------------------------------------------------
 */
bool
CSClear(
	CSBuf_p strp)			/* CS to clear */
{
    CSerrno = CSE_NOERROR;

    if (NULL == strp){
	return TRUE;
    }

    if (CS_IS_DYNAMIC(strp)) {

	if (! CSRealloc(strp, 1)) {
	    CSerrno = CSE_NOMEM;
	}
    }
    if (CSerrno != CSE_NOERROR) return FALSE;
    strp->cs_length = 1;
    * (char *) (strp->cs_bufp) = '\0';

    return TRUE;
} /* CSClear() */

/*------------------------------------------------------------------------
 *  Function	: CSDestroy
 *  Purpose	: De-allocate a (fully-)dynamic counted string
 *  Method	: Realloc with size 0.
 *
 *  History:
 *  15-11-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  13-06-2002	: sutherlr : Added support for fully-dynamic strings
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  This function must be called when a dynamic counted string goes out of
 *  scope, or when a fully-dynamic counted string needs to be de-allocated.
 *  This means that care must be taken with presence of break and return
 *  statements.
 *
 *  When all is ok, TRUE is returned.
 *
 *  Notes	: Should not return FALSE
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *  	CS_PARAM(string);
 *
 *  	if (!CSDestroy(string)) {
 *  	    ERWrite(CSerrno, fname, "CSDestroy() failed");
 *  	}
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
CSDestroy(
	CSBuf_p	strp)			/* CS to destroy */
{
    CSerrno = CSE_NOERROR;

    if ((strp != NULL) && CS_IS_DYNAMIC(strp)) {
	if (! CSRealloc(strp, 0)) {
	    /* This should not be happening ! */
	    CSerrno = CSE_NOMEM;
	} else {
	    strp->cs_length = 0;
	}
	if (strp->cs_type == CST_FULLY_DYNAMIC) {
	    free(strp);
	}
    }
    return (CSerrno == CSE_NOERROR) ;
} /* CSDestroy() */

/*------------------------------------------------------------------------
 *  Function	: CSAppend
 *  Purpose	: Append a '\0'-terminated string to a counted string
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  02-04-2003	: paul     : Use CSSetSize()
 *  23-10-2006	: heesakkm : Call CSSetSize() only when needed.
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  string must point to a '\0'-terminated string.
 *
 *  CSerrno is set to CSE_LENGTH when the size of the CS is insufficient, or to
 *  CSE_NOMEM if not enough memory could be allocated (in case of a
 *  (fully-)dynamic CS).
 *
 *  This function will return TRUE when the string is changed.
 *
 *  Notes	: This function cannot be used for binary information.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string)
 *
 *	(void) CSSetString(string, "hello");
 *	if (!CSAppend(string, " there")) {
 *	    ERWrite(CSerrno, fname, "CSAppend() failed");
 *	}
 *
 *  Returns	: FALSE if the result would be too long, TRUE on success
 *------------------------------------------------------------------------
 */
bool
CSAppend(
	CSBuf_p		strp,		/* CS to append to */
	const char	*string)	/* string to append */
{
    size_t oldlen = CSLength(strp);
    size_t maxlen = CSSize(strp);
    size_t addlen;

    CSerrno = CSE_NOERROR;
    if (string == NULL) string = CSEmptyString;
    addlen = strlen(string);
    if (addlen == 0) {
	return TRUE;
    }

    if (oldlen + addlen > maxlen) {
	CSerrno = CSE_LENGTH;
    } else {
	if (strp->cs_length + addlen > strp->cs_size) {
	    (void) CSSetSize(strp, oldlen + addlen);
	}
    }

    if (CSerrno != CSE_NOERROR) return FALSE;

    /* Do the real work. */
    (void) strcpy((char *)(CSContents(strp) + oldlen), string);
    strp->cs_length += addlen;

    return TRUE;
} /* CSAppend() */

/*------------------------------------------------------------------------
 *  Function	: CSAppendN
 *  Purpose	: Append a string with given length to a counted string
 *
 *  History:
 *  06-04-1998	: BdB      : creation
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  02-04-2003	: paul     : if appending a NULL pointer then length = 0
 *                         : Use CSSetSize()
 *  20-10-2006	: heesakkm : Call CSSetSize() only when needed.
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  string must point to an array of ‘length’ characters, which does not have
 *  to be null-terminated.
 *
 *  ‘length’ is the number of bytes to append, if it equals (size_t)-1, the
 *  length will be determined with strlen(), in which case the strp should be
 *  null-terminated.
 *
 *  CSerrno is set to CSE_LENGTH when the size of the CS is insufficient, or to
 *  CSE_NOMEM if not enough memory could be allocated (in case of a
 *  (fully-)dynamic CS).
 *
 *  This function will return TRUE if no errors occur.
 *
 *  Notes	:
 *  This function can be used for binary information if non-minus-one length is
 *  supplied.
 *
 *  Appending a NULL pointer results in empty string being appended and the
 *  value of the length parameter being ignored.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string)
 *
 *	(void) CSSetString(string, "hello");
 *	if (!CSAppendN(string, " there", 6)) {
 *	    ERWrite(CSerrno, fname, "CSAppendN() failed");
 *	}
 *
 *  Returns	: FALSE if the result would be too long, TRUE on success
 *------------------------------------------------------------------------
 */
bool
CSAppendN(
	CSBuf_p		strp,		/* CS to append to */
	const char	*string,	/* array to append */
	size_t		length)		/* length of array to append */
{
    size_t oldlen = CSLength(strp);
    size_t maxlen = CSSize(strp);

    CSerrno = CSE_NOERROR;
    if (string == NULL) {
	string = CSEmptyString;
	length = 0;
    } else if (length == (size_t)-1) {
	length = strlen(string);
    }
    if (length == 0) {
	return TRUE;
    }

    if (oldlen + length > maxlen) {
	CSerrno = CSE_LENGTH;
    } else {
	if (strp->cs_length + length > strp->cs_size) {
	    (void) CSSetSize(strp, oldlen + length);
	}
    }

    if (CSerrno != CSE_NOERROR) return FALSE;

    /* Do the real work. */
    (void) memcpy((char *)(CSContents(strp) + oldlen), string, length);
    strp->cs_length += length;
    *(char *)(CSContents(strp) + CSLength(strp)) = '\0';

    return TRUE;
} /* CSAppendN() */

/*------------------------------------------------------------------------
 *  Function	: CSAppendC
 *  Purpose	: Append a character to a counted string
 *  Method	:
 *
 *  History:
 *  17-11-1993	: eck      : creation
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  02-04-2003	: paul     : Use CSSetSize()
 *  03-04-2003	: paul     : Reinserted a lost line
 *  20-10-2006	: heesakkm : Call CSSetSize() only when needed.
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  ch may be '\0', because strp may contain binary data.
 *
 *  CSerrno is set to CSE_LENGTH when the size of the CS is insufficient,
 *  or to CSE_NOMEM if not enough memory could be allocated (in case of
 *  a (fully-)dynamic CS).
 *
 *  This function will return TRUE when the string is changed.
 *
 *  Notes	:
 *  Copied from CSAppend() and adapted a bit.
 *
 *  Appending a '\0' is useful if the counted string is to contain binary
 *  information.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	if (!CSAppendC(string, 'p')) {
 *	    ERWrite(CSerrno, fname, "CSAppendC() failed");
 *	}
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
CSAppendC(
	CSBuf_p 	strp,		/* CS to append to */
	char		ch)		/* character to append */
{
    size_t oldlen = CSLength(strp);
    size_t maxlen = CSSize(strp);

    CSerrno = CSE_NOERROR;
    if (oldlen + 1 > maxlen) {
	CSerrno = CSE_LENGTH;
    } else {
	if (strp->cs_length + 1 > strp->cs_size) {
	    (void) CSSetSize(strp, oldlen + 1);
	}
    }

    if (CSerrno != CSE_NOERROR) return FALSE;

    /* Do the real work. */
    *(char *)(CSContents(strp) + oldlen) = ch;
    *(char *)(CSContents(strp) + oldlen + 1) = '\0';

    strp->cs_length += 1;
    return TRUE;
} /* CSAppendC() */

/*------------------------------------------------------------------------
 *  Function	: CSDeletePart
 *  Purpose	: Delete a part of a counted string
 *  Method	: memmove() over the data to delete and adjust length
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  02-04-2003	: paul     : use memcpy() instead of strcpy()
 *                         : Fixed bug: prohibit possibility of deleting one
 *                           byte too many
 *  31-05-2007	: heesakkm : Replaced memcpy() with memmove().
 *  07-06-2013	: paul     : CBS-4494 Deleting zero bytes is not an error
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  start must be smaller than the length field (that is, it must lie inside
 *  the counted string).
 *
 *  start + length must be smaller than the length field (so the end lies
 *  inside the counted string).
 *
 *  This function will return TRUE when the deletion succeeded. When the start
 *  or the length is illegal, FALSE is returned and CSerrno is set to
 *  CSE_LENGTH. In this case, the counted string is left unchanged.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(specified length too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	if (!CSDeletePart(string, 3, 5)) {
 *	    ERWrite(CSerrno, fname, "CSDeletePart() failed");
 *	}
 *
 *  Returns	: FALSE if start or length wrong, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSDeletePart(
	CSBuf_p strp,		/* CS to delete in */
	size_t	start,		/* first pos to delete */
	size_t	length)		/* the # of chars to delete */
{
    size_t curlen = CSLength(strp);

    CSerrno = CSE_NOERROR;
    /* Check the size. start and length > 0 */
    if (start + length > curlen) {
	CSerrno = CSE_LENGTH;
	return FALSE;
    }

    /* Deleting zero bytes is not an error */
    if (length == 0) {
	return TRUE;
    }

    /* Do the real work. */
    (void) memmove((char *) strp->cs_bufp + start,
		    strp->cs_bufp + start + length,
		    curlen + 1 - start - length);	/* copies '\0' too */
    strp->cs_length -= length;

    if (CS_IS_DYNAMIC(strp)) {
	if (! CSRealloc(strp, strp->cs_length)) {
	    /* We're deleting something: this can't be happening! */
	    CSerrno = CSE_NOMEM;
	}
    }

    return (CSerrno == CSE_NOERROR);
} /* CSDeletePart() */

/*------------------------------------------------------------------------
 *  Function	: CSInsertN
 *  Purpose	: Insert an array of characters into a counted string
 *  Method	: memmove() and memcpy()
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  02-04-2003	: paul     : Generalized from CSInsert() to handle binary info
 *                         : Use CSSetSize()
 *  03-02-2016	: paul     : CBS-6398 Don't shrink buffer: PPToInternAppendCS()
 *                           would be less efficient when expected size has
 *                           been set by CSSetSize()
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  start must be smaller than the length field (that is, it must lie inside
 *  the counted string).
 *
 *  str must point to an array of ‘len’ characters, which does not have to be
 *  null-terminated.
 *
 *  len is the number of bytes to append, if it equals -1, the length will be
 *  determined with strlen(), in which case the str should be a
 *  null-terminated-string.
 *
 *  The new length of the counted string must be smaller than the maximum
 *  size. This means that CSLength(strp) + strlen(str) must be no larger than
 *  CSSize(strp).
 *
 *  CSerrno is set to CSE_LENGTH when the start is illegal or the size of the
 *  CS is insufficient. When not enough memory could be allocated it is set to
 *  CSE_NOMEM (in case of a (fully-)dynamic CS).
 *
 *  This function will return TRUE if the insertion succeeded.
 *
 *  Notes	:
 *  This function can be used for binary information if non-negative length is
 *  supplied.
 *
 *  Inserting a NULL pointer results in empty string being inserted and the
 *  value of the len parameter being ignored
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	// insert only the first 2 characters of "there"
 *	if (!CSInsertN(string, 5, "there", 2)) {
 *	    ERWrite(CSerrno, fname, "CSInsertN() failed");
 *	}
 *
 *  Returns	:
 *  FALSE if length is insufficient or start is too big, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSInsertN(
    CSBuf_p 	strp,		/* CS to insert in */
    size_t	start,		/* position to insert to */
    const char	*str,		/* array to insert */
    int		len)		/* length of the array */
{
    size_t maxlen = CSSize(strp);
    size_t curlen = CSLength(strp);
    size_t inslen;

    CSerrno = CSE_NOERROR;
    if (str == NULL) {
	str = CSEmptyString;
	inslen = 0;
    } else if (len < 0) {
	inslen = strlen(str);
    } else {
	inslen = len;
    }

    /* Check size + length */
    if ((curlen + inslen > maxlen)
        || (start > curlen)) {
	CSerrno = CSE_LENGTH;
    } else if (strp->cs_length + inslen > strp->cs_size) {
	(void) CSSetSize(strp, curlen + inslen);
    }
    if (CSerrno != CSE_NOERROR) return FALSE;

    /* Do the real work. */
    (void) memmove((void *) (strp->cs_bufp + start + inslen),
		    strp->cs_bufp + start,
		    curlen - start + 1);
    (void) memcpy((void *) (strp->cs_bufp + start), str, inslen);

    strp->cs_length += inslen;

    return TRUE;
} /* CSInsertN() */

/*------------------------------------------------------------------------
 *  Function	: CSInsert
 *  Purpose	: Insert a null-terminated-string into a counted string
 *  Method	: CSInsertN() with len -1
 *
 *  History:
 *  14-10-1993	: eck      : creation
 *  15-11-1993	: eck      : Added dynamic strings
 *  13-06-2002	: sutherlr : Added CS_IS_DYNAMIC() to check for either
 *                           dynamic or fully-dynamic strings
 *  02-04-2003	: paul     : Code moved to CSInsertN()
 *  22-02-2008	: sieburgw : Added CSPurge
 *  05-03-2008	: sieburgw : Renamed CSPurge to CSTearOff
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  start must be smaller than the length field (that is, it must lie inside
 *  the counted string).
 *
 *  The new length of the counted string must be smaller than the maximum
 *  size. This means that CSLength(strp) + strlen(str) must be no larger than
 *  CSSize(strp).
 *
 *  CSerrno is set to CSE_LENGTH when the start is illegal or the size of the
 *  CS is insufficient. When not enough memory could be allocated it is set to
 *  CSE_NOMEM (in case of a (fully-)dynamic CS).
 *
 *  This function will return TRUE if the insertion succeeded.
 *
 *  Notes	:
 *  This function is a wrapper around CSInsertN(). It cannot be used for binary
 *  information.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_NOMEM	(not enough memory for operation)
 *	CSE_LENGTH	(resulting length of operation too large)
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	if (!CSInsert(string, 5, "there")) {
 *	    ERWrite(CSerrno, fname, "CSInsert() failed");
 *	}
 *  Returns	:
 *  FALSE if length is insufficient or start is too big, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
CSInsert(
	CSBuf_p		strp,		/* CS to insert in */
	size_t		start,		/* position to insert to */
	const char	*str)		/* string to insert */
{
    return CSInsertN(strp, start, str, -1);
} /* CSInsert() */

/*------------------------------------------------------------------------
 *  Function	: CSTearOff
 *  Purpose	: 'steal' the buffer from a dynamic CS.
 *  Method	:
 *
 *  History:
 *  22-02-2008	: sieburgw : creation
 *
 *  Notes	:
 *  The caller is responsible for deallocating the returned buffer.
 *  The 'cs' is left in a state as if it has just been created.
 *
 *  Returns	: Pointer to string
 *------------------------------------------------------------------------
 */
char *
CSTearOff(
	CSBuf_p	cs)
{
    char *res; 
    if ((cs == NULL) || !CS_IS_DYNAMIC(cs)){
        return NULL;
    }
    res = (char*)cs->cs_bufp;
    if (res == NULL){
        /*
         * Return an empty string, in case we have no cs_bufp.
         * This prevents the need to check for NULL
         * in the caller
         */
        res = malloc(1);
        *res = '\0';
    }
    cs->cs_bufp = NULL;
    cs->cs_size = 0;
    cs->cs_length = 0;
    return res;
} /* CSTearOff() */

/*------------------------------------------------------------------------
 *  Function	: CSCreateStringNOpt
 *  Purpose	: Create and populate counted string
 *  Method	: CSCreate() and CSSetStringN()
 *
 *  History:
 *  19-07-2012	: martinz  : creation
 *
 *  Notes	: This is a PM_strdup() equivalent for counted strings.
 *
 *  Returns	: Created counted string on success, NULL otherwise.
 *------------------------------------------------------------------------
 */
CSBuf_p
CSCreateStringNOpt(
    const char	*string,	/* string to set */
    int		len,		/* string length */
    const char	*options)	/* create options */
{
    CSBuf_p buf;		/* return value */

    if ((buf = CSCreateOpt(options)) == NULL) {
	return NULL;
    }
    if (!CSSetStringN(buf, string, len)) {
	CSDestroy(buf);
	return NULL;
    }
    return buf;
} /* CSCreateStringNOpt() */
