/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: csprintf.c
 *  Description	: The print-routines for counted strings
 *  Author(s)	: Hans van Eck
 *  Creation	: 15-10-1993
 *
 *  History:
 *  15-10-1993	: eck      : creation
 *  19-05-1994	: eck      : added VSPrintf()
 *  30-09-2004	: eck      : Added an implementation to use vsnprintf()
 *  22-03-2007	: jaap/eck : Use vsnprintf() on Linux
 *  17-04-2007	: jaap     : Use va_copy() in CSVPrintf_vsnprintf()
 *
 *  Notes	:
 *  This routine has it's own file, so that when it is not used, the
 *  stdio-library will not be linked when using counted string.
 *  The function vsnprintf() is in the newer ANSI_C standard, but on Solaris,
 *  we still compile using the old one. We still want to use vsnprintf(), so
 *  there some "#define" fiddling going on.
 *========================================================================
 */

#pragma ident "$Id: csprintf.c 307506 2014-01-22 10:52:31Z sieburgw $"

#if defined(__STDC_VERSION__)
#if (__STDC_VERSION__ >= 199901L)
#define USE_VSNPRINTF
#endif
#endif

#if !defined(USE_VSNPRINTF)

#if defined(__sun) && ! defined(__EXTENSIONS__)
/* On SUN/Solaris the #define is a dirty hack necessary for the declaration of
 * vsnprintf() and va_copy().
 */
#define __EXTENSIONS__
#define USE_VSNPRINTF
#endif

#if defined(__linux__)
/* On Linux we simply assume that vsnprintf() and va_copy() are available. */
#define USE_VSNPRINTF
#endif

#endif /* !defined(USE_VSNPRINTF) */

#include <stdarg.h>
#include <stdio.h>
#include <pica.h>
#include <cs.h>

#if defined(USE_VSNPRINTF)
/*------------------------------------------------------------------------
 *  Function	: CSVPrintfAppend_vsnprintf
 *  Purpose	: Print to the end of a CS.
 *  Method	: Calls vsnprintf()
 *
 *  History:
 *  30-09-2004	: eck      : creation
 *
 *  Notes	:
 *  If the counted string is not large enough, it will be filled to
 *  CSSize(strp) characters and FALSE is returned. CSerrno will be set to
 *  CSE_LENGTH.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_LENGTH	(resulting length of operation too large)
 *	CSE_SPRINTF	(sprintf returned a negative value)
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *  Returns	:
 *  - FALSE if an error has occurred; CSerrno should be checked:
 *    - CSE_SPRINTF means error from sprintf()
 *    - CSE_NOMEM means a resize failed
 *  - TRUE if the string was successfully assigned;
 *    If CSerrno is set to CSE_LENGTH, the string is filled to its maximum.
 *------------------------------------------------------------------------
 */
static bool
CSVPrintfAppend_vsnprintf(
	CSBuf_p 	strp,		/* the CS to print to */
	const char	*fmt,		/* format-string */
	va_list		argp)		/* argument-list pointer */
{
    int		len;
    va_list	argp_copy;

    CSerrno = CSE_NOERROR;

    /* Ensure that there is a cs_bufp, otherwise the vsnprintf() will fail.
     * CSClear() will allocate a buffer, but an implementation change might
     * change this, so the extra test and CSSetLength() is defensive
     * programming.
     */
    if (strp->cs_bufp == NULL) {
	CSClear(strp);
	if (strp->cs_bufp == NULL) {
	    CSSetLength(strp, 0);
	}
    }

    /* Remember that strp->cs_length contains the length of the string
     * including the '\0' terminator.
     * At this point strp->cs_length may not be equal to 0.
     */
    if (strp->cs_length == 0) {
	CSerrno = CSE_NOMEM;
	return FALSE;
    }

    /* vsnprintf() returns the number of characters it would write if
     * strp->cs_size were large enough, not counting the terminating null
     * character.
     */
    va_copy(argp_copy, argp);
    len = vsnprintf((char *)strp->cs_bufp + (strp->cs_length - 1),
		    strp->cs_size - (strp->cs_length - 1), fmt, argp_copy);
    va_end(argp_copy);

    if (len < 0) {
	(void) CSClear(strp);

	CSerrno = CSE_SPRINTF;
	return FALSE;
    } else if (len + (strp->cs_length - 1) >= strp->cs_size) {
	if (CS_IS_DYNAMIC(strp)) {
	    /* Re-size the counted string, so we can print into the buffer
	     * immediately.
	     */
	    if (! CSSetSize(strp, len + (strp->cs_length - 1))) {
		CSErrno_t	err = CSerrno;
		CSClear(strp);
		CSerrno = err;
		return FALSE;
	    }
	    len = vsnprintf((char *)strp->cs_bufp + (strp->cs_length - 1),
			    strp->cs_size - (strp->cs_length - 1), fmt, argp);
	}
	if (len + (strp->cs_length - 1) >= strp->cs_size) {
	    len = strp->cs_size - (strp->cs_length - 1) - 1;
	    CSerrno = CSE_LENGTH;
	}
    }
    strp->cs_length += len;
    return TRUE;
} /* CSVPrintfAppend_vsnprintf() */

#else

/* prt_buffer is used to print into. There is some extra space. */
static char prt_buffer[CSMAX_SPRINTF_LEN + 32];

/*------------------------------------------------------------------------
 *  Function	: CSVPrintfAppend_orig
 *  Purpose	: Print to the end a CS
 *  Method	: Calls vsprintf()
 *
 *  History:
 *  30-09-2004	: eck      : creation, used to be CSVPrintf()
 *
 *  Notes	:
 *  - Does only a limited check on buffer overflow.
 *  - Does not use ERWrite() to avoid dependency loops (which can cause
 *    CSerrno to change).
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_LENGTH	(resulting length of operation too large)
 *	CSE_SPRINTF	(sprintf returned a negative value)
 *	CSE_OVFL	(length of argument list > CSMAX_SPRINTF_LEN)
 *
 *  Returns	:
 *  - FALSE if an error has occurred. CSerrno should be checked.
 *    - CSE_OVFL should be fatal
 *    - CSE_SPRINTF means error from sprintf()
 *  - TRUE if the string was successfully assigned;
 *    If CSerrno is set to CSE_LENGTH, the string is filled to its maximum.
 *    If it is set to CSE_SPRINTF, a '\0'- byte was found.
 *------------------------------------------------------------------------
 */
bool
CSVPrintfAppend_orig(
	CSBuf_p 	strp,		/* the CS to print to */
	const char	*fmt,		/* format-string */
	va_list		argp)		/* argument-list pointer */
{
    int		len;
    size_t	maxlen;

    CSerrno = CSE_NOERROR;

    (void) CSClear(strp);
    maxlen = CSSize(strp);

    len = vsprintf(prt_buffer, fmt, argp);

    if (len > CSMAX_SPRINTF_LEN) {
	/* This is serious, folks! */
	CSerrno = CSE_OVFL;
	return FALSE;
    } else if (len < 0) {
	CSerrno = CSE_SPRINTF;
	return FALSE;
    } else {
	size_t	oldlen = CSLength(strp);

	if (len > maxlen) {
	    /* Since CSAppend clears CSerrno, don't set it just yet. */
	    prt_buffer[maxlen] = '\0';
	}
	if (CSAppend(strp, prt_buffer)) {
	    /* Try to catch '\0'-bytes. */
	    if (oldlen + len != CSLength(strp)) {
		CSerrno = CSE_SPRINTF;
	    } else if (len > maxlen) {
		CSerrno = CSE_LENGTH;
	    }
	}
	return TRUE;
    }
} /* CSVPrintfAppend_orig() */

#endif

/*------------------------------------------------------------------------
 *  Function	: CSVPrintfAppend
 *  Purpose	: Print to the end of a CS
 *  Method	:
 *  Wrapper around CSVPrintfAppend_orig() or CSVPrintfAppend_vsnprintf()
 *
 *  History:
 *  20-01-2005	: eck      : creation
 *
 *  Description	:
 *  strp must be a valid counted string.
 *
 *  fmt must be a valid format string like in the printf-family. The format and
 *  it's arguments may not result in a '\0'-character being printed.
 *
 *  va_list must be initialized with a va_start() (and possible va_end()) call
 *  and is supposed to be a regular va_list containing the arguments for the
 *  printf.
 *
 *  Error values:
 *
 *	CSE_NOERROR	(ok)
 *	CSE_LENGTH	(resulting length of operation too large)
 *	CSE_SPRINTF	(sprintf returned a negative value)
 *
 *	CSVPrintfAppend_vsnprintf():
 *	CSE_NOMEM	(not enough memory for operation)
 *
 *	CSVPrintfAppend_orig():
 *	CSE_OVFL	(length of argument list > CSMAX_SPRINTF_LEN)
 *
 *  Returns	:
 *  - FALSE if an error has occurred; CSerrno should be checked:
 *    - CS_OVFL should be fatal
 *    - CS_SPRINTF means error from sprintf()
 *    - CSE_NOMEM means a resize failed
 *  - TRUE if the string was successfully assigned;
 *    If CSerrno is set to CSE_LENGTH, the string is filled to its maximum.
 *    If it is set to CSE_SPRINTF, a '\0'- byte was found.
 *------------------------------------------------------------------------
 */
bool
CSVPrintfAppend(
	CSBuf_p 	strp,		/* the CS to print to */
	const char	*fmt,		/* format-string */
	va_list		argp)		/* argument-list pointer */
{
#if defined(USE_VSNPRINTF)
    return CSVPrintfAppend_vsnprintf(strp, fmt, argp);
#else
    return CSVPrintfAppend_orig(strp, fmt, argp);
#endif
} /* CSVPrintfAppend() */

/*------------------------------------------------------------------------
 *  Function	: CSVPrintf
 *  Purpose	: Print to a CS
 *  Method	: Call CSVPrintfAppend()
 *
 *  History:
 *  15-02-2012	: paul     : creation
 *
 *  Description	: This routine is a wrapper around routine CSVPrintfAppend().
 *
 *  Example	:
 *
 *	#include <stdarg.c>
 *	#include <cs.h>
 *	...
 *	bool
 *	CSPrintf(
 *	 	CS_PARAM(string),
 *	 	const char *fmt,
 *	 	...)
 *	bool	retval;
 *	va_list	pvar;
 *	va_start(pvar, fmt)
 *	retval = CSVPrintf(string, fmt, pvar);
 *	va_end(pvar);
 *
 *	return retval;
 *
 *  Returns	: See CSVPrintfAppend()
 *------------------------------------------------------------------------
 */
bool
CSVPrintf(
	CSBuf_p 	strp,		/* the CS to print to */
	const char	*fmt,		/* format-string */
	va_list		argp)		/* argument-list pointer */
{
    if (!CSSetLength(strp, 0)) {
	return FALSE;
    }
    return CSVPrintfAppend(strp, fmt, argp);
} /* CSVPrintf() */

/*------------------------------------------------------------------------
 *  Function	: CSPrintf
 *  Purpose	: Print formatted to a counted string
 *  Method	: Call CSVPrintf()
 *  Author(s)	: Hans van Eck
 *  Creation	: 15-10-1993
 *
 *  History:
 *  15-10-1993	: eck      : creation
 *  19-05-1994	: eck      : Code moved to CSVPrintfAppend()
 *
 *  Description	: This routine is a wrapper around routine CSVPrintf().
 *
 *  Returns	: See CSVPrintf()
 *------------------------------------------------------------------------
 */
bool
CSPrintf(
	CSBuf_p 	strp,		/* the CS to print to */
	const char	*fmt,		/* format string */
			...)		/* optional format arguments */
{
    va_list	pvar;
    bool	retval;

    va_start(pvar, fmt);
    retval = CSVPrintf(strp, fmt, pvar);
    va_end(pvar);

    return retval;
} /* CSPrintf() */

/*------------------------------------------------------------------------
 *  Function	: CSPrintfAppend
 *  Purpose	: Print to the end a CS
 *  Method	: Call CSVPrintfAppend()
 *
 *  History:
 *  15-02-2012	: paul     : creation
 *
 *  Description	: This routine is a wrapper around routine CSVPrintfAppend().
 *
 *  Returns	: See CSVPrintfAppend()
 *------------------------------------------------------------------------
 */
bool
CSPrintfAppend(
	CSBuf_p 	strp,		/* the CS to print to */
	const char	*fmt,		/* format string */
			...)		/* optional format arguments */
{
    va_list	pvar;
    bool	retval;

    va_start(pvar, fmt);
    retval = CSVPrintfAppend(strp, fmt, pvar);
    va_end(pvar);

    return retval;
} /* CSPrintfAppend() */
