/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: envinit.c
 *  Description	: Initialize environment
 *  Author(s)	: Hans van Eck
 *  Creation	: 23-11-1993
 *
 *  History:
 *  23-11-1993	: eck      : creation
 *  14-12-1993  : clemens  : extra parameter ENVInit(...,appl_version)
 *                           ENVReadFile(..) no longer static
 *  04-07-1994	: eck      : Turn off group write-protection for files
 *			     The errorlog- and account- files should be shared
 *  05-04-1995	: eck      : Added variable-expansion, empty- and comment-lines
 *  15-06-1995	: eck      : Changes for dependency-generation
 *  21-11-1996  : PvdM     : Adjusted for Windows compatability.
 *  22-06-1999	: eck      : Added -L and -S options for Hermes (for loopback
 *			   : and script-name)
 *  24-08-1999	: eck      : Added socket initialisation for Tandem
 *  02-09-1999	: eck      : Socket initialisation under Guardian (duh)
 *  11-09-2003	: paul     : Added ENVutf8 (Unicode Master Switch)
 *  24-02-2005	: sutherlr : Made ENVExpandVariables() public.
 *  23-03-2005	: paul     : Added ENV_max_copies
 *  23-01-2008  : fb       : Moved define for PATH_MAX to pica.h (Windows)
 *  20-04-2011	: heesakkm : Removed __TANDEM and _TNS_R_TARGET defines.
 *  27-02-2012	: heesakkm : Per request of HvE: removed support for -L
 *			     (HM_LOOPBACK) and reused -S (SCRIPT_NAME) for
 *			     storing option string values in the environment.
 *  27-02-2012	: huug     : changed variable actabl_str with value ACTABL
 *			     to actabl_cat_str with value ACTABL_CAT
 *  17-07-2012	: heesakkm : Support hyphenated environment variable names.
 *  20-05-2016	: heesakkm : Removed VMS support.
 *
 *========================================================================
 */

#pragma ident "$Id: envinit.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <pica.h>
#include <picamisc.h>
#include <picatypes.h>
#include <cs.h>
#include <er.h>
#include <env.h>

#include "envversion"

static const char *applname_str		= "APPLNAME";
static const char *applident_str	= "APPLIDENT";
static const char *filemap_str		= "FILEMAP";
static const char *actabl_catStr	= "ACTABL_CAT";
static const char *tb_inputfile_str	= "TB_INPUTFILE"; /* matches value in TB lib */
static const char *extended_opts_str	= "EXTOPTS";
static const char *errorlog_str		= "ERRORLOG";
static const char *utf8_str		= "UTF8";
static const char *max_copies_str	= "MAX_COPIES";

/* The Unicode Master Switch is set if and only if UTF8=ON */
bool	ENVutf8 = FALSE;

/* The default highest copy number is 99 */
PICCopyNum_t	ENV_max_copies = 99;


/*------------------------------------------------------------------------
 *  Function	: ENVExpandVariables
 *  Purpose	: Do variable substitution
 *  Method	:
 *  Look for variables in input and change them if found in environment.
 *  All is done in only one buffer by keeping track of the '$'-sign.
 *
 *  History:
 *  05-04-1995	: eck      : creation
 *  21-04-1995	: eck      : added support for variables between parenthesis
 *  15-10-1999	: eck      : Added support for variables in curly brackets
 *  24-02-2005	: sutherlr : Made this function public.
 *  25-10-2011	: heesakkm : Handle NULL input.
 *  17-07-2013	: heesakkm : Support hyphenated variable names.
 *
 *  Notes:
 *  - value_in is the character string to be expanded.
 *
 *  The function will parse value_in, expand any shell variables found, and
 *  return the expanded result.
 *
 *  The function can handle shell variables in the following formats:
 *  1. $VARIABLE
 *  2. $(VARIABLE)
 *  3. ${VARIABLE}
 *
 *  There is a limit on the output.
 *
 *  Unrecognised variables are NOT subsituted.
 *
 *  The function does not expand $TODAY.
 *
 *  Example:
 *
 *      #include <env.h>
 *      .
 *      const char *expanded;
 *      expanded = ENVExpandVariables("$RUNDIR/output_file");
 *      .
 *
 *  Returns	: A buffer with the expanded string
 *------------------------------------------------------------------------
 */
const char *
ENVExpandVariables(
	const char	*value_in)	/* Buffer to expand */
{
    static char	out_buf[BUFSIZ];
    int		out_len;
    char	paren;

    out_len = 0;
    out_buf[out_len] = '\0';

    if (value_in == NULL) return out_buf;

    while (*value_in != '\0') {
	while ((*value_in != '\0')
		&& (*value_in != '$')) {
	    if (out_len < sizeof(out_buf)-1) {
		out_buf[out_len++] = *(value_in++);
	    }
	}
	if (*value_in == '$') {
	    int	sav_len = out_len;

	    /* Copy the '$' */
	    if (out_len < sizeof(out_buf)-1) {
		out_buf[out_len++] = *(value_in++);
	    }
	    paren = '\0';
	    if (*value_in == '(') paren = ')';
	    if (*value_in == '{') paren = '}';
	    if (paren != '\0') {
		if (out_len < sizeof(out_buf)-1) {
		    out_buf[out_len++] = *(value_in++);
		}
	    }

	    /**
	     * Copy the variable. The test for (out_len - sav_len > 2)
	     * will make sure that $<digit> is not a variable.
	     */
	    while ((*value_in != '\0')
		    && (isalpha(*value_in)
			|| (*value_in == '_')
			|| (*value_in == '-')
			|| ((out_len - sav_len > 2) && isdigit(*value_in)))) {
		if (out_len < sizeof(out_buf)-1) {
		    out_buf[out_len++] = *(value_in++);
		}
	    }
	    if (paren != '\0') {
		if (*value_in != paren) continue;
		value_in++;
	    }
	    out_buf[out_len] = '\0';

	    /* Test if there was a variable (remember the '$'). */
	    if (out_len - sav_len > 1) {
		const char	*subst;

		if (paren == '\0') subst = ENVGetVar(out_buf + sav_len + 1);
		else subst = ENVGetVar(out_buf + sav_len + 2);

		/* If we found one, substitute it. */
		if (subst != NULL) {
		    out_len = sav_len;
		    while (*subst != '\0') {
			if (out_len < sizeof(out_buf)-1) {
			    out_buf[out_len++] = *(subst++);
			}
		    }
		    continue;
		}
	    }
	    if (paren != '\0') {
		if (out_len < sizeof(out_buf)-1) {
		    out_buf[out_len++] = paren;
		}
	    }
	} /* if (value_in == '$') */
    } /* while (*value_in != '\0') */

    /* Terminate the buffer */
    out_buf[out_len] = '\0';

    return out_buf;
} /* ENVExpandVariables() */

/*-----------------------------------------------------------------------
 *  Function	: ENVReadFile
 *  Purpose	: Read a filemap-file
 *  Method	: Read the file line by line
 *  Author(s)	: Hans van Eck
 *  Creation	: 24-11-1993
 *
 *  History:
 *  24-11-1993	: eck      : creation
 *  14-12-1993  : clemens  : function externally available
 *  05-04-1995	: eck      : Added variable expansion and some input checking
 *  17-07-2012	: heesakkm : Support hyphenated variable names
 *
 *  Notes:
 *  filnam is the name of the file containing environment variables. The
 *  file must be a standard textfile (stream-IO).
 *
 *  The function will open filnam and read until EOF. Every line must
 *  contain an environment definition like `<name>=<value>`. These
 *  values are internally stored.
 *
 *  Both `<name>` and `<value>` may contain variables, defined in the
 *  environment of the running process. Those variables are automatically
 *  expanded by this routine. A variable is a string, beginning with a
 *  dollar sign ($), followed by a letter or underscore and followed by a
 *  sequence of letters, ciphers or underscores. Any other character is
 *  interpreted as the end of the variable. For example: $HOME/bin, where
 *  $HOME may be a variable.
 *
 *  Empty lines in the file are ignored. Preceding spaces are removed in
 *  a stream record. Comment lines must begin with a hash character (#)
 *  and are ignored. An example of this kind of file is the filemapping
 *  file.
 *
 *  In erroneous situations FALSE is returned and ENVerrno should be examined.
 *
 *  ENVReadFile() does very little error checking.
 *
 *  Example:
 *
 *      #include <env.h>
 *      .
 *      ok = ENVReadFile("FILEMAP");
 *      .
 *
 *  Returns	: Indication of success
 *-----------------------------------------------------------------------
 */
bool
ENVReadFile(const char *name) /* Name of the filemap file */
{
    FILE	*fil = fopen(name, "r");
    char	buffer[BUFSIZ];
    int		lineno = 0;

    ENVerrno = ENVE_NOERROR;

    if (fil == NULL) {
	ENVerrno = ENVE_IO;
	(void) ERWrite(ENVerrno, "ENVReadFile.fopen",
			"%s: %s", name, strerror(errno));
    }
    while ((ENVerrno == ENVE_NOERROR)
	    && (fgets(buffer, BUFSIZ, fil) != NULL)) {
	char	*ptr;
	char	*buf = buffer;

	lineno++;
	while (PM_is_ascii_space(*buf)) buf++;

	if ((*buf != '\0') && (*buf != '#')) {
	    ptr = buf;

	    while ((*ptr != '\0')
		    && (isalpha(*ptr)
			|| (*ptr == '_')
			|| (*ptr == '-')
			|| ((ptr != buf) && (isdigit(*ptr))))) {
		ptr++;
	    }

	    if ((ptr == buf) || (*ptr != '=')) {
		ENVerrno = ENVE_FILEMAP;
		(void) ERWrite(ENVerrno, "ENVReadFile",
				"file %s, line=%d", name, lineno);
	    } else {
		int	len;

		*ptr++ = '\0';			/* remove '=' */
		len = (int) strlen(ptr) - 1;
		if (*(ptr + len) == '\n') {
		    *(ptr + len) = '\0';	/* remove '\n' */
		}

		/* will set errno on error */
		(void) ENVSetVar(buf, ENVExpandVariables(ptr));
	    }
	}
    } /* while (fgets...) */

    if (ENVerrno == ENVE_NOERROR) {
	if (! feof(fil)) {
	    ENVerrno = ENVE_IO;
	    (void) ERWrite(ENVerrno, "ENVReadFile.feof", "expected EOF!");
	}
    }
    if (fil != NULL) {
	(void) fclose(fil);
    }

    return (ENVerrno == ENVE_NOERROR);
} /* ENVReadFile() */

/*------------------------------------------------------------------------
 *  Function	: ENVBasename
 *  Purpose	: Retrieve the last part of a path
 *  Method	: look backwards from the end for significant chars
 *
 *  History:
 *  07-02-1995	: eck      : creation
 *  21-04-1995	: eck      : Added VMS detection
 *
 *  Notes	:
 *  This routine is system dependent. On a vax, both __VMS and
 *  _POSIX_SOURCE are defined. This is independent of whether we use
 *  the posix layer. Therefore, the code tries to detect whether the path
 *  is a VMS or a POSIX-path. When it does not contain a ']' or a ':', it
 *  is assumed POSIX.
 *
 *  Returns	: The base-name of the path.
 *------------------------------------------------------------------------
 */
const char *
ENVBasename(
	const char	*path)
{
    const char	*app_start;

    app_start = path + strlen(path) - 1;

    while (app_start > path) {
#if defined(_POSIX_SOURCE) || defined(_XOPEN_SOURCE)
	if (*app_start == '/') { return app_start + 1; }
#endif
	app_start--;
    }
    return path;
} /* ENVBasename() */

/*------------------------------------------------------------------------
 *  Function	: create_optstring
 *  Purpose	: Create a string for ENVGetOpt() with all the options
 *  Method	: Append standard option-string with the application one.
 *
 *  History:
 *  21-06-1994	: eck      : creation
 *
 *  Notes	:
 *  Does some checking: applications may only use lower-case option-letters.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
create_optstring(
	CSBuf_p 	all_opts,	/* CS for the concatenation-result */
	const char	*std_opts,	/* Standard options */
	const char	*appl_opts)	/* Application options */
{
    const char	*tmp = appl_opts;
    bool	ok = TRUE;

    CSSetString(all_opts, std_opts);

    if (tmp != NULL) {
	while (ok && *tmp) {
	    if (isupper(*tmp)) {
		ENVerrno = ENVE_OPTION;
		ERWrite(ENVerrno, "create_optstring", "Reserved option '%c' specified in %s", *tmp, appl_opts);
		ok = FALSE;
	    } else if (*(tmp + 1) == ':') {
		tmp += 2;
	    } else {
		tmp++;
	    }
	} /* while (ok && *tmp) */
    } /* if (tmp != NULL) */

    if (ok && (appl_opts != NULL)) {
	CSAppend(all_opts, appl_opts);
    }
    return ok;  
} /* create_optstring() */

/*------------------------------------------------------------------------
 *  Function	: set_env_opts
 *  Purpose	: Store option string values in the environment
 *  Method	:
 *  Retrieve the `<name>=<value>` pair(s) from the options string.
 *  Store the expanded value into the local nametable with ENVSetVar().
 *
 *  History:
 *  27-02-2012	: heesakkm : creation
 *
 *  Notes	: The `<name>=` part is mandatory, `<value>` is optional.
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
set_env_opts(
	const char	*options)
{
    static const char	fname[] = "set_env_opts";
    CS_DYNAMIC(		csname);
    CS_DYNAMIC(		csval);
    const char		*name_start, *name_end;
    const char		*val_start, *val_end;
    const char		*ptr;

    ptr = options;
    while ((ptr != NULL) && (*ptr != '\0')) {
	name_start = PM_opt_findname(ptr);
	name_end = PM_opt_findnameend(ptr);
	val_start = PM_opt_findval(ptr);
	val_end = PM_opt_findvalend(ptr);

	(void) CSSetStringN(csname, name_start, name_end - name_start);
	(void) CSSetStringN(csval, val_start, val_end - val_start);

	if ((name_end == val_start) || (CSLength(csname) == 0)) {
	    ENVerrno = ENVE_OPTION;
	    ERWrite(ENVerrno, fname,
		    "Invalid option \"%s%s%s\" in string \"%s\"",
		    CSContents(csname),
		    (name_end == val_start) ? "" : "=",
		    CSContents(csval), options);
	    break;
	}
	if (!ENVSetVar(CSContents(csname),
			ENVExpandVariables(CSContents(csval)))) {
	    /* errno is set */
	    break;
	}
	ptr = PM_opt_findnext(ptr);
    }
    CSDestroy(csname);
    CSDestroy(csval);

    return (ENVerrno == ENVE_NOERROR);
} /* set_env_opts() */

/*-----------------------------------------------------------------------
 *  Function	: ENVInit
 *  Purpose	: Initialize the environment of an application
 *  Method	: ENVGetOpt()-calls and ENVSetVar()
 *
 *  History:
 *  23-11-1993	: eck      : creation
 *  21-06-1994	: eck      : Added extra params for application-options
 *  17-05-1996	: eck      : Changed srand() call: now uses time.
 *  06-11-1996	: eck      : Changed srand() call again: also use pid.
 *  21-11-1996  : PvdM     : Adjusted for Windows.
 *  21-09-1998	: eck      : Don't set $TODAY anymore (interpreted dynamically)
 *  07-08-2002	: jaap     : Support for reading multiple configuration files
 *  11-09-2003	: paul     : Added ENVutf8 (Unicode Master Switch)
 *  27-02-2012	: heesakkm : Per request of HvE: removed obsolete -L, reused -S.
 *
 *  Notes	:
 *  'argc' is the number of command-line arguments the application was
 *  invoked with.
 *
 *  'argv' is a pointer to an array of character strings that
 *  contain the arguments, one per string. By convention argv[0] is the
 *  name by which the program was invoked, so argc is at least 1.
 *  Additionally, conform the C standard, argv[argc] must be a null
 *  pointer.
 *
 *  'appl_version' is (a pointer to) a string containing the version of the
 *  application. This version is displayed when the -V option is passed as
 *  command line argument.
 *
 *  'appl_options' is a (pointer to) string that indicates the options
 *  that the application accepts. This string is passed to the
 *  standard getopts()-function. Only lower-case options are allowed:
 *  upper-case letters are reserved for standard options. When there
 *  are no application options, appl_options should be NULL.
 *
 *  'func' is a pointer to a function that handles the options for the
 *  application. When an option is recognised as an application-option,
 *  opt_func is called with the option as argument. When there are no
 *  application options (and thus no function), ENV_NOFUNC can be used as
 *  argument.
 *
 *  ENVInit() uses the (standard) getopt()-routine to parse the
 *  arguments. Since getopt() uses a number of external variables,
 *  these variables have their counterpart in the ENV-library. These
 *  variables are:
 *
 *  - ENVoptarg - argument of an option (if it accepts one)
 *  - ENVoptind - the index of the next element of the argv[] vector to be processed
 *  - ENVopterr - if set to 1 before calling ENVInit(), getopt() will
 *    write diagnostic messages on stderr
 *  - ENVoptopt - the option on which an error occurred.
 *
 *  These variables (apart from ENVopterr) are not valid before calling
 *  ENVInit(). They can be used inside opt_func() (ENVoptarg) and
 *  after ENVInit() has finished (ENVoptind).
 *
 *  The function will internally store all standard arguments from the
 *  argument list and determine the other environment variables.
 *  Filemapping assignments will be stored by scanning the proper
 *  filemap (see chapter 1). In erroneous situations FALSE is
 *  returned and ENVerrno should be examined.
 *
 *  Example
 *
 *      #include <env.h>
 *      .
 *      static void
 *      do_option(int option)
 *      {
 *          switch (option) {
 *          case 'a':	a_flag++; break;
 *          case 'n':	count = atoi(ENVoptarg); break;
 *          }
 *      }
 *
 *      int
 *      main(int argc, char **argv)
 *      {
 *          int idx;
 *
 *          ENVopterr = 1;
 *          if (! ENVInit(argc, argv, "3.1", "an:", do_option) {
 *             return EXIT_FAILURE;
 *          }
 *          idx = ENVoptind;
 *
 *          while (idx < argc) {
 *              printf("%d : %s\n", idx, argv[idx]);
 *              idx++;
 *          }
 *          return EXIT_SUCCESS;
 *      }
 *
 *  Returns	: Indication of success
 *-----------------------------------------------------------------------
 */
bool
ENVInit(
	int		argc,		/* As passed to main() */
	char		*argv[],	/* ... ditto */
	const char	*appl_version,	/* Application version */
	const char	*appl_options,	/* options for application */
	void		(*func)(int option))	/* application option-handler */
{
    int		opt;
    bool	ok;
    CS_LOCAL(	buffer, PATH_MAX);
    const char 	*stdopts = "C:E:M:S:T:U:VX:";
    CS_LOCAL(	allopts, 80);

#if defined(_POSIX_SOURCE) || defined(_XOPEN_SOURCE)
    mode_t	mask;

    mask = umask((mode_t) 0);
    (void) umask(mask & ~S_IWGRP);	/* Turn off group write-protection */
#endif /* defined(_POSIX_SOURCE) */

    ENVerrno = ENVE_NOERROR;

    ok = ENVSetVar(applname_str, ENVBasename(argv[0]));

    if (ok) {
#if defined(_POSIX_SOURCE) || defined(_XOPEN_SOURCE)
	(void) CSPrintf(buffer, "%ld", (long) getpid());
#else
	CSSetString(buffer, "UNKNOWN");
#endif
	ok = ENVSetVar(applident_str, CSContents(buffer));
    }
    if (ok) {
	if ((appl_options != NULL) && (*appl_options != '\0')
	    && (func == ENV_NOFUNC)) {
	    /* Application has given options but no function to handle them */
	    ok = FALSE;
	    ENVerrno = ENVE_OPTION;
	} else {
	    ok = create_optstring(allopts, stdopts, appl_options);
	}
    }
    while (ok && ((opt = ENVGetOpt(argc, argv, CSContents(allopts))) != -1)) {
	switch (opt) {
	case 'C':
		ok = ENVSetVar(actabl_catStr, ENVoptarg);
		break;
	case 'E':
		ok = ENVSetVar(errorlog_str, ENVoptarg);
		break;
	case 'M':
		ok = ENVSetVar(filemap_str, ENVoptarg);
		break;
	case 'S':
		ok = set_env_opts(ENVoptarg);
		break;
	case 'T':
		ok = ENVSetVar(tb_inputfile_str, ENVoptarg);
		break;
	case 'V':
		if (appl_version) {
		    fprintf(stderr, "%s ", appl_version);
		}
		fprintf(stderr, "%s\n", ENVversion ? ENVversion : "");
		exit(EXIT_SUCCESS);
		break;
	case 'X':
		ok = ENVSetVar(extended_opts_str, ENVoptarg);
		break;
	case ':':
	case '?':
		ok = FALSE;
		ENVerrno = ENVE_OPTION;
		break;
	default:
		if (func != ENV_NOFUNC) {
		    func(opt);
		}
	} /* switch(opt) */
    }
    if (ok) {
	const char	*map = ENVGetVar(filemap_str);

	if (map == NULL) {
	    map = filemap_str;
	}
	{
	    CS_LOCAL(path, PATH_MAX);
	    const char *next = map;
	    do {
		const char *colon = strchr(next, ':');
		if (colon == NULL) {
		    ok = CSSetString(path, next);
		    next = NULL;
		} else {
		    CSClear(path);
		    ok = CSAppendN(path, next, colon - next);
		    next = colon + 1;
		}
		if (ok) {
		    ok = ENVReadFile(CSContents(path));
		} else {
		    ENVerrno = ENVE_FILEMAP;
		    ERWrite(ENVerrno, "ENVInit",
			    "Failed to parse FILEMAP environment setting");
		}
	    } while (ok && next != NULL && *next != '\0');
	}
    }
    /* Set the Unicode Master Switch */
    if (ok) {
	const char	*utf8_val = ENVGetVar(utf8_str);

	ENVutf8 = utf8_val != NULL && STREQ(utf8_val, "ON");
    }
    if (ok) {
	const char	*max_copies_val = ENVGetVar(max_copies_str);
	int		max_copies_int;

	if (max_copies_val != NULL) {
	    max_copies_int = atoi(max_copies_val);
	    if (max_copies_int != 99 && max_copies_int != 999) {
		ERWrite(ENVE_NOERROR, "ENVInit",
			"%s=%s expected either 99 or 999",
			max_copies_str, max_copies_val);
	    } else {
		ENV_max_copies = (PICCopyNum_t)max_copies_int;
	    }
	}
    }
    if (ok) {
	time_t	curtime;
	int	xtra;

	curtime = time(NULL);
#if defined(_POSIX_SOURCE) || defined(_XOPEN_SOURCE)
	xtra = atoi(ENVGetVar(applident_str));
#else
	xtra = 0;
#endif
	/* Initialize the random generator (for multiple port trick in NIO). */
	srand((unsigned) curtime + xtra);
    }
    return (ENVerrno == ENVE_NOERROR);
} /* ENVInit() */
