/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: envvar.c
 *  Description	: Handling of the environment variables
 *  Author(s)	: Hans van Eck
 *  Creation	: 23-11-1993
 *
 *  History:
 *  23-11-1993	: eck      : creation
 *  21-09-1998	: eck      : Added special handling of $TODAY
 *  15-08-2001  : lutzj    : Changed to use a nametable to hold the 
 *			   : environment, added ENVWalk
 *
 *========================================================================
 */

#pragma ident "$Id: envvar.c 335828 2016-06-10 11:33:48Z heesakkm $"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <pica.h>
#include <env.h>
#include <nmt.h>
#include <cs.h>
#include <er.h>

/* Pointer to nametable that holds the environment (key, value) pairs */
static NMTHandle_p ENVtable = NULL;

// Pointer to nametable used in substitute_today
static	NMTHandle_p	ENV_substitute_today_nmt;

/*------------------------------------------------------------------------
 *  Function	: shutdown_env_lib
 *  Purpose	: destroy static variables, to help searching for
 *		  memory leaks 
 *  Method	: to be installed with atexit() 
 *
 *  History:
 *  17-09-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static void
shutdown_env_lib(void)
{
    NMTDestroy(ENVtable);
    ENVtable = NULL;
    NMTDestroy(ENV_substitute_today_nmt);
    ENV_substitute_today_nmt = NULL;
}

/*------------------------------------------------------------------------
 *  Function	: substitute_today
 *  Purpose	: substitute $TODAY in an environment variable
 *  Method	:
 *
 *  History:
 *  15-09-1998	: eck      : creation
 *  15-10-1999	: eck      : Added curly-bracket handling
 *
 *  Notes	:
 *  - Expands logfile_$TODAY into logfile_980915 (or so, depending on date).
 *  - Expands multiple $TODAY occurrences per value: $TODAY/$TODAY works.
 *  - The allowed forms are $TODAY, $(TODAY) and ${TODAY}.
 *  - Perhaps a sequence of
 *	ENVSetVar("TODAY", ...);
 *	ENVSetVar(name, ENVExpandVariables(orig_value));
 *	ENVSetVar("TODAY", NULL);
 *    would have been easier, but slightly less accurate (it might expand
 *    other environment variables as well).
 *    Might you decide to ignore this, calling ENVExpandVariables() in
 *    ENVGetVar(), and having a special case for $TODAY would be even
 *    better (though you would have to watch out for infinite recursion).
 *  - A nametable is used for memory management.
 *
 *  Returns	: The expanded value
 *------------------------------------------------------------------------
 */
static const char *
substitute_today(
	const char	*name,		/* name of variable to expand */
	const char	*orig_value)	/* value to expand */
{
    const char	*tmpp;
    struct tm	*tim_struct;
    time_t	curtime;
    char	buff[80];
    int		var_len;
    CS_DYNAMIC(	result_value);
    bool	xpand;
    bool	copyvar;
    const char	*value = orig_value;

    if (value == NULL) return value;
    if (! strstr(value, "TODAY")) return value;

    if ((curtime = time(NULL)) == -1) {
	buff[0] = '\0';
    } else {
	tim_struct = localtime(&curtime);
	(void) strftime(buff, sizeof(buff), "%y%m%d", tim_struct);
    }
    do {
	/* Test if there is really a $TODAY or $(TODAY). */
	tmpp = strstr(value, "$TODAY");

	copyvar = FALSE;
	if (tmpp != NULL) {
	    const char *xp = tmpp + sizeof("$TODAY") - 1;
	    xpand = (*xp  == '\0')
		    || ((! isalnum(*xp)) && (*xp != '_'));
	    copyvar = ! xpand;
	    var_len = sizeof("$TODAY") - 1;
	} else {
	    tmpp = strstr(value, "$(TODAY)");
	    if (tmpp == NULL) tmpp = strstr(value, "${TODAY}");
	    xpand = (tmpp != NULL);
	    var_len = sizeof("$(TODAY)") - 1;
	}
	if (xpand) {
	    (void) CSAppendN(result_value, value, tmpp - value);
	    (void) CSAppend(result_value, buff);
	    value = tmpp + var_len;
	} else if (copyvar) {
	    (void) CSAppendN(result_value, value, tmpp - value + var_len);
	    value = tmpp + var_len;
	}
    } while (xpand || copyvar);

    (void) CSAppend(result_value, value);

    if (ENV_substitute_today_nmt == NULL) {
	ENV_substitute_today_nmt = NMTCreate("size=5", NMTNewOldString);
    }
    if ((ENV_substitute_today_nmt == NULL)
	|| (NMTPut(ENV_substitute_today_nmt, name, CSContents(result_value)) == NULL)) {
	if (NMTerrno == NMTE_NOMEM) ENVerrno = ENVE_NOMEM;
	(void) ERWrite(NMTerrno, "ENVGetVar.substitute_today",
			"expanding variable %s", name);
    }
    orig_value = NMTGet(ENV_substitute_today_nmt, name);
    (void) CSDestroy(result_value);
    return orig_value;
} /* substitute_today() */

/*-----------------------------------------------------------------------
 *  Function	: ENVGetVar
 *  Purpose	: Get the value of a variable
 *  Method	: Try NMTGet(). If nothing found, try with getenv()
 *  Author(s)	: Hans van Eck
 *  Creation	: 24-11-1993
 *
 *  History:
 *  24-11-1993	: eck      : creation
 *  15-08-2001  : lutzj    : use nametable
 *  17-09-2012	: sieburgw : CBS-4081: Give a one-time error-message
 *			     if called before ENVInit.
 *
 *  Notes:
 *  'name' is the name of the variable to be retrieved.
 *
 *  This function will first look in the internal table. If the variable
 *  is not found, the function will call the C function getenv().
 *
 *  This function will return a pointer to the value of the environment
 *  variable, or NULL if that variable is not defined.
 *
 *  Example:
 *
 *      #include <env.h>
 *      .
 *      homedir = ENVGetVar("HOME");
 *      .
 *
 *  Returns	: Pointer to value or NULL
 *-----------------------------------------------------------------------
 */
const char *
ENVGetVar(
	const char *name)		/* variable to get */
{
    const char *value;
    static bool been_here = FALSE;

    if (!been_here && (ENVtable == NULL)){
	been_here = TRUE;
	/* CBS-4081 */
	ERWrite(NMTerrno, "ENVGetVar", "EnvGetVar called before ENVInit ??");
    }

    if (ENVtable != NULL) {
	value = NMTGet(ENVtable, name);
	if (value != NULL) {
	    return substitute_today(name, value);
	}
    }

    /* nothing found, look in standard environment (don't substitute $TODAY) */
    return getenv(name);
} /* ENVGetVar() */

/*-----------------------------------------------------------------------
 *  Function	: ENVSetVar
 *  Purpose	: Set or delete an environment variable from the local nametable
 *  Method	: Call NMTPut()
 *  Author(s)	: Hans van Eck
 *  Creation	: 24-11-1993
 *
 *  History:
 *  24-11-1993	: eck      : creation
 *  15-08-2001  : lutzj    : use nametable
 *
 *  Notes:
 *  'name' is the name of the variable.
 *
 *  'value' is the value of the variable.
 *
 *  A NULL pointer will remove the variable from the internal table
 *  (although it may be present in the standard environment, which is also
 *  searched by ENVGetVar(...) ) .
 *
 *  This function sets environment variables. When 'name' is unknown
 *  it will be defined otherwise its value is redefined by 'value'.
 *  In erroneous situations FALSE is returned and ENVerrno should be
 *  examined.
 *
 *  Example:
 *
 *      #include <env.h>
 *      .
 *      (void) ENVSetVar("WARNING", "ON");
 *      .
 *
 *  Returns	: Indication of success
 *-----------------------------------------------------------------------
 */
bool
ENVSetVar(
	const char	*name,	/* Name of variable to be set */
	const char	*value)	/* Value of variable to be set */
{
    ENVerrno = ENVE_NOERROR;

    if (ENVtable == NULL) {
	ENVtable = NMTCreate(NULL, NMTNewOldString);
	atexit (shutdown_env_lib);
	if (NMTerrno != NMTE_NOERROR) {
	    (void) ERWrite(NMTerrno, "ENVSetVar", "NMTCreate failed");
	    ENVerrno = ENVE_NOMEM;	/* Assume NMTE_NOMEM */
	    return FALSE;
	}
    }

    (void) NMTPut(ENVtable, name, value);
    if (NMTerrno != NMTE_NOERROR) {
	(void) ERWrite(NMTerrno, "ENVSetVar", "NMTPut failed");
	ENVerrno = ENVE_NOMEM;		/* Assume NMTE_NOMEM */
	return FALSE;
    }

    return TRUE;
} /* ENVSetVar() */

/*------------------------------------------------------------------------
 *  Function	: ENVWalk
 *  Purpose	: Walk through environment nametable
 *  Method	: Call NMTWalk()
 *
 *  History:
 *  15-08-2001	: lutzj    : creation
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
ENVWalk(
	bool	(*HandleElt)(const char *name, const char *value, 
			     voidp xtra),	/* Callback function */
	voidp	xtra)		/* Extra value for the callback function */
{
    if (ENVtable != NULL) {
	return NMTWalk(ENVtable, (bool (*) (const char *, cvoidp, voidp))
		    HandleElt, xtra);
    } else {
	return TRUE;
    }
} /* ENVWalk() */
