<!--
#
#  Copyright (c) 2013 by OCLC
#
#  File		: lib_env.md
#  Description	: General description of the ENV library
#  Author(s)	: Eric Helsper
#  Creation	: 22-11-2012
#
#  History:
#  22-11-2012	: eric     : creation
#
#  SVN ident	: $Id: lib_env.md 307506 2014-01-22 10:52:31Z sieburgw $
-->

The ENV library
===============

No application works in a vacuum ; at least, not known to the authors.
There is always an environment. Important environment information is, for
instance, the name of a file the application wants to access. A simple approach
would be to hard-code that name into the program and to require that the file
will indeed be found where it is supposted to be found. This approach causes
problems in case, for whatever reason, the system manager wants to re-locate
the file. The ENV library offers a flexible solution to environment handling.

The ENV library supports a hidden datastructure containing environment
variables. The datastructure may be described as:

    <Environment Variables Table> = {<var-name><var-value>}...

At startup, an application may be passed command-line arguments. The ENVInit()
routine scans all passed arguments and stores the defined ones into the
datastructure. The following arguments, all optional, are defined:

- -E<errorlogfile>
- -M<filemapping>
- -V (version, see 1.1)

Example:

    POPC -M/pica/filemaps/popc

(note : all examples are indicative)

The ENV library will at least support the following variables:

variable name| description
-------------|--------------------------------------
TODAY        | current date ('yymmdd')
APPLNAME     | name of application
APPLIDENT    | process identification of application
FILEMAP      | name of filemapping file (see below)
ERRORLOG     | name of error logfile

TODAY - determined by the standard C date&time routines.

APPLNAME - conform the C standard, argv[0] contains the name of the application

APPLIDENT - determined by (the Posix call) getpid()

FILEMAP

1. argument to -M option
2. from standard environment (call getenv("FILEMAP") )
3. default "FILEMAP"

ERRORLOG

1. from filemap file, definition of "ERRORLOG"
2. argument to -E option
3. from standard environment (call getenv("ERRORLOG") )
4. ERR-library will assume <stderr>

The -E errorlog option is specially meant to be used in case something
goes wrong during program start-up. Normally the definition from the
filemap file will be used.

Environment variables may be set or retrieved by routines ENVSetVar and
ENVGetVar.

After scanning the command-line arguments, ENVInit() will open a special
file, the filemapping file. This file, implemented as a standard textfile,
contains a list of logical filenames coupled to their physical names.
The structure of the file may be described as:

    <filemapping file> =
	{<logical filename> "=" <physical filename>}...

Example:

    ERRORLOG=/pica/table/popc/errlog
    CMDSYNTAX=/pica/table/popc/cmdsyn
    GLOSSARY=/pica/table/popc/wordlist

ENVInit() will read this file and store each filemapping as environment
variable.

Suppose the application wants to access the command syntax file.
The proper way to handle this would be:

    x = ENVGetVar("CMDSYNTAX");

    if ((x != NULL) && (*x != '\0')) {
	fopen(x, "r");
    }


A filemapping file may be used by any number of applications. When files are
re-located only the filemapping files have to be adjusted, not the software.

NB 1: In the FILEMAP you may use dollar-variables ; a dollar-variable is a
previously, implicitly or explicitly defined variable whose value will
automatically by assumed when its name is encountered. A dollar-variable is
recognized by a dollar sign, followed by an alphanumeric string.

Example 1:

    $TABLEROOT=/pica/table/popc (explicit definition)
    ERRORLOG=$TABLEROOT/errorlog

Example 2:

    ERRORLOG=/pica/table/popc/errorlog_$TODAY

where $TODAY represents the implicitly defined string 'yymmdd' (see above).

NB 2: Parameters may be spread over multiple FILEMAP files. The proper syntax
to use these is with the –M option and a colon as follows

    -M<filemap1>:<filemap2>[:<filemap3>].

Example 3:

    csfpsch -M confdir/FILEMAP:confdir/FILEMAP_BIBSERV

Version Control
---------------

The ENV library supports a special feature concerning version control.
When the command line argument -V is passed, the ENVInit() routine will write
the version of the application (the appl_version input parameter of ENVInit()
to the standard error file. All other arguments are ignored and the application
is terminated. The system manager can check if the right version is used.

ENVInit Example
---------------
\code
#include <env.h>
static void

do_option(int option)
{
    switch (option) {
    case 'a':	a_flag++; break;
    case 'n':	count = atoi(ENVoptarg); break;
    }
}

int
main(int argc, char **argv)
{
    int idx;

    ENVopterr = 1;	/* Option errors to stdout */
    if (! ENVInit(argc, argv, "3.1", "an:", do_option) {
	return EXIT_FAILURE;
    }
    idx = ENVoptind;

    while (idx < argc) {
	printf("%d : %s\n", idx, argv[idx]);
	idx++;
    }
    return EXIT_SUCCESS;
 } /* main() */
\endcode
