/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: erall.c
 *  Description	: All the routines in the ER-library
 *
 *  Author(s)	: Clemens Buys
 *  Creation	: 10-12-1993
 *
 *  History:
 *  10-12-1993	: clemens  : creation
 *  10-03-1994	: eck      : Added ERWarning(), added errcod to ERTrace().
 *  21-03-1994	: HSi	   : Changed date format in errorlog.
 *  05-04-1995	: eck      : Changed all interfaces to ER*() routines.
 *  21-09-1998	: eck      : Open errorlog-box when ERRORLOG changes.
 *			   : This fixes the problem with $TODAY.
 *  23-01-2008  : fb       : CALL_CLOEXEC not defined for Windows either.
 *  12-09-2013	: martinz  : Added documentation from ERLIB.DOC.
 *========================================================================
 */

#pragma ident "$Id: erall.c 378287 2020-07-20 11:56:52Z heesakkm $"

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <pica.h>
#include <env.h>
#include <er.h>

ERROR_DEFINE(ERE_NOERROR);
ERROR_DEFINE(ERE_OPEN);
ERROR_DEFINE(ERE_IO);

ERErrno_t ERerrno = ERE_NOERROR;

static FILE	*logfile = NULL;

/* Use a 2K output buffer */
#define ERR_BUFSIZ	2048

static char	ERRbuf[ERR_BUFSIZ];
static char	ERfilename[PATH_MAX];

static bool	ER_atexit_installed;

static void
set_cloexec(
	FILE	*fil)
{
    int	fildes = fileno(fil);
    int flags;                  /* fcntl() flags */

    flags = fcntl(fildes, F_GETFD, 0);
    if (flags == -1) return;
    flags |= FD_CLOEXEC;
    fcntl(fildes, F_SETFD, flags);
} /* set_cloexec() */

/*------------------------------------------------------------------------
 *  Function	: shutdown_er_lib
 *  Purpose	: close logfile when application exits.
 *  Method	: to be installed with atexit() 
 *
 *  History:
 *  17-09-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static void
shutdown_er_lib(void)
{
    if (logfile != NULL){
	fclose(logfile);
	logfile = NULL;
    }
}

/*------------------------------------------------------------------------
 *  Function	: ERDoWrite
 *  Purpose	: Write the error to the errorlog-file
 *  Method	: Write the entries line by line
 *
 *  History:
 *  05-04-1995	: eck      : creation, code copied from ERWrite()
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
ERDoWrite(
	ERCode_t	errcod,		/* The error code */
	const char	*errloc,	/* The location of the error */
	const char	*errcom,	/* The comment (a format) */
	va_list		argp)		/* argument list pointer */
{
    char	buffer[100];
    struct tm	*tim_struct;
    time_t	curtime;
    const char	*value;
    const char	*filename;

    ERerrno = ERE_NOERROR;

    filename = ENVGetVar("ERRORLOG");
    if (filename == NULL) {
	filename = "ERRORLOG";
    }
    if ((logfile != NULL) && (! STREQ(filename, ERfilename))) {
	(void) fclose(logfile);
	logfile = NULL;
    }
    if (logfile == NULL) {
	logfile = fopen(filename, "a");

	if (logfile == NULL) {
	    fprintf(stderr,"could not open %s, errno = %d (%s)\n",
			    filename, errno, strerror(errno));
	    ERerrno = ERE_OPEN;
	} else {
	    (void) strncpy(ERfilename, filename, sizeof(ERfilename));
	    ERfilename[sizeof(ERfilename) - 1] = '\0';
	    setvbuf(logfile, ERRbuf, _IOFBF, sizeof(ERRbuf));
	    set_cloexec(logfile);

	    if (!ER_atexit_installed){
		// Only once for application : Register an atexit() function.
		atexit (shutdown_er_lib);
		ER_atexit_installed = TRUE;
	    }
	}
    }
    if (ERerrno == ERE_NOERROR) {
	fputs("---\n", logfile);
	/* fprintf(logfile, "---\n"); */
	if ((curtime = time(NULL)) != -1) {
	    tim_struct = localtime(&curtime);
	    strftime(buffer, sizeof(buffer),"%Y-%m-%d %H:%M:%S", tim_struct);
	    fprintf(logfile, "TIM : %s\n", buffer);
	}
	value = ENVGetVar("APPLIDENT");
	if (value != NULL) {
	    fprintf(logfile, "PID : %s\n", value);
	}
	value = ENVGetVar("REQUESTID");
	if (value != NULL) {
	    fprintf(logfile, "REQ : %s\n", value);
	}
	value = ENVGetVar("APPLNAME");
	if (value != NULL) {
	    fprintf(logfile, "NAM : %s\n", value);
	}
	if (errcod != NULL) {
	    fprintf(logfile, "ERR : %s\n", errcod);
	} else {
	    fprintf(logfile, "ERR : NONE\n");
	}
	if (errloc != NULL) {
	    fprintf(logfile, "LOC : %s\n", errloc);
	}
	if (errcom != NULL) {
	    fputs("COM : ", logfile);
	    vfprintf(logfile, errcom, argp);
	    fputc('\n', logfile);
	}
	fputs("---\n", logfile);
	/* fprintf(logfile, "---\n"); */
	fflush(logfile);
    }
    return (ERerrno == ERE_NOERROR);
} /* ERDoWrite() */

/*-----------------------------------------------------------------------
 *  Function	: ERWrite
 *  Purpose	: Log error
 *  Method	:
 *
 *  History:
 *  13-12-1993	: clemens  : creation
 *  21-03-1994	: HSi	   : changed dateformat in errorlog.
 *  05-04-1995	: eck      : Moved code to ERDoWrite() and used parenthesis
 *
 *  Notes	: If the logfile is not opened it is opened.
 *		  If the open fails ERE_OPEN is returned.
 *
 *  Description	:
 *  - errcod is the code of the error to log.
 *  - errloc is a zero-terminated string containing the location where the
 *    error was detected (e.g. name of function).
 *  - errcom is a zero-terminated (empty or filled) string containing
 *    additional information describing the error.
 *    errcom may be a standard C format-string ; in that case additional
 *    arguments must be passed in the function call (...).
 *
 *  In erroneous situations false is returned and ERerrno should be examined.
 *
 *  Notes	:
 *  Generally spoken, it will not be checked whether ERWrite() has completed
 *  successfully or not.
 *
 *  Error values:
 *
 *	ERE_NOERROR	(ok)
 *	ERE_IO		(io error)
 *
 *  Example	:
 *
 *	#include <er.h>
 *	.
 *	if (!DBOpen("file")) {
 *		ERWrite(DBerrno, jobname, "opening file");
 *	} else { // just to show the use of the format feature
 *              ERTrace(DBerrno, jobname, "user %d has opened dbs %s",
 * 	                usernumber, "file");
 * 	}
 *	.
 *
 *  Returns	: FALSE if fileopen or write fails, otherwise TRUE
 *-----------------------------------------------------------------------
 */
bool
ERWrite(
	ERCode_t	errcod,		/* The error code */
	const char	*errloc,	/* The location of the error */
	const char	*errcom,	/* The comment (a format) */
	...)				/* optional arguments */
{
    bool	ok;
    va_list	pvar;

    va_start(pvar, errcom);
    ok = ERDoWrite(errcod, errloc, errcom, pvar);
    va_end(pvar);

    return ok;
} /* ERWrite() */

/*-----------------------------------------------------------------------
 *  Function	: ERAbort
 *  Purpose	: Log error and abort application
 *  Method	:
 *
 *  Author(s)	: Clemens Buys
 *  Creation	: 13-12-1993
 *
 *  History:
 *  13-12-1993	: clemens  : creation
 *  05-04-1995	: eck      : Used parenthesis
 *
 *  Notes	: After the error has been logged the routine calls abort()
 *
 *  Description	:
 *  - errcod is the code of the error to log.
 *  - errloc is a zero-terminated string containing the location where the
 *    error was detected (e.g. name of function).
 *  - errcom is a zero-terminated (empty or filled) string containing
 *    additional information concerning the error.
 *    errcom may be a standard C format-string ; in that case additional
 *    arguments must be passed in the function call (...).
 *
 *  Error values:
 *
 *	(none)
 *
 *  Example	:
 *
 *	#include <er.h>
 *	.
 *	if (!TBRead("mode=exact", pkey, NULL, 0, CBbuf, sizeof(CBbuf))) {
 *	    ERAbort(TBerrno, jobname, "reading table");
 *	    // application will not return here
 *	}
 *	.
 *
 *  Returns	: TRUE
 *-----------------------------------------------------------------------
 */
bool
ERAbort(
	ERCode_t	errcod,		/* The error code */
	const char	*errloc,	/* The location of the error */
	const char	*errcom,	/* The comment (a format) */
	...)				/* optional arguments */
{
    va_list	pvar;

    va_start(pvar, errcom);
    (void) ERDoWrite(errcod, errloc, errcom, pvar);
    va_end(pvar);

    abort();

    /* NOTREACHED */
    return TRUE;
} /* ERAbort() */

/*------------------------------------------------------------------------
 *  Function	: ERWarning
 *  Purpose	: Log error if warning-mode is set to ON
 *
 *  History:
 *  09-03-1994	: eck      : creation
 *  05-04-1995	: eck      : Used parenthesis
 *
 *  Notes	:
 *  The warning parameter is retrieved from the ENV library.
 *
 *  Generally spoken, the application will not check whether the ERTrace() has
 *  performed its function successfully or not.
 *
 *  Description	:
 *  - errloc is a zero-terminated string containing the location where the
 *    trace info was composed (e.g. name of function).
 *  - errcom is a zero-terminated (empty or filled) string containing
 *    additional information concerning the trace info.
 *    errcom may be a standard C format-string ; in that case additional
 *    arguments must be passed in the function call (...).
 *
 *  The routine will only log the warning info if global flag WARNING is set to
 *  ON, otherwise it will return success immediately.
 *  In erroneous situations false is returned and ERerrno should be examined.
 *
 *  Error values:
 *
 *	ERE_NOERROR	(ok)
 *	ERE_IO		(io error)
 *
 *  Example	:
 *
 *	#include <er.h>
 *	.
 *	if (!WMGetVar("TA", taal, sizeof(taal))) {
 *		ERWarning(WMerrno, jobname, "getting variable");
 * 	}
 * 	.
 *
 *  Returns	: FALSE if warning is ON and write fails, otherwise TRUE
 *------------------------------------------------------------------------
 */
bool
ERWarning(
	ERCode_t	errcod,		/* Error number */
	const char	*errloc,	/* Error location */
	const char	*errcom,	/* Comment (NULL if none) */
	...)				/* optional arguments */
{
    const char *warning_val;
    bool	ok;

    ERerrno = ERE_NOERROR;
    ok = TRUE;

    /* Assign && test */
    if ((warning_val = ENVGetVar("WARNING")) != NULL) {
	if (STREQ(warning_val, "ON")) {
	    va_list	pvar;

	    va_start(pvar, errcom);
	    ok = ERDoWrite(errcod, errloc, errcom, pvar);
	    va_end(pvar);
	}
    }
    return ok;
} /* ERWarning() */

/*-----------------------------------------------------------------------
 *  Function	: ERTrace
 *  Purpose	: Log error if trace-mode is set to ON
 *  Method	:
 *  Author(s)	: Clemens Buys
 *  Creation	: 13-12-1993
 *
 *  History:
 *  13-12-1993	: clemens  : creation
 *  09-03-1994	: eck      : Added errcod parameter, used STREQ() for strcmp()
 *  05-04-1995	: eck      : Used parenthesis
 *
 *  Notes	:
 *  The trace parameter is retrieved from the ENV library.
 *
 *  Generally spoken, the application will not check whether the ERTrace has
 *  performed its function successfully or not.
 *
 *  Description	:
 *  - errloc is a zero-terminated string containing the location where the
 *    trace info was composed (e.g. name of function).
 *  - errcom is a zero-terminated (empty or filled) string containing
 *    additional information concerning the trace info.
 *    errcom may be a standard C format-string ; in that case additional
 *    arguments must be passed in the function call (...).
 *
 *  The routine will only log the trace info if global flag TRACE is set to ON,
 *  otherwise it will return success immediately.
 *  In erroneous situations false is returned and ERerrno should be examined.
 *
 *  Error values:
 *
 *	ERE_NOERROR	(ok)
 *	ERE_IO		(io error)
 *
 *  Example	:
 *
 *	#include <er.h>
 *	.
 *	if (!WMGetVar("TA", taal, sizeof(taal))) {
 *		ERTrace(WMerrno, jobname, "getting variable");
 *	}
 *	.
 *
 *  Returns	: FALSE if trace is ON and write fails, otherwise TRUE
 *-----------------------------------------------------------------------
 */
bool
ERTrace(
	ERCode_t	errcod,		/* The error/trace code */
	const char	*errloc,	/* The location of the error */
	const char	*errcom,	/* The comment (a format) */
	...)				/* optional arguments */
{
    const char *trace_val;
    bool	ok;

    ERerrno = ERE_NOERROR;
    ok = TRUE;

    /* Assign && test */
    if ((trace_val = ENVGetVar("TRACE")) != NULL) {
	if (STREQ(trace_val, "ON")) {
	    va_list	pvar;

	    va_start(pvar, errcom);
	    ok = ERDoWrite(errcod, errloc, errcom, pvar);
	    va_end(pvar);
	}
    }
    return ok;
} /* ERTrace() */

/*------------------------------------------------------------------------
 *  Function	: ERExit
 *  Purpose	: Log message and exit with error
 *  Method	:
 *
 *  History:
 *  10-10-2002	: jaap     : creation
 *
 *  Returns	: TRUE
 *------------------------------------------------------------------------
 */
bool
ERExit(
	ERCode_t	errcod,		/* Code */
	const char	*errloc,	/* Location */
	const char	*errcom,	/* Message (format) */
	...)				/* Format arguments */
{
    va_list		pvar;

    va_start(pvar, errcom);
    (void)ERDoWrite(errcod, errloc, errcom, pvar);
    va_end(pvar);

    exit(EXIT_FAILURE);

    /* NOTREACHED */
    return TRUE;
} /* ERExit() */
