/*
 *  Copyright (c) 1993 by Pica
 *
 *  File	: fep.c
 *  Description	: Frontend for Pica-3 communications
 *  Author(s)	: Rob Koopman
 *  Creation	: 24-11-1993
 *
 *  History:
 *  24-11-1993	: rob      : creation
 *  09-03-1994  : rob      : disconnect cpu loop removed
 *  17-03-1994  : rob      : assume no line errors! ... otherwise
 *			   : it's allmost impossible to do it right
 *			   : using standard unix IO
 *  21-03-1994	: eck      : initialise errno before read() in Receive()
 *  25-03-1994	: eck      : call ERTrace() (not ERWrite()) on (dis)connect
 *  11-08-1994	: wimo	   : added FESendAbort
 *  11-08-1994  : wimo	   : added socket support
 *  18-08-1994  : wimo	   : include NIO
 *  26-08-1994	: wimo	   : include time in active pdu
 *  29-09-1994	: eck      : changed some types FEContextType
 *			   : call FEWFNDProc() if connection breaks
 *  03-11-1994	: eck      : (re-) added Receive() routine. Removed
 *			   : sscanf()-s (gave problems in Windows DLL).
 *  15-02-1995	: eck      : Wait for DIsconnect-PDU (or other).
 *			   : Fixed length bug in RetransmitPDU()
 *  31-03-1995	: eck      : Use less NIORecv() calls in Receive.
 *  05-04-1995	: eck      : Changed an ERWrite() to ERWarning() in Receive().
 *  05-02-1996	: jv       : In FEStart: no NIOpen then init FEIT[i]....
 *  20-11-1996  : PvdM     : Adjusted for Windows compatibility.
 *
 *  Notes	:
 */

#pragma ident "$Id: fep.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pica.h>
#include <fep.h>
#include <nio.h>
#include <env.h>
#include <er.h>

ERROR_DEFINE(FEE_NOERROR);
ERROR_DEFINE(FEE_TIMEOUT);
ERROR_DEFINE(FEE_DISCONNECT);
ERROR_DEFINE(FEE_CONNABORT);

enum { SendSignal, ReadSignal, OpenSignal, CloseSignal, AbortSignal };

typedef enum {
    FEFREE,	/* not yet used */
    FEERROR,	/* not available */
    FECLOSEDA,	/* closed active frontend */
    FECLOSEDP,	/* closed passive frontend */
    FEWFNC,	/* wait for network connect */
    FEWFAP,	/* wait for accept pdu */
    FEOPEN,	/* wait for client request */
    FEOPENR,	/* open, wait for pdu */
    FEWFND,	/* wait for network disconnect */
    FEWFDP	/* wait for disconnect pdu */
} FEStateType;

#define FEMAXFRONTENDS 60
#define PDU_HEADLENGTH	6

typedef struct {
    FEHandle_t	id;
    FEConnect_t	connect;	/* FEConnected - FEDisconnected */
    NIOHandle_t	NIOHandle;
    FEType_t	type;		/* active / passive */
    FEStateType	state;		/* state of protocol machine */
    FEErrno_t	errors;		/* latest error */
    int		retrys;		/* nr of times data is retransmitted */
    char	*pdu;		/* latest sent pdu (for retransmit!) */
    char	*iobuf;		/* io buffer */
} FEContextType;

static FEContextType	FEIT[FEMAXFRONTENDS+1];
static int		FirstTime = 1;
static FEHandle_t	NextHandle = 314;

static int FEWFNDProc(int fi);

static void
SendPDU(int fi)
{
    int i = strlen(FEIT[fi].iobuf);
    /* send a pdu */
    if (!NIOSend(FEIT[fi].NIOHandle, FEIT[fi].iobuf, i)) return;
    if (strncmp(FEIT[fi].iobuf, "DA", 2) == 0) {
	/* retransmit buffer */
	FEIT[fi].retrys = 0;
	(void) strcpy(FEIT[fi].pdu, FEIT[fi].iobuf);
    } else {
	FEIT[fi].pdu[0] = 0;
    }
} /* SendPDU() */

static int
RetransmitPDU(int fi )
{
    int		i = strlen(FEIT[fi].pdu);
    const char	*retry_wait;
    int		sleeptime;

    if ((FEIT[fi].pdu[0]!=0) && (FEIT[fi].retrys < 3)) {
	retry_wait = ENVGetVar("FEP_RETRY_WAIT");
	if (retry_wait == NULL) {
	    sleeptime = 1;
	} else {
	    sleeptime = atoi(retry_wait);
	}
	if (sleeptime > 10) {
	    sleeptime = 10;
	}
	if (sleeptime > 0) {
	    (void) sleep((unsigned) sleeptime);
	}

	if (!NIOSend(FEIT[fi].NIOHandle, FEIT[fi].pdu, i)) return 0;
	FEIT[fi].retrys+=1;
	return 1;
    }
    if (FEIT[fi].pdu[0] != '\0') {
	(void) ERWrite(APPLE_FILEWRITE, __func__, "More than 3 retries!");
    }
    return 0;
} /* RetransmitPDU() */

/*
 * Receive - receive a PDU
 *
 * remarks:
 *	io should be no waited and non blocking
 *
 *	returns	>0	-> everything ok
 *		0	-> eof
 *		-1	-> serious error (see errno)
 */
static int
Receive(int nio_handle, char *buf, unsigned len)
{
    unsigned	lt;
    int		lr;
    unsigned	pdul;
    unsigned	i;

    /*
     * Read as much as possible in one go and keep index to the start of a
     * PDU (a PDU starts with a capital).  The rest is supposed to be
     * telnet stuff.
     */
    do {
	lt = 0;
	lr = NIORecv(nio_handle, buf, len);
	if (lr <= 0) {
	    ERWarning(FEE_DISCONNECT, "FEReceive - 1", "Unexpected line loss");
	    return 0;
	}
	/* skip leading TELNET commands by searching for PDU-start */
	while ((lt < (unsigned)lr) && (! isupper((unsigned char) buf[lt]))) {
	    lt++;
	}
    } while (lt == (unsigned)lr);

    /* Copy the PDU to the start of the buffer (if necessary). */
    if (lt != 0) {
	for (i = lt; i < (unsigned)lr ; i++) {
	    buf[i - lt] = buf[i];
	}
	lr -= lt;
    }

    /* Read until at least the header of the PDU is present. */
    lt = lr;
    while (lt < PDU_HEADLENGTH) {
	lr = NIORecv(nio_handle, buf + lt, len - lt);
	if (lr <= 0) {
	    ERWrite(FEE_DISCONNECT, "FEReceive - 2", "Unexpected line loss");
	    return 0;
	}
	lt += lr;
    }

    /* Compute the length of the rest of the PDU. */
    pdul = 0;
    for (i = 2; i < PDU_HEADLENGTH; i++) {
	pdul *= 10;
	pdul += buf[i] - '0';
    }
    if ((pdul < 3) || (pdul > len-PDU_HEADLENGTH)) {
	/* error in pdu -> abort connection! */
	ERWrite(FEE_DISCONNECT, "FEReceive - 3", "unexpected PDU length");
	return 0;
    }

    /* Now read the rest of the PDU. pdul contains the total size. */
    pdul += PDU_HEADLENGTH;
    while (lt < pdul) {
	lr = NIORecv(nio_handle, buf + lt, pdul - lt);
	if (lr <= 0) {
	    ERWrite(FEE_DISCONNECT, "FEReceive - 4", "Unexpected line loss");
	    return 0;
	}
	lt += lr;
    }

    /* And store a '\0' byte at the end. */
    buf[lt] = '\0';

    return lt;
} /* Receive() */

static void
ReceivePDU(int fi)
{
    int		l;
    int		lr;
    char	c;

    /* receive a pdu */
    /* for the first character we wait forever */
    /* read chars until etx or length or timeout */
    if ((lr = Receive(FEIT[fi].NIOHandle, FEIT[fi].iobuf, FE_PDUMAX))) {
	FEIT[fi].iobuf[lr] = '\0';
	/* we read something */
	c = FEIT[fi].iobuf[6];
        FEIT[fi].iobuf[6] = 0;
        l = atoi(FEIT[fi].iobuf+2);
        FEIT[fi].iobuf[6] = c;

	if ((l > 0) && (l+7 < FE_PDUMAX) && (l+6 == lr)) {
	    /* we expect things to be right */
	    FEIT[fi].iobuf[l+6] = 0;
	} else {
	    /* the PDU is garbled */
	    (void) strcpy(FEIT[fi].iobuf, "GB000399\003");
	}
    } else if (lr == -2) {
	/* probably bad line serious */
	(void) strcpy(FEIT[fi].iobuf, "GB000399\003");
    } else {
	/* connection lost forever */
	FEIT[fi].iobuf[0] = 0;
	FEIT[fi].errors = FEE_DISCONNECT;
	FEWFNDProc(fi);
    }
    return;
} /* ReceivePDU() */


static void
ZendPdu(int fi, const char *pdu, const char *data, const char *worm)
{
    int		l;
    int		lw;
    char	*b;

    /* de header */
    b	= FEIT[fi].iobuf;
    l	= strlen(data)+3;
    lw	= strlen(worm);
    if (lw > 0) { l += lw+1;}

    b[0] = pdu[0];
    b[1] = pdu[1];

    /* de lengte met voorloop nullen */
    (void) sprintf(&b[2], "%04d", l);

    /* de inhoud */
    (void) strcat(b, data);
    if (lw > 0) {
	(void) strcat(b, "10");
	(void) strcat(b, worm);
	b[strlen(b)-1] = 0;
    }

    /* de etx */
    (void) strcat(b, "99\003");

    SendPDU(fi);
} /* ZendPdu() */

static void
WormToPdu(int fi, const char *worm)
{
    /* da llll 10 worm 99 etx */
    /* da llll 01 tttt 10 worm 99 etx */

    char	ts[20];

    if (FEIT[fi].type == FEPassive) {
	/* no timer */
	ts[0] = 0;
    } else {
	unsigned	elapsed = NIOElapsed(FEIT[fi].NIOHandle);

	/* Make elapsed time fit in 4 bytes. */
	if (elapsed >= 10000) {
	    elapsed = 9999;
	}
	(void) sprintf(ts, "01%04d", elapsed);
    }

    /* stuur ook maar weg */
    ZendPdu(fi, "DA", ts, worm);

} /* WormToPdu() */

static void
PduToWorm(int fi, char *worm, int worml)
{
    /* da llll 01 worm 99 etx */
    /* da llll 10 tttt 01 worm 99 etx */
    /* --> worm etx */
    /* SHOULD READ NO MORE THAN WORML CHARACTERS! */

    char	*pdup;
    char	*wormp;
    int		l;
    char	c;

    pdup = FEIT[fi].iobuf;

    if (worm == NULL) {
	return;
    }

    if (strncmp(pdup, "DA", 2) == 0) {
	/* data -> worm */
	c = pdup[6];
        pdup[6] = 0;
        l = atoi(pdup+2);
        pdup[6] = c;

	if (strncmp(pdup+6, "01", 2) == 0) {
	    wormp = pdup+14;
	    l -= (6+2+3);
	} else {
	    wormp = pdup+8;
	    l -= (2+3);
	}
	if (l < 0) {
	    l = 0;
	}

	/* append etx */
	if (l < worml-2) {
	    (void) strncpy(worm, wormp, l);
	    worm[l]	= 3;
	    worm[l+1]	= 0;
	} else {
	    /* overflow error */
	    worm[0]	= 3;
	    worm[1]	= 0;
	}
    } else {
	/* no data return empty worm */
	worm[0] = 3;
	worm[1] = 0;
    }
} /* PduToWorm() */

static void
FEInit(void)
{
    int		i;

    FirstTime = 0;
    for (i = 0; i <= FEMAXFRONTENDS; i++) {
	FEIT[i].id		= FE_NOHANDLE;
	FEIT[i].connect		= FEDisconnected;
	FEIT[i].NIOHandle	= -1;
	FEIT[i].state		= FEFREE;
	FEIT[i].pdu		= NULL;
	FEIT[i].iobuf		= NULL;
	FEIT[i].type		= -1;
	FEIT[i].errors		= 0;
	FEIT[i].retrys		= 0;
    }

    /* last is sentinal */
    FEIT[FEMAXFRONTENDS].state = FEERROR;
} /* FEInit() */

static int
FEC(FEHandle_t h)
{
    int		i;

    i = 0;
    FEIT[FEMAXFRONTENDS].id = h;
    while (FEIT[i].id != h) { i++; }

    return i;
} /* FEC() */


static int
FEFreeIndex(void)
{
    int		i;

    for (i=0 ; i < FEMAXFRONTENDS; i++) {
	if (FEIT[i].state == FEFREE) {
	    return i;
	}
    }

    return FEMAXFRONTENDS;	/* The sentinel */
} /* FEFreeIndex() */


static int
FECLOSEDAProc(int fi)
{
    ZendPdu(fi, "CO", "010020", "");
    FEIT[fi].state = FEWFAP;
    return 1;
} /* FECLOSEDAProc() */

static int
FEWFNCProc(int fi)
{
    ReceivePDU(fi);
    if (strncmp(FEIT[fi].iobuf, "CO", 2) == 0) {
	ZendPdu(fi, "AC", "", "");
	FEIT[fi].state = FEOPEN;
	ERTrace(0, "FEWFNCProc", "Connection accepted");
    } else {
	/* weten we nog niet */
	FEIT[fi].state = FEWFND;
	return 0;
    }
    return 1;
}

static int
FEWFAPProc(int fi)
{
    ReceivePDU(fi);
    if (strncmp(FEIT[fi].iobuf, "AC", 2) == 0) {
	FEIT[fi].state = FEOPEN;
	return 1;
    }
    return 0;
} /* FEWFAPProc() */

/* application should read, write or stop */

static int
FEOPENProc(int fi, int sign, const char *worm, int worml)
{
    int		cont = 0;

    (void) worml;	// Keep lint happy

    switch (sign) {
    case ReadSignal :
	FEIT[fi].state = FEOPENR;
	cont = 1;
	break;

    case SendSignal :
	/* send data pdu worm */
	/* go back to user */
	WormToPdu(fi, worm);
	cont = 0;
	break;

    case CloseSignal :
	/* send finish pdu */
	ZendPdu(fi, "FI", "", "");
	FEIT[fi].state = FEWFDP;
	cont = 1;
	break;

    case AbortSignal :
	ZendPdu(fi, "AB", "171", "");
	cont = 0;
	break;
    }
    return cont;
} /* FEOPENProc() */

/* application should want to read a pdu */

static int
FEOPENRProc(int fi, int sign, char *worm, int worml)
{
    /* read pdu */
    char	*bp;
    int		cont;

    (void) sign;	// Keep lint happy

    /* en wat als het misgaat?? */
    (void) ReceivePDU(fi);

    /* if wrong no worm */
    worm[0] = 3;
    worm[1] = 0;

    bp = FEIT[fi].iobuf;

    cont = 1;

    if (strncmp("DA", bp, 2) == 0) {
	/* data -> return data */
	PduToWorm(fi, worm, worml);
	FEIT[fi].state = FEOPEN;
	cont = 0;
    } else if (bp[0] == 0) {
	FEIT[fi].state = FEWFND;
    } else if (strncmp("ST", bp, 2) == 0) {
	if (strncmp("00", &bp[11], 2) == 0) {
	/* status -> sent status pdu */
	    if (FEIT[fi].state == FEOPENR) {
	        ZendPdu(fi, "ST", "1015106", "");
	        /* state doesn't change state ? */
	        FEIT[fi].state = FEOPEN;
	        cont = 0;
	    }
	}
    } else if (strncmp("ER", bp, 2) == 0) {
	/* error -> send old data pdu */
	FEIT[fi].state = FEOPEN;
	cont = RetransmitPDU(fi);
    } else if (strncmp("GB", bp, 2) == 0) {
	/* garbage -> send error pdu */
	ZendPdu(fi, "ER", "", "");
    } else if (strncmp("TO", bp, 2) == 0) {
	/* send status pdu */
	ZendPdu(fi, "ST", "", "");
    } else if (strncmp("AB", bp, 2) == 0) {
	/* close connection */
	FEIT[fi].state = FEWFND;
    } else if (strncmp("FI", bp, 2) == 0) {
	ZendPdu(fi, "DI", "", "");
	FEIT[fi].state = FEWFND;
    } else {
	ZendPdu(fi, "ER", "", "");
    }
    return cont;
} /* FEOPENRProc() */

static int
FEWFNDProc(int fi)
{

    FEIT[fi].state = FEERROR;
    FEIT[fi].connect = FEDisconnected;
    ERTrace(0, "FEWFNDProc", "Connection closed");

    /* return handle to original state */
    NIOClose(FEIT[fi].NIOHandle);
    return 0;
} /* FEWFNDProc() */

static int
FEWFDPProc(int fi)
{
    /* wait for disconnect pdu (I am impatient) */
    ReceivePDU(fi);
    if (strncmp(FEIT[fi].iobuf, "DI", 2) == 0) {
	FEIT[fi].state = FEWFND;
	return 1;
    } else {
	FEIT[fi].state = FEWFND;
	return 1;
	
    }
} /* FEWFDPProc() */

/* FE is the frontend procedure.			*/
/* It is called from sent, receive, open and close	*/

static void
FE(
	int	fi,
	int	signal,
	char	*worm,
	int	worml)

{
    int		cont = 1;

    while (cont) {
	switch (FEIT[fi].state) {

	case FEFREE:	/* not yet used */
	    return;
	    /*break;*/

	case FEERROR:	/* not available or state error */
	    return;
	    /*break;*/

	case FECLOSEDA:	/* closed active frontend */
	    cont = FECLOSEDAProc(fi);
	    break;

	case FECLOSEDP:	/* closed passive frontend */
	    FEIT[fi].state = FEWFNC;
	    break;

	case FEWFNC:	/* wait for network connect */
	    cont = FEWFNCProc(fi);
	    break;

	case FEWFAP:	/* wait for accept pdu */
	    cont = FEWFAPProc(fi);
	    break;

	case FEOPEN:	/* wait for client request */
	    cont = FEOPENProc(fi, signal, worm, worml);
	    break;

	case FEOPENR:	/* open, wait for pdu */
	    cont = FEOPENRProc(fi, signal, worm, worml);
	    break;

	case FEWFND:	/* wait for network disconnect */
	    cont = FEWFNDProc(fi);
	    break;

	case FEWFDP:	/* wait for disconnect pdu */
	    cont = FEWFDPProc(fi);
	    break;

	default:
	    ERAbort(FEE_CONNABORT,"FE", "fi = %d, state =  %d", fi, FEIT[fi].state);
	}
    } /* while (cont) */
} /* FE() */

FEHandle_t
FEStart(FEType_t type, const char *FEAddress)
{

    int			i;
    char		options[80];

    /* first time user? init */
    if (FirstTime) { FEInit(); }

    /* find free frontend */
    i = FEFreeIndex();

    /* initieer blok */
    if (FEIT[i].state == FEFREE) {
	if (type == FEPassive) {
	    FEIT[i].state = FECLOSEDP;
	    (void) strcpy(options, "mode=passive");
	} else {
	    FEIT[i].state = FECLOSEDA;
	    (void) strcpy(options, "mode=active;");
	    (void) strcat(options, FEAddress);
	} 
	FEIT[i].NIOHandle = NIOOpen(options);
	if (FEIT[i].NIOHandle == -1) {
	    FEIT[i].state	= FEFREE;
	    (void) ERWrite(NIOerrno, "FEStart", "opening network");
	    return FE_NOHANDLE;
	}
	FEIT[i].id = NextHandle++;
	FEIT[i].type = type;
	FEIT[i].connect	= FEConnected;
	FEIT[i].pdu = (char *) malloc(FE_PDUMAX);
	FEIT[i].iobuf = (char *) malloc(FE_PDUMAX);
	return FEIT[i].id;
    } else {				/* (FEIT[i].state != FEFREE) */
	/* error */
	(void) ERWrite(APPLE_PARAM, __func__, "No free frontend handles");
	return FE_NOHANDLE;
    }
} /* FEStart */

FEConnect_t
FESend(FEHandle_t FEHandle, char *FEBuffer, int FELen)
{
    FE(FEC(FEHandle), SendSignal, FEBuffer, FELen);
    return FEIT[FEC(FEHandle)].connect;
} /* FESend() */

FEConnect_t
FESendAbort(FEHandle_t FEHandle)
{
    FE(FEC(FEHandle), AbortSignal, NULL, 0);
    return FEIT[FEC(FEHandle)].connect;
}

FEConnect_t
FEReceive(FEHandle_t FEHandle, char *FEBuffer, int FELen)
{
    FE(FEC(FEHandle), ReadSignal, FEBuffer, FELen);
    return FEIT[FEC(FEHandle)].connect;
} /* FEReceive() */

FEConnect_t
FEStop(FEHandle_t FEHandle)
{
    int	i = FEC(FEHandle);

    FE(i, CloseSignal, NULL, 0);

    FEIT[i].id		= FE_NOHANDLE;
    FEIT[i].connect	= FEDisconnected;
    FEIT[i].NIOHandle	= -1;
    FEIT[i].state	= FEFREE;
    free(FEIT[i].pdu);
    FEIT[i].pdu		= NULL;
    free(FEIT[i].iobuf);
    FEIT[i].iobuf	= NULL;
    FEIT[i].type	= -1;
    FEIT[i].errors	= 0;
    FEIT[i].retrys	= 0;

    return FEDisconnected;
} /* FEStop() */

FEErrno_t
FEError(FEHandle_t FEHandle)
{
    (void) FEHandle;	// Keep lint happy

    return FEE_NOERROR;
} /* FEError() */

int
FE_fildes(FEHandle_t FEHandle)
{
    int	i = FEC(FEHandle);

    if (i == FE_NOHANDLE) return -1;

    return NIO_fildes(FEIT[i].NIOHandle);
} /* FE_fildes() */
