/*========================================================================
 *  Copyright (c) 2004 by OCLC PICA
 *
 *  File	: hshtall.c
 *  Description	: All the routines for the HSHT-library
 *  Author(s)	: Maarten Heesakkers
 *  Creation	: 06-09-2004
 *
 *  History:
 *  06-09-2004	: heesakkm : creation (copied and adapted from NMT library)
 *  27-02-2012	: heesakkm : Added sanity check to HSHTDestroy() and HSHTClean()
 *  12-03-2014	: heesakkm : Added HSHTGetNrEntries().
 *
 *========================================================================
 */

#pragma ident "$Id: hshtall.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <hsht.h>
#include <er.h>

ERROR_DEFINE(HSHTE_NOERROR);
ERROR_DEFINE(HSHTE_NOMEM);

HSHTErrno_t HSHTerrno = HSHTE_NOERROR;

#define DEFAULT_TABLE_SIZE	211	/* A prime */
#define DEFAULT_MAXLOAD		5.0

typedef struct HSHTEntry_t *HSHTEntry_p;

struct HSHTEntry_t {
    HSHTKey_t	he_key;			/* key of the entry */
    HSHTEntry_p	he_next;		/* next entry */
    cvoidp	he_value;		/* value to return */
};

struct HSHTable_t {
    uint32	hsht_size;		/* size of the hash-table */
    uint32	hsht_nrentries;		/* nr. of hash-table entries */
    double	hsht_maxload;		/* max. of avg. nr. entries in bucket */
    int		hsht_walknesting;	/* nesting level of HSHTWalk() */
    cvoidp	(*hsht_newold)(		/* callback for (de-)allocation */
		    HSHTKey_t key, cvoidp newp, cvoidp oldp);
    HSHTEntry_p	*hsht_table;		/* the array with ptrs to entries */
};

/*------------------------------------------------------------------------
 *  Function	: HSHTCreate
 *  Purpose	: Create an empty hash-table
 *  Method	: Allocate and initialize data structures.
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  The following options are supported in the options-string:
 *  - "size=<nr>", where "<nr>" is the size of the hash-table. The default
 *    value is 211 (defined by DEFAULT_TABLE_SIZE). If the size is 1, you will
 *    get a linked list.
 *  - "maxload=<factor>", where factor is a the maximum load factor for this
 *    table. The default factor is 5 (defined by DEFAULT_MAXLOAD). The factor
 *    is a floating-point number. Usually, a lower load-factor means better the
 *    performance.
 *
 *  NewOld is an allocation/deallocation routine. This routine is called during
 *  an HSHTPut() and HSHTDestroy().
 *
 *  Description	:
 *  HSHTCreate() allocates the initial datastructures for a hash-table.
 *  The implementation uses a table of "<size>" entries. Elements that have
 *  congruent keys mod "<size>", are put in a linked list. When the number of
 *  elements causes the load to be higher than the maxload, the table-size is
 *  doubled.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *	HSHTE_NOMEM		(out of memory)
 *
 *  Returns	: A handle to the newly created hash-table or NULL
 *------------------------------------------------------------------------
 */
HSHTHandle_p
HSHTCreate(
	const char	*options,
	cvoidp		(*NewOld)(HSHTKey_t key, cvoidp newp, cvoidp oldp))
{
    char		valbuf[PM_INTLEN_MAX];
    int			tablesize = DEFAULT_TABLE_SIZE;
    double		maxload = DEFAULT_MAXLOAD;
    int			i;
    HSHTHandle_p	handle;

    HSHTerrno = HSHTE_NOERROR;

    if (PM_opt_fetch(options, "size", valbuf, sizeof(valbuf))) {
	tablesize = atoi(valbuf);
	if (tablesize <= 0) tablesize = DEFAULT_TABLE_SIZE;
    }
    if (PM_opt_fetch(options, "maxload", valbuf, sizeof(valbuf))) {
	maxload = strtod(valbuf, NULL);
	if (maxload <= 0) maxload = DEFAULT_MAXLOAD;
    }

    handle = malloc(sizeof(*handle));
    if (handle != NULL) {
	handle->hsht_table = malloc(tablesize * sizeof(HSHTEntry_p));
    }
    if ((handle == NULL) || (handle->hsht_table == NULL)) {
	HSHTerrno = HSHTE_NOMEM;
	ERWrite(HSHTerrno, __func__, "%s", options);

	if (handle != NULL) free(handle);
	handle = NULL;
    } else {
	handle->hsht_size = tablesize;
	handle->hsht_nrentries = 0;
	handle->hsht_maxload = maxload;
	handle->hsht_walknesting = 0;
	handle->hsht_newold = NewOld;
	for (i = 0; i < tablesize; i++) {
	    handle->hsht_table[i] = NULL;
	}
    }
    return handle;
} /* HSHTCreate() */

/*------------------------------------------------------------------------
 *  Function	: HSHTNewEntry
 *  Purpose	: Create a new hash-entry
 *  Method	: Allocate and initialize data structure
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Returns	: The entry allocated or NULL
 *------------------------------------------------------------------------
 */
static HSHTEntry_p
HSHTNewEntry(
	HSHTKey_t	key)		/* the key */
{
    static const char	fname[] = "HSHTNewEntry";
    HSHTEntry_p		ptr;

    ptr = malloc(sizeof(* ptr));
    if (ptr != NULL) {
	ptr->he_key = key;
	ptr->he_next = NULL;
	ptr->he_value = NULL;
    } else {
	HSHTerrno = HSHTE_NOMEM;
	ERWrite(HSHTerrno, fname, "key=%lu", (unsigned long) key);
    }
    return ptr;
} /* HSHTNewEntry() */

/*------------------------------------------------------------------------
 *  Function	: AddToList
 *  Purpose	: Add an element to a linked list
 *  Method	: obvious
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  This routine inserts the element. Appending might take longer (depending
 *  on the load).
 *  This routine is used for lists in the table as well as the free list.
 *------------------------------------------------------------------------
 */
static void
AddToList(
	HSHTEntry_p	*list,		/* the list to add to */
	HSHTEntry_p	eltp)		/* the element to add */
{
    eltp->he_next = *list;
    *list = eltp;
} /* AddToList() */

/*------------------------------------------------------------------------
 *  Function	: HSHTFreeEntry
 *  Purpose	: Free an entry
 *  Method	: obvious.
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  In the NMT library, a list of free entries was maintained here.
 *------------------------------------------------------------------------
 */
static void
HSHTFreeEntry(
	HSHTEntry_p	entryp)		/* Entry to free */
{
    free(entryp);
} /* HSHTFreeEntry() */

/*------------------------------------------------------------------------
 *  Function	: HSHTResizeTable
 *  Purpose	: Increase the size of the table
 *  Method	:
 *  Malloc a new array for the pointers and move each element from the old
 *  table to the new one.
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  The routine checks for resizing itself, so it can be called from
 *  multiple places.
 *------------------------------------------------------------------------
 */
static void
HSHTResizeTable(
	HSHTHandle_p	table)		/* the hash-table to expand */
{
    uint32	new_size;
    HSHTEntry_p	*new_table;
    uint32	i;

    if (table->hsht_walknesting > 0) {
	return;
    }

    /* First compute if resizeing is necessary. */
    new_size = table->hsht_size;
    while (((double) table->hsht_nrentries / new_size) > table->hsht_maxload) {
	new_size *= 2;
    }
    if (new_size == table->hsht_size) return;

    new_table = malloc(new_size * sizeof(HSHTEntry_p));
    if (new_table == NULL) return;
    for (i = 0; i < new_size; i++) {
	new_table[i] = NULL;
    }

    for (i = 0 ; i < table->hsht_size ; i++) {
	HSHTEntry_p	tmp = table->hsht_table[i];

	while ((tmp = table->hsht_table[i]) != NULL) {
	    table->hsht_table[i] = tmp->he_next;
	    tmp->he_next = NULL;

	    if (tmp->he_value != NULL) {
		AddToList(&new_table[tmp->he_key % new_size], tmp);
	    } else {
		HSHTFreeEntry(tmp);
		table->hsht_nrentries--;
	    }
	}
    }

    free(table->hsht_table);
    table->hsht_table = new_table;
    table->hsht_size = new_size;
} /* HSHTResizeTable() */

/*------------------------------------------------------------------------
 *  Function	: HSHT
 *  Purpose	: Get, Put or Delete a value
 *  Method	:
 *  - Determine hash-value and index in the table
 *  - Walk down linked list until the address of the pointer to the right entry
 *    is found (this eases a possible store)
 *  - if we shouldn't put a value, return the value found
 *  - otherwise call the user-supplied NewOld() routine and put the value
 *    returned by it into the table
 *  - if the value stored in the table is NULL, free the entry
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  Don't change the code if you are not used to pointers to pointers.
 *
 *  Returns	: The value as found/stored in the table
 *------------------------------------------------------------------------
 */
cvoidp
HSHT(
	HSHTHandle_p	table,		/* the hash-table to search in */
	HSHTKey_t	key,		/* the key */
	cvoidp		value,		/* the value (ignored for get) */
	bool		mustput)	/* TRUE if called from HSHTPut */
{
    uint32	idx = key % (table->hsht_size);
    HSHTEntry_p	*entry_pp = &(table->hsht_table[idx]);
    HSHTEntry_p	entry_p;
    bool	entry_added = FALSE;

    HSHTerrno = HSHTE_NOERROR;

    while (*entry_pp != NULL) {
	entry_p = *entry_pp;

	if (entry_p->he_key == key) {
	    break;
	}
	entry_pp = &(entry_p->he_next);
    }
    if (! mustput) {
	if (*entry_pp == NULL) return NULL;
	return (*entry_pp)->he_value;
    }

    /* Only create an entry when not deleting. */
    if ((value != NULL) && (*entry_pp == NULL)) {
	table->hsht_nrentries++;
	*entry_pp = HSHTNewEntry(key);

	/* We might need to resize the table. */
	entry_added = TRUE;
    }
    if (*entry_pp == NULL) {
	return NULL;
    }
    entry_p = *entry_pp;

    /* Call the callback (if any). */
    if (table->hsht_newold != HSHT_NOFUNC) {
	entry_p->he_value = table->hsht_newold(entry_p->he_key, value,
						   entry_p->he_value);
    } else {
	entry_p->he_value = value;
    }
    if (entry_p->he_value == NULL) {
	/* entry_pp still points to a pointer to entry_p. Change this pointer
	 * so entry_p is removed from the linked list.
	 */
	*entry_pp = entry_p->he_next;
	entry_p->he_next = NULL;
	HSHTFreeEntry(entry_p);
	table->hsht_nrentries--;
	return NULL;
    }
    if (entry_added) {
	HSHTResizeTable(table);
    }
    return entry_p->he_value;
} /* HSHT() */

/*------------------------------------------------------------------------
 *  Function	: HSHTGet
 *  Purpose	: Retrieve a value from a hash-table
 *  Method	: Use the HSHTGet() macro
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  - 'table' must be a handle to a previously HSHTCreate()-ed table (and not
 *    yet HSHTDestroy()-ed).
 *  - 'key' is the key of the element to get.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *
 *  Returns	: A pointer to the found value or NULL
 *------------------------------------------------------------------------
 */
cvoidp
(HSHTGet)(
	HSHTHandle_p	table,		/* the hash-table handle */
	HSHTKey_t	key)		/* the key to look for */
{
    return HSHTGet(table, key);
} /* HSHTGet() */

/*------------------------------------------------------------------------
 *  Function	: HSHTPut
 *  Purpose	: Put a value into a hash-table
 *  Method	: Use the HSHTPut() macro
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  - 'handle' must be a handle to a previously HSHTCreate()-ed table (and not
 *    yet HSHTDestroy()-ed).
 *  - 'key' is the key of the element to put.
 *  - 'value' is the value of the element to put. If it is NULL, the old value
 *    will be deleted.
 *
 *  Description	:
 *  This function will cause the NewOld()-function (the second argument to
 *  HSHTCreate()) to be called with 'key' as the first parameter, 'value' as
 *  its second parameter, and the current value from the table as its third
 *  parameter. This means that the NewOld()-function can determine its own
 *  allocation/deallocation strategy. It is expected, though, that the routine
 *  will cause a deallocation of the first argument if the second argument is
 *  NULL. The NewOld() routine must return the pointer to store in the table.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *	HSHTE_NOMEM		(out of memory)
 *
 *  Returns	:
 *  The value stored in the table (pointer returned by the NewOld()-routine)
 *------------------------------------------------------------------------
 */
cvoidp
(HSHTPut)(
	HSHTHandle_p	table,		/* the hash-table handle */
	HSHTKey_t	key,		/* the key to put */
	cvoidp		value)		/* value to put */
{
    return HSHTPut(table, key, value);
} /* HSHTPut() */

/*------------------------------------------------------------------------
 *  Function	: HSHTDel
 *  Purpose	: Delete a key from a hash-table
 *  Method	: Use the HSHTDel() macro
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  - 'table' must be a handle to a previously HSHTCreate()-ed table (and not
 *    yet HSHTDestroy()-ed).
 *  - 'key' is the key of the element to delete.
 *
 *  HSHTDel() will not always delete the whole key (scope). This depends on the
 *  programmer supplied NewOld() routine.
 *
 *  Description	:
 *  This function will cause the NewOld()-function (the second argument to
 *  HSHTCreate()) to be called with 'key' as the first parameter, NULL as its
 *  second parameter, and the current value from the table as its third
 *  parameter.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *	HSHTE_NOMEM		(out of memory)
 *
 *  Returns	:
 *  The value stored in the table (pointer returned by the NewOld()-routine)
 *------------------------------------------------------------------------
 */
cvoidp
(HSHTDel)(
	HSHTHandle_p	table,		/* the hash-table handle */
	HSHTKey_t	key)		/* the key to delete */
{
    return HSHTDel(table, key);
} /* HSHTDel() */

/*------------------------------------------------------------------------
 *  Function	: HSHTWalk
 *  Purpose	: Call a user-supplied function for every element in the table
 *  Method	:
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  - 'table' must be a handle to a previously HSHTCreate()-ed table (and not
 *    yet HSHTDestroy()-ed).
 *  - 'HandleElt' is a (pointer to a) function that handles the element.
 *  - 'xtraparam' is an extra parameter that can be used to pass some extra
 *    information to HandleElt() (and back).
 *
 *  Perhaps this should have been done as a pair of get-first/get-next
 *  routines, but that would make it hard/impossible to graciously handle
 *  changes to the hash-table while walking down the list.
 *
 *  Description	:
 *  This function will call HandleElt() for each element in the hash-table in
 *  an unspecified order.
 *
 *  The HandleElt() function will receive the key of the entry as the first
 *  parameter and its value as the second. The value will be equal to
 *  HSHTGet(table, key). The third parameter, extra_arg, is the same as the
 *  ExtraArg given to HSHTWalk().
 *  If this routine returns TRUE, HSHTWalk() will continue normally. If it
 *  returns FALSE, HSHTWalk() returns immediately with the return value FALSE.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *
 *  Example	:
 *
 *	#include	<hsht.h>
 *
 *	static bool
 *	Printer(
 *	    HSHTKey_t	key,
 *	    cvoidp	value,
 *	    voidp	stream)
 *	{
 *	    (void) fprintf((FILE *)stream, "%d = %s\n", (int) key, (char *) value);
 *	    return TRUE;
 *	}
 *	...
 *	HSHWalk(hsht_handle, Printer, stdout);
 *
 *  Returns	:
 *  TRUE if the callback routine always returned TRUE. FALSE if the callback
 *  routine returned FALSE.
 *------------------------------------------------------------------------
 */
bool
HSHTWalk(
	HSHTHandle_p	table,
	HSHTHandleElt_f HandleElt,
	voidp		xtraparam)
{
    uint32	i;
    HSHTEntry_p	tmp;
    bool	go_on_flag = TRUE;

    HSHTerrno = HSHTE_NOERROR;

    table->hsht_walknesting++;
    for (i = 0 ; go_on_flag && i < table->hsht_size; i++) {
	tmp = table->hsht_table[i];
	while (go_on_flag && tmp != NULL) {
	    if (tmp->he_value != NULL) {
		go_on_flag = HandleElt(tmp->he_key, tmp->he_value, xtraparam);
	    }
	    tmp = tmp->he_next;
	}
    }
    table->hsht_walknesting--;

    /* Optionally resize the table. This is necessary when the callback
     * adds elements.
     */
    HSHTResizeTable(table);
    return go_on_flag;
} /* HSHTWalk() */

/*------------------------------------------------------------------------
 *  Function	: HSHTDeleteList
 *  Purpose	: Delete a linked list of hash-table-entries.
 *  Method	:
 *  Walk down linked list, deleting value(s) and (when del_all is TRUE), the
 *  key and the element itself.
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  - The value should be de-allocated in the programmer-supplied NewOld()-
 *    routine. This routine is called in a loop, so this library can be used to
 *    maintain scopes.
 *  - The pointers free()-ed are set to NULL. This simplifies catching errors.
 *------------------------------------------------------------------------
 */
static void
HSHTDeleteList(
	HSHTEntry_p	entryp,		/* list of entries to delete */
	cvoidp		(*newold)(HSHTKey_t key, cvoidp newp, cvoidp oldp))
{
    HSHTEntry_p	tmp;

    while (entryp != NULL) {
	tmp = entryp;
	entryp = entryp->he_next;

	if (newold != HSHT_NOFUNC) {
	    /* This should deallocate */
	    while (tmp->he_value != NULL) {
		tmp->he_value = newold(tmp->he_key, NULL, tmp->he_value);
	    }
	} else {
	    tmp->he_value = NULL;	 	/* safety ! */
	}
	HSHTFreeEntry(tmp);
    }
} /* HSHTDeleteList() */

/*------------------------------------------------------------------------
 *  Function	: HSHTClean
 *  Purpose	: Free the memory associated with the values in a hash-table.
 *  Method	:
 *  Call HSHTDeletList() with FALSE to indicate value-only deletion.
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *  27-02-2012	: heesakkm : Added sanity check on table.
 *
 *  Notes	:
 *  'table' must be a handle to a previously HSHTCreate()-ed table (and not yet
 *  HSHTDestroy()-ed).
 *
 *  This routine will cause the hash-table to be 'cleaned'. That is: all
 *  values are deleted, but the table itself will remain.
 *
 *  Description	:
 *  This routine will free memory associated with values stored in a
 *  hash-table. The table itself will not be deallocated. The NewOld()-function
 *  (given on HSHTCreate()) will be called for every value in the table. This
 *  might involve calling NewOld() in a loop for some keys.
 *
 *  After an HSHTClean(), the hash-table will be empty, but HSHThandle will
 *  still be valid.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *------------------------------------------------------------------------
 */
void
HSHTClean(
	HSHTHandle_p	table)		/* the hash-table to clean */
{
    uint32	i;

    HSHTerrno = HSHTE_NOERROR;

    if (table == NULL) return;

    for (i = 0 ; i < table->hsht_size; i++) {
	if (table->hsht_table[i] != NULL) {
	    (void) HSHTDeleteList(table->hsht_table[i], table->hsht_newold);
	    table->hsht_table[i] = NULL;
	}
    }
    table->hsht_nrentries = 0;
} /* HSHTClean() */

/*------------------------------------------------------------------------
 *  Function	: HSHTDestroy
 *  Purpose	: Free all memory associated with a hash-table.
 *  Method	:
 *  - call HSHTClean() to free all linked lists in the table
 *  - free the table with the pointers
 *  - free the handle
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *  27-02-2012	: heesakkm : Added sanity check on table.
 *
 *  Notes	:
 *  'table' must be a handle to a previously HSHTCreate()-ed table (and not
 *  yet HSHTDestroy()-ed).
 *
 *  The pointers free()-ed are set to NULL. This simplifies catching errors.
 *
 *  Description	:
 *  This routine will free all memory associated with the HSHThandle. The
 *  NewOld()-function (given on HSHTCreate()) will be called for every value in
 *  the table. This might involve calling NewOld() in a loop for some keys.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *------------------------------------------------------------------------
 */
void
HSHTDestroy(
	HSHTHandle_p	table)		/* the hash-table to destroy */
{
    HSHTerrno = HSHTE_NOERROR;

    if (table != NULL) {
	HSHTClean(table);
	free(table->hsht_table);
	table->hsht_table = NULL;
	free(table);
    }
} /* HSHTDestroy() */

/*------------------------------------------------------------------------
 *  Function	: HSHTNewOldString
 *  Purpose	: Allocation/deallocation for strings.
 *  Method	:
 *  If the first argument is NULL, then free oldp. Otherwise use realloc() to
 *  reduce memory overhead (though this causes copying)
 *
 *  History:
 *  07-09-2004	: heesakkm : creation
 *
 *  Notes	:
 *  This routine can be used as second argument in a HSHTCreate()-call.
 *  If there is already a value in the table, the new one will replace it.
 *
 *  Error values:
 *
 *	HSHTE_NOERROR		(ok)
 *	HSHTE_NOMEM		(out of memory)
 *
 *  Returns	: Pointer to the element to put into hash-table
 *------------------------------------------------------------------------
 */
cvoidp
HSHTNewOldString(
	HSHTKey_t	key,	/* the key of the record */
	cvoidp		newp,	/* the value given to HSHTPut */
	cvoidp		oldp)	/* the value currently in the table */
{
    static const char	fname[] = "HSHTNewOldString";
    const char		*news = newp;	/* newp as string */
    char		*tmp;

    (void) key;	// Keep lint happy

    HSHTerrno = HSHTE_NOERROR;

    if (newp == NULL) {
	free((voidp) oldp);
	return NULL;
    }
    tmp = realloc((voidp) oldp, strlen((char *)newp) + 1);
    if (tmp != NULL) {
	(void) strcpy(tmp, news);
    } else {
	HSHTerrno = HSHTE_NOMEM;
	ERWrite(HSHTerrno, fname, "no room to realloc");
	free((voidp) oldp);		/* Just delete the old one */
    }
    return tmp;
} /* HSHTNewOldString() */

/*------------------------------------------------------------------------
 *  Function	: HSHTGetNrEntries
 *  Purpose	: Retrieve the number of entries in the hash-table
 *  Method	:
 *
 *  History:
 *  12-03-2014	: heesakkm : creation
 *
 *  Returns	: The number of hash-table entries
 *------------------------------------------------------------------------
 */
uint32
HSHTGetNrEntries(
	HSHTHandle_p	table)
{
    return (table != NULL) ? table->hsht_nrentries : 0;
} /* HSHTGetNrEntries() */
