<!--
#
#  Copyright (c) 2013 by OCLC
#
#  File		: lib_hsht.md
#  Description	: General description of the ENV library
#  Author(s)	: Maarten Heesakkers
#  Creation	: 01-10-2013
#
#  History:
#  01-10-2013	: heesakkm : creation
#
#  SVN ident	: $Id: lib_hsht.md 307506 2014-01-22 10:52:31Z sieburgw $
-->

The HSHT library
================

The HSHT library contains functions for creating an abstract interface to
hash-tables.

The hash-table
--------------

This implementation of hash-tables is copied and adapted from the NMT library.
The main differences are:

1. no names, but numbers as keys
2. no unused entries in the table
3. no list of free entries

The hashtable-routines maintain a hash-table with key/value-pairs, with linked
lists for clashes.

The following datastructures are used in the implementation:
- HSHTEntry_t:
    - identifier key
    - pointer to next element (for clashes)
    - pointer to user-supplied value
- HSHTHandle_t:
    - a pointer to a table (array of pointers)
    - size of the table
    - an allocation/deallocation routine
    - nr. entries in the table
    - maximum load allowed (== size/nr.entries)
    - nesting level of HSHTWalk()

In order to have most flexibility, the allocation/deallocation routine can be
supplied by the programmer. Its purpose is to free the old value and return the
new value to insert. This gives a programmer a lot of flexibility in how to
handle different situations (fi. two HSHTPut()-s with different non-NULL
values). It is even possible to create scope-like functionality with this
mechanism.
See the NewOld-routine paragraph below for details about this routine.

The load balancing uses 2 variables. The nr. of entries is used to determine
whether or not to resize. The nesting-level for HSHTWalk()-s prevents
load-balancing during HSHTWalk()-s, which might otherwise cause problems.

The NewOld-routine
------------------

The HSHTCreate()-routine has a routine as it's second parameter: the NewOld()
routine. The NewOld() routine is called during an HSHTPut(), HSHTDel(),
HSHTClean() and HSHTDestroy(). Its job is to perform allocation/deallocation
when necessary. This routine must usually be supplied by the programmer, but
there is a default routine for strings that performs (re-)allocation. If no
allocation/deallocation is necessary (by the library), HSHT_NOFUNC can be used.

**Interface**

    cvoidp
    NewOld(
	    HSHTKey_t	Key,		/* The key of the element */
	    cvoidp	Newp,		/* The new value */
	    cvoidp	Oldp);		/* Value currently in the table */

Key is the key with which the value can be retrieved.
Newp is the value given in an HSHTPut().
Oldp is the value that is currently in the hash-table.

In general, the NewOld() routine should free the memory associated with Oldp
(which can be a complicated structure), but it could also refuse to do so if it
was not a deletion (in which case Oldp should be returned). This happens on two
successive HSHTPut()-calls with different (non-NULL) values. Since this
behavior is determined by the programmer, (s)he can test for this situation
after an HSHTPut().

When Newp is NULL, this must be seen as a deletion. The return value need not
necessarily be NULL, but it should do some form of 'delete', since the NewOld()
routine will be called in a loop during an HSHTDestroy(). This loop ends when
NewOld() returns NULL. During the loop, the NewOld()-routine will be called
with the value previously returned as second argument. This facility makes it
possible to create 'scopes'.

When Oldp is NULL, it means that the key was not in the table.

The NewOld()-routine must return the value to put into the table. Usually (but
not always), this will be the same as Newp.

There are two NewOld() interfaces supplied with the library: HSHTNewOldString()
and HSHT_NOFUNC. The annotated source of HSHTNewOldString() can be found on the
next page. HSHT_NOFUNC can be seen as a function that just returns Newp.

**A string example**

If the values are strings, deallocation and allocation could be combined into a
reallocation followed by a string-copy. This gives the following function:
\code
    cvoidp
    HSHTNewOldString(			/* Does string allocation */
	    HSHTKey_t	key,		/* The key (ignored here) */
	    cvoidp	newp,		/* Value to put */
	    cvoidp	oldp)		/* Value to delete */
    {
	const char	*news = (const char *) newp;
	char		*tmp;

	/* Test for deletion first */
	if (newp == NULL) {
	    free((voidp) oldp);
	    return NULL;
	}
	/* If we want to enforce deletion by programmer before inserting a new
	 * value, we must uncomment the statement:
	 *     if (oldp != NULL) return oldp;
	 */

	/* Free the old string and allocate room for the new one.
	 * Using a realloc() might be too expensive in some situations, since
	 * it might copy the old string into newly allocated memory.
	 * On the other hand, when oldp == newp, a free()/malloc() would not
	 * work properly: either this situation should be tested or
	 * allocation should be done using malloc()/free().
	 */
	tmp = realloc((voidp) oldp, strlen((char *) newp) + 1);
	if (tmp != NULL) {
	    (void) strcpy(tmp, news);
	} else {
	    /* The reallocation failed. This implies that oldp was not
	     * free()-ed. We do that here. If we want it to remain in the
	     * table, it should be returned as function value.
	     */
	    HSHTerrno = HSHTE_NOMEM;
	    (void) ERWrite(HSHTerrno, "HSHTNewOldString",
			     "no room to realloc");
	    free((voidp) oldp);
	}
	return tmp;
    } /* HSHTNewOldString() */
\endcode

This routine can be called with both static and dynamic strings: it does its
own allocation.
