/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: idb.c
 *  Description	: Routines of the Identified Buffers library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *  07-03-2014	: martinz  : Moved documentation from Word file.
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: idb.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <pica.h>
#include <idb.h>
#include <er.h>

#define MAXIDENT 50
#define MAXBUFFER 4000

typedef struct {
    char id[3];
    int  ip;
    int  il;
} ident_t;

struct IDB_t {
    ident_t *IdentPointers;
    int     IdentCount;
    int     IdentSize;
    char    *Buffer;
    int	    BufferLength;		
    int     BufferSize;
};

ERROR_DEFINE(IDBE_NOERROR);
ERROR_DEFINE(IDBE_NOMEM);
ERROR_DEFINE(IDBE_NOTFOUND);

IDBErrno_t IDBerrno = IDBE_NOERROR;

/*------------------------------------------------------------------------
 *  Function	: SizeBuf
 *  Purpose	: resize Buffer to size
 *------------------------------------------------------------------------
*/
static
bool
SizeBuf(IDBHandle_p handle, size_t size)
{
    char	*b;
    size_t	bl;

    if ((handle->BufferLength + size) > handle->BufferSize){
	bl = ((handle->BufferLength + size) / MAXBUFFER + 1) * MAXBUFFER;
	if (handle->Buffer == NULL) {
	    b = (char *) malloc (bl);
	} else {
	    b = (char *) realloc(handle->Buffer, bl);
	}	

        if (b == NULL) {
	    return FALSE;
	}

        handle->BufferSize = bl;
	handle->Buffer = b;
    }
    return TRUE;
} /* SizeBuf() */

/*------------------------------------------------------------------------
 *  Function	: SizeIdent
 *  Purpose	: resize IdentPointers to size
 *------------------------------------------------------------------------
*/
static
bool
SizeIdent(IDBHandle_p handle, size_t size)
{
    ident_t *i;
    size_t  il;

    if ((handle->IdentCount + size) > handle->IdentSize){
	il = ((handle->IdentCount + size) / MAXIDENT + 1) * MAXIDENT;
	if (handle->IdentPointers == NULL){
	    i = (ident_t *) malloc ( il*sizeof(ident_t) );
	} else {
	    i = (ident_t *)
                realloc( handle->IdentPointers, il*sizeof(ident_t) );
	}	

        if (i == NULL) {
	    return FALSE;
        }
	handle->IdentSize = il;
	handle->IdentPointers = i;
    }

    return TRUE;
} /* SizeIdent() */

/*------------------------------------------------------------------------
 *  Function	: IDBOpen
 *  Purpose	: Makes room for an Identified Buffer
 *  Method	: uses malloc and calloc
 *
 *  History:
 *  21-12-1993	: lolke    : creation
 *
 *  Description	:
 *
 *  The hidden structure IDB_t is allocated using malloc and calloc. Counters
 *  are initialized.
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	.
 *	handle = IDBOpen();
 *	if (handle == NULL) {
 *		ERWrite(IDBerrno, jobname, "opening idb");
 * 	}
 *	.
 *
 *  Error values:
 *
 *      IDBE_NOERROR	(ok)
 *      IDBE_NOMEM	(not enough memory)
 *
 *  Returns	: NULL if malloc or calloc fails
 *
 *------------------------------------------------------------------------
 */
IDBHandle_p
IDBOpen(void)
{
    IDBHandle_p handle;

    IDBerrno = IDBE_NOERROR;
    if ((handle = (IDBHandle_p) malloc(sizeof(struct IDB_t))) != NULL) {
	handle->IdentCount    = 0;
	handle->IdentSize     = 0;
	handle->BufferSize    = 0;
	handle->BufferLength  = 0;
    	handle->IdentPointers = NULL;
        handle->Buffer        = NULL;
        SizeIdent(handle,10);
        SizeBuf(handle,1000);
        return handle;
    }
    IDBerrno = IDBE_NOMEM;
    (void) ERWrite(IDBerrno, "IDBOpen", "Out of memory");
    return NULL;
} /* IDBOpen() */

/*------------------------------------------------------------------------
 *  Function	: IDBEmpty
 *  Purpose	: Empties an IDB
 *  Method	: resets counters
 *------------------------------------------------------------------------
 */
void
IDBEmpty(IDBHandle_p handle)
{
    handle->IdentCount    = 0;
    handle->BufferLength  = 0;
} /* IDBEmpty() */

/*------------------------------------------------------------------------
 *  Function	: IDBClose
 *  Purpose	: Free the room used with the handle given
 *  Method	: uses free
 *
 *  History:
 *  21-12-1993	: lolke    : creation
 *  25-04-1994	: eck      : Added test for NULL so a hendle returned by a
 *			   : failed IDBOpen() can be IDBClose()-d
 *
 *  Description	:
 *
 *  All the memory used by the handle is freed with the use of free. IDBerrno
 *  is set to IDBE_NOERROR.
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	.
 *	(void) IDBClose(handle);
 *	.
 *
 *  Error values:
 *
 *      IDBE_NOERROR		(ok)
 *
 *------------------------------------------------------------------------
 */
void
IDBClose(
	IDBHandle_p handle)	/* the handle to free */
{
    IDBerrno = IDBE_NOERROR;

    if (handle != NULL) {
	free(handle->IdentPointers);
	free(handle->Buffer);

	handle->IdentPointers = NULL;	/* prevent accidental usage */
	handle->Buffer = NULL;		/* ditto */
	free(handle);
    }
} /* IDBClose() */

/*------------------------------------------------------------------------
 *  Function	: FindIdent
 *  Purpose	: Search for id in IdentPointers
 *  Method	: Linear search
 *  Returns	: -1 if not found
 *
 *------------------------------------------------------------------------
 */
static int
FindIdent(IDBHandle_p handle, const char *id, int *ip, int *il)
{
    int i;

    for (i = 0;
	i < handle->IdentCount && strcmp(handle->IdentPointers[i].id, id) != 0;
	i++) {
        /* empty body */
    }
    if (i < handle->IdentCount) {
	*ip = handle->IdentPointers[i].ip;
	*il = handle->IdentPointers[i].il;
	return i;
    }
    return -1;
} /* FindIdent() */

/*------------------------------------------------------------------------
 *  Function	: IDBDel
 *  Purpose	: Find id and then delete contents and identification
 *
 *  History:
 *  22-12-1993	: lolke    : creation
 *
 *  Description	:
 *
 *  If 'id' is found delete id and contents.
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	.
 *	(void) IDBDel(handle, "ER");
 *	.
 *
 *  Error values:
 *
 *      IDBE_NOERROR	(ok)
 *
 *  Returns	: TRUE
 *
 *------------------------------------------------------------------------
 */
bool
IDBDel(
	IDBHandle_p	handle,		/* handle to hidden struct */
	const char	*id)		/* id to delete */
{
    int ip;
    int il;
    int i;
    int j;
    int len;

    IDBerrno = IDBE_NOERROR;
    if ((i = FindIdent(handle, id, &ip, &il)) != -1) {
	len = handle->BufferLength;
	for (j = ip; j <= len - il; j++) {
	    handle->Buffer[j] = handle->Buffer[j + il];
	}
	for (j = 0; j < handle->IdentCount; j++) {
	    if (handle->IdentPointers[j].ip > ip) {
		handle->IdentPointers[j].ip -= il;
	    }
	}
	for (j = i; j < handle->IdentCount; j++) {
	    handle->IdentPointers[j] = handle->IdentPointers[j + 1];
	}
	handle->BufferLength-=il;
	handle->IdentCount--;
    }
    return TRUE;
} /* IDBDel() */

/*------------------------------------------------------------------------
 *  Function	: IDBGetLength
 *  Purpose	: Get length of contents of id
 *  Returns	: -1 if not found
 *
 *------------------------------------------------------------------------
 */
int
IDBGetLength(IDBHandle_p handle, char *id)
{
    int ip;
    int il;

    IDBerrno = IDBE_NOERROR;
    if (FindIdent(handle, id, &ip, &il) != -1) {
	return il;
    }
    return -1;
} /* IDBGetLength() */

/*------------------------------------------------------------------------
 *  Function	: IDBGet
 *  Purpose	: Get contents of id
 *  Returns	: FALSE if not found, set IDBE_NOMEM if too big
 *
 *  Description	:
 *
 *  If 'id' is found and size is big enough return TRUE, otherwise FALSE. 'id'
 *  may consist of wildcards.
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	char		contents[30];
 *	.
 *	if (!IDBGet(handle, "ER", contents, sizeof(contents))) {
 *		ERWrite(IDBerrno, jobname, "getting contents");
 *	}
 * 	.
 *
 *  Error values:
 *
 *      IDBE_NOERROR	(ok)
 *      IDBE_NOMEM	(not enough memory or contents too long)
 *
 *------------------------------------------------------------------------
 */
bool
IDBGet(
	IDBHandle_p 	handle,		/* handle to hidden struct */
	const char	*id,		/* (returned) id to get */
	char		*contents,	/* the returned contents */
	size_t		size)		/* size of contents */

{
    int ip;
    int il;

    IDBerrno = IDBE_NOERROR;
    if (FindIdent(handle, id, &ip, &il) != -1) {
	if (il <= size) { size = il; } else { IDBerrno = IDBE_NOMEM;}
	if (size>0){
	    (void) strncpy(contents, handle->Buffer + ip, size);
	    contents[size-1] = 0;
	}
    } else {
	IDBerrno = IDBE_NOTFOUND;
        contents[0] = 0;
        return FALSE;
    }
    return IDBerrno == IDBE_NOERROR;
} /* IDBGet() */

/*------------------------------------------------------------------------
 *  Function	: IDBGetN
 *  Purpose	: Get contents of id
 *  Returns	: FALSE if not found, set IDBE_NOMEM if too big
 *
 *  Description	:
 *
 *  If next 'id' is found and size is big enough return TRUE, otherwise FALSE.
 *  'id' may consist of wildcards. The returned 'id' is the 'id' of the
 *  contents.
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	char		contents[30];
 *	.
 *	if (!IDBGetN(handle, 2, "ER", contents, sizeof(contents))) {
 *		ERWrite(IDBerrno, jobname, "getting contents");
 *	}
 *	.
 *
 *  Error values:
 *
 *      IDBE_NOERROR	(ok)
 *      IDBE_NOMEM	(not enough memory)
 *      IDBE_NOTFOUND	(buffer not found)
 *
 *------------------------------------------------------------------------
 */
bool
IDBGetN (
	IDBHandle_p 	handle,		/* handle to hidden struct */
	int		i,		/* number of next buffer */
	char		*id,		/* returned id to get next */
	char		*contents,	/* the returned contents */
	size_t		size)		/* size of contents */
{
    int ip;
    int il;

    IDBerrno = IDBE_NOERROR;
    if (i < handle->IdentCount) {
	ip    = handle->IdentPointers[i].ip;
	il    = handle->IdentPointers[i].il;
	id[0] = handle->IdentPointers[i].id[0];
	id[1] = handle->IdentPointers[i].id[1];
	id[2] = 0;
	
	if (il <= size) { size = il; } else { IDBerrno = IDBE_NOMEM;}
	if (size>0){
	    (void) strncpy(contents, handle->Buffer + ip, size);
	    contents[size-1] = 0;
	}
    } else {
	id[0] = 0;
	id[1] = 0;
	contents[0] = 0;
        IDBerrno = IDBE_NOTFOUND;
    }
    return IDBerrno == IDBE_NOERROR;
} /* IDBGetN() */

/*------------------------------------------------------------------------
 *  Function	: IDBPutA
 *  Purpose	: Put id and contents, even if already exists
 *
 *  Returns	: FALSE if out of memory
 *
 *  Description	:
 *
 *  If 'id' is found append contents, otherwise put id and contents.
 *  Return TRUE if there was enough room.
 *
 *  Error values:
 *
 *      IDBE_NOERROR	(ok)
 *      IDBE_NOMEM	(not enough memory)
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	.
 *	if (!IDBPutA(handle, "ER", "MEER TEKST")) {
 *		ERWrite(IDBerrno, jobname, "appending to variable");
 *	}
 *	.
 *
 *------------------------------------------------------------------------
 */
bool
IDBPutA(
	IDBHandle_p 	handle,		/* handle to hidden struct */
	const char	*id,		/* id to put */
	const char	*contents)	/* the contents */
{
    int i;
    int len;

    len = strlen(contents)+1;

    IDBerrno = IDBE_NOERROR;

    if (SizeBuf (handle, len) && SizeIdent (handle, 1)) {
	i = handle->IdentCount;
	handle->IdentCount++;
	(void) strcpy(handle->IdentPointers[i].id, id);

        handle->IdentPointers[i].ip = handle->BufferLength;
        handle->IdentPointers[i].il = len;
        (void) strcpy(handle->Buffer+handle->BufferLength, contents);
	handle->BufferLength += len;	
        return TRUE;
    } else {
	/* out of memory */
	IDBerrno = IDBE_NOMEM;
	return FALSE;
    }
} /* IDBPutA() */

/*------------------------------------------------------------------------
 *  Function	: IDBPut
 *  Purpose	: Put id and contents overwrite old id
 *  Returns	: FALSE if pointers no longer valid ie out of memory
 *
 *  Description	:
 *
 *  If 'id' is found overwrite contents, otherwise put id and contents.
 *  Return TRUE if there was enough room.
 *
 *  Example	:
 *
 *	#include <idb.h>
 *	IDBHandle_p	handle;
 *	.
 *	if (!IDBPut(handle, "ER", "tekst")) {
 *		ERWrite(IDBerrno, jobname, "putting idb");
 *	}
 *	.
 *
 *  Error values:
 *
 *      IDBE_NOERROR		(ok)
 *      IDBE_NOMEM		(not enough memory)
 *
 *------------------------------------------------------------------------
 */
bool
IDBPut(
	IDBHandle_p 	handle,		/* handle to hidden struct */
	const char	*id,		/* id to put */
	const char	*contents)	/* the contents */
{
    IDBDel(handle, id);

    return IDBPutA(handle, id, contents);
} /* IDBPut() */
