/*========================================================================
 *  Copyright (c) 2011 by OCLC
 *
 *  File	: ihl.c
 *  Description	: IP address handling functions
 *  Author(s)	: Hans van Eck
 *  Creation	: 16-03-2011
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *  15-07-2011	: heesakkm : Added IHL_Userdata_Get().
 *  15-02-2013	: sieburgw : Added IHL_CheckAddressInRange (CBS-4519).
 *
 *  Notes	: copied and adapted from hms_host.c
 *========================================================================
 */

#pragma ident "$Id: ihl.c 352214 2018-01-17 16:52:17Z vaneckh $"

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include <stdio.h>

#include <pica.h>
#include <picamisc.h>
#include <er.h>
#include <ihl.h>
#include <pa.h>

ERROR_DEFINE(IHLE_NOERROR);
ERROR_DEFINE(IHLE_NOMEM);
ERROR_DEFINE(IHLE_INTERNAL);

/* An addresss-type can be either IPv4 or IPv6. */
typedef enum {
    IHL_TYPE_IPv4,
    IHL_TYPE_IPv6
} IHL_Type_t;

/* An addresss-range can be a single address, a CIDR representation or a range
 * of addresses.
 */
typedef enum {
    IHL_RANGETYPE_SINGLE,
    IHL_RANGETYPE_CIDR,
    IHL_RANGETYPE_RANGE
} IHL_RangeType_y;

typedef uint32	IHL_IPv4_Address_t;

/* An IPv6 address is 128 bit, and we have no native type for that. We
 * implement it as 2 64-bit unsigned longs.
 */
typedef struct {
    uint64	ii6a_low;	/* Low 64-bit part of IPv6 address */
    uint64	ii6a_high;	/* High 64-bit part of IPv6 address */
} IHL_IPv6_Address_t;

/* Note that allocation/deallocation of IHL_Entry_t's occurs at multiple
 * places, so if this struct gets dynamically allocated stuff added to it, this
 * must be incorporated in multiple places.
 */
typedef struct IHL_Entry_t {
    IHL_Type_t		ihle_type;
    IHL_RangeType_y	ihle_rangetype;
    cvoidp		ihle_userdata;
    union {
	struct {
	    IHL_IPv4_Address_t	ihlea4_first;	/* 1st address in range */
	    IHL_IPv4_Address_t	ihlea4_last;	/* Last address in range */
	    int			ihlea4_nbits;	/* # bits in mask (for CIDR) */
	} ihlea_ipv4;
	struct {
	    IHL_IPv6_Address_t	ihlea6_first;	/* 1st address in range */
	    IHL_IPv6_Address_t	ihlea6_last;	/* Last address in range */
	    int			ihlea6_nbits;	/* # bits in mask (for CIDR) */
	} ihlea_ipv6;
    } ihle_address;
} IHL_Entry_t;

#define ihle4_first	ihle_address.ihlea_ipv4.ihlea4_first
#define ihle4_last	ihle_address.ihlea_ipv4.ihlea4_last
#define ihle4_nbits	ihle_address.ihlea_ipv4.ihlea4_nbits

#define ihle6_first	ihle_address.ihlea_ipv6.ihlea6_first
#define ihle6_last	ihle_address.ihlea_ipv6.ihlea6_last
#define ihle6_nbits	ihle_address.ihlea_ipv6.ihlea6_nbits

/* IHL_Handle_t consists of only 1 single PAHandle_p, in which all address
 * types and all addresses are stored in the order they are supplied.
 * We could split it up into IPv4 and IPv6 or do some kind of range ordering,
 * but we'll leave that to the future (if ever). At least the interface allows
 * it.
 */
struct IHL_Handle_t {
    PAHandle_p		ihl_list;
};

/*------------------------------------------------------------------------
 *  Function	: IHL_Print_Entry
 *  Purpose	: Print an IHL entry
 *  Method	:
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	: This function is only intended for debugging.
 *------------------------------------------------------------------------
 */
void
IHL_Print_Entry(
	IHL_Entry_p	ihep)
{
    if (ihep != NULL) {
	printf("type = %s\n", ihep->ihle_type == IHL_TYPE_IPv6 ? "IHL_TYPE_IPv6" : "IHL_TYPE_IPv4");
	printf("Range-type = ");
	switch (ihep->ihle_rangetype) {
	case IHL_RANGETYPE_SINGLE: printf("IHL_RANGETYPE_SINGLE"); break;
	case IHL_RANGETYPE_CIDR: printf("IHL_RANGETYPE_CIDR"); break;
	case IHL_RANGETYPE_RANGE: printf("IHL_RANGETYPE_RANGE"); break;
	default: printf("UNKNOWN"); break;
	}
	printf("\n");
	printf("%s\n", IHL_EntryToString(NULL, ihep));
    }
} /* IHL_Print_Entry() */

/*------------------------------------------------------------------------
 *  Function	: IHL_IPv6_Address_Cmp
 *  Purpose	: Compare 2 IPv6 addresses
 *  Method	: Obvious
 *
 *  History:
 *  20-06-2011	: eck      : creation
 *
 *  Notes	:
 *  This function is necessare since the IPv6 addresses are 128 bits and split
 *  into 2 64-bit parts.
 *
 *  Returns	:
 *  <0: left smaller than right part
 *  =0: parts are equal
 *  >0: left bigger than right part
 *------------------------------------------------------------------------
 */
static int64
IHL_IPv6_Address_Cmp(
	IHL_IPv6_Address_t	left,
	IHL_IPv6_Address_t	right)
{
    if (left.ii6a_high == right.ii6a_high) return left.ii6a_low - right.ii6a_low;
    return left.ii6a_high - right.ii6a_high;
} /* IHL_IPv6_Address_Cmp() */

/*------------------------------------------------------------------------
 *  Function	: IHL_NewOld_Entry
 *  Purpose	: Callback for PA_Add
 *  Method	: Copy the struct
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *
 *  Notes	:
 *  This function does it's own allocation/deallocation. The argument old value
 *  (a struct) is copied.
 *  The IHL_Entry_p's can also be handled in
 *  IHL_Entry_Parse()/IHL_Entry_Destroy().
 *
 *  Returns	: pointer to the new value
 *------------------------------------------------------------------------
 */
static cvoidp
IHL_NewOld_Entry(
	cvoidp	newval,
	cvoidp	oldval)
{
    if (newval == NULL) {
	free((voidp) oldval);
	return NULL;
    } else if (oldval == NULL) {
	oldval = malloc(sizeof(IHL_Entry_t));
    }
    memcpy((voidp) oldval , newval, sizeof(IHL_Entry_t));
    return oldval;
} /* IHL_NewOld_Entry() */

/*------------------------------------------------------------------------
 *  Function	: IHL_IPv4_Entry_Fill
 *  Purpose	: Fill an entry with an IPv4 address
 *  Method	: Parse the octets
 *
 *  History:
 *  20-06-2011	: eck      : creation
 *
 *  Notes	: Parses dotted IPv4 addresses.
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
IHL_IPv4_Entry_Fill(
	const char	**in,
	IHL_Entry_p	out)
{
    long int		octet[4];
    int			i;
    IHL_IPv4_Address_t	ipv4_out = 0;

    memset(out, '\0', sizeof(*out));

    for (i = 0; i < 4; i++) {
	char *endptr;

	octet[i] = strtol(*in, &endptr, 10);
	if (octet[i] < 0 || octet[i] > 255) break;
	if (*endptr == '\0') {
	    if (i < 3) break;
	} else {
	    if (endptr == *in
		|| (i < 3 && *endptr != '.')) break;
	}
	*in = endptr + (i < 3);
	ipv4_out = (ipv4_out << 8) + octet[i];
    }
    if (i == 4) {
	out->ihle_type = IHL_TYPE_IPv4;
	out->ihle_rangetype = IHL_RANGETYPE_SINGLE;
	out->ihle4_first = ipv4_out;
	out->ihle4_last = ipv4_out;
    }

    return i == 4;
} /* IHL_IPv4_Entry_Fill() */

/*------------------------------------------------------------------------
 *  Function	: my_hexstringtol
 *  Purpose	: Converts string with hex-digits to integer
 *  Method	: Read hex digits 1-by-1 and put them into the result value
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	:
 *  This function will mimic the behaviour of strtol(in, end, 16), except that
 *  it will not recognise hex strings with a leading 0x.
 *  It will also not skip leading white-space.
 *
 *  Returns	: The value found
 *------------------------------------------------------------------------
 */
static long
my_hexstringtol(
	const char	*in,
	char		**end)
{
    unsigned long	val = 0;
    bool		overflow = FALSE;

    while (isxdigit(*in)) {
	if (val & ((unsigned long) 0xf << (CHAR_BIT * sizeof(val) - 4))) {
	    overflow = TRUE;
	    break;
	}
	val = val << 4;
	val += isdigit(*in) ? *in - '0':
		isupper(*in) ? *in - 'A' + 10
			    : *in - 'a' + 10;
	in++;
    }
    if (end != NULL) {
	*end = (char *) in;
    }
    if (overflow) {
	errno = ERANGE;
	val = LONG_MAX;
    }
    return val;
} /* my_hexstringtol() */

/*------------------------------------------------------------------------
 *  Function	: IHL_IPv6_Entry_Fill
 *  Purpose	: Fill an entry with an IPv6 address
 *  Method	: Parse an IPv6 address string
 *
 *  History:
 *  20-06-2011	: eck      : creation
 *
 *  Notes	:
 *  Parses IPv6 addresses, except when they end in IPv4 addresses ("IPv4-mapped
 *  IPv6 address" and "IPv4-compatible IPv6 addresses"), which is not handled
 *  at the moment.
 *  An IPv6 address string can look like:
 *  1234:5678:90ab:cdef:1234:5678:90ab:cdef
 *  1234::5678		(meaning 1234:0000:0000:0000:0000:0000:0000:5678)
 *  1234:5678::cdef	(meaning 1234:5678:0000:0000:0000:0000:0000:cdef)
 *  ::1			(meaning 0000:0000:0000:0000:0000:0000:0000:0001)
 *  ::			(meaning 0000:0000:0000:0000:0000:0000:0000:0000)
 *
 *  For future reference:
 *  An IPv4-mapped IPv6 address is of the form
 *  ::192.87.44.12	(meaning 0000:0000:0000:0000:0000:0000:c057:2c0c)
 *
 *  An IPv4-compatible IPv6 addresses is of the form
 *  ::ffff:192.87.44.12	(meaning 0000:0000:0000:0000:0000:ffff:c057:2c0c)
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
IHL_IPv6_Entry_Fill(
	const char	**in,
	IHL_Entry_p	out)
{
#define	IPv6_NUM_WORDS	(8)
    unsigned long	word[IPv6_NUM_WORDS + 1];
    const char		*empty_colon = NULL;
    int			empty_at = -1;
    int			i;
    IHL_IPv6_Address_t	ipv6_out = { 0, 0 };
    bool		ok = TRUE;

    memset(out, '\0', sizeof(*out));
    memset(word, '\0', sizeof(word));

    if (STREQ(*in , "::")) {
	*in +=2;
    } else {
	i = 0;

	/* Let empty_colon point to the seconf colon in a "::" sequence. */
	empty_colon = strstr(*in, "::");
	if (empty_colon != NULL) {
	    if (*in == empty_colon) (*in)++;
	    empty_colon++;
	}

	/* Loop until the word array is full (it contains 1 extra element in
	 * order to catch specification errors).
	 */
	while (i < (sizeof(word) / sizeof(word[0]))) {
	    char *endptr;

	    /* If we're at the second colon of a :: sequence, set empty_at and
	     * advance the input pointer.
	     */
	    if (*in == empty_colon) {
		empty_at = i;
		(*in)++;
	    }

	    word[i] = my_hexstringtol(*in, &endptr);
	    if (*in == endptr) break;
	    if (word[i] > 0xffff) {
		ok = FALSE;
		break;
	    }

	    i++;
	    *in = endptr;

	    if (**in != ':') break;

	    (*in)++;
	}
	/* The string was only ok if we saw IPv6_NUM_WORDS words and no empty
	 * part or we saw fewer than IPv6_NUM_WORDS and an empty part.
	 */
	ok = ok && (((i == IPv6_NUM_WORDS) && (empty_at < 0))
		    || ((i < IPv6_NUM_WORDS) && (empty_at >= 0)));
	if (ok && (empty_at >= 0)) {
	    /* Fixup the "word" array by shifting the part after the :: to the
	     * end and clearing out the middle part.
	     */
	    memmove(&word[IPv6_NUM_WORDS - (i - empty_at)], &word[empty_at], (i - empty_at) * (sizeof(word[0])));
	    memset(&word[empty_at], '\0', (IPv6_NUM_WORDS - i) * sizeof(word[0]));
	}
    }

    if (ok) {
	/* The array "word" is now filled with the words. Put them in ipv6_out.
	 * The loop fills the high and low part in the same iteration.
	 */
	for (i = 0; i < (IPv6_NUM_WORDS / 2); i++) {
	    ipv6_out.ii6a_high = (ipv6_out.ii6a_high << 16) | word[i];
	    ipv6_out.ii6a_low = (ipv6_out.ii6a_low << 16) | word[i + (IPv6_NUM_WORDS / 2)];

	}
	out->ihle_type = IHL_TYPE_IPv6;
	out->ihle_rangetype = IHL_RANGETYPE_SINGLE;
	out->ihle6_first = ipv6_out;
	out->ihle6_last = ipv6_out;
    }

    return ok;
} /* IHL_IPv6_Entry_Fill() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Entry_Fill
 *  Purpose	: Attempt to read an IP host address
 *  Method	:
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *  17-06-2011	: eck      : Split into IPv4 and IPv6
 *  18-07-2011	: heesakkm : Assign userdata after filling the address entry.
 *
 *  Returns	: TRUE if an address was read, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
IHL_Entry_Fill(
	const char	**in,
	IHL_Entry_p	out,
	cvoidp		userdata)
{
    const char	*colon = strchr(*in, ':');
    const char	*dot = strchr(*in, '.');
    bool	ok;

    if (colon == NULL) {
	ok = IHL_IPv4_Entry_Fill(in, out);
    } else if (dot == NULL) {
	ok = IHL_IPv6_Entry_Fill(in, out);
    } else if (dot < colon) {
	ok = IHL_IPv4_Entry_Fill(in, out);
    } else {
	ok = IHL_IPv6_Entry_Fill(in, out);
    }
    out->ihle_userdata = userdata;
    return ok;
} /* IHL_Entry_Fill() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Entry_Parse
 *  Purpose	: Parse an IPv4/IPv6 address and return the host entry
 *  Method	: IHL_Entry_Fill()
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	:
 *  Advances the host pointer so the calling function can know where to continue
 *  (if necessary).
 *
 *  Returns	: the host entry (with range-type IHL_RANGETYPE_SINGLE) or NULL
 *------------------------------------------------------------------------
 */
IHL_Entry_p
IHL_Entry_Parse(
	const char	**host,
	cvoidp		userdata)
{
    IHL_Entry_p	entry = malloc(sizeof(*entry));

    memset(entry, '\0', sizeof(*entry));

    if (! IHL_Entry_Fill(host, entry, userdata)) {
	free(entry);
	entry = NULL;
    }
    return entry;
} /* IHL_Entry_Parse() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Entry_Destroy
 *  Purpose	: Destroy an IHL_Entry_p
 *  Method	: free()
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	:
 *  There is also malloc()/free() of IHL_Entry_p's in IHL_NewOld_Entry().
 *------------------------------------------------------------------------
 */
void
IHL_Entry_Destroy(
	IHL_Entry_p	ihe)
{
    free(ihe);
} /* IHL_Entry_Destroy() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Mask_Set
 *  Purpose	: Set a mask for an IPv4 or IPv6 address.
 *  Method	: Bit twiddling.
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	:
 *  The numbits means the number of bits of the network mask, (so 8 means that
 *  only the leftmost 8 bits are set).
 *
 *  The IHL_TYPE_IPv6 code mimics a native 128-bit integer type.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
IHL_Mask_Set(
	IHL_Entry_p	address,
	int		numbits)
{
    switch (address->ihle_type) {
    case IHL_TYPE_IPv4:
	{
	    IHL_IPv4_Address_t	mask;

	    if ((numbits < 0) || (numbits > 32)) return FALSE;

	    mask = (numbits != 0) ? 0xffffffff << (32 - numbits) : 0;

	    address->ihle_rangetype = IHL_RANGETYPE_CIDR;
	    address->ihle4_nbits = numbits;
	    address->ihle4_first &= mask;
	    address->ihle4_last |= ~mask;

	    return TRUE;
	}
    case IHL_TYPE_IPv6:
	{
	    IHL_IPv6_Address_t	mask;
	    uint64		native_mask;

	    if ((numbits < 0) || (numbits > 128)) return FALSE;

	    address->ihle_rangetype = IHL_RANGETYPE_CIDR;
	    if (numbits >= 64) {
		native_mask = (~ (uint64) 0) << (64 - (numbits - 64));

		mask.ii6a_low = native_mask;
		mask.ii6a_high = (~ (uint64) 0);	/* All 1's */
	    } else {
		native_mask = (~ (uint64) 0) << (64 - numbits);

		mask.ii6a_low = 0;
		mask.ii6a_high = native_mask;
	    }

	    address->ihle6_nbits = numbits;
	    address->ihle6_first.ii6a_low &= mask.ii6a_low;
	    address->ihle6_first.ii6a_high &= mask.ii6a_high;
	    address->ihle6_last.ii6a_low |= ~mask.ii6a_low;
	    address->ihle6_last.ii6a_high |= ~mask.ii6a_high;

	    return TRUE;
	}
    default:
	return FALSE;
    }
} /* IHL_Mask_Set() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Create
 *  Purpose	: Allocate an IHL handle
 *  Method	: PA_Create()
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Returns	: The newly created IHL_Handle_p
 *------------------------------------------------------------------------
 */
IHL_Handle_p
IHL_Create(void)
{
    static const char	fname[] = "IHL_Create";
    IHL_Handle_p	ihlh = malloc(sizeof(*ihlh));

    if (ihlh == NULL) {
	ERWrite(IHLE_NOMEM, fname, "Could not allocate host list");
	return NULL;
    }
    ihlh->ihl_list = PA_Create("size=4", IHL_NewOld_Entry);
    if (ihlh->ihl_list == NULL) {
	ERWrite(PAerrno, fname, "Could not allocate host list");
	free(ihlh);
	return NULL;
    }
    return ihlh;
} /* IHL_Create() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Add
 *  Purpose	: Create/Add-to hostlist of an IP-range
 *  Method	:
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *
 *  Notes	:
 *  The address list may contain single IPv4 addresses, ranges
 *  ("a.b.c.d-p.q.r.s") and CIDR address blocks ("a.b.c.d/n"). Values are
 *  separated by commas/spaces.
 *  The IPv6 support allows the same possibilities.
 *
 *  Description	:
 *  This function will add a hostrange to a hostlist. If hostlist is NULL, it
 *  will be created.
 *
 *  The hostrange can contain multiple ranges (separated by space or comma),
 *  but that means some errors may not be reported.
 *
 *  The function will return the new hostlist, NULL will be returned if there
 *  was an error and the hostlist argument was NULL.
 *
 *  If the number of elements in the hostlist wasn't changed by a call to
 *  IHL_Add, there was an error.
 *
 *  Returns	: The list containing the IP-ranges
 *------------------------------------------------------------------------
 */
IHL_Handle_p
IHL_Add(
	IHL_Handle_p	hostlist_arg,
	const char	*liststring,
	cvoidp		userdata)
{
    static const char *const fname = "IHL_Add";
    IHL_Handle_p	ihlh = hostlist_arg;
    const char		*orig_liststring = liststring;
    bool		ok = TRUE;

    if (ihlh == NULL) {
	ihlh = IHL_Create();
	if (ihlh == NULL) {
	    ERWrite(IHLE_NOMEM, fname, "Could not allocate host list");
	    return NULL;
	}
    }

    while (ok) {
	IHL_Entry_t	address;

	memset(&address, '\0', sizeof(address));

	while (PM_is_ascii_space((unsigned char)*liststring) || *liststring == ',') liststring++;
	if (*liststring == '\0') break;

	if (IHL_Entry_Fill(&liststring, &address, userdata) == FALSE) {
	    ERWrite(APPLE_PARAM, fname, "Invalid IP address in host list (%s)",
					orig_liststring);
	    ok = FALSE;
	    break;
	}
	if (*liststring == '/') {
	    char		*endptr;
	    long int		bits;

	    liststring++;
	    bits = strtol(liststring, &endptr, 10);
	    if ((endptr == liststring)
		|| (! IHL_Mask_Set(&address, bits))) {
		ERWrite(APPLE_PARAM, fname, "Invalid CIDR notation in host list");
		ok = FALSE;
		break;
	    }
	    liststring = endptr;
	} else if (*liststring == '-') {
	    IHL_Entry_t	last;
	    liststring++;
	    if (IHL_Entry_Fill(&liststring, &last, NULL) == FALSE) {
		ERWrite(APPLE_PARAM, fname, "Invalid IP range in host list");
		ok = FALSE;
		break;
	    }
	    if (last.ihle_type != address.ihle_type) {
		ERWrite(APPLE_PARAM, fname, "Mixed IP versions in range");
		ok = FALSE;
		break;
	    }
	    switch (last.ihle_type) {
	    case IHL_TYPE_IPv4:
		if (address.ihle4_first > last.ihle4_first) {
		    ERWrite(APPLE_PARAM, fname, "Invalid IP range in host list");
		    ok = FALSE;
		} else {
		    address.ihle_rangetype = IHL_RANGETYPE_RANGE;
		    address.ihle4_last = last.ihle4_first;
		}
		break;
	    case IHL_TYPE_IPv6:
		if (IHL_IPv6_Address_Cmp(address.ihle6_first, last.ihle6_first) > 0) {
		    ERWrite(APPLE_PARAM, fname, "Invalid IP range in host list");
		    ok = FALSE;
		} else {
		    address.ihle_rangetype = IHL_RANGETYPE_RANGE;
		    address.ihle6_last = last.ihle6_first;
		}
		break;
	    }
	}

	if (ok && ! PA_Add(ihlh->ihl_list, &address)) {
	    ERWrite(APPLE_PARAM, fname, "Could not add IP range to host list");
	    ok = FALSE;
	    break;
	}
    }

    if (! ok && hostlist_arg == NULL) {
	IHL_Destroy(ihlh);
	ihlh = NULL;
    }
    return ihlh;
} /* IHL_Add() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Destroy
 *  Purpose	: Free a hostlist and all its elements
 *  Method	: Warpper around PA_Destroy().
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *------------------------------------------------------------------------
 */
void
IHL_Destroy(
	IHL_Handle_p	hostlist)
{
    if (NULL == hostlist){
	return;
    }
    PA_Destroy(hostlist->ihl_list);
    free(hostlist);
} /* IHL_Destroy() */

/*------------------------------------------------------------------------
 *  Function	: IHL_CheckAddressInRange
 *  Purpose	: Test if an address fits in range
 *  Method	:
 *
 *  History:
 *  13-02-2013	: sieburgw : creation
 *
 *  Returns	: Whether the address fits in the supplied range
 *------------------------------------------------------------------------
 */
bool
IHL_CheckAddressInRange(
	IHL_Entry_p range,
	IHL_Entry_p address)
{
    IHL_IPv4_Address_t	ipv4_address = 0;
    IHL_IPv6_Address_t	ipv6_address = {0, 0};

    if (!range | ! address){
	return FALSE;
    }
    switch (address->ihle_type) {
    case IHL_TYPE_IPv4:
	ipv4_address  = address->ihle4_first;
	break;
    case IHL_TYPE_IPv6:
	ipv6_address  = address->ihle6_first;
	break;
    default:
	return FALSE;
    }

    if (address->ihle_type == range->ihle_type) {
	switch (address->ihle_type) {
	case IHL_TYPE_IPv4:
	    if ((ipv4_address >= range->ihle4_first)
		&& (ipv4_address <= range->ihle4_last)) {
		return TRUE;
	    }
	    break;
	case IHL_TYPE_IPv6:
	    if ((IHL_IPv6_Address_Cmp(ipv6_address, range->ihle6_first) >= 0)
		&& (IHL_IPv6_Address_Cmp(ipv6_address, range->ihle6_last) <= 0)) {
		return TRUE;
	    }
	    break;
	}
    }
    return FALSE;
} /* IHL_CheckAddressInRange() */

/*------------------------------------------------------------------------
 *  Function	: IHL_CheckAddressInRangeList
 *  Purpose	: Check if an address occurs in the list of ranges
 *  Method	: Call IHL_CheckAddressInRange() for each hostlist entry
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *
 *  Notes	: This function has some preliminary IPv6 support.
 *
 *  Returns	: Pointer to the address range or NULL
 *------------------------------------------------------------------------
 */
static IHL_Entry_p
IHL_CheckAddressInRangeList(
	IHL_Handle_p	ihl,
	IHL_Entry_p	address)
{
    int			i;
    PAHandle_p		hostlist = ihl->ihl_list;
    int			max = PA_GetNrEntries(hostlist);

    for (i = 0; i < max; i++) {
	IHL_Entry_p	address_range = (IHL_Entry_p) PA_GetN(hostlist, i);

	if (IHL_CheckAddressInRange(address_range, address)){
	    return address_range;
	}
    }
    return NULL;
} /* IHL_CheckAddressInRangeList() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Check
 *  Purpose	: Check if an address occurs in a range of addresses
 *  Method	:
 *  Parse the host with IHL_Entry_Fill() and check it against the list with
 *  IHL_CheckAddressInRangeList().
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *
 *  Description	:
 *  This function checks if host is on the list. The requested host must be
 *  a dotted IP-address (IPv4), not a DNS name.
 *
 *  Returns	: Pointer to the address range or NULL
 *------------------------------------------------------------------------
 */
IHL_Entry_p
IHL_Check(
	IHL_Handle_p	hostlist,
	const char	*host)
{
    IHL_Entry_t address;

    if (hostlist == NULL) return NULL;

    if (IHL_Entry_Fill(&host, &address, NULL) == FALSE) return NULL;

    return IHL_CheckAddressInRangeList(hostlist, &address);
} /* IHL_Check() */

/*------------------------------------------------------------------------
 *  Function	: IHL_IsValidAddress
 *  Purpose	: Check if an address is a valid IP address
 *  Method	: IHL_Entry_Fill()
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *
 *  Returns	: Whether or not the address is valid
 *------------------------------------------------------------------------
 */
bool
IHL_IsValidAddress(
	const char	*host)
{
    IHL_Entry_t address;

    return (IHL_Entry_Fill(&host, &address, NULL) && (*host == '\0'));
} /* IHL_IsValidAddress() */

/*------------------------------------------------------------------------
 *  Function	: IHL_IPv4_AddressToString
 *  Purpose	: Convert an IHL_IPv4_Address_t to string
 *  Method	: Obvious
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	: The stbuf is to make it easier to call from a debugger.
 *
 *  Returns	: buffer with the string representation
 *------------------------------------------------------------------------
 */
static const char *
IHL_IPv4_AddressToString(
	char			*buf,		// OUT
	IHL_IPv4_Address_t	ipv4_address)	// IN
{
    byte		octet[4];
    int			i;
    static char		stbuf[IHL_ADDRBUF_SIZE];

    if (buf == NULL) buf = stbuf;

    for (i = 3; i >= 0; --i) {
	octet[i] = (byte) ipv4_address & 0xff;
	ipv4_address >>= 8;
    }
    sprintf(buf, "%d.%d.%d.%d", octet[0], octet[1], octet[2], octet[3]);

    return buf;
} /* IHL_IPv4_AddressToString() */

/*------------------------------------------------------------------------
 *  Function	: IHL_IPv6_AddressToString
 *  Purpose	: Convert an IHL_IPv6_Address_t to string
 *  Method	:
 *  1. Convert the high and low part of the IPv6 address into an array of words
 *  2. Determine the longest seqence of zero's
 *  3. Output the address, reducing a sequence of zero's into "::"
 *
 *  History:
 *  21-06-2011	: eck      : creation
 *
 *  Notes	: The stbuf is to make it easier to call from a debugger.
 *
 *  Returns	: buffer with the string representation
 *------------------------------------------------------------------------
 */
static const char *
IHL_IPv6_AddressToString(
	char			*buf,		// OUT
	IHL_IPv6_Address_t	ipv6_address)	// IN
{
    unsigned int	word[IPv6_NUM_WORDS];
    int			i;
    uint64		high = ipv6_address.ii6a_high;
    uint64		low = ipv6_address.ii6a_low;
    static char		stbuf[IHL_ADDRBUF_SIZE];
    char		*ptr;
    int			zero_count = 0;
    int			max_zero_start = -1;
    int			max_zero_count = -1;
    int			max_zero_count_end = -1;

    if (buf == NULL) buf = stbuf;
    ptr = buf;

    /* Fill word starting at the least significant parts of low and high. */
    for (i = (sizeof(high)/sizeof(uint16)) - 1 ; i >= 0 ; --i) {
	word[i] = (uint16) high & 0xffff;
	word[i + (IPv6_NUM_WORDS / 2)] = (uint16) low & 0xffff;
	high >>= 16;
	low >>= 16;
    }

    /* Determine the end of the longest sequence of zero's. */
    for (i = 0; i < IPv6_NUM_WORDS; i++) {
	if (word[i] == 0) {
	    zero_count++;
	    if (zero_count > max_zero_count) {
		max_zero_count = zero_count;
		max_zero_count_end = i;
	    }
	} else {
	    zero_count = 0;
	}
    }

    if (max_zero_count > 0)  {
	/* If we saw 1 or more zero's, determine the start position.
	 * Only there is only 1 zero at start or end, the replacement by "::"
	 * saves no space, so we don't do it.
	 * This also avoids issues with IPv6 address parsers.
	 */
	if (max_zero_count > 1) {
	    max_zero_start = max_zero_count_end - max_zero_count + 1;
	} else if ((max_zero_count_end != 0)
		    && (max_zero_count_end != (IPv6_NUM_WORDS - 1))) {
	    max_zero_start = max_zero_count_end - max_zero_count + 1;
	}
    }

    /* A bit of a weird loop to accomodate the :: handling. */
    for (i = 0; i < IPv6_NUM_WORDS; i++) {
	if ((i < max_zero_start) || (i > max_zero_count_end)) {
	    sprintf(ptr, "%x", word[i]);
	    ptr += strlen(ptr);
	    if (i < (IPv6_NUM_WORDS -1)) {
		*ptr++ = ':';
	    }
	} else if (i == max_zero_start) {
	    if (i == 0) *ptr++ = ':';
	    *ptr++ = ':';
	}
    }
    *ptr = '\0';

    return buf;
} /* IHL_IPv6_AddressToString() */

/*------------------------------------------------------------------------
 *  Function	: IHL_EntryToString
 *  Purpose	: Convert a host address to string
 *  Method	:
 *
 *  History:
 *  16-03-2011	: eck      : creation
 *
 *  Notes	:
 *  - If buf argument is NULL, a static buffer is used.
 *  - Assumes IPv4 (for now).
 *  - Does not yet handle ranges and CIDR specification properly.
 *
 *  Description	:
 *  - This function prints an IHL entry (IP-address range) to a buffer.
 *  - The buffer must be at least IHL_ADDRBUF_SIZE bytes long.
 *  - If the buf parameter is NULL, a single static buffer is used and
 *    returned.
 *
 *  Returns	: Pointer to output buffer
 *------------------------------------------------------------------------
 */
const char *
IHL_EntryToString(
	char		*buf,		// OUT
	IHL_Entry_p	address)	// IN
{
    static char		stbuf[IHL_ADDRBUF_SIZE];
    char		*rest;

    if (buf == NULL) buf = stbuf;
    switch (address->ihle_type) {
    case IHL_TYPE_IPv4:
	IHL_IPv4_AddressToString(buf, address->ihle4_first);
	break;
    case IHL_TYPE_IPv6:
	IHL_IPv6_AddressToString(buf, address->ihle6_first);
	break;
    }

    rest = buf + strlen(buf);

    switch (address->ihle_rangetype) {
    case IHL_RANGETYPE_SINGLE:
	break;
    case IHL_RANGETYPE_CIDR:
	switch (address->ihle_type) {
	case IHL_TYPE_IPv4:
	    sprintf(rest, "/%d", address->ihle4_nbits);
	    break;
	case IHL_TYPE_IPv6:
	    sprintf(rest, "/%d", address->ihle6_nbits);
	    break;
	}
	break;
    case IHL_RANGETYPE_RANGE:
	*rest++ = '-';
	switch (address->ihle_type) {
	case IHL_TYPE_IPv4:
	    IHL_IPv4_AddressToString(rest, address->ihle4_last);
	    break;
	case IHL_TYPE_IPv6:
	    IHL_IPv6_AddressToString(rest, address->ihle6_last);
	    break;
	}
    }
    return buf;
} /* IHL_EntryToString() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Walk
 *  Purpose	: Perform an action on all elements in a host list.
 *  Method	: Call the callback function for each hostlist entry.
 *
 *  History:
 *  27-04-2011	: eck      : creation
 *
 *  Description	:
 *  This function walks down a hostlist and calls a callback for each element.
 *  The order in which this happens is not guaranteed.
 *  If the callback returns FALSE, IHL_Walk() stops.
 *
 *  Returns	: FALSE if any callback returned FALSE, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
IHL_Walk(
	IHL_Handle_p	ihl,
	bool		(*HandleElt)(IHL_Entry_p entry, voidp xtraparam),
	voidp		userdata)
{
    PAHandle_p		hostlist;
    int			max;
    int			i;

    if (ihl == NULL) return TRUE;

    hostlist = ihl->ihl_list;
    max = PA_GetNrEntries(hostlist);
    for (i = 0; i < max; i++) {
	if (! HandleElt((IHL_Entry_p) PA_GetN(hostlist, i), userdata)) return FALSE;
    }
    return TRUE;
} /* IHL_Walk() */

/*------------------------------------------------------------------------
 *  Function	: IHL_Userdata_Get
 *  Purpose	: Retrieve the userdata value from an entry
 *  Method	:
 *
 *  History:
 *  15-07-2011	: heesakkm : creation
 *
 *  Returns	: Pointer to userdata
 *------------------------------------------------------------------------
 */
cvoidp
IHL_Userdata_Get(
	IHL_Entry_p	ihe)
{
    return (ihe != NULL) ? ihe->ihle_userdata : NULL;
} /* IHL_Userdata_Get() */
