/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File        : ccv.h
 *  Description : Character ConVersion interface
 *  Author(s)   : Hans van Eck
 *  Creation    : 17-02-1994
 *
 *  History:
 *  17-02-1994  : eck      : creation
 *  09-10-1997  : BdB      : Added CCV2 routines
 *  26-04-2001  : paul     : CCVDiagnostic added
 *  24-01-2007  : paul     : Added CCV2ConvertCS() and CCV2ConvertM8()
 *  26-01-2007  : paul     : Added CCV2ConvertM8Append()
 *  08-10-2013	: martinz  : Copied documentation from CCVLIB.DOC.
 *
 *========================================================================
 */

#if ! defined(CCV_H_INCLUDED)
#define CCV_H_INCLUDED  1

#pragma ident "$Id: ccv.h 305171 2013-11-01 13:41:52Z sieburgw $"

#include <pica.h>
#include <cs.h>

/*
 * The CCVErrno_t is not necessary the moment, but we might have a
 * preprocessing phase later (for faster matching), or a compilation phase
 * for dynamic loading of tables.
 */
ERROR_TYPE(CCVErrno_t);

ERROR_DECLARE(CCVE_NOERROR);	/* Ok */
ERROR_DECLARE(CCVE_NOMEM);      /* Out of memory. */
ERROR_DECLARE(CCVE_FORMAT);     /* Syntax error in conversion table. */
ERROR_DECLARE(CCVE_LIMIT);      /* Hit arbitrary limit in routines. */
ERROR_DECLARE(CCVE_PARAM);      /* Illegal parameter. */

extern CCVErrno_t CCVerrno;

typedef struct {
    const char  *ce_intern;
    const char  *ce_extern;
} CCVElement_t, *CCVElement_p;

typedef enum {
    CCV_TO_INTERN, CCV_TO_EXTERN
} CCVDirection_t;

#define CCV_DIAG_SIZE           500
extern char     CCVDiagnostic[CCV_DIAG_SIZE];

const char *CCV(CCVElement_t cv_table[], const char *text_in, int
                text_in_len, int *text_in_used, CCVDirection_t dir);

const char *CCVIn(CCVElement_t cv_table[], const char *text_in,
                    int text_in_len, int *text_in_used);

const char *CCVOut(CCVElement_t cv_table[], const char *text_in,
                    int text_in_len, int *text_in_used);

#define CCVIn(tab, t_in, t_inlen, t_in_used)    \
                        CCV(tab, t_in, t_inlen, t_in_used, CCV_TO_INTERN)

#define CCVOut(tab, t_in, t_inlen, t_in_used)   \
                        CCV(tab, t_in, t_inlen, t_in_used, CCV_TO_EXTERN)

bool CCVCompile(const char *name, const char *definition);

void CCVDestroy(const char *name);

CCVElement_p CCVGetTable(const char *name);


/**
 *  Notes on CCV2: The following quantities have arbitrary limits due to limits
 *  of data types used in implementation:
 *  - Input and output strings: UCHAR_MAX (usually 255) characters each
 *  - Length of all input strings together: maximum depends on the contents but
 *    is at least USHRT_MAX/2 (usually 32767).
 *
 *  NULLs in input and output strings are allowed.
 */

/* Handle to compiled conversion table, opaque. */
typedef const struct CCV2_handle_t *CCV2Handle_p;

/* Compile a table of conversion items to a CCV2Handle_p. Return NULL on error.
   Convert from first strings to second strings unless 'reverse'. */
CCV2Handle_p
CCV2Compile(const char *table, bool reverse);

/* Destroy a handle returned by CCV2Compile. */
void
CCV2Destroy(CCV2Handle_p handle);

/* Convert a string according to a previously compiled table. Returns number
   of characters written into outstr. Puts number of input characters used in
   *iunused. Cannot return any errors. */
int
CCV2Convert(CCV2Handle_p handle,
            const char *instr, int inlen, int *inused,
            char *outstr, int outlen);

CSBuf_p
CCV2ConvertCS(CCV2Handle_p handle, CSBuf_p outstr, const char *instr,
	      long inlen, const char *options);

CSBuf_p
CCV2ConvertM8(CCV2Handle_p handleG0, bool extendedG0,
	      CCV2Handle_p handleG1, bool extendedG1,
	      CSBuf_p outstr, const char *instr, long inlen,
	      const char *options);

bool
CCV2ConvertM8Append(CCV2Handle_p handleG0, bool	extendedG0,
		    CCV2Handle_p handleG1, bool	extendedG1,
		    CSBuf_p outstr, const char *instr, long inlen,
		    const char *options);

/* Associate a name with a CCV2Handle_p. */
bool
CCV2Name(const char *name, CCV2Handle_p handle);

/* Get the CCV2Handle_p associated with a name. */
CCV2Handle_p
CCV2GetByName(const char *name);

#endif /* ! defined(CCV_H_INCLUDED) */
