/*
 *  Copyright (c) 1998 by Pica
 *
 *  File	: cookie.h
 *  Description	: Interface for cookie server operations
 *  Author(s)	: Bas de Bakker
 *  Creation	: 30-01-1998
 *
 *  History:
 *  30-01-1998	: BdB	: creation
 *  20-03-1998	: BdB	: Added formatting (CookieAttr*) routines
 *  27-03-1998	: BdB	: Renamed Cookie* to CK*
 *  04-08-1998	: LB 	: Added CKRelease
 *  04-08-1998	: BdB	: Change internal defs to accomodate CKRelease
 *  03-09-1999	: BdB	: Move internal defs to ckinternals.h
 *  05-04-2006	: paul  : removed CKDelOldest()
 *  02-05-2011	: martinz  : Added CKServerOpen2() and func typedefs.
 *  21-10-2016	: zwaalm   : CBS-6957 Added session context options.
 *
 *  Notes	:
 *
 */

#ifndef COOKIE_H_INCLUDED
#define COOKIE_H_INCLUDED

#pragma ident "$Id: cookie.h 356505 2018-05-31 13:37:08Z smitp $"


/* Various statuses returned by CK*() routines. */
typedef enum {
    COOKIE_IMPOSSIBLE,		/* Shouldn't be used. */
    COOKIE_PRESENT,		/* Attempted creation of existing cookie. */
    COOKIE_NOSUCH,		/* Validated/destroyed non-existing cookie. */
    COOKIE_NETWORK,		/* A network error occurred. */
    COOKIE_SERVER,		/* An error occurred in the server. */
    COOKIE_NOMEM,		/* An out of memory error occurred. */
    COOKIE_INTERNAL,		/* An internal error. */
    COOKIE_PARAM,		/* Illegal parameters to function. */
    COOKIE_OK			/* Everything is ok. */
} CKReturn_t;

/* The following are the cookie fields as currently understood. They properly
 * belong to the hermes tree, but there is no logical place to put them (one
 * could be added at the expense of som refactoring).
 * As this is a hermes specific library anyway, I (HvE) have just put it here.
 *
 * Field   Meaning         Hermes variable
 * 'U'     user number     USERNUM        
 * 'K'     user_key                       
 * 'D'     database-id     DB?            
 * 'E'     session-id      SID            
 * 'I'     iln             ILN            
 * 'C'     channel                        
 * 'B'     library         LIBID          
 * 'S'     accounting      ACCOUNT        
 * 'A'     adimax                         
 * 'H'     authority                      
 * 'N'     user_name       USERNAME       
 * 'P'     nickname        NICKNAME       
 * 'T'     SNS-banned?     ABUSED         
 * 'L'     enduser key     USERKEY        
 * 'R'     remote address                 
 * 'F'     Using defaultlogin?            
 * 'X'     auth_method     RAPRF          
 * 'Z'     CBS session id  SESID          
 * 'G'     usertype        USERTYPE       
 * 'Q'     filter_class    FILTER_CLASS   
 * 'J'     context options
 */
#define CKF_USER_NUMBER		'U'
#define CKF_USER_KEY		'K'
#define CKF_DATABASE_ID		'D'
#define CKF_SESSION_ID		'E'
#define CKF_ILN			'I'
#define CKF_CHANNEL		'C'
#define CKF_LIBRARYID		'B'
#define CKF_DOACCOUNT		'S'
#define CKF_ADIMAX		'A'
#define CKF_AUTHORITY		'H'
#define CKF_USER_NAME		'N'
#define CKF_NICK_NAME		'P'
#define CKF_SNS_BANNED		'T'
#define CKF_ENDUSER_KEY		'L'
#define CKF_REMOTE_ADDRESS	'R'
#define CKF_USING_DEFAULTLOGIN	'F'
#define CKF_RAPRF		'X'	// Remote authentication method key
#define CKF_CBS_SESSION_ID	'Z'
#define CKF_USERTYPE		'G'
#define CKF_FILTER_CLASS	'Q'
#define CKF_CONTEXT_OPTIONS	'J'	// Central System context options

/* A handle for cookie servers. */
typedef struct CKServer_t *CKServer_p;

typedef int			/* file descriptor to opened server */
	(*server_open_f)(	/* callback function to open server */
	voidp userdata);	/* user data */

typedef bool			/* indication of success */
	(*server_close_f)(	/* callback function to close server */
	voidp userdata);	/* user data */

/* Open a connection to a cookie server, specified in 'options'.  Puts handle
   to connection in '*server' (if return value == COOKIE_OK). */
CKReturn_t
CKServerOpen2(
	const char 	*options,
	CKServer_p 	*server,
	server_open_f 	server_open,
	server_close_f 	server_close,
	voidp 		server_userdata);

CKReturn_t
CKServerOpen(const char *options, CKServer_p *server);

/* Close a connection to a cookie server and free memory associated with the
   server handle. */
CKReturn_t
CKServerClose(CKServer_p server);

/* Check if a cookie is present.  If it is, update its use time. */
CKReturn_t
CKValidate(CKServer_p server, const char *cookie);

/* Create a new cookie on a cookie server. */
CKReturn_t
CKCreate(CKServer_p server, const char *cookie);

/* Unconditionally destroy a cookie on a server. */
CKReturn_t
CKDestroy(CKServer_p server, const char *cookie);

/* Delete cookies that have attributes equal to 'values', until the number left
   is less than or equal to 'total'.  Return COOKIE_OK if remaining cookies
   with attributes equal to 'values' is less than or equal to 'total'.  Remove
   only cookies older than 'seconds' */
CKReturn_t
CKRelease(CKServer_p server, const char *values, int total, 
	   unsigned int seconds);

/* Add a name/value attribute pair to a cookie.  This will create duplicate
   attributes if the name is already present, so if you're not building a
   cookie from scratch, make sure to first delete the name from the cookie.
   Returns FALSE on overflow. */
bool
CKAttrAdd(char *cookie, size_t cookiesize,
	      char name, const char *value);

/* Delete a name/value attribute pair from a cookie.  Returns FALSE if name
   wasn't present. */
bool
CKAttrDel(char *cookie, size_t cookiesize, char name);

/* Retrieve a value from a cookie attribute.  Returns FALSE if name wasn't
   present or value overflowed.  In the first case, value will be the empty
   string, in the second it will contain the start of the cookie. */
bool
CKAttrGet(const char *cookie, char name, char *value, size_t valuesize);

/* See if cookie attributes 'attr' are all present in 'cookie', return TRUE if
   so, FALSE if not.  Note that an empty 'attr' string will match all
   cookies. */
bool
CKAttrMatch(const char *cookie, const char *attr);

/* Retrieve the cookies that match the provided attributes. When successful, it
   returns COOKIE_OK, puts the number of cookies found in ncookies and puts the
   matching cookies in cookies as a '\0' separated string. Otherwise it will
   return an error value. */
CKReturn_t
CKSelect(CKServer_p server, const char *attr, int *ncookies,
	 const char **cookies);

#endif /* COOKIE_H_INCLUDED */
