/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: cs.h
 *  Description	: Counted String library routines
 *  Author(s)	: Hans van Eck
 *  Creation	: 01-10-1993
 *
 *  History:
 *  01-10-1993	: eck      : creation
 *  15-11-1993	: eck      : added dynamic strings
 *  17-11-1993	: eck      : added CSAppendC()
 *  19-05-1994	: eck      : added CSVPrintf()
 *  07-04-1995	: eck      : Removed initialisation of the buffer of locals
 *			   : It is now done in CS_LOCALINIT().
 *  05-06-1996	: eck      : Removed function declarations of CSLength(),
 *			   : CSSize() and CSContents().
 *  06-04-1998	: BdB	   : Added CSAppendN().
 *  25-08-1998	: BdB	   : CSContents never returns NULL
 *  13-06-2002	: sutherlr : added fully-dynamic strings: macro CS_IS_DYNAMIC(),
 *                         : function CSCreate(), and type CSBuf_p
 *  21-03-2003	: eck      : added CSSetStringN()
 *  27-03-2003	: paul     : added CSInsertN() and CSSetSize()
 *  08-04-2003	: paul     : Added CSSetLength()
 *  15-10-2003	: paul     : Added cs_chunk_size member to CSBuf_t
 *                         : Moved CS_DEFAULT_CHUNK_SIZE from cs.c
 *                         : Added CSCreateOpt()
 *                         : Added CSChunkSize() and CSSetChunkSize()
 *  22-02-2008	: sieburgw : Added CSPurge()
 *  05-03-2008	: sieburgw : Renamed CSPurge to CSTearOff
 *  18-07-2011	: paul     : CBS-3077 Made CSMAX_SPRINTF_LEN publicly available
 *  19-07-2012	: martinz  : Added CSCreateStringNOpt, CSCreateStringOpt(),
 *  			     CSCreateStringN() and CSCreateString().
 *  01-10-2013	: martinz  : Added CSE_PARAM.
 *
 *  Notes	:
 *  - The CSBuf_t type may be used only by the CS-library.
 *  - Variables must be defined with CS_GLOBAL(name, size), CS_LOCAL(name, size),
 *    CS_PARAM(name), CS_DYNAMIC(name).
 *  - CSBuf_p (initialized with CSCreate()) can be used to create a
 *    fully-dynamic counted string.
 *========================================================================
 */

#if ! defined(CS_H_INCLUDED)
#define CS_H_INCLUDED

#pragma ident "$Id: cs.h 368646 2019-07-16 16:25:21Z vaneckh $"

#include <limits.h>
#include <stddef.h>
#include <stdarg.h>
#include <pica.h>

#define CSMAX_SPRINTF_LEN	2048

#define CS_DEFAULT_CHUNK_SIZE	128	/* Allocate in 128 bytes chunks */

enum CSType_t { CST_GLOBAL, CST_LOCAL, CST_DYNAMIC, CST_FULLY_DYNAMIC };

/**
 * CSBuf_t contains the size and length of a counted string, and a pointer
 * to the buffer with the contents.
 * This type may only be used directly by the CS-library.
 */
typedef struct CSBuf_t {
    size_t		cs_size;
    size_t		cs_length;
    size_t		cs_chunk_size;
    enum CSType_t	cs_type;
    const char		*cs_bufp;
} CSBuf_t, *CSBuf_p;

/** Declaration of CS-library error variable. */
ERROR_TYPE(CSErrno_t);

ERROR_DECLARE(CSE_NOERROR);
ERROR_DECLARE(CSE_NOMEM);
ERROR_DECLARE(CSE_LENGTH);
ERROR_DECLARE(CSE_OVFL);
ERROR_DECLARE(CSE_SPRINTF);
ERROR_DECLARE(CSE_PARAM);

extern CSErrno_t CSerrno;

/*
 * The global declaration doesn't need an initializer for the buffer since
 * the data is guaranteed to be 0.
 * A local buffer, however, is not guaranteed, so we put an empty string in.
 */

#define CS_GLOBAL_X(buf, siz, id)	\
	    static char CSBUF_ ## id [ siz + 1 ] ; \
	    CSBuf_t CSVAR_ ## id = { siz + 1, 1, 0, \
				     CST_GLOBAL, CSBUF_ ## id }; \
	    CSBuf_t *buf = & CSVAR_ ## id 

#define CS_GLOBAL_ID(buf, siz, id)	CS_GLOBAL_X(buf, siz, id)
/** CS_GLOBAL() defines a global CS variable with name and size. */
#define CS_GLOBAL(buf, siz)		CS_GLOBAL_ID(buf, siz, __LINE__)

#define CS_LOCAL_X(buf, siz, id)   \
	    char CSBUF_ ## id [ siz + 1 ] ; \
	    CSBuf_t CSVAR_ ## id = CS_LOCALINIT(siz + 1, CSBUF_ ## id) ; \
	    CSBuf_t *buf = & CSVAR_ ## id 

#define CS_LOCAL_ID(buf, siz, id)	CS_LOCAL_X(buf, siz, id)
/** CS_LOCAL() defines a local CS variable with name and size. */
#define CS_LOCAL(buf, siz)		CS_LOCAL_ID(buf, siz, __LINE__)

#define CS_DYNAMIC_X(buf, id, csize)	\
	    CSBuf_t CSVAR_ ## id = { 0, 0, csize, CST_DYNAMIC, NULL }; \
	    CSBuf_t *buf = & CSVAR_ ## id

#define CS_DYNAMIC_ID(buf, id, csize)	CS_DYNAMIC_X(buf, id, csize)
#define CS_DYNAMIC_CHUNK(buf, csize)	CS_DYNAMIC_ID(buf, __LINE__, csize)
/**
 * CS_DYNAMIC() defines a local or global CS with dynamic size.
 * A CS defined this way must be CSDestroy()-ed before going out of scope.
 * A dynamic CS is automatically intialized.
 */
#define CS_DYNAMIC(buf) \
	CS_DYNAMIC_CHUNK(buf, CS_DEFAULT_CHUNK_SIZE)

/** CS_PARAM() defines a CS parameter with an implicit size. */
#define CS_PARAM(buf)	CSBuf_p buf

#define CS_IS_DYNAMIC(strp)	((strp->cs_type == CST_DYNAMIC) || \
				    (strp->cs_type == CST_FULLY_DYNAMIC))

CSBuf_t CS_LOCALINIT(size_t size, char *buf);

CSBuf_p CSCreate(void);
CSBuf_p CSCreateOpt(const char *options);

bool CSClear(CSBuf_p strp);

bool CSDestroy(CSBuf_p strp);

bool CSSetString(CSBuf_p strp, const char *string);

bool CSSetStringN(CSBuf_p strp, const char *string, int len);

void CSSetChunkSize(CSBuf_p strp, size_t newsize);

bool CSSetSize(CSBuf_p strp, size_t newsize);

bool CSSetLength(CSBuf_p strp, size_t newlength);

bool CSAppend(CSBuf_p strp, const char *string);

bool CSAppendC(CSBuf_p strp, char ch);

bool CSAppendN(CSBuf_p strp, const char *string, size_t length);

bool CSDeletePart(CSBuf_p strp, size_t start, size_t length);

bool CSInsert(CSBuf_p strp, size_t start, const char *str);

bool CSInsertN(CSBuf_p strp, size_t start,const char *str,int len);

extern char *CSTearOff(CSBuf_p cs);

/*PRINTFLIKE2*/
bool CSPrintf(CSBuf_p strp, const char *fmt, ...) __attribute__((format(printf, 2, 3)));
bool CSVPrintf(CSBuf_p strp, const char *fmt, va_list argp) __attribute__((format(printf, 2, 0)));

/*PRINTFLIKE2*/
bool CSPrintfAppend(CSBuf_p strp, const char *fmt, ...) __attribute__((format(printf, 2, 3)));
bool CSVPrintfAppend(CSBuf_p strp, const char *fmt, va_list argp) __attribute__((format(printf, 2, 0)));

/* The following macros should improve CS-performance. The tests are for
 * dynamic CS's
 */

/**
 *  CSSize() returns the size of the counted string that can be used.
 *
 *  The size is equal to 'cs_size' field minus 1: a '\0'-byte is not included.
 *  A (fully-)dynamic CS will have UINT_MAX as maximum size.
 *
 *  Notes	: strp must be a valid counted string.
 *
 *  Error values: none
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	len = CSSize(string);
 *
 *  Returns	: The maximum size of a counted string
 */
#define CSSize(strp)		(CS_IS_DYNAMIC(strp) \
				    ? UINT_MAX : ((strp)->cs_size - 1))

/**
 *  CSLength() returns the current length of a counted string.
 *
 *  The length is equal to 'cs_length' field minus 1: the terminating '\0'-byte
 *  is not included. Since a counted string may contain a '\0'-byte, CSLength()
 *  is not necessarily equal to a strlen() of the contents.
 *
 *  Notes	: strp must be a valid counted string.
 *
 *  Error values: none
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	CS_PARAM(string);
 *
 *	len = CSLength(string);
 *
 *  Returns	: The length of the counted string that is used or 0
 *
 *  @param strp		the counted string
 */
#define CSLength(strp)		(((strp)->cs_bufp == NULL) \
				    ? 0 : ((strp)->cs_length - 1))

#define CSChunkSize(strp)	((strp)->cs_chunk_size)

extern const char *const CSEmptyString;
/**
 *  CSContents() returns a pointer to the contents of a counted string.
 *
 *  - There will always be a '\0' byte at the end of the buffer.
 *  - The contents may not be changed, hence the 'const' in the return-type.
 *  - The string can be used for printing, comparison, etc.
 *  - The standard C-library functions can be used with the result, except, of
 *    course, for functions that change the contents.
 *
 *  Notes	:
 *  - strp must be a valid counted string.
 *  - If strp is a (fully-)dynamic CS, the pointer returned by CSContents() may
 *    become obsolete after changes to the CS.
 *
 *  Error values: none
 *
 *  Example	:
 *
 *	#include <cs.h>
 *
 *	char      *tmp;
 *	CS_PARAM( string);
 *
 *	tmp = CSContents(string);
 *
 *  Returns	: Pointer to buffer of the counted string, never NULL.
 *
 *  @param strp		the counted string
 */
#define CSContents(strp)	((strp)->cs_bufp == NULL \
				 ? CSEmptyString : (strp)->cs_bufp)

CSBuf_p CSCreateStringNOpt(const char *string, int len, const char *opt);

#define CSCreateStringOpt(str,opt)	(CSCreateStringNOpt((str),-1,(opt)))
#define CSCreateStringN(str,len)	(CSCreateStringNOpt((str),(len),NULL))
#define CSCreateString(str)		(CSCreateStringNOpt((str),-1,NULL))

#endif /* ! defined(CS_H_INCLUDED) */
