/*========================================================================
 *  Copyright (c) 1999 by Pica
 *
 *  File	: ctx.h
 *  Description	: Interface for the CTX-library
 *  Author(s)	: Hans van Eck
 *  Creation	: 25-11-1999
 *
 *  History:
 *  25-11-1999	: eck      : creation
 *  09-02-2000	: eck      : Changed order of messages (while still possible)
 *  11-02-2000	: eck      : Changed some interfaces.
 *  28-05-2003	: clemens  : Added CTXFildeslistGet().
 *  05-09-2007	: heesakkm : Added CTXT_SESLIST to CTXType_t.
 *  12-10-2007	: heesakkm : Added CTX_SessionlistGet().
 *  12-04-2011	: martinz  : Added CTXPSICONCreate().
 *
 *  Notes	:
 *  The CTX-library handles contexts for the Hermes system.
 *========================================================================
 */

#ifndef CTX_H_INCLUDED
#define CTX_H_INCLUDED

#pragma ident "$Id: ctx.h 302220 2013-08-07 10:11:14Z sieburgw $"

#if defined (CTX_INTERNALS)
#include <rpc/rpc.h>	/* Just for the XDR type definition */
#endif

#include <pica.h>
#include <nmt.h>

ERROR_TYPE(CTXErrno_t);

ERROR_DECLARE(CTXE_NOERROR);
ERROR_DECLARE(CTXE_NOMEM);
ERROR_DECLARE(CTXE_KEYUNKNOWN);
ERROR_DECLARE(CTXE_NETIO);
ERROR_DECLARE(CTXE_INTERNAL);

extern CTXErrno_t CTXerrno;

/* CTX_IDLENGTH is used for the secret key as well as the session-id. This
 * makes the protocol easier to implement (communication to context server
 * will always contain an identifier).
 */
#define CTX_IDLENGTH	32	/* key/session-id max 32 chars */

typedef struct CTXHandle_t *CTXHandle_p;

#if defined (CTX_INTERNALS)
typedef int (*ctxt_xdrfunc_f)(XDR *xdrs, void *);
struct CTXTable_t {
    const char	*ctxt_name;
    ctxt_xdrfunc_f  ctxt_xdrfunc;
};

typedef struct CTXTable_t CTXTable_t, *CTXTable_p;

CTXHandle_p
CTXCONCreate(
	const char	*loopback,	/* hm_loopback */
	const char	*prefix,	/* Name of context */
	CTXTable_p	ctxtable);	/* Table with keys/XDR-routines */

CTXHandle_p
CTXPSICONCreate(
	const char	*loopback,	/* hm_loopback */
	const char	*prefix,	/* Name of context */
	CTXTable_p	ctxtable);	/* Table with keys/XDR-routines */
#endif

/* The following enum contains the numbers just to ensure them. The values are
 * part of the protocol, so changing them is out of the question. This means
 * that new message-type should be added at the end.
 */
typedef enum {
    /* Normal client requests: */
    CTXT_INIT		= 0,	/* Initialise session with context server */
    CTXT_SESCREATE	= 1,	/* Create a new user-session */
    CTXT_SESDESTROY	= 2,	/* Delete user-session */
    CTXT_GET		= 3,	/* Get data */
    CTXT_PUT		= 4,	/* Store data */
    /* Server replies: */
    CTXT_OK		= 5,	/* Action (partly) successful */
    CTXT_ERROR		= 6,	/* Action failed */

    CTXT_SESLIST	= 7	/* Retrieve list of user-sessions */
} CTXType_t;

void CTXCONDestroy(CTXHandle_p contexthandle);

NMTHandle_p
CTX_SessionlistGet(
	CTXHandle_p	ctxp);

/* Simple send-and-receive interface for clients */
bool
CTXSendRecv(CTXHandle_p ctxp, const char *ident,
	CTXType_t *reqtyp, NMTHandle_p data_tab);

/* Send a request (in client) */
bool CTXReqSend( int fildes, const char *ident, CTXType_t reqtyp,
	NMTHandle_p data_tab, NMTHandle_p xdr_tab); 

/* Receive a request (in server) */
bool CTXReqRecv(int fildes, char ident[CTX_IDLENGTH + 1], CTXType_t *p_reqtyp,
	NMTHandle_p data_tab);

/* Send a response (in client) */
bool CTXRespSend(int fildes, CTXType_t reqtyp,
	NMTHandle_p data_tab);

/* Receive a response (in server) */
bool CTXRespRecv(int fildes, CTXType_t *p_reqtyp,
	NMTHandle_p data_tab, NMTHandle_p xdr_tab);

/* Retrieve file descriptors connecting to context servers */
int CTXFildeslistGet(CTXHandle_p ctxp,
	 int fildeslist[], size_t maxnrfildes);

/* The data-binding support interfaces */
/* The function CTXDataBind_NewOld() is the callback for the NMTCreate() for
 * data sent/received through CTX*Send()/CTX*Recv(). The callback will do
 * deallocation of the dynamic data when necessary.
 */
extern cvoidp CTXDataBind_NewOld(const char *name, cvoidp newp, cvoidp oldp);
typedef struct {
    voidp	ctd_data;		/* The data */
    size_t	ctd_size;		/* Size of the data */
    bool	*ctd_pfilled;		/* Boolean (for client) */
    bool	ctd_dynamic;		/* Dynamically allocated (in server) */
} *CTXDataDef_p;

bool CTXDataBind_Add(NMTHandle_p data_tab, const char *name, voidp pvar,
		    size_t psize, bool *have_pvar);

CTXDataDef_p CTXData_New(void);
bool CTXData_Resize(CTXDataDef_p datap, int newsiz);

#endif /* CTX_H_INCLUDED */
