/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: http.h
 *  Description	: Header file for HTTP library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 09-04-1996
 *
 *  History:
 *  09-04-1996	: lolke    : creation
 *  27-05-1997	: eck      : Changed interface to HTTPHeaderComplete()
 *  15-05-1998	: BdB	   : Added HTTPUrlEncode* and HTTPParseCookies
 *  08-04-1999	: eck      : Added status HTTP_INTERNAL_REDIRECT and
 *			   : HTTPParseResponseString()
 *  10-10-2003	: paul     : Added HTTPUrlEncode()
 *  23-10-2012	: martinz  : Moved documentation from Word file.
 *  22-03-2017	: heesakkm : CBS-7015: Added HTTP_UNPROCESSABLE_ENTITY
 *  04-10-2018	: zwaalm   : CBS-8399 Added url-decode functions.
 *
 *========================================================================
 */

#if ! defined(HTTP_H_INCLUDED)
#define HTTP_H_INCLUDED

#pragma ident "$Id: http.h 374278 2020-02-13 08:40:07Z smitp $"

#include <stddef.h>
#include <time.h>

#include <pica.h>
#include <cs.h>
#include <nmt.h>
#include <cb.h>

#define HTTP_NONE				0	// no HTTP status set
#define HTTP_OK					200
#define HTTP_CREATED				201
#define HTTP_ACCEPTED				202
#define HTTP_NON_AUTHORITIVE_INFORMATION	203
#define HTTP_NO_CONTENT				204
#define HTTP_MULTIPLE_CHOICES			300
#define HTTP_MOVED_PERMANENTLY			301
#define HTTP_MOVED_TEMPORARILY			302
#define HTTP_SEE_OTHER				303
#define HTTP_NOT_MODIFIED			304
#define HTTP_INTERNAL_REDIRECT			399	/* For dispatcher */
#define HTTP_BAD_REQUEST			400
#define HTTP_UNAUTHORIZED			401
#define HTTP_PAYMENT_REQUIRED			402
#define HTTP_FORBIDDEN				403
#define HTTP_NOT_FOUND				404
#define HTTP_METHOD_NOT_ALLOWED			405
#define HTTP_NONE_ACCEPTABLE			406
#define HTTP_PROXY_AUTHENTICATION_REQUIRED	407
#define HTTP_REQUEST_TIMEOUT			408
#define HTTP_CONFLICT				409
#define HTTP_GONE				410
#define HTTP_LENGTH_REQUIRED			411
#define HTTP_UNPROCESSABLE_ENTITY		422
#define HTTP_INTERNAL_SERVER_ERROR		500
#define HTTP_NOT_IMPLEMENTED			501
#define HTTP_BAD_GATEWAY			502
#define HTTP_SERVICE_UNAVAILABLE		503
#define HTTP_GATEWAY_TIMEOUT			504

// Possible values of charset param of HTTP Accept header
// HTTP identification of Unicode-utf8 encoding
// HTTP 1.1 says that the default charset is ISO-8859-1.
#define HTTP_CHARSET_UTF8	"utf-8"

typedef struct {
    char method[10];
    char *URL;
    char version[10];
    char code[4];
    char phrase[128];
    char *entity;
    long entlen;
} HTTPHandle_t;

typedef HTTPHandle_t *HTTPHandle_p;

ERROR_TYPE(HTTPErrno_t);

ERROR_DECLARE(HTTPE_ERROR);             /* unspecified internal error */
ERROR_DECLARE(HTTPE_NOERROR);           /* ok */
ERROR_DECLARE(HTTPE_NOMEM);             /* not enough memory */
ERROR_DECLARE(HTTPE_OVERFLOW);          /* buffer overflow */
ERROR_DECLARE(HTTPE_SOCKET);		/* socket error */
ERROR_DECLARE(HTTPE_INVALID_RESPONSE);	/* response error */
ERROR_DECLARE(HTTPE_INVALID_REQUEST);   /* request error */
ERROR_DECLARE(HTTPE_INVALID_HEADER);    /* header error */
ERROR_DECLARE(HTTPE_EOF);               /* EOF before any data read */

extern HTTPErrno_t HTTPerrno;

#define HTTP_NOFUNC ((void (*)(const char *, const char*, void*)) 0)

void HTTPFreeHandle(HTTPHandle_p handle);

HTTPHandle_p HTTPParseResponse(int fildes,
			    void (*func)(const char *name, const char *value, void *userdata), void *userdata);

HTTPHandle_p HTTPParseResponseString(const char *in_buf, long in_len,
			    void (*func)(const char *name, const char *value, void *userdata), void *userdata);

HTTPHandle_p HTTPParseRequest(int fildes,
			    void (*func)(const char *name, const char *value, void *userdata), void *userdata);

HTTPHandle_p HTTPParseRequestString(const char *in_buf, long in_len,
			    void (*func)(const char *name, const char *value, void *userdata), void *userdata);

int HTTPHeaderComplete(char *in_buf, long in_len, long *body_len);

bool HTTPPutResponseLine(int fildes, int maj_vers, int minor_vers, int status);

bool HTTPSPutResponseLine(int fildes, int maj_vers, int minor_vers, int status);


bool HTTPPutRequestLine(int fildes, const char *method, const char *URL,
			int maj_vers, int minor_vers);

bool HTTPPutHeaderLine(int fildes, const char *name, const char *value);

bool HTTPPutBody(int fildes, const char *buf, int bufsiz);

void HTTPParseNameValue(const char *buf, CBBuf_t *cbbuf, size_t cbsiz);

void HTTPGetReasonPhrase(int status, char *buf, size_t bufsiz);

bool HTTPNormalToSave(const char *inbuf, char *outbuf, size_t outsiz);

bool HTTPSaveToNormal(const char *inbuf, char *outbuf, size_t outsiz);

void HTTPMakeDateString(const time_t tim, char *outbuf, size_t outsiz);

/* Destroy memory returned by HTTPUrlEncodeCreate. */
void HTTPUrlEncodeDestroy(const char *p);

/* URL encode input string.  Return encoded string in memory which must be
   destroyed with HTTPUrlEncodeDestroy or NULL on out of memory. */
const char *HTTPUrlEncodeCreate(const char *in);
long HTTPUrlEncode(const char *in, long len, char *out, size_t out_size);

const char *HTTPUrlDecodeCreate(const char *in);
void HTTPUrlDecodeDestroy(const char *p);

/* Parse a Cookie: HTTP header line value into separate cookies and call
   callback for each cookie.  Values will not be URL decoded by this routine.
   Although it takes a 'const char *cookies', it will temporarily modify it, so
   don't pass it a string literal. */
bool
HTTPParseCookies(const char *cookies,
		 void (*func)(const char *name, const char *value,
			       void *userdata),
		 void *userdata);

const char *
HTTPAcceptHeaderFind(const char		*http_accept,
		     const char		*search_value,
		     const char		*param_name,
		     NMTHandle_p	supported_values,
		     const char		*options);

#endif /* ! defined(HTTP_H_INCLUDED) */
