/*========================================================================
 *  Copyright (c) 2004 by OCLC PICA
 *
 *  File	: idxinternals.h
 *  Description	: Details that were (and should be) local to the IDX
 *		  library but are now also used by the IDG library
 *  Author(s)	: Jaap Lutz
 *  Creation	: 18-05-2004
 *
 *  History:
 *  18-05-2004	: jaap     : creation (from idxlocal.h)
 *  30-06-2004  : paul     : added IDXFilterToInternField() and
 *                           IDXAdiFieldValueToExtern() and
 *                           IDXAdiFieldToExternalFilter()
 *  10-05-2007	: paul     : Allow for a list of intervals in adifilter
 *  05-11-2008	: paul     : Increased the max length of an ADI filter, for BSZ
 *  04-12-2009  : daalmanf : added macro's for IDXDEF_ADITYPE_LONG_NUMBER
 *                         : and enlarge IDXAdiFilterFieldInterval_t to 32-bit
 *                         : ranges (Jaap Lutz, GBV, for CollectionWS/PerLis)
 *  12-07-2017	: zwaalm   : CBS-7349 Changed comment about adi structure.
 *  12-11-2018	: zwaalm   : CBS-8478 Moved ADI-fields to idx.h.
 *
 *  Notes	:
 *
 *========================================================================
 */

#ifndef IDXINTERNALS_H_INCLUDED
#define IDXINTERNALS_H_INCLUDED

#pragma ident "$Id: idxinternals.h 363677 2019-01-10 15:10:10Z heesakkm $"

#include <pica.h>
#include <idxtypes.h>
#include <idx.h>

/*-----------------------------------------------------------------------------
 * IDX_CHAR_BIT is defined to be all of the following:
 *     1. The minimum number of bits that a char must contain, if it is to be
 *        able to support Pica ADI and data.
 *     2. The maximum number of bits that ADI uses per char.
 *     3. The maximum number of bits that data tag 001@ uses per char.
 *     4. Half of the number of bits that an uint16 contains. (Since for any
 *        uint16 value it must be possible to represent it in 2 unsigned
 *        char values and vice versa)
 * IDX_UCHAR_MAX is the maximum value of an unsigned char that uses exactly
 *     IDX_CHAR_BIT bits
 * IDXSetOfBytes_t is the type of an array containing one bit per char,
 *     indicating whether the char is present (1) in the set or not (0)
 */
#define IDX_CHAR_BIT         8
#define IDX_UCHAR_MAX        255
#define IDX_SIZEOF_SET_OF_BYTES ((IDX_UCHAR_MAX + 1) / IDX_CHAR_BIT)
typedef byte   IDXSetOfBytes_t[IDX_SIZEOF_SET_OF_BYTES];
#define IDX_BYTE_IN_SET(byt, set) \
   ((  (unsigned)(set)[(int)(byt) / (int)IDX_CHAR_BIT] \
     & (1U << (  (unsigned)IDX_CHAR_BIT - 1U \
               - ((unsigned)(byt) % (unsigned)IDX_CHAR_BIT)))) \
    != 0U)

/* The maximum number of ranges in an interval ADI filter
 * BSZ needs this number to be at least 85
 */
#define IDX_INTERVAL_COUNT_MAX	128

/*---------------------------------------------------------------------------
 *   Defines for (un-)packing ADI
 *---------------------------------------------------------------------------
 */
#define IDX_PACK_UINT16_ADI(adifld, uint16val) \
    ((uint8 *)(adifld))[0] = (uint16val) / (uint16)(IDX_UCHAR_MAX + 1); \
    ((uint8 *)(adifld))[1] = (uint16val) % (uint16)(IDX_UCHAR_MAX + 1);

/*
 * The packing of int16 ADI is done using offset binary, also called excess-K
 * or biased representation. We can't use either ('/' and '%') or ('>>' and
 * '&') on these offset binary numbers. What here is done should match
 * lib/idg/idgvm_machine.c:number_as_string(). Note that we use MSB first.
 */
#define IDX_PACK_INT16_ADI(adifld, int16val) \
    IDX_PACK_UINT16_ADI((adifld), (uint16)((long)(int16val) - (long)INT16MIN))
#define IDX_UNPACK_UINT16_ADI(adifld, uint16val) \
    (uint16val) = (uint16)(IDX_UCHAR_MAX + 1) \
                  * (uint16)((const uint8 *)(adifld))[0] \
                  + (uint16)((const uint8 *)(adifld))[1]
#define IDX_UNPACK_INT16_ADI(adifld, int16val) { \
    uint16       uint16val; \
\
    IDX_UNPACK_UINT16_ADI((adifld), uint16val); \
    (int16val) = (int16)((long)uint16val + (long)INT16MIN); \
}
#define IDX_PACK_UINT32_ADI(adifld, uint32val) PM_PACK4((byte *)adifld, uint32val)
#define IDX_UNPACK_UINT32_ADI(adifld, uint32val) uint32val = PM_UNPACK4((byte *)adifld)
#define IDX_UNPACK_ADI_FIELD(adifld, adibuf, fldoffset, fldlen) \
    switch((adifld)->type) { \
    case IDXDEF_ADITYPE_ASCII: \
        (adifld)->val.ascii.buf = (const char *)(adibuf) + (fldoffset); \
        (adifld)->val.ascii.len = (size_t)(fldlen); \
        break; \
    case IDXDEF_ADITYPE_BITMAP: \
        (adifld)->val.bitmap.buf = (adibuf) + (fldoffset); \
        (adifld)->val.bitmap.len = (size_t)(fldlen); \
        break; \
    case IDXDEF_ADITYPE_INTERVAL: \
        IDX_UNPACK_UINT16_ADI((adibuf) + (fldoffset), \
                              (adifld)->val.interval.lo); \
        IDX_UNPACK_UINT16_ADI((adibuf) + (fldoffset) \
                                       + IDXDEF_ADISIZE_NUMBER, \
                              (adifld)->val.interval.hi); \
        break; \
    case IDXDEF_ADITYPE_SIGNED_INTERVAL: \
        IDX_UNPACK_INT16_ADI((adibuf) + (fldoffset), \
                             (adifld)->val.signed_interval.lo); \
        IDX_UNPACK_INT16_ADI((adibuf) + (fldoffset) \
                                      + IDXDEF_ADISIZE_NUMBER, \
                             (adifld)->val.signed_interval.hi); \
        break; \
    case IDXDEF_ADITYPE_NUMBER: \
        IDX_UNPACK_UINT16_ADI((adibuf) + (fldoffset), \
                              (adifld)->val.number); \
        break; \
    case IDXDEF_ADITYPE_SHORT_NUMBER: \
        (adifld)->val.short_number = (adibuf)[(fldoffset)]; \
        break; \
    case IDXDEF_ADITYPE_LONG_NUMBER: \
        IDX_UNPACK_UINT32_ADI((adibuf) + (fldoffset), \
                              (adifld)->val.long_number); \
        break; \
    case IDXDEF_ADITYPE_SHORT_INTERVAL: \
        (adifld)->val.short_interval.lo = (adibuf)[(fldoffset)]; \
        (adifld)->val.short_interval.hi = \
            (adibuf)[(fldoffset) + IDXDEF_ADISIZE_SHORT_NUMBER]; \
        break; \
    default: IDXerrno = IDXE_INTERNAL; \
    }
/*
 * Typedef of the structure that defines how to select ADI fields.
 */

typedef struct IDXAdiFilterFieldIntervalSigned_t {
    int32		lo, hi;
} IDXAdiFilterFieldIntervalSigned_t, *IDXAdiFilterFieldIntervalSigned_p;
typedef struct IDXAdiFilterFieldInterval_t {
    uint32		lo, hi;
} IDXAdiFilterFieldInterval_t, *IDXAdiFilterFieldInterval_p;    
typedef struct IDXAdiFilterFieldRecType_t {
    IDXIktNum_t          iktnum;
    bool                 filter;
    union {
        struct {
            IDXSetOfBytes_t set;
        }              ascii_char;
        struct {
            char            buf[IDX_SIZEOF_SET_OF_BYTES];
        }              ascii_string;
        struct {
            IDXSetOfBytes_t set;
        }              bitmap;
	IDXAdiFilterFieldIntervalSigned_t
			signed_interval[IDX_INTERVAL_COUNT_MAX];
        IDXAdiFilterFieldInterval_t
			interval[IDX_INTERVAL_COUNT_MAX];
        struct {
            IDXSetOfBytes_t set;
        }              short_number;
        struct {
            IDXSetOfBytes_t set;
        }              short_interval;
    }                    val;
} IDXAdiFilterFieldRecType_t;
typedef struct IDXAdiFilterField_t {
    IDXAdiFilterFieldRecType_t title;
    IDXAdiFilterFieldRecType_t entry;
    IDXDefAdiType_t            type; 
    IDXDefAdiAccess_t          access;
    size_t                     offset;
    size_t                     len;
} IDXAdiFilterField_t;
typedef struct IDXAdiFilter_t {
    IDXRecType_t         rectypemask;
    size_t               len;
    IDXAdiFilterField_t  fld[IDX_ADIFLD_MAX];
} IDXAdiFilter_t;

/*
 * IDXNumber_t is the type of any int value that is read from the source table
 */
#define IDX_NUMBER_MIN    0
#define IDX_NUMBER_MAX    USHRT_MAX
typedef uint16            IDXNumber_t;

#define IDX_IKTCOUNT_MAX      400
#define IDX_IKTMNEMO_SIZE         3
#define IDX_IKTMNEMO_DESC_SIZE    20
#define IDX_ADIIKTCODE_DESC_SIZE  20
#define IDX_ADIIKTCODECOUNT_MAX   400
#define IDX_IKTMNEMOS_BUF_SIZE \
    (  (IDX_IKTCOUNT_MAX * (IDX_IKTMNEMO_SIZE + IDX_IKTMNEMO_DESC_SIZE)) \
     + (IDX_ADIIKTCODECOUNT_MAX * IDX_ADIIKTCODE_DESC_SIZE))
#define IDX_IKTELTCOUNT_MAX       (IDX_IKTCOUNT_MAX + IDX_ADIIKTCODECOUNT_MAX)
typedef union IDXIktMnemoElt_t {
    IDXIktMnemo_t                       ikt;
    IDXAdiCode_t                        adicode;
} IDXIktMnemoElt_t;
typedef struct IDXIktMnemos_t {
    size_t                              len;
    size_t                              iktcnt;
    IDXIktMnemoElt_t                    elt[IDX_IKTELTCOUNT_MAX];
    char                                buf[IDX_IKTMNEMOS_BUF_SIZE];
    char                               *bufend;
} IDXIktMnemos_t;

IDXDefAdiType_t IDXAdiAccessToType(IDXDefAdiAccess_t access, IDXNumber_t len);

void
IDXAddAdiCode(
        const char          *str,
        IDXIktMnemos_t      *iktmnemos,
        IDXIktNum_t          iktnum,
        IDXDefAdiAccess_t    adiaccess,
        size_t               adilen,
        size_t               firstelt);

bool
IDXAdiFieldToExternalFilter(
    const struct IDXAdiField_t	*adifld,
    char			*ext,
    size_t			ext_size);

bool
IDXFilterToInternField(
    const char                       *value,
    const IDXAdiCode_t               *adicode,
    IDXAdiFilterField_t      	     *adifld,
    IDXAdiFilterFieldRecType_t       *adifldrect,
    const IDXAdiFilterField_t        *maxfld,
    const IDXAdiFilterFieldRecType_t *maxfldrect);

bool
IDXAdiFieldValueToExtern(
        const IDXAdiFilterField_t        *adifld,
        const IDXAdiFilterFieldRecType_t *adifldrect,
	const char                       *extadifilterend,
        char                             **extadifilter,
	const IDXAdiCode_t               *adicode,
	bool                              ranges);

#endif /* IDXINTERNALS_H_INCLUDED */
