/*
 *  Copyright (c) 1993 by Pica
 *
 *  File        : picamisc.h
 *  Description : Miscellaneous routines for general use.
 *  Author(s)   : Hans van Eck
 *  Creation    : 26-11-1993
 *
 *  History:
 *  26-11-1993  : eck      : creation
 *  06-01-1994  : eck      : Added PM_gensym()
 *  10-01-1994  : eck      : Added PM_utoa() and PM_itoa()
 *  09-02-1994  : eck      : Added PM_hash()
 *  25-03-1994  : eck      : Return types of PM_*toa() now (const char *)
 *  13-01-1995  : eck      : Added PM_range_clean() and PM_RANGEBUF_SIZE
 *                         : Added PM_ltoa() and PM_ultoa()
 *  06-02-1996  : eck      : Added PM_range_*() routines.
 *  07-02-1996  : eck      : Increased PM_INTLEN_MAX (for alpha).
 *  15-03-1996  : eck      : Added PM_Mod11*() routines.
 *  20-05-1996  : eck      : Added PM_strcmpi() and PM_strncmpi().
 *  03-07-1996  : rt       : Security functions added.
 *  14-03-1997  : paul     : Added PM_ppn_*() routines
 *  18-08-1997  : eck      : Changed return type of PM_hash().
 *  11-11-1997  : BdB      : Added PM_hyphenate_is{b,s}n routines.
 *  13-01-1998  : BdB      : Added PM_DES*() routines
 *  30-01-1998  : BdB      : Added PM_CRC32().
 *  13-03-1998  : lolke    : PMPpnTab_t no longer hidden (temporary).
 *  23-10-1998  : teresita : Added PM_time_ functions
 *  26-02-1999  : eck      : Added PM_Base64[En|De]code() functions
 *  01-03-1999  : martin   : Added PM_time_DifferDateTime function
 *  29-04-1999  : eric     : Added PM_MakeSortString(), PM_UsedByIlns*().
 *  24-03-2000  : teresita : Added PM_LBSStamp()
 *  19-05-2000  : eck      : Added functions for creating/accepting connections
 *  22-05-2000  : eck      : Moved connection-functions to con.h
 *  22-06-2000  : eck      : Added PM_strdup()
 *  26-07-2000  : paul     : PM_ISBN_SIZE and moved from pmisbn.c
 *  27-07-2000  : eck      : Added PM_strncpy()
 *  20-09-2000  : evd      : Added PM_epn_* routines
 *  17-10-2000  : paul     : PM_RANGE_ERROR, PM_range_from_bits() and
 *                           PM_range_to_bits() introduced
 *  12-01-2001  : eck      : Added PM_bit_count() and PM_userbit_test()
 *  12-03-2001  : paul     : PM_ISBN_HYPHEN introduced
 *  05-11-2001  : paul     : PM_opt_findvalend() introduced
 *  27-11-2001  : paul     : Added PM_range_intersect() + PM_range_unite() +
 *                           PM_RANGE_BUF_EMPTY + PM_range_error()
 *  31-03-2003  : HvdW     : Tandem ifdefs removed
 *  21-10-2003  : Guy      : added PM_PACK* and PM_UNPACK* macros (copied 
 *                           from idx library)
 *  18-11-2003  : paul     : Added Prototypes for pm_utf.c
 *  21-11-2003  : paul     : Added PM_Utf32CharRefPut()
 *  20-04-2004  : Guy      : Moved PM_APPEND_UNSAFE from pm_utf.c to here 
 *  21-04-2004  : Guy      : Added PM_toupper & PM_tolower
 *  28-06-2004  : paul     : Added PM_range_to_bits_check()
 *  02-09-2004  : paul	   : Added PM_ISSN_EAN_PREFIX +PM_ISBN_EAN_PREFIX*
 *  27-09-2004	: heesakkm : Added PM_UTF32_COMBINING_GRAPHEME_JOINER
 *  12-04-2005  : Guy      : Changed PM_ISBN_SIZE to 13
 *  08-08-2005  : paul     : Changed PM_ISBN_SIZE back to 10
 *                         : Introduced PM_BOOKLAND_EAN_SIZE with value 13
 *  23-09-2005  : paul     : Added PM_atosize()
 *  30-09-2006  : paul     : Added PM_month2l() and PM_roman2l()
 *  07-07-2006	: henk     : PM_userbit_previous and PM_userbit_next added
 *  07-07-2006	: heesakkm : Added PM_range_x_*() routines.
 *  29-08-2006  : paul     : Added PM_EAN_*() functions
 *  12-03-2008  : paul     : Added defines of private use characters
 *  07-03-2008  : sutherlr : added PM_isbn_13_to_10() and PM_isbn_10_to_13()
 *  24-07-2008  : paul     : added line marking characters for bidirectional
 *                           presentations
 *  24-10-2008	: paul     : added PM_UTF32_REPLACEMENT_CHARACTER
 *  10-11-2008	: clemens  : added PM_version_strcmp().
 *  04-05-2009	: paul     : added PM_UTF32_NEWLINE (temporarily as everything)
 *  22-06-2009  : evd      : added PM_pinyin_syllables
 *  23-06-2009  : evd      : Moved PM_pinyin_syllables to PU library. (PU8_....)
 *  05-08-2009	: sutherlr : Removed functions PM_range_x_from_bits(), 
 *                           PM_range_x_to_bits_check(), and
 *                           PM_range_x_to_bits().
 *  27-04-2010	: paul     : CBS-1581 Added PM_Mod11GenBNF()
 *  11-10-2010	: martinz  : Added PM_time_local_to_UTC().
 *  24-03-2011	: paul     : CBS-2776 automatically create ISNI in 003E
 *  14-01-2014  : evd      : CBS-4106 Added PM_Mod29GenBNF()
 *  10-03-2017	: smitp    : moved PM_UTF8_IS_SINGLE() PM_UTF8_IS_LEAD()
 *                           PM_UTF8_IS_TRAIL() from <pu.h>
 *  30-10-2017	: smitp    : CBS-7763 Added defines for escape sequences in
 *                           normalized cataloguing format
 *
 *  Notes       :
 *  PM_PACK*            - macros to convert, in a portable manner,  
 *                        16/32/48/64 bit integers to big-endian byte strings
 *  PM_UNPACK*          - macros to convert, in a portable manner,
 *                        big-endian byte strings to 16/32/48/64 bit integers
 *  PM_opt_*            - general routines for parsing DB-options
 *  PM_gensym           - generate unique string
 *  PM_*toa             - generate strings from int or unsigned
 *  PM_hash             - generate (integer) hash from a '\0'-terminated string
 *  PM_range_*          - routines dealing with range-strings
 *  PM_Utf*		- Utf8 and utf32 character (aka code point) functions
 *  PM_toupper          - convert one pica character to uppercase 
 *  PM_tolower          - convert one pica character to lowercase
 */
#if ! defined(PICAMISC_H_INCLUDED)
#define PICAMISC_H_INCLUDED

#pragma ident "$Id: picamisc.h 378221 2020-07-16 13:46:16Z vaneckh $"

#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include <pica.h>
#include <picatypes.h>
#include <cb.h>

#ifdef  __cplusplus
#ifndef bool
#define __bool_defined_here
#define bool Cbool
#elif bool != Cbool
#error "When #define'd 'bool' must be 'Cbool'"
#endif
extern "C" {
#endif

// Defines needed in normalized cataloguing format used in Pica3 protocol

// The char that terminates a field in normalized cataloguing format
#define PM_CR  "\015"

// The char that forms an escape-sequence in combination with subsequent char
#define PM_ESC "\033"

// Characters used in escape sequences (behind ESC character)
// 'P' is not just used for protecting tags: also for fields in a window
// 'H' for Hide and 'L' for P3Local variable may be added later
#define	PM_ESC_INACCESSIBLE_OPEN	"I"
#define	PM_ESC_INACCESSIBLE_CLOSE	"N"
#define PM_ESC_PROTECT_OPEN		"P"
#define PM_ESC_PROTECT_CLOSE		"D"
// S-O for "isSelected" property in CCweb
#define PM_ESC_SELECT_OPEN		"S"
#define PM_ESC_SELECT_CLOSE		"O"
// M-Q for "isModified" property in CCweb
#define PM_ESC_MODIFIED_OPEN		"M"
#define PM_ESC_MODIFIED_CLOSE		"Q"

// Escape sequences around PPN-expansion in normalized cataloguing format
#define	PM_PPNEXPANSION_OPEN		(PM_ESC PM_ESC_INACCESSIBLE_OPEN)
#define	PM_PPNEXPANSION_CLOSE		(PM_ESC PM_ESC_INACCESSIBLE_CLOSE)

// Escape sequences around protected field in normalized cataloguing format
#define PM_TAG_PROTECT_OPEN             (PM_ESC PM_ESC_PROTECT_OPEN)
#define PM_TAG_PROTECT_CLOSE            (PM_ESC PM_ESC_PROTECT_CLOSE)

// Escape sequences around selected field in normalized cataloguing format
#define PM_TAG_SELECT_OPEN             (PM_ESC PM_ESC_SELECT_OPEN)
#define PM_TAG_SELECT_CLOSE            (PM_ESC PM_ESC_SELECT_CLOSE)

// Escape sequences around modified field in normalized cataloguing format
#define PM_TAG_MODIFIED_OPEN             (PM_ESC PM_ESC_MODIFIED_OPEN)
#define PM_TAG_MODIFIED_CLOSE            (PM_ESC PM_ESC_MODIFIED_CLOSE)

// Defines used for marking literal strings arguments in CBS commands
#define PM_MARKER_SINGLE_QUOTE		'\''
#define PM_MARKER_DOUBLE_QUOTE		'"'
#define PM_MARKER_SLASH			'/'
#define PM_MARKER_BACK_QUOTE		'`'
#define PM_MARKER_PIPE			'|'
#define PM_MARKER_LNK_EXACT_1		PM_MARKER_SINGLE_QUOTE
#define PM_MARKER_LNK_EXACT_2		PM_MARKER_DOUBLE_QUOTE
#define PM_MARKER_LNK_TRUNCATED		PM_MARKER_PIPE

/* in the definitions of PM_PACK* and PM_UNPACK* we assume CHAR_BIT == 8;
 * results will not be correct when this is not the case */
 
#define PM_PACK2(p,x) {                                       \
                          (p)[0] = (byte)((uint16)(x) >> 8);  \
                          (p)[1] = (byte)((uint16)(x));       \
                      }

#define PM_PACK4(p,x) {                                       \
                          (p)[0] = (byte)((uint32)(x) >> 24); \
                          (p)[1] = (byte)((uint32)(x) >> 16); \
                          (p)[2] = (byte)((uint32)(x) >> 8);  \
                          (p)[3] = (byte)((uint32)(x));       \
                      }

#define PM_PACK6(p,x) {                                       \
                          (p)[0] = (byte)((uint64)(x) >> 40); \
                          (p)[1] = (byte)((uint64)(x) >> 32); \
                          (p)[2] = (byte)((uint64)(x) >> 24); \
                          (p)[3] = (byte)((uint64)(x) >> 16); \
                          (p)[4] = (byte)((uint64)(x) >> 8);  \
                          (p)[5] = (byte)((uint64)(x));       \
                      }

#define PM_PACK8(p,x) {                                       \
                          (p)[0] = (byte)((uint64)(x) >> 56); \
                          (p)[1] = (byte)((uint64)(x) >> 48); \
                          (p)[2] = (byte)((uint64)(x) >> 40); \
                          (p)[3] = (byte)((uint64)(x) >> 32); \
                          (p)[4] = (byte)((uint64)(x) >> 24); \
                          (p)[5] = (byte)((uint64)(x) >> 16); \
                          (p)[6] = (byte)((uint64)(x) >> 8);  \
                          (p)[7] = (byte)((uint64)(x));       \
                     }

#define PM_UNPACK2(p)  (((uint16)(p)[0] << 8) +  \
                        ((uint16)(p)[1]))

#define PM_UNPACK4(p)  (((uint32)(p)[0] << 24) + \
                        ((uint32)(p)[1] << 16) + \
                        ((uint32)(p)[2] << 8) +  \
                        ((uint32)(p)[3]))

#define PM_UNPACK6(p)  (((uint64)(p)[0] << 40) + \
                        ((uint64)(p)[1] << 32) + \
                        ((uint64)(p)[2] << 24) + \
                        ((uint64)(p)[3] << 16) + \
                        ((uint64)(p)[4] << 8) +  \
                        ((uint64)(p)[5]))

#define PM_UNPACK8(p)  (((uint64)(p)[0] << 56) + \
                        ((uint64)(p)[1] << 48) + \
                        ((uint64)(p)[2] << 40) + \
                        ((uint64)(p)[3] << 32) + \
                        ((uint64)(p)[4] << 24) + \
                        ((uint64)(p)[5] << 16) + \
                        ((uint64)(p)[6] << 8) +  \
                        ((uint64)(p)[7]))

bool PM_opt_equal(const char *optstring, const char *optname);

const char *PM_opt_findstart(const char *options);

const char *PM_opt_findname(const char *options);

const char *PM_opt_findnameend(const char *options);

const char *PM_opt_findval(const char *options);

const char *PM_opt_findvalend(const char *options);

const char *PM_opt_findend(const char *options);

int PM_opt_copyval(const char *options, char *target, size_t maxlen);

const char *PM_opt_findnext(const char *options);

int PM_opt_fetch(const char *options, const char *name,
                char *target, size_t maxlen);

/**
 * PM_SYMLEN_MIN is a constant for the minimum size of a unique symbol.
 * The actual minimum length is 11, but there is some extra lenght for the
 * start of the symbol. Also, when using 64-bit unique numbers, this will
 * suffice.
 */
#define PM_SYMLEN_MIN   32

const char *PM_gensym(const char *SymStart);

/**
 * PM_INTLEN_MAX is a constant for the maximum possible length of an
 * (long) integer string. It is also good enough for unsigned ints.
 * The length includes a possible '-' sign and the trailing '\0'.
 */
#define PM_INTLEN_MAX   22

const char *PM_itoa(char *string, int value);

const char *PM_ltoa(char *string, long value);

const char *PM_utoa(char *string, unsigned value);

const char *PM_ultoa(char *string, unsigned long value);

uint32 PM_hash(const char *string);

/**
 * PM_RANGE_BUF_SIZE is a constant for the maximum length for a string
 * containing a range. It was previously necessary when PM_range_clean()
 * did not have a buf_size parameter. It is now for convenience only.
 * NOTE: do NOT use this constant for ILN ranges. Please refer to 
 * picatypes.h for the relevant constants. 
 */
#define PM_RANGEBUF_SIZE        200

/**
 * PM_RANGE_ERROR is the unsigned long value that indicates error from
 * PM_range_from_bits().
 */
#define PM_RANGE_ERROR          ULONG_MAX

/**
 * PM_RANGE_BUF_EMPTY is the string denoting a range containing no number
 */
#define PM_RANGE_BUF_EMPTY       "-"

unsigned long PM_range_from_bits(char *rangebuf, size_t buf_size,
                                 const byte *bitarr, size_t bitlen,
                                 unsigned long lowest);

unsigned long PM_range_to_bits(byte *bitarr, size_t bitsize, size_t *bitlen,
                               const char *rangebuf, unsigned long lowest);

unsigned long PM_range_to_bits_check(byte *bitarr, size_t bitsize,
				     size_t *bitlen, const char *rangebuf,
				     long rangebuflen, unsigned long   lowest);

const char *PM_range_error(const char *rangebuf);

const char *PM_range_x_error(const char *rangebuf);

unsigned long PM_range_clean(char *rangebuf, size_t buf_size,
                            unsigned long low, unsigned long high);

unsigned long PM_range_x_clean(char *rangebuf, size_t buf_size,
                            unsigned long low, unsigned long high);

unsigned long PM_range_add(char *rangebuf, size_t buf_size,
                            unsigned long first, unsigned long last);

unsigned long PM_range_x_add(char *rangebuf, size_t buf_size,
                            unsigned long first, unsigned long last);

unsigned long PM_range_del_number(char *rangebuf, size_t buf_size,
                            unsigned long number);

unsigned long PM_range_x_del_number(char *rangebuf, size_t buf_size,
                            unsigned long number);

const char *PM_range_get_bounds(const char *rangebuf,
                            unsigned long *first, unsigned long *last);

const char *PM_range_x_get_bounds(const char *rangebuf,
                            unsigned long *first, unsigned long *last);

bool PM_range_test_number(const char *rangebuf, unsigned long number);

bool PM_range_x_test_number(const char *rangebuf, unsigned long number);

unsigned long PM_range_intersect(char *destbuf, size_t buf_size,
				 const char *src1buf, const char *src2buf);

unsigned long PM_range_x_intersect(char *destbuf, size_t buf_size,
				 const char *src1buf, const char *src2buf);

unsigned long PM_range_unite(char *destbuf, size_t buf_size,
			     const char *src1buf, const char *src2buf);

unsigned long PM_range_x_unite(char *destbuf, size_t buf_size,
			     const char *src1buf, const char *src2buf);

typedef enum {
    PMRC_OP_NONE,
    PMRC_OP_UNITE,
    PMRC_OP_INTERSECT,
    PMRC_OP_SUBTRACT
} PM_range_combine_op_t;
unsigned long PM_range_combine(PM_range_combine_op_t op, 
			       char *destbuf, size_t buf_size, size_t *buf_size_req,
			       const char *src1buf, const char *src2buf,
			       int base);

unsigned long PM_range_count(const char *rangebuf);
unsigned long PM_range_x_count(const char *rangebuf);

/** PM_ISNI_LEN defines the length of an ISNI, including check digit */
#define PM_ISNI_LEN		16

/**
 * PM_ISNI_CHECKDIGIT_NONE is returned if the ISNI check digit couldn't be
 * computed.
 */
#define PM_ISNI_CHECKDIGIT_NONE	'?'
char PM_Mod11GenISNI(const char *str, int len);

char PM_Mod11GenBNF(const char *str, int len);

char PM_Mod29GenBNF(const char *str);

char PM_Mod11Gen(const char *str, int len);

bool PM_Mod11Test(const char *str, int len);

char PM_EAN_gen(const char *str, int len);

bool PM_EAN_test(const char *str, int len);

/*  PMPpnTab_p points to the table needed to be able to convert PPN's
 *  to IPN's vice versa.
 */
#define PM_PPN_TAB_SIZE            14
#define PM_PPN_YEAR_SIZE           2

typedef struct PMPpnTab_t {
    int                             len;
    struct
    {
        char                            name[PM_PPN_YEAR_SIZE + 1];
        PICIpn_t                        ipn;
    } year[PM_PPN_TAB_SIZE];
} PMPpnTab_t;
typedef const struct PMPpnTab_t  *PMPpnTab_p;

PMPpnTab_p PM_ppn_tab_create(CBBuf_t *CBBuf, size_t CBSize);

void PM_ppn_tab_destroy(PMPpnTab_p ppntab);

PICIpn_t PM_ppn_to_ipn(PMPpnTab_p ppntab, const char *ppn);

const char *PM_ppn_from_ipn(PMPpnTab_p ppntab, PICIpn_t ipn, char *ppn);

PICIpn_t PM_epn_to_ipn(const char *epn);

const char *PM_epn_from_ipn(PICIpn_t ipn, char *epn);

int PM_strcmpi(const char *str1, const char *str2);

int PM_strncmpi(const char *str1, const char *str2, size_t len);

char *PM_strdup(const char *str, int len);
char *PM_strncpy(char *target, const char *source, size_t siz);

int PM_old_scramble(byte *outstring, size_t outsize, const char *instring,
                                int used_len, int32 seed);

int32 PM_scramble(const char *secret, const char *public_);

size_t PM_hex_unpack(char *out, size_t outsiz, const byte *in, int packedlen);

size_t PM_hex_pack(byte *out, size_t outsiz, const char *in, int unpackedlen);

/*
 * The International Article Numbering Association (EAN) has introduced a
 * unique number for each commercial product. If this product is a book, the
 * number is derived from its ISBN; if it is a serial publication, the number
 * is derived from its ISSN. EAN numbers come in several flavors. The one used
 * for ISBN's is EAN.UCC-13, which contains exactly 13 digits: 3 digits prefix
 * and 9 digits ISBN and 1 checkdigit. In general, the EAN prefix identifies a
 * country and need not have length 3. The prefix for ISBN's identifies the
 * imaginary country "Bookland". a second prefix has already been reserved in
 * case the range of numbers of the first prefix overflows.
 */
#define PM_ISSN_EAN_PREFIX		"977"
#define PM_ISBN_EAN_PREFIX		"978"
#define PM_ISBN_EAN_PREFIX_OVFLOW	"979"

/**
 * The punctuation that is used in both PM_hyphenate_isbn() and
 * PM_hyphenate_issn().
 */
#define PM_ISBN_HYPHEN  '-'

/**
 * Maximum length of a hyphenated ISBN (or ISSN) or Bookland EAN number created
 * by PM_hyphenate_is{b,s}n(). This is a conservative estimate.
 */
#define PM_ISBN_LEN_MAX 20
/** Supposed length of an ISBN. */
#define PM_ISBN_SIZE 10
/** Supposed length of a "Bookland" EAN number i.e. a 13digit ISBN. */
#define PM_BOOKLAND_EAN_SIZE 13

/* Return value of PM_hyphenate_isbn() PM_hyphenate_issn() */
#define PM_ISBN_ERROR	((size_t)-1)

/* Routines to put hyphens in ISBNs and ISSNs respectively. If the input
   string already contains any hyphens, it will be copied to the output string
   unchanged. */

size_t PM_hyphenate_isbn(const char *in, char *out, size_t outsiz);

size_t PM_hyphenate_issn(const char *in, char *out, size_t outsiz);

/* Routines to convert back and forth from 10-digit ISBNs and 
   13-digit ISBNs (Bookland EAN numbers) */

bool PM_isbn_10_to_13(
	const char *isbn10,
	char       *isbn13,
	size_t      isbn13size);

bool PM_isbn_13_to_10(
	const char *isbn13,
	char       *isbn10,
	size_t      isbn10size);

/* Return value of PM_isni_tetragram() */
#define PM_ISNI_ERROR	((size_t)-1)

/* ISNI functions:
 * PM_isni_tetragram() turns "0000000065440788" into: "0000 0000 6544 0788"
 */
size_t PM_isni_tetragram(
	const char	*in,
	char		*out,
	size_t		outsiz);
    
/**
 * Mode for all block ciphers, either Electronic Code Book, Cipher Block
 * Chaining or Propagating Cipher Block Chaining.
 */
typedef enum {
    PM_DESMODE_ECB,
    PM_DESMODE_CBC,
    PM_DESMODE_PCBC
} PM_DESMode_t;

/** A handle type for DES keys */
typedef const struct PM_DESKey_t *PM_DESKey_p;

/* Create a key for DES. 'multi' is 1 for normal and 3 for triple DES;
   'keystr' must point to 8 * 'multi' bytes; mode is one of the modes mentioned
   above; initval points to 8 bytes with the initial value for (normal or
   propagating) cipher block chaining. Returns handle or NULL on out of
   memory. */
PM_DESKey_p
PM_DESKeyCreate(byte *keystr, int multi, PM_DESMode_t mode,
                byte *initval);

/* Destroy a DES key, no-op if 'key' == NULL. */
void PM_DESKeyDestroy(PM_DESKey_p key);

/* Encode an input of 'inlen' 8-byte blocks, using the key 'key'. 'output'
   should point to 8 * 'inlen' bytes. */
void PM_DESEncode(PM_DESKey_p key, const byte *input, size_t inlen,
                  byte *output);

/* As in PM_DESEncode, but (surprise!) decode. */
void PM_DESDecode(PM_DESKey_p key, const byte *input, size_t inlen,
                  byte *output);


/* Calculate the CRC32 Cyclic Redundancy Check. 'crc' is the crc of any
   previous part of the data, use 0 if this is the first and/or only part. */
uint32 PM_CRC32(uint32 crc, const byte *in, size_t insize);

/* PM_time_ functions (same as in cbstime) */

void PM_time_GetDateTime(char *, size_t, const char *);
void PM_time_MakeCTime(const char *, time_t *, const char *);
void PM_time_CalculateDateTime(const char *, char *, size_t, int,
                               const char *);
bool PM_time_DifferDateTime(const char *, const char *, double *,
                               const char *);

// PM_datetime library 
typedef struct {
    int		dt_year;	// year
    int		dt_mon;		// month of year [1,12]
    int		dt_mday;	// day of month [1,31]
    int		dt_hour;	// hour [0,23]
    int		dt_min;		// minutes [0,59]
    int		dt_sec;		// seconds [0,59]
    long	dt_micro;	// microseconds [0,999999]
} PM_datetime_t, *PM_datetime_p;
void PM_datetime_clear(PM_datetime_p dt);
const char *PM_datetime_scan(PM_datetime_p dt, const char *dtstr,
			     const char *fmt);
    
/* The PM_Base64[En|De]code() do a simple base64 [en|de]coding, which is
 * sometimes used in HTTP. They do not support the 76 chars-per-line limit
 * of MIME. The base64 scheme is explained in RFC 1521.
 */
int PM_Base64Decode(const char *in, byte *out, size_t outsiz);
int PM_Base64Encode(byte *in, int inlen, char *out, size_t outsiz);

/* Prototypes pmsortstring.c */

bool PM_MakeSortString(
        const char *year_sbf,
        const char *month_sbf,
        const char *day_sbf,
        const char *volume_sbf,
        const char *issue_sbf,
        const char *page_sbf,
        char *sortstring,
        const size_t sortstring_size);

/* Prototypes pmuserbits.c */

bool PM_UsedByIlnsAscToBin(
        const char *asc,
        char *bin,
        const size_t binsize,
        int16 *binused);

bool PM_UsedByIlnsBinToAsc(
        const char *bin,
        const int16 binlength,
        char *asc,
        const size_t ascsize);

uint32 PM_bit_count(const byte *bit_buffer, size_t byte_length);
bool PM_userbit_test(const byte *bit_buffer, size_t byte_length,
        uint32 bit_num);

uint32 PM_userbit_next(const byte *bit_buffer, size_t byte_length,
    	uint32 iln);

uint32 PM_userbit_previous(const byte *bit_buffer, size_t byte_length,
    	uint32 iln);

/* Prototype Non-posixy part pmtime.c */

void PM_LBSStamp(char *stamp);

/**
 * PM_TIME_UTC_SIZE_MIN defines the minimum size of a UTC (if no fractions of a
 * second present). Such a UTC has the format "%Y-%m-%dT%H:%M:%SZ"
 * Thus the size of the 'buf' parameter of PM_time_local_to_UTC() must be at
 * least equal to PM_TIME_UTC_SIZE_MIN + the length of the fraction in
 * 'localtime' including the period.
 */
#define PM_TIME_UTC_SIZE_MIN	21
bool PM_time_local_to_UTC(
	const char *localtime,
	const char *options,
	char *buf,
	size_t bufsize);

/* Prototypes for pm_utf.c */
/* Defines of a few important combining characters */
#define PM_UTF32_COMBINING_GRAVE_ACCENT		0x0300
#define PM_UTF32_COMBINING_ACUTE_ACCENT		0x0301
#define PM_UTF32_COMBINING_CIRCUMFLEX_ACCENT	0x0302
#define PM_UTF32_COMBINING_DIAERESIS		0x0308
#define PM_UTF32_COMBINING_CEDILLA		0x0327
#define PM_UTF32_COMBINING_GRAPHEME_JOINER	0x034f
/* Defines of private use characters */
#define PM_UTF32_BIDI_ELT_BOUNDARY		0xe000
#define PM_UTF32_BIDI_ELT_BOUNDARY_LTR		0xe001
#define PM_UTF32_BIDI_ELT_BOUNDARY_RTL		0xe002
#define PM_UTF32_BIDI_LINE_ALIGN_LEFT		0xe003
#define PM_UTF32_BIDI_LINE_ALIGN_LEFT_MONORTL	0xe004
#define PM_UTF32_BIDI_LINE_CONTINUE		0xe005
#define PM_UTF32_NEWLINE			0xe008
#define PM_UTF32_RECORDBUILDER_PPN_EXPANSION  	0xe009

/**
 * The size of a buffer that can contain the utf8 representation of any code
 * point and the terminating 0-byte. Equal to U8_MAX_LENGTH + 1
 */
#define PM_UTF32_CHAR_LEN_MAX  5
const char *PM_Utf32CharToUtf8(int32 u32c, char *dest);

#define PM_UTF32_CHAR_NONE		(0xffff)
#define PM_UTF32_REPLACEMENT_CHARACTER	(0xfffd)
#define PM_HEX_INT32_LEN_MAX		(32 / 4)
#define PM_HEX_CHARREF_LEN_MAX		(3 + PM_HEX_INT32_LEN_MAX + 1 + 1)

int32 PM_Utf32CharRefGet(const char *in, const char **end);
int PM_Utf32CharRefPut(int32 u32, char out[PM_HEX_CHARREF_LEN_MAX]);

// Low level defines copied from ICU library
#define PM_UTF8_IS_SINGLE(c)	(((c)&0x80)==0)
#define PM_UTF8_IS_LEAD(c)	((uint8)((c)-0xc0)<0x3e)
#define PM_UTF8_IS_TRAIL(c)	(((c)&0xc0)==0x80)

/**
 * Macro copy-pinched from ICU marco U8_LENGTH():
 *
 * How many code units (bytes) are used for the UTF-8 encoding
 * of this Unicode code point?
 * @param c 32-bit code point
 * @return 1..4, or 0 if c is a surrogate or not a Unicode code point
 * @stable ICU 2.4
 */
#define PM_UTF8_LENGTH(c) \
    ((uint32)(c)<=0x7f ? 1 : \
        ((uint32)(c)<=0x7ff ? 2 : \
            ((uint32)(c)<=0xd7ff ? 3 : \
                ((uint32)(c)<=0xdfff || (uint32)(c)>0x10ffff ? 0 : \
                    ((uint32)(c)<=0xffff ? 3 : 4)\
                ) \
            ) \
        ) \
    )

/**
 * Macro same as PM_UTF8_LENGTH() but never returning 0
 * Equivalent to 'i' argument of PM_APPEND_UNSAFE() macro
 */
#define PM_UTF8_LENGTH_UNSAFE(c) \
    ((uint32)(c)<=0x7f ? 1 : \
        ((uint32)(c)<=0x7ff ? 2 : \
            ((uint32)(c)<=0xdfff ? 3 : 4) \
        ) \
    )

/**
 * Macro copy-pinched from ICU marco U8_APPEND_UNSAFE():
 *
 * Append a code point to a string, overwriting 1 to 4 bytes.
 * The offset points to the current end of the string contents
 * and is advanced (post-increment).
 *
 * "Unsafe" macro, assumes a valid code point and sufficient space in the
 * string. Otherwise, the result is undefined.
 *
 * @param s const char * string buffer
 * @param i string offset
 * @param c code point to append
 * @draft ICU 2.4
 */
#define PM_APPEND_UNSAFE(s, i, c) { \
    if((uint32)(c)<=0x7f) { \
        (s)[(i)++]=(uint8)(c); \
    } else { \
        if((uint32)(c)<=0x7ff) { \
            (s)[(i)++]=(uint8)(((c)>>6)|0xc0); \
        } else { \
            if((uint32)(c)<=0xffff) { \
                (s)[(i)++]=(uint8)(((c)>>12)|0xe0); \
            } else { \
                (s)[(i)++]=(uint8)(((c)>>18)|0xf0); \
                (s)[(i)++]=(uint8)((((c)>>12)&0x3f)|0x80); \
            } \
            (s)[(i)++]=(uint8)((((c)>>6)&0x3f)|0x80); \
        } \
        (s)[(i)++]=(uint8)(((c)&0x3f)|0x80); \
    } \
}

unsigned int
PM_toupper(
           unsigned int char_in
           );

unsigned int
PM_tolower(
           unsigned int char_in
           );

int
PM_is_ascii_space(
	int	    c
	);

long
PM_atosize(
	const char	*s);

const char *PM_sizetoa(char *string, long value);

long PM_roman2l(const char *s, const char **end);

long PM_month2l(const char *cs, const char **end);

int
PM_version_strcmp(
	const char	*desired_version,	/* Fi. "3.3" */
	const char	*client_version);	/* Fi. "3.3.0.1" */

/**
 * A cluster type is a string that can be the value of the CLUSTER argument
 * of PSI URL's. Two values are special:
 * - Y means: set contains default clusters
 * - N means: set contains records, not clusters (i.e. don't cluster)
 */
#define PM_CLUSTER_TYPE_NONE		"N"
#define PM_CLUSTER_TYPE_DEFAULT		"Y"
bool PM_cluster_type_valid(const char *cluster_type);

const char *PM_cluster_num_phrase(long cluster_num, const char *cluster_type);
const char *PM_cluster_num_phrase_pos(long cluster_num, char wildcard,
				      const char *cluster_type);
long PM_cluster_num_from_phrase(const char *cluster_num_phrase);

// Prototypes pm_math.c

uint32 PM_math_pow2roundup(uint32 x);

uint32 *
PM_simhsh_compute(
	uint32		*minhsh,
	int		minhsh_count,
	const char	*text,
	uint32		flags);


// Prototypes pm_lang.c

const char *
PM_LangToISO639_1(
	const char	*langCbs);
const char *
PM_LangFromISO639_1(
	const char	*langIso);

// Prototypes pm_file.c
char *
PM_file_tmpnam(
	char	*tmpnamBuf);

char *
PM_file_tempnam(
	const char	*directory,
	const char	*prefix);

bool PM_ensurePath(const char *directoryPath, int mode);

#ifdef  __cplusplus
} /* extern "C" */
#ifdef __bool_defined_here
#undef bool
#endif
#endif

#endif /* ! defined(PICAMISC_H_INCLUDED) */
