/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: options.c
 *  Description	: Parsing of Db-options
 *  Author(s)	: Hans van Eck
 *  Creation	: 26-11-1993
 *
 *  History:
 *  26-11-1993	: eck      : creation
 *  29-05-1995	: eck      : Added PM_opt_fetch()
 *  05-11-2001  : paul     : PM_opt_findvalend() introduced
 *  05-06-2009	: martinz  : Added PM_opt_findname() and PM_opt_findnameend().
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *  Notes	:
 *  All the PM_opt_*-routines have as their first argument a pointer into an
 *  options-string.  The functions that return a pointer, will return a pointer
 *  into the options-string passed.  This pointer can then be used as argument
 *  to other PM_opt_*-routines.
 *
 *  - PM_opt_findstart() returns a pointer to the start of an option.  It does
 *    this by skipping spaces an semicolons.  If there is no option, a pointer
 *    to the end of the string is returned.
 *  - PM_opt_equal() test if the option that starts the options-string is equal
 *    to a certain name.
 *  - PM_opt_findval() returns a pointer to the value part of an option.  It
 *    does this by first skipping all characters (except semicolons) to a '='-
 *    character.  After that, all blanks after the '=' are skipped.
 *  - PM_opt_findend() returns as pointer to the end of an option.  It does
 *    this by searching for a semicolon.  If there is none, a pointer to the
 *    end of the string is returned.
 *  - PM_opt_findnext() returns a pointer to the next option.  It is simply a
 *    call to PM_opt_findend() followed by a call to PM_opt_findstart().
 *  - PM_opt_copyval() copies the value part of an option to a buffer.  It will
 *    ensure '\0'-termination of the string. The value returned is the length
 *    copied, not including the '\0'-character at the end (so the length
 *    returned will always be smaller than the maxlen parameter).
 *  - PM_opt_fetch() is a combination of the routines described above : a
 *    specific option is searched for in the option string and copied to an
 *    output buffer. See further the comment at routine PM_opt_copyval(..).
 *
 *  All these routines are fairly simple and assume a well-formed
 *  options-string.
 *========================================================================
 */

#pragma ident "$Id: options.c 339156 2016-11-18 18:26:14Z heesakkm $"

#include <ctype.h>
#include <stddef.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function	: PM_opt_equal
 *  Purpose	: Test if the options in an options-string is equal to a
 *		  specific option-variable
 *  Method	:
 *  Compare over the length of the option-variable. If they match: skip the
 *  blanks in the options-string; '\0', '=' and ';' are valid option-variable
 *  ends
 *
 *  History:
 *  26-11-1993	: eck      : creation
 *
 *  Returns	: indication of match
 *------------------------------------------------------------------------
 */
bool
PM_opt_equal(	const char	*optstring,
		const char	*optname)
{
    int	len = strlen(optname);

    if (strncmp(optstring, optname, len) == 0) {
	/* They are equal in the first 'len' characters */
	while (PM_is_ascii_space(optstring[len])) len++;

	return ((optstring[len] == '\0')
		|| (optstring[len] == '=')
		|| (optstring[len] == ';'));
    } else {
	return FALSE;
    }
} /* PM_opt_equal() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findname
 *  Purpose	: Find the name-part of an option
 *  Method	: Skip to the first non-blank.
 *
 *  History:
 *  05-06-2009	: martinz  : creation
 *
 *  Returns	: pointer to the start of the value (if there is one)
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findname(const char *options)
{
    while ((*options != '\0') && PM_is_ascii_space(*options)) {
	options++;
    }
    return options;
} /* PM_opt_findname() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findval
 *  Purpose	: Find the value-part of an option
 *  Method	: Skip to the first non-blank after an '='
 *
 *  History:
 *  26-11-1993	: eck      : creation
 *
 *  Returns	: pointer to the start of the value (if there is one)
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findval(const char *options)
{
    while ((*options != '\0') && (*options != '=') && (*options != ';')) {
	options++;
    }
    if (*options == '=') {
	options++;
	while (PM_is_ascii_space(*options)) options++;
    }
    return options;
} /* PM_opt_findval() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findstart
 *  Purpose	: Find the start of an option
 *  Method	: skip blanks and ';'-signs
 *
 *  History:
 *  26-11-1993	: eck      : creation
 *
 *  Returns	: pointer to the start of an option (if there is one)
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findstart(const char *options)
{
    if (options != NULL) {
	while (PM_is_ascii_space(*options) || (*options == ';')) options++;
    }
    return options;
} /* PM_opt_findstart() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findend
 *  Purpose	: Find the end of an option in an options-string
 *  Method	: Skip characters until a '\0' or a ';' is found
 *
 *  History:
 *  26-11-1993	: eck      : creation
 *
 *  Returns	: pointer to the end of (the value of) an option
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findend(const char *options)
{
    while ((*options != '\0') && (*options != ';')) {
	options++;
    }
    return options;
} /* PM_opt_findend() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findnameend
 *  Purpose	: Find the end of the name of an option in an options-string
 *  Method	: PM_opt_findval() and skip trailing blanks
 *
 *  History:
 *  05-06-2009	: martinz  : creation
 *
 *  Returns	: pointer to the end of the name of an option
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findnameend(const char *options)
{
    const char *nameend = PM_opt_findval(options);

    while (nameend > options
        && (PM_is_ascii_space(nameend[-1])
	    || nameend[-1] == '=')) {
	nameend--;
    }
    return nameend;
} /* PM_opt_findnameend() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findvalend
 *  Purpose	: Find the end of the value of an option in an options-string
 *  Method	: PM_opt_findend() and skip trailing blanks
 *
 *  History:
 *  05-11-2001	: paul     : creation
 *
 *  Returns	: pointer to the end of the value of an option
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findvalend(const char *options)
{
    const char *valend = PM_opt_findend(options);

    while (valend > options && PM_is_ascii_space(valend[-1])) valend--;
    return valend;
} /* PM_opt_findvalend() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_findnext
 *  Purpose	: Find the next option in the options-string
 *  Method	: Find the end of an option, themn the start of a new one
 *
 *  History:
 *  26-11-1993	: eck      : creation
 *
 *  Returns	: pointer to the start of a next options (if there is one)
 *------------------------------------------------------------------------
 */
const char *
PM_opt_findnext(const char *options)
{
    options = PM_opt_findend(options);
    options = PM_opt_findstart(options);
    return options;
} /* PM_opt_findnext() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_copyval
 *  Purpose	: Copy the value from an options-string to a buffer
 *  Method	: Find the end and copy  (but not more than maxlen-1 chars)
 *  Author(s)	: Hans van Eck
 *  Creation	: 20-01-1994
 *
 *  History:
 *  20-01-1994	: eck      : creation
 *  21-03-1994	: eck      : target[maxlen] was sometimes referenced
 *
 *  Returns	: The length copied (not including the  '\0')
 *------------------------------------------------------------------------
 */
int
PM_opt_copyval(
	const char	*options,	/* the options-string to copy from */
	char		*target,	/* place to put the value */
	size_t		maxlen)		/* size of the target */
{
    const char	*val_end = PM_opt_findend(options);
    ptrdiff_t	len = val_end - options;

    if (len >= maxlen) len = maxlen - 1;
    if (len >= 0) {
	(void) memcpy(target, options, len);
	target[len] = '\0';
    }
    return len;
} /* PM_opt_copyval() */

/*------------------------------------------------------------------------
 *  Function	: PM_opt_fetch
 *  Purpose	: Get an option by name
 *  Method	:
 *
 *  History:
 *  24-05-1995	: eck      : creation
 *  11-06-2013	: paul     : Fixed bug: first option not found if options
 *                           string starts with space
 *
 *  Returns	: The length copied (not including the  '\0')
 *------------------------------------------------------------------------
 */
int
PM_opt_fetch(
	const char	*options,	/* the options-string */
	const char	*name,		/* option to find */
	char		*target,	/* place to put the value */
	size_t		maxlen)		/* size of the target */
{
    *target = '\0';

    /* Skip leading whitespace and semicolons */
    options = PM_opt_findstart(options);

    while ((options != NULL) && (*options != '\0')) {
	if (PM_opt_equal(options, name)) {
	    options = PM_opt_findval(options);
	    return PM_opt_copyval(options, target, maxlen);
	}
	options = PM_opt_findnext(options);
    }
    return 0;			/* Not found */
} /* PM_opt_fetch() */
