/*========================================================================
 *  Copyright (c) 2013 by OCLC
 *
 *  File	: pm_cluster.c
 *  Description	: Miscellaneous cluster related functions and defines
 *  Author(s)	: Paul Smit
 *  Creation	: 06-02-2013
 *
 *  History:
 *  06-02-2013	: paul     : CBS-4317 creation
 *
 *========================================================================
 */

#pragma ident "$Id: pm_cluster.c 339156 2016-11-18 18:26:14Z heesakkm $"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <pica.h>
#include <er.h>

#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function	: PM_cluster_type_valid
 *  Purpose	: Validate a cluster type
 *  Method	:
 *
 *  History:
 *  07-02-2013	: paul     : creation
 *  19-02-2013	: paul     : CBS-4515 Last char of type must be alpha
 *
 *  Notes	:
 *  Cluster type must be able to function as:
 *  - a directory
 *  - part of the key of a CLUSTERSPEC standard table
 *  - part of the cluster number index phrase
 *
 *  Returns	: Whether the cluster type is acceptable
 *------------------------------------------------------------------------
 */
bool
PM_cluster_type_valid(
	const char	*cluster_type)
{
    const char	*p;

    if (cluster_type == NULL) {
	return FALSE;
    }
    for (p = cluster_type; *p != '\0'; p++) {
	if (!isalnum((unsigned char)*p) && *p != '-') {
	    return FALSE;
	}
    }

    /* Last char must be alpha, for PM_cluster_num_from_phrase() to work */
    if (p > cluster_type && !isalpha((unsigned char)p[-1])) {
	return FALSE;
    }

    return TRUE;
} /* PM_cluster_type_valid() */

/*------------------------------------------------------------------------
 *  Function	: PM_cluster_num_from_phrase
 *  Purpose	: Extract the cluster number from its index phrase
 *  Method	:
 *
 *  History:
 *  19-02-2013	: paul     : CBS-4515 creation
 *
 *  Returns	:
 *  The number from the index phrase or -1 if index phrase was incorrect
 *------------------------------------------------------------------------
 */
long
PM_cluster_num_from_phrase(
	const char	*cluster_num_phrase)
{
    const char	*p;

    if (cluster_num_phrase == NULL) {
	return -1;
    }

    /* Scan digits from right to left */
    for (p = cluster_num_phrase + strlen(cluster_num_phrase);
	 p > cluster_num_phrase;
	 p--) {
	if (!isdigit((unsigned char)p[-1])) {
	    break;
	}
    }

    /* Return -1 if phrase didn't end with digits */
    return (*p != '\0') ? atol(p) : -1;
} /* PM_cluster_num_from_phrase() */

/*------------------------------------------------------------------------
 *  Function	: PM_cluster_num_phrase
 *  Purpose	: Print a cluster number index phrase
 *  Method	:
 *
 *  History:
 *  12-02-2013	: paul     : CBS-4510 creation
 *
 *  Returns	:
 *  Pointer to static buffer containing the index phrase or NULL in case of
 *  error
 *------------------------------------------------------------------------
 */
const char *
PM_cluster_num_phrase(
	long		cluster_num,
	const char	*cluster_type)
{
    static const char * const fname = "PM_cluster_num_phrase";

    static char		*phrase_buf = NULL;
    static size_t	phrase_siz = 0;
    size_t		new_phrase_siz;

    if (cluster_type == NULL) {
	cluster_type = PM_CLUSTER_TYPE_DEFAULT;
    }

    new_phrase_siz = strlen(cluster_type) + PM_INTLEN_MAX;

    /* Allocate enough memory for this index phrase */
    if (phrase_siz < new_phrase_siz) {
	voidp	tmpp = realloc((voidp)phrase_buf, new_phrase_siz);

	if (tmpp == NULL) {
	    ERWrite(APPLE_NOMEM, fname, "Can't allocate cluster_phrase");
	    return NULL;
	}
	phrase_buf = (char *)tmpp;
	phrase_siz = new_phrase_siz;
    }

    /* Print the index phrase into the static buffer */
    sprintf(phrase_buf, "%s%ld",
	    STREQ(cluster_type, PM_CLUSTER_TYPE_DEFAULT)
	    ? ""
	    : cluster_type,
	    cluster_num);

    return phrase_buf;
} /* PM_cluster_num_phrase() */

/*------------------------------------------------------------------------
 *  Function	: PM_cluster_num_phrase_pos
 *  Purpose	: Print a cluster number index phrase that can be used in
 *                IDXPhrasePosition()
 *  Method	:
 *
 *  History:
 *  19-02-2013	: paul     : CBS-4515 creation
 *
 *  Returns	: Pointer to cluster number index phrase
 *------------------------------------------------------------------------
 */
const char *
PM_cluster_num_phrase_pos(
	long		cluster_num, 
	char		wildcard,
	const char	*cluster_type)
{
    static const char * const fname = "PM_cluster_num_phrase_pos";

    static char		*phrase_buf = NULL;
    static size_t	phrase_siz = 0;
    size_t		new_phrase_siz;

    if (cluster_type == NULL) {
	cluster_type = PM_CLUSTER_TYPE_DEFAULT;
    }

    new_phrase_siz = strlen(cluster_type) + 1;
    new_phrase_siz += (cluster_num >= 0) ? PM_INTLEN_MAX : 1;

    /* Allocate enough memory for this index phrase */
    if (phrase_siz < new_phrase_siz) {
	voidp	tmpp = realloc((voidp)phrase_buf, new_phrase_siz);

	if (tmpp == NULL) {
	    ERWrite(APPLE_NOMEM, fname, "Can't allocate cluster_phrase");
	    return NULL;
	}
	phrase_buf = (char *)tmpp;
	phrase_siz = new_phrase_siz;
    }

    /* Print the index phrase into the static buffer */
    if (cluster_num >= 0) {
	sprintf(phrase_buf, "%s%ld%c",
		STREQ(cluster_type, PM_CLUSTER_TYPE_DEFAULT)
		? ""
		: cluster_type,
		cluster_num,
		wildcard);
    } else {
	sprintf(phrase_buf, "%s%c",
		STREQ(cluster_type, PM_CLUSTER_TYPE_DEFAULT)
		? ""
		: cluster_type,
		wildcard);
    }

    return phrase_buf;
} /* PM_cluster_num_phrase_pos() */
