/*========================================================================
 *  Copyright (c) 2014 by OCLC
 *
 *  File	: pm_datetime.c
 *  Description	: Code for handling the datetime-structure
 *
 *  History:
 *  14-10-2014	: paul     : CBS-5825 creation
 *
 *  Notes	:
 *  The datetime-structure contains the following members:
 *  year		(Common Era)
 *  month of year	[1,12]
 *  day of month	[1,31]
 *  hour		[0,23]
 *  minutes		[0,59]
 *  seconds		[0,59]
 *  microseconds	[0,999999]
 *
 *========================================================================
 */

#pragma ident "$Id: pm_datetime.c 319408 2014-10-17 10:52:15Z heesakkm $"

#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <pica.h>
#include <picamisc.h>
#include <er.h>

// The PM_datetime_scan format element that matches microseconds
#define PM_DATETIME_FMT_MICRO_strptime	"%:"

// The PM_datetime_scan format element that matches millseconds
#define PM_DATETIME_FMT_MILLI_strptime	"%."

// The sscanf format element that matches microseconds
#define PM_DATETIME_FMT_MICRO_sscanf	"%ld"

// The year offset used by struct tm
#define PM_DATETIME_STRUCT_TM_YEAR_OFFSET	1900

// The month offset used by struct tm
#define PM_DATETIME_STRUCT_TM_MONTH_OFFSET	1

/*------------------------------------------------------------------------
 *  Function	: PM_datetime_clear
 *  Purpose	: Clear a datetime structure
 *
 *  History:
 *  14-10-2014	: paul     : creation
 *
 *  Notes	:
 *  Use zero for unfilled members
 *  Thus, printing a date without time will result in: "00:00:00.000"
 *
 *------------------------------------------------------------------------
 */
void
PM_datetime_clear(
	PM_datetime_p	dt)
{
    dt->dt_year = 0;
    dt->dt_mon = 0;
    dt->dt_mday = 0;
    dt->dt_hour = 0;
    dt->dt_min = 0;
    dt->dt_sec = 0;
    dt->dt_micro = 0;
} // PM_datetime_clear()

/*------------------------------------------------------------------------
 *  Function	: PM_datetime_strptime_fmt
 *  Purpose	: Create the format for strptime()
 *
 *  History:
 *  14-10-2014	: paul     : creation
 *
 *  Description :
 *  The PM_datetime_scan() format is supposed to consist of 2 parts:
 *  1. in front: the strptime() format
 *  2. behind: the format element matching either:
 *     - microseconds "%:" plus trailing characters
 *     - milliseconds "%." plus trailing characters
 *
 *  This function duplicates the first part.
 *  The second part may be missing.
 *
 *  Example     :
 *  PM_datetime_scan() format:
 *  "%Y-%m-%d %H:%M:%S.%.Z"
 *
 *  Return value:
 *  "%Y-%m-%d %H:%M:%S.Z"
 *
 *  Returns	: The strptime() format
 *                or NULL && APPLE_NOMEM in case of memory failure
 *                or NULL && APPLE_PARAM if illegal PM_datetime_scan() format
 *
 *------------------------------------------------------------------------
 */
static char *
PM_datetime_strptime_fmt(
	const char	*fmt,		// in: PM_datetime_scan() format
	const char	**msec_fmt)	/* out: pointer to either "%:" or "%."
					   or to end of 'fmt' if missing */
{
    char		*strptime_fmt = NULL;	// return value

    APPLerrno = APPLE_NOERROR;

    // Find the first occurrence of both "%." and "%:" in fmt
    if (APPLerrno == APPLE_NOERROR) {
	const char	*p, *q;
	
	p = strstr(fmt, PM_DATETIME_FMT_MICRO_strptime);
	q = strstr(fmt, PM_DATETIME_FMT_MILLI_strptime);
	if (p == NULL && q == NULL) {
	    *msec_fmt = fmt + strlen(fmt);
	} else if (p == NULL) {
	    *msec_fmt = q;
	} else if (q == NULL) {
	    *msec_fmt = p;
	} else if (p < q) {
	    *msec_fmt = p;
	} else {
	    *msec_fmt = q;
	}
    }

    // Check that "%" doesn't occur behind "%."
    if (APPLerrno == APPLE_NOERROR && **msec_fmt != '\0') {
	const char	*percent = *msec_fmt;

	while ((percent = strchr(percent + 1, '%')) != NULL) {
	    if (percent[1] != '%') {
		APPLerrno = APPLE_PARAM;
		ERWrite(APPLerrno, __func__, "Illegal format: \"%s\"", fmt);
	    }
	    percent++;
	}
    }

    // Create the format used by strptime()
    if (APPLerrno == APPLE_NOERROR) {
	strptime_fmt = PM_strdup(fmt, *msec_fmt - fmt);
	if (strptime_fmt == NULL) {
	    APPLerrno = APPLE_NOMEM;
	    ERWrite(APPLerrno, __func__, "Can't duplicate strptime_fmt");
	}
    }

    if (APPLerrno != APPLE_NOERROR) {
	free((voidp)strptime_fmt);
	return NULL;
    }
    
    return strptime_fmt;
} // PM_datetime_strptime_fmt()
	
/*------------------------------------------------------------------------
 *  Function	: PM_datetime_sscanf_fmt
 *  Purpose	: Create the format for sscanf to scan microseconds/milliseconds
 *
 *  History:
 *  14-10-2014	: paul     : creation
 *
 *  Example     :
 *  PM_datetime_scan() format:
 *  "%Y-%m-%d %H:%M:%S.%.Z"
 *
 *  Return value:
 *  "%ldZ%n"
 *
 *  Returns	: The sscanf() format
 *                or NULL && APPLE_NOMEM in case of memory failure
 *
 *------------------------------------------------------------------------
 */
static char *
PM_datetime_sscanf_fmt(
	const char	*msec_fmt)	/* in: pointer to either "%:" or "%." */
{
    char		*sscanf_fmt = NULL;	// return value
    size_t		sscanf_fmt_siz;		// size of return value

    APPLerrno = APPLE_NOERROR;

    // Create the buffer for the sscanf() format
    if (APPLerrno == APPLE_NOERROR) {
	sscanf_fmt_siz = (strlen(PM_DATETIME_FMT_MICRO_sscanf)
			  + strlen(msec_fmt)
			  + strlen("%n")
			  + 1);
	sscanf_fmt = malloc(sscanf_fmt_siz);
	if (sscanf_fmt == NULL) {
	    APPLerrno = APPLE_NOMEM;
	    ERWrite(APPLerrno, __func__, "Can't malloc sscanf_fmt");
	}
    }

    // Fill the buffer with the sscanf() format
    if (APPLerrno == APPLE_NOERROR) {
	sprintf(sscanf_fmt, "%s%s%s",
		PM_DATETIME_FMT_MICRO_sscanf,
		msec_fmt + strlen(PM_DATETIME_FMT_MICRO_strptime),
		"%n");
    }

    if (APPLerrno != APPLE_NOERROR) {
	free((voidp)sscanf_fmt);
	return NULL;
    }
    
    return sscanf_fmt;
} // PM_datetime_sscanf_fmt()

/*------------------------------------------------------------------------
 *  Function	: PM_datetime_scan
 *  Purpose	: Scan a datetime-string and put values in datetime-struct
 *
 *  History:
 *  14-10-2014	: paul     : creation
 *
 *  Notes	:
 *  PM_datetime_scan() will set unscanned members of the datetime-struct to zero.
 *
 *  PM_datetime_scan() will validate the output date only superficially.
 *  Day 31 of February will be accepted; day 32 will cause a NULL return.
 *
 *  Description :
 *  %y matches a 2 digit year:
 *  "69" refers to 1969
 *  "68" refers to 2068
 *  See strptime() for all details.
 *
 *  Note that PM_datetime_scan() format may contain conversion specifications
 *  such as %a (day of week) in which case a corresponding day name needs to be
 *  present in the string to scan, but its value is not part of the
 *  datetime-struct.
 *
 *  The following conversion specifications are supported in addition to those
 *  in strptime():
 *  "%." matches milliseconds
 *  "%:" matches microseconds
 *  These must occur as the last specification in the format.
 *
 *  Example     :
 *  fmt:   %d-%m-%y %H:%M:%S.%.X
 *  dtstr: 31-02-68 23:59:59.789XYZ
 *  dt:    2068-02-31 23:59:59.789000
 *  return: pointer to the "YZ" substring of dtstr
 *
 *  fmt:   %Y-%m-%dT%H:%M:%S.%:
 *  dtstr: 1968-02-31T23:59:59.456789
 *  dt:    1968-02-31 23:59:59.456789
 *
 *  fmt:   %a%d%b%y%H%p%z
 *  dtstr: Mon13Oct6908AM+0200
 *  dt:    1969-10-13 00:00:00.000000
 *
 *  Returns	: a pointer to the character following the last character parsed
 *                or NULL && APPLE_NOERROR in case format didn't match input
 *                or NULL && APPLE_NOMEM in case of memory failure
 *                or NULL && APPLE_PARAM if illegal PM_datetime_scan() format
 *
 *------------------------------------------------------------------------
 */
const char *
PM_datetime_scan(
	PM_datetime_p	dt,	// the scanned values
	const char	*dtstr,	// the string to scan f.e: 31-02-70 23:59:59.000
	const char	*fmt)	// the format         f.e: %d-%m-%y %H:%M:%S.%.
{
    char		*strptime_fmt = NULL;
    char		*sscanf_fmt = NULL;
    struct tm		tm_struct;
    const char		*msec_fmt;	// either micro- or milli-seconds-fmt
    const char		*end = NULL;
    int			scanlen = 0;
    unsigned long	micro = 0;
    bool		ok = TRUE;

    APPLerrno = APPLE_NOERROR;

    if (dt == NULL) {
	return NULL;
    }
    
    PM_datetime_clear(dt);

    // Create the format for strptime()
    ok = ok && (strptime_fmt = PM_datetime_strptime_fmt(fmt, &msec_fmt)) != NULL;

    // Scan all date elements in the string except microseconds
    if (ok) {
	// Init struct tm because strptime() may not do it
	memset(&tm_struct, '\0', sizeof(tm_struct));

	// Unscanned year results in a datetime-struct year of: zero
	tm_struct.tm_year = -PM_DATETIME_STRUCT_TM_YEAR_OFFSET;

	// Unscanned year results in a datetime-struct month of: zero
	tm_struct.tm_mon = -PM_DATETIME_STRUCT_TM_MONTH_OFFSET;

	end = strptime(dtstr, strptime_fmt, &tm_struct);
	ok = end != NULL;
    }

    // Create the format used by sscanf(), if "%." or "%:" supplied
    if (ok && *msec_fmt != '\0') {
	ok = (sscanf_fmt = PM_datetime_sscanf_fmt(msec_fmt)) != NULL;
    }

    // Scan microseconds, if "%." or "%:" supplied
    if (ok && *msec_fmt != '\0') {
	ok = sscanf(end, sscanf_fmt, &micro, &scanlen) == 1;
    }

    // number of microseconds is 1000 times number of milliseconds
    if (ok && *msec_fmt != '\0') {
	if (strncmp(msec_fmt, PM_DATETIME_FMT_MILLI_strptime,
		    strlen(PM_DATETIME_FMT_MILLI_strptime)) == 0) {
	    
	    micro *= 1000;
	}
    }

    // Fill return value
    if (ok) {
	dt->dt_year  = tm_struct.tm_year + PM_DATETIME_STRUCT_TM_YEAR_OFFSET;
	dt->dt_mon   = tm_struct.tm_mon + PM_DATETIME_STRUCT_TM_MONTH_OFFSET;
	dt->dt_mday  = tm_struct.tm_mday;
	dt->dt_hour  = tm_struct.tm_hour;
	dt->dt_min   = tm_struct.tm_min;
	dt->dt_sec   = tm_struct.tm_sec;
	dt->dt_micro = micro;
    }

    // Prevent memory leaks
    free((voidp)strptime_fmt);
    free((voidp)sscanf_fmt);

    if (!ok) {
	return NULL;
    }

    return end + scanlen;
} // PM_datetime_scan()

