/*========================================================================
 *  Copyright (c) 1998 by Pica
 *
 *  File	: pm_des.c
 *  Description	: DES encryption and decryption routines
 *  Author(s)	: Bas de Bakker
 *  Creation	: 07-01-1998
 *
 *  History:
 *  07-01-1998	: BdB      : creation
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *  Notes	:
 *  Much of this code has been based on the public domain DES implementation by
 *  Phil Karn. You can find his web page at http://people.qualcomm.com/karn/,
 *  but because of ridiculous US export restrictions he cannot put the code
 *  there. Instead, I found it in
 *  ftp://ftp.funet.fi/pub/crypt/cryptography/symmetric/des
 *
 *========================================================================
 */

#pragma ident "$Id: pm_des.c 362304 2018-11-27 14:08:45Z heesakkm $"

#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>

/* A handle for single DES keys. */
typedef struct {
    byte kn[16][8];
} deskey_t;

/* A handle for multiple DES keys and encryption mode info. */
typedef struct PM_DESKey_t {
    int multi;
    PM_DESMode_t mode;
    byte initval[8];		/* Unused in ECB mode. */
    deskey_t *keys;		/* Points to 'multi' instances of deskey_t. */
} PM_DESKey_t;


/* Whether the static tables below have been initialized. */
static bool initialized = FALSE;

/* Some static tables with data that is only set at initialization time. */
static uint32 sp[8][64];
static byte iperm[16][16][8];
static byte fperm[16][16][8];


/* Tables defined in the Data Encryption Standard documents */

/* initial permutation IP */
static const byte ip[] = {
	58, 50, 42, 34, 26, 18, 10,  2,
	60, 52, 44, 36, 28, 20, 12,  4,
	62, 54, 46, 38, 30, 22, 14,  6,
	64, 56, 48, 40, 32, 24, 16,  8,
	57, 49, 41, 33, 25, 17,  9,  1,
	59, 51, 43, 35, 27, 19, 11,  3,
	61, 53, 45, 37, 29, 21, 13,  5,
	63, 55, 47, 39, 31, 23, 15,  7
};

/* final permutation IP^-1 */
static const byte fp[] = {
	40,  8, 48, 16, 56, 24, 64, 32,
	39,  7, 47, 15, 55, 23, 63, 31,
	38,  6, 46, 14, 54, 22, 62, 30,
	37,  5, 45, 13, 53, 21, 61, 29,
	36,  4, 44, 12, 52, 20, 60, 28,
	35,  3, 43, 11, 51, 19, 59, 27,
	34,  2, 42, 10, 50, 18, 58, 26,
	33,  1, 41,  9, 49, 17, 57, 25
};

/* permuted choice table (key) */
static const byte pc1[] = {
	57, 49, 41, 33, 25, 17,  9,
	 1, 58, 50, 42, 34, 26, 18,
	10,  2, 59, 51, 43, 35, 27,
	19, 11,  3, 60, 52, 44, 36,
	63, 55, 47, 39, 31, 23, 15,
	 7, 62, 54, 46, 38, 30, 22,
	14,  6, 61, 53, 45, 37, 29,
	21, 13,  5, 28, 20, 12,  4
};

/* number left rotations of pc1 */
static const byte totrot[] = {
	1,2,4,6,8,10,12,14,15,17,19,21,23,25,27,28
};

/* permuted choice key (table) */
static const byte pc2[] = {
	14, 17, 11, 24,  1,  5,
	 3, 28, 15,  6, 21, 10,
	23, 19, 12,  4, 26,  8,
	16,  7, 27, 20, 13,  2,
	41, 52, 31, 37, 47, 55,
	30, 40, 51, 45, 33, 48,
	44, 49, 39, 56, 34, 53,
	46, 42, 50, 36, 29, 32
};

/* The (in)famous S-boxes */
static const byte si[8][64] = {
    {	/* S1 */
	14,  4, 13,  1,  2, 15, 11,  8,  3, 10,  6, 12,  5,  9,  0,  7,
	 0, 15,  7,  4, 14,  2, 13,  1, 10,  6, 12, 11,  9,  5,  3,  8,
	 4,  1, 14,  8, 13,  6,  2, 11, 15, 12,  9,  7,  3, 10,  5,  0,
	15, 12,  8,  2,  4,  9,  1,  7,  5, 11,  3, 14, 10,  0,  6, 13
    },{
	/* S2 */
	15,  1,  8, 14,  6, 11,  3,  4,  9,  7,  2, 13, 12,  0,  5, 10,
	 3, 13,  4,  7, 15,  2,  8, 14, 12,  0,  1, 10,  6,  9, 11,  5,
	 0, 14,  7, 11, 10,  4, 13,  1,  5,  8, 12,  6,  9,  3,  2, 15,
	13,  8, 10,  1,  3, 15,  4,  2, 11,  6,  7, 12,  0,  5, 14,  9
    },{
	/* S3 */
	10,  0,  9, 14,  6,  3, 15,  5,  1, 13, 12,  7, 11,  4,  2,  8,
	13,  7,  0,  9,  3,  4,  6, 10,  2,  8,  5, 14, 12, 11, 15,  1,
	13,  6,  4,  9,  8, 15,  3,  0, 11,  1,  2, 12,  5, 10, 14,  7,
	 1, 10, 13,  0,  6,  9,  8,  7,  4, 15, 14,  3, 11,  5,  2, 12
    },{
	/* S4 */
	 7, 13, 14,  3,  0,  6,  9, 10,  1,  2,  8,  5, 11, 12,  4, 15,
	13,  8, 11,  5,  6, 15,  0,  3,  4,  7,  2, 12,  1, 10, 14,  9,
	10,  6,  9,  0, 12, 11,  7, 13, 15,  1,  3, 14,  5,  2,  8,  4,
	 3, 15,  0,  6, 10,  1, 13,  8,  9,  4,  5, 11, 12,  7,  2, 14
    },{
	/* S5 */
	 2, 12,  4,  1,  7, 10, 11,  6,  8,  5,  3, 15, 13,  0, 14,  9,
	14, 11,  2, 12,  4,  7, 13,  1,  5,  0, 15, 10,  3,  9,  8,  6,
	 4,  2,  1, 11, 10, 13,  7,  8, 15,  9, 12,  5,  6,  3,  0, 14,
	11,  8, 12,  7,  1, 14,  2, 13,  6, 15,  0,  9, 10,  4,  5,  3
    },{
	/* S6 */
	12,  1, 10, 15,  9,  2,  6,  8,  0, 13,  3,  4, 14,  7,  5, 11,
	10, 15,  4,  2,  7, 12,  9,  5,  6,  1, 13, 14,  0, 11,  3,  8,
	 9, 14, 15,  5,  2,  8, 12,  3,  7,  0,  4, 10,  1, 13, 11,  6,
	 4,  3,  2, 12,  9,  5, 15, 10, 11, 14,  1,  7,  6,  0,  8, 13
    },{
	/* S7 */
	 4, 11,  2, 14, 15,  0,  8, 13,  3, 12,  9,  7,  5, 10,  6,  1,
	13,  0, 11,  7,  4,  9,  1, 10, 14,  3,  5, 12,  2, 15,  8,  6,
	 1,  4, 11, 13, 12,  3,  7, 14, 10, 15,  6,  8,  0,  5,  9,  2,
	 6, 11, 13,  8,  1,  4, 10,  7,  9,  5,  0, 15, 14,  2,  3, 12
    },{
	/* S8 */
	13,  2,  8,  4,  6, 15, 11,  1, 10,  9,  3, 14,  5,  0, 12,  7,
	 1, 15, 13,  8, 10,  3,  7,  4, 12,  5,  6, 11,  0, 14,  9,  2,
	 7, 11,  4,  1,  9, 12, 14,  2,  0,  6, 10, 13, 15,  3,  5,  8,
	 2,  1, 14,  7,  4, 10,  8, 13, 15, 12,  9,  0,  3,  5,  6, 11
    }
};

/* 32-bit permutation function P used on the output of the S-boxes */
static const byte p32i[] = {	
	16,  7, 20, 21,
	29, 12, 28, 17,
	 1, 15, 23, 26,
	 5, 18, 31, 10,
	 2,  8, 24, 14,
	32, 27,  3,  9,
	19, 13, 30,  6,
	22, 11,  4, 25
};

/* End of DES-defined tables */


/* bit 0 is left-most in byte */
static int bytebit[] = {
	0200,0100,040,020,010,04,02,01
};

static int nibblebit[] = {
	 010,04,02,01
};


/* initialize a perm array */
static void
perminit(byte (*perm)[16][8], const byte *p)
{
    int l, j, k;
    int i, m;

    /* Clear the permutation array */
    memset(perm, 0, 16 * 16 * 8);

    for (i=0; i<16; i++)	/* each input nibble position */
	for (j = 0; j < 16; j++) /* each possible input nibble */
	    for (k = 0; k < 64; k++) { /* each output bit position */
		l = p[k] - 1;	/* where does this bit come from*/
		if ((l >> 2) != i) /* does it come from input posn?*/
		    continue;	/* if not, bit k is 0	 */
		if (!(j & nibblebit[l & 3]))
		    continue;	/* any such bit in input? */
		m = k & 07;	/* which bit is this in the byte*/
		perm[i][j][k>>3] |= bytebit[m];
	    }
} /* perminit() */

/* Initialize the lookup table for the combined S and P boxes */
static void
spinit(void)
{
    byte pbox[32];
    int p, i, s, j, rowcol;
    uint32 val;

    /* Compute pbox, the inverse of p32i.
     * This is easier to work with
     */
    for (p=0; p<32; p++) {
	for (i=0; i<32; i++) {
	    if (p32i[i] - 1 == p) {
		pbox[p] = (byte)i;
		break;
	    }
	}
    }

    for (s = 0; s < 8; s++) {	/* For each S-box */
	for (i=0; i<64; i++) {	/* For each possible input */
	    val = 0;
	    /* The row number is formed from the first and last
	     * bits; the column number is from the middle 4
	     */
	    rowcol = (i & 32) | ((i & 1) ? 16 : 0) | ((i >> 1) & 0xf);
	    for (j=0; j<4; j++) { /* For each output bit */
		if (si[s][rowcol] & (8 >> j)) {
		    val |= 1L << (31 - pbox[4 * s + j]);
		}
	    }
	    sp[s][i] = val;
	}
    }
} /* spinit() */

/* Set key (initialize key schedule array) */
static void
single_key(byte (*kn)[8], const byte *key)
{
    byte pc1m[56];		/* place to modify pc1 into */
    byte pcr[56];		/* place to rotate pc1 into */
    int i, j, l;
    int m;

    /* Clear key schedule */
    memset(kn, 0, 16 * 8);

    for (j=0; j<56; j++) {	/* convert pc1 to bits of key */
	l = pc1[j] - 1;		/* integer bit location	 */
	m = l & 07;		/* find bit		 */
	pc1m[j] = (key[l>>3] &	/* find which key byte l is in */
		   bytebit[m])	/* and which bit of that byte */
	    ? 1 : 0;		/* and store 1-bit result */
    }
    for (i=0; i<16; i++) {	/* key chunk for each iteration */
	for (j=0; j<56; j++)	/* rotate pc1 the right amount */
	    pcr[j] = pc1m[(l = j+totrot[i]) < (j < 28 ? 28 : 56) ? l : l-28];
	/* rotate left and right halves independently */
	for (j=0; j<48; j++){	/* select bits individually */
	    /* check bit that goes to kn[j] */
	    if (pcr[pc2[j]-1]){
		/* mask it in if it's there */
		l= j % 6;
		kn[i][j/6] |= bytebit[l] >> 2;
	    }
	}
    }
} /* single_key() */

/* Permute inblock with perm */
static void
permute(byte *inblock, byte (*perm)[16][8], byte *outblock)
{
    byte *ib, *ob;		/* ptr to input or output block */
    byte *p, *q;
    int j;

    /* Clear output block */
    memset(outblock, 0, 8);

    ib = inblock;
    for (j = 0; j < 16; j += 2, ib++) { /* for each input nibble */
	ob = outblock;
	p = perm[j][(*ib >> 4) & 0xf];
	q = perm[j + 1][*ib & 0xf];
	/* and each output byte, OR the masks together */
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
	*ob++ |= *p++ | *q++;
    }
} /* permute() */


/* The nonlinear function f(r,k), the heart of DES */
static uint32
f(uint32 r, byte *subkey)
{
    uint32 *spp;
    uint32 rval, rt;
    uint32 er;

    /* Run E(R) ^ K through the combined S & P boxes.
     * This code takes advantage of a convenient regularity in
     * E, namely that each group of 6 bits in E(R) feeding
     * a single S-box is a contiguous segment of R.
     */
    subkey += 7;

    /* Compute E(R) for each block of 6 bits, and run thru boxes */
    er = (r << 1) | ((r & 0x80000000) ? 1 : 0);
    spp = &sp[7][0];
    rval = spp[(er ^ *subkey--) & 0x3f];
    spp -= 64;
    rt = r >> 3;
    rval |= spp[(rt ^ *subkey--) & 0x3f];
    spp -= 64;
    rt >>= 4;
    rval |= spp[(rt ^ *subkey--) & 0x3f];
    spp -= 64;
    rt >>= 4;
    rval |= spp[(rt ^ *subkey--) & 0x3f];
    spp -= 64;
    rt >>= 4;
    rval |= spp[(rt ^ *subkey--) & 0x3f];
    spp -= 64;
    rt >>= 4;
    rval |= spp[(rt ^ *subkey--) & 0x3f];
    spp -= 64;
    rt >>= 4;
    rval |= spp[(rt ^ *subkey--) & 0x3f];
    spp -= 64;
    rt >>= 4;
    rt |= (r & 1) << 5;
    rval |= spp[(rt ^ *subkey) & 0x3f];
    return rval;
} /* f() */


/* The original code did this somewhat faster, but at the cost of endian-ness
   #ifdef's. This is more portable. (BdB) */
static uint32
from_bigendian(byte *b)
{
    return (b[0] << 24) + (b[1] << 16) + (b[2] << 8) + b[3];
} /* from_bigendian() */

static void
to_bigendian(uint32 u, byte *b)
{
    b[0] = u >> 24;
    b[1] = u >> 16;
    b[2] = u >> 8;
    b[3] = (byte) u;
} /* to_bigendian() */

/* In-place encryption of 64-bit block */
static void
encrypt_block(byte (*kn)[8], byte *block)
{
    uint32 left, right;
    byte *knp;
    byte work[8];

    permute(block, iperm, work); /* Initial Permutation */

    left = from_bigendian(work);
    right = from_bigendian(work + 4);

    /* Do the 16 rounds.
     * The rounds are numbered from 0 to 15. On even rounds
     * the right half is fed to f() and the result exclusive-ORs
     * the left half; on odd rounds the reverse is done.
     */
    knp = &kn[0][0];
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);
    knp += 8;
    left ^= f(right,knp);
    knp += 8;
    right ^= f(left,knp);

    /* Left/right half swap */
    to_bigendian(right, work);
    to_bigendian(left, work + 4);

    permute(work, fperm, block); /* Inverse initial permutation */
} /* encrypt_block() */

/* In-place decryption of 64-bit block. This function is the mirror
 * image of encryption; exactly the same steps are taken, but in
 * reverse order
 */
static void
decrypt_block(byte (*kn)[8], byte *block)
{
    uint32 left, right;
    byte *knp;
    byte work[8];		/* Working data storage */

    permute(block, iperm, work); /* Initial permutation */

    /* Left/right half swap*/
    right = from_bigendian(work);
    left = from_bigendian(work + 4);

    /* Do the 16 rounds in reverse order.
     * The rounds are numbered from 15 to 0. On even rounds
     * the right half is fed to f() and the result exclusive-ORs
     * the left half; on odd rounds the reverse is done.
     */
    knp = &kn[15][0];
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);
    knp -= 8;
    right ^= f(left,knp);
    knp -= 8;
    left ^= f(right,knp);

    to_bigendian(left, work);
    to_bigendian(right, work + 4);

    permute(work, fperm, block); /* Inverse initial permutation */
} /* decrypt_block() */

/*------------------------------------------------------------------------
 *  Function	: PM_DESKeyCreate
 *  Purpose	: Create key structure for DES encryption
 *  Method	:
 *  Author(s)	: Bas de Bakker
 *  Creation	: 07-01-1998
 *
 *  History:
 *  07-01-1998	: BdB      : creation
 *
 *  Notes	:
 *  - 'multi' is 1 for normal and 3 for triple DES;
 *  - 'keystr' must point to 8 * 'multi' bytes;
 *  - 'mode' is one of the modes mentioned above;
 *  - 'initval' points to 8 bytes with the initial value for (normal or
 *    propagating) cipher block chaining.
 *
 *  Description	:
 *  This function creates a key from a key string.
 *
 *  'keystr' is an array of 8 * multi bytes. Each set of 8 bytes is used for
 *  one of the iterations of the encryption. The array can contain null bytes.
 *
 *  'multi' is the number of times to apply DES encryption. Usual values are 1
 *  for normal DES and 3 for so-called triple DES. The name "triple DES" is
 *  also used for the sequence encode/decode/encode, which is not explicitly
 *  supported by these routines.
 *
 *  'mode' is one of the following:
 *  - PM_DESMODE_ECB: Electronic Code Book (encode each 8 byte block
 *    separately).
 *  - PM_DESMODE_CBC: Cipher Block Chaining (xor each block with the encrypted
 *    previous block).
 *  - PM_DESMODE_PCBC: Propagating Cipher Block Chaining (xor each block with
 *    both plain text and cipher text of previous block).
 *
 *  'initval' is NULL or points to 8 bytes. If not NULL, it is the value to xor
 *  the first block with for the Cipher Block Chaining modes. If NULL, all bits
 *  0 is used. This parameter is ignored in PM_DESMODE_ECB.
 *
 *  Returns	: Handle to key or NULL on out of memory
 *------------------------------------------------------------------------
 */
PM_DESKey_p
PM_DESKeyCreate(byte *keystr, int multi, PM_DESMode_t mode,
		byte *initval)
{
    PM_DESKey_t *key;
    int i;

    if (initialized == FALSE) {
	spinit();
	perminit(iperm,ip);
	perminit(fperm,fp);
	initialized = TRUE;
    }

    key = malloc(sizeof *key);
    if (key == NULL) {
	return NULL;
    }
    key->keys = malloc(multi * sizeof *key->keys);
    if (key->keys == NULL) {
	free(key);
	return NULL;
    }

    key->multi = multi;
    key->mode = mode;
    if (initval == NULL) {
	memset(key->initval, 0, 8);
    } else {
	memcpy(key->initval, initval, 8);
    }

    for (i = 0; i < multi; ++i) {
	single_key(key->keys[i].kn, keystr + 8 * i);
    }

    return key;
} /* PM_DESKeyCreate() */

/*------------------------------------------------------------------------
 *  Function	: PM_DESKeyDestroy
 *  Purpose	: Destroy a key created by PM_DESKeyCreate()
 *  Method	: free() associated memory
 *  Author(s)	: Bas de Bakker
 *  Creation	: 07-01-1998
 *
 *  History:
 *  07-01-1998	: BdB      : creation
 *
 *  Notes	: No-op if 'key' == NULL.
 *
 *  Description	:
 *  PM_DESKeyDestroy destroys a key previously created with PM_DESKeyCreate by
 *  freeing the associated memory. If 'key' is NULL, nothing happens. No
 *  value is returned and no errors can occur.
 *------------------------------------------------------------------------
 */
void
PM_DESKeyDestroy(PM_DESKey_p key) /* Points to a DES key object */
{
    if (key != NULL) {
	free(key->keys);
    }
    free((void *)key);
} /* PM_DESKeyDestroy() */

/*------------------------------------------------------------------------
 *  Function	: xor_block
 *  Purpose	: xor first block of 8 bytes with second
 *  Method	:
 *  Author(s)	: Bas de Bakker
 *  Creation	: 07-01-1998
 *
 *  History:
 *  07-01-1998	: BdB      : creation
 *
 *------------------------------------------------------------------------
 */
static void
xor_block(byte *dest, const byte *src)
{
    int i;

    for (i = 0; i < 8; ++i) {
	dest[i] ^= src[i];
    }
} /* xor_block() */

/*------------------------------------------------------------------------
 *  Function	: PM_DESEncode
 *  Purpose	: Encode an input of 'inlen' 8-byte blocks, using 'key'.
 *  Method	:
 *  Author(s)	: Bas de Bakker
 *  Creation	: 07-01-1998
 *
 *  History:
 *  07-01-1998	: BdB      : creation
 *
 *  Notes	:
 *  'output' should point to 8 * 'inlen' bytes.
 *
 *  Because I mainly wrote wrappers around the existing code, the loop over
 *  multi currently does the initial and final permutation each time (within
 *  encrypt_block), while once would suffice. This is certainly the first thing
 *  to look at if you want to increase the performance.
 *------------------------------------------------------------------------
 */
void
PM_DESEncode(PM_DESKey_p key, const byte *input, size_t inlen,
	     byte *output)
{
    int i;

    for (i = 0; i < inlen; ++i) {
	const byte *const src = input + 8 * i;
	byte *const dest = output + 8 * i;
	int j;

	memcpy(dest, src, 8);

	switch (key->mode) {
	case PM_DESMODE_ECB:
	    /* empty */
	    break;
	case PM_DESMODE_PCBC:
	    if (i > 0) {
		xor_block(dest, src - 8);
	    }
	    /* FALLTHROUGH */
	case PM_DESMODE_CBC:
	    xor_block(dest, i > 0 ? dest - 8 : key->initval);
	    break;
	}

	for (j = 0; j < key->multi; ++j) {
	    encrypt_block(key->keys[j].kn, dest);
	}
    }
} /* PM_DESEncode() */

/*------------------------------------------------------------------------
 *  Function	: PM_DESDecode
 *  Purpose	: Decode an input of 'inlen' 8-byte blocks, using 'key'.
 *  Method	:
 *  Author(s)	: Bas de Bakker
 *  Creation	: 07-01-1998
 *
 *  History:
 *  07-01-1998	: BdB      : creation
 *
 *  Notes	: 'output' should point to 8 * 'inlen' bytes.
 *------------------------------------------------------------------------
 */
void
PM_DESDecode(PM_DESKey_p key, const byte *input, size_t inlen,
	     byte *output)
{
    int i;

    for (i = 0; i < inlen; ++i) {
	const byte *const src = input + 8 * i;
	byte *const dest = output + 8 * i;
	int j;

	memcpy(dest, src, 8);

	for (j = key->multi - 1; j >= 0; --j) {
	    decrypt_block(key->keys[j].kn, dest);
	}

	switch (key->mode) {
	case PM_DESMODE_ECB:
	    /* empty */
	    break;
	case PM_DESMODE_PCBC:
	    if (i > 0) {
		xor_block(dest, dest - 8);
	    }
	    /* FALLTHROUGH */
	case PM_DESMODE_CBC:
	    xor_block(dest, i > 0 ? src - 8 : key->initval);
	    break;
	}	
    }
} /* PM_DESDecode() */
