/*==============================================================================================================================
 *  Copyright (c) 2020 by OCLC
 *
 *  File	: pm_file.c
 *  Description	: Miscellaneous file & directory functions
 *  Author(s)	: Maarten Heesakkers
 *  Creation	: 05-02-2020
 *
 *  History:
 *  05-02-2020	: heesakkm : CBS-9124 creation
 *  16-07-2020	: vaneckh  : Added PM_ensurePath()
 *
 *  Notes	:
 *
 *==============================================================================================================================
 */

#pragma ident "$Id: pm_file.c 378227 2020-07-16 15:44:23Z vaneckh $"

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <pica.h>
#include <picamisc.h>
#include <env.h>
#include <er.h>

#define PM_FILE_FILE_PATTERN		"XXXXXX"	// last six characters of the filename template as required by mkstemp()
#define PM_FILE_DEFAULT_PREFIX		"file"		// default filename prefix, backward compatible with tmpnam()/tempnam()
#define PM_FILE_MAX_PREFIX_LEN		5		// maximum filename prefix length, backward compatible with tmpnam()/tempnam()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: IsValidTempDirectory
 *  Purpose	: Test if dirName is a directory that is writable by the current user
 *  Method	: Get the file status with stat(), check if it is a directory with S_ISDIR() and user writable with access()
 *
 *  History:
 *  05-02-2020	: heesakkm : creation
 *
 *  Returns	: Pointer to an existing directory that is writable by the current user or NULL
 *------------------------------------------------------------------------------------------------------------------------------
 */
static const char *
IsValidTempDirectory(
	const char	*dirName,
	struct stat	*statbuf)	// OUT
{
    if ((dirName == NULL) || (*dirName == '\0')) {	// dirName is optional, so no error if it's missing
	return NULL;
    }
    if (stat(dirName, statbuf) == -1) {			// Retrieve information about dirName
	APPLerrno = APPLE_FILEOPEN;
	ERWrite(APPLerrno, __func__, "stat: %s: %s", dirName, strerror(errno));
    } else if (S_ISDIR(statbuf->st_mode) == 0) {	// Test if dirName is a directory
	APPLerrno = APPLE_PARAM;
	ERWrite(APPLerrno, __func__, "\"%s\" is not a directory!", dirName);
    } else if (access(dirName, W_OK) == -1) {		// Test if directory is writable by current user
	APPLerrno = APPLE_FILEWRITE;
	ERWrite(APPLerrno, __func__, "Directory \"%s\" is not writable by user: %s", dirName, strerror(errno));
    } else {
	return dirName;
    }
    return NULL;
} // IsValidTempDirectory()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: GetDirectory
 *  Purpose	: Determine the directory to store the temporary file in
 *  Method	:
 *  Check the following paths in the specified order. If defined, it must be a user writable directory or this function will fail.
 *  - the supplied directory,
 *  - directory in environment variable "TMPDIR",
 *  - system directory "/tmp"
 *
 *  History:
 *  05-02-2020	: heesakkm : creation
 *
 *  Notes	:
 *  This function is an adaptation of the path search algorithm that is used by tmpnam() and tempnam(). In the orignal algorithm,
 *  tmpnam() always writes into the system's /tmp and tempnam() first tries "TMPDIR" before testing the supplied directory.
 *
 *  Returns	: Pointer to an existing directory or NULL
 *------------------------------------------------------------------------------------------------------------------------------
 */
static const char *
GetDirectory(
	const char	*directory)	// optional file location
{
    // statbuf is declared here and not in IsValidTempDirectory(), to avoid Fortify warning about a File System Access race
    // condition between stat() in IsValidTempDirectory() and unlink() in CreateTempFileName(), the caller of this function.
    struct stat	statbuf;

    // Try the supplied directory
    directory = IsValidTempDirectory(directory, &statbuf);
    if ((directory != NULL) || (APPLerrno != APPLE_NOERROR)) return directory;

    // Try environment variable "TMPDIR"
    directory = IsValidTempDirectory(ENVGetVar("TMPDIR"), &statbuf);
    if ((directory != NULL) || (APPLerrno != APPLE_NOERROR)) return directory;

    // Try "/tmp"
    directory = IsValidTempDirectory("/tmp", &statbuf);
    if ((directory != NULL) || (APPLerrno != APPLE_NOERROR)) return directory;

    return NULL;
} // GetDirectory()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: SetFileTemplate
 *  Purpose	: Create a template suitable for use by mkstemp()
 *  Method	:
 *  Write the directory into the output buffer, followed by up to five characters of the supplied prefix if it is not NULL, and
 *  ending in "XXXXXX" as prescribed by mkstemp(). If no prefix is supplied, then "file" is used.
 *
 *  History:
 *  05-02-2020	: heesakkm : creation
 *
 *  Notes	: This function is based on the path search algorithm that is used by tmpnam() and tempnam().
 *
 *  Returns	: FALSE if there's not enough space in output buffer
 *------------------------------------------------------------------------------------------------------------------------------
 */
static bool
SetFileTemplate(
	char		*buf,		// OUT: result buffer
	int		bufSize,	// size of result buffer
	const char	*directory,	// file location
	const char	*prefix)	// optional file prefix
{
    // Exclude trailing slashes
    int dirLen = strlen(directory);
    while (dirLen > 1 && directory[dirLen - 1] == '/') {
	dirLen--;
    }

    // Use the default file prefix if none is supplied
    if ((prefix == NULL) || (*prefix == '\0')) {
	prefix = PM_FILE_DEFAULT_PREFIX;
    }
    // Limit the length of the prefix
    int prefixLen = strlen(prefix);
    if (prefixLen > PM_FILE_MAX_PREFIX_LEN) {
	prefixLen = PM_FILE_MAX_PREFIX_LEN;
    }

    // Set the result buffer, it must have enough room for "${directory}/${prefix}XXXXXX\0"
    int len = snprintf(buf, bufSize, "%.*s/%.*s%s", (int) dirLen, directory, (int) prefixLen, prefix, PM_FILE_FILE_PATTERN);
    if (len >= bufSize) {
	APPLerrno = APPLE_BUFFEROVERFLOW;
	ERWrite(APPLerrno, __func__, "Buffer for temporary filename too small (size %d < needed %d)", bufSize, len+1);
	return FALSE;
    }
    return TRUE;
} // SetFileTemplate()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: CreateTempFileName
 *  Purpose	: Create an unique name for a temporary file
 *  Method	:
 *  - Determine the directory to store the temporary file in.
 *  - Set the file mode creation mask so that the file will be created with mode 0600.
 *  - Create a filename template for use by mkstemp().
 *  - Generate the unique filename with mkstemp(), which returns an open file descriptor to the new temporary file.
 *  - Truncate result buffer if mkstemp() fails, otherwise close the file descriptor and remove the temporary file.
 *
 *  History:
 *  05-02-2020	: heesakkm : creation
 *
 *  Notes	:
 *  Although mkstemp() uses mode 0600 on modern systems, the POSIX specification of mkstemp() does not say anything about file
 *  modes. Hence the umask() call prior to creating the file. That also prevents the "Insecure Temporary File" warning from Fortify.
 *
 *  mkstemp() replaces the "XXXXXX" part in the supplied pathBuf with a string that makes the filename unique. It creates and opens
 *  the file, and returns an open file descriptor for the file. Because we're only interested in the filename, we close the file
 *  descriptor and remove the file.
 *
 *  Returns	: Whether the result buffer contains a unique filename
 *------------------------------------------------------------------------------------------------------------------------------
 */
static bool
CreateTempFileName(
	char		*buf,		// OUT: result buffer
	int		bufSize,	// size of result buffer
	const char	*directory,	// optional file location
	const char	*prefix)	// optional file prefix
{
    bool	ok = TRUE;

    // Determine the directory to create the temporary file in.
    directory = GetDirectory(directory);
    if (directory == NULL) {
	return FALSE;
    }

    // Create a filename template into buf for use by mkstemp()
    if (!SetFileTemplate(buf, bufSize, directory, prefix)) {
	return FALSE;
    }

    // Set the file-creation mask to 0177 to ensure that the new file has restrictive permissions, i.e. owner read/write permission
    // and no group/other permissions:
    // S_IXUSR 0100   X mask for owner -> file mode 0600 (read 0400 / write 0200)
    // S_IRWXG 0070 RWX mask for group -> file mode 0000 (no permission)
    // S_IRWXO 0007 RWX mask for other -> file mode 0000 (no permission)
    //
    // Explicitly setting the mask prevents the "Insecure Temporary File" warning from Fortify.
    // umask stores the current mask into oldmask
    mode_t	oldmask = umask(S_IXUSR|S_IRWXG|S_IRWXO);

    // Generate a unique temporary filename using the template in buf
    errno = 0;
    int fd = mkstemp(buf);

    if (fd < 0) {
	APPLerrno = APPLE_FILEOPEN;
	ERWrite(APPLerrno, __func__, "mkstemp(%s) failed: %s", buf, strerror(errno));
	// Truncate the result buffer
	buf[0] = '\0';
	ok = FALSE;
    } else {
	(void) unlink(buf);
	(void) close(fd);
    }

    // Restore the original umask
    (void) umask(oldmask);

    return ok;
} // CreateTempFileName()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: PM_file_tmpnam
 *  Purpose	: Generate a unique temporary filename using prefix "file"
 *  Method	:
 *  Create the temporary filename with CreateTempFileName() into a buffer.
 *  If no output buffer is supplied, then the filename is generated in an internal static buffer and may be
 *  overwritten by the next call to PM_file_tmpnam().
 *
 *  History:
 *  05-02-2020	: heesakkm : CBS-9124 creation
 *
 *  Notes	:
 *  This function replaces the thread unsafe function tmpnam() in order to get rid of the compiler warning: "the use of 'tmpnam' is
 *  dangerous, better use 'mkstemp'". Note that by keeping the same interface as tmpnam(), this function is thread unsafe too.
 *  To be thread safe, the file descriptor created by mkstemp() should be returned.
 *
 *  Example	:
 *
 *	char	filename[50];
 *
 *	if (PM_file_tmpnam(filename) != NULL) {
 *	    FILE *fp = fopen(filename, "wb+");
 *	    ...
 *	    fclose(fp);
 *	}
 *
 *  Returns	: Pointer to buffer with unique filename or NULL on error
 *------------------------------------------------------------------------------------------------------------------------------
 */
char *
PM_file_tmpnam(
	char	*tmpnamBuf)	// OUT: pointer to output buffer or NULL
{
    static char	resultBuf[FILENAME_MAX];
    char	buf[L_tmpnam];
    char	*bufp = (tmpnamBuf != NULL) ? tmpnamBuf : buf;

    APPLerrno = APPLE_NOERROR;

    // Use the buffer pointed to by tmpnamBuf if non-NULL although we don't know the size.
    // We limit the size to L_tmpnam characters in any case, which is backward compatible with tmpnam().
    if (!CreateTempFileName(bufp, sizeof(buf), NULL, NULL)) {
	return NULL;
    }
    if (tmpnamBuf == NULL) {
	PM_strncpy(resultBuf, bufp, sizeof(resultBuf));
	tmpnamBuf = resultBuf;
    }
    return tmpnamBuf;
} // PM_file_tmpnam()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: PM_file_tempnam
 *  Purpose	: Generate a unique temporary filename, optionally using up to five characters of the supplied prefix
 *  Method	: Create the temporary filename with CreateTempFileName() and duplicate the result with PM_strdup().
 *
 *  History:
 *  05-02-2020	: heesakkm : CBS-9124 creation
 *
 *  Notes	:
 *  This function replaces the thread unsafe function tempnam() in order to get rid of the compiler warning: "the use of 'tempnam'
 *  is dangerous, better use 'mkstemp'". Note that by keeping the same interface as tempnam(), this function is thread unsafe too.
 *  To be thread safe, the file descriptor created by mkstemp() should be returned.
 *
 *  Example	:
 *
 *	char *filename = PM_file_tempnam(".", "tmp");
 *
 *	if (filename != NULL) {
 *	    FILE *fp = fopen(filename, "wb+");
 *	    ...
 *	    fclose(fp);
 *	    free(filename);
 *	}
 *
 *  Returns	: Pointer to unique filename (which caller must deallocate) or NULL on error
 *------------------------------------------------------------------------------------------------------------------------------
 */
char *
PM_file_tempnam(
	const char	*directory,	// IN: optional file location
	const char	*prefix)	// IN: optional file prefix
{
    char	retBuf[FILENAME_MAX];

    APPLerrno = APPLE_NOERROR;

    if (!CreateTempFileName(retBuf, sizeof(retBuf), directory, prefix)) {
	return NULL;
    }
    return PM_strdup(retBuf, -1);
} // PM_file_tempnam()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: PM_ensurePathRecursively
 *  Purpose	: Ensure a full directory path exists
 *  Method	: 
 *  1- If the full path exists, return TRUE
 *  2- Create the path up to the last part. This is a recursive call.
 *  3- Create the last part with the mkdir() system call
 *
 *  History:
 *  16-07-2020	: vaneckh  : creation
 *
 *  Returns	: Indication of success.
 *------------------------------------------------------------------------------------------------------------------------------
 */
static bool
PM_ensurePathRecursively(
	char	*directoryPath,
	int	mode)
{
    struct stat		statbuf;

    // Check the filesystem to see if the directory exists. This will also work
    // properly when the directory to test is actually a symbolic link to another
    // directory.
    if (stat(directoryPath, &statbuf) == 0) {
	if ((statbuf.st_mode & S_IFDIR) != 0) {
	    return TRUE;		// directory already exists
	} else {
	    APPLerrno = APPLE_FILEOPEN;
	    ERWrite(APPLerrno, __func__, "%s is not a directory", directoryPath);
	    return FALSE;
	}
    }

    // The directoryPath does not exist (or is a dead symbolic link). Check the
    // parent recusively before trying to make it.
    char	*lastslash = strrchr(directoryPath, '/');
    if (lastslash == directoryPath) {
	// We're at root. That shouldn't happen.
	APPLerrno = APPLE_PARAM;
	ERWrite(APPLerrno, __func__, "Root of the path does not exist: \"%s\"", directoryPath);
	return FALSE;
    }
    // lastslash can be NULL for a relative path, which means that we should skip the recursion
    if (lastslash != NULL)  {
	// Ensure existence of the parent directory. Patch the fullname (temporarily), so the last part is not present.
	*lastslash = '\0';
	if (! PM_ensurePathRecursively(directoryPath, mode)) {
	    return FALSE;
	}
	*lastslash = '/';
    }

    // We've established existence of the parent directory. Now go make the directory.
    if (0 == mkdir(directoryPath, mode)){
	return TRUE;
    }

    // We couldn't create the directory, but it might have been created elsewhere.
    // For example in another process, or when the path contains a sequence of slashes.
    if ((errno == EEXIST) && (stat(directoryPath, &statbuf) == 0)) {
	if ((statbuf.st_mode & S_IFDIR) != 0) {
	    return TRUE;		// directory just created elsewhere. It exists, so we're good.
	} else {
	    APPLerrno = APPLE_FILEOPEN;
	    ERWrite(APPLerrno, __func__, "%s is not a directory", directoryPath);
	    return FALSE;
	}
    }

    // Drats, no luck.
    APPLerrno = APPLE_FILEWRITE;
    ERWrite(APPLerrno, __func__, "Error creating directory %s (%d, %s)", directoryPath, errno, strerror(errno));
    return FALSE;
} // PM_ensurePathRecursively()

/*------------------------------------------------------------------------------------------------------------------------------
 *  Function	: PM_ensurePath
 *  Purpose	: Ensure a full directory path exists
 *  Method	: Call of PM_ensurePathRecursively()
 *
 *  History:
 *  16-07-2020	: vaneckh  : creation
 *
 *  Notes	:
 *  This is a wrapper function to allow a const char * in the interface and be able to log the full path that failed in case of an
 *  error.
 *
 *  Description	:
 *  Will ensure that a full path exists (provided is allowed). The directoryPath argument is taken to be the path of the directory
 *  to create. This can be absolute and relative paths. Creating a top-level directory is not supported.
 *  If a directory does no exist, it is created with the mode as argument to the mkdir() system-call (meaning that the file-system
 *  type and the process' umask are taken into account).
 *
 *  Example	: PM_ensurePath("some/directory/path", 0755);
 *
 *  Returns	: Indication of success.
 *------------------------------------------------------------------------------------------------------------------------------
 */
bool
PM_ensurePath(
	const char	*directoryPath,		// Directory name to ensure
	int		mode)			// mkdir() permission mode
{
    if (directoryPath == NULL) {
	APPLerrno = APPLE_PARAM;
	ERWrite(APPLerrno, __func__, "Called with NULL path");
	return FALSE;
    }

    char	fullpath[strlen(directoryPath) + 1];
    bool	ok = TRUE;

    const char	*in = directoryPath;
    char	*out = fullpath;
    bool	sawSlash = FALSE;

    // Clean up the path, so it doesn't contain sequences of slashes anymore.
    // Not really necessary but prevent EEXIST errors and gives a cleaner "strace" output.
    while (*in != '\0') {
	if (*in == '/') {
	    // Skip the slash if it wasn't the 1st one in a sequence
	    if (sawSlash) {
		in++;
		continue;
	    }
	    sawSlash = TRUE;
	} else {
	    sawSlash = FALSE;
	}
	*out++ = *in++;
    }
    *out = '\0';

    ok = PM_ensurePathRecursively(fullpath, mode);
    if (! ok) {
	// Log the error with the original input argument.
	ERWrite(APPLerrno, __func__, "Could not ensure existence of directory path \"%s\"", directoryPath);
    }
    return ok;
} // PM_ensurePath()
