/*========================================================================
 *  Copyright (c) 2015 by OCLC
 *
 *  File	: pm_isni.c
 *  Description	: Miscellaneous functions for ISNI
 *
 *  History:
 *  26-02-2015	: paul     : CBS-6075 creation
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: pm_isni.c 339156 2016-11-18 18:26:14Z heesakkm $"

#include <string.h>

#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function	: PM_isni_tetragram
 *  Purpose	: Split an ISNI number into 4 blocks of 4 digits
 *
 *  History:
 *  26-02-2015	: paul     : CBS-6075 creation
 *
 *  Notes	:
 *  The only Unicode chars acceptable for this function are ASCII
 *
 *  Description :
 *  PM_isni_tetragram will put spaces in an unsplit ISNI number. If the input
 *  string already contains a space, the output string will be equal to
 *  the input string.
 *
 *  Example     :
 *  in: "0000000065440788"
 *  out: "0000 0000 6544 0788"
 *
 *  Returns	: Number of characters written in 'out' (including '\0')
 *                or PM_ISNI_ERROR if the algorithm doesn't like your input
 *                or 'outsiz' is too low.
 *
 *------------------------------------------------------------------------
 */
size_t
PM_isni_tetragram(
	const char	*in,	/* Input ISNI number */
	char		*out,	/* Output ISNI number */
	size_t		outsiz)	/* Size of 'out' */
{
    size_t	inlen = strlen(in);
    size_t	outwritten;

    if (in == NULL || out == NULL || outsiz == 0) {
	return PM_ISNI_ERROR;
    }

    /* If input contains spaces: assume ISNI is already split */
    if (strchr(in, ' ') != NULL) {
	outwritten = inlen >= outsiz ? outsiz : inlen + 1;
	strncpy(out, in, outwritten - 1);
	out[outwritten - 1] = '\0';
	return inlen >= outsiz ? PM_ISNI_ERROR : outwritten;
    }

    --outsiz;                   /* Reserve byte for terminating '\0'. */
    outwritten = 0;

    while (inlen != 0 && outsiz != 0) {
	size_t	tocopy = 4;

        /* Copy the next tetragram */
        if (tocopy > inlen) {
            tocopy = inlen;
        }
        if (tocopy > outsiz) {
            tocopy = outsiz;
        }
        (void)memcpy(out, in, tocopy);
        in += tocopy;
        inlen -= tocopy;
        out += tocopy;
        outsiz -= tocopy;
        outwritten += tocopy;

        /* Add a space if possible and necessary. */
        if (outsiz != 0 && inlen != 0) {
            *out = ' ';
            ++out;
            --outsiz;
            ++outwritten;
        }
    }

    /* Add terminating '\0'. */
    *out = '\0';
    ++outwritten;

    return inlen != 0 ? PM_ISNI_ERROR : outwritten;
} /* PM_isni_tetragram() */

