/*========================================================================
 *  Copyright (c) 2014 by OCLC
 *
 *  File	: pm_simhsh.c
 *  Description	: Code for similarity hashing
 *
 *  History:
 *  14-04-2015	: paul     : CBS-6055 extracted from pccclus_neighbour.c
 *
 *  Notes	:
 *  From Wikipedia:
 *  QUOTE
 *  MinHash (or the min-wise independent permutations locality sensitive
 *  hashing scheme) is a technique for quickly estimating how similar two sets
 *  are. The scheme was invented by Andrei Broder (1997), and initially used in
 *  the AltaVista search engine to detect duplicate web pages and eliminate
 *  them from search results. It has also been applied in large-scale
 *  clustering problems, such as clustering documents by the similarity of
 *  their sets of words.
 *  UNQUOTE
 *
 *  In this module, a 'simhash' is derived from a text in such a way as to
 *  maximize probability of "collision" of similar texts: similar texts should
 *  have similar (if not identical) simhashes.
 *
 *  A simhash is a sequence of 1 or more minhashes.  Each minhash is computed
 *  with the same algorithm but different parameters; each minhash has its own
 *  'hash-type'.
 *
 *  The alogorithm extracts pentagrams (groups of 5 bytes) from the text. For
 *  each pentagram a hash value of each hash-type is computed. For each
 *  hash-type the lowest of the pentagram-hashes is retained.
 *
 *========================================================================
 */

#pragma ident "$Id: pm_simhsh.c 335412 2016-05-27 11:01:38Z heesakkm $"

#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <er.h>

#include <picamisc.h>

/* FNV hashes are designed to be fast while maintaining a low
 * collision rate. The FNV speed allows one to quickly hash lots
 * of data while maintaining a reasonable collision rate.  See:
 *
 *      http://www.isthe.com/chongo/tech/comp/fnv/index.html
 *
 * for more details as well as other forms of the FNV hash.
 *
 * 32 bit magic FNV-0 and FNV-1 prime
 */
#define FNV_32_PRIME ((uint32)0x01000193)

/*
 * 32 bit FNV-1 and FNV-1a non-zero initial basis
 *
 * The FNV-1 initial basis is the FNV-0 hash of the following 32 octets:
 *
 *              chongo <Landon Curt Noll> /\../\
 *
 * NOTE: The \'s above are not back-slashing escape characters.
 * They are literal ASCII  backslash 0x5c characters.
 *
 * NOTE: The FNV-1a initial basis is the same value as FNV-1 by definition.
 */
#define FNV1_32_INIT ((uint32)0x811c9dc5)
#define FNV1_32A_INIT FNV1_32_INIT

/*------------------------------------------------------------------------
 *  Function	: fnv_compute
 *  Purpose	: Perform a 32 bit Fowler/Noll/Vo hash on a buffer
 *
 *  History:
 *  19-03-2014	: paul     : copied from fnv_32_buf() from:
 *                           http://www.isthe.com/chongo/src/fnv/hash_32.c
 *  14-04-2015	: paul     : CBS-6055 extracted from pccclus_neighbour.c
 *
 *  Notes	:
 *  To use the recommended 32 bit FNV-1 hash, use FNV1_32_INIT as the hshval
 *  argument on the first call to fnv_compute()
 *
 *  Returns	: 32 bit FNV-1 hash value
 *
 *------------------------------------------------------------------------
 */
static uint32
fnv_compute(
	const unsigned char	*buf,	// start of buffer to hash
	size_t			len,	// length of buffer in octets
	uint32			hshval)	/* previous hash value
					 * or FNV1_32_INIT if first call */
{
    const unsigned char	*bp = buf;	// start of buffer
    const unsigned char	*be = bp + len;	// beyond end of buffer

    // FNV-1 hash each octet of the buffer
    while (bp < be) {

	// multiply by the 32 bit FNV magic prime mod 2^32
	hshval *= FNV_32_PRIME;

	// xor the bottom with the current octet
	hshval ^= (uint32)*bp++;
    }

    // return our new hash value
    return hshval;
} // fnv_compute()

/*------------------------------------------------------------------------
 *  Function	: hshtyp_create
 *  Purpose	: Create the array of random numbers to use for generating
 *                the different types of minhashes
 *
 *  History:
 *  25-03-2014	: paul     : CBS-5458 creation
 *  14-04-2015	: paul     : CBS-6055 extracted from pccclus_neighbour.c
 *
 *  Returns	: The allocated array of random numbers
 *                or NULL if an error occurred
 *
 *------------------------------------------------------------------------
 */
static uint32 *
hshtyp_create(
	int	minhsh_count)	// size of the simhash
{
    uint32	*hshtyp;

    hshtyp = malloc(sizeof(*hshtyp) * minhsh_count);
    if (hshtyp == NULL) {
	APPLerrno = APPLE_NOMEM;
	ERWrite(APPLerrno, __func__, "Can't allocate hshtyp");
	return NULL;
    }

    /* Use the same random numbers in every run or else clustering becomes
     * unpredictable
     */
    srand(1);    

    for (int i = 0; i < minhsh_count; i++) {
	/* The highest bit returned by rand() is always zero
	 * Make sure it too is really random
	 */
	hshtyp[i] = ((uint32)rand() << 16) ^ (uint32)rand();
    }
    
    return hshtyp;
} // hshtyp_create()

/*------------------------------------------------------------------------
 *  Function	: PM_simhsh_compute
 *  Purpose	: Compute the simhash of supplied text
 *
 *  History:
 *  25-03-2014	: paul     : CBS-5458 creation
 *  14-04-2015	: paul     : CBS-6055 extracted from pccclus_neighbour.c
 *
 *  Returns	: Pointer to simhash
 *                or NULL if an error occurred
 *
 *------------------------------------------------------------------------
 */
uint32 *
PM_simhsh_compute(
	uint32		*minhsh,	/* destination simhash or NULL */
	int		minhsh_count,	/* elt. count of 'minhsh' buffer;
					   requested size of the simhash */
	const char	*text,		/* string from which to generate */
	uint32		flags)		/* options */
{
    static uint32	*static_minhsh = NULL;
    static int		static_minhsh_count = 0;
    static uint32	*hshtyp = NULL;
    const int		pentagram_size = 5;

    unsigned int	text_len = (unsigned int) strlen(text);
    int			pentagram_len = 0;
    unsigned int	space = ' ';

    (void)flags;	// to please lint

    // Compute values to use for the different hash types
    if (hshtyp == NULL) {
	hshtyp = hshtyp_create(minhsh_count);
	if (hshtyp == NULL) {
	    return NULL;
	}
    }

    // If we need the static buffer: check if it is big enough
    if (minhsh == NULL && minhsh_count > static_minhsh_count) {
	voidp	tmpp = realloc(static_minhsh,
			       sizeof(*static_minhsh) * (minhsh_count));

	if (tmpp == NULL) {
	    APPLerrno = APPLE_NOMEM;
	    ERWrite(APPLerrno, __func__, "Can't allocate static_minhsh");
	    return NULL;
	}
	
	static_minhsh = (uint32 *)tmpp;
	static_minhsh_count = minhsh_count;
    }

    // Use the static buffer for output
    if (minhsh == NULL) {
	minhsh = static_minhsh;
    }

    // Initialize minhashes with maximum values
    for (int i = 0; i < minhsh_count; i++) {
	minhsh[i] = UINT32MAX;
    }

    // Compute the minhashes
    for (unsigned int j = 0; j < text_len; j++) {
	pentagram_len++;

	// Check if we're starting with a new word
	if ((unsigned)text[j] <= space){
	    pentagram_len = 0;
	    continue;
	}

	// If pentagram is full: update the array of minhashes
	while (pentagram_len == pentagram_size
	       || ((unsigned)text[j + 1] <= space
		   && pentagram_len > 0)) {

	    byte	*buf = (byte *)text + j + 1 - pentagram_len;
	    uint32	fnv = fnv_compute(buf, pentagram_len, FNV1_32_INIT);
	    
	    for (int i = 0; i < minhsh_count; i++) {
		if (minhsh[i] > fnv * hshtyp[i]) {
		    minhsh[i] = fnv * hshtyp[i];
		}
	    }

	    // Prepare for the next pentagram
	    pentagram_len--;
	}
    }

    return minhsh;
} // PM_simhsh_compute()

