/*========================================================================
 *  Copyright (c) 2003 by Pica
 *
 *  File	: pm_utf.c
 *  Description	: Utf8 and utf32 character (aka code point) functions
 *  Author(s)	: Paul Smit
 *  Creation	: 20-11-2003
 *
 *  History:
 *  20-11-2003	: paul     : creation
 *  21-11-2003	: paul     : Added PM_Utf32CharRefPut()
 *  20-04-2004  : Guy      : moved PM_APPEND_UNSAFE to picamisc.h
 *
 *========================================================================
 */

#pragma ident "$Id: pm_utf.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <pica.h>
#include <picamisc.h>

/*--------------------------------------------------------------------
 * Function   : PM_Utf32CharRefGet
 * Purpose    : Get an XML character reference
 * Author     : Paul Smit
 *
 * History    :
 * 24-10-2003 : paul     : Creation
 * 30-10-2003 : paul     : Don't change end if char ref not recognized
 *                       : end of char ref is behind the ';', not on it.
 * 20-11-2003 : paul     : Renamed from PU8_CharRefGet()
 *
 * Notes      :
 * An XML character reference has one of two forms:
 *
 *	&#xFFFFF;
 *	&#999;
 *
 * where F is any hex digit and 9 is any decimal digit and where there can be
 * up to 8 hex digits and up to 9 decimal digits. (Even though the current
 * maximum code point consists of only 6 hex digits or 7 decimal digits)
 *
 * The value of the digits is the code point in Unicode.
 *
 * This function is not in the PU library because it is used in the CCV library
 * and we don't want to link the whole ICU library just because we need the CCV
 * library.
 *
 * Returns    :
 * The code point or PM_UTF32_CHAR_NONE if either incorrect escape sequence
 * or code point greater than maximum code point
 *-------------------------------------------------------------------
 */
int32
PM_Utf32CharRefGet(
    const char	*in,		/* points to the ampersand */
    const char	**end)		/* points to char behind the ';' if non-NULL */
{
    int			len;
    unsigned long	u32;
    int			base;

    if (*(in++) != '&' || *(in++) != '#') return PM_UTF32_CHAR_NONE;

    if (*in == 'x') {
	++in;
	len = strspn(in, "0123456789abcdefABCDEF");
	if (len > 8) {
	    return PM_UTF32_CHAR_NONE;
	}
	base = 16;
    } else {
	len = strspn(in, "0123456789");
	if (len > 9) {
	    return PM_UTF32_CHAR_NONE;
	}
	base = 10;
    }

    if (len == 0) {
	return PM_UTF32_CHAR_NONE;
    }
    if (in[len] != ';') return PM_UTF32_CHAR_NONE;

    u32 = strtoul(in, (char **) end, base);

    if (end != NULL) ++*end; /* skip ';' */

    /* the return cannot overflow due to the limit on the number of digits, so
       just cast this number to Utf32: */
    return (int32) u32;
} /* PM_Utf32CharRefGet() */

/*--------------------------------------------------------------------
 * Function   : PM_Utf32CharRefPut
 * Purpose    : Put an hexadecimal XML character reference
 * Author     : Paul Smit
 *
 * History    :
 * 21-11-2003 : paul     : Creation
 *
 * Notes      :
 * The structure of a XML character reference is described at
 * PM_Utf32CharRefGet()
 *
 * This function assumes that PM_HEX_CHARREF_LEN_MAX or more bytes are
 * available in the output buffer, which may not be NULL.
 *
 * Returns    :
 * The length of the XML character reference or a negative number
 * in case of error
 *-------------------------------------------------------------------
 */
int
PM_Utf32CharRefPut(
    int32	u32,				/* the code point to write */
    char	out[PM_HEX_CHARREF_LEN_MAX])	/* the output buffer or NULL */
{
    return sprintf(out, "&#x%04lx;", (long) u32);
} /* PM_Utf32CharRefPut() */

/*------------------------------------------------------------------------
 *  Function	: PM_Utf32CharToUtf8
 *  Purpose	: Put a utf32 char into a utf8 string
 *  Method	: PM_APPEND_UNSAFE()
 *
 *  History:
 *  30-10-2003	: paul     : creation
 *
 *  Notes	:
 *  If a NULL buffer is supplied, the static ret_buffer is used.
 *  If the buffer is non-NULL, its size is assumed to be at least
 *  PM_UTF32_CHAR_LEN_MAX
 *
 *  Returns	: The utf8 char or NULL
 *------------------------------------------------------------------------
 */
const char *
PM_Utf32CharToUtf8(
    int32		u32c,
    char		*dest)
{
    static char	ret_buffer[PM_UTF32_CHAR_LEN_MAX];
    int		len = 0;

    if (dest == NULL) dest = ret_buffer;

    PM_APPEND_UNSAFE(dest, len, u32c);
    dest[len] = '\0';

    return dest;
} /*PM_Utf32CharToUtf8() */
