/*========================================================================
 *  Copyright (c) 1999 by Pica
 *
 *  File	: pmbase64.c
 *  Description	: Base64 Encoding and decoding
 *  Author(s)	: Hans van Eck
 *  Creation	: 23-02-1999
 *
 *  History:
 *  23-02-1999	: eck      : creation
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *  Notes	:
 *  The routunes do base64 [en|de]coding. It differs a little from the
 *  encoding described in RFC1521, since it does not support the 76
 *  characters-per-line limit. This is done because this limit is not
 *  supported in HTTP.
 *
 *  Base64 encoding converts 3 input-bytes into 4 by regrouping them into
 *  groups of 6 bits and then converting to ASCII.
 *
 *  Description	:
 *  Base64 encoding is used in MIME to transfer binary data in ASCII.  The
 *  specification for base64 encoding can be found in RFC 1521. The routines
 *  specified here do not comply completely with this specification: they do
 *  not have the 76 characters-per-line limitation because HTTP does not have
 *  this limitation.
 *
 *========================================================================
 */

#pragma ident "$Id: pmbase64.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <stdlib.h>
#include <string.h>
#include <pica.h>
#include <picamisc.h>

static const char table[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/*------------------------------------------------------------------------
 *  Function	: PM_Base64Decode
 *  Purpose	: Decode message in base 64 format
 *  Method	:
 *
 *  History:
 *  01-12-1997	: BdB      : creation
 *  23-02-1999	: eck      : Moved to library
 *
 *  Notes	: Result will be null-terminated.
 *
 *  Description	:
 *  PM_Base64Decode() converts a sequence of 4 input characters into 3 output
 *  characters. It will ensure that the result is '\0'-terminated. The return
 *  value is the number of characters in the output buffer (excluding the
 *  terminating '\0').
 *
 *  This routine will decode a base64-encoded input string.
 *  The parameter 'in' is a pointer to the input string.
 *  The parameter 'out' is a pointer to the output buffer.
 *  The parameter 'outsiz' contains the size of the buffer.
 *
 *  The proper size for the output buffer can be determined from the input
 *  lenght (approximately input length * 3/4  + some extra for rounding off and
 *  termination).
 *
 *  Returns	:
 *  Number of decoded chars (excluding terminating '\0') or -1 on error (can't
 *  happen currently).
 *------------------------------------------------------------------------
 */
int
PM_Base64Decode(
	const char	*in,		/* input string '\0'-terminated */
	byte		*out,		/* output buffer */
	size_t		outsiz)		/* size of output buffer */
{
    int			outp;
    unsigned int	shiftreg;
    int			bits;
    char		p;

    outp = 0;
    shiftreg = 0;
    bits = 0;

    while ((p = *in++) != '\0' && outp < outsiz - 1) {
	const char *const val = strchr(table, p);
	if (val == NULL) {
	    break;
	}
	shiftreg <<= 6;
	shiftreg |= val - table;
	bits += 6;
	if (bits >= 8) {
	    bits -= 8;
	    out[outp++] = shiftreg >> bits;
	}
    }

    out[outp] = '\0';
    return outp;
} /* PM_Base64Decode() */

/*------------------------------------------------------------------------
 *  Function	: PM_Base64Encode
 *  Purpose	: Base64 encoding of a string
 *  Method	:
 *  The input bytes are stored in a 'bitbucket' (by shifting what is in
 *  the bucket). Then, the highest 6 bits from the input are output until the
 *  number of bits left < 6. When the input is exhausted, the remaining bits
 *  are output and padding bytes are added.
 *
 *  History:
 *  23-02-1999	: eck      : creation
 *
 *  Description	:
 *  PM_Base64Encode() will convert a piece of memory (or '\0'-terminated
 *  string) into a base64-encoded string. This is done by converting 3 bytes
 *  into 4 printable characters.
 *
 *  The parameter 'in' is a pointer to the input string.
 *  The parameter 'inlen' should contain the length of input (may be -1 when
 *  'in' is a  '\0'-terminated string).
 *  The parameter 'out' is a pointer to the output buffer.
 *  The parameter 'outsiz' contains the size of the buffer.
 *
 *  The result will always be '\0'-terminated.
 *  The value returned is the length of the output string (excluding the
 *  terminating '\0'-byte). If the output buffer was large enough, this is
 *  always a multiple of 4.
 *
 *  The proper size for the output buffer can be determined from the input
 *  length (approximately input length * 4/3 + some extra for rounding off,
 *  padding and termination).
 *
 *  Returns	: Length of the encoded buffer (excluding terminating '\0').
 *------------------------------------------------------------------------
 */
int
PM_Base64Encode(
	byte		*in,		/* input buffer */
	int		inlen,		/* size of input buffer */
	char		*out,		/* output buffer */
	size_t		outsiz)		/* size of output buffer */
{
    int		bitsleft = 0;
    uint16	bitbucket = 0;		/* could be uint16 */
    size_t	sizleft = outsiz;
    int		padbytes;

    if (inlen < 0) {
	inlen = strlen((char *) in);
    }

    /* Assume CHAR_BIT == 8, otherwise we have a problem. */
    while ((inlen > 0) && (sizleft != 0)) {
	bitbucket = (bitbucket << 8) | *in++;
	bitsleft += 8;
	inlen--;

	while ((sizleft != 0) && (bitsleft >= 6)) {
	    *out++ = table[(bitbucket >> (bitsleft - 6)) & 0x3f];
	    sizleft--;
	    bitsleft -= 6;
	}
    }
    if ((sizleft != 0) && (bitsleft != 0)) {
	/* Pending input bits are 0-padded. */
	*out++ = table[(bitbucket << (6 - bitsleft)) & 0x3f];
	sizleft--;

	/* Output is '='-padded. #pad-bytes == (6 - bitsleft) / 2 */
	padbytes = (6 - bitsleft) / 2;
	while ( (sizleft != 0) && (padbytes != 0)) {
	    *out++ = '=';
	    sizleft--;
	    padbytes--;
	}
    }
    if (sizleft == 0) out--;
    *out = '\0';

    return outsiz - sizleft;
} /* PM_Base64Encode() */
