/*========================================================================
 *  Copyright (c) 2006 by Pica
 *
 *  File	: pmean.c
 *  Description	: EAN checkdigit routines.
 *
 *  History:
 *  29-08-2006	: paul      : creation
 *
 *  Notes	:
 *  This code was copied from pmmod11.c and adapted
 *
 *  These functions works for EAN numbers of any length, including:
 *  GTIN-8	( 8 digit Global Trade Item Number)
 *  GTIN-12	(12 digit Global Trade Item Number)
 *  GTIN-13	(13 digit Global Trade Item Number)
 *  GTIN-14	(14 digit Global Trade Item Number)
 *  SSCC	(18 digit Serial Shipping Container Code)
 *========================================================================
 */

#pragma ident "$Id: pmean.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function	: PM_EAN_gen
 *  Purpose	: Generate an EAN check digit from a digit-string.
 *  Method	:
 *  1. Digits are processed in reverse: last one first.
 *  2. First a sum of the digits on even positions is computed.
 *  3. This sum is multiplied by 3.
 *  4. Then the digits on odd positions are added.
 *  5. The checksum is then computed using the expression (10 - (sum % 10)) % 10,
 *     which gives the distance of the sum to the next multiple of 10 (and 0
 *     when the sum can be divided by 10).
 *  6. The result is then the check-digit.
 *
 *  Consider the string "457".
 *  The '7' is multiplied by 3, the '5' is is multiplied by 1 and the '4'
 *  is multiplied by 3. This gives a total of 3*7 + 1*5 + 3*4 = 38. The
 *  next mupltiple is 40, giving a difference of 2. This means that the
 *  check-digit is '2'.
 *
 *  History:
 *  29-08-2006	: paul      : creation
 *
 *  Notes	:
 *  The input string is checked for len bytes, all are assumed digits. If
 *  len is negative, str must be '\0'-terminated.
 *
 *  Returns	: The check digit.
 *------------------------------------------------------------------------
 */
char
PM_EAN_gen(
	const char	*str,	/* input string (digits only) */
	int		len)	/* length of str (without check-digit) */
{
    int	sum;
    int	i;

    if (len < 0) len = (int) strlen(str);

    sum = 0;
    for (i = len - 1; i >= 0; i -= 2) {
	sum += str[i] - '0';
    }
    sum *= 3;
    for (i = len - 2; i >= 0; i -= 2) {
	sum += str[i] - '0';
    }
    sum = (10 - (sum % 10)) % 10;

    return '0' + sum;
} /* PM_EAN_gen() */

/*------------------------------------------------------------------------
 *  Function	: PM_EAN_test
 *  Purpose	: Test an EAN check-digit in a string.
 *  Method	:
 *  PM_EAN_gen() is used to generate a new check-digit by passing (len - 1)
 *  as the string lengh. This is then compared with one in the string.
 *
 *  History:
 *  29-08-2006	: paul      : creation
 *
 *  Returns	: Whether the number has the right check digit
 *------------------------------------------------------------------------
 */
bool
PM_EAN_test(
	const char	*str,	/* input string */
	int		len)	/* length of str (including check-digit) */
{
    char	chk_digit;

    if (len < 0) len = (int) strlen(str);

    if (len == 0) return FALSE;		/* empty (check digit only is ok) */

    len--;	/* The length now doesn't include the checkdigit */

    chk_digit = PM_EAN_gen(str, len);

    return (chk_digit == str[len]);
} /* PM_EAN_test() */
