/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: pmgensym.c
 *  Description	: Generate a unique symbol.
 *  Author(s)	: Hans van Eck
 *  Creation	: 06-01-1994
 *
 *  History:
 *  06-01-1994	: eck      : creation
 *  11-01-1994	: eck      : moved PM_utoa() to pmtoasc.c
 *  12-01-1994	: eck      : fixed initialisation bug
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: pmgensym.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function	: PM_gensym
 *  Purpose	: Generate a unique string.
 *
 *  History:
 *  06-01-1994	: eck      : creation
 *  12-01-1994	: eck      : initialised var when SymStart == NULL.
 *
 *  Notes	:
 *  'SymStart' should point to a string containing the start of the unique
 *  symbol. This can be used to generate symbols which have the same start,
 *  which is handy for grouping symbols. SymStart should not end in digits,
 *  since this would defeat the uniqueness algorithm (which is simply to
 *  generate a integer-string after the SymStart).
 *
 *  The length of SymStart should be smaller than PM_SYMLEN_MIN -
 *  PM_INTLEN_MAX, which is about 20 characters at present. In order to copy
 *  the value, an array of PM_SYMLEN_MIN characters should be defined.
 *
 *  When SymStart is NULL, the string "UNIQUE" will precede the integer.
 *
 *  The return value points to static area, so the symbol should be copied
 *  before it can be used safely.
 *
 *  Returns	: pointer to string
 *------------------------------------------------------------------------
 */
const char *
PM_gensym(const char *SymStart)		/* String to start the symbol with */
{
    static unsigned	id_count = 1;
    static char		var[PM_SYMLEN_MIN];
    int			len;

    if (SymStart == NULL) {
	SymStart = "UNIQUE";
    }
    len = strlen(SymStart);

    if (len > (PM_SYMLEN_MIN - PM_INTLEN_MAX)) {
	len = PM_SYMLEN_MIN - PM_INTLEN_MAX;
    }
    (void) strncpy(var, SymStart, len);
    (void) PM_utoa(var + len, id_count++);

    return var;
} /* PM_gensym() */
