/*========================================================================
 *  Copyright (c) 1997 by Pica
 *
 *  File        : pmisbn.c
 *  Description : Hyphenate ISBNs and ISSNs
 *  Author(s)   : Bas de Bakker
 *  Creation    : 11-11-1997
 *
 *  History:
 *  11-11-1997  : BdB      : creation
 *  26-07-2000  : paul     : ISBN_SIZE renamed to PM_ISBN_SIZE and moved to
 *                           <picamisc.h>
 *                           Eliminated lint warnings by casting -1 to (size_t)
 *  12-03-2001  : paul     : PM_ISBN_HYPHEN used in both PM_hyphenate_isbn()
 *                           and PM_hyphenate_issn()
 *  24-10-2003  : huug	   : new rule in PM_hyphenate_isbn:
 *			     908[0-4]nnnnnx -> 90-8[0-4]nnnn-n-x
 *			     908[5-9]nnnnmx -> 90-8[5-9]nn-nnn-x
 *			     old rule was:
 *			     908nnnnnnx -> 90-8nnnnn-n-x
 *  02-09-2004  : paul	   : implemented 13 digit ISBN's
 *  18-10-2004	: huug	   : treat 9978 ISBNs like 9977 ISBNs
 *  07-03-2008	: sutherlr : added PM_isbn_13_to_10() and PM_isbn_10_to_13()
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: pmisbn.c 339156 2016-11-18 18:26:14Z heesakkm $"

#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include <pica.h>
#include <picamisc.h>

#define MAX_ISBN 18
#define isISBNdigit(c)        (isdigit(c) || ((c) == 'X') || ((c) == 'x'))
#define skip_non_ISBN_digit(p)    while (*p && !isISBNdigit(*p)) p++
static void
squeezeISBN(
	char	    *out,
	const char  *in)
{
    const char *i = in;
    char	*o = out;

    while (*i != 0 && o-out < MAX_ISBN) {
	if (isISBNdigit(*i)) {
	    *o++ = *i;
	}
	i++;
    }
    *o = '\0';
} /* squeezeISBN() */

static int
in_ISBN_range(
	const char  *begin,
	const char  *ISBN,
	const char  *end)
{
    /* Compare the countrygroupnumber-publishernumber part of ISBN
       against the range (begin, end), and return -1 (less than),
       0 (in range), or +1 (greater than). */

    char begin_prefix[MAX_ISBN];
    char end_prefix[MAX_ISBN];
    char ISBN_prefix[MAX_ISBN];

    squeezeISBN(begin_prefix, begin);
    squeezeISBN(ISBN_prefix,ISBN);

    if (strncmp(ISBN_prefix,begin_prefix,strlen(begin_prefix)) < 0)
        return (-1);

    squeezeISBN(end_prefix,end);
    if (strncmp(end_prefix,ISBN_prefix,strlen(end_prefix)) < 0)
        return (1);

    return (0);
} /* in_ISBN_range() */

/*------------------------------------------------------------------------
 *  Function	: hyphenate_one_ISBN
 *  Purpose	: copy ISBN while adding hyphens according to prefix patern
 *  Method	:
 *
 *  History:
 *  28-11-2012	: huug     : creation
 *  29-10-2015	: paul     : CBS-6374 No functional change:
 *                           replaced '-' by PM_ISBN_HYPHEN
 *
 *  Notes	:
 *
 *  Returns	: length of hyphenated ISBN, including terminating '\0'
 *
 *------------------------------------------------------------------------
 */
static int
hyphenate_one_ISBN(
	char	    *new_ISBN,
	const char  *prefix,
	const char  *ISBN)
{
    /*******************************************************************
      Given a countrygroupnumber-publishernumber prefix, and an ISBN
      optionally containing spaces and hyphens, return a pointer to an
      unmodifiable properly-hyphenated ISBN stored in an internal buffer
      that is overwritten on subsequent calls, or NULL if the correct
      number of ISBN digits is not found.

      The input ISBN can contain optional leading and trailing text,
      such as a line from a BibTeX .bib file, like this:

         ISBN =         "0-387-09823-2 (paperback)",

     ******************************************************************/

    int k;

    /* loop up to the last character of ISBN, the checkdigit */
    for (k = strlen(new_ISBN); ISBN[1] ; ) {
        if (*prefix == PM_ISBN_HYPHEN) {
            new_ISBN[k++] = PM_ISBN_HYPHEN;
            prefix++;
        } else if (*prefix) {
            if (*ISBN == '\0') break;
            new_ISBN[k++] = *ISBN++;
            prefix++;
            if ((*prefix == '\0')) {
                new_ISBN[k++] = PM_ISBN_HYPHEN;
	    }
        } else { /* past prefix */
            if (*ISBN == '\0') break;
            new_ISBN[k++] = *ISBN++;
        }
    }
    new_ISBN[k++] = PM_ISBN_HYPHEN;
    new_ISBN[k++] = *ISBN;
    new_ISBN[k++] = '\0';
    return k;
} /* hyphenate_one_ISBN() */

/*------------------------------------------------------------------------
 *  Function	: fix_ISBN
 *  Purpose	: make a hyphenated copy of the ISBN
 *  Method	: find the proper range and hyphenate accordingly
 *
 *  History:
 *  28-11-2012	: huug     : creation
 *  21-10-2015	: paul     : CBS-6362 Regenerated ranges from RangeMessage.xml
 *  29-10-2015	: paul     : CBS-6374 No functional change:
 *                           replaced '-' by PM_ISBN_HYPHEN
 *
 *  Notes	: when no range is found, the output buffer will be unaffected
 *
 *  Returns	:
 *  -1: improper ISBN or ISBN not in any range
 *  >0 nr bytes used in out, including terminating '\0'
 *------------------------------------------------------------------------
 */
static int
fix_ISBN(
	const char  *ISBN,
	int	    ISBNlen,
	char	    *out,
	int	    iPrefix)
{
    size_t  i;
    typedef struct {
        const char *begin;
        const char *end;
    } range_t, *range_p;

    /*
     * ISBN ranges are documented by the International ISBN Agency:
     * http://www.isbn-international.org/
     * http://www.isbn-international.org/page/ranges
     */
    /* the following table only applies to 978- ISBNs
     * 979- ISBNs are managed by a separate table (which follows the 978 table)
     */
    static range_t ISBN_range978[] =
    {

	/* English language */
	{ "0-00",          "0-19" },
	{ "0-200",         "0-699" },
	{ "0-7000",        "0-8499" },
	{ "0-85000",       "0-89999" },
	{ "0-900000",      "0-949999" },
	{ "0-9500000",     "0-9999999" },

	/* English language */
	{ "1-00",          "1-09" },
	{ "1-100",         "1-399" },
	{ "1-4000",        "1-5499" },
	{ "1-55000",       "1-86979" },
	{ "1-869800",      "1-998999" },
	{ "1-9990000",     "1-9999999" },

	/* French language */
	{ "2-00",          "2-19" },
	{ "2-200",         "2-349" },
	{ "2-35000",       "2-39999" },
	{ "2-400",         "2-699" },
	{ "2-7000",        "2-8399" },
	{ "2-84000",       "2-89999" },
	{ "2-900000",      "2-949999" },
	{ "2-9500000",     "2-9999999" },

	/* German language */
	{ "3-00",          "3-02" },
	{ "3-030",         "3-033" },
	{ "3-0340",        "3-0369" },
	{ "3-03700",       "3-03999" },
	{ "3-04",          "3-19" },
	{ "3-200",         "3-699" },
	{ "3-7000",        "3-8499" },
	{ "3-85000",       "3-89999" },
	{ "3-900000",      "3-949999" },
	{ "3-9500000",     "3-9539999" },
	{ "3-95400",       "3-96999" },
	{ "3-9700000",     "3-9899999" },
	{ "3-99000",       "3-99499" },
	{ "3-99500",       "3-99999" },

	/* Japan */
	{ "4-00",          "4-19" },
	{ "4-200",         "4-699" },
	{ "4-7000",        "4-8499" },
	{ "4-85000",       "4-89999" },
	{ "4-900000",      "4-949999" },
	{ "4-9500000",     "4-9999999" },

	/* former U.S.S.R */
	{ "5-00000",       "5-00499" },
	{ "5-0050",        "5-0099" },
	{ "5-01",          "5-19" },
	{ "5-200",         "5-420" },
	{ "5-4210",        "5-4299" },
	{ "5-430",         "5-430" },
	{ "5-4310",        "5-4399" },
	{ "5-440",         "5-440" },
	{ "5-4410",        "5-4499" },
	{ "5-450",         "5-699" },
	{ "5-7000",        "5-8499" },
	{ "5-85000",       "5-89999" },
	{ "5-900000",      "5-909999" },
	{ "5-91000",       "5-91999" },
	{ "5-9200",        "5-9299" },
	{ "5-93000",       "5-94999" },
	{ "5-9500000",     "5-9500999" },
	{ "5-9501",        "5-9799" },
	{ "5-98000",       "5-98999" },
	{ "5-9900000",     "5-9909999" },
	{ "5-9910",        "5-9999" },

	/* Iran */
	{ "600-00",        "600-09" },
	{ "600-100",       "600-499" },
	{ "600-5000",      "600-8999" },
	{ "600-90000",     "600-99999" },

	/* Kazakhstan */
	{ "601-00",        "601-19" },
	{ "601-200",       "601-699" },
	{ "601-7000",      "601-7999" },
	{ "601-80000",     "601-84999" },
	{ "601-85",        "601-99" },

	/* Indonesia */
	{ "602-00",        "602-07" },
	{ "602-0800",      "602-0899" },
	{ "602-0900",      "602-1099" },
	{ "602-1100",      "602-1199" },
	{ "602-1200",      "602-1399" },
	{ "602-14000",     "602-14999" },
	{ "602-1500",      "602-1699" },
	{ "602-17000",     "602-17999" },
	{ "602-18000",     "602-18999" },
	{ "602-19000",     "602-19999" },
	{ "602-200",       "602-674" },
	{ "602-6750",      "602-6999" },
	{ "602-70000",     "602-74999" },
	{ "602-7500",      "602-7999" },
	{ "602-8000",      "602-9499" },
	{ "602-95000",     "602-99999" },

	/* Saudi Arabia */
	{ "603-00",        "603-04" },
	{ "603-05",        "603-49" },
	{ "603-500",       "603-799" },
	{ "603-8000",      "603-8999" },
	{ "603-90000",     "603-99999" },

	/* Vietnam */
	{ "604-0",         "604-4" },
	{ "604-50",        "604-89" },
	{ "604-900",       "604-979" },
	{ "604-9800",      "604-9999" },

	/* Turkey */
	{ "605-01",        "605-02" },
	{ "605-030",       "605-039" },
	{ "605-04",        "605-09" },
	{ "605-100",       "605-399" },
	{ "605-4000",      "605-5999" },
	{ "605-60000",     "605-89999" },
	{ "605-9000",      "605-9999" },

	/* Romania */
	{ "606-0",         "606-0" },
	{ "606-10",        "606-49" },
	{ "606-500",       "606-799" },
	{ "606-8000",      "606-9199" },
	{ "606-92000",     "606-99999" },

	/* Mexico */
	{ "607-00",        "607-39" },
	{ "607-400",       "607-749" },
	{ "607-7500",      "607-9499" },
	{ "607-95000",     "607-99999" },

	/* Macedonia */
	{ "608-0",         "608-0" },
	{ "608-10",        "608-19" },
	{ "608-200",       "608-449" },
	{ "608-4500",      "608-6499" },
	{ "608-65000",     "608-69999" },
	{ "608-7",         "608-9" },

	/* Lithuania */
	{ "609-00",        "609-39" },
	{ "609-400",       "609-799" },
	{ "609-8000",      "609-9499" },
	{ "609-95000",     "609-99999" },

	/* Thailand */
	{ "611",	"611" },

	/* Peru */
	{ "612-00",        "612-29" },
	{ "612-300",       "612-399" },
	{ "612-4000",      "612-4499" },
	{ "612-45000",     "612-49999" },
	{ "612-50",        "612-99" },

	/* Mauritius */
	{ "613-0",         "613-9" },

	/* Lebanon */
	{ "614-00",        "614-39" },
	{ "614-400",       "614-799" },
	{ "614-8000",      "614-9499" },
	{ "614-95000",     "614-99999" },

	/* Hungary */
	{ "615-00",        "615-09" },
	{ "615-100",       "615-499" },
	{ "615-5000",      "615-7999" },
	{ "615-80000",     "615-89999" },

	/* Thailand */
	{ "616-00",        "616-19" },
	{ "616-200",       "616-699" },
	{ "616-7000",      "616-8999" },
	{ "616-90000",     "616-99999" },

	/* Ukraine */
	{ "617-00",        "617-49" },
	{ "617-500",       "617-699" },
	{ "617-7000",      "617-8999" },
	{ "617-90000",     "617-99999" },

	/* Greece */
	{ "618-00",        "618-19" },
	{ "618-200",       "618-499" },
	{ "618-5000",      "618-7999" },
	{ "618-80000",     "618-99999" },

	/* Bulgaria */
	{ "619-00",        "619-14" },
	{ "619-150",       "619-699" },
	{ "619-7000",      "619-8999" },
	{ "619-90000",     "619-99999" },

	/* Mauritius */
	{ "620-0",         "620-9" },

	/* Philippines */
	{ "621-00",        "621-29" },
	{ "621-400",       "621-599" },
	{ "621-8000",      "621-8999" },
	{ "621-95000",     "621-99999" },

	/* China, People's Republic */
	{ "7-00",          "7-09" },
	{ "7-100",         "7-499" },
	{ "7-5000",        "7-7999" },
	{ "7-80000",       "7-89999" },
	{ "7-900000",      "7-999999" },

	/* former Czechoslovakia */
	{ "80-00",         "80-19" },
	{ "80-200",        "80-699" },
	{ "80-7000",       "80-8499" },
	{ "80-85000",      "80-89999" },
	{ "80-900000",     "80-999999" },

	/* India */
	{ "81-00",         "81-19" },
	{ "81-200",        "81-699" },
	{ "81-7000",       "81-8499" },
	{ "81-85000",      "81-89999" },
	{ "81-900000",     "81-999999" },

	/* Norway */
	{ "82-00",         "82-19" },
	{ "82-200",        "82-689" },
	{ "82-690000",     "82-699999" },
	{ "82-7000",       "82-8999" },
	{ "82-90000",      "82-98999" },
	{ "82-990000",     "82-999999" },

	/* Poland */
	{ "83-00",         "83-19" },
	{ "83-200",        "83-599" },
	{ "83-60000",      "83-69999" },
	{ "83-7000",       "83-8499" },
	{ "83-85000",      "83-89999" },
	{ "83-900000",     "83-999999" },

	/* Spain */
	{ "84-00",         "84-13" },
	{ "84-140",        "84-149" },
	{ "84-15000",      "84-19999" },
	{ "84-200",        "84-699" },
	{ "84-7000",       "84-8499" },
	{ "84-85000",      "84-89999" },
	{ "84-9000",       "84-9199" },
	{ "84-920000",     "84-923999" },
	{ "84-92400",      "84-92999" },
	{ "84-930000",     "84-949999" },
	{ "84-95000",      "84-96999" },
	{ "84-9700",       "84-9999" },

	/* Brazil */
	{ "85-00",         "85-19" },
	{ "85-200",        "85-549" },
	{ "85-5500",       "85-5999" },
	{ "85-60000",      "85-69999" },
	{ "85-7000",       "85-8499" },
	{ "85-85000",      "85-89999" },
	{ "85-900000",     "85-959999" },
	{ "85-96",         "85-97" },
	{ "85-98000",      "85-99999" },

	/* former Yugoslavia */
	{ "86-00",         "86-29" },
	{ "86-300",        "86-599" },
	{ "86-6000",       "86-7999" },
	{ "86-80000",      "86-89999" },
	{ "86-900000",     "86-999999" },

	/* Denmark */
	{ "87-00",         "87-29" },
	{ "87-400",        "87-649" },
	{ "87-7000",       "87-7999" },
	{ "87-85000",      "87-94999" },
	{ "87-970000",     "87-999999" },

	/* Italy */
	{ "88-00",         "88-19" },
	{ "88-200",        "88-599" },
	{ "88-6000",       "88-8499" },
	{ "88-85000",      "88-89999" },
	{ "88-900000",     "88-909999" },
	{ "88-910",        "88-929" },
	{ "88-9300",       "88-9399" },
	{ "88-940000",     "88-949999" },
	{ "88-95000",      "88-99999" },

	/* Korea, Republic */
	{ "89-00",         "89-24" },
	{ "89-250",        "89-549" },
	{ "89-5500",       "89-8499" },
	{ "89-85000",      "89-94999" },
	{ "89-950000",     "89-969999" },
	{ "89-97000",      "89-98999" },
	{ "89-990",        "89-999" },

	/* Netherlands */
	{ "90-00",         "90-19" },
	{ "90-200",        "90-499" },
	{ "90-5000",       "90-6999" },
	{ "90-70000",      "90-79999" },
	{ "90-800000",     "90-849999" },
	{ "90-8500",       "90-8999" },
	{ "90-90",         "90-90" },
	{ "90-94",         "90-94" },

	/* Sweden */
	{ "91-0",          "91-1" },
	{ "91-20",         "91-49" },
	{ "91-500",        "91-649" },
	{ "91-7000",       "91-7999" },
	{ "91-85000",      "91-94999" },
	{ "91-970000",     "91-999999" },

	/* International NGO Publishers and EU Organizations */
	{ "92-0",          "92-5" },
	{ "92-60",         "92-79" },
	{ "92-800",        "92-899" },
	{ "92-9000",       "92-9499" },
	{ "92-95000",      "92-98999" },
	{ "92-990000",     "92-999999" },

	/* India */
	{ "93-00",         "93-09" },
	{ "93-100",        "93-499" },
	{ "93-5000",       "93-7999" },
	{ "93-80000",      "93-94999" },
	{ "93-950000",     "93-999999" },

	/* Netherlands */
	{ "94-000",        "94-599" },
	{ "94-6000",       "94-8999" },
	{ "94-90000",      "94-99999" },

	/* Argentina */
	{ "950-00",        "950-49" },
	{ "950-500",       "950-899" },
	{ "950-9000",      "950-9899" },
	{ "950-99000",     "950-99999" },

	/* Finland */
	{ "951-0",         "951-1" },
	{ "951-20",        "951-54" },
	{ "951-550",       "951-889" },
	{ "951-8900",      "951-9499" },
	{ "951-95000",     "951-99999" },

	/* Finland */
	{ "952-00",        "952-19" },
	{ "952-200",       "952-499" },
	{ "952-5000",      "952-5999" },
	{ "952-60",        "952-65" },
	{ "952-6600",      "952-6699" },
	{ "952-67000",     "952-69999" },
	{ "952-7000",      "952-7999" },
	{ "952-80",        "952-94" },
	{ "952-9500",      "952-9899" },
	{ "952-99000",     "952-99999" },

	/* Croatia */
	{ "953-0",         "953-0" },
	{ "953-10",        "953-14" },
	{ "953-150",       "953-509" },
	{ "953-51",        "953-54" },
	{ "953-55000",     "953-59999" },
	{ "953-6000",      "953-9499" },
	{ "953-95000",     "953-99999" },

	/* Bulgaria */
	{ "954-00",        "954-28" },
	{ "954-2900",      "954-2999" },
	{ "954-300",       "954-799" },
	{ "954-8000",      "954-8999" },
	{ "954-90000",     "954-92999" },
	{ "954-9300",      "954-9999" },

	/* Sri Lanka */
	{ "955-0000",      "955-1999" },
	{ "955-20",        "955-40" },
	{ "955-41000",     "955-43999" },
	{ "955-44000",     "955-44999" },
	{ "955-4500",      "955-4999" },
	{ "955-50000",     "955-54999" },
	{ "955-550",       "955-749" },
	{ "955-7500",      "955-7999" },
	{ "955-8000",      "955-9499" },
	{ "955-95000",     "955-99999" },

	/* Chile */
	{ "956-00",        "956-19" },
	{ "956-200",       "956-699" },
	{ "956-7000",      "956-9999" },

	/* Taiwan */
	{ "957-00",        "957-02" },
	{ "957-0300",      "957-0499" },
	{ "957-05",        "957-19" },
	{ "957-2000",      "957-2099" },
	{ "957-21",        "957-27" },
	{ "957-28000",     "957-30999" },
	{ "957-31",        "957-43" },
	{ "957-440",       "957-819" },
	{ "957-8200",      "957-9699" },
	{ "957-97000",     "957-99999" },

	/* Colombia */
	{ "958-00",        "958-56" },
	{ "958-57000",     "958-59999" },
	{ "958-600",       "958-799" },
	{ "958-8000",      "958-9499" },
	{ "958-95000",     "958-99999" },

	/* Cuba */
	{ "959-00",        "959-19" },
	{ "959-200",       "959-699" },
	{ "959-7000",      "959-8499" },
	{ "959-85000",     "959-99999" },

	/* Greece */
	{ "960-00",        "960-19" },
	{ "960-200",       "960-659" },
	{ "960-6600",      "960-6899" },
	{ "960-690",       "960-699" },
	{ "960-7000",      "960-8499" },
	{ "960-85000",     "960-92999" },
	{ "960-93",        "960-93" },
	{ "960-9400",      "960-9799" },
	{ "960-98000",     "960-99999" },

	/* Slovenia */
	{ "961-00",        "961-19" },
	{ "961-200",       "961-599" },
	{ "961-6000",      "961-8999" },
	{ "961-90000",     "961-94999" },

	/* Hong Kong, China */
	{ "962-00",        "962-19" },
	{ "962-200",       "962-699" },
	{ "962-7000",      "962-8499" },
	{ "962-85000",     "962-86999" },
	{ "962-8700",      "962-8999" },
	{ "962-900",       "962-999" },

	/* Hungary */
	{ "963-00",        "963-19" },
	{ "963-200",       "963-699" },
	{ "963-7000",      "963-8499" },
	{ "963-85000",     "963-89999" },
	{ "963-9000",      "963-9999" },

	/* Iran */
	{ "964-00",        "964-14" },
	{ "964-150",       "964-249" },
	{ "964-2500",      "964-2999" },
	{ "964-300",       "964-549" },
	{ "964-5500",      "964-8999" },
	{ "964-90000",     "964-96999" },
	{ "964-970",       "964-989" },
	{ "964-9900",      "964-9999" },

	/* Israel */
	{ "965-00",        "965-19" },
	{ "965-200",       "965-599" },
	{ "965-7000",      "965-7999" },
	{ "965-90000",     "965-99999" },

	/* Ukraine */
	{ "966-00",        "966-12" },
	{ "966-130",       "966-139" },
	{ "966-14",        "966-14" },
	{ "966-1500",      "966-1699" },
	{ "966-170",       "966-199" },
	{ "966-2000",      "966-2789" },
	{ "966-279",       "966-289" },
	{ "966-2900",      "966-2999" },
	{ "966-300",       "966-699" },
	{ "966-7000",      "966-8999" },
	{ "966-90000",     "966-90999" },
	{ "966-910",       "966-949" },
	{ "966-95000",     "966-97999" },
	{ "966-980",       "966-999" },

	/* Malaysia */
	{ "967-00",        "967-00" },
	{ "967-0100",      "967-0999" },
	{ "967-10000",     "967-19999" },
	{ "967-300",       "967-499" },
	{ "967-5000",      "967-5999" },
	{ "967-60",        "967-89" },
	{ "967-900",       "967-989" },
	{ "967-9900",      "967-9989" },
	{ "967-99900",     "967-99999" },

	/* Mexico */
	{ "968-01",        "968-39" },
	{ "968-400",       "968-499" },
	{ "968-5000",      "968-7999" },
	{ "968-800",       "968-899" },
	{ "968-9000",      "968-9999" },

	/* Pakistan */
	{ "969-0",         "969-1" },
	{ "969-20",        "969-22" },
	{ "969-23000",     "969-23999" },
	{ "969-24",        "969-39" },
	{ "969-400",       "969-749" },
	{ "969-7500",      "969-9999" },

	/* Mexico */
	{ "970-01",        "970-59" },
	{ "970-600",       "970-899" },
	{ "970-9000",      "970-9099" },
	{ "970-91000",     "970-96999" },
	{ "970-9700",      "970-9999" },

	/* Philippines */
	{ "971-000",       "971-015" },
	{ "971-0160",      "971-0199" },
	{ "971-02",        "971-02" },
	{ "971-0300",      "971-0599" },
	{ "971-06",        "971-49" },
	{ "971-500",       "971-849" },
	{ "971-8500",      "971-9099" },
	{ "971-91000",     "971-95999" },
	{ "971-9600",      "971-9699" },
	{ "971-97",        "971-98" },
	{ "971-9900",      "971-9999" },

	/* Portugal */
	{ "972-0",         "972-1" },
	{ "972-20",        "972-54" },
	{ "972-550",       "972-799" },
	{ "972-8000",      "972-9499" },
	{ "972-95000",     "972-99999" },

	/* Romania */
	{ "973-0",         "973-0" },
	{ "973-100",       "973-169" },
	{ "973-1700",      "973-1999" },
	{ "973-20",        "973-54" },
	{ "973-550",       "973-759" },
	{ "973-7600",      "973-8499" },
	{ "973-85000",     "973-88999" },
	{ "973-8900",      "973-9499" },
	{ "973-95000",     "973-99999" },

	/* Thailand */
	{ "974-00",        "974-19" },
	{ "974-200",       "974-699" },
	{ "974-7000",      "974-8499" },
	{ "974-85000",     "974-89999" },
	{ "974-90000",     "974-94999" },
	{ "974-9500",      "974-9999" },

	/* Turkey */
	{ "975-00000",     "975-01999" },
	{ "975-02",        "975-24" },
	{ "975-250",       "975-599" },
	{ "975-6000",      "975-9199" },
	{ "975-92000",     "975-98999" },
	{ "975-990",       "975-999" },

	/* Caribbean Community */
	{ "976-0",         "976-3" },
	{ "976-40",        "976-59" },
	{ "976-600",       "976-799" },
	{ "976-8000",      "976-9499" },
	{ "976-95000",     "976-99999" },

	/* Egypt */
	{ "977-00",        "977-19" },
	{ "977-200",       "977-499" },
	{ "977-5000",      "977-6999" },
	{ "977-700",       "977-849" },
	{ "977-85000",     "977-89999" },
	{ "977-90",        "977-99" },

	/* Nigeria */
	{ "978-000",       "978-199" },
	{ "978-2000",      "978-2999" },
	{ "978-30000",     "978-79999" },
	{ "978-8000",      "978-8999" },
	{ "978-900",       "978-999" },

	/* Indonesia */
	{ "979-000",       "979-099" },
	{ "979-1000",      "979-1499" },
	{ "979-15000",     "979-19999" },
	{ "979-20",        "979-29" },
	{ "979-3000",      "979-3999" },
	{ "979-400",       "979-799" },
	{ "979-8000",      "979-9499" },
	{ "979-95000",     "979-99999" },

	/* Venezuela */
	{ "980-00",        "980-19" },
	{ "980-200",       "980-599" },
	{ "980-6000",      "980-9999" },

	/* Singapore */
	{ "981-00",        "981-16" },
	{ "981-17000",     "981-19999" },
	{ "981-200",       "981-299" },
	{ "981-3000",      "981-3099" },
	{ "981-310",       "981-399" },
	{ "981-4000",      "981-9999" },

	/* South Pacific */
	{ "982-00",        "982-09" },
	{ "982-100",       "982-699" },
	{ "982-70",        "982-89" },
	{ "982-9000",      "982-9799" },
	{ "982-98000",     "982-99999" },

	/* Malaysia */
	{ "983-00",        "983-01" },
	{ "983-020",       "983-199" },
	{ "983-2000",      "983-3999" },
	{ "983-40000",     "983-44999" },
	{ "983-45",        "983-49" },
	{ "983-50",        "983-79" },
	{ "983-800",       "983-899" },
	{ "983-9000",      "983-9899" },
	{ "983-99000",     "983-99999" },

	/* Bangladesh */
	{ "984-00",        "984-39" },
	{ "984-400",       "984-799" },
	{ "984-8000",      "984-8999" },
	{ "984-90000",     "984-99999" },

	/* Belarus */
	{ "985-00",        "985-39" },
	{ "985-400",       "985-599" },
	{ "985-6000",      "985-8999" },
	{ "985-90000",     "985-99999" },

	/* Taiwan */
	{ "986-00",        "986-11" },
	{ "986-120",       "986-559" },
	{ "986-5600",      "986-7999" },
	{ "986-80000",     "986-99999" },

	/* Argentina */
	{ "987-00",        "987-09" },
	{ "987-1000",      "987-1999" },
	{ "987-20000",     "987-29999" },
	{ "987-30",        "987-35" },
	{ "987-3600",      "987-3999" },
	{ "987-4000",      "987-4199" },
	{ "987-42",        "987-43" },
	{ "987-4400",      "987-4499" },
	{ "987-45000",     "987-48999" },
	{ "987-4900",      "987-4999" },
	{ "987-500",       "987-899" },
	{ "987-9000",      "987-9499" },
	{ "987-95000",     "987-99999" },

	/* Hong Kong, China */
	{ "988-00",        "988-11" },
	{ "988-12000",     "988-14999" },
	{ "988-15000",     "988-16999" },
	{ "988-17000",     "988-19999" },
	{ "988-200",       "988-769" },
	{ "988-77000",     "988-79999" },
	{ "988-8000",      "988-9699" },
	{ "988-97000",     "988-99999" },

	/* Portugal */
	{ "989-0",         "989-1" },
	{ "989-20",        "989-54" },
	{ "989-550",       "989-799" },
	{ "989-8000",      "989-9499" },
	{ "989-95000",     "989-99999" },

	/* Cambodia */
	{ "9924-30",       "9924-39" },
	{ "9924-500",      "9924-649" },
	{ "9924-9000",     "9924-9999" },

	/* Cyprus */
	{ "9925-0",        "9925-2" },
	{ "9925-30",       "9925-54" },
	{ "9925-550",      "9925-734" },
	{ "9925-7350",     "9925-9999" },

	/* Bosnia and Herzegovina */
	{ "9926-0",        "9926-1" },
	{ "9926-20",       "9926-39" },
	{ "9926-400",      "9926-799" },
	{ "9926-8000",     "9926-9999" },

	/* Qatar */
	{ "9927-00",       "9927-09" },
	{ "9927-100",      "9927-399" },
	{ "9927-4000",     "9927-4999" },

	/* Albania */
	{ "9928-00",       "9928-09" },
	{ "9928-100",      "9928-399" },
	{ "9928-4000",     "9928-4999" },

	/* Guatemala */
	{ "9929-0",        "9929-3" },
	{ "9929-40",       "9929-54" },
	{ "9929-550",      "9929-799" },
	{ "9929-8000",     "9929-9999" },

	/* Costa Rica */
	{ "9930-00",       "9930-49" },
	{ "9930-500",      "9930-939" },
	{ "9930-9400",     "9930-9999" },

	/* Algeria */
	{ "9931-00",       "9931-29" },
	{ "9931-300",      "9931-899" },
	{ "9931-9000",     "9931-9999" },

	/* Lao People's Democratic Republic */
	{ "9932-00",       "9932-39" },
	{ "9932-400",      "9932-849" },
	{ "9932-8500",     "9932-9999" },

	/* Syria */
	{ "9933-0",        "9933-0" },
	{ "9933-10",       "9933-39" },
	{ "9933-400",      "9933-899" },
	{ "9933-9000",     "9933-9999" },

	/* Latvia */
	{ "9934-0",        "9934-0" },
	{ "9934-10",       "9934-49" },
	{ "9934-500",      "9934-799" },
	{ "9934-8000",     "9934-9999" },

	/* Iceland */
	{ "9935-0",        "9935-0" },
	{ "9935-10",       "9935-39" },
	{ "9935-400",      "9935-899" },
	{ "9935-9000",     "9935-9999" },

	/* Afghanistan */
	{ "9936-0",        "9936-1" },
	{ "9936-20",       "9936-39" },
	{ "9936-400",      "9936-799" },
	{ "9936-8000",     "9936-9999" },

	/* Nepal */
	{ "9937-0",        "9937-2" },
	{ "9937-30",       "9937-49" },
	{ "9937-500",      "9937-799" },
	{ "9937-8000",     "9937-9999" },

	/* Tunisia */
	{ "9938-00",       "9938-79" },
	{ "9938-800",      "9938-949" },
	{ "9938-9500",     "9938-9999" },

	/* Armenia */
	{ "9939-0",        "9939-4" },
	{ "9939-50",       "9939-79" },
	{ "9939-800",      "9939-899" },
	{ "9939-9000",     "9939-9999" },

	/* Montenegro */
	{ "9940-0",        "9940-1" },
	{ "9940-20",       "9940-49" },
	{ "9940-500",      "9940-899" },
	{ "9940-9000",     "9940-9999" },

	/* Georgia */
	{ "9941-0",        "9941-0" },
	{ "9941-10",       "9941-39" },
	{ "9941-400",      "9941-899" },
	{ "9941-9000",     "9941-9999" },

	/* Ecuador */
	{ "9942-00",       "9942-84" },
	{ "9942-8500",     "9942-8999" },
	{ "9942-900",      "9942-984" },
	{ "9942-9850",     "9942-9999" },

	/* Uzbekistan */
	{ "9943-00",       "9943-29" },
	{ "9943-300",      "9943-399" },
	{ "9943-4000",     "9943-9749" },
	{ "9943-975",      "9943-999" },

	/* Turkey */
	{ "9944-0000",     "9944-0999" },
	{ "9944-100",      "9944-499" },
	{ "9944-5000",     "9944-5999" },
	{ "9944-60",       "9944-69" },
	{ "9944-700",      "9944-799" },
	{ "9944-80",       "9944-89" },
	{ "9944-900",      "9944-999" },

	/* Dominican Republic */
	{ "9945-00",       "9945-00" },
	{ "9945-010",      "9945-079" },
	{ "9945-08",       "9945-39" },
	{ "9945-400",      "9945-569" },
	{ "9945-57",       "9945-57" },
	{ "9945-580",      "9945-849" },
	{ "9945-8500",     "9945-9999" },

	/* Korea, P.D.R. */
	{ "9946-0",        "9946-1" },
	{ "9946-20",       "9946-39" },
	{ "9946-400",      "9946-899" },
	{ "9946-9000",     "9946-9999" },

	/* Algeria */
	{ "9947-0",        "9947-1" },
	{ "9947-20",       "9947-79" },
	{ "9947-800",      "9947-999" },

	/* United Arab Emirates */
	{ "9948-00",       "9948-39" },
	{ "9948-400",      "9948-849" },
	{ "9948-8500",     "9948-9999" },

	/* Estonia */
	{ "9949-0",        "9949-0" },
	{ "9949-10",       "9949-39" },
	{ "9949-400",      "9949-749" },
	{ "9949-75",       "9949-89" },
	{ "9949-9000",     "9949-9999" },

	/* Palestine */
	{ "9950-00",       "9950-29" },
	{ "9950-300",      "9950-849" },
	{ "9950-8500",     "9950-9999" },

	/* Kosova */
	{ "9951-00",       "9951-39" },
	{ "9951-400",      "9951-849" },
	{ "9951-8500",     "9951-9999" },

	/* Azerbaijan */
	{ "9952-0",        "9952-1" },
	{ "9952-20",       "9952-39" },
	{ "9952-400",      "9952-799" },
	{ "9952-8000",     "9952-9999" },

	/* Lebanon */
	{ "9953-0",        "9953-0" },
	{ "9953-10",       "9953-39" },
	{ "9953-400",      "9953-599" },
	{ "9953-60",       "9953-89" },
	{ "9953-9000",     "9953-9999" },

	/* Morocco */
	{ "9954-0",        "9954-1" },
	{ "9954-20",       "9954-39" },
	{ "9954-400",      "9954-799" },
	{ "9954-8000",     "9954-9999" },

	/* Lithuania */
	{ "9955-00",       "9955-39" },
	{ "9955-400",      "9955-929" },
	{ "9955-9300",     "9955-9999" },

	/* Cameroon */
	{ "9956-0",        "9956-0" },
	{ "9956-10",       "9956-39" },
	{ "9956-400",      "9956-899" },
	{ "9956-9000",     "9956-9999" },

	/* Jordan */
	{ "9957-00",       "9957-39" },
	{ "9957-400",      "9957-649" },
	{ "9957-65",       "9957-69" },
	{ "9957-70",       "9957-84" },
	{ "9957-8500",     "9957-8799" },
	{ "9957-88",       "9957-99" },

	/* Bosnia and Herzegovina */
	{ "9958-00",       "9958-01" },
	{ "9958-020",      "9958-029" },
	{ "9958-0300",     "9958-0399" },
	{ "9958-040",      "9958-089" },
	{ "9958-0900",     "9958-0999" },
	{ "9958-10",       "9958-18" },
	{ "9958-1900",     "9958-1999" },
	{ "9958-20",       "9958-49" },
	{ "9958-500",      "9958-899" },
	{ "9958-9000",     "9958-9999" },

	/* Libya */
	{ "9959-0",        "9959-1" },
	{ "9959-20",       "9959-79" },
	{ "9959-800",      "9959-949" },
	{ "9959-9500",     "9959-9699" },
	{ "9959-970",      "9959-979" },
	{ "9959-98",       "9959-99" },

	/* Saudi Arabia */
	{ "9960-00",       "9960-59" },
	{ "9960-600",      "9960-899" },
	{ "9960-9000",     "9960-9999" },

	/* Algeria */
	{ "9961-0",        "9961-2" },
	{ "9961-30",       "9961-69" },
	{ "9961-700",      "9961-949" },
	{ "9961-9500",     "9961-9999" },

	/* Panama */
	{ "9962-00",       "9962-54" },
	{ "9962-5500",     "9962-5599" },
	{ "9962-56",       "9962-59" },
	{ "9962-600",      "9962-849" },
	{ "9962-8500",     "9962-9999" },

	/* Cyprus */
	{ "9963-0",        "9963-1" },
	{ "9963-2000",     "9963-2499" },
	{ "9963-250",      "9963-279" },
	{ "9963-2800",     "9963-2999" },
	{ "9963-30",       "9963-54" },
	{ "9963-550",      "9963-734" },
	{ "9963-7350",     "9963-7499" },
	{ "9963-7500",     "9963-9999" },

	/* Ghana */
	{ "9964-0",        "9964-6" },
	{ "9964-70",       "9964-94" },
	{ "9964-950",      "9964-999" },

	/* Kazakhstan */
	{ "9965-00",       "9965-39" },
	{ "9965-400",      "9965-899" },
	{ "9965-9000",     "9965-9999" },

	/* Kenya */
	{ "9966-000",      "9966-149" },
	{ "9966-1500",     "9966-1999" },
	{ "9966-20",       "9966-69" },
	{ "9966-7000",     "9966-7499" },
	{ "9966-750",      "9966-959" },
	{ "9966-9600",     "9966-9999" },

	/* Kyrgyz Republic */
	{ "9967-00",       "9967-39" },
	{ "9967-400",      "9967-899" },
	{ "9967-9000",     "9967-9999" },

	/* Costa Rica */
	{ "9968-00",       "9968-49" },
	{ "9968-500",      "9968-939" },
	{ "9968-9400",     "9968-9999" },

	/* Uganda */
	{ "9970-00",       "9970-39" },
	{ "9970-400",      "9970-899" },
	{ "9970-9000",     "9970-9999" },

	/* Singapore */
	{ "9971-0",        "9971-5" },
	{ "9971-60",       "9971-89" },
	{ "9971-900",      "9971-989" },
	{ "9971-9900",     "9971-9999" },

	/* Peru */
	{ "9972-00",       "9972-09" },
	{ "9972-1",        "9972-1" },
	{ "9972-200",      "9972-249" },
	{ "9972-2500",     "9972-2999" },
	{ "9972-30",       "9972-59" },
	{ "9972-600",      "9972-899" },
	{ "9972-9000",     "9972-9999" },

	/* Tunisia */
	{ "9973-00",       "9973-05" },
	{ "9973-060",      "9973-089" },
	{ "9973-0900",     "9973-0999" },
	{ "9973-10",       "9973-69" },
	{ "9973-700",      "9973-969" },
	{ "9973-9700",     "9973-9999" },

	/* Uruguay */
	{ "9974-0",        "9974-2" },
	{ "9974-30",       "9974-54" },
	{ "9974-550",      "9974-749" },
	{ "9974-7500",     "9974-9099" },
	{ "9974-91",       "9974-94" },
	{ "9974-95",       "9974-99" },

	/* Moldova */
	{ "9975-0",        "9975-0" },
	{ "9975-100",      "9975-299" },
	{ "9975-3000",     "9975-3999" },
	{ "9975-4000",     "9975-4499" },
	{ "9975-45",       "9975-89" },
	{ "9975-900",      "9975-949" },
	{ "9975-9500",     "9975-9999" },

	/* Tanzania */
	{ "9976-0",        "9976-5" },
	{ "9976-60",       "9976-89" },
	{ "9976-900",      "9976-989" },
	{ "9976-9900",     "9976-9999" },

	/* Costa Rica */
	{ "9977-00",       "9977-89" },
	{ "9977-900",      "9977-989" },
	{ "9977-9900",     "9977-9999" },

	/* Ecuador */
	{ "9978-00",       "9978-29" },
	{ "9978-300",      "9978-399" },
	{ "9978-40",       "9978-94" },
	{ "9978-950",      "9978-989" },
	{ "9978-9900",     "9978-9999" },

	/* Iceland */
	{ "9979-0",        "9979-4" },
	{ "9979-50",       "9979-64" },
	{ "9979-650",      "9979-659" },
	{ "9979-66",       "9979-75" },
	{ "9979-760",      "9979-899" },
	{ "9979-9000",     "9979-9999" },

	/* Papua New Guinea */
	{ "9980-0",        "9980-3" },
	{ "9980-40",       "9980-89" },
	{ "9980-900",      "9980-989" },
	{ "9980-9900",     "9980-9999" },

	/* Morocco */
	{ "9981-00",       "9981-09" },
	{ "9981-100",      "9981-159" },
	{ "9981-1600",     "9981-1999" },
	{ "9981-20",       "9981-79" },
	{ "9981-800",      "9981-949" },
	{ "9981-9500",     "9981-9999" },

	/* Zambia */
	{ "9982-00",       "9982-79" },
	{ "9982-800",      "9982-989" },
	{ "9982-9900",     "9982-9999" },

	/* Gambia */
	{ "9983-80",       "9983-94" },
	{ "9983-950",      "9983-989" },
	{ "9983-9900",     "9983-9999" },

	/* Latvia */
	{ "9984-00",       "9984-49" },
	{ "9984-500",      "9984-899" },
	{ "9984-9000",     "9984-9999" },

	/* Estonia */
	{ "9985-0",        "9985-4" },
	{ "9985-50",       "9985-79" },
	{ "9985-800",      "9985-899" },
	{ "9985-9000",     "9985-9999" },

	/* Lithuania */
	{ "9986-00",       "9986-39" },
	{ "9986-400",      "9986-899" },
	{ "9986-9000",     "9986-9399" },
	{ "9986-940",      "9986-969" },
	{ "9986-97",       "9986-99" },

	/* Tanzania */
	{ "9987-00",       "9987-39" },
	{ "9987-400",      "9987-879" },
	{ "9987-8800",     "9987-9999" },

	/* Ghana */
	{ "9988-0",        "9988-2" },
	{ "9988-30",       "9988-54" },
	{ "9988-550",      "9988-749" },
	{ "9988-7500",     "9988-9999" },

	/* Macedonia */
	{ "9989-0",        "9989-0" },
	{ "9989-100",      "9989-199" },
	{ "9989-2000",     "9989-2999" },
	{ "9989-30",       "9989-59" },
	{ "9989-600",      "9989-949" },
	{ "9989-9500",     "9989-9999" },

	/* Bahrain */
	{ "99901-00",      "99901-49" },
	{ "99901-500",     "99901-799" },
	{ "99901-80",      "99901-99" },

	/* Reserved Agency */
	{ "99902",         "99902" },

	/* Mauritius */
	{ "99903-0",       "99903-1" },
	{ "99903-20",      "99903-89" },
	{ "99903-900",     "99903-999" },

	/* Curaçao */
	{ "99904-0",       "99904-5" },
	{ "99904-60",      "99904-89" },
	{ "99904-900",     "99904-999" },

	/* Bolivia */
	{ "99905-0",       "99905-3" },
	{ "99905-40",      "99905-79" },
	{ "99905-800",     "99905-999" },

	/* Kuwait */
	{ "99906-0",       "99906-2" },
	{ "99906-30",      "99906-59" },
	{ "99906-600",     "99906-699" },
	{ "99906-70",      "99906-89" },
	{ "99906-90",      "99906-94" },
	{ "99906-950",     "99906-999" },

	/* Malawi */
	{ "99908-0",       "99908-0" },
	{ "99908-10",      "99908-89" },
	{ "99908-900",     "99908-999" },

	/* Malta */
	{ "99909-0",       "99909-3" },
	{ "99909-40",      "99909-94" },
	{ "99909-950",     "99909-999" },

	/* Sierra Leone */
	{ "99910-0",       "99910-2" },
	{ "99910-30",      "99910-89" },
	{ "99910-900",     "99910-999" },

	/* Lesotho */
	{ "99911-00",      "99911-59" },
	{ "99911-600",     "99911-999" },

	/* Botswana */
	{ "99912-0",       "99912-3" },
	{ "99912-400",     "99912-599" },
	{ "99912-60",      "99912-89" },
	{ "99912-900",     "99912-999" },

	/* Andorra */
	{ "99913-0",       "99913-2" },
	{ "99913-30",      "99913-35" },
	{ "99913-600",     "99913-604" },

	/* Suriname */
	{ "99914-0",       "99914-4" },
	{ "99914-50",      "99914-89" },
	{ "99914-900",     "99914-999" },

	/* Maldives */
	{ "99915-0",       "99915-4" },
	{ "99915-50",      "99915-79" },
	{ "99915-800",     "99915-999" },

	/* Namibia */
	{ "99916-0",       "99916-2" },
	{ "99916-30",      "99916-69" },
	{ "99916-700",     "99916-999" },

	/* Brunei Darussalam */
	{ "99917-0",       "99917-2" },
	{ "99917-30",      "99917-89" },
	{ "99917-900",     "99917-999" },

	/* Faroe Islands */
	{ "99918-0",       "99918-3" },
	{ "99918-40",      "99918-79" },
	{ "99918-800",     "99918-999" },

	/* Benin */
	{ "99919-0",       "99919-2" },
	{ "99919-300",     "99919-399" },
	{ "99919-40",      "99919-69" },
	{ "99919-70",      "99919-79" },
	{ "99919-800",     "99919-849" },
	{ "99919-850",     "99919-899" },
	{ "99919-900",     "99919-999" },

	/* Andorra */
	{ "99920-0",       "99920-4" },
	{ "99920-50",      "99920-89" },
	{ "99920-900",     "99920-999" },

	/* Qatar */
	{ "99921-0",       "99921-1" },
	{ "99921-20",      "99921-69" },
	{ "99921-700",     "99921-799" },
	{ "99921-8",       "99921-8" },
	{ "99921-90",      "99921-99" },

	/* Guatemala */
	{ "99922-0",       "99922-3" },
	{ "99922-40",      "99922-69" },
	{ "99922-700",     "99922-999" },

	/* El Salvador */
	{ "99923-0",       "99923-1" },
	{ "99923-20",      "99923-79" },
	{ "99923-800",     "99923-999" },

	/* Nicaragua */
	{ "99924-0",       "99924-1" },
	{ "99924-20",      "99924-79" },
	{ "99924-800",     "99924-999" },

	/* Paraguay */
	{ "99925-0",       "99925-3" },
	{ "99925-40",      "99925-79" },
	{ "99925-800",     "99925-999" },

	/* Honduras */
	{ "99926-0",       "99926-0" },
	{ "99926-10",      "99926-59" },
	{ "99926-600",     "99926-869" },
	{ "99926-87",      "99926-89" },
	{ "99926-90",      "99926-99" },

	/* Albania */
	{ "99927-0",       "99927-2" },
	{ "99927-30",      "99927-59" },
	{ "99927-600",     "99927-999" },

	/* Georgia */
	{ "99928-0",       "99928-0" },
	{ "99928-10",      "99928-79" },
	{ "99928-800",     "99928-999" },

	/* Mongolia */
	{ "99929-0",       "99929-4" },
	{ "99929-50",      "99929-79" },
	{ "99929-800",     "99929-999" },

	/* Armenia */
	{ "99930-0",       "99930-4" },
	{ "99930-50",      "99930-79" },
	{ "99930-800",     "99930-999" },

	/* Seychelles */
	{ "99931-0",       "99931-4" },
	{ "99931-50",      "99931-79" },
	{ "99931-800",     "99931-999" },

	/* Malta */
	{ "99932-0",       "99932-0" },
	{ "99932-10",      "99932-59" },
	{ "99932-600",     "99932-699" },
	{ "99932-7",       "99932-7" },
	{ "99932-80",      "99932-99" },

	/* Nepal */
	{ "99933-0",       "99933-2" },
	{ "99933-30",      "99933-59" },
	{ "99933-600",     "99933-999" },

	/* Dominican Republic */
	{ "99934-0",       "99934-1" },
	{ "99934-20",      "99934-79" },
	{ "99934-800",     "99934-999" },

	/* Haiti */
	{ "99935-0",       "99935-2" },
	{ "99935-30",      "99935-59" },
	{ "99935-600",     "99935-699" },
	{ "99935-7",       "99935-8" },
	{ "99935-90",      "99935-99" },

	/* Bhutan */
	{ "99936-0",       "99936-0" },
	{ "99936-10",      "99936-59" },
	{ "99936-600",     "99936-999" },

	/* Macau */
	{ "99937-0",       "99937-1" },
	{ "99937-20",      "99937-59" },
	{ "99937-600",     "99937-999" },

	/* Srpska, Republic of */
	{ "99938-0",       "99938-1" },
	{ "99938-20",      "99938-59" },
	{ "99938-600",     "99938-899" },
	{ "99938-90",      "99938-99" },

	/* Guatemala */
	{ "99939-0",       "99939-5" },
	{ "99939-60",      "99939-89" },
	{ "99939-900",     "99939-999" },

	/* Georgia */
	{ "99940-0",       "99940-0" },
	{ "99940-10",      "99940-69" },
	{ "99940-700",     "99940-999" },

	/* Armenia */
	{ "99941-0",       "99941-2" },
	{ "99941-30",      "99941-79" },
	{ "99941-800",     "99941-999" },

	/* Sudan */
	{ "99942-0",       "99942-4" },
	{ "99942-50",      "99942-79" },
	{ "99942-800",     "99942-999" },

	/* Albania */
	{ "99943-0",       "99943-2" },
	{ "99943-30",      "99943-59" },
	{ "99943-600",     "99943-999" },

	/* Ethiopia */
	{ "99944-0",       "99944-4" },
	{ "99944-50",      "99944-79" },
	{ "99944-800",     "99944-999" },

	/* Namibia */
	{ "99945-0",       "99945-5" },
	{ "99945-60",      "99945-89" },
	{ "99945-900",     "99945-999" },

	/* Nepal */
	{ "99946-0",       "99946-2" },
	{ "99946-30",      "99946-59" },
	{ "99946-600",     "99946-999" },

	/* Tajikistan */
	{ "99947-0",       "99947-2" },
	{ "99947-30",      "99947-69" },
	{ "99947-700",     "99947-999" },

	/* Eritrea */
	{ "99948-0",       "99948-4" },
	{ "99948-50",      "99948-79" },
	{ "99948-800",     "99948-999" },

	/* Mauritius */
	{ "99949-0",       "99949-1" },
	{ "99949-20",      "99949-89" },
	{ "99949-900",     "99949-999" },

	/* Cambodia */
	{ "99950-0",       "99950-4" },
	{ "99950-50",      "99950-79" },
	{ "99950-800",     "99950-999" },

	/* Reserved Agency */
	{ "99951",         "99951" },

	/* Mali */
	{ "99952-0",       "99952-4" },
	{ "99952-50",      "99952-79" },
	{ "99952-800",     "99952-999" },

	/* Paraguay */
	{ "99953-0",       "99953-2" },
	{ "99953-30",      "99953-79" },
	{ "99953-800",     "99953-939" },
	{ "99953-94",      "99953-99" },

	/* Bolivia */
	{ "99954-0",       "99954-2" },
	{ "99954-30",      "99954-69" },
	{ "99954-700",     "99954-879" },
	{ "99954-88",      "99954-99" },

	/* Srpska, Republic of */
	{ "99955-0",       "99955-1" },
	{ "99955-20",      "99955-59" },
	{ "99955-600",     "99955-799" },
	{ "99955-80",      "99955-99" },

	/* Albania */
	{ "99956-00",      "99956-59" },
	{ "99956-600",     "99956-859" },
	{ "99956-86",      "99956-99" },

	/* Malta */
	{ "99957-0",       "99957-1" },
	{ "99957-20",      "99957-79" },
	{ "99957-800",     "99957-999" },

	/* Bahrain */
	{ "99958-0",       "99958-4" },
	{ "99958-50",      "99958-93" },
	{ "99958-940",     "99958-949" },
	{ "99958-950",     "99958-999" },

	/* Luxembourg */
	{ "99959-0",       "99959-2" },
	{ "99959-30",      "99959-59" },
	{ "99959-600",     "99959-999" },

	/* Malawi */
	{ "99960-0",       "99960-0" },
	{ "99960-10",      "99960-94" },
	{ "99960-950",     "99960-999" },

	/* El Salvador */
	{ "99961-0",       "99961-3" },
	{ "99961-40",      "99961-89" },
	{ "99961-900",     "99961-999" },

	/* Mongolia */
	{ "99962-0",       "99962-4" },
	{ "99962-50",      "99962-79" },
	{ "99962-800",     "99962-999" },

	/* Cambodia */
	{ "99963-00",      "99963-49" },
	{ "99963-500",     "99963-919" },
	{ "99963-92",      "99963-99" },

	/* Nicaragua */
	{ "99964-0",       "99964-1" },
	{ "99964-20",      "99964-79" },
	{ "99964-800",     "99964-999" },

	/* Macau */
	{ "99965-0",       "99965-3" },
	{ "99965-40",      "99965-62" },
	{ "99965-630",     "99965-999" },

	/* Kuwait */
	{ "99966-0",       "99966-2" },
	{ "99966-30",      "99966-69" },
	{ "99966-700",     "99966-799" },
	{ "99966-80",      "99966-94" },

	/* Paraguay */
	{ "99967-0",       "99967-1" },
	{ "99967-20",      "99967-59" },
	{ "99967-600",     "99967-899" },

	/* Botswana */
	{ "99968-0",       "99968-3" },
	{ "99968-400",     "99968-599" },
	{ "99968-60",      "99968-89" },
	{ "99968-900",     "99968-999" },

	/* Oman */
	{ "99969-0",       "99969-4" },
	{ "99969-50",      "99969-79" },
	{ "99969-800",     "99969-999" },

	/* Haiti */
	{ "99970-0",       "99970-4" },
	{ "99970-50",      "99970-89" },
	{ "99970-900",     "99970-999" },

	/* Myanmar */
	{ "99971-0",       "99971-5" },
	{ "99971-60",      "99971-84" },
	{ "99971-850",     "99971-999" },

	/* Faroe Islands */
	{ "99972-0",       "99972-4" },
	{ "99972-50",      "99972-89" },
	{ "99972-900",     "99972-999" },

	/* Mongolia */
	{ "99973-0",       "99973-3" },
	{ "99973-40",      "99973-79" },
	{ "99973-800",     "99973-999" },

	/* Bolivia */
	{ "99974-40",      "99974-79" },
	{ "99974-800",     "99974-999" },

	/* Tajikistan */
	{ "99975-0",       "99975-3" },
	{ "99975-40",      "99975-79" },
	{ "99975-800",     "99975-999" },

	/* Srpska, Republic of */
	{ "99976-0",       "99976-1" },
	{ "99976-20",      "99976-59" },
	{ "99976-600",     "99976-799" },

	/* Rwanda */
	{ "99977-0",       "99977-1" },
	{ "99977-40",      "99977-69" },
	{ "99977-700",     "99977-799" },

	/* From http://www.isbn-international.org/en/agencies.html:
	   The publisher prefix ranges in the English group (U.S., U.K.,
	   Canada, Australia, New Zealand, etc) are as follows:

	   Group                   If Number Ranges         Insert Hyphens
	   Identifier "0"           are Between              After
	   -------------------------------------------------------------------
	   00.......19                00-19             1st  3rd  9th digit
	   200......699               20-69              "   4th       "
	   7000.....8499              70-84              "   5th       "
	   85000....89999             85-89              "   6th       "
	   900000...949999            90-94              "   7th       "
	   9500000..9999999           95-99              "   8th       "



	   Group                   If Number Ranges        Insert Hyphens
	   Identifier "1"           are Between             After
	   -------------------------------------------------------------------
	   55000....86979           5500-8697           1st  6th  9th digit
	   869800...998999          8698-9989            "   7th       "
	   9990000..9999999         9990-9999            "   8th       "



	   The following table gives the range distribution of the group identifiers:

	   0 - 7
	   80 - 94
	   950 - 995
	   9960 - 9989
	   99900 - 99999

	   These ranges are already partially covered by ranges above, but
	   we include them here for completeness, unless they are identical
	   to one above, in which case, they are commented out. I have yet
	   to find comparable data for other language groups. */

	{ (const char*)NULL,        (const char*)NULL }        /* do not delete ! */
    };

    /* separate table for prefix 979- */
    static range_t ISBN_range979[] = {
     /* France */
     { "10-00",            "10-19" },
     { "10-200",           "10-699" },
     { "10-7000",          "10-8999" },
     { "10-90000",         "10-97599" },
     { "10-976000",        "10-999999" },

     /* Korea, Republic */
     { "11-00" ,           "11-24" },
     { "11-250" ,          "11-549" },
     { "11-5500" ,         "11-8499" },
     { "11-85000" ,        "11-94999" },
     { "11-950000" ,       "11-999999" },

     /* Italy */
     { "12-200",           "12-200" },

     { (const char*)NULL,        (const char*)NULL }        /* do not delete ! */
    };

    range_p actualISBNrange;
    if (979 == iPrefix) {
	actualISBNrange=ISBN_range979;
    } else {
	actualISBNrange=ISBN_range978;
    }

    *out = '\0';
    /* with ISNBN-13, copy the first 3 characters */
    if (ISBNlen > PM_ISBN_SIZE) {
	strncpy(out, ISBN, 3);
	out[3] = PM_ISBN_HYPHEN;
	ISBN +=3;
	out[4]='\0';
    }
    for (i = 0; (actualISBNrange[i].begin != (const char*)NULL); ++i) {
	if (in_ISBN_range(actualISBNrange[i].begin, ISBN, actualISBNrange[i].end)
		== 0) {
	    return (hyphenate_one_ISBN(out, actualISBNrange[i].begin, ISBN));
	}
    }
    /* no fitting range found use default pattern */

    (void) hyphenate_one_ISBN(out, "", ISBN);
    return -1;
} /* fix_ISBN() */

/*------------------------------------------------------------------------
 *  Function    : PM_hyphenate_isbn
 *  Purpose     : Put hyphens in ISBN if not yet present
 *  Method      : The Input ISBN is compared with all known ISBN ranges.
 *  Author(s)   : Bas de Bakker
 *  Creation    : 11-11-1997
 *
 *  History:
 *  11-11-1997  : BdB      : creation
 *  24-10-2003  : huug	   : new rule in PM_hyphenate_isbn:
 *			     908[0-4]nnnnnx -> 90-8[0-4]nnnn-n-x
 *			     908[5-9]nnnnmx -> 90-8[5-9]nn-nnn-x
 *			     old rule was:
 *			     908nnnnnnx -> 90-8nnnnn-n-x
 *  02-09-2004  : paul	   : implemented acceptation of EAN.UCC-13 ISBN's
 *  18-10-2004	: huug	   : treat 9978 ISBNs like 9977 ISBNs
 *  24-10-2012	: huug     : CBS-4154 handle (978)3* (German language) ISBNs 
 *			     properly
 *  28-11-2012	: huug     : All new algorithm, based on official ISBN ranges
 *  29-10-2015	: paul     : CBS-6374 No functional change:
 *                           replaced 10 by PM_ISBN_SIZE
 *                           replaced 13 by PM_BOOKLAND_EAN_SIZE
 *
 *  Notes       :
 *  The ranges for ISBN's can be found on the international ISBN website:
 *  http://www.isbn-international.org/
 *  It is recommended that the ranges are updated regularly.
 *
 *  Desciption	:
 *  PM_hyphenate_isbn will put hyphens in an unhyphenated ISBN. If the input
 *  string already contains any hyphens, the output string will be equal to
 *  the input string.
 *
 *  The header file picamisc.h defines the constant PM_ISBN_LEN_MAX, which is
 *  the maximum size that a hyphenated ISBN can have.
 *
 *  Returns     :
 *  Number of characters written in 'out' (including '\0')
 *  or PM_ISBN_ERROR when the algorithm doesn't like your input
 *  or outsiz is too low.
 *------------------------------------------------------------------------
 */
size_t
PM_hyphenate_isbn(
	const char  *in,
	char	    *out,
	size_t	    outsiz)
{
    int inlen = strlen(in);

    int prefix = 978;
    if (strchr(in, PM_ISBN_HYPHEN) != NULL) {
        /* Already hyphenated, just copy to output, if possible. */
        if (inlen > outsiz - 1) {
            return PM_ISBN_ERROR; /* does not fit */
        }
        (void)memcpy(out, in, inlen);
        out[inlen] = '\0';
        return inlen + 1;
    }
    if (inlen == PM_ISBN_SIZE) {
	if (outsiz < PM_ISBN_SIZE + 4) return PM_ISBN_ERROR;
    } else if (inlen == PM_BOOKLAND_EAN_SIZE) {
	if (outsiz < PM_BOOKLAND_EAN_SIZE + 5) return PM_ISBN_ERROR;
	if (sscanf(in, "%3d", &prefix) != 1) {
	    return PM_ISBN_ERROR;
	}
    } else {
	return PM_ISBN_ERROR; /* improper input length */
    }
    return fix_ISBN(in, inlen, out, prefix);
} /* PM_hyphenate_isbn() */

/*------------------------------------------------------------------------
 *  Function    : PM_hyphenate_issn
 *  Purpose     : Put hyphens in ISSN if not yet present
 *  Method      : Insert hyphen after 4th character (in paranoid way).
 *  Author(s)   : Bas de Bakker
 *  Creation    : 11-11-1997
 *
 *  History:
 *  11-11-1997  : BdB      : creation
 *
 *  Notes       :
 *  Even if the output buffer is too short, it will still be null-terminated
 *  after this routine returns.
 *
 *  Complaints  : Safe string manipulation in C is such a bother.
 *
 *  Desciption	:
 *  PM_hyphenate_issn will put hyphens in an unhyphenated ISSN. If the input
 *  string already contains any hyphens, the output string will be equal to
 *  the input string.
 *
 *  Returns     : Number of characters written in 'out' (including '\0').
 *------------------------------------------------------------------------
 */
size_t
PM_hyphenate_issn(const char *in, char *out, size_t outsiz)
{
    size_t insiz;
    size_t outwritten;
    size_t tocopy;

    insiz = strlen(in);
    --outsiz;                   /* Reserve byte for terminating '\0'. */
    outwritten = 0;

    if (strchr(in, PM_ISBN_HYPHEN) == NULL) {
        /* Not yet hyphenated, copy 4 characters and add PM_ISBN_HYPHEN. */
        tocopy = 4;
        if (tocopy > insiz) {
            tocopy = insiz;
        }
        if (tocopy > outsiz) {
            tocopy = outsiz;
        }
        (void)memcpy(out, in, tocopy);
        in += tocopy;
        insiz -= tocopy;
        out += tocopy;
        outsiz -= tocopy;
        outwritten += tocopy;

        /* Add PM_ISBN_HYPHEN if space available. */
        if (outsiz != 0) {
            *out = PM_ISBN_HYPHEN;
            ++out;
            --outsiz;
            ++outwritten;
        }
    }

    /* Copy the rest (or all if the string already included a hyphen). */
    tocopy = insiz;
    if (tocopy > outsiz) {
        tocopy = outsiz;
    }
    (void)memcpy(out, in, tocopy);
    out += tocopy;
    outwritten += tocopy;

    /* Add terminating '\0'. */
    *out = '\0';
    ++outwritten;

    return outwritten;
} /* PM_hyphenate_issn() */

/*------------------------------------------------------------------------
 *  Function	: PM_isbn_10_to_13
 *  Purpose	: Convert a 10-digit ISBN into a 13-digit ISBN
 *                (also known as a Bookland EAN number)
 *  Method	:
 *
 *  History:
 *  07-03-2008	: sutherlr : creation
 *
 *  Notes	:
 *  This function allocates no memory. Buffers of the necessary sizes must be
 *  pre-allocated.
 *
 *  This function conforms to the rules found in the publication "ISBN-13 for
 *  Dummies" (http://www.bisg.org/isbn-13/for.dummies.html)
 *
 *  Resulting isbn13 string is unhyphenated.
 *
 *  Returns	: TRUE on successful conversion, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
PM_isbn_10_to_13(
	const char *isbn10, 
	char       *isbn13,
	size_t      isbn13size) 
{
    int i;
    char loopChar;
    char checkdigit;

    if ((isbn10 == NULL) || (isbn13 == NULL)) {
	return FALSE;
    }

    if (strlen(isbn10) < PM_ISBN_SIZE) {
	return FALSE;
    }

    if (isbn13size < PM_BOOKLAND_EAN_SIZE + 1 /*include space for NULL byte*/ ) {
	return FALSE;
    }

    if (!PM_Mod11Test(isbn10, PM_ISBN_SIZE)) {
	return FALSE;
    }

    isbn13[0] = '9';
    isbn13[1] = '8';
    isbn13[2] = '7';

    for (i=0; i<9; i++) {
	loopChar = isbn10[i];
	if (!isdigit(loopChar)) {
	    break;
	}
	isbn13[i+3] = loopChar;
    }
    if (i!=9) {
	return FALSE;
    }

    checkdigit = PM_EAN_gen(isbn13, 12);
    isbn13[i+3] = checkdigit;
    isbn13[PM_BOOKLAND_EAN_SIZE] = '\0';
    return TRUE;
} /* PM_isbn_10_to_13() */

/*------------------------------------------------------------------------
 *  Function	: PM_isbn_13_to_10
 *  Purpose	: Convert a 13-digit ISBN (Bookland EAN number) into a
 *                10-digit ISBN
 *  Method	:
 *
 *  History:
 *  07-03-2008	: sutherlr : creation
 *  29-10-2015	: paul     : CBS-6374 No functional change:
 *                           replaced 10 by PM_ISBN_SIZE
 *                           replaced 13 by PM_BOOKLAND_EAN_SIZE
 *
 *  Notes	:
 *  This function allocates no memory. Buffers of the necessary sizes must be
 *  pre-allocated.
 *
 *  This function conforms to the rules found in the publication "ISBN-13 for
 *  Dummies" (http://www.bisg.org/isbn-13/for.dummies.html)
 *
 *  Only ISBN13's beginning with prefix "978" may be converted back to
 *  ISBN10's.
 *
 *  Resulting isbn10 string is unhyphenated.
 *
 *  Returns	: TRUE on successful conversion, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
PM_isbn_13_to_10(
	const char *isbn13, 
	char       *isbn10,
	size_t      isbn10size) 
{
    int i;
    char loopChar;
    char checkdigit;

    size_t len;

    if ((isbn13 == NULL) || (isbn10 == NULL)) {
	return FALSE;
    }

    len = strlen(isbn13);
    if (len < PM_BOOKLAND_EAN_SIZE) {
	return FALSE;
    }

    if (isbn10size < PM_ISBN_SIZE + 1 /*include space for NULL byte*/ ) {
	return FALSE;
    }

    if (!PM_EAN_test(isbn13, PM_BOOKLAND_EAN_SIZE)) {
	return FALSE;
    }

    if ( (*(isbn13) != '9')
	|| (*(isbn13 + 1) != '7')
	|| (*(isbn13 + 2) != '8')) {

	/* not allowed to convert from isbn13 to isbn10
	 * if prefix is not "978"
	 */
	return FALSE;
    }

    for (i=0; i<9; i++) {
	loopChar = isbn13[i+3];
	if (!isdigit(loopChar)) {
	    break;
	}
	isbn10[i] = loopChar;
    }
    if (i!=9) {
	return FALSE;
    }

    checkdigit = PM_Mod11Gen(isbn10, 9);
    isbn10[i] = checkdigit;
    isbn10[PM_ISBN_SIZE] = '\0';
    return TRUE;
} /* PM_isbn_13_to_10() */
