/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: pmmod.c
 *  Description	: Checkdigit routines.
 *  Author(s)	: Hans van Eck
 *  Creation	: 15-03-1996
 *
 *  History:
 *  15-03-1996	: eck      : creation
 *  27-04-2010	: paul     : CBS-1581 Added PM_Mod11GenBNF()
 *  24-03-2011	: paul     : CBS-2776 automatically create ISNI in 003E
 *  14-01-2014  : evd      : CBS-4106 Added PM_Mod29GenBNF
 *
 *  Notes	:
 *  This is an adapted version of the TAL-original.
 *========================================================================
 */

#pragma ident "$Id: pmmod.c 335412 2016-05-27 11:01:38Z heesakkm $"

#include <ctype.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function	: PM_Mod11Gen
 *  Purpose	: Generate a check digit from a digit-string.
 *  Method	: 
 *  First a sum is computed, in which each digit is multiplied by
 *  'distance from the end + 1'.
 *  The results are added and the checksum is then computed using the
 *  expression (11 - (sum % 11)) % 11, which gives the distance of the sum
 *  to the next multiple of 11 (and 0 when the sum can be divided by 11).
 *  The result is then the check-digit, except whem the result is 10, in
 *  which case the check-diogit is 'X'.
 *
 *  Consider the string "457".
 *  The '4' is multiplied by 4, the '5' is is multiplied by 3 and the '7'
 *  is multiplied by 2.  This gives a total of 4*4 + 3*5 + 2*7 = 45.  The
 *  next mupltiple is 55, giving a difference of 10.  This means that the
 *  check-digit is 'X'.
 *
 *  History:
 *  15-03-1996	: eck      : creation
 *
 *  Notes	:
 *  Although this version does works different than the original TAL version,
 *  it still computes the same value.
 *  The input string is checked for len bytes, all are assumed digits.  If
 *  len is negative, str must be '\0'-terminated.
 *
 *  Returns	: The check digit.
 *------------------------------------------------------------------------
 */
char
PM_Mod11Gen(
	const char	*str,	/* input string (digits only) */
	int		len)	/* length of str (without check-digit) */
{
    int	sum;
    int	i;

    if (len < 0) len = (int) strlen(str);

    sum = 0;
    for (i = len + 1; i > 1; i--) {
	sum += (*str++ - '0') * i;
    }
    sum = (11 - (sum % 11)) % 11;

    if (sum == 10) return 'X';
    else return '0' + sum;
} /* PM_Mod11Gen() */

/*------------------------------------------------------------------------
 *  Function	: PM_Mod11Test
 *  Purpose	: Test a check-digit in a string.
 *  Method	: 
 *  PM_Mod11Gen() is used to generate a new check-digit by passing (len - 1)
 *  as the string lengh.  This is then compared with one in the string.
 *
 *  History:
 *  15-03-1996	: eck      : creation
 *
 *  Notes	:
 *  The comparison of the check-digit uses toupper() since it might be 'x'.
 *
 *  Returns	: indication of match
 *------------------------------------------------------------------------
 */
bool
PM_Mod11Test(
	const char	*str,	/* input string */
	int		len)	/* length of str (including check-digit) */
{
    char	chk_digit;

    if (NULL == str){
	return TRUE;
    }

    if (len < 0) len = (int) strlen(str);

    if (len == 0) return FALSE;		/* empty (check digit only is ok) */

    len--;	/* The length now doesn't include the checkdigit */

    chk_digit = PM_Mod11Gen(str, len);

    return (chk_digit == toupper(str[len]));
} /* PM_Mod11Test() */

/*------------------------------------------------------------------------
 *  Function	: PM_Mod11GenBNF
 *  Purpose	: Generate a BNF check digit from a digit-string.
 *  Method	: 
 *
 *  History:
 *  27-04-2010	: paul     : creation
 *
 *  Notes	:
 *
 *  Returns	: The BNF check digit.
 *
 *------------------------------------------------------------------------
 */
char
PM_Mod11GenBNF(
	const char	*str,	/* input string (digits only) */
	int		len)	/* length of str (without check-digit) */
{
    int	sum;
    int	i;

    if (len < 0) len = (int) strlen(str);

    sum = 0;
    for (i = 1; i <= len; i++) {
	sum += (*str++ - '0') * i;
    }
    sum %= 11;

    if (sum == 10) return 'X';
    else return '0' + sum;
} /* PM_Mod11GenBNF() */

/*------------------------------------------------------------------------
 *  Function	: PM_Mod29GenBNF
 *  Purpose	: Generate a BNF ARK check digit from a digit-string.
 *  Method	: 
 *
 *  History:
 *  14-01-2014	: evd      : creation
 *
 *  Notes	:
 *
 *  Returns	: The BNF ARK check digit if ok or '\0' at error (str length exact 10)
 *
 *------------------------------------------------------------------------
 */
char
PM_Mod29GenBNF(
	const char	*str)	/* input string  (only chars from checkstr) */
{
    const char	*checkstr = "0123456789bcdfghjkmnpqrstvwxz";
    int		sum;
    int		i;
    char	*p;

    if (str == NULL) return '\0';
    if (strlen(str) != 10) return '\0';

    sum = 0;
    for (i = 1; i <= 10; i++) {
	p = strchr(checkstr, *str++);
	if (p == NULL) return '\0';
	sum += (p - checkstr) * i;
    }
    sum %= 29;

    return checkstr[sum];
} /* PM_Mod29GenBNF() */

/*------------------------------------------------------------------------
 *  Function	: PM_Mod11GenISNI
 *  Purpose	: Generate an ISNI check digit from a digit-string.
 *  Method	: 
 *
 *  History:
 *  23-03-2011	: paul     : creation
 *
 *  Notes	:
 *  From "Information and documentation -- International standard name
 *  identifier (ISNI)"
 *  Annex A: Method for calculating the check character of an ISNI
 *  
 *  The check character shall be calculated in accordance with ISO/IEC
 *  7064:2003, MOD 11-2. The check character shall always be located in the
 *  extreme right (low order) position.
 *
 *  Multiply the character values by their weights, then add the
 *  products. Strings including the check character are valid if the sum of
 *  these products is congruent to 1 (mod 11).
 *  
 *  The method for calculating the check character, which may be carried out
 *  automatically in a computer, is shown in , using the ISNI number
 *  245 863 573 047 as an example.
 *
 *  Table A.1 Method for calculating the check character of 1422 4586 3573 047
 *  
 *  1 Take the first twelve digits of the ISNI (the check character will the
 *    thirteenth and last character)
 *    2 4 5 8 6 3 5 7 3 0 4 7
 *  2 Take the weighting factors associated with each digit
 *    13 12 11 10 9 8 7 6 5 4 3 2
 *  3 Multiply each digit in turn by its weighting factor
 *    26 48 55 80 54 24 35 42 15 0 12 14
 *  4 Add these numbers together
 *    26+48+55+80+54+24+35+42+15+0+12+14=405
 *  5 Divide this sum by the modulus 11
 *    405 : 11 = 36 remainder 9
 *  6 Subtract the remainder from 11
 *    11 – 9 = 2
 *  7 Append the remainder, which is the check character, to the extreme right
 *    (low order) position of the base number of the ISNI
 *    245 863
 *
 *  The weight w(i) is calculated using w(i) = pow(2, i-1) % 11
 *  where i is the digit position starting from the right and %11 is Modulus 11
 *
 *  Returns	: The checkdigit
 *                or PM_ISNI_CHECKDIGIT_NONE if it couldn't be computed
 *
 *------------------------------------------------------------------------
 */
char
PM_Mod11GenISNI(
	const char	*str,	/* input string (digits only) */
	int		len)	/* length of str (without check-digit) */
{
    static int	weight[64 - 1];		/* array of weights */
    static int	weight_count = 0;	/* number of initialized weights */
    int		sum;
    int		i;

    if (len < 0) len = (int) strlen(str);
    if (len >= 64) {
	return PM_ISNI_CHECKDIGIT_NONE;
    }

    if (weight_count < len) {
	/* Initialize the new weights */
	for (i = weight_count; i < len; i++) {
	    weight[i] = ((uint64)1 << (i + 1)) % 11;
	}
	weight_count = len;
    }

    sum = 0;
    for (i = 1; i <= len; i++) {
	sum += (*str++ - '0') * weight[len - i];
    }
    sum = (11 - (sum % 11) + 1) % 11;

    if (sum == 10) return 'X';
    else return '0' + sum;
} /* PM_Mod11GenISNI() */

