/*========================================================================
 *  Copyright (c) 1997 by Pica
 *
 *  File        : pmppn.c
 *  Description : PPN-related routines
 *  Author(s)   : Paul Smit
 *  Creation    : 14-03-1997
 *
 *  History:
 *  14-03-1997  : paul     : creation
 *  26-02-1998  : lolke    : Content PPNTAB standard table changed. No ERAbort anymore.
 *  27-07-1999	: BdB	: Allow for larger PPNs (up to PIC_PPN_MAXLEN).
 *  19-09-2000  : EvD   : Added PM_epn_from_ipn, PM_epn_to_ipn
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *  05-02-2016	: heesakkm : CBS-6521: use strtol() instead of atol().
 *
 *========================================================================
 */

#pragma ident "$Id: pmppn.c 372553 2019-12-02 16:32:40Z heesakkm $"

#include <ctype.h>
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <pica.h>
#include <picatypes.h>
#include <picamisc.h>
#include <er.h>
#include <cb.h>

#define PM_PPN_TAB_KEY             "PPNTAB"
#define PM_PPN_YEAR_ATTR           "OFFSET"
#define PM_PPN_YEAR_NONE           "00"
#define PM_PPN_IPN_ATTR            "IPN"
/* Made global for old pp3tab definition
#define PM_PPN_TAB_SIZE            14
#define PM_PPN_YEAR_SIZE           2
*/

typedef enum PMPpnStat_t {
    PM_PPNSTAT_OK,
    PM_PPNSTAT_NONE,
    PM_PPNSTAT_ERROR,
    PM_PPNSTAT_EOF
} PMPpnStat_t;

/* Made global for old pp3tab definition
typedef struct PMPpnTab_t {
   int                             len;
   struct
   {
      char                            name[PM_PPN_YEAR_SIZE + 1];
      PICIpn_t                        ipn;
   } year[PM_PPN_TAB_SIZE];
} PMPpnTab_t;
*/

/*------------------------------------------------------------------------
 *  Function	: StringToLong
 *  Purpose	: Convert a string to a valid non-negative long integer
 *  Method	: strtol()
 *
 *  History:
 *  05-02-2016	: heesakkm : creation
 *  16-06-2016	: martinz  : Renamed from StringToIpn(), return -1 on error.
 *
 *  Returns	: A valid long integer or -1.
 *------------------------------------------------------------------------
 */
static long
StringToLong(
	const char	*str)
{
    char	*endptr = NULL;
    long        l;

    if ((str == NULL) || (*str == '\0')) return -1;

    errno = 0;
    l = strtol(str, &endptr, 10);

    if ((l == 0) && (errno == EINVAL))	return -1;	// conversion error
    else if (*endptr != '\0')		return -1;	// contains garbage
    else if (errno == ERANGE)		return -1;	// too large to represent
    else if (l < 0)			return -1;	// too small

    return l;
} /* StringToLong() */

/*------------------------------------------------------------------------
 *  Function    : PM_ppn_to_ipn
 *  Purpose     : Convert a PPN to an IPN
 *  Method      :
 *
 *  History:
 *  14-03-1997  : paul     : Successor of tal routine PPN2IPN
 *  27-07-1999	: BdB	: Allow for larger PPNs (up to PIC_PPN_MAXLEN).
 *  24-03-2010	: sieburgw : Allow NULL for ppn.
 *  05-02-2016	: heesakkm : CBS-6521: call StringToIpn() instead of atol().
 *  16-06-2016	: martinz  : CBS-6759: Fixed ppn offset 0.
 *
 *  Notes       :
 *  Does not check checkdigit; does require that ppn is '\0' terminated.
 *
 *  Description	:
 *  PM_ppn_tab_create must have created ppntab; if ppntab == NULL, ERAbort()
 *  will be called. PIC_IPN_NONE will be returned if and only if the conversion
 *  failed. The conversion doesn't require the ppn to be terminated by a
 *  0-byte, even a checkdigit need not be present.
 *
 *  Example	:
 *
 *	PICIpn_t	 ipn;
 *	char		 ppn[PIC_PPN_LEN + 1];
 *
 *	...
 *	if ((ipn = PM_ppn_to_ipn(ppn)) == PIC_IPN_NONE) {
 *	    ERWrite(APPLE_PARAM, __func__, "PM_ppn_to_ipn(%s) failed", ppn);
 *	}
 *
 *  Returns     : The IPN on success or PIC_IPN_NONE on failure
 *------------------------------------------------------------------------
 */
PICIpn_t
PM_ppn_to_ipn(
        PMPpnTab_p     ppntab,
        const char    *ppn)
{
    int len = 0;
    char p[PIC_PPN_MAXLEN];
    int      i;
    PICIpn_t ipn;

    if (NULL != ppn){
	len = strlen(ppn) - 1; /* Subtract checkdigit. */
    }

    if (len < 1 || len > PIC_PPN_MAXLEN - 1) return PIC_IPN_NONE;
    if (ppntab == NULL) {
        ERWrite(APPLE_FATAL, "PM_ppn_to_ipn()", "ppntab == NULL");
        return PIC_IPN_NONE;
    }
    /* Get rid of checkdigit. */
    memcpy(p, ppn, len);
    p[len] = '\0';
    /* Years only apply to PPNs of length PIC_PPN_LEN. New style longer PPNs
       (or erroneous shorter PPNs missing leading zeros) never have a year
       prefix. */
    if (ppntab->len > 0 && len == PIC_PPN_LEN - 1 &&
        memcmp(ppntab->year[0].name, p, PM_PPN_YEAR_SIZE) <= 0) {
        for (i = 0; i < ppntab->len; i++) {
            if (memcmp(ppntab->year[i].name, p, PM_PPN_YEAR_SIZE) == 0) {
		long offset;	/* offset relative to year found in ppntab */

		offset = StringToLong(p + PM_PPN_YEAR_SIZE);
		if (offset < 0) 	return PIC_IPN_NONE;	// error
                ipn = ppntab->year[i].ipn + offset;
                if (ipn < PIC_IPN_MIN) 	return PIC_IPN_NONE;	// too small
                if (ipn > PIC_IPN_MAX) 	return PIC_IPN_NONE;	// too large
		return ipn;
            }
        }
    }
    ipn = StringToLong(p);
    if (ipn < 0) 		return PIC_IPN_NONE;	// error
    if (ipn < PIC_IPN_MIN) 	return PIC_IPN_NONE;	// too small
    if (ipn > PIC_IPN_MAX) 	return PIC_IPN_NONE;	// too large
    return ipn;
} /* PM_ppn_to_ipn */

/*------------------------------------------------------------------------
 *  Function    : PM_ppn_from_ipn
 *  Purpose     : Convert an IPN to a PPN
 *  Method      :
 *
 *  History:
 *  14-03-1997  : paul     : Successor of tal routine IPN2PPN
 *  27-07-1999	: BdB	: Allow for larger PPNs (up to PIC_PPN_MAXLEN).
 *
 *  Notes       :
 *  Size of output ppn must be large enough to contain '\0' byte.
 *  If output ppn == NULL, the value will be written into ppn_buffer.
 *
 *  Description	:
 *  PM_ppn_tab_create must have created ppntab; if ppntab == NULL, ERAbort()
 *  will be called. If ppn is equal to NULL an internal buffer will be used to
 *  store the output PPN, otherwise ppn must point to an array of at least
 *  PIC_PPN_LEN characters plus 1 extra for the terminating 0-byte. NULL will
 *  be returned if and only if the conversion failed.
 *
 *  Example	:
 *
 *	PICIpn_t	ipn;
 *	const char	*ppn;
 *
 *	if ((ppn = PM_ppn_from_ipn(ipn, NULL)) == NULL) {
 *	    ERWrite(APPLE_PARAM, __func__, "PM_ppn_from_ipn(%ld) failed", (long) ipn);
 *	}
 *
 *  Returns     : A pointer to the PPN on success or NULL on failure.
 *------------------------------------------------------------------------
 */
const char *
PM_ppn_from_ipn(
        PMPpnTab_p     ppntab,
        PICIpn_t       ipn,
        char          *ppn)
{
    static char    ppn_buffer[PIC_PPN_MAXLEN + 1];
    int            i;
    bool           found;
    int len;

    if (ppntab == NULL) {
        ERWrite(APPLE_FATAL, "PM_ppn_from_ipn()", "ppntab == NULL");
	return NULL;
    }
    if (ppn == NULL) ppn = ppn_buffer;
    if (ipn < PIC_IPN_MIN || PIC_IPN_MAX < ipn) {
        return NULL;
    }
    for (found = FALSE, i = ppntab->len - 1; i >= 0 && !found; i--) {
        if (ipn >= ppntab->year[i].ipn) {
            found = TRUE;
            if (i == ppntab->len - 1) {
                len = sprintf(ppn, "%0*ld", (int)(PIC_PPN_LEN - 1),
			      (long int)ipn);
		if (len < 0) return NULL;
            } else {
                len = sprintf(ppn, "%s%0*ld", ppntab->year[i].name,
			      (int)(PIC_PPN_LEN - PM_PPN_YEAR_SIZE - 1),
			      (long int)(ipn - ppntab->year[i].ipn));
		if (len < 0) return NULL;
            }
        }
    }
    if (!found) {
	len = sprintf(ppn, "%0*ld", (int)(PIC_PPN_LEN - 1), (long int)ipn);
	if (len < 0) return NULL;
    }
    ppn[len] = PM_Mod11Gen(ppn, len);
    ppn[len + 1] = '\0';
    return ppn;
} /* PM_ppn_from_ipn */

/*------------------------------------------------------------------------
 *  Function    : get_year_ipn
 *  Purpose     : Get OFFSET=<year>=<ipn> from PPNTAB
 *  Method      :
 *
 *  History:
 *  14-03-1997  : paul     : Creation
 *  11-03-2004	: heesakkm : Use long_ipn for sscanf().
 *
 *  Returns     : Whether another year/ipn pair could be retrieved
 *------------------------------------------------------------------------
 */
static PMPpnStat_t
get_year_ipn(
        CBBuf_t          *CBBuf,
        size_t            CBSize,
        CBElm_p          *CBElm,
        char             *year,
        PICIpn_t         *ipn)
{
    const               char *text;
    int                 i;
    int                 scanlen;
    long		long_ipn;	/* for sscanf() */

    if (*CBElm == NULL) {
        *CBElm = CBGetAttribute(CBBuf, CBSize, NULL, &text);
    } else {
        *CBElm = CBNextAttribute(CBBuf, CBSize, *CBElm, &text);
    }
    if (*CBElm == NULL) {
        return PM_PPNSTAT_EOF;
    }
    (void) memset(year, ' ', PM_PPN_YEAR_SIZE);
    year[PM_PPN_YEAR_SIZE] = '\0';
    *ipn = PIC_IPN_NONE;
    if (!STREQ((const char *)*CBElm, PM_PPN_YEAR_ATTR)) return PM_PPNSTAT_ERROR;
    if (strncmp(text, PM_PPN_YEAR_NONE, PM_PPN_YEAR_SIZE) == 0) {
        return PM_PPNSTAT_NONE;
    }
    for (i = 0; i < PM_PPN_YEAR_SIZE; i++) {
        if (!isdigit(text[i])) return PM_PPNSTAT_ERROR;
    }
    if (*(text + PM_PPN_YEAR_SIZE) != '=') { return PM_PPNSTAT_ERROR; }
    (void) strncpy(year, text, PM_PPN_YEAR_SIZE);

    if (strlen(text + PM_PPN_YEAR_SIZE + 1) > (size_t)(PIC_PPN_LEN - 1)
        || sscanf(text + PM_PPN_YEAR_SIZE + 1, "%ld%n", &long_ipn, &scanlen) != 1
        || (size_t)scanlen != strlen(text + PM_PPN_YEAR_SIZE + 1)
        || long_ipn + 1 < PIC_IPN_MIN || PIC_IPN_MAX < long_ipn) {
        return PM_PPNSTAT_ERROR;
    }
    *ipn = (PICIpn_t) long_ipn;
    return PM_PPNSTAT_OK;
} /* get_year_ipn */

/*------------------------------------------------------------------------
 *  Function    : PM_ppn_tab_create
 *  Purpose     : Create a PPN-table
 *  Method      :
 *
 *  History:
 *  14-03-1997  : paul     : Creation
 *
 *  Notes       :
 *  If CBBuf is equal to NULL, PPN-tab is set to it's default value: no year
 *  offsets. The same is true if CBBuf only contains \OFFSET=00=0
 *
 *  If CBBuf does not contain a valid PPN-table, ERWrite() will be called.
 *
 *  Description	:
 *  PM_ppn_tab_create() creates the table that is needed to convert PPN's to
 *  IPN's vice versa and returns whether it could be created. If initialization
 *  failed, an ERWrite() will be performed. CBBuf must contain the PPN-table as
 *  read from the standard table with TBRead().
 *
 *  A PPN conversion table must have the following syntax:
 *
 *	PPN_TABLE	::= <YEAROFFSET>*
 *	YEAROFFSET	::= '\YEAR=' <YEAR> <NEWLINE> '\IPN=' <IPN> <NEWLINE>
 *	YEAR		::= <DIGIT> <DIGIT> | 'NONE'
 *	IPN		::= <DIGIT>*
 *
 * Yearoffsets must appear in ascending order. If an IPN is less than the first
 * IPN in the table or if an IPN is greater than or equal to the last (highest)
 * IPN in the table or if the table only contains \YEAR=NONE, int to decimal
 * conversion will be used to compute the PPN.
 *
 *------------------------------------------------------------------------
 */
PMPpnTab_p
PM_ppn_tab_create(
        CBBuf_t       *CBBuf,
        size_t         CBSize)
{
    const char     *errloc = "PM_ppn_tab_create";
    CBElm_p         CBElm;
    PMPpnStat_t     status;
    PMPpnTab_t     *ppntab;
    bool            ppntab_ok = FALSE;

    if ((ppntab = malloc(sizeof(PMPpnTab_t))) == NULL) {
        ERWrite(APPLE_NOMEM, errloc, "cannot malloc() PPN-table");
        return NULL;
    } else if (CBBuf == NULL) {
        ppntab->len = 0;
        ppntab_ok = TRUE;
    } else {
        CBElm = NULL;
        ppntab->len = 0;
        ppntab_ok = TRUE;
        while (   ppntab_ok
               && ppntab->len < PM_PPN_TAB_SIZE
               && (status = get_year_ipn(CBBuf, CBSize,
                                &CBElm, ppntab->year[ppntab->len].name,
                                &ppntab->year[ppntab->len].ipn)
                  ) == PM_PPNSTAT_OK
               ) {
/*
 *          Check that years and ipns are in correct order
 */
            if (ppntab->len != 0) {
                if (strcmp(ppntab->year[ppntab->len - 1].name,
                           ppntab->year[ppntab->len    ].name)
                    >= 0) {
                    ERWrite(APPLE_PARAM, errloc, "%s=%s out of order",
                            PM_PPN_YEAR_ATTR,
                            ppntab->year[ppntab->len].name);
                    ppntab_ok = FALSE;
                } else if (ppntab->year[ppntab->len - 1].ipn >
                           ppntab->year[ppntab->len    ].ipn) {
                    ERWrite(APPLE_PARAM, errloc, "%s=%ld out of order",
                            PM_PPN_IPN_ATTR,
                            (long) ppntab->year[ppntab->len].ipn);
                    ppntab_ok = FALSE;
                }
            }
            ppntab->len++;
        }
        if (ppntab_ok) switch(status) {
        case PM_PPNSTAT_ERROR:
            ERWrite(APPLE_PARAM, errloc, "error in yearoffset nr. %d",
                    ppntab->len + 1);
            ppntab_ok = FALSE;
            break;
        case PM_PPNSTAT_EOF:
            if (ppntab->len == 0) {
                ERWrite(APPLE_PARAM, errloc, "%s=%s expected", PM_PPN_YEAR_ATTR,
                        PM_PPN_YEAR_NONE);
                ppntab_ok = FALSE;
            }
            break;
        case PM_PPNSTAT_NONE:
            if (ppntab->len != 0) {
                ERWrite(APPLE_PARAM, errloc, "unexpected %s=%s", PM_PPN_YEAR_ATTR,
                        PM_PPN_YEAR_NONE);
                ppntab_ok = FALSE;
            }
            break;
        case PM_PPNSTAT_OK:
            if (   ppntab->len == PM_PPN_TAB_SIZE
                && get_year_ipn(CBBuf, CBSize,
                                &CBElm, ppntab->year[0].name,
                                &ppntab->year[0].ipn)
                       != PM_PPNSTAT_EOF) {
                ERWrite(APPLE_PARAM, errloc, "too many yearoffsets");
                ppntab_ok = FALSE;
            }
            break;
        }
    }
    if (!ppntab_ok) {
        PM_ppn_tab_destroy(ppntab);
        ppntab = NULL;
    }
    return ppntab;
} /* PM_ppn_tab_create */

/*------------------------------------------------------------------------
 *  Function    : PM_ppn_tab_destroy
 *  Purpose     : Destroys a PPN-table
 *  Method      :
 *
 *  History:
 *  14-03-1997  : paul     : Creation
 *
 *  Notes       :
 *  If ppntab is equal to NULL, ppntab is assumed to be already destroyed.
 *
 *  Example	:
 *
 *	PMPpnTab_p	ppntab;
 *
 *	PM_ppn_tab_destroy(ppntab);
 *	ppntab = NULL;
 *
 *------------------------------------------------------------------------
 */
void
PM_ppn_tab_destroy(
        PMPpnTab_p     ppntab)
{
    if (ppntab != NULL) free((PMPpnTab_t *)ppntab);
} /* PM_ppn_tab_destroy */

/*------------------------------------------------------------------------
 *  Function    : PM_epn_from_ipn
 *  Purpose     : Convert an IPN to an EPN
 *  Method      :
 *
 *  History:
 *  19-09-2000  : evd   : PM_ppn_from_ipn without ppntab
 *
 *  Notes       :
 *  Size of output epn must be large enough to contain '\0' byte.
 *  If output epn == NULL, the value will be written into epn_buffer.
 *
 *  Returns     : A pointer to the PPN on success or NULL on failure.
 *------------------------------------------------------------------------
 */
const char *
PM_epn_from_ipn(
        PICIpn_t       ipn,
        char          *epn)
{
    static char    epn_buffer[PIC_PPN_MAXLEN + 1];
    int len;

    if (epn == NULL) epn = epn_buffer;
    if (ipn < PIC_IPN_MIN || PIC_IPN_MAX < ipn) {
        return NULL;
    }
    len = sprintf(epn, "%0*ld", (int)(PIC_PPN_LEN - 1), (long int)ipn);
    if (len < 0) return NULL;
    epn[len] = PM_Mod11Gen(epn, len);
    epn[len + 1] = '\0';
    return epn;
} /* PM_epn_from_ipn */

/*------------------------------------------------------------------------
 *  Function    : PM_epn_to_ipn
 *  Purpose     : Convert an EPN to an IPN
 *  Method      :
 *
 *  History:
 *  20-09-2000  : EvD     : PM_ppn_to_ipn without ppntab
 *  05-02-2016	: heesakkm : CBS-6521: call StringToIpn() instead of atol().
 *  16-06-2016	: martinz  : CBS-6759: Fixed ppn offset 0.
 *
 *  Notes       :
 *  Does not check checkdigit; does require that ppn is '\0' terminated.
 *
 *  Returns     : The IPN on success or PIC_IPN_NONE on failure
 *------------------------------------------------------------------------
 */
PICIpn_t
PM_epn_to_ipn( const char    *epn)
{
    const int len = strlen(epn) - 1; /* Subtract checkdigit. */
    char p[PIC_PPN_MAXLEN];
    PICIpn_t ipn;

    if (len < 1 || len > PIC_PPN_MAXLEN - 1) return PIC_IPN_NONE;
    /* Get rid of checkdigit. */
    memcpy(p, epn, len);
    p[len] = '\0';
    ipn = StringToLong(p);
    if (ipn < 0) 		return PIC_IPN_NONE;	// error
    if (ipn < PIC_IPN_MIN) 	return PIC_IPN_NONE;	// too small
    if (ipn > PIC_IPN_MAX) 	return PIC_IPN_NONE;	// too large
    return ipn;
} /* PM_ppn_to_ipn */
