/*========================================================================
 *  Copyright (c) 1995 by Pica
 *
 *  File        : pmrange.c
 *  Description : Clean a range-buffer
 *  Author(s)   : Hans van Eck
 *  Creation    : 13-01-1995
 *
 *  History:
 *  13-01-1995  : eck      : creation
 *  07-02-1996  : eck      : Added a number of new routines.  This slightly
 *                         : changes the input-syntax of PM_range_clean().
 *  15-09-1998  : fb       : Fixed possible use of NULL pointer in
 *                           output_rangelist().
 *  17-10-2000  : paul     : PM_range_from_bits() and
 *                           PM_range_to_bits() introduced
 *  27-11-2001  : paul     : Added PM_range_intersect() + PM_range_unite()
 *                           + PM_range_error()
 *  29-11-2001  : paul     : Don't print r_high if it is equal to ULONG_MAX
 *  28-06-2004  : paul     : Added PM_range_to_bits_check()
 *  05-07-2006	: heesakkm : Added support for hexadecimal ranges.
 *  05-08-2009	: sutherlr : Function range_from_bits() now builds the 
 *                           range string itself.
 *  21-03-2012	: martinz  : Fixed range_from_bits().
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *  Notes       :
 *  This code tries to be careful about just about everything. The
 *  algorithm is not too smart, but is does the job.
 *  It parses the range-list and stores every range in a linked list,
 *  sorted on the low-order part of the range. These are then collated so
 *  adjacent and/or overlapping entries are merged. The resulting list is
 *  then written back to the string.
 *
 *  Description	:
 *
 *  All the PM_range_*-routines have as their first argument a pointer into an
 *  range-string. The functions that return an unsigned long, will return the
 *  number of element in the resulting range. The following descriptions also
 *  apply to the PM_range_x_* routines.
 *
 *  - PM_range_clean() accepts an unsorted range-buffer. It will then create
 *    a cleaned version of the buffer by sorting and collating the ranges.
 *    Open ranges (like "12-") are closed using the 'high' parameter. The low
 *    and high give the lower- and higher-bounds between which the ranges must
 *    lie. In case of an error (like format, buffer size, number out of
 *    bounds), PM_range_clean() will return 0 (otherwise it will return the
 *    number of elements).
 *  - PM_range_add() accepts an unsorted range-buffer and first- and
 *    last-element of the range to add. The new range is added and the
 *    range-buffer is cleaned. In case of an error (like format, buffer size),
 *    PM_range_add() will return 0 (otherwise it will return the number of
 *    elements).
 *  - PM_range_del_number() accepts an unsorted range-buffer and a number as
 *    argument. The number is deleted and the range-buffer is cleaned. In
 *    case of an error (like format, buffer size), PM_range_add() will return 0
 *    (otherwise it will return the number of elements).
 *  - PM_range_get_bounds() accepts an unsorted range-buffer as input argument.
 *    It will use first and last as pointers to store the first range in the
 *    range buffer. PM_range_get_bounds() will return a pointer to the start
 *    of the next range in the buffer (or NULL when there was not a range or in
 *    case of an error).
 *  - PM_range_test_number() accepts an unsorted range-buffer and a number as
 *    arguments. PM_range_test_number() will return TRUE if the number occurs
 *    in the range-buffer, FALSE otherwise.
 *  - PM_range_to_bits() accepts a bit array, a range-string and a number as
 *    arguments. The number is the number indicated by the first bit.
 *    PM_range_to_bits() will return PM_RANGE_ERROR if it couldn't convet
 *    the range-string to a bit array, either because the range-string was
 *    incorrect or because the bit array was too small. Otherwise it will
 *    return the number of elements in the range-string (even if that number
 *    was zero, which is not an error).
 *  - PM_range_from_bits() accepts a range-buffer, a bit array and a number as
 *    arguments. The number is the number indicated by the first bit.
 *    PM_range_from_bits() will return PM_RANGE_ERROR if it couldn't convert
 *    the bit array to a range-string, because the range-buffer was too small.
 *    Otherwise it will return the number of elements in the range-string (even
 *    if that number was zero, which is not an error).
 *
 *  Example 1	:
 *
 *      char			buffer[100] = "15 4 3 2 1 19";
 *      const char		*tmp;
 *	unsigned long	nr_elts, low, high;
 *
 *	nr_elts = PM_range_clean(buffer, sizeof(buffer), 0, ULONG_MAX);
 * 	// nr_elts == 6, buffer == "1-4,15,19"
 *
 *	nr_elts = PM_range_add(buffer, sizeof(buffer), 16, 18);
 * 	// nr_elts == 9, buffer == "1-4,15-19"
 *
 *	nr_elts = PM_range_del_number(buffer, sizeof(buffer), 17);
 * 	// nr_elts == 8, buffer == "1-4,15-16,18-19"
 *
 *	tmp = buffer;
 *	while ((tmp = PM_range_get_bounds(tmp, &low, &high)) != NULL) {
 *	    // low and high are (1,4), (15,16) and (18,19) in subsequent
 *	    // iterations of the loop.
 *	    //
 *	    ... some meaningful statements with low and high ...
 *
 *	    // Use the following statement. A test for tmp != NULL after the
 *	    // loop will then indicate format problem.
 *	    //
 *	    if (*tmp == '\0') break;
 *	}
 *
 *      if (PM_range_test_number(buffer, 3)) // is TRUE ;
 *
 *	if (PM_range_test_number(buffer, 17)) // is FALSE ;
 *
 *  Example 2	:
 *
 *      char		rangebuf[100];
 *      byte		bitbuf[100];
 *      size_t		bitlen;
 *
 *      if (PM_range_to_bits(bitbuf, sizeof bitbuf, &bitlen,
 *          "1-4,15-16,18-19", 1) == PM_RANGE_ERROR) {
 *          if (bitlen > sizeof bitbuf) {
 *              fprintf(stderr, "PM_range_to_bits(): buffer overflow"
 *	    } else {
 *              fprintf(stderr, "PM_range_to_bits(): incorrect range string"
 *	    }
 *	} else if (PM_range_from_bits(rangebuf, sizeof rangebuf, bitbuf,i
 *	    bitlen, 1) == PM_RANGE_ERROR) {
 *	    fprintf(stderr, "PM_range_to_bits(): buffer overflow"
 *	} else {
 *	    // rangebuf contains "1-4,15-16,18-19"
 *	    ...
 *	}
 *
 *========================================================================
 */

#pragma ident "$Id: pmrange.c 363677 2019-01-10 15:10:10Z heesakkm $"

#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>

#include <string.h>
#include <pica.h>
#include <picamisc.h>

typedef struct range_t  *range_p;

typedef struct range_t {
    range_p             r_next;
    unsigned long       r_low;
    unsigned long       r_high;
} range_t;

/*
 * The allocation stuff keeps a permanent linked list. This avoids having
 * to call malloc() too often, at the expence of some extra memory.
 */
#define NR_RAN_ALLOC    10
static range_p  free_ran_list = NULL;

/*------------------------------------------------------------------------
 *  Function    : ran_alloc
 *  Purpose     : Allocate a range-element and initialise it.
 *  Method      : Take it from linked list if possible. Otherwise allocate.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *  17-10-2000  : paul     : return NULL if malloc() fails.
 *
 *  Notes       :
 *  Together with ran_free(), ran_alloc() maintains a linked list with free
 *  elements.
 *
 *  Returns     : The element allocated (and initialised) or NULL.
 *------------------------------------------------------------------------
 */
static range_p
ran_alloc(
        unsigned long   low,            /* lowest value in range */
        unsigned long   high)           /* highest value in range */
{
    range_p     tmp;
    int         i;

    if (free_ran_list == NULL) {
        if ((free_ran_list = malloc(NR_RAN_ALLOC * sizeof(range_t))) == NULL) {
            return NULL;
        }

        /* Initialise the array as linked list. */
        for (i = 0; i < NR_RAN_ALLOC - 1; i++) {
            free_ran_list[i].r_next = &free_ran_list[i+1];
        }
        free_ran_list[NR_RAN_ALLOC - 1].r_next = NULL;
    }
    tmp = free_ran_list;
    free_ran_list = free_ran_list->r_next;

    tmp->r_next = NULL;
    tmp->r_low = low;
    tmp->r_high = high;
    return tmp;
} /* ran_alloc() */

/*------------------------------------------------------------------------
 *  Function    : ran_free
 *  Purpose     : Free a range-element.
 *  Method      : Link it back into the free list.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *------------------------------------------------------------------------
 */
static void
ran_free(
        range_p elt)                    /* range-element to free */
{
    elt->r_next = free_ran_list;
    free_ran_list = elt;
} /* ran_free() */

/*------------------------------------------------------------------------
 *  Function    : clear_rangelist
 *  Purpose     : Deallocate a whole range-list.
 *  Method      :
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *------------------------------------------------------------------------
 */
static void
clear_rangelist(
        range_p ran_list)               /* range-list to deallocate() */
{
    range_p     rp;

    rp = ran_list;
    while (rp != NULL) {
        ran_list = ran_list->r_next;
        ran_free(rp);
        rp = ran_list;
    }
} /* clear_rangelist() */

/*------------------------------------------------------------------------
 *  Function    : add_range
 *  Purpose     : Add a range to the linked list.
 *  Method      :
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Notes       :
 *  This routine keeps the elements sorted on the r_low's. This fact is
 *  used by collate_rangelist().
 *
 *  Returns     : A pointer to the first element of the list or NULL.
 *------------------------------------------------------------------------
 */
static range_p
add_range(
        range_p         ran_list,       /* original range-list */
        unsigned long   low,            /* lowest value in ranges */
        unsigned long   high)           /* highest value in range */
{
    range_p     rp;
    range_p     prev_rp;

    if (ran_list == NULL) {
        return ran_alloc(low, high);
    }
    rp = ran_list;
    prev_rp = NULL;
    do {
        if (rp->r_low > low) break;
        prev_rp = rp;
        rp = rp->r_next;
    } while (rp != NULL);

    /* prev_rp now points just before the element after which to insert
     * the new element. If prev_rp == NULL, the new element is the first
     * element of the list.
     */
    if ((rp = ran_alloc(low, high)) == NULL) {
        clear_rangelist(ran_list);
        return NULL;
    }
    if (prev_rp == NULL) {
        rp->r_next = ran_list;
        return rp;
    }
    rp->r_next = prev_rp->r_next;
    prev_rp->r_next = rp;
    return ran_list;
} /* add_range() */

/*------------------------------------------------------------------------
 *  Function    : del_number
 *  Purpose     : Remove a number from a range-list
 *  Method      :
 *  First find the element in which the number belongs.
 *  If it is the only element, delete it from linked list.
 *  If number is the r_high or the r_low, adjust the boundary
 *  Otherwise split the elemnt into 2 pieces.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Notes       : Assumes that the range-list is collated.
 *
 *  Returns     : The new range-list or NULL
 *------------------------------------------------------------------------
 */
static range_p
del_number(
        range_p                 ran_list,       /* the range_list */
        unsigned long const     number)         /* number to delete */
{
    range_p     rp;
    range_p     tmp;
    range_p     prevptr;

    prevptr = NULL;
    tmp = ran_list;

    /* Look for the right element */
    while ((tmp != NULL) && (tmp->r_low <= number)) {
        if (tmp->r_high >= number) {
            /* Got it! */
            break;
        }
        prevptr = tmp;
        tmp = tmp->r_next;
    }
    /* The number may not have been in the range-list. */
    if ((tmp == NULL) || (tmp->r_low > number)) return ran_list;

    if (number == tmp->r_low) {
        /* Adjust or delete */
        if (number == tmp->r_high) {
            /* Delete the whole thing. */
            if (prevptr == NULL) {
                ran_list = tmp->r_next;
            } else {
                prevptr->r_next = tmp->r_next;
            }
            ran_free(tmp);
        } else {
            /* Adjust low part. */
            tmp->r_low++;
        }
    } else if (number == tmp->r_high) {
        /* Adjust high part */
        tmp->r_high--;
    } else {
        /* Split into two parts */
        if ((rp = ran_alloc(number + 1, tmp->r_high)) == NULL) {
            clear_rangelist(ran_list);
            return NULL;
        }
        rp->r_next = tmp->r_next;
        tmp->r_high = number - 1;
        tmp->r_next = rp;
    }
    return ran_list;
} /* del_number() */

/*------------------------------------------------------------------------
 *  Function    : range_get_bounds
 *  Purpose     : Get the lower and upper bound of a range from a string
 *  Method      : Manual (well, sort of) parsing of numbers.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Notes       :
 *  A ',' is a valid range or number separator.
 *  A string of the form "7-,1" is accepted.
 *
 *  Returns     :
 *  Pointer to start of next range or NULL. NULL indicates no (valid)
 *  range found.
 *------------------------------------------------------------------------
 */
static const char *
range_get_bounds(
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   *firstp,        /* pointer to store low value */
        unsigned long   *lastp,         /* pointer to store high value */
        int             base)
{
    char        *endp;

    /* Do we have something? */
    if ((rangebuf == NULL) || (*rangebuf == '\0')) return NULL;

    *firstp = strtoul(rangebuf, &endp, base);

    /* Test validity of number. */
    if (endp == rangebuf) return NULL;

    /* Skip spaces. */
    while (*endp == ' ') endp++;

    /* '-' indicates a range (instead of single number). */
    if (*endp == '-') {
        rangebuf = endp + 1;
        *lastp = strtoul(rangebuf, &endp, base);

        if (endp == rangebuf) {
            /* No valid number. Skip spaces to see the cause. */
            while (*endp == ' ') endp++;

            if ((*endp == '\0') || (*endp == ',')) {
                /* It is an open range. */
                *lastp = ULONG_MAX;
            } else {
                /* It is definitely wrong. */
                return NULL;
            }
        }
        if (*lastp < *firstp) return NULL;
    } else {
        *lastp = *firstp;
    }

    while (*endp == ' ') endp++;        /* skip spaces */

    if (*endp == ',') endp++;           /* skip ',' */

    return endp;                        /* start next one from here */
} /* range_get_bounds() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_get_bounds
 *  Purpose	: Get the lower and upper bound of a range from a string
 *  Method	: Call range_get_bounds() with base 10
 *
 *  History:
 *  02-07-1996  : eck      : creation
 *
 *  Returns     :
 *  Pointer to start of next range or NULL. NULL indicates no (valid)
 *  range found.
 *------------------------------------------------------------------------
 */
const char *
PM_range_get_bounds(
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   *firstp,        /* pointer to store low value */
        unsigned long   *lastp)         /* pointer to store high value */
{
    return range_get_bounds(rangebuf, firstp, lastp, 10);
} /* PM_range_get_bounds() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_x_get_bounds
 *  Purpose	: Get the lower and upper bound of a range from a hex-string
 *  Method	: Call range_get_bounds() with base 16
 *
 *  History:
 *  05-07-2006	: heesakkm : creation
 *
 *  Returns     :
 *  Pointer to start of next range or NULL. NULL indicates no (valid)
 *  range found.
 *------------------------------------------------------------------------
 */
const char *
PM_range_x_get_bounds(
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   *firstp,        /* pointer to store low value */
        unsigned long   *lastp)         /* pointer to store high value */
{
    return range_get_bounds(rangebuf, firstp, lastp, 16);
} /* PM_range_x_get_bounds() */

/*------------------------------------------------------------------------
 *  Function    : collate_rangelist
 *  Purpose     : Collate overlapping ranges in the range-list
 *  Method      :
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Notes       : Assumes that the range-list is sorted on r_low.
 *
 *  Returns     : The collated list.
 *------------------------------------------------------------------------
 */
static range_p
collate_rangelist(
        range_p         ran_list)       /* The list of ranges */
{
    range_p     rp = ran_list;
    range_p     next_rp;

    /* Walk down linked list (sorted on r_low). */
    while (rp != NULL) {
        next_rp = rp->r_next;

        if (next_rp == NULL) break;

        /* Watch out for overflow on (rp->rhigh + 1). */
        if ((rp->r_high == ULONG_MAX)
            || (rp->r_high + 1 >= next_rp->r_low)) {
            /* We can collate! The new r_high is the largest of the two. */
            if (rp->r_high < next_rp->r_high) {
                rp->r_high = next_rp->r_high;
            }
            rp->r_next = next_rp->r_next;
            ran_free(next_rp);
            /* Do not advance rp. */
        } else {
            rp = rp->r_next;
        }
    }
    return ran_list;
} /* collate_rangelist() */

/*------------------------------------------------------------------------
 *  Function    : intersect_rangelist
 *  Purpose     : Intersect ranges in the range-lists
 *  Method      :
 *
 *  History:
 *  27-11-2001  : paul     : creation
 *
 *  Notes       : Assumes that the range-lists are sorted on r_low.
 *
 *  Returns     :
 *  - The intersected list
 *  - or NULL && ok if intersection is empty
 *  - or NULL && !ok if an allocation error occurred
 *
 *------------------------------------------------------------------------
 */
static range_p
intersect_rangelist(
        range_p         src1_list,       /* source list of ranges */
        range_p         src2_list,       /* other source list of ranges */
	bool		*ok)		// out: if an error occurred
{
    range_p         s1 = src1_list, s2 = src2_list, dest_list = NULL;
    unsigned long   low;
    unsigned long   high;

    /* Walk down linked lists (sorted on r_low). */
    while (s1 != NULL && s2 != NULL) {
        if (s1->r_high < s2->r_low) {
            s1 = s1->r_next;
        } else if (s2->r_high < s1->r_low) {
            s2 = s2->r_next;
        } else {
/*          We can intersect!
 *          The new r_high is the smallest of the two;
 *	    the new r_low is the largest of the two.
 */
	    low = s1->r_low >= s2->r_low ? s1->r_low : s2->r_low;
	    high = s1->r_high >= s2->r_high ? s2->r_high : s1->r_high;
	    dest_list = add_range(dest_list, low, high);
	    if (dest_list == NULL) {
		*ok = FALSE;
		return NULL;
	    }
	    if (s1->r_high < s2->r_high) {
		s1 = s1->r_next;
	    } else {
		s2 = s2->r_next;
	    }
        }
    }

    *ok = TRUE;
    return dest_list;
} /* intersect_rangelist() */

/*------------------------------------------------------------------------
 *  Function    : unite_rangelist
 *  Purpose     : Unite ranges in the range-lists
 *  Method      :
 *
 *  History:
 *  27-11-2001  : paul     : creation
 *
 *  Notes       : Assumes that the range-lists are sorted on r_low.
 *                NOTE: unite_rangelist will overwrite both input rangelists!
 *
 *  Returns     : The united list or NULL == empty unison or error.
 *------------------------------------------------------------------------
 */
static range_p
unite_rangelist(
        range_p         src1_list,       /* source list of ranges */
        range_p         src2_list,       /* other source list of ranges */
	bool		*ok)		// out: if an error occurred
{
    range_p         s1 = src1_list, s2 = src2_list, dest_list = NULL;
    unsigned long   low;

    /* Walk down linked lists (sorted on r_low). */
    while (s1 != NULL && s2 != NULL) {
        /* Watch out for overflow on (rp->rhigh + 1). */
        if (s1->r_high != ULONG_MAX
            && s1->r_high + 1 < s2->r_low) {
	    dest_list = add_range(dest_list, s1->r_low, s1->r_high);
	    if (dest_list == NULL) {
		*ok = FALSE;
		return NULL;
	    }
            s1 = s1->r_next;
        } else if (s2->r_high != ULONG_MAX
		   && s2->r_high + 1 < s1->r_low) {
	    dest_list = add_range(dest_list, s2->r_low, s2->r_high);
	    if (dest_list == NULL) {
		*ok = FALSE;
		return NULL;
	    }
            s2 = s2->r_next;
        } else {
/*          We can unite!
 *          The new r_high is the largest of the two;
 *	    the new r_low is the smallest of the two.
 */
	    low = s1->r_low >= s2->r_low ? s2->r_low : s1->r_low;
	    if (s1->r_high < s2->r_high) {
		s2->r_low = low;
		s1 = s1->r_next;
	    } else {
		s1->r_low = low;
		s2 = s2->r_next;
	    }
        }
    }
    while (s1 != NULL) {
	dest_list = add_range(dest_list, s1->r_low, s1->r_high);
	s1 = s1->r_next;
    }
    while (s2 != NULL) {
	dest_list = add_range(dest_list, s2->r_low, s2->r_high);
	s2 = s2->r_next;
    }

    *ok = TRUE;
    return dest_list;
} /* unite_rangelist() */

/*------------------------------------------------------------------------
 *  Function	: subtract_rangelist
 *  Purpose	: Subtract the second list of ranges from the first
 *
 *  History:
 *  22-03-2016	: paul     : CBS-6594 creation
 *
 *  Example     :
 *  range1: 1-
 *  range2: 2,4,6,8
 *  No less than 5 ranges emerge from subtraction:
 *  1,3,5,7,9-
 *
 *  Returns	:
 *  - The created range containing the subtraction
 *  - or NULL && ok if subtraction is empty
 *  - or NULL && !ok if an allocation error occurred
 *
 *------------------------------------------------------------------------
 */
static range_p
subtract_rangelist(
        range_p         src1_list,	// list of ranges to subtract from
        range_p         src2_list,	// list of ranges to subtract
	bool		*ok)		// out: if an error occurred
{
    range_p         s1 = src1_list;
    range_p         s2 = src2_list;
    range_p         dest_list = NULL;

    // Walk down linked lists (sorted on r_low).
    while (s1 != NULL && s2 != NULL) {

	// s1 in front of s2 without overlap
        if (s1->r_high < s2->r_low) {
	    dest_list = add_range(dest_list, s1->r_low, s1->r_high);
	    if (dest_list == NULL) {
		*ok = FALSE;
		return NULL;
	    }
            s1 = s1->r_next;
	    continue;
        }

	// s2 in front of s1 without overlap
	if (s2->r_high < s1->r_low) {
            s2 = s2->r_next;
	    continue;
	}
	
	// Overlapping ranges:

	// Copy the lower part of the first range, if it sticks out
	// For example add_range "1" if s1="1-" and s2="2"
	if (s1->r_low < s2->r_low) {
	    dest_list = add_range(dest_list, s1->r_low, s2->r_low - 1);
	    if (dest_list == NULL) {
		*ok = FALSE;
		return NULL;
	    }
	}

	// Skip the range with the lowest higher boundary
	if (s1->r_high <= s2->r_high) {
	    s1 = s1->r_next;
	} else {
	    // Erase the processed part of range1 and skip range2
	    // For example change s1 from "1-" to "3-" and skip s2="2"
	    s1->r_low = s2->r_high + 1;
	    s2 = s2->r_next;
	}
    }

    // Copy the remaining ranges in the first list
    while (s1 != NULL) {
	dest_list = add_range(dest_list, s1->r_low, s1->r_high);
	s1 = s1->r_next;
    }

    *ok = TRUE;
    return dest_list;
} /* subtract_rangelist() */

/*------------------------------------------------------------------------
 *  Function    : parse_ranges
 *  Purpose     : parse the input string and build the linked list
 *  Method      :
 *  Every element (with a lower- and upper- bound) is added to the linked
 *  linked list in the order of their lower-bound (by add_range()).
 *
 *  History:
 *  13-01-1995  : eck      : creation
 *  06-02-1996  : eck      : Use PM_ramge_get_bounds().
 *                         : This also changes the syntax.
 *
 *  Notes       :
 *  Open ended ranges are allowed.
 *  A number can be preceded by spaces. A comma can be used as separator.
 *  The variable errno is not checked for numbers larger than ULONG_MAX.
 *
 *  Returns     : The linked list with all the ranges.
 *------------------------------------------------------------------------
 */
static range_p
parse_ranges(
        const char      *rangebuf,      /* pointer to start of buffer */
        unsigned long   low,            /* lowest allowed value in a range */
        unsigned long   high,           /* highest allowed value in a range */
        int             base)
{
    range_p             ran_list = NULL;
    unsigned long       first = 0, last = 0;
    bool                ok = TRUE;

    do {
        rangebuf = range_get_bounds(rangebuf, &first, &last, base);

        ok = (rangebuf != NULL);

        if (ok) {
            if (last == ULONG_MAX) last = high;
            ok = ((first >= low) && (last <= high));
        }
        if (ok) {
            ran_list = add_range(ran_list, first, last);
            ok = ran_list != NULL;
        }
    } while (ok && (*rangebuf != '\0'));

    if (! ok) {
        clear_rangelist(ran_list);
        return NULL;
    }
    return ran_list;
} /* parse_ranges() */

/*------------------------------------------------------------------------
 *  Function    : output_rangelist
 *  Purpose     : Print the range-list into a buffer
 *  Method      :
 *  Run down linked-list and output every element separately.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *  15-09-1998  : fb       : moved while condition to top of loop to avoid
 *                           using a NULL pointer.
 *  17-10-2000  : paul     : 0-terminate the output buffer even in case of
 *                           error.
 *  29-11-2001  : paul     : Don't print r_high if it is equal to ULONG_MAX
 *  22-03-2016	: paul     : CBS-6594 Return required size if buffer overlows
 *
 *  Notes       :
 *  Uses a buffer to avoid overflowing the output buffer.
 *
 *  Returns     : Number of element in the list or 0 in case of overflow.
 *------------------------------------------------------------------------
 */
static unsigned long
output_rangelist(
        range_p         ran_list,       /* the collated range-list */
        char            *o_buf,         // output buffer */
        size_t          buf_size,       /* size of the output buffer */
        size_t          *buf_size_req,	/* out: required size of output buffer
					   or 0 if an error occurred
					   or NULL if caller isn't interested */
        int             base)
{
#define PM_RANGE_APPEND(o_buf, tmp_buf) \
    { \
	unsigned		cpy_len = strlen(tmp_buf); \
	req_size += cpy_len; \
	if (cpy_len > buf_size - tot_len) { \
	    cpy_len = buf_size - tot_len; \
	} \
	(void)strncpy(*(o_buf), (tmp_buf), buf_size - tot_len); \
	*(o_buf) += cpy_len; \
	tot_len += cpy_len; \
    }
    
    unsigned long       nr_elts = 0;
    unsigned            tot_len = 0;
    unsigned            req_size = 0;	// required length
    char                tmp_buf[PM_INTLEN_MAX];

    while (ran_list != NULL) {
        nr_elts += ran_list->r_high - ran_list->r_low + 1;

	if (base == 10) {
	    (void) PM_ultoa(tmp_buf, ran_list->r_low);
	} else {
	    sprintf(tmp_buf, "%lx", ran_list->r_low);
	}
	
	PM_RANGE_APPEND(&o_buf, tmp_buf);

        if (ran_list->r_low != ran_list->r_high) {
	    if (ran_list->r_high != ULONG_MAX) {
		if (base == 10) {
		    (void) PM_ultoa(tmp_buf, ran_list->r_high);
		} else {
		    sprintf(tmp_buf, "%lx", ran_list->r_high);
		}
	    } else {
		tmp_buf[0] = '\0';
	    }

	    PM_RANGE_APPEND(&o_buf, "-");
	    PM_RANGE_APPEND(&o_buf, tmp_buf);
        }

        ran_list = ran_list->r_next;
        if (ran_list)  {
	    PM_RANGE_APPEND(&o_buf, ",");
        }
    }

    // Append terminating 0-byte
    if (buf_size - tot_len >= 1) {
	*o_buf = '\0';
    } else if (buf_size != 0) {
	// Truncate the output range-list
	o_buf[-1] = '\0';
    }

    // Required size includes 0-byte
    req_size++;

    // Return the required size, if requested
    if (buf_size_req != NULL) {
	*buf_size_req = req_size;
    }

    // Return 0 in case of buffer overflow
    if (req_size > buf_size) {
	return 0;
    }

    return nr_elts;
} /* output_rangelist() */

/*------------------------------------------------------------------------
 *  Function    : range_clean
 *  Purpose     : Clean a string with ranges so there is no overlap
 *  Method      :
 *  - Generate a linked list with all the elements from the string.
 *  - Collate the linked list, so that consecutive elements do not overlap.
 *  - Write the result back to the string.
 *  - Deallocate the linked list.
 *
 *  History:
 *  13-01-1995  : eck      : creation
 *
 *  Returns     : The total number of elements (or 0 on error).
 *------------------------------------------------------------------------
 */
static unsigned long
range_clean(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   low,            /* lower bound for the elements */
        unsigned long   high,           /* upper bound for the elements */
        int             base)
{
    range_p             ran_list = NULL;
    unsigned long       nr_elts;

    if (low > high) return 0;

    ran_list = parse_ranges(rangebuf, low, high, base);
    if (ran_list == NULL) return 0;

    ran_list = collate_rangelist(ran_list);
    nr_elts = output_rangelist(ran_list, rangebuf, buf_size, NULL, base);
    clear_rangelist(ran_list);

    return nr_elts;
} /* range_clean() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_clean
 *  Purpose     : Clean a string with ranges so there is no overlap
 *  Method      :
 *  Method      : Call range_clean() with base 10
 *
 *  History:
 *  02-07-1996  : eck      : creation
 *
 *  Returns     : The total number of elements (or 0 on error).
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_clean(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   low,            /* lower bound for the elements */
        unsigned long   high)           /* upper bound for the elements */
{
    return range_clean(rangebuf, buf_size, low, high, 10);
} /* PM_range_clean() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_x_clean
 *  Purpose     : Clean a hexadecimal string with ranges so there is no overlap
 *  Method      : Call range_clean() with base 16
 *
 *  History:
 *  05-07-2006	: heesakkm : creation
 *
 *  Returns     : The total number of elements (or 0 on error).
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_x_clean(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   low,            /* lower bound for the elements */
        unsigned long   high)           /* upper bound for the elements */
{
    return range_clean(rangebuf, buf_size, low, high, 16);
} /* PM_range_x_clean() */

/*------------------------------------------------------------------------
 *  Function    : range_add
 *  Purpose     : Add a range to an existing range-buffer.
 *  Method      :
 *  Convert the range-buffer to an element-list, add the new element and
 *  collate.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Notes       : Cleans up the range-buffer.
 *
 *  Returns     :
 *  The number of element in the new range-buffer or 0 if something went
 *  wrong.
 *------------------------------------------------------------------------
 */
static unsigned long
range_add(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   low,            /* Lowest to add */
        unsigned long   high,           /* highest to add */
        int             base)
{
    unsigned long       nr_elts;
    range_p             ran_list = NULL;

    if (low > high) return 0;

    ran_list = parse_ranges(rangebuf, 0, ULONG_MAX, base);
    if (ran_list == NULL) return 0;

    if ((ran_list = add_range(ran_list, low, high)) == NULL) return 0;

    ran_list = collate_rangelist(ran_list);
    nr_elts = output_rangelist(ran_list, rangebuf, buf_size, NULL, base);
    clear_rangelist(ran_list);

    return nr_elts;
} /* range_add() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_add
 *  Purpose	: Add a range to an existing range-buffer.
 *  Method	: Call range_add() with base 10
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Notes       : Cleans up the range-buffer.
 *
 *  Returns	:
 *  The number of element in the new range-buffer or 0 if something went
 *  wrong.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_add(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   low,            /* Lowest to add */
        unsigned long   high)           /* highest to add */
{
    return range_add(rangebuf, buf_size, low, high, 10);
} /* PM_range_add() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_x_add
 *  Purpose	: Add a range to an existing hexadecimal range-buffer.
 *  Method	: Call range_add() with base 16
 *
 *  History:
 *  05-07-2006	: heesakkm : creation
 *
 *  Notes       : Cleans up the range-buffer.
 *
 *  Returns	:
 *  The number of element in the new range-buffer or 0 if something went
 *  wrong.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_x_add(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   low,            /* Lowest to add */
        unsigned long   high)           /* highest to add */
{
    return range_add(rangebuf, buf_size, low, high, 16);
} /* PM_range_x_add() */

/*------------------------------------------------------------------------
 *  Function    : range_del_number
 *  Purpose     : Delete a number from a range-buffer
 *  Method      :
 *  Convert the buffer to a range-list, then delete the number from the
 *  range-list and convert the range-lits back into a string.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Returns     : The number of elements in the new range_buffer.
 *------------------------------------------------------------------------
 */
static unsigned long
range_del_number(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   number,         /* Number to delete */
        int             base)
{
    unsigned long       nr_elts;
    range_p             ran_list = NULL;

    ran_list = parse_ranges(rangebuf, 0, ULONG_MAX, base);
    if (ran_list == NULL) return 0;

    ran_list = collate_rangelist(ran_list);
    ran_list = del_number(ran_list, number);

    nr_elts = output_rangelist(ran_list, rangebuf, buf_size, NULL, base);
    clear_rangelist(ran_list);

    return nr_elts;
} /* range_del_number() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_del_number
 *  Purpose	: Delete a number from a range-buffer
 *  Method	: Call range_del_number() with base 10
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Returns     : The number of elements in the new range_buffer.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_del_number(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   number)         /* Number to delete */
{
    return range_del_number(rangebuf, buf_size, number, 10);
} /* PM_range_del_number() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_x_del_number
 *  Purpose	: Delete a number from a hexadecimal range-buffer
 *  Method	: Call range_del_number() with base 16
 *
 *  History:
 *  05-07-2006	: heesakkm : creation
 *
 *  Returns     : The number of elements in the new range_buffer.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_x_del_number(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the buffer */
        unsigned long   number)         /* Number to delete */
{
    return range_del_number(rangebuf, buf_size, number, 16);
} /* PM_range_x_del_number() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_combine
 *  Purpose	: Unite/intersect/subtract two range-buffers
 *  Method      :
 *  Convert the two buffer to range-lists, then walk thru their ranges in
 *  paralel to find their combination and create range-list and convert the
 *  range-list back into a string.
 *
 *  History:
 *  27-11-2001  : paul     : creation
 *  22-03-2016	: paul     : generalized from range_intersect() / range_unite()
 *  25-04-2018	: zwaalm   : CBS-8174 Fixed wrong (empty) result when
 *  			     destbuf equals src1buf or src2buf.
 *
 *  Notes       : This function uses the convention that the string
 *                PM_RANGE_BUF_EMPTY denotes an empty range.
 *
 *  Returns	:
 *  - The number of elements in the non-empty combination
 *  - or 0 && *buf_size_req==1 in case of empty combination
 *  - or 0 && *buf_size_req>1 in case of buffer overflow
 *  - or 0 && *buf_size_req==0 in case of error
 *
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_combine(
	PM_range_combine_op_t	op,		// unite, intersect or subtract
        char			*destbuf,	// output buffer with the ranges
        size_t			buf_size,	// size of the output buffer
        size_t			*buf_size_req,	/* out: required size of output buffer
						   or 0 if an error occurred
						   or NULL if caller isn't interested */
	const char		*src1buf,	// input buffer with ranges
	const char		*src2buf,	// input buffer with ranges
        int             	base)		// either 10 or 16
{
    unsigned long       nr_elts;
    range_p             dest_list = NULL, src1_list = NULL, src2_list = NULL;
    bool		ok = TRUE;

    // Parse the first input buffer
    if (!STREQ(src1buf, PM_RANGE_BUF_EMPTY)) {
	src1_list = parse_ranges(src1buf, 0, ULONG_MAX, base);
	if (src1_list == NULL) {
	    ok = FALSE;
	}
    }

    // Parse the second input buffer
    if (!STREQ(src2buf, PM_RANGE_BUF_EMPTY)) {
	src2_list = parse_ranges(src2buf, 0, ULONG_MAX, base);
	if (src2_list == NULL) {
	    ok = FALSE;
	}
    }

    if (ok) {
	// Normalize both ranges
	src1_list = collate_rangelist(src1_list);
	src2_list = collate_rangelist(src2_list);

	// Create the range that is the combination of the input ranges
	switch (op) {
	case PMRC_OP_UNITE:
	    dest_list = unite_rangelist(src1_list, src2_list, &ok);
	    break;
	case PMRC_OP_INTERSECT:
	    dest_list = intersect_rangelist(src1_list, src2_list, &ok);
	    break;
	case PMRC_OP_SUBTRACT:
	    dest_list = subtract_rangelist(src1_list, src2_list, &ok);
	    break;
	default:
	    ok = FALSE;
	    break;
	}
    }

    clear_rangelist(src1_list);
    clear_rangelist(src2_list);

    // Initialize output parameter.
    if (buf_size_req != NULL) {
	*buf_size_req = 0;
    }

    // Support destbuf to be equal to src1buf or src2buf
    // by initializing destbuf not too early.
    if (buf_size != 0) {
	destbuf[0] = '\0';
    }

    // Ensure that *buf_size_req (if requested) and destbuf are set on error.
    // We want to keep compatible with clients that don't check our result.
    if (!ok) {
	return 0;
    }

    // Print the subtraction range
    nr_elts = output_rangelist(dest_list, destbuf, buf_size, buf_size_req, base);

    // Destroy the subtraction range
    clear_rangelist(dest_list);

    return nr_elts;
} // PM_range_combine()

unsigned long
PM_range_intersect(
        char            *destbuf,  /* output buffer with the ranges */
        size_t          buf_size,  /* size of the output buffer */
	const char      *src1buf,  /* input buffer with ranges */
	const char      *src2buf)  /* input buffer with ranges */
{
    return PM_range_combine(PMRC_OP_INTERSECT, destbuf, buf_size, NULL,
			    src1buf, src2buf, 10);
} /* PM_range_intersect() */

unsigned long
PM_range_x_intersect(
        char            *destbuf,  /* output buffer with the ranges */
        size_t          buf_size,  /* size of the output buffer */
	const char      *src1buf,  /* input buffer with ranges */
	const char      *src2buf)  /* input buffer with ranges */
{
    return PM_range_combine(PMRC_OP_INTERSECT, destbuf, buf_size, NULL,
			    src1buf, src2buf, 16);
} /* PM_range_x_intersect() */

unsigned long
PM_range_unite(
        char            *destbuf,  /* output buffer with the ranges */
        size_t          buf_size,  /* size of the output buffer */
	const char      *src1buf,  /* input buffer with ranges */
	const char      *src2buf)  /* input buffer with ranges */
{
    return PM_range_combine(PMRC_OP_UNITE, destbuf, buf_size, NULL,
			    src1buf, src2buf, 10);
} /* PM_range_unite() */

unsigned long
PM_range_x_unite(
        char            *destbuf,  /* output buffer with the ranges */
        size_t          buf_size,  /* size of the output buffer */
	const char      *src1buf,  /* input buffer with ranges */
	const char      *src2buf)  /* input buffer with ranges */
{
    return PM_range_combine(PMRC_OP_UNITE, destbuf, buf_size, NULL,
			    src1buf, src2buf, 16);
} /* PM_range_x_unite() */

/*------------------------------------------------------------------------
 *  Function    : set_bit_range
 *  Purpose     : Set a range of bits in a single byte to 1.
 *  Method      :
 *
 *  History:
 *  16-10-2000  : paul     : creation
 *
 *  Notes       :
 *
 *------------------------------------------------------------------------
 */
static void
set_bit_range(
        byte            *bitarr,        /* byte to set bits in */
        unsigned        first,          /* first bit to set to 1 */
        unsigned        last)           /* last bit to set to 1 */
{
    unsigned       mask1, mask2;

    mask1 = UINT8MAX >> first;
    mask2 = (UINT8MAX << (PIC_CHARBIT - last - 1));
    *bitarr = *bitarr | (mask1 & mask2);
} /* set_bit_range() */

/*------------------------------------------------------------------------
 *  Function    : scan_bit_range
 *  Purpose     : Search the next 0- or 1-bit in a bit array.
 *  Method      :
 *
 *  History:
 *  16-10-2000  : paul     : creation
 *
 *  Notes       :
 *
 *  Returns     :
 *  The sequence number of the bit or if not found, 1 plus the sequence
 *  number of the last bit in the array.
 *------------------------------------------------------------------------
 */
static
unsigned long
scan_bit_range(
        unsigned        bitval,     /* the bit to search for */
        const byte      *bitarr,    /* buffer with the bits */
        unsigned long   begnr,      /* the seq. nr. of the first bit to test */
        unsigned long   endnr)      /* the number of bits in bitarr */
{
    unsigned long       bytenr;
    unsigned            bitnr;

    if (begnr > endnr) begnr = endnr;
    for (; begnr < endnr; begnr++) {
        bytenr = begnr / PIC_CHARBIT;
        bitnr = (unsigned)(begnr % PIC_CHARBIT) + 1;
        if (   ((bitarr[bytenr] >> (PIC_CHARBIT - bitnr)) & (unsigned)0x01)
            == bitval) {
            break;
        }
    }
    return begnr;
} /* scan_bit_range() */

/*------------------------------------------------------------------------
 *  Function    : range_from_bits
 *  Purpose     : Convert a bit-array to a range-buffer.
 *  Method      : Use scan_bit_range() and add_range() to create an
 *                element-list and output.
 *
 *  History:
 *  16-10-2000  : paul     : creation
 *  05-08-2009	: sutherlr : Now builds the range string itself;
 *                           new algorithms works in linear time.
 *  21-03-2012	: martinz  : Fixed crash on NULL rangebuf.
 *
 *  Returns     :
 *  The number of elements in the new range-buffer
 *  or PM_RANGE_ERROR if something went wrong.
 *------------------------------------------------------------------------
 */
static unsigned long
range_from_bits(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the ranges-buffer */
        const byte      *bitarr,        /* buffer with the bits */
        size_t          bitlen,         /* size of bit-buffer */
        unsigned long   lowest)         /* number indicated by bit begnr 0 */
{
    unsigned long       nr_elts = 0;
    unsigned long       begnr, endnr, low, high;
    char		*rangebuf_end = rangebuf;

    if (buf_size == 0)    return PM_RANGE_ERROR;
    if (bitarr == NULL)   return PM_RANGE_ERROR;
    if (rangebuf == NULL) return PM_RANGE_ERROR;
    rangebuf[0] = '\0';

    begnr = 0;
    endnr = PIC_CHARBIT * bitlen;
    begnr = scan_bit_range(1, bitarr, begnr, endnr);
    if (begnr >= endnr) {
/*
 *      Empty bit array. That's not an error, just zero elements.
 */
        rangebuf[0] = '\0';
        return 0;
    }
    do {
	char	num_buf[PM_INTLEN_MAX];
	int	num_len;

        low = begnr + lowest;

        begnr = scan_bit_range(0, bitarr, begnr + 1, endnr);
        high = begnr - 1 + lowest;

	PM_ultoa(num_buf, low);
	num_len = strlen(num_buf);
	if (num_len + (rangebuf_end - rangebuf) + 1 >= buf_size) {
	    return PM_RANGE_ERROR;
	}

	if (rangebuf != rangebuf_end) {
	    *rangebuf_end++ = ',';
	}

	strcpy(rangebuf_end, num_buf);
	rangebuf_end += num_len;

	if (low != high) {
	    PM_ultoa(num_buf, high);
	    num_len = strlen(num_buf);
	    if (num_len + (rangebuf_end - rangebuf) + 1 >= buf_size) {
		return PM_RANGE_ERROR;
	    }
	    *rangebuf_end++ = '-';
	    strcpy(rangebuf_end, num_buf);
	    rangebuf_end += num_len;
	}
	nr_elts += (high - low) + 1;

        begnr = scan_bit_range(1, bitarr, begnr + 1, endnr);
    } while (begnr < endnr);

    return nr_elts;
} /* range_from_bits() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_from_bits
 *  Purpose	: Convert a bit-array to a range-buffer.
 *  Method	: Call range_from_bits()
 *
 *  History:
 *  17-10-2000  : paul     : creation
 *  05-07-2006	: heesakkm : Call range_from_bits().
 *
 *  Returns	:
 *  The number of elements in the new range-buffer
 *  or PM_RANGE_ERROR if something went wrong.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_from_bits(
        char            *rangebuf,      /* buffer with the ranges */
        size_t          buf_size,       /* size of the ranges-buffer */
        const byte      *bitarr,        /* buffer with the bits */
        size_t          bitlen,         /* size of bit-buffer */
        unsigned long   lowest)         /* number indicated by bit begnr 0 */
{
    return range_from_bits(rangebuf, buf_size, bitarr, bitlen, lowest);
} /* PM_range_from_bits() */

/*------------------------------------------------------------------------
 *  Function    : range_to_bits_check
 *  Purpose     : Convert a range-buffer to a bit-array and check that all
 *                chars in the range-buffer are used.
 *  Method      :
 *  Convert the range-buffer to an element-list, add the new element and
 *  collate.
 *
 *  History:
 *  28-06-2004  : paul     : creation
 *
 *  Notes       :
 *  PM_RANGE_ERROR indicates buffer overflow if *bitlen > bitsize. Otherwise
 *  it indicates an incorrect range in rangebuf.
 *
 *  Returns     :
 *  The number of elements in the new bit-array
 *  or PM_RANGE_ERROR if something went wrong, for example if not all chars in
 *  the rangebuf were used when encountering a 0-byte.
 *------------------------------------------------------------------------
 */
static unsigned long
range_to_bits_check(
        byte            *bitarr,        /* buffer with the bits */
        size_t          bitsize,        /* size of bit-buffer */
        size_t          *bitlen,        /* length of bit-buffer */
        const char      *rangebuf,      /* buffer with the ranges */
	long		rangebuflen,	/* length of the ranges-buffer */
        unsigned long   lowest,         /* number indicated by bit begnr 0 */
        int             base)
{
    unsigned long       nr_elts = 0;
    unsigned long       first;
    unsigned long       last;
    unsigned long       first_bytenr, last_bytenr;
    unsigned            first_bitnr , last_bitnr;
    const char          *range = rangebuf;

    if (bitarr == NULL)   return PM_RANGE_ERROR;
    if (rangebuf == NULL) return PM_RANGE_ERROR;

    (void) memset(bitarr, '\0', bitsize);
    *bitlen = 0;
    if (rangebuf[0] == '\0') {
/*
 *      Empty range buffer. That's not an error, just zero elements.
 */
        return 0;
    }

    /* Check all elements. They need not be ordered. */
    while ((range = range_get_bounds(range, &first, &last, base)) != NULL) {
        nr_elts += last - first + 1;
/*
 *      Set the bits corresponding to first through last
 */
        if (first < lowest) return PM_RANGE_ERROR;
        first_bytenr = (first - lowest) / PIC_CHARBIT;
        last_bytenr  = (last - lowest)  / PIC_CHARBIT;
        first_bitnr  = (unsigned)((first - lowest) % PIC_CHARBIT);
        last_bitnr   = (unsigned)((last  - lowest) % PIC_CHARBIT);
        if (*bitlen < last_bytenr + 1) {
            *bitlen = last_bytenr + 1;
        }
        if (*bitlen > bitsize) return PM_RANGE_ERROR;
        if (first_bytenr == last_bytenr) {
            set_bit_range(bitarr + first_bytenr, first_bitnr, last_bitnr);
        } else {
            set_bit_range(bitarr + first_bytenr, first_bitnr, PIC_CHARBIT - 1);
            for (first_bytenr++;
                 first_bytenr <= last_bytenr - 1;
                 first_bytenr++) {
                bitarr[first_bytenr] = UINT8MAX;
            }
            set_bit_range(bitarr + last_bytenr, 0, last_bitnr);
        }

        if (*range == '\0') {
	    if (rangebuflen < 0 || rangebuflen == range - rangebuf) {
	    /*
	     * A correct rangebuf has been completely processed.
	     */
		return nr_elts;
	    } else {
	    /*
	     * rangebuf contains 0-bytes: return error
	     */
		return PM_RANGE_ERROR;
	    }
	}
    }

    return PM_RANGE_ERROR;
} /* range_to_bits_check() */

/*------------------------------------------------------------------------
 *  Function    : PM_range_to_bits_check
 *  Purpose     : Convert a range-buffer to a bit-array and check that all
 *                chars in the range-buffer are used.
 *  Method      : Call range_to_bits_check() with base 10
 *
 *  History:
 *  05-07-2006	: heesakkm : creation
 *
 *  Notes       :
 *  PM_RANGE_ERROR indicates buffer overflow if *bitlen > bitsize. Otherwise
 *  it indicates an incorrect range in rangebuf.
 *
 *  Returns     :
 *  The number of elements in the new bit-array
 *  or PM_RANGE_ERROR if something went wrong, for example if not all chars in
 *  the rangebuf were used when encountering a 0-byte.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_to_bits_check(
        byte            *bitarr,        /* buffer with the bits */
        size_t          bitsize,        /* size of bit-buffer */
        size_t          *bitlen,        /* length of bit-buffer */
        const char      *rangebuf,      /* buffer with the ranges */
	long		rangebuflen,	/* length of the ranges-buffer */
        unsigned long   lowest)         /* number indicated by bit begnr 0 */
{
    return range_to_bits_check(bitarr, bitsize, bitlen, rangebuf, rangebuflen,
				lowest, 10);
} /* PM_range_to_bits_check() */

/*------------------------------------------------------------------------
 *  Function    : range_to_bits
 *  Purpose     : Convert a range-buffer to a bit-array.
 *  Method      :
 *  Convert the range-buffer to an element-list, add the new element and
 *  collate.
 *
 *  History:
 *  16-10-2000  : paul     : creation
 *  28-06-2004  : paul     : code moved to PM_range_to_bits_check()
 *
 *  Notes       :
 *  PM_RANGE_ERROR indicates buffer overflow if *bitlen > bitsize. Otherwise
 *  it indicates an incorrect range in rangebuf.
 *
 *  Returns     :
 *  The number of elements in the new bit-array
 *  or PM_RANGE_ERROR if something went wrong.
 *------------------------------------------------------------------------
 */
static unsigned long
range_to_bits(
        byte            *bitarr,        /* buffer with the bits */
        size_t          bitsize,        /* size of bit-buffer */
        size_t          *bitlen,        /* length of bit-buffer */
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   lowest,         /* number indicated by bit begnr 0 */
        int             base)
{
    return range_to_bits_check(bitarr, bitsize, bitlen, rangebuf, -1,
				lowest, base);
} /* range_to_bits() */

/*------------------------------------------------------------------------
 *  Function	: PM_range_to_bits
 *  Purpose     : Convert a range-buffer to a bit-array.
 *  Method	: Call range_to_bits() with base 10
 *
 *  History:
 *  20-10-2000  : paul     : creation
 *  05-07-2006	: heesakkm : Call range_to_bits().
 *
 *  Notes	:
 *  PM_RANGE_ERROR indicates buffer overflow if *bitlen > bitsize. Otherwise
 *  it indicates an incorrect range in rangebuf.
 *
 *  Returns	:
 *  The number of elements in the new bit-array
 *  or PM_RANGE_ERROR if something went wrong.
 *------------------------------------------------------------------------
 */
unsigned long
PM_range_to_bits(
        byte            *bitarr,        /* buffer with the bits */
        size_t          bitsize,        /* size of bit-buffer */
        size_t          *bitlen,        /* length of bit-buffer */
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   lowest)         /* number indicated by bit begnr 0 */
{
    return range_to_bits(bitarr, bitsize, bitlen, rangebuf, lowest, 10);
} /* PM_range_to_bits() */

/*------------------------------------------------------------------------
 *  Function    : range_error
 *  Purpose     : Find the first error in a range-buffer
 *  Method      : range_get_bounds()
 *
 *  History:
 *  03-12-2001  : paul     : creation
 *
 *  Notes       :
 *
 *  Returns     :
 *  Pointer to the error in the range or NULL. NULL indicates valid range.
 *------------------------------------------------------------------------
 */
static const char *
range_error(
        const char      *rangebuf,      /* buffer to check */
        int             base)
{
    const char         *cur, *prev;
    unsigned long       first, last;

    prev = rangebuf;
    while ((cur = range_get_bounds(prev, &first, &last, base)) != NULL) {
	prev = cur;
    }
    while (*prev == ' ') prev++;        /* skip spaces */
    if (*prev != '\0') {
	return prev;
    }

    return NULL;
} /* range_error() */

const char *
PM_range_error(
        const char      *rangebuf)      /* buffer to check */
{
    return range_error(rangebuf, 10);
} /* PM_range_error() */

const char *
PM_range_x_error(
        const char      *rangebuf)      /* buffer to check */
{
    return range_error(rangebuf, 16);
} /* PM_range_x_error() */

static bool
range_test_number(
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   number,         /* number to test */
        int             base)
{
    unsigned long       first;
    unsigned long       last;

    /* Check all elements. They need not be ordered. */
    while ((rangebuf = range_get_bounds(rangebuf, &first, &last, base)) != NULL) {
        if ((first <= number) && (number <= last)) return TRUE;
    }
    return FALSE;
} /* range_test_number() */

/*------------------------------------------------------------------------
 *  Function    : PM_range_test_number
 *  Purpose     : Test if a range-buffer contains a certain number.
 *  Method      : Repeated calls of range_get_bounds() and simple tests.
 *
 *  History:
 *  06-02-1996  : eck      : creation
 *
 *  Returns     : indication of whether number is in the range-buffer.
 *------------------------------------------------------------------------
 */
bool
PM_range_test_number(
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   number)         /* number to test */
{
    return range_test_number(rangebuf, number, 10);
} /* PM_range_test_number() */

bool
PM_range_x_test_number(
        const char      *rangebuf,      /* buffer with the ranges */
        unsigned long   number)         /* number to test */
{
    return range_test_number(rangebuf, number, 16);
} /* PM_range_x_test_number() */

/*------------------------------------------------------------------------
 *  Function	: range_count
 *  Purpose	: Count the number of elements in a range
 *  Method	: 
 *
 *  History:
 *  21-02-2012	: paul     : creation
 *
 *  Notes	:
 *
 *  Returns	: The number of elements in a range
 *                or 0 in case of parse error
 *
 *------------------------------------------------------------------------
 */
static unsigned long
range_count(
        const char      *rangebuf,      /* buffer with the ranges */
        int             base)
{
    unsigned long       nr_elts = 0;
    range_p             ran_list = NULL;
    range_p             ran_elt;

    ran_list = parse_ranges(rangebuf, 0, ULONG_MAX, base);
    if (ran_list == NULL) return 0;

    ran_list = collate_rangelist(ran_list);

    for (ran_elt = ran_list; ran_elt != NULL; ran_elt = ran_elt->r_next) {
        nr_elts += ran_elt->r_high - ran_elt->r_low + 1;
    }

    clear_rangelist(ran_list);

    return nr_elts;
} /* range_count() */

unsigned long
PM_range_count(
        const char      *rangebuf)      /* buffer with the ranges */
{
    return range_count(rangebuf, 10);
} /* PM_range_count() */

unsigned long
PM_range_x_count(
        const char      *rangebuf)      /* buffer with the ranges */
{
    return range_count(rangebuf, 16);
} /* PM_range_x_count() */

