/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: pmscramble.c
 *  Description	:
 *  Author(s)	: Rudy Tan
 *  Creation	: 27-08-1996
 *
 *  History:
 *  27-08-1996	: rudy     : creation
 *  31-01-1997	: eck      : Changed scarmbling drastically.
 *  06-02-1997	: lolke    : Type cast in PM_scramble doesn't work on TANDEM.
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: pmscramble.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <picamisc.h>

/*------------------------------------------------------------------------
 *  Function    : PM_old_scramble
 *  Purpose     : Scramble NO encryption of code.
 *  Method      : Yes,in a way.
 *  Author(s)   : Rudy Tan
 *  Creation    : 01-07-1996
 *
 *  History:
 *  01-07-1996  : rudy          : creation
 *  19-08-1996  : rudy/lolke    : refined
 *  27-08-1996  : HvE           : deleted code
 *  03-09-1996  : LB            : changed return type to int
 *  30-01-1997	: eck      : Renamed to PM_old_scramble().
 *
 *  Returns     : Amount of scrambled bytes, -1 on error.
 *------------------------------------------------------------------------
 */
int
PM_old_scramble(
    byte *outstring,      /* o scrambled version of instring */
    size_t size_out,      /* i size of outbuffer */
    const char *instring, /* i original string */
    int used_len,         /* i used_length from instring */
    int32 seed)           /* i seed */
{
    unsigned char *ptr;
    byte byte_buf[4];
    int i,j,k;
    int shift = ((seed % 3) + 1);
    unsigned char masks[10] = {0xFF,0xFE,0xFC,0xF8,0xF0,
                               0xFF,0xEF,0xCF,0x8F,0x0F};

    if (used_len < 0){
	used_len = strlen(instring);
    }
    if(size_out < used_len){
        return -1;
    }
    byte_buf[0] = (byte) ((seed >> 24) & 0xFF);
    byte_buf[1] = (byte) ((seed >> 16) & 0xFF);
    byte_buf[2] = (byte) ((seed >> 8) & 0xFF);
    byte_buf[3] = (byte) (seed & 0xFF);
    ptr = (unsigned char *) byte_buf;
    for (i = 0; i < 4; i++) {
        for (j = 0; j < 4; j++) {
            if (i != j){
                ptr[i] ^= (ptr[j] >> j & masks[5+j]);
            }
        }
    }
    for (i = j = 0; i < used_len; j = (++i % 4)) {
        outstring[i] = (byte) (instring[i] ^ (ptr[j] + (char)(i * ptr[j])));
        k = ((int) ptr[j]) % shift;
        outstring[i] = (outstring[i] << k & masks[k]);
    }
    return used_len;
} /* PM_old_scramble() */

/*------------------------------------------------------------------------
 *  Function	: PM_scramble
 *  Purpose     : Scramble passwords and system-id's
 *  Method	: Hashing of public and private parts.
 *
 *  History:
 *  30-01-1997	: eck      : creation, algorithm by Rob Koopman
 *
 *  Notes	: It will never replace DES.
 *
 *  PM_scramble() is used for hashing secret strings (passwords and client-type
 *  id's). The seed is returned by a server. For safety reasons, the secret
 *  string should be as short-lived as possible. For instance, when it consists
 *  of a string of digits, the original integer(s) should be stored. Then, the
 *  string should be made, PM_scramble() be called, and the string be cleared
 *  (fi. with memset()). This procedure makes it less easy for an outsider to
 *  learn the secret string.
 *
 *  Example	:
 *
 *      int32	val;
 *	val = PM_scramble("bicycle", "12345");	// returns 815987
 *	val = PM_scramble("bicycle", "12346");	// returns 513618
 *	val = PM_scramble("fiets", "12345");	// returns 49403
 *
 *  Returns	: A hash of the combination of secret and seed.
 *------------------------------------------------------------------------
 */
int32
PM_scramble(
	const char	*secret,	/* password or sys. id. */
	const char	*public)	/* public seed */
{
    uint32		retval;
    uint32		fornow;
    const unsigned char	*cp;

    /* first get a starting value */
    retval = 4711;

    /* simply mingle password */
    cp = (const unsigned char *) secret;
    while (*cp != '\0') {
	retval = (retval*107 + 1234567 * (*cp++)) % 3141557;
    }

    /* add salt */
    cp = (const unsigned char *) public;
    while (*cp != '\0') {
	retval = (retval*107 + 1234567 * (*cp++)) % 3141557;
    }
    fornow = retval;

    /* a bit more password */
    cp = (const unsigned char *) secret;
    while (*cp != '\0') {
	retval = (retval*317 + 2718281 * (*cp++)) % 3141557;
    }

    /* stop to easy stuff */
    retval ^= fornow / 256;
    retval ^= fornow * 256;
    retval %= 1000000;

    return (int32) retval;
} /* PM_scramble() */
