/*========================================================================
 *  Copyright (c) 1997 by Pica
 *
 *  File	: pmsortstring.c
 *  Description	: Create the sortstring for level 0 bibliographic records
 *  Author(s)	: Eric Helsper
 *  Creation	: 29-04-1999
 *
 *  History:
 *  29-04-1999	: Eric : creation
 *  20-09-1999	: Lolke : corrected the routine
 *  26-03-2004  : paul  : Fixed bug: neither "mrt" nor "dec" are Roman numbers
 *  05-04-2004  : paul  : Fixed bug: coredump if month is absent
 *  30-05-2006	: paul     : Roman() moved to pmtoint.c as PM_roman2l()
 *                         : Month() moved to pmtoint.c as PM_month2l()
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: pmsortstring.c 307506 2014-01-22 10:52:31Z sieburgw $"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

#include <pica.h>
#include <picamisc.h>

/*------------------------------------------------------------------------
 * Function    : PM_MakeSortString
 * Purpose     : Formats the sort string with year, month, day, volume,
 *               number, page.
 * Author(s)   : Eric Helsper
 * Creation    : 05-01-1998
 *
 * History:
 * 05-01-1998  : eric - creation
 * 24-07-1998  : eric - changed calculation of page component from sbf h
 *               to be more compatible with the old way (which was faulty).
 * 29-04-1999  : eric - Moved to PM; interface changed and some things more
 *               robust because of promotion to library function.
 *  11-03-2004	: heesakkm : Changed int32 to long.
 *
 * Description	:
 * This produces the so-called sort string, typically present in level 0
 * records as 039$x. The string is composed from the first six input arguments,
 * as listed in the table below. The input values are usually taken from the
 * tag 031A of the title being processed (column Source) but may be modified
 * or even generated by the caller program.
 *
 * |Item      | Source |Allowed input format                | Output width |
 * |:---------|:-------|:-----------------------------------|:-------------|
 * |year      | 031A$j |arabic                              | 4            |
 * |month     | 031A$c |arabic/roman/month name/season name | 2            |
 * |day       | 031A$b |arabic				    | 2            |
 * |volume    | 031A$d |arabic/roman		 	    | 3            |
 * |issue     | 031A$e |arabic/roman			    | 4            |
 * |pagination| 031A$h |arabic, see *)			    | 3            |
 *
 * *) (999 - (page - 1) mod 1000), ignore alphas
 *
 * Output fields are written into the sort string with leading zeroes, no
 * separators. Numerical fields are modulated with a power of 10 to never
 * exceed the field width. PM_MakeSortString has a few help functions for
 * conversion of Roman numbers and month and season names.
 *
 * Returns     : success
 *------------------------------------------------------------------------
 */
bool
PM_MakeSortString(
	const char     *year_sbf,       /* j */
	const char     *month_sbf,      /* c */
	const char     *day_sbf,        /* b */
	const char     *volume_sbf,     /* d */
	const char     *issue_sbf,      /* e */
	const char     *page_sbf,       /* h */
	char           *sortstring,	/* The output string */
	const size_t   sortstring_size)	/* The size allocated for the
					   sortstring. This should be at least
					   the sum of output field widths in
					   the table plus one for the trailing
					   zero.*/
{
#define WIDTH_YEAR    4
#define WIDTH_MONTH   2
#define WIDTH_DAY     2
#define WIDTH_VOLUME  3
#define WIDTH_ISSUE   4
#define WIDTH_PAGE    3
    static long tenpower[] =
    { 1, 10, 100, 1000, 10000, 100000, 1000000 };
    long year = 0;
    long month = 0;
    long day = 0;
    long volume = 0;
    long issue = 0;
    long page = 0;
    const char  *cp;

    if (volume_sbf != NULL) {
	month_sbf = "";
	day_sbf = "";
    }
    if (sortstring_size <= WIDTH_YEAR +
	                   WIDTH_MONTH +
	                   WIDTH_DAY +
	                   WIDTH_VOLUME +
	                   WIDTH_ISSUE +
	                   WIDTH_PAGE)  return(FALSE);

    if (year_sbf != NULL) {
        year = atol(year_sbf);
    } else {
	year = 0;
    }
    if (month_sbf != NULL) {
        month = atol(month_sbf);
    } else {
	month = 0;
    }
    if (day_sbf != NULL) {
        day = atol(day_sbf);
    } else {
	day = 0;
    }
    if (volume_sbf != NULL) {
        volume = atol(volume_sbf);
    } else {
	volume = 0;
    }
    if (issue_sbf != NULL) {
        issue = atol(issue_sbf);
    } else {
	issue = 0;
    }

    if (month_sbf != NULL && month == 0) {
	month = PM_roman2l(month_sbf, NULL);
    }
    if (month_sbf != NULL && month == 0) {
	month = PM_month2l(month_sbf, NULL);
    }
    if (volume_sbf != NULL && volume == 0) {
	volume = PM_roman2l(volume_sbf, NULL);
    }
    if (issue_sbf != NULL && issue == 0) {
	issue = PM_roman2l(issue_sbf, NULL);
    }

    if (page_sbf != NULL) {
        for (cp = page_sbf; (*cp != '\0') && (!isdigit(*(byte *)cp)); cp++);
        page = atol(cp);
        if (page < 1) page = 1;
    } else {
	page = 1;
    }
    page = (page - 1) % tenpower[WIDTH_PAGE] + 1;
    page = tenpower[WIDTH_PAGE] - page;

    (void) sprintf(sortstring, "%0*ld%0*ld%0*ld%0*ld%0*ld%0*ld", 
		   WIDTH_YEAR, year % tenpower[WIDTH_YEAR],
		   WIDTH_MONTH, month % tenpower[WIDTH_MONTH],
		   WIDTH_DAY, day % tenpower[WIDTH_DAY],
		   WIDTH_VOLUME, volume % tenpower[WIDTH_VOLUME],
		   WIDTH_ISSUE, issue % tenpower[WIDTH_ISSUE],
		   WIDTH_PAGE, page % tenpower[WIDTH_PAGE]);
    return(TRUE);
} /* PM_MakeSortString() */
