/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: pmstring.c
 *  Description	: String routines.
 *  Author(s)	: Hans van Eck
 *  Creation	: 20-05-1996
 *
 *  History:
 *  20-05-1996	: eck      : creation
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *  Notes	:
 *  PM_strcmpi() and PM_strncmpi() are case insensitive versions of strcmp()
 *  and strncmp() respectively.
 *
 *  If the two strings are equal except for their case, the value returned is
 *  0. If not, the value returned is positive when the first string precedes
 *  the second one (case insensitively), negative otherwise. Comparisons will
 *  stop at '\0'-bytes. The routine PM_strncmpi() tests at most len bytes.
 *
 *========================================================================
 */

#pragma ident "$Id: pmstring.c 331081 2015-11-12 14:23:40Z heesakkm $"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <er.h>

/*------------------------------------------------------------------------
 *  Function	: PM_strcmpi
 *  Purpose	: Case-insensitive string compare.
 *  Method	:
 *
 *  History:
 *  20-05-1996	: eck      : creation
 *
 *  Example	:
 *
 *	//Both the comparisons will succeed.
 *	if (PM_strcmpi("test", "TEST") == 0)  printf("Equal\n");
 *	if (PM_strcmpi("abc", "ABD") < 0)  printf("Smaller\n");
 *
 *  Returns	: Difference of the lower-case chars.
 *------------------------------------------------------------------------
 */
int
PM_strcmpi(
	const char	*str1,
	const char	*str2)
{
    while ((*str1 != '\0') && (*str2 != '\0')) {
	if (tolower(*(unsigned char *) str1)
		!= tolower(*(unsigned char *) str2)) {
	    break;
	}
	str1++;
	str2++;
    }
    return tolower(*(unsigned char *) str1) - tolower(*(unsigned char *) str2);
} /* PM_strcmpi() */

/*------------------------------------------------------------------------
 *  Function	: PM_strcmpi
 *  Purpose	: Case-insensitive string compare with length.
 *  Method	:
 *
 *  History:
 *  20-05-1996	: eck      : creation
 *
 *  Returns	: Difference of the lower-case chars.
 *------------------------------------------------------------------------
 */
int
PM_strncmpi(
	const char	*str1,
	const char	*str2,
	size_t		len)
{
    while ((*str1 != '\0') && (*str2 != '\0') && (len != 0)) {
	if (tolower(*(unsigned char *) str1)
		!= tolower(*(unsigned char *) str2)) {
	    break;
	}
	str1++;
	str2++;
	len--;
    }
    if (len == 0) return 0;
    return tolower(*(unsigned char *) str1) - tolower(*(unsigned char *) str2);
} /* PM_strcmpi() */

/*------------------------------------------------------------------------
 *  Function	: PM_strdup
 *  Purpose	: Duplicate a string
 *  Method	: malloc()/strcpy()
 *
 *  History:
 *  16-06-2000	: eck      : creation
 *
 *  Notes	:
 *  This function will not dump core when there is not enough memory.
 *  The caller must free() the string itself.
 *
 *  Returns	: Copy of string or NULL.
 *------------------------------------------------------------------------
 */
char *
PM_strdup(
	const char	*str,
	int		len)
{
    char	*retval;

    if (len < 0) len = strlen(str);
    retval = malloc(len + 1);
    if (retval == NULL) {
	ERWrite(APPLE_NOMEM, "PM_strdup", "Out of memory!");
    } else {
	strncpy(retval, str, len);
	retval[len] = '\0';
    }
    return retval;
} /* PM_strdup() */

/*------------------------------------------------------------------------
 *  Function	: PM_strncpy
 *  Purpose	: Copy a string over a length & enforce '\0'-termination
 *  Method	: Use strncpy() over siz - 1 and force last byte to be '\0'.
 *
 *  History:
 *  27-07-2000	: eck      : creation
 *
 *  Notes	:
 *  This function will ensure that the target is '\0'-terminated (unless the
 *  size of the target is 0).
 *
 *  Avoid problems with a NULL source, so it can be used in code like
 *  PM_strncpy(value, ENVGetVar("YOUZE"), sizeof(value));
 *
 *  Returns	: The target string (for convenience, like strcpy())
 *------------------------------------------------------------------------
 */
char *
PM_strncpy(
	char		*target,	/* Target string */
	const char	*source,	/* Source string (or NULL) */
	size_t		siz)		/* Size of the target */
{
    if (siz == 0) {
	/* Pathetic situation: target will not be '\0' terminated */
	return target;
    }
    if (source == NULL) {
	*target = '\0';
	return target;
    }
    strncpy(target, source, siz - 1);
    target[siz - 1] = '\0';

    return target;
} /* PM_strncpy() */
