/*========================================================================
 *  Copyright (c) 1998 by Pica
 *
 *  File	: pmtime.c
 *  Description	:
 *  Author(s)	: Teresita Frizzarin
 *  Creation	: 23-10-1998
 *
 *  History:
 *  23-10-1998	: teresita  : creation
 *  02-03-1999	: martin    : PM_time_DifferDateTime added
 *  24-03-2000  : teresita  : Added PM_LBSStamp
 *  22-05-2000  : lolke     : Removed SIZE_LBS_STAMP 
 *  11-10-2010	: martinz  : Added PM_time_local_to_UTC().
 *  27-10-2010	: martinz  : Accept 2nd input format in PM_time_local_to_UTC().
 *  20-04-2011	: heesakkm : Removed __TANDEM define for PM_LBSStamp().
 *
 *========================================================================
 */

#pragma ident "$Id: pmtime.c 372490 2019-11-29 15:32:21Z heesakkm $"

#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <ctype.h>

#include <pica.h>
#include <picatypes.h>

#include <picamisc.h>
#include <er.h>

/*------------------------------------------------------------------------
 *  Function	: PM_time_GetDateTime
 *  Purpose	: Get current time in specified format.
 *  Method	: <time.h> routines
 *  Author(s)	: Teresita Frizzarin
 *  Creation	: 23-10-1998
 *
 *  History:
 *  23-10-1998	: teresita  : creation
 *
 *  Notes	:
 *  The Options string is used to regulate the format of the output.
 *
 *  Supported options:
 *
 *	format ( "DATE", "DATETIME" "TIMESTAMP" )
 *
 *------------------------------------------------------------------------
 */
void 
PM_time_GetDateTime(char *s, size_t smax, const char *Options)
{
    time_t tt;
    char tmp[20];

    if ((tt = time(NULL)) == -1) {
	*s = '\0';
	return;
    }
    tmp[0] = '\0';
    PM_opt_fetch(Options, "format", tmp, sizeof(tmp));
    if (strcmp(tmp, "DATE") == 0) {
	(void) strftime(s, smax, "%Y-%m-%d", localtime(&tt));
    } else if (strcmp(tmp, "DATETIME") == 0) {
	(void) strftime(s, smax, "%Y-%m-%d:%H:%M", localtime(&tt));
    } else if (strcmp(tmp, "TIMESTAMP") == 0) {
	(void) strftime(s, smax, "%Y-%m-%d:%H:%M:%S.000000", localtime(&tt));
    } else {
	*s = '\0';
    }
} /* PM_time_GetDateTime */

/*------------------------------------------------------------------------
 *  Function	: PM_time_MakeCTime
 *  Purpose	: Convert time strings to time_t.
 *  Method	: string parsing and <time.h> routines.
 *  Author(s)	: Teresita Frizzarin
 *  Creation	: 23-10-1998
 *
 *  History:
 *  23-10-1998	: teresita : copied
 *
 *  Notes	:
 *  The Options string contains the format of the input, output and the amount
 *  to be added.
 *
 *  Supported options:
 *
 *	input ( "DATE", "DATETIME", "TIMESTAMP" )
 *
 *------------------------------------------------------------------------
 */
void 
PM_time_MakeCTime(const char *in, time_t *tt, const char *Options)
{
    struct tm tim;
    char tmp[20];
    char *ip = (char *) in;
    char *cp;

    (void) memset(&tim, '\0', sizeof(struct tm));
    tmp[0] = '\0';
    PM_opt_fetch(Options, "input", tmp, sizeof(tmp));
    if (strcmp(tmp, "DATE") == 0 || strcmp(tmp, "DATETIME") == 0 || strcmp(tmp, "TIMESTAMP") == 0) {
	if ((cp = strchr(ip, '-')) != NULL) {
	    tim.tm_year = atoi(ip) - 1900;
	    ip = cp + 1;
	}
	if ((cp = strchr(ip, '-')) != NULL) {
	    tim.tm_mon = atoi(ip) - 1;
	    ip = cp + 1;
	}
	if (cp != NULL) {
	    tim.tm_mday = atoi(ip);
	}
    }
    if (strcmp(tmp, "DATETIME") == 0 || strcmp(tmp, "TIMESTAMP") == 0) {
	if ((ip = strchr(ip, ':')) != NULL) {
	    ip += 1;
	    if ((cp = strchr(ip, ':')) != NULL) {
	        tim.tm_hour = atoi(ip);
	        ip = cp + 1;
	    }
	    if (cp != NULL) {
	        tim.tm_min = atoi(ip);
	    }
        }
    }
    if (strcmp(tmp, "TIMESTAMP") == 0) {
	if ((ip = strchr(ip, ':')) != NULL) {
	    ip += 1;
	    if ((cp = strchr(ip, '.')) != NULL) {
	        tim.tm_sec = atoi(ip);
	        ip = cp + 1;
	    }
        }
    }
    *tt = mktime(&tim);
} /* PM_time_MakeCTime */

/*------------------------------------------------------------------------
 *  Function	: PM_time_CalculateDateTime
 *  Purpose	: Add amount to time.
 *  Method	: string parsing and <time.h> routines.
 *  Author(s)	: Teresita Frizzarin
 *  Creation	: 23-10-1998
 *
 *  History:
 *  23-10-1998	: teresita : creation
 *
 *  Notes	:
 *  The Options string contains the format of the input, output and the amount
 *  to be added.
 *
 *  Supported options:
 *
 *	input  ( "DATE", "DATETIME", "TIMESTAMP" )
 *	output ( "DATE", "DATETIME", "TIMESTAMP" )
 *	amount ( "DAYS" , "HOURS", "MINUTES" )
 *
 *  If no "amount" is given, amount will be taken in seconds.
 *------------------------------------------------------------------------
 */
void 
PM_time_CalculateDateTime(const char *in, char *out, size_t outmax, int amount, const char *Options)
{
    char tmp[20];
    time_t tt;

    PM_time_MakeCTime(in, &tt, Options);
    PM_opt_fetch(Options, "amount", tmp, sizeof(tmp));
    if (strcmp(tmp, "DAYS") == 0) {
	tt += ((time_t) amount) * 60 * 60 * 24;
    } else if (strcmp(tmp, "HOURS") == 0) {
	tt += ((time_t) amount) * 60 * 60;
    } else if (strcmp(tmp, "MINUTES") == 0) {
	tt += ((time_t) amount) * 60;
    } else {
	tt += (time_t) amount;
    }
    tmp[0] = '\0';
    PM_opt_fetch(Options, "output", tmp, sizeof(tmp));
    if (strcmp(tmp, "DATE") == 0) {
	(void) strftime(out, outmax, "%Y-%m-%d", localtime(&tt));
    } else if (strcmp(tmp, "DATETIME") == 0) {
	(void) strftime(out, outmax, "%Y-%m-%d:%H:%M", localtime(&tt));
    } else if (strcmp(tmp, "TIMESTAMP") == 0) {
	(void) strftime(out, outmax, "%Y-%m-%d:%H:%M:%S.000000", localtime(&tt));
    } else {
	*out = '\0';
    }
} /* PM_time_CalculateDateTime */

/*------------------------------------------------------------------------
 *  Function	: PM_time_DifferDateTime
 *  Purpose	: Differ time.
 *  Method	: string parsing and <time.h> routines.
 *  Author(s)	: Martin Talmon
 *  Creation	: 01-03-1999
 *
 *  History:
 *  01-03-1999	: martin  : creation
 *
 *  Notes	:
 *  The Options string contains the format of the input, output and the amount
 *  to be added.
 *
 *  Supported options:
 *
 *	input  ( "DATE", "DATETIME", "TIMESTAMP" )
 *	amount ( "DAYS" , "HOURS", "MINUTES" )
 *
 *  If no "amount" is given, amount will be taken in seconds.
 *------------------------------------------------------------------------
 */
bool 
PM_time_DifferDateTime(const char *date1, const char *date2, double  *amount, const char *Options)
{
    char tmp[20];
    time_t tt1;
    time_t tt2;
    double difference;

    PM_time_MakeCTime(date1, &tt1, Options);
    if (tt1 == -1) return FALSE;
    PM_time_MakeCTime(date2, &tt2, Options);
    if (tt2 == -1) return FALSE;
    difference = difftime(tt2, tt1);
    PM_opt_fetch(Options, "amount", tmp, sizeof(tmp));
    if (strcmp(tmp, "DAYS") == 0) {
	*amount = difference/(60L*60L*24L);
    } else if (strcmp(tmp, "HOURS") == 0) {
	*amount = difference/(60L*60L);
    } else if (strcmp(tmp, "MINUTES") == 0) {
	*amount = difference/(60L);
    } else {
	*amount = difference;
    }
    return TRUE;
} /* PM_time_DifferDateTime */

/*------------------------------------------------------------------------
 *  Function	: PM_LBSStamp
 *  Purpose	: Get a timestamp
 *  Method	:
 *  Author(s)	: Teresita Frizzarin
 *  Creation	: 24-03-2000
 *
 *  History:
 *  24-03-2000	: teresita  : moved from hmrsysdep.c
 *------------------------------------------------------------------------
 */
void
PM_LBSStamp(char *stamp)
{
    struct timeval tv;
    struct tm *tim;

    gettimeofday(&tv, NULL);
    tim = localtime(&tv.tv_sec);
    /* I've split the format string, because sccs was interpreting it as sccs
       keywords. */
    strftime(stamp, 16 + 1, "%Y""%m""%d""%H""%M""%S", tim);
    sprintf(stamp + 14, "%02d", (int)(tv.tv_usec / 10000));
} /* PM_LBSStamp */

/*------------------------------------------------------------------------
 *  Function	: PM_time_local_to_UTC
 *  Purpose	: Convert local time to UTC (ISO 8601).
 *  Method	:
 *
 *  History:
 *  11-10-2010	: martinz  : creation
 *  13-10-2010	: paul     : Fixed check on buffer overflow
 *  27-10-2010	: martinz  : Added support for 2nd input format.
 *
 *  Notes	:
 *  Supported input formats:
 *
 *	YYYY-MM-DD hh:mm:ss.s
 *	YYYY-MM-DD hh.mm.ss.s
 *
 *  where the trailing ".s" is a decimal fraction of a second (optional) which
 *  should be zero, two or more than two bytes long. All characters (except
 *  delimiters "-", " ", ":" and ".") should be digits.
 *
 *  Output format (ISO 8601):
 *
 *	YYYY-MM-DDThh:mm:ss.sZ
 *
 *  where "YYYY", "MM", "DD", "hh", "mm", "ss", "s" are digits and all other
 *  characters are literals. The decimal fraction ".s" will only be used if it
 *  exists in the input and will be copied 'as is'.
 *
 *  Returns	: Indication of success.
 *------------------------------------------------------------------------
 */
bool
PM_time_local_to_UTC(
	const char *local_time,	/* local time in YYYY-MM-DD hh:mm:ss.s */
	const char *options,	/* "ermess_key" = key to mention in ERWrite */
	char *buf,		/* output buffer */
	size_t bufsize)		/* output buffer size in bytes */
{
    const char fname[] = "PM_time_local_to_UTC";

    struct tm t;		/* parsed local time */
    time_t t2;			/* local time internal format */
    struct tm *t3;		/* UTC time internal format */
    int len = 0;		/* string length */
    const char *frac = NULL;	/* fraction of a second including dot '.' */
    const char *c;		/* pointer to character of fraction */
    char ermess_key[100];

    PM_opt_fetch(options, "ermess_key", ermess_key, sizeof(ermess_key));

    if (buf == NULL) {
	ERWrite(APPLE_PARAM, fname, "%s: buf is NULL.", ermess_key);
	return FALSE;
    }

/*
 * Check if localtime has 19 or more than 20 characters.
 */
    if (local_time == NULL
	|| (len = strlen(local_time)) < 19
	|| len == 20) {
	ERWrite(APPLE_PARAM, fname,
		local_time ? "%s: localtime has invalid length: %d."
		: "%s: localtime is NULL.", ermess_key, len);
	return FALSE;
    }

/*
 * Check first 19 characters.
 */

    if (!isdigit(local_time[0])
	|| !isdigit(local_time[1])
	|| !isdigit(local_time[2])
	|| !isdigit(local_time[3])
	|| local_time[4] != '-'
	|| !isdigit(local_time[5])
	|| !isdigit(local_time[6])
	|| local_time[7] != '-'
	|| !isdigit(local_time[8])
	|| !isdigit(local_time[9])
	|| local_time[10] != ' '
	|| !isdigit(local_time[11])
	|| !isdigit(local_time[12])
	|| (local_time[13] != ':' && local_time[13] != '.')
	|| !isdigit(local_time[14])
	|| !isdigit(local_time[15])
	|| local_time[16] != local_time[13]
	|| !isdigit(local_time[17])
	|| !isdigit(local_time[18])) {
	ERWrite(APPLE_PARAM, fname, "%s: Invalid localtime: \"%s\"",
		ermess_key, local_time);
	return FALSE;
    }

    if (len > 19) {

/*
 * Format is YYYY-MM-DD hh:mm:ss.s
 * Check all fraction characters.
 */
	frac = &local_time[19];
	if (*frac != '.') {
	    ERWrite(APPLE_PARAM, fname, "%s: Invalid localtime: \"%s\"",
		    ermess_key, local_time);
	    return FALSE;
	}
	c = frac + 1;
	while (*c != '\0') {
	    if (!isdigit(*c)) {
	        ERWrite(APPLE_PARAM, fname, "%s: Invalid localtime: \"%s\"",
			ermess_key, local_time);
		return FALSE;
	    }
	    c++;
	}
    }

/*
 * Convert substrings to integers.
 */
    t.tm_year = atoi(&local_time[0]) - 1900;
    t.tm_mon = atoi(&local_time[5]) - 1;
    t.tm_mday = atoi(&local_time[8]);
    t.tm_hour = atoi(&local_time[11]);
    t.tm_min = atoi(&local_time[14]);
    t.tm_sec = atoi(&local_time[17]);
    t.tm_isdst = -1;

/*
 * Convert to UTC.
 */
    if ((t2 = mktime(&t)) == -1) {
	ERWrite(APPLE_INTERNAL, fname, "%s: Failed mktime().", ermess_key);
	return FALSE;
    }
    if ((t3 = gmtime(&t2)) == NULL) {
	ERWrite(APPLE_INTERNAL, fname, "%s: Failed gmtime().", ermess_key);
	return FALSE;
    }

/*
 * Convert to string.
 * strftime() will fail on too small buf.
 */
    if ((len = strftime(buf, bufsize, "%Y-%m-%dT%H:%M:%S", t3)) == 0) {
	ERWrite(APPLE_PARAM, fname, "%s: Failed strftime(); possibly bufsize" \
		" too small: %lu.", ermess_key, (unsigned long)bufsize);
	return FALSE;
    }

/*
 * Check whether buf is large enough for fraction and last character.
 */
    if (frac != NULL) {
	if ((long)bufsize - len - 2 < (long)strlen(frac)) {
	    ERWrite(APPLE_PARAM, fname, "%s: bufsize is too small: %lu.",
		    ermess_key, (unsigned long)bufsize);
	    return FALSE;
	}
        strcat(buf, frac);
    } else {
	if ((long)bufsize - len - 2 < 0) {
	    ERWrite(APPLE_PARAM, fname, "%s: bufsize is too small: %lu.",
		    ermess_key, (unsigned long)bufsize);
	    return FALSE;
	}
    }
    strcat(buf, "Z");
    return TRUE;
} /* PM_time_local_to_UTC() */
