/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: pmtoasc.c
 *  Description	: Miscellaneous routines for conversion to ASCII.
 *  Author(s)	: Hans van Eck
 *  Creation	: 10-01-1994
 *
 *  History:
 *  10-01-1994	: eck      : creation
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *  Notes	:
 *  These routines convert int's, long's, unsigned's and unsigned long's to
 *  ASCII-strings. The buffers to write into should be at least
 *  PM_INTLEN_MAX-bytes large. When NULL-pointers are passed, these routines
 *  will use and return (a pointer to) an internal buffer, otherwise, buffer is
 *  returned.
 *
 *  The internal buffer will be reused at every call, so it should be copied
 *  before it can be used safely.
 *========================================================================
 */

#pragma ident "$Id: pmtoasc.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <picamisc.h>

static char ret_buffer[PM_INTLEN_MAX];

/*------------------------------------------------------------------------
 *  Function	: PM_utoa
 *  Purpose	: Convert an unsigned int to an ASCII-string
 *  Method	: determine length first, then fill from end to start
 *
 *  History:
 *  06-01-1994	: eck      : creation
 *  10-01-1994	: eck      : moved from pmgensym.c to pmtoasc.c
 *  21-03-1994	: eck      : use own buffer if target == NULL && return buffer
 *  13-01-1995	: eck      : cleaned up a bit
 *
 *  Notes	: Does not check anything
 *
 *  Returns	: pointer to buffer
 *------------------------------------------------------------------------
 */
const char *
PM_utoa(
	char		*string,	/* target string */
	unsigned	value)		/* value to convert */
{
    int		idx = 0;
    unsigned	tmp;

    if (string == NULL) string = ret_buffer;

    tmp = value;
    while (tmp >= 10) {
	tmp /= 10;
	idx++;
    }

    string[idx + 1] = '\0';
    while (value >= 10) {
	string[idx--] = (value % 10) + '0';
	value /= 10;
    }
    *string = value + '0';

    return string;
} /* PM_utoa() */

/*------------------------------------------------------------------------
 *  Function	: PM_itoa
 *  Purpose	: Convert an integer to an ASCII-string
 *  Method	: Generate sign if necessary, then call PM_utoa().
 *
 *  History:
 *  10-01-1994	: eck      : creation
 *  21-03-1994	: eck      : use own buffer if target == NULL && return buffer
 *  13-01-1995	: eck      : cleaned up a bit
 *
 *  Notes	: Does not check anything
 *
 *  Returns	: pointer to buffer
 *------------------------------------------------------------------------
 */
const char *
PM_itoa(
	char	*string,	/* target string */
	int	value)		/* value to convert */
{
    if (string == NULL) string = ret_buffer;

    if (value < 0) {
	*string = '-';
	(void) PM_utoa(string + 1, (unsigned) -value);
    } else {
	(void) PM_utoa(string, (unsigned) value);
    }

    return string;
} /* PM_itoa() */

/*------------------------------------------------------------------------
 *  Function	: PM_ultoa
 *  Purpose	: Convert an unsigned long to an ASCII-string
 *  Method	: determine length first, then fill from end to start
 *
 *  History:
 *  13-01-1995	: eck      : Copied from PM_utoa()
 *
 *  Notes	: Does not check anything
 *
 *  Returns	: pointer to buffer
 *------------------------------------------------------------------------
 */
const char *
PM_ultoa(
	char		*string,	/* target string */
	unsigned long	value)		/* value to convert */
{
    int			idx = 0;
    unsigned long	tmp;

    if (string == NULL) string = ret_buffer;

    tmp = value;
    while (tmp >= 10) {
	tmp /= 10;
	idx++;
    }

    string[idx + 1] = '\0';
    while (value >= 10) {
	string[idx--] = (value % 10) + '0';
	value /= 10;
    }
    *string = value + '0';

    return string;
} /* PM_ultoa() */

/*------------------------------------------------------------------------
 *  Function	: PM_ltoa
 *  Purpose	: Convert a long to an ASCII-string
 *  Method	: Generate sign if necessary, then call PM_ultoa().
 *
 *  History:
 *  13-01-1995	: eck      : Copied from PM_itoa().
 *
 *  Notes	: Does not check anything
 *
 *  Returns	: pointer to buffer
 *------------------------------------------------------------------------
 */
const char *
PM_ltoa(
	char	*string,	/* target string */
	long	value)		/* value to convert */
{
    if (string == NULL) string = ret_buffer;

    if (value < 0) {
	*string = '-';
	(void) PM_ultoa(string + 1, (unsigned long) -value);
    } else {
	(void) PM_ultoa(string, (unsigned long) value);
    }

    return string;
} /* PM_ltoa() */

/*------------------------------------------------------------------------
 *  Function	: PM_sizetoa
 *  Purpose	: Convert a value to a size string (when possible)
 *  Method	: Duh
 *
 *  History:
 *  23-03-2011	: eck      : creation
 *
 *  Notes	:
 *  Assumes the target string is long enough for an extra character.
 *
 *  Returns	: pointer to buffer
 *------------------------------------------------------------------------
 */
const char *
PM_sizetoa(
	char	*string,
	long	value)
{
    char	suffix = '\0';
    const char	*ret;
    char	*end;

    if (value > 0) {
	if (value % 1024 == 0) {
	    value /= 1024;
	    suffix = 'K';
	}
	if (value % 1024 == 0) {
	    value /= 1024;
	    suffix = 'M';
	}
	if (value % 1024 == 0) {
	    value /= 1024;
	    suffix = 'G';
	}
    }
    ret = PM_ltoa(string, value);
    if (suffix != '\0') {
	end = (char *) ret + strlen(ret);
	*end++ = suffix;
	*end = '\0';
    }
    return ret;
} /* PM_sizetoa() */
