/*========================================================================
 *  Copyright (c) 2005 by Pica
 *
 *  File	: pmtoint.c
 *  Description	: Miscellaneous routines for conversion from ASCII to all kind
 *                of integers.
 *  Author(s)	: Paul Smit
 *  Creation	: 23-09-2005
 *
 *  History:
 *  30-05-2006	: paul     : PM_roman2l() moved from pmsortstring.c
 *                         : PM_month2l() moved from pmsortstring.c
 *
 *========================================================================
 */

#pragma ident "$Id: pmtoint.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <picamisc.h>
#include <pica.h>

/*------------------------------------------------------------------------
 *  Function	: PM_atosize
 *  Purpose	: Convert string with size suffix to number
 *  Method	:
 *
 *  History:
 *  16-09-2004	: eck      : creation
 *
 *  Notes	:
 *  - Originally copied from HMIMemsizeGet() (Hermes indexing)
 *  - Handles strings like "128K", "500M", etc. Fractions (like "0.5G") are not
 *    supported.
 *  - On overflow, LONG_MAX is returned.
 *
 *  Returns	: The size or -1 on error
 *------------------------------------------------------------------------
 */
long
PM_atosize(
	const char	*s)
{
    long	result;
    char	*t;
    long	multiplier = 1;

    errno = 0;
    result = strtol(s, &t, 10);
    if (errno != 0) return -1;		/* errno == EINVAL or ERANGE */
    if (result < 0) return -1;		/* Can't be right */

    if (t != NULL && *t) {
	switch(toupper(*t)) {
	case 'G': multiplier *= 1024;	/* FALLTHROUGH */
	case 'M': multiplier *= 1024;	/* FALLTHROUGH */
	case 'K': multiplier *= 1024;	/* FALLTHROUGH */
	}
    }
    if ((LONG_MAX / multiplier) < result) return LONG_MAX;
    return result * multiplier;
}  /* PM_atosize() */

/*------------------------------------------------------------------------
 *  Function	: PM_roman2l
 *  Purpose	: Convert Roman numerals to integer
 *  Author(s)	: Eric Helsper
 *  Creation	: V-I-MCMXCVIII
 *
 *  History:
 *  05-01-1998	: eric     : creation
 *  26-03-2004	: paul     : "mrt" is not the Roman number 1000,
 *                           nor is "dec" the Roman number 500
 *                         : Added check that if a letter has been used to add
 *                           to the total, it may not also be used to
 *                           substract from it
 *  05-04-2004	: paul     : Fixed bug: dereferencing NULL pointer
 *  30-05-2006	: paul     : renamed from Roman() and moved from pmsortstring.c
 *                         : added end parameter
 *
 *  Notes	:
 *  "dici" is no longer recognized as the roman number 600. The reason is that
 *  the "i" is used for both addition and substraction. We can reject this
 *  notation without risking to reject too much, since no one in his right mind
 *  would write "dici" if "dc" denotes the same number. It is important to
 *  reject "dici" because this string is recognized by PM_month2l() as an
 *  indication of the month december.
 *
 *  Returns	: the converted number, 0 if failed
 *------------------------------------------------------------------------
 */
long
PM_roman2l(
	const char	*s,
	const char	**end)
{
    /* usage remembers how roman digits have been used:
     * 0 indicates: not yet used
     * 1 indicates: used for addition
     * -1 indicates: used for substraction
     */
    struct {
	int	I;
	int	V;
	int	X;
	int	L;
	int	C;
	int	D;
	int	M;
    }			usage = { 0, 0, 0, 0, 0, 0, 0 };
    int			*usagep = NULL, *lastusagep;
    long i = 0;
    long v;
    long lastv = 10000;

    for (  ; ; s++) {
	lastusagep = usagep;
	switch (toupper(*s)) {
	case 'I':   v = 1;      usagep = &usage.I; break;
	case 'V':   v = 5;      usagep = &usage.V; break;
	case 'X':   v = 10;     usagep = &usage.X; break;
	case 'L':   v = 50;     usagep = &usage.L; break;
	case 'C':   v = 100;    usagep = &usage.C; break;
	case 'D':   v = 500;    usagep = &usage.D; break;
	case 'M':   v = 1000;   usagep = &usage.M; break;
	default:
	    /* last letter may have been used before, but for addition only */
	    if (lastusagep == NULL) return 0; 
	    if (*lastusagep == -1) return 0; 
	    if (isalnum(*(byte *)s)) return 0;
	    if (end != NULL) {
		*end = s;
	    }
	    return(i);
	}
	i += v;
	if (lastv < v) {
	    if (lastusagep != NULL) {
		/* this letter may not have been used before */
		if (*lastusagep != 0) return 0;
		/* this letter is used to substract: */
		*lastusagep = -1;
	    }
	    i -= lastv;   /* if last v smaller, we had to subtract */
	    i -= lastv;   /* instead of add */
	} else if (lastusagep != NULL) {
	    /* this letter may have been used before, but for addition only */
	    if (*lastusagep == -1) return 0; 
	    *lastusagep = 1;
	}
	lastv = v;
    }
} /* PM_roman2l */

/*------------------------------------------------------------------------
 *  Function	: PM_month2l
 *  Purpose	: Convert Month / Season name to integer
 *  Author(s)	: Eric Helsper
 *  Creation	: 05-01-1998
 *
 *  History:
 *  05-01-1998	: eric - creation
 *  24-07-1998	: eric - can handle ranges, will only look at first
 *  30-05-2006	: paul     : renamed from Month() and moved from pmsortstring.c
 *                         : month must match start of a known month-string
 *                         : added end parameter
 *
 *  Notes	:
 *  Uses hard-coded table. Names of months are not likely to change often.
 *
 *  Returns	: the month number, 0 if failed
 *------------------------------------------------------------------------
 */
long
PM_month2l(
	const char	*cs,
	const char	**end)
{
    static const char *month[] = {
/* jan */  " janu janv genn ener jane",
/* feb */  " febr fevr febb feve",
/* mar */  " maar mars maer marz marc mrt",
/* apr */  " apri avri abri",
/* may */  " mei may mai magg mayo maio",
/* jun */  " juni june juin giug junh",
/* jul */  " juli july juil lugl julh",
/* aug */  " augu aout agos",
/* sep */  " sept sett sete",
/* oct */  " okto octo otto octu otub",
/* nov */  " nove novi",
/* dec */  " dece deze dice dici"
       };

    static const char *season[] = {
/* mar */  " lent voor spri frue prin prim",
/* jun */  " zome summ somm ete esta vera",
/* sep */  " herf naja autu fall herb auto autu oton outo",
/* dec */  " wint hive inve invi"
       };

    char	s[6];
    int		slen;
    int16	i;

    if (*cs == '\0' || !isalpha(*(byte *)cs)) {
	return 0;
    }
    s[0] = ' ';
    for (i = 0; i < 4; i++) {
	if ((cs[i] == '\0') || !isalpha(((byte *)cs)[i])) break;
	s[i+1] = tolower((unsigned char)cs[i]);
    }
    slen = i + 1;
    s[slen] = '\0';
    for (i = 0; i < 12; i++) {
	if (strstr(month[i], s) != NULL) {
	    if (end != NULL) {
		*end = cs + slen - 1;
	    }
	    return(i + 1);
	}
    }
    for (i = 0; i < 4; i++) {
	if (strstr(season[i], s) != NULL) {
	    if (end != NULL) {
		*end = cs + slen - 1;
	    }
	    return((i + 1) * 3);
	}
    }

    return(0);
} /* PM_month2l */
