/*========================================================================
 *  Copyright (c) 1997 by Pica
 *
 *  File	: pmuserbits.c
 *  Description	: Conversion between userbits and readable iln ranges
 *  Author(s)	: Eric Helsper
 *  Creation	: 29-04-1999
 *
 *  History:
 *  29-04-1999	: Eric : creation
 *  10-07-2006	: henk     :  added PM_userbit_previous and PM_userbit_next
 *  05-08-2009	: sutherlr : PM_UsedByIlnsBinToAsc() algorithm replaced with
 *                           call to PM_range_from_bits().
 *  12-09-2013	: martinz  : Added documentation from PMLIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: pmuserbits.c 364831 2019-02-11 17:26:10Z vaneckh $"

#include <string.h>
#include <limits.h>

#include <pica.h>
#include <picamisc.h>

/*-----------------------------------------------------------------------------
 * Function    : PM_UsedByIlnsAscToBin
 * Purpose     : Convert used-by-ilns (ranges) to binary (userbits) form
 * Author(s)   : Eric Helsper
 * Creation    : 29-04-1999
 *
 * History:
 * 29-04-1999  : eric - creation (heavily modified from fiottlens.c)
 *
 * Notes       :
 * The required binsize is (the highest iln in asc) / 8.
 * The function can handle very high "ilns" (but we consider 400 to be the
 * highest possible iln).
 *
 * Returns     :
 * TRUE: Successful conversion. This includes the case when there are no ilns.
 * The *binused is then 0.
 *
 * FALSE: An iln required an index into the bin string which was beyond its
 * length.
 * Or: scanning of asc stopped because of a ranges "syntax" error. The string
 * should be correct from begin to end. *binused is set to the number of bytes
 * used after the last correct iln (range).
 *
 * The function will return unsuccessful if binsize is too small to represent
 * all ilns in the asc string.
 *-----------------------------------------------------------------------------
 */
bool
PM_UsedByIlnsAscToBin(
	const char   *asc,	/* The input string in ranges form */
	char         *bin,	/* The output array for the bits */
	const size_t binsize,	/* The maximum number of bytes to be used in
				   bin */
	int16        *binused)	/* Will be set to the actual number of bytes
				   used (possibly 0) */
{
    int16         binidx;
    const char    *tmp;
    const char    *last;
    unsigned long lowiln;
    unsigned long highiln;

    memset(bin, '\0', binsize);
    *binused = 0;
    if (asc == NULL) return(FALSE);
    last = asc;

    tmp = PM_range_get_bounds(asc, &lowiln, &highiln);

    while (tmp != NULL) {
	unsigned long iln1; /* = iln - 1, because iln 1 sets bit 0, etc. */

	last = tmp;
	if (lowiln == 0) break;
	for (iln1 = lowiln - 1; iln1 < highiln; iln1++) {
	    binidx = iln1 / 8;
	    if (binidx >= binsize) return(FALSE);
	    bin[binidx] |= (0x80 >> (iln1 % 8));

	    /* Ranges are not necessarily in ascending order, so any range may
	     * be the highest. So we have to check *binused each time.
	     */
	    if (*binused <= binidx) *binused = binidx + 1;
	}
	tmp = PM_range_get_bounds(tmp, &lowiln, &highiln);
    }
    /* we require that the asc string is processed up to the end. In that
     * case *last points to the final zero byte.
     */
    return(*last == '\0');
} /* PM_UsedByIlnsAscToBin() */

/*-----------------------------------------------------------------------------
 * Function    : PM_UsedByIlnsBinToAsc
 * Purpose     : Convert userbits to used-by-ilns (ranges)
 * Author(s)   : Eric Helsper
 * Creation    : 29-04-1999
 *
 * History:
 * 29-04-1999  : eric - creation (heavily modified from fiodbtitle.c)
 * 31-07-2009  : sutherlr : replaced algorithm with call to 
 *                          PM_range_from_bits().
 *
 * Notes       :
 * ascsize needed depends on the number of ilns in bin (but should at least be 8).
 * A typical value is sizeof(FIOTitle_t.used_by_ilns.val).
 * The longest asc representation arises when relatively few ranges can be
 * formed, like 1,3,5,7,9,...
 * The maximum possible required ascsize is PIC_ILN_RANGEBUF_SIZE.
 *
 *               The asc string returned is always cleaned and zero-ended.
 *
 * Description	:
 * Converts used-by-ilns from binary (userbits) form to readable (ranges)
 * form. The reverse of PM_UsedByIlnsAscToBin(). The function will return
 * unsuccessful if ascsize is to small to represent all ilns in the bin
 * string.
 *
 * Returns     :
 * TRUE: Successful conversion. This includes the case when there are no ilns.
 * The asc string will then be empty.
 *
 * FALSE: There were too many ilns to fit within ascsize.
 * The asc string will then contain all iln ranges that fit.
 *-----------------------------------------------------------------------------
 */
bool
PM_UsedByIlnsBinToAsc(
	const char   *bin,	/* The input array for the bits */
	const int16  binlength,	/* The size of the input array */
	char         *asc,	/* The output string for the ranges */
	const size_t ascsize)	/* The maximum number of bytes to be used in
				   asc, including the final zero byte */
{
    int ret = PM_range_from_bits(asc, ascsize, (byte*)bin, 
				 binlength, PIC_ILN_MIN);

    if (ret == 0) {
	 return asc[0] == '\0';
    }
    return ret != PM_RANGE_ERROR;
} /* PM_UsedByIlnsBinToAsc() */

/*-----------------------------------------------------------------------------
 *  Function    : PM_bit_count
 *  Purpose     : Count the number of set bits in the supplied bit string
 *  Method      :
 *
 *  History:
 *  08-12-2000  : craig    : creation
 *
 *  Returns     : The number of set bits is returned
 *-----------------------------------------------------------------------------
 */
uint32
PM_bit_count(
    const byte *bit_buffer,         /* ptr to array containing bit string */
    size_t      byte_length)        /* length of bit string (in bytes) */

{
    /*****
     * Local variables
     ****/

    uint32 count;                   /* number of set bits */
    byte   data;                    /* byte from input bit string */
    int    i;                       /* loop index */

    /*****
     * Loop to count the number of bits set
     ****/

    for (i = 0, count = 0; i < byte_length; ++ i) {
	data = bit_buffer[i];
	while (data != 0) {
	    ++ count;
	    data &= (data - 1);
	}
    }

    /*****
     * Return the count to the caller
     ****/

    return (count);
} /* PM_bit_count() */

/*-----------------------------------------------------------------------------
 *  Function    : PM_userbit_test
 *  Purpose     : Test if specified user bit (from 001@ $0) is set.
 *  Method      :
 *
 *  History:
 *  08-12-2000  : craig    : creation
 *
 *  Notes       :
 *  Numbering of the user bits starts with 1.
 *
 *  The user bits are numbered as follows:
 *
 *                               1111111
 *                     12345678 90123456
 *                    +--------+--------+
 *                    | byte 0 | byte 1 | ...
 *                    +--------+--------+
 *
 *  Returns     :
 *  TRUE if the specified bit is set.
 *  FALSE if the specified bit is not set (or out of range)
 *-----------------------------------------------------------------------------
 */
bool
PM_userbit_test(
    const byte *bit_buffer,         /* ptr to array containing bit string */
    size_t      byte_length,        /* length of bit string (in bytes) */
    uint32      bit_num)            /* number of bit to be tested */

{
    /*****
     * Local variables
     ****/

    byte   mask;                    /* mask for testing the bit */
    byte   data;                    /* byte from input bit string */

    /*****
     * Make sure bit number is not out of range
     ****/

    if ((bit_num == 0) || (bit_num > (byte_length * CHAR_BIT))) {
	return (FALSE);
    } else {
	-- bit_num;
    }

    /*****
     * Test the bit and return the results
     ****/

    mask = (byte) (0x01 << (7 - (int) (bit_num % CHAR_BIT)));
    data = bit_buffer[(bit_num / CHAR_BIT)];
    return ((data & mask) != 0);
} /* PM_userbit_test() */

/*------------------------------------------------------------------------
 *  Function	: PM_userbit_next
 *  Purpose	: Get index of first set userbit after reference iln
 *  Method	:
 *
 *  History:
 *  03-07-2006	: henk     : creation
 *  26-08-2008  : henk     : fixed bug in precedence of operators
 *
 *  Notes	:
 *
 *  Returns	: Index of first set userbit after reference iln else 0.
 *------------------------------------------------------------------------
 */
uint32 
PM_userbit_next(
    const byte *bit_buffer,         /* ptr to array containing bit string */
    size_t      byte_length,        /* length of bit string (in bytes) */
    uint32      iln)                /* reference iln */
{
    byte 	mask, elt;
    uint32	idx, bit_idx;
    int		n;

    if (iln == 0) {
	idx = 0;
	if (byte_length == 0) {
	    return 0;
	}
	elt = bit_buffer[idx];
    } else {
	idx = (iln - 1) / CHAR_BIT;
	bit_idx = (iln - 1) % CHAR_BIT;
	if (byte_length < idx + 1) {
	    return 0;
	}
	elt = bit_buffer[idx];
	if (bit_buffer[idx] != 0) {
	    if (bit_idx == CHAR_BIT - 1) {
		mask = 0;
	    } else {
		mask = ~(~0U << (CHAR_BIT - bit_idx - 1));
	    }
	    elt &= mask;
	}
    }
    if (elt == 0) {
	idx++;
	while (idx < byte_length) {
	    if (bit_buffer[idx] != 0) {
		break;
	    }
	    idx++;
	}
	if (idx >= byte_length) {
	    return 0;
	}
	elt = bit_buffer[idx];
    }

    /* number of leading zeroes */
    n = 1;
    if ((elt & 0xf0) == 0) { n += 4; elt = elt << 4; }
    if ((elt & 0xc0) == 0) { n += 2; elt = elt << 2; }
    if ((elt & 0x80) == 0) { n += 1; }
    return idx * CHAR_BIT + n ;
} /* PM_userbit_next() */

/*------------------------------------------------------------------------
 *  Function	: PM_userbit_previous
 *  Purpose	: Get index of last set userbit before reference iln
 *  Method	:
 *
 *  History:
 *  03-07-2006	: henk     : creation
 *
 *  Returns	:
 *  Index of last set userbit before reference iln.
 *  If iln = 0, gives the last bit that is set.
 *  If no bits found, returns 0.
 *------------------------------------------------------------------------
 */
uint32 
PM_userbit_previous(
    const byte *bit_buffer,         /* ptr to array containing bit string */
    size_t      byte_length,        /* length of bit string (in bytes) */
    uint32      iln)                /* reference iln */
{
    byte 	mask, elt;
    uint32	idx, bit_idx;
    int		n;

    if (iln == 0) { 
	if (byte_length == 0) {
	    return 0;
	}
	idx = byte_length - 1;
	elt = bit_buffer[idx];
    } else {
	idx = (iln - 1) / CHAR_BIT;
	bit_idx = (iln - 1) % CHAR_BIT;
	if (byte_length < idx + 1) {
	    idx = byte_length - 1;
	    elt = bit_buffer[idx];
	} else {
	    elt = bit_buffer[idx];
	    if (bit_buffer[idx] != 0) {
		if (bit_idx == 0) {
		    mask = 0;
		} else {
		    mask = (byte) 0;
		    mask = ~ mask;
		    mask = mask >> bit_idx;
		    mask = ~ mask;
		}
		elt &= mask;
	    }
	}
    }
    if (elt == 0) {
	if (idx == 0) return 0;
	idx--;
	while (idx > 0) {
	    if (bit_buffer[idx] != 0) {
		break;
	    }
	    idx--;
	}
	elt = bit_buffer[idx];
    }
    if (elt == 0) return 0;
    /* number of trailing zeroes */

    n = 1;
    if ((elt & 0x0f) == 0) { n += 4; elt = elt >> 4; }
    if ((elt & 0x03) == 0) { n += 2; elt = elt >> 2; }
    return idx * CHAR_BIT + n - (elt & 0x01);
} /* PM_userbit_previous() */
