/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: nioall.c
 *  Description	: Sourcefile for NIO library functions
 *  Author(s)	: Pica maintenance
 *  Creation	: 16-08-1994
 *
 *  History:
 *  16-08-1994	: pica     : creation
 *  20-11-1996	: eck      : Added NIO_fildes() (returns file-descriptor).
 *			    : Increased MAXCONNECTION.
 *  20-11-1996  : PvdM     : Adjusted for Windows compatibility.
 *  26-02-1998	: BdB	   : Trivial additions to compile on Tandem OSS
 *  09-12-1998  : LB       : Trivial additions to compile on Tandem Guardian
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: nioall.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <string.h>
#include <errno.h>
#include <sys/times.h>
#include <time.h>

#include <pica.h>
#include <picamisc.h>
#include <er.h>
#include <nio.h>
#include "niolocal.h"

// \cond DOXYGEN_EXCLUDE

#define MAXCONNECTION 60

typedef enum {
    NIO_UNKNOWN = 0, NIO_TCPIP, NIO_DECNET, NIO_X25
} NIOProtocol_t;

typedef enum {
    NIO_NOMODE = 0, NIO_PASSIVE, NIO_ACTIVE
} NIOMode_t;

ERROR_DEFINE(NIOE_NOERROR);
ERROR_DEFINE(NIOE_MAXCONN);
ERROR_DEFINE(NIOE_OPENERR);
ERROR_DEFINE(NIOE_OPTION);
ERROR_DEFINE(NIOE_HANDLE);
ERROR_DEFINE(NIOE_SENDERR);
ERROR_DEFINE(NIOE_RECVERR);

NIOErrno_t	NIOerrno = NIOE_NOERROR;

static struct {
    NIOProtocol_t	protocol;
    NIOMode_t		mode;
    int			in_handle;
    int			out_handle;
    clock_t		time_send;
    clock_t		time_recv;
} NIOConnection[MAXCONNECTION];

static struct tms timbuf;

static NIOHandle_t
freehandle(void)
{
    NIOHandle_t		i;

    for (i = 0; i < MAXCONNECTION; i++) {
	if (NIOConnection[i].mode == NIO_NOMODE) return i;
    }
    NIOerrno = NIOE_MAXCONN;
    return -1;
}

static bool
valid_handle(
	NIOHandle_t	handle)		/* the handle to the connection */
{
    if ((handle < 0)
	|| (handle >= MAXCONNECTION)
	|| (NIOConnection[handle].mode == NIO_NOMODE)) {
	return  FALSE;
    }
    return TRUE;
}

static NIOMode_t
det_mode(const char *ptr)
{
    char	mode[32];

    ptr = PM_opt_findstart(ptr);
    while (*ptr) {
	if (PM_opt_equal(ptr, "mode")) {
	    ptr = PM_opt_findval(ptr);
	    (void) PM_opt_copyval(ptr, mode, sizeof(mode) - 1);
	    if (!strcmp(mode, "passive")) return NIO_PASSIVE;
	    if (!strcmp(mode, "active")) return NIO_ACTIVE;
	    NIOerrno = NIOE_OPTION;
	    return (NIOMode_t)-1;
	}
	ptr = PM_opt_findnext(ptr);
    }
    NIOerrno = NIOE_OPTION;
    return NIO_NOMODE;
}

static NIOProtocol_t
det_protocol(const char *ptr)
{
    char	protocol[32];

    ptr = PM_opt_findstart(ptr);
    while (*ptr) {
	if (PM_opt_equal(ptr, "protocol")) {
	    ptr = PM_opt_findval(ptr);
	    (void) PM_opt_copyval(ptr, protocol, sizeof(protocol) - 1);
	    if (!strcmp(protocol, "tcp/ip")) return NIO_TCPIP;
	    if (!strcmp(protocol, "decnet")) return NIO_DECNET;
	    if (!strcmp(protocol, "x25")) return NIO_X25;
    	    NIOerrno = NIOE_OPTION;
	    return (NIOProtocol_t)-1;
	}
	ptr = PM_opt_findnext(ptr);
    }
    NIOerrno = NIOE_OPTION;
    return (NIOProtocol_t)-1;
}

// \endcond

/*------------------------------------------------------------------------
 *  Function	: NIOOpen
 *  Purpose	: Open a connection, given protocol and address
 *  Method	: Determine mode and protocol and call appropriate function
 *
 *  History:
 *  21-02-2014	: eck      : creation (of comment)
 *
 *  Notes	:
 *  This function opens a connection to the address, specified in NIOAddress.
 *  This address is of the form "mode=mmm;protocol=xxx[;...]". In this string,
 *  the mode is either 'active' or 'passive'. Only one passive connection may
 *  be opened: it will be associated to standard input and standard output. The
 *  protocol does not need to be specified.
 *
 *  More than one active connection may be opened. In this case protocol must
 *  be specified. Only the protocol "tcp/ip" is supported. The addressing
 *  information is of the form "address=yyy;port=zzz".
 *
 *  When the connection could not be set up, NIO_NOHANDLE is returned. The
 *  variable NIOerrno contains the error. When successful, the function returns
 *  a handle to the connection.
 *
 *  Example     :
 *
 *	NIOHandle_t	nioh = NIOOpen("address=cbs.pica.nl;port=1040");
 *
 *  Returns	: A valid NIOHandle_t or NIO_NOHANDLE.
 *
 *  Error values:
 *
 *	NIOE_NOERROR	(ok)
 *	NIOE_OPTION	(error in option)
 *	NIOE_IO		(io error)
 *
 *------------------------------------------------------------------------
 */
NIOHandle_t
NIOOpen(
	const char	*NIOAddress)	/* in- or outgoing address */
{
    int			h_in, h_out;
    NIOHandle_t		handle;
    NIOMode_t		mode;

    handle = freehandle();
    if (handle == -1) return NIO_NOHANDLE;

    mode = det_mode(NIOAddress);
    if (mode == (NIOMode_t) -1) return NIO_NOHANDLE;

    if (mode == NIO_ACTIVE) {
	NIOConnection[handle].protocol = det_protocol(NIOAddress);
	switch(NIOConnection[handle].protocol) {
	case NIO_UNKNOWN:
	    return NIO_NOHANDLE;
	case NIO_TCPIP:
	    if (!NIOtcpip_open(NIOAddress, &h_in)) {
		NIOerrno = NIOE_OPENERR;
		ERWrite(NIOerrno, __func__, "%s", NIOAddress);
		return NIO_NOHANDLE;
	    }
	    h_out = h_in;
	    break;
	case NIO_DECNET:
	    return NIO_NOHANDLE;
	    /* break; */
	case NIO_X25:
	    return NIO_NOHANDLE;
	    /* break; */
	}
    } else {
	if (!NIOpassive_open(&h_in, &h_out)) return NIO_NOHANDLE;
    }
    NIOConnection[handle].in_handle = h_in;
    NIOConnection[handle].out_handle = h_out;
    NIOConnection[handle].mode = mode;
    NIOConnection[handle].time_send = 0;
    NIOConnection[handle].time_recv = 0;
    return handle;
} /* NIOOpen() */

/*------------------------------------------------------------------------
 *  Function	: NIOSend
 *  Purpose	: Send an application buffer to a connection
 *  Method	: Determine mode and protocol and call appropriate function.
 *
 *  History:
 *  21-02-2014	: eck      : creation (of comment)
 *
 *  Notes	:
 *  This function will send a databuffer to the connection. FALSE is returned
 *  when the function failed. NIOerrno will contain the error.
 *
 *  Example     :
 *
 *	if (!NIOSend(handle, buffer)) {
 *	    ERWrite(NIOerrno, __func__, "sending buffer");
 *	    return;
 *	}
 *
 *  Returns	:
 *
 *	TRUE	the buffer was written completely
 *	FALSE	the buffer not or partially written
 *
 *  Error values:
 *
 *	NIOE_NOERROR	(ok)
 *	NIOE_HANDLE	(invalid handle)
 *	NIOE_SENDERR	(io error)
 *------------------------------------------------------------------------
 */
bool
NIOSend(
	NIOHandle_t	handle,		/* the handle to the connection */
	char		*buffer,	/* the buffer to send */
	size_t		size)		/* length of the buffer */
{
    NIOerrno = NIOE_NOERROR;
    if (!valid_handle(handle)) {
	NIOerrno = NIOE_HANDLE;
	return FALSE;
    }
    NIOConnection[handle].time_send = times(&timbuf);
    if (NIOConnection[handle].mode == NIO_ACTIVE) {
	switch (NIOConnection[handle].protocol) {
	case NIO_TCPIP:
	    NIOtcpip_send(NIOConnection[handle].out_handle,
		    buffer, size);
	    break;
	default:
	    break;
	}
    } else {
	NIOpassive_send(NIOConnection[handle].out_handle, buffer, size);
    }
    if (NIOerrno != NIOE_NOERROR) {
	ERWrite(NIOerrno, __func__, "%s", strerror(errno));
    }
    return (NIOerrno == NIOE_NOERROR);
} /* NIOSend() */

/*------------------------------------------------------------------------
 *  Function	: NIORecv
 *  Purpose	: Receive a buffer from the connection
 *  Method	: Determine mode and protocol and call appropriate function
 *
 *  History:
 *  21-02-2014	: eck      : creation (of comment)
 *
 *  Description :
 *  This function receives a buffer from the connection.
 *
 *  bufsiz should specify the length of the buffer. The number of bytes that
 *  are read is returned. The buffer will hold the string that was read from
 *  the handle.
 *
 *  Example     :
 *
 *	if (!NIORecv(handle, buffer, sizeof(buffer)-1)) {
 *	    ERWrite(NIOerrno, __func__, "receiving buffer");
 *	    return;
 *	}
 *
 *  Returns	: The number of bytes read or -1 on error
 *
 *  Error values:
 *
 *	NIOE_NOERROR	(ok)
 *	NIOE_HANDLE	(invalid handle)
 *	NIOE_RECVERR	(io error)
 *------------------------------------------------------------------------
 */
size_t
NIORecv(
	NIOHandle_t	handle,		/* the connection handle */
	char		*buffer,	/* the buffer to receive */
	size_t		bufsiz)		/* length of buffer   */
{
    int ir = 0;

    NIOerrno = NIOE_NOERROR;
    if (!valid_handle(handle)) {
	NIOerrno = NIOE_HANDLE;
	return (size_t)-1;
    }
    if (NIOConnection[handle].mode == NIO_ACTIVE) {
	switch (NIOConnection[handle].protocol) {
	case NIO_TCPIP:
	    ir = NIOtcpip_recv(NIOConnection[handle].in_handle,
		    buffer, bufsiz);
	    break;
	default:
	    break;
	}
    } else {
	ir = NIOpassive_recv(NIOConnection[handle].in_handle, buffer, bufsiz);
    }
    if ((ir == 0) && (errno != 0)) {
	ERWrite(NIOerrno, __func__, "%s", strerror(errno));
    }
    NIOConnection[handle].time_recv = times(&timbuf);
    return ir;
} /* NIORecv() */

/*------------------------------------------------------------------------
 *  Function	: NIOClose
 *  Purpose	: Close an open connection
 *  Method	: Determine mode and protocol and call appropriate function
 *
 *  History:
 *  21-02-2014	: eck      : creation (of comment)
 *
 *  Notes	:
 *  This function will close down the connection, as specified by the handle.
 *  It will try to do this in a proper way. In case of a failure, FALSE is
 *  returned.
 *
 *  Returns	: TRUE if the conenction valid handle, FALSE otherwise
 *
 *  Error values:
 *
 *	NIOE_HANDLE	handle was invalid
 *------------------------------------------------------------------------
 */
bool
NIOClose(
	NIOHandle_t	handle)		/* Handle to close */
{
    if (!valid_handle(handle)) {
	NIOerrno = NIOE_HANDLE;
	return FALSE;
    }
    if (NIOConnection[handle].mode == NIO_ACTIVE) {
	switch (NIOConnection[handle].protocol) {
	case NIO_TCPIP:
	    NIOtcpip_close(NIOConnection[handle].in_handle);
	    break;
	default:
	    break;
	}
    } else {
	NIOpassive_close(NIOConnection[handle].in_handle,
		NIOConnection[handle].out_handle);
    }
    NIOConnection[handle].mode = NIO_NOMODE;
    return TRUE;
} /* NIOClose() */

/*------------------------------------------------------------------------
 *  Function	: NIOElapsed
 *  Purpose	: determine elapsed time
 *  Method	: time_recv - time_send
 *
 *  History:
 *  21-02-2014	: eck      : creation (of comment)
 *
 *  Notes	:
 *  This function returns the time that has elapsed between the start of the
 *  last NIOSend() and the completion of the last NIOReceive(), expressed in
 *  1/100 seconds.
 *
 *  Example     :
 *
 *          unsigned	elapsed = NIOElapsed(nioh);
 *
 *  Returns	: Elapsed time in 1/100 seconds or -1 for an invalid handle
 *------------------------------------------------------------------------
 */
unsigned int
NIOElapsed(
	NIOHandle_t	handle)		/* The connection handle */
{
    clock_t	diff;

    if (!valid_handle(handle)) {
	NIOerrno = NIOE_HANDLE;
	return (unsigned int)-1;
    }
    diff = NIOConnection[handle].time_recv - NIOConnection[handle].time_send;
    return (diff / (CLOCKS_PER_SEC / 100));
} /* NIOElapsed() */

/*------------------------------------------------------------------------
 *  Function	: NIO_fildes
 *  Purpose	: Get the file-descriptor for a NIOHandle_t
 *  Method	: obvious
 *
 *  History:
 *  21-02-2014	: eck      : creation (of comment)
 *
 *  Notes	:
 *  This returns the file-decsriptor for a NIOHandle_t, so POSIX functions can
 *  be used on it (like select()).
 *
 *  Example     :
 *
 *  Returns	: The file-decsriptor or -1 for an invalid handle
 *------------------------------------------------------------------------
 */
int
NIO_fildes(
	NIOHandle_t	handle)		/* The connection handle */
{
    if (!valid_handle(handle)) {
	return -1;
    }
    return NIOConnection[handle].in_handle;
} /* NIO_fildes() */
