// \cond DOXYGEN_EXCLUDE

/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: niopassiveall.c
 *  Description	:
 *  Author(s)	: Wim Oosterhof
 *  Creation	: 18-08-1994
 *
 *  History:
 *  18-08-1994	: wimo     : creation
 *  03-11-1994	: eck      : Made NIOpassive_recv() like NIOtcpip_recv().
 *  14-07-1995  : Jv       : In NIOpassive_recv en sendv the processing of
 *                           the returned status was not handled correct and
 *                           the wrong iosb-vmstatus was printed.
 *                           If the iosb-vmstatus = 8428 (disconnect) or
 *                           8412 (connection failed) this
 *                           will be ignored to avoid unnecessary
 *                           errormessages! The disconnection/connectionfailed
 *                           is activated mostly through the timeout in
 *                           osak/vots.
 *  09-11-1995	: eck      : Call NIOtcpip_setkeepalive() in NIOpassive_open().
 *			     This protects against client crashes.  This is
 *			     only important when started by a deamon process,
 *			     and does not work under VMS.
 *  20-11-1996  : PvdM     : Adjusted for Windows 'compatibility'.
 *  26-02-1998	: BdB	   : Trivial additions to compile on Tandem OSS
 *  20-05-2016	: heesakkm : Removed VMS support.
 *
 *  Notes	:
 *  NIOpassive_recv() previously assumed only Pica3-PDU's and skipped telnet
 *  commands.  This is now handled by the frontend (again).  The was
 *  necessary because read() is not guaranteed to return the complete PDU,
 *  so a subsequent NIOpassive_recv() action should not assume that it
 *  starts at a PDU-boundary.
 *========================================================================
 */

#pragma ident "$Id: niopassiveall.c 335412 2016-05-27 11:01:38Z heesakkm $"

#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include <pica.h>
#include <nio.h>
#include "niolocal.h"

static int	ttymin, ttytime, ttylflag, ttyoflag;

bool
NIOpassive_open(int *h_in, int *h_out)
{
    struct termios	ts;

    if (isatty(STDIN_FILENO)) {
    	(void) tcgetattr(STDIN_FILENO, &ts);
    	ttymin = ts.c_cc[VMIN];
    	ttytime = ts.c_cc[VTIME];
    	ttylflag = ts.c_lflag;
    	ts.c_cc[VMIN] = 100;	// Changed from 4000 (& deleted BdB's comment)
    	ts.c_cc[VTIME] = 1;
    	ts.c_lflag = 0;
    	(void) tcsetattr(STDIN_FILENO, TCSANOW, &ts);
    } else {
	/* Assume TCP/IP socket: set keepalive option. */
	NIOtcpip_setkeepalive(STDIN_FILENO);
    }

    if (isatty(STDOUT_FILENO)) {
    	(void) tcgetattr(STDOUT_FILENO, &ts);
    	ttyoflag = ts.c_oflag;
    	ts.c_oflag = 0;
    	(void) tcsetattr(STDOUT_FILENO, TCSANOW, &ts);
    }

    *h_in = STDIN_FILENO;
    *h_out = STDOUT_FILENO;
    return TRUE;
}

void
NIOpassive_send(int handle, char *buffer, size_t size)
{
    if (write(handle, buffer, size) != size) NIOerrno = NIOE_SENDERR;
}

size_t
NIOpassive_recv(int handle, char *buffer, size_t bufsiz)
{
    int		i;

    errno = 0;
    i = read(handle, buffer, bufsiz);
    if (i <= 0) NIOerrno = NIOE_RECVERR;
    return i;
}

void
NIOpassive_close(int h_in, int h_out)
{
    struct termios	ts;

    if (isatty(h_in)) {
    	(void) tcgetattr(h_in, &ts);
    	ts.c_cc[VMIN] = (cc_t)ttymin;
    	ts.c_cc[VTIME] = (cc_t)ttytime;
    	ts.c_lflag = ttylflag;
    	(void) tcsetattr(h_in, TCSANOW, &ts);
    }

    if (isatty(h_out)) {
    	(void) tcgetattr(h_out, &ts);
    	ts.c_oflag = ttyoflag;
    	(void) tcsetattr(h_out, TCSANOW, &ts);
    }
}

// \endcond
