// \cond DOXYGEN_EXCLUDE

/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: niotcpall.c
 *  Description	: Support routines for TCP/IP connections
 *  Author(s)	: Wim Oosterhof
 *  Creation	: 16-08-1994
 *
 *  History:
 *  16-08-1994	: wimo     : creation
 *  11-10-1994	: eck      : Improved NIOtcpip_open().
 *  09-11-1995	: eck      : Added NIOtcpip_setkeepalive()
 *  20-11-1996  : PvdM     : Adjusted for Windows compatibility.
 *  26-02-1998	: BdB	   : Trivial additions to compile on Tandem OSS
 *  09-12-1998	: LB	   : Trivial additions to compile on Tandem Guardian
 *  18-01-1999	: BdB	   : Set keepalive on client connections
 *  20-04-2011	: heesakkm : Removed _GUARDIAN_HOST test.
 *  10-05-2011	: heesakkm : Added use of the CONN library for IPv6 support.
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: niotcpall.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <ctype.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <pica.h>
#include <picamisc.h>
#include <er.h>
#include <cs.h>
#include <conn.h>
#include <nio.h>
#include "niolocal.h"

static bool
det_address(char *remhost, int size, const char *ptr)
{
    ptr = PM_opt_findstart(ptr);
    while (*ptr) {
	if (PM_opt_equal(ptr, "address")) {
	    ptr = PM_opt_findval(ptr);
	    (void) PM_opt_copyval(ptr, remhost, size);
	    return TRUE;
	}
	ptr = PM_opt_findnext(ptr);
    }
    return FALSE;
}

static bool
det_port(char *portno, int size, const char *ptr)
{
    ptr = PM_opt_findstart(ptr);
    while (*ptr) {
	if (PM_opt_equal(ptr, "port")) {
	    ptr = PM_opt_findval(ptr);
	    (void) PM_opt_copyval(ptr, portno, size);
	    return TRUE;
	}
	ptr = PM_opt_findnext(ptr);
    }
    return FALSE;
}

bool
NIOtcpip_open(const char *buffer, int *h_in)
{
    int			handle;
    char		remhost[256];
    char		portno[80];
    CS_DYNAMIC(		address);
    uint16		port_low, port_high, port_first, port_cur;
    const char		*tmpp;

    if (!det_address(remhost, sizeof(remhost) - 1, buffer)) return FALSE;
    if (!det_port(portno, sizeof(portno) - 1, buffer)) return FALSE;

    port_low = atoi(portno);
    port_high = port_low;
    if (port_low == 0) return FALSE;

    tmpp = strchr(portno, '/');
    if ((tmpp != NULL) && (*tmpp != '\0')) {
	port_high = (uint16) (port_low - 1 + atoi(tmpp+1));
	if (port_high < port_low) port_high = port_low;
    }

    errno = 0;

    port_first = port_low + (unsigned) rand() % (port_high - port_low + 1);
    port_cur = port_first;
    for (;;) {
	CSPrintf(address, "%s:%d", remhost, port_cur);

	handle = CONN_connect_tcp(CSContents(address));
	if (handle != -1) {
	    break;
	}
	port_cur++;
	if (port_cur > port_high) port_cur = port_low;
	if (port_cur == port_first) {
	    ERWrite(APPLE_FILEOPEN, __func__, "CONN_connect_tcp() failed: %s", strerror(errno));
	    CSDestroy(address);
	    return FALSE;
	}
    }
    CSDestroy(address);

    /* Socket option SO_KEEPALIVE is implicitly set by CONN_connect_tcp(). */
    /* NIOtcpip_setkeepalive(handle); */

    *h_in = handle;
    return TRUE;
}

void
NIOtcpip_send(int handle, char *buffer, size_t size)
{
    if (write(handle, buffer, size) != (int)size) NIOerrno = NIOE_SENDERR;
    return;
}

size_t
NIOtcpip_recv(int handle, char *buffer, size_t bufsiz)
{
    int i;

    i = read(handle, buffer, bufsiz);
    if (i <= 0) NIOerrno = NIOE_RECVERR;
    return i;
}

void
NIOtcpip_close(int handle)
{
    (void) close(handle);
    return;
}

void
NIOtcpip_setkeepalive(int handle)
{
    int		optval, optsize;

    /* Set the keepalive option. Ignore errors. */
    optval = 1;
    optsize = sizeof(optval);
    (void) setsockopt(handle, SOL_SOCKET, SO_KEEPALIVE,
			(char *) &optval, optsize);
} /* NIOtcpip_setkeepalive() */

// \endcond
