/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: nmtall.c
 *  Description	: All the routines for the NMT-library
 *  Author(s)	: Hans van Eck
 *  Creation	: 15-04-1996
 *
 *  History:
 *  15-04-1996	: eck      : creation
 *  18-04-1996	: eck      : Changed some functions into macro's,
 *			   : Added NMTDel().
 *  19-04-1996	: eck      : Added name parameter to NewOld() interface.
 *  18-07-1997	: eck      : Added load-balancing for better performance.
 *			   : Added NMTWalk().
 *  14-08-1997	: eck      : Added parameter to NMTWalk() and its callback.
 *  24-08-1998	: BdB	   : Allow NMTWalk() to quit before end of table
 *  15-01-1999	: eck      : Changed some ints to uint32s: problems on Tandem
 *  09-09-2004	: eck      : Fixed typos in comments
 *  12-02-2008	: sieburgw : Added NMTVisit (had to change some internal
 *			     interfaces for that) 
 *  05-03-2008	: sieburgw : Renamed NMTVisit to NMTWalkOrdered
 *                           and made the interface like NMTWalk.
 *  05-03-2008	: sieburgw : Renamed some internal structure members.
 *  12-09-2013	: martinz  : Added documentation from NMTLIB.DOC.
 *  10-10-2013	: martinz  : Corrected error values documentation.
 *
 *  Notes	:
 *  The NMT-library consists of a set of functions creating an abstract
 *  interface to nametables. A nametable maps a names into values. This can be
 *  done in many ways (like binary-tree, linked-list). The implementation
 *  chosen uses a hash-table.
 *
 *  A nametable can be thought of as an associative array. The values in the
 *  array can basically be anything, since the value is determined by the
 *  application. In order to have most flexibility, the programmer must specify
 *  an allocation/deallocation routine. Details about this routine can be found
 *  below.
 *
 *  The nametable-routines maintain a resizing hash-table with
 *  name/value-pairs, with linked lists for clashes.
 *
 *  The following datastructures are use in the implementation:
 *  - NMTEntry_t:
 *	- identifier name
 *	- identifier hash value (for quick comparison)
 *	- pointer to next element (for clashes)
 *	- pointer to user-supplied value
 *  - NMTHandle_t:
 *	- a pointer to a table (array of pointers)
 *	- size of the table
 *	- an allocation/deallocation routine
 *	- nr. entries in the table
 *	- nr. filled entries in the table (where value != NULL)
 *	- maximum load allowed (== size/nr.entries)
 *	- nesting level of NMTWalk().
 *
 *  The allocation/deallocation routine can be supllied by the programmer.
 *  Its purpose is to free the old value and return the new value to
 *  insert. This gives a programmer a lot of flexibility in how to handle
 *  different situations (fi. two NMTPut()-s with different non-NULL
 *  values). It is even possible to create scope-like functionality with
 *  this mechanism.
 *  The load balancing uses 3 variables. The nr. of entries is used to
 *  determine whether or not to resize. The nr. filled entries is used for
 *  the new size. During resizing, unused entries are put on a free list.
 *  The free list has a maximum number of elements.
 *  The nesting-level for NMTWalk()-s prevents load-balancing during
 *  NMTWalk()-s, which might otherwise cause problems.
 *  The  nesting-level is also increased/decreased when creating/deleting
 *  iterators.
 *
 *  The NewOld-routine
 *  ------------------
 *  The NMTCreate()-routine has a routine as it's second parameter: the
 *  NewOld() routine. The NewOld() routine is called during an NMTPut() and
 *  NMTDestroy(). Its job is to perform allocation/deallocation when
 *  necessary. This routine must usually be supplied by the programmer, but
 *  there is a default routine for strings that performs (re-)allocation. If
 *  no allocation/deallocation is necessary (by the library), NMT_NOFUNC can
 *  be used.
 *
 *	voidp
 *	NewOld(
 *   	    const char	*Name,	// The name of the value
 *	    cvoidp	Newp,	// The new value
 *	    cvoidp	Oldp);	// Value currently in the table
 *
 *  @param Name is the name with which the value can be retrieved.
 *  @param Newp is the value given in an NMTPut().
 *  @param Oldp is the value that is currently in the nametable.
 *
 *  In general, the NewOld() routine should free the memory associated with
 *  Oldp (which can be a complicated structure), but it could also refuse to do
 *  so if it was not a deletion (in which case Oldp should be returned). This
 *  happens on two successive NMTPut()-calls with different (non-NULL) values.
 *  Since this behavior is determined by the programmer, (s)he can test for
 *  this situation after an NMTPut().
 *
 *  When Newp is NULL, this must be seen as a deletion. The return value need
 *  not necessarily be NULL, but it should do some form of 'delete', since the
 *  NewOld() routine will be called in a loop during an NMTDestroy(). This
 *  loop ends when NewOld() returns NULL. During the loop, the NewOld()-
 *  routine will be called with the value previously returned as second
 *  argument. This facility makes it possible to create 'scopes'.
 *
 *  When Oldp is NULL, it means that the name was not in the table.
 *
 *  The NewOld()-routine must return the value to put into the table. Usually
 *  (but not always), this will be the same as Newp.
 *
 *  The Name can be used to store the name of the record inside the record.
 *  This is the actual name that is allocated by the library, so it never not
 *  be mofified or deallocated.
 *
 *  There are two NewOld() interfaces supplied with the library:
 *  NMTNewOldString() and NMT_NOFUNC. The annotated source of
 *  NMTNewOldString() can be found in this source file. NMT_NOFUNC can be
 *  seen as a function that just returns Newp.
 *
 *  A string example
 *  ----------------
 *  When the values are strings, deallocation and allocation could be combined
 *  into a reallocation followed by a string-copy. This gives the function
 *  NMTNewOldString().
 *
 *  A scope example
 *  ---------------
 *
 *      voidp
 *      NewOldIdent(		// Does allocation
 *          const char    *name	// name of the record
 *          cvoidp         newp,// New value (actually an (ident_t *))
 *          cvoidp         oldp)// Current value (actually an (ident_t *))
 *      {
 *          if (newp == NULL) {		// delete from scope
 *             ident_t	*tmp = oldp;
 *
 *             oldp = tmp->id_next;
 *             tmp->id_next = NULL;
 *             FreeIdent(tmp);
 *             return oldp;
 *          }
 *          ((ident_t *)newp)->id_next = oldp;
 *          ((ident_t *)newp)->id_name = name;	// might come in handy
 *          return newp;
 *      } // NewOldIdent()
 *
 *========================================================================
 */
#pragma ident "$Id: nmtall.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <nmt.h>
#include <er.h>
#include <cs.h>

ERROR_DEFINE(NMTE_NOERROR);
ERROR_DEFINE(NMTE_NOMEM);

NMTErrno_t NMTerrno = NMTE_NOERROR;

#define DEFAULT_TABLE_SIZE	211	/* A prime */
#define DEFAULT_MAXLOAD		5.0
#define MAX_FREE_ENTRIES	200	/* Max. nr. entries on free-list */

typedef struct NMTEntry_t *NMTEntry_p;

struct NMTEntry_t {
    char	*ne_name;		/* name of the entry */
    uint32	ne_hashval;		/* its hash (for quick comparison) */
    NMTEntry_p	ne_next;		/* next entry */
    cvoidp	ne_value;		/* value to return */
    NMTEntry_p  ne_pred;		
    NMTEntry_p  ne_succ;	/*
				 * ne_pred and ne_succ:
				 *   list of entries in order
				 *   of insertion
				 *   nmt_head and nmt_tail are in
				 *   the main table
				 */
};

struct NMTable_t {
    uint32	nmt_size;
    uint32	nmt_nrentries;
    uint32	nmt_nrfilled;
    double	nmt_maxload;
    int		nmt_walknesting;
    cvoidp	(*nmt_newold)(const char *name, cvoidp newp, cvoidp oldp);
    bool	(*nmt_tostring)(const char *, cvoidp eltp, voidp xtra);
    NMTEntry_p	*nmt_table;
    NMTEntry_p	nmt_head;
    NMTEntry_p	nmt_tail;
};

struct NMT_Iterator_t {
    NMTEntry_p	ni_curr;
    NMTHandle_p	ni_nmt;
};

/* The list of free entries. The nr_free_entries should not exceed
 * MAX_FREE_ENTRIES.
 */
static NMTEntry_p	free_entries = NULL;
static int		nr_free_entries = 0;

static bool		NMT_atexit_installed = FALSE;

/*------------------------------------------------------------------------
 *  Function	: shutdown_nmt_lib
 *  Purpose	: destroy static variables, to help searching for
 *		  memory leaks 
 *  Method	: to be installed with atexit() 
 *
 *  History:
 *  17-09-2013	: sieburgw : creation
 *  16-11-2015	: martinz  : Made more robust.
 *
 *  Notes	: Because atexit() handlers trigger in reverse order, atexit()
 *  		  code that use the NMT-library should register its atexit()
 *  		  handler after shutdown_nmt_lib() has been registered.
 *  		  See e.g. ENVInit(). But even without doing so, our code
 *  		  should not crash. So we set free_entries to NULL here.
 *
 *  Returns	: Nothing.
 *
 *------------------------------------------------------------------------
 */
static void
shutdown_nmt_lib(void)
{
    while (free_entries != NULL) {
	NMTEntry_p p = free_entries;

	free_entries = p->ne_next;
	free(p->ne_name);
	free(p);
    }
    nr_free_entries = 0;
} /* shutdown_nmt_lib() */

/*------------------------------------------------------------------------
 *  Function	: NMT_Create
 *  Purpose	: Create an empty nametable
 *  Method	: Malloc and initialise datastructures.
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *  30-03-2012	: paul     : added parameter: EltToString
 *
 *  Notes	:
 *  The options recognized are "size=..." and "maxload=...".
 *
 *  Description	:
 *  NMToptions is an options-string. The possible options are:
 *  - "size=<nr>", where <nr> is the size of the hash-table.
 *    The default value is 211. If the size is 1, you will get a linked list.
 *  - "maxload=<factor>", where factor is a the maximum load factor for this
 *    table. The default factor is 5. The factor is a floating-point number.
 *    Usually, a lower load-factor means better the performance.
 *
 *  NewOld is an allocation/deallocation routine. This routine is called
 *  during an NMTPut() and NMTDestroy(). It is describe in more detail in
 *  NMTNewOldString().
 *
 *  NMTCreate() allocates the initial datastructures for a nametable. The
 *  implementation uses a hash table of <size> entries. Elements that have
 *  congruent hash-values mod <size>, are put in a linked list. When the number
 *  of elements causes the load to be higher than the maxload, the table-size
 *  is doubled.
 *
 *  Error values:
 *
 *	NMTE_NOERROR	(ok)
 *	NMTE_NOMEM	(out of memory)
 *
 *  Returns	: A handle to the nametable created or NULL
 *------------------------------------------------------------------------
 */
NMTHandle_p
NMT_Create(
	const char	*options,		/* options-string */
	cvoidp		(*NewOld)(const char *, cvoidp newp, cvoidp oldp),
	bool		(*EltToString)(const char *, cvoidp eltp, voidp xtra))
{
    char	valbuf[PM_INTLEN_MAX];
    int		tablesize = DEFAULT_TABLE_SIZE;
    double	maxload = DEFAULT_MAXLOAD;
    int		i;
    NMTHandle_p	handle;

    NMTerrno = NMTE_NOERROR;

    if (PM_opt_fetch(options, "size", valbuf, sizeof(valbuf))) {
	tablesize = atoi(valbuf);
	if (tablesize <= 0) tablesize = DEFAULT_TABLE_SIZE;
    }
    if (PM_opt_fetch(options, "maxload", valbuf, sizeof(valbuf))) {
	maxload = strtod(valbuf, NULL);
	if (maxload <= 0) maxload = DEFAULT_MAXLOAD;
    }

    handle = malloc(sizeof(*handle));
    if (handle != NULL) {
	handle->nmt_table = malloc(tablesize * sizeof(NMTEntry_p));
    }
    if ((handle == NULL) || (handle->nmt_table == NULL)) {
	NMTerrno = NMTE_NOMEM;
	(void) ERWrite(NMTerrno, __func__, "%s", options);

	if (handle != NULL) free(handle);
	handle = NULL;
    } else {
	handle->nmt_size = tablesize;
	handle->nmt_nrentries = 0;
	handle->nmt_nrfilled = 0;
	handle->nmt_maxload = maxload;
	handle->nmt_walknesting = 0;
	handle->nmt_newold = NewOld;
	handle->nmt_tostring = EltToString;
	for (i = 0; i < tablesize; i++) {
	    handle->nmt_table[i] = NULL;
	}
	handle->nmt_head = NULL;
	handle->nmt_tail = NULL;
    }
    if (!NMT_atexit_installed){
	// Only once for application : Register an atexit() function.
	atexit (shutdown_nmt_lib);
	NMT_atexit_installed = TRUE;
    }
    return handle;
} /* NMT_Create() */

/*------------------------------------------------------------------------
 *  Function	: NMTCreate
 *  Purpose	: Create an empty nametable
 *  Method	: NMT_Create()
 *
 *  History:
 *  30-03-2012	: paul     : creation
 *
 *  Notes	:
 *
 *  Returns	: A handle to the nametable created or NULL
 *------------------------------------------------------------------------
 */
NMTHandle_p
NMTCreate(
	const char	*options,
	cvoidp		(*NewOld)(const char *, cvoidp newp, cvoidp oldp))
{
    return NMT_Create(options, NewOld,
		      (NewOld == &NMTNewOldString
		       ? NMTStringToString
		       : NMTPointerToString));
} /* NMTCreate() */

/*------------------------------------------------------------------------
 *  Function	: NMTNewEntry
 *  Purpose	: Create a new hash-entry
 *  Method	: malloc()-s
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *  18-07-1997	: eck      : Added use of free_entries
 *
 *  Notes	:
 *  When an entry is put on the free list, its' name remains allocated.
 *
 *  Returns	: The entry allocated or NULL
 *------------------------------------------------------------------------
 */
static NMTEntry_p
NMTNewEntry(
	NMTHandle_p	table,		/* main handle */
	uint32		hash_value,	/* hash-value of the name */
	const char	*name)		/* the name */
{
    NMTEntry_p	ptr;

    if (free_entries != NULL) {
	ptr = free_entries;
	free_entries = free_entries->ne_next;
	nr_free_entries--;
    } else {
	ptr = malloc(sizeof(* ptr));
	ptr->ne_name = NULL;
    }
    if (ptr != NULL) {
	ptr->ne_name = realloc(ptr->ne_name, strlen(name) + 1);
	if (ptr->ne_name == NULL) {
	    free(ptr);
	    ptr = NULL;
	}
    }
    if (ptr != NULL) {
	(void) strcpy(ptr->ne_name, name);
	ptr->ne_hashval = hash_value;
	ptr->ne_next = NULL;
	ptr->ne_value = NULL;
	ptr->ne_succ = NULL;
	if (table->nmt_head == NULL){
	    table->nmt_head = ptr;
	}
	if (table->nmt_tail != NULL){
	    table->nmt_tail->ne_succ = ptr;
	}
	ptr->ne_pred = table->nmt_tail;
	table->nmt_tail = ptr;
    } else {
	NMTerrno = NMTE_NOMEM;
	(void) ERWrite(NMTerrno, __func__, "%s", name);
    }
    return ptr;
} /* NMTNewEntry() */

/*------------------------------------------------------------------------
 *  Function	: AddToList
 *  Purpose	: Add an element to a linked list
 *  Method	:
 *
 *  History:
 *  17-07-1997	: eck      : creation
 *
 *  Notes	:
 *  This routine inserts the element. Appending might take longer (depending
 *  on the load).
 *  This routine is used for lists in the table as well as the free list.
 *------------------------------------------------------------------------
 */
static void
AddToList(
	NMTEntry_p	*list,		/* List to add to */
	NMTEntry_p	eltp)		/* Element to add */
{
    eltp->ne_next = *list;
    *list = eltp;
} /* AddToList() */

/*------------------------------------------------------------------------
 *  Function	: NMTFreeEntry
 *  Purpose	: Add entries to the free list.
 *  Method	:
 *
 *  History:
 *  18-07-1997	: eck      : creation
 *
 *  Notes	:
 *  Not more than MAX_FREE_ENTRIES are on the list.
 *------------------------------------------------------------------------
 */
static void
NMTFreeEntry(
	NMTHandle_p	table,		/* main nmt handle */
	NMTEntry_p	entryp)		/* Entry to free */
{
    if (entryp->ne_pred != NULL){
	entryp->ne_pred->ne_succ = entryp->ne_succ;
    }
    if (entryp->ne_succ != NULL){
	entryp->ne_succ->ne_pred = entryp->ne_pred;
    }
    if (table->nmt_head == entryp){
	table->nmt_head = entryp->ne_succ;
    }
    if (table->nmt_tail == entryp){
	table->nmt_tail = entryp->ne_pred;
    }

    if (nr_free_entries < MAX_FREE_ENTRIES) {
	entryp->ne_next = NULL;
	AddToList(&free_entries, entryp);
	nr_free_entries++;
    } else {
	free(entryp->ne_name);
	entryp->ne_name = NULL;
	free(entryp);
    }

} /* NMTFreeEntry() */

/*------------------------------------------------------------------------
 *  Function	: NMTResizeTable
 *  Purpose	: Increase the size of the table
 *  Method	:
 *  Malloc a new array for the pointers and move each element from the old
 *  table to the new one.
 *
 *  History:
 *  17-07-1997	: eck      : creation
 *
 *  Notes	:
 *  The routine checks for resizing itself, so it can be called from
 *  multiple places.
 *------------------------------------------------------------------------
 */
static void
NMTResizeTable(
	NMTHandle_p	table)		/* table to expand */
{
    uint32	new_size;
    NMTEntry_p	*new_table;
    uint32	i;

    if (table->nmt_walknesting > 0) {
	return;
    }

    /* First compute if resizeing is necessary. */
    new_size = table->nmt_size;
    while (((double) table->nmt_nrentries / new_size) > table->nmt_maxload) {
	new_size *= 2;
    }
    if (new_size == table->nmt_size) return;

    /* Now, compute the 'real' new size, depending on the actual entries. */
    new_size = table->nmt_size;
    while (((double) table->nmt_nrfilled / new_size) > table->nmt_maxload) {
	new_size *= 2;
    }

    new_table  = malloc(new_size * sizeof(NMTEntry_p));
    if (new_table == NULL) return;
    for (i = 0; i < new_size; i++) {
	new_table[i] = NULL;
    }

    for (i = 0 ;  i < table->nmt_size ; i++) {
	NMTEntry_p	tmp = table->nmt_table[i];

	while ((tmp = table->nmt_table[i]) != NULL) {
	    table->nmt_table[i] = tmp->ne_next;
	    tmp->ne_next = NULL;

	    if (tmp->ne_value != NULL) {
		AddToList(&new_table[tmp->ne_hashval % new_size], tmp);
	    } else {
		NMTFreeEntry(table, tmp);
		table->nmt_nrentries--;
	    }
	}
    }

    free(table->nmt_table);
    table->nmt_table = new_table;
    table->nmt_size = new_size;
} /* NMTResizeTable() */

/*------------------------------------------------------------------------
 *  Function	: NMT
 *  Purpose	: Get, Put or Delete a value
 *  Method	:
 *  - Determine hash-value and index in the table
 *  - Walk down linked list until the address of the pointer to the right
 *    entry is found (this eases a possible store)
 *  - if we shouldn't put a value, return the value found
 *  - otherwise call the user-supplied NewOld() routine and put the value
 *    returned by it into the table
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *  18-07-1997	: eck      : Added load-balancing
 *
 *  Notes	:
 *  Don't change the code if you are not used to pointers to pointers.
 *
 *  Returns	: The value as found/stored in the table
 *------------------------------------------------------------------------
 */
cvoidp
NMT(
	NMTHandle_p	table,		/* table to search in */
	const char	*name,		/* the name */
	cvoidp		value,		/* the value (ignored for get) */
	bool		mustput)	/* TRUE when called from NMTPut() */
{
    uint32	hash_value;
    uint32	idx;
    NMTEntry_p	*entry_pp;
    cvoidp	retval;
    bool	entry_added = FALSE;

    if ((NULL == table) || (NULL == name)){
	return NULL;
    }

    hash_value = PM_hash(name);
    idx = hash_value % (table->nmt_size);
    entry_pp = &(table->nmt_table[idx]);

    NMTerrno = NMTE_NOERROR;

    while (*entry_pp != NULL) {
	NMTEntry_p	entry_p = *entry_pp;

	if ((entry_p->ne_hashval == hash_value)
	    && STREQ(entry_p->ne_name, name)) {
	    break;
	}
	entry_pp = &(entry_p->ne_next);
    }
    if (! mustput) {
	if (*entry_pp == NULL) return NULL;
	return (*entry_pp)->ne_value;
    }

    /* Only create an entry when not deleting. */
    if ((value != NULL) && (*entry_pp == NULL)) {
	table->nmt_nrentries++;
	*entry_pp = NMTNewEntry(table, hash_value, name);

	/* We might need to resize the table. */
	entry_added = TRUE;
    }
    if (*entry_pp != NULL) {
	cvoidp	(*newold)(const char *, cvoidp, cvoidp) = table->nmt_newold;
	bool	was_filled, is_filled;

	was_filled = ((*entry_pp)->ne_value != NULL);
	if (newold != NMT_NOFUNC) {
	    (*entry_pp)->ne_value = newold((*entry_pp)->ne_name,
					    value, (*entry_pp)->ne_value);
	} else {
	    (*entry_pp)->ne_value = value;
	}
	is_filled = ((*entry_pp)->ne_value != NULL);
	if (is_filled != was_filled) {
	    /* Safer than "table->nmt_nrfilled += is_filled - was_filled" */
	    if (is_filled) {
		table->nmt_nrfilled++;
	    } else {
		table->nmt_nrfilled--;
	    }
	}
    } else {
	return NULL;
    }
    retval = (*entry_pp)->ne_value;
    if (entry_added) {
	NMTResizeTable(table);
    }
    return retval;
} /* NMT() */

/*------------------------------------------------------------------------
 *  Function	: NMTGet
 *  Purpose	: Retrieve a value from a nametable
 *  Method	: Use the macro
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *
 *  Notes	: It's just a simple wrapper.
 *
 *  Description	:
 *  NMThandle must be a handle to a previously NMTCreate()-ed table (and not
 *  yet NMTDestroy()-ed).
 *  Name is the name of the element to get.
 *
 *  The function will return a pointer to the value or NULL.
 *
 *  Error values:
 *
 *	NMTE_NOERROR	(ok)
 *
 *  Returns	: The value found (or NULL)
 *------------------------------------------------------------------------
 */
cvoidp
(NMTGet)(
	NMTHandle_p	table,		/* the nametable */
	const char	*name)		/* the name to look for */
{
    return NMTGet(table, name);
} /* NMTGet() */

/*------------------------------------------------------------------------
 *  Function	: NMTPut
 *  Purpose	: Put a value into a nametable
 *  Method	: Use the macro
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *
 *  Notes	:
 *  An NMTPut() of a NULL value will cause a deletion.
 *
 *  Description	:
 *  NMThandle must be a handle to a previously NMTCreate()-ed table (and not
 *  yet NMTDestroy()-ed).
 *  Name is the name of the element to put.
 *  Value is the value of the element to put. When it is NULL, the old value
 *  will be deleted.
 *
 *  This function will cause the NewOld()-function (the second argument to
 *  NMTCreate()) to be called with (a copy of) Name as first parameter, Value
 *  as its second parameter, and the current value from the table as its third
 *  parameter. This means that the NewOld()-function can determine its own
 *  allocation/deallocation strategy. It is expected, though, that the routine
 *  will cause a deallocation of the first argument if the second argument is
 *  NULL. The NewOld() routine must return the pointer to store in the table.
 *  The first parameter can be used to store the name of the record inside the
 *  record without having to mallocing and copying it.
 *
 *  Error values:
 *
 *	NMTE_NOERROR	(ok)
 *	NMTE_NOMEM	(out of memory)
 *
 *  Returns	: The value stored in the table
 *------------------------------------------------------------------------
 */
cvoidp
(NMTPut)(
	NMTHandle_p	table,		/* the nametable */
	const char	*name,		/* the name to put */
	cvoidp		value)		/* value to put */
{
    return NMTPut(table, name, value);
} /* NMTPut() */

/*------------------------------------------------------------------------
 *  Function	: NMTDel
 *  Purpose	: Delete a name from a table
 *  Method	: Use the macro
 *
 *  History:
 *  18-04-1996	: eck      : creation
 *
 *  Notes	:
 *  This will not always delete the whole name (in case of scopes). This
 *  depends on the programmer supplied NewOld() routine.
 *
 *  Description	:
 *  NMThandle must be a handle to a previously NMTCreate()-ed table (and not
 *  yet NMTDestroy()-ed).
 *  Name is the name of the element to delete.
 *
 *  This function will cause the NewOld()-function (the second argument to
 *  NMTCreate()) to be called with (a copy of) Name as first parameter, NULL as
 *  its second parameter, and the current value from the table as its third
 *  parameter.
 *
 *  Error values:
 *
 *	NMTE_NOERROR	(ok)
 *	NMTE_NOMEM	(out of memory)
 *
 *  Returns	: The value stored in the table
 *------------------------------------------------------------------------
 */
cvoidp
(NMTDel)(
	NMTHandle_p	table,		/* the nametable */
	const char	*name)		/* the name to delete */
{
    return NMTDel(table, name);
} /* NMTDel() */

/*------------------------------------------------------------------------
 *  Function	: NMTWalk
 *  Purpose	: Call a user-supplied function for every element in the table
 *  Method	:
 *
 *  History:
 *  18-07-1997	: eck      : creation
 *  14-08-1997	: eck      : Added extra parameter for callback.
 *  24-08-1998	: BdB	   : Allow NMTWalk() to quit before end of table
 *  04-12-2012	: martinz  : Handle NULL table input parameter.
 *
 *  Notes	:
 *  Perhaps this should have been done as a pair of get-first/get-next
 *  routines, but that would make it hard/impossible to graciously handle
 *  NMTPut()-s while walking down the list.
 *
 *  Actually, I am using an NMTDel() inside the callback of NMTWalk() to remove
 *  all entries with a name satisfying certain criteria (in csnimp_process.c).
 *  Hope that's ok.
 *
 *  Description	:
 *  NMThandle must be a handle to a previously NMTCreate()-ed table (and not
 *  yet NMTDestroy()-ed).
 *  HandleElt is a (pointer to a) function which handles the element.
 *  ExtraArg is an extra parameter that can be used to pass some extra
 *  information to HandleElt() (and back).
 *
 *  NMTWalk() will call HandleElt() for each element in the nametable in an
 *  unspecified order.
 *
 *  The HandleElt() function will receive the name of the entry as first
 *  parameter and its' value as the second. The value will be equal to
 *  NMTGet(NMThandle, name). The third parameter, extra_arg, is the same as the
 *  ExtraArg given to NMTWalk().
 *  If this routine returns TRUE, NMTWalk() will continue normally. If it
 *  returns FALSE, NMTWalk() returns immediately with the return value FALSE.
 *
 *  Error values:
 *
 *  - none
 *
 *  Example	:
 *
 *	#include	<nmt.h>
 *
 *	static bool
 *	Printer(
 *	    const char	*name,
 *	    cvoidp	value,
 *	    voidp	stream) {
 *	    (void) fprintf((FILE *)stream, "%s = %s\n", name, (char *) value);
 * 	    return TRUE;
 *	}
 *	...
 *	NMTWalk(nmt_handle, Printer, stdout);
 *
 *  Returns	:
 *  - TRUE  : The callback routine always returned TRUE.
 *  - FALSE : The callback routine returned FALSE.
 *
 *------------------------------------------------------------------------
 */
bool
NMTWalk(
	NMTHandle_p	table,		/* name-table handle */
	bool		(*HandleElt)(const char *, cvoidp eltp, voidp xtra),
	voidp		xtra)
{
    uint32	i;
    NMTEntry_p	tmp;
    bool go_on_flag = TRUE;

    NMTerrno = NMTE_NOERROR;

    if (table == NULL) {
	return TRUE;
    }
    table->nmt_walknesting++;
    for (i = 0 ; go_on_flag && i < table->nmt_size; i++) {
	tmp = table->nmt_table[i];
	while (go_on_flag && tmp != NULL) {
	    if (tmp->ne_value != NULL) {
		go_on_flag = HandleElt(tmp->ne_name, tmp->ne_value, xtra);
	    }
	    tmp = tmp->ne_next;
	}
    }
    table->nmt_walknesting--;

    /* Optionally resize the table. This is necessary when the callback
     * adds elements.
     */
    NMTResizeTable(table);
    return go_on_flag;
} /* NMTWalk() */

/*------------------------------------------------------------------------
 *  Function	: NMTWalkOrdered
 *  Purpose	: Call a user-supplied function for every element in the table
 *  Method	:
 *
 *  History:
 *  11-02-2008	: sieburgw : creation
 *  28-06-2012	: sieburgw : Fix CBS-3920
 *  04-12-2012	: martinz  : Handle NULL table input parameter.
 *
 *  Notes	:
 *  Copy of NMTWalk(), but the visit is done in the order as the keys have been
 *  inserted in the NMT.
 *
 *  Caveat: the order of insertion is a bit sloppy definition;
 *  What would it mean if 'a', 'b' and 'c' are added, 'b' is deleted and 'b' is
 *  added again. Is the order now 'a','b', 'c', or is it 'a', 'c', 'b'?
 *  No guarantee about one of the two, so do not rely on it.
 *------------------------------------------------------------------------
 */
bool
NMTWalkOrdered(
	NMTHandle_p	table,
	bool		(*HandleElt)(const char *, cvoidp eltp, voidp xtra),
	void 		*xtra)
{
    NMTEntry_p	node;
    bool go_on_flag = TRUE;

    NMTerrno = NMTE_NOERROR;

    if (table == NULL) {
	return TRUE;
    }
    table->nmt_walknesting++;
    node = table->nmt_head;
    while(go_on_flag && (node != NULL)){

	if (node->ne_value != NULL) {
	    /*
	     * Do not visit deleted nodes
	     */
	    go_on_flag = HandleElt(node->ne_name, node->ne_value, xtra);
	}
	node = node->ne_succ;
    }
    table->nmt_walknesting--;

    /* Optionally resize the table. This is necessary when the callback
     * adds elements.
     */
    NMTResizeTable(table);
    return go_on_flag;
} /* NMTWalkOrdered() */

/*------------------------------------------------------------------------
 *  Function	: NMT_IteratorCreate
 *  Purpose	: Create an iterator, to be used for  NMT_IteratorFirst,
 *		  NMT_IteratorNext etc.
 *  Method	: An iterator treats an NMT as if it is a (double) linked
 *		  list. The order of the list is the order of insertion
 *		  of the keys.
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	: Should be destroyed with NMT_IteratorDestroy
 *
 *  Returns	: pointer to NMT_Iterator
 *
 *------------------------------------------------------------------------
 */
NMT_Iterator_p
NMT_IteratorCreate(NMTHandle_p nmt)
{
    NMT_Iterator_p res;

    if (!nmt){
	return NULL;
    }
    res = calloc(1, sizeof(*res));
    if (!res){
	return NULL;
    }
    res->ni_nmt = nmt;
    nmt->nmt_walknesting++;
    return res;
}

/*------------------------------------------------------------------------
 *  Function	: nmt_iterator_values
 *  Purpose	: Get key and value for current iterator entry. 
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	: Will reset key and value if iter is NULL.
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static void
nmt_iterator_values(
	NMT_Iterator_p	iter,
	const char**	key,
	cvoidp*		value)
{
    if (key){
	if (iter && iter->ni_curr){
	    *key = iter->ni_curr->ne_name; 
	} else {
	    *key = NULL;
	}
    }

    if (value){
	if (iter && iter->ni_curr){
	    *value = iter->ni_curr->ne_value; 
	} else {
	    *value = NULL;
	}
    }
}

/*------------------------------------------------------------------------
 *  Function	: nmt_iterator_next
 *  Purpose	: Advances the iterator to the next valid entry.
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static void
nmt_iterator_next(
	NMT_Iterator_p	iter)
{
    if (!iter || !iter->ni_nmt){
	return;
    }

    if (!iter->ni_curr){
	iter->ni_curr = iter->ni_nmt->nmt_head;
    } else {
	iter->ni_curr = iter->ni_curr->ne_succ;
    }

    while(iter->ni_curr){
	if (
		(iter->ni_curr->ne_name)
	    &&  (iter->ni_curr->ne_value)
	){
	    break;
	}
	iter->ni_curr = iter->ni_curr->ne_succ;
    }
}

/*------------------------------------------------------------------------
 *  Function	: nmt_iterator_prev
 *  Purpose	: Reverses the iterator to the previous valid entry.
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static void
nmt_iterator_prev(
	NMT_Iterator_p	iter)
{
    if (!iter || !iter->ni_nmt){
	return;
    }

    if (!iter->ni_curr){
	iter->ni_curr = iter->ni_nmt->nmt_tail;
    } else {
	iter->ni_curr = iter->ni_curr->ne_pred;
    }

    while(iter->ni_curr){
	if (
		(iter->ni_curr->ne_name)
	    &&  (iter->ni_curr->ne_value)
	){
	    break;
	}
	iter->ni_curr = iter->ni_curr->ne_pred;
    }
}

/*------------------------------------------------------------------------
 *  Function	: NMT_IteratorFirst
 *  Purpose	: Resets the iterator to first valid entry, starting from
 *		  the head of the list (and get the key and value).
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	: FALSE if no key/value could be read, TRUE otherwise
 *
 *------------------------------------------------------------------------
 */
bool
NMT_IteratorFirst(
	NMT_Iterator_p	iter,
	const char**	key,
	cvoidp*		value)
{
    if (iter && iter->ni_nmt){
	iter->ni_curr = NULL;
	nmt_iterator_next(iter);
    }
    nmt_iterator_values(iter, key, value);

    if (!iter || !key || !*key || !value || !*value){
	return FALSE;
    }
    return TRUE;
}

/*------------------------------------------------------------------------
 *  Function	: NMT_IteratorLast
 *  Purpose	: Resets the iterator to the last valid entry, starting from
 *		  the tail of the list (and get the key and value).
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	: FALSE if no key/value could be read, TRUE otherwise
 *
 *------------------------------------------------------------------------
 */
bool
NMT_IteratorLast(
	NMT_Iterator_p	iter,
	const char**	key,
	cvoidp*		value)
{
    if (iter && iter->ni_nmt){
	iter->ni_curr = NULL;
	nmt_iterator_prev(iter);
    }
    nmt_iterator_values(iter, key, value);

    if (!iter || !key || !*key || !value || !*value){
	return FALSE;
    }
    return TRUE;
}

/*------------------------------------------------------------------------
 *  Function	: NMT_IteratorNext
 *  Purpose	: Advance iterator and get key/value
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	: If used on a just created iterator, the effect is the
 *		  same as NMT_IteratorGetFirst
 *
 *  Returns	: FALSE if no key/value could be read, TRUE otherwise
 *
 *------------------------------------------------------------------------
 */
bool
NMT_IteratorNext(
	NMT_Iterator_p	iter,
	const char**	key,
	cvoidp*		value)
{
    nmt_iterator_next(iter);
    nmt_iterator_values(iter, key, value);

    if (!iter || !key || !*key || !value || !*value){
	return FALSE;
    }
    return TRUE;
}

/*------------------------------------------------------------------------
 *  Function	: NMT_IteratorPrev
 *  Purpose	: Reverse iterator and get key/value 
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	: If used on a just created iterator, the effect is the
 *		  same as NMT_IteratorGetLast
 *
 *  Returns	: FALSE if no key/value could be read, TRUE otherwise
 *
 *------------------------------------------------------------------------
 */
bool
NMT_IteratorPrev(
	NMT_Iterator_p	iter,
	const char**	key,
	cvoidp*		value)
{
    nmt_iterator_prev(iter);
    nmt_iterator_values(iter, key, value);

    if (!iter || !key || !*key || !value || !*value){
	return FALSE;
    }
    return TRUE;
}

/*------------------------------------------------------------------------
 *  Function	: NMT_IteratorDestroy
 *  Purpose	: Destroy an iterator created with NMT_IteratorCreate 
 *  Method	: 
 *
 *  History:
 *  31-10-2013	: sieburgw : creation
 *
 *  Notes	:
 *
 *  Returns	: void
 *
 *------------------------------------------------------------------------
 */
void
NMT_IteratorDestroy(
	NMT_Iterator_p iter)
{
    if (iter){
	NMTHandle_p nmt = iter->ni_nmt;
	if (nmt){
	    nmt->nmt_walknesting--;
	}
	free(iter);
    }
}

/*------------------------------------------------------------------------
 *  Function	: NMTDeleteList
 *  Purpose	: Delete a linked list of nametable-entries.
 *  Method	:
 *  Walk down linked list, deleting value(s) and (when del_all is TRUE),
 *  the name and the element itself.
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *  08-05-1996	: eck      : Added del_all parameter
 *  18-07-1997	: eck      : Call NMTFreeEntry() instead of free()-ing
 *
 *  Notes	:
 *  The value should be de-allocated in the programmer-supplied NewOld()-
 *  routine. This routine is called in a loop, so this library can be
 *  used to maintain scopes.
 *
 *  The pointers free()-ed are set to NULL. This simplifies catching errors.
 *------------------------------------------------------------------------
 */
static void
NMTDeleteList(
	NMTHandle_p	table,		/* main nmt handle */
	NMTEntry_p	entryp,		/* list of entries to delete */
	cvoidp		(*newold)(const char *name, cvoidp newp, cvoidp oldp),
	bool		del_all)	/* delete entries and name as well */
{
    NMTEntry_p	tmp;

    while (entryp != NULL) {
	tmp = entryp;
	entryp = entryp->ne_next;

	if (newold != NMT_NOFUNC) {
	    /* This should deallocate */
	    while (tmp->ne_value != NULL) {
		tmp->ne_value = newold(tmp->ne_name, NULL, tmp->ne_value);
	    }
	} else {
	    tmp->ne_value = NULL;	 	/* safety ! */
	}
	if (del_all) {
	    NMTFreeEntry(table, tmp);
	}
    }
} /* NMTDeleteList() */

/*------------------------------------------------------------------------
 *  Function	: NMTDestroy
 *  Purpose	: Free all memory associated with a nametable.
 *  Method	:
 *  - free all linked lists in the table (all nmt_size)
 *  - free the table with the pointers
 *  - free the handle
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *  08-05-1996	: eck      : Added option TRUE to NMTDeleteList(): delete all.
 *
 *  Notes	:
 *  The pointers free()-ed are set to NULL. This simplifies catching errors.
 *
 *  Description	:
 *  NMThandle must be a handle to a previously NMTCreate()-ed table (and not
 *  yet NMTDestroy()-ed).
 *
 *  This routine will free all memory associated with the NMThandle. The
 *  NewOld()-function (given on NMTCreate()) will be called for every value in
 *  the table. This might involve calling NewOld() in a loop for some names.
 *
 *  Error values:
 *
 *	NMTE_NOERROR	(ok)
 *
 *------------------------------------------------------------------------
 */
void
NMTDestroy(
	NMTHandle_p	table)		/* name-table handle */
{
    uint32	i;

    NMTerrno = NMTE_NOERROR;

    if (table == NULL) {
	return;
    }

    for (i = 0 ; i < table->nmt_size; i++) {
	if (table->nmt_table[i] != NULL) {
	    (void) NMTDeleteList(table, table->nmt_table[i], table->nmt_newold, TRUE);
	    table->nmt_table[i] = NULL;	/* safety ! */
	}
    }
    free(table->nmt_table);
    table->nmt_table = NULL;

    free(table);
} /* NMTDestroy() */

/*------------------------------------------------------------------------
 *  Function	: NMTClean
 *  Purpose	: Free the memory associated with the values in a nametable.
 *  Method	:
 *  Call NMTDeletList() with FALSE to indicate value-only deletion.
 *
 *  History:
 *  08-05-1996	: eck      : creation
 *  18-07-1997	: eck      : Added (re-)setting of table->nmt_nrfilled
 *  04-12-2012	: martinz  : Handle NULL table input parameter.
 *
 *  Notes	:
 *  This routine will cause the nametable to be 'cleaned'. That is: all
 *  values are deleted, but the table itself will remain.
 *
 *  Description	:
 *  NMThandle must be a handle to a previously NMTCreate()-ed table (and not
 *  yet NMTDestroy()-ed).
 *
 *  This routine will free memory associated with values stored in a nametable.
 *  The table itself will not be deallocated. The NewOld()-function (given on
 *  NMTCreate()) will be called for every value in the table. This might
 *  involve calling NewOld() in a loop for some names.
 *
 *  After an NMTClean(), the nametable will be empty, but NMThandle will still
 *  be valid.
 *
 *  Error values:
 *
 *	NMTE_NOERROR	(ok)
 *
 *------------------------------------------------------------------------
 */
void
NMTClean(
	NMTHandle_p	table)		/* Name-table to clean */
{
    uint32	i;

    NMTerrno = NMTE_NOERROR;

    if (table == NULL) {
	return;
    }
    for (i = 0 ; i < table->nmt_size; i++) {
	if (table->nmt_table[i] != NULL) {
	    (void) NMTDeleteList(table, table->nmt_table[i], table->nmt_newold, FALSE);
	}
    }
    table->nmt_nrfilled = 0;
} /* NMTClean() */

/*------------------------------------------------------------------------
 *  Function	: NMTNewOldString
 *  Purpose	: Allocation/deallocation for strings.
 *  Method	:
 *  If the first argument is NULL, the free() oldp. Otherwise use realloc() to
 *  reduce memory overhead (though this causes copying).
 *
 *  History:
 *  16-04-1996	: eck      : creation
 *
 *  Notes	:
 *  This routine can be used as second argument in a NMTCreate()-call.
 *  If there is already a value in the table, the new one will replace it.
 *  This routine can be called with both static and dynamic strings: it does
 *  its own allocation. This routine will allocate and deallocate
 *  '\0'-terminated strings.
 *
 *  Returns	: Pointer to the element to put into nametable
 *------------------------------------------------------------------------
 */
cvoidp
NMTNewOldString(
	const char	*name,	/* The name of the record */
	cvoidp		newp,	/* The value given to NMTPut() */
	cvoidp		oldp)	/* The value currently in the table */
{
    const char	*news = newp;		/* newp as string */
    char	*tmp;

    (void) name;	// Keep lint happy

    /* Test for deletion first */
    if (newp == NULL) {
	free((voidp) oldp);
	return NULL;
    }
    /* If we want to enforce deletion before inserting a new value,
     * we must uncomment the statement:
     *     if (oldp != NULL) return oldp;
     */

    /* Free the old string and allocate room for the new one.
     * Using a realloc() might be too expensive in some situations, since
     * it might copy the old string into newly allocated memory.
     * On the other hand, when oldp == newp, a free()/malloc() would not
     * work properly: either this situation should be tested or
     * allocation should be done using malloc()/free().
     */
    tmp = realloc((voidp) oldp, strlen((char *)newp) + 1);
    if (tmp != NULL) {
	(void) strcpy(tmp, news);
    } else {
        /* The reallocation failed. This implies that oldp was not
         * free()-ed. We do that here. If we want it to remain in the
         * table, it should be returned as function value.
         */
	NMTerrno = NMTE_NOMEM;
	(void) ERWrite(NMTerrno, "NMTNewOldString", "no room to realloc");
	free((voidp) oldp);		/* Just delete the old one */
    }
    return tmp;
} /* NMTNewOldString() */

/*------------------------------------------------------------------------
 *  Function	: NMTStringToString
 *  Purpose	: Convert a nametable entry created with NMTNewOldString to
 *                string, for use in debugger
 *  Method	:
 *
 *  History:
 *  30-03-2012	: paul     : creation
 *
 *  Returns	: If NMTWalk() should continue printing the next entry
 *------------------------------------------------------------------------
 */
bool
NMTStringToString(
	const char	*name,	/* Name of the element to print */
	cvoidp		eltp,	/* Value of the element to print or NULL */
	voidp		xtra)	/* User data */
{
    CSBuf_p	buf = (CSBuf_p)xtra;

    /* Separate entries with a double space */
    if (eltp != NULL) {
	CSPrintfAppend(buf, "'%s' -> '%s'  ", name, (char *)eltp);
    } else {
	CSPrintfAppend(buf, "'%s' -> NULL  ", name);
    }
    return TRUE;
} /* NMTStringToString() */

/*------------------------------------------------------------------------
 *  Function	: NMTPointerToString
 *  Purpose	: Convert a nametable entry to string, for use in debugger
 *                From the entry, only the address will be printed.
 *  Method	:
 *
 *  History:
 *  30-03-2012	: paul     : creation
 *
 *  Returns	: If NMTWalk() should continue printing the next entry
 *------------------------------------------------------------------------
 */
bool
NMTPointerToString(
	const char	*name,	/* Name of the element to print */
	cvoidp		eltp,	/* Value of the element to print or NULL */
	voidp		xtra)	/* User data */
{
    CSBuf_p	buf = (CSBuf_p)xtra;

    /* Separate entries with a double space */
    if (eltp != NULL) {
	CSPrintfAppend(buf, "'%s' -> %p  ", name, eltp);
    } else {
	CSPrintfAppend(buf, "'%s' -> NULL  ", name);
    }
    return TRUE;
} /* NMTPointerToString() */

/*------------------------------------------------------------------------
 *  Function	: NMTToStringSet
 *  Purpose	: Set the default 'ToString' callback function
 *  Method	:
 *
 *  History:
 *  30-03-2012	: paul     : creation
 *------------------------------------------------------------------------
 */
void
NMTToStringSet(
	NMTHandle_p	table,
	bool		(*EltToString)(const char *, cvoidp eltp, voidp xtra))
{
    if (table != NULL) {
	table->nmt_tostring = EltToString;
    }
} /* NMTToStringSet() */

/*------------------------------------------------------------------------
 *  Function	: NMTToString
 *  Purpose	: Convert a nametable to string, for use in debugger
 *  Method	:
 *
 *  History:
 *  29-03-2012	: paul     : creation
 *  04-12-2012	: martinz  : Handle NULL table input parameter.
 *
 *  Notes	:
 *  The interface is kept as simple as possible, in order to debug easily.
 *  Newlines are not printed between entries, because the debugger will
 *  probably show "\n", which is definitely not what you want.
 *  You might want to print such newlines in other applications, but then you
 *  need to define a different function which does something similar to this
 *  one.
 *
 *  Returns	:
 *  Pointer to static buffer containing the nametable string or string "NULL"
 *  in case of error
 *------------------------------------------------------------------------
 */
const char *
NMTToString(
	NMTHandle_p	table)		/* Name-table to print */
{
    static CSBuf_p	buf = NULL;

    if (table == NULL) {
	return NULL;
    }
    if (buf == NULL) {
	buf = CSCreate();
	if (buf == NULL) {
	    return "NULL";
	}
    }
    CSClear(buf);

    NMTWalk(table, table->nmt_tostring, buf);

    return CSContents(buf);
} /* NMTToString() */

/*------------------------------------------------------------------------
 *  Function	: NMTToStringElt
 *  Purpose	: Convert a nametable element to string, for use in debugger
 *  Method	:
 *
 *  History:
 *  30-03-2012	: paul     : creation
 *  04-12-2012	: martinz  : Handle NULL table input parameter.
 *
 *  Returns	:
 *  Pointer to static buffer containing the element string or string "NULL"
 *  in case of error
 *------------------------------------------------------------------------
 */
const char *
NMTToStringElt(
	NMTHandle_p	table,	/* Name-table containing element */
	const char	*name)	/* Name of element to print */
{
    static CSBuf_p	buf = NULL;
    cvoidp		eltp;

    if (table == NULL) {
	return NULL;
    }
    if (buf == NULL) {
	buf = CSCreate();
	if (buf == NULL) {
	    return "NULL";
	}
    }
    CSClear(buf);

    eltp = NMTGet(table, name);
    table->nmt_tostring(name, eltp, buf);

    return CSContents(buf);
} /* NMTToStringElt() */
