/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3connection.c
 *  Description	: Local stuff for Pica3-connections
 *  Author(s)	: Hans van Eck
 *  Creation	: 06-09-1994
 *
 *  History:
 *  06-09-1994	: eck      : creation
 *
 *  Notes	:
 *  There are two data-structures that are maintained here:
 *  The connection-list and the connection-ADB.
 *
 *  The connection-list is a linked list that contains the P3Handle, the
 *  FEHandle and the WMContext. This could not be stored into an ADB,
 *  since they have (more or less) abstract datatypes.
 *
 *  The connection-ADB is use for other information about the connection.
 *  Into this ADB, "map"-s are generated for DBList() and "id-type"-s and
 *  "id-name"-s for DBNameInfo().
 *========================================================================
 */

#pragma ident "$Id: p3connection.c 335412 2016-05-27 11:01:38Z heesakkm $"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <adb.h>
#include <cs.h>
#include <p3.h>
#include <er.h>

#include "p3local.h"

static p3_connection_p	P3ConnectionList = NULL;
static ADBHandle_p	conn_adb = NULL;
static p3_XLWMconnection_p	P3XLWMConnectionList = NULL;
static ADBHandle_p	XLWMconn_adb = NULL;

/*------------------------------------------------------------------------
 *  Function	: P3_ADBInit
 *  Purpose	: Initialise the connection ADB
 *  Method	: obvious.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Notes	: Should not fail
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
P3_ADBInit(void)
{
    if (conn_adb != NULL) return TRUE;

    conn_adb = ADBCreate(ADBT_FULLMEM);
    if (conn_adb == NULL) {
	P3errno = P3E_NOMEM;
	(void) ERWrite(ADBerrno, "P3_ADBInit", "creating connection ADB");
    }
    return (P3errno == P3E_NOERROR);
} /* P3_ADBInit() */

/*------------------------------------------------------------------------
 *  Function	: FreeConnection
 *  Purpose	: Deallocate a connection
 *  Method	: Destroy the context, delete the connection from the ADB
 *		  free the P3Handle and free the element from the list.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *------------------------------------------------------------------------
 */
static void
FreeConnection(
	p3_connection_p	p3_conn)	/* connection to deallocate */
{
    CS_LOCAL(	key, ADB_KEYLENGTH_MAX);

    WMDeleteContext(p3_conn->p3cn_wm_context);
    (void) CSPrintf(key, "connection[id=%s]", p3_conn->p3cn_p3_handle);

    /* Ignore any error messages  from ADBDelVar() */
    (void) ADBDelVar(conn_adb, CSContents(key));

    /* Cast to get rid of const */
    free((voidp) p3_conn->p3cn_p3_handle);
    free(p3_conn);
} /* FreeConnection() */

/*------------------------------------------------------------------------
 *  Function	: NewConnection
 *  Purpose	: Allocate datastructures for a new connection
 *  Method	:
 *  1- Allocate connection element for linked list
 *  2- Create new P3-handle
 *  3- Create new WM context
 *  4- Create new entry in ADB
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Notes	:
 *  When something fails, everything up to that point will be undone.
 *
 *  Returns	: pointer to new connection-list element (or NULL).
 *------------------------------------------------------------------------
 */
static p3_connection_p
NewConnection(
	FEHandle_t	fe_handle)	/* FEhandle of the new connection */
{
    p3_connection_p	tmp_conn;
    const char		*p3_handle;
    char		*tmp_p3_handle = NULL;

    if (! P3_ADBInit()) {
	return NULL;
    }
    tmp_conn = malloc(sizeof(* tmp_conn));
    if (tmp_conn != NULL) {
	tmp_conn->p3cn_fe_handle = fe_handle;
	tmp_conn->p3cn_next = NULL;

	p3_handle = PM_gensym("P3_");
	tmp_p3_handle = malloc(strlen(p3_handle) + 1);
	if (tmp_p3_handle == NULL) {
	    free(tmp_conn);
	    tmp_conn = NULL;
	} else {
	    (void) strcpy(tmp_p3_handle, p3_handle);
	    tmp_conn->p3cn_p3_handle = tmp_p3_handle;
	}
    }
    if (tmp_conn != NULL) {
	tmp_conn->p3cn_wm_context = WMCreateContext(CLIENT);
	if (tmp_conn->p3cn_wm_context == WM_NOCONTEXT) {
	    /* Cast to get rid of const */
	    free((voidp) tmp_conn->p3cn_p3_handle);
	    tmp_conn->p3cn_p3_handle = NULL;
	    free(tmp_conn);
	    tmp_conn = NULL;
	}
    }
    if (tmp_conn != NULL) {
	if ((! ADBNewVar(conn_adb, "connection", NULL))
	    || (! ADBNewVar(conn_adb, "connection[:last].id",
					tmp_conn->p3cn_p3_handle))) {
	    FreeConnection(tmp_conn);
	    tmp_conn = NULL;
	}
    }
    if (tmp_conn == NULL) {
	P3errno = P3E_NOMEM;
    }
    return tmp_conn;
} /* NewConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_AddConnection
 *  Purpose	: Add a new connection to the existing list.
 *  Method	: Create connection and link it before the rest.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: pointer to new connection-list element (or NULL).
 *------------------------------------------------------------------------
 */
p3_connection_p
P3C_AddConnection(
	FEHandle_t	fe_handle)	/* FEhandle of the new connection */
{
    p3_connection_p p3_conn = NewConnection(fe_handle);

    if (p3_conn != NULL) {
	p3_conn->p3cn_next = P3ConnectionList;
	P3ConnectionList = p3_conn;
    }
    return p3_conn;
} /* P3C_AddConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_GetConnection
 *  Purpose	: Get connection-list element using the P3Handle.
 *  Method	: Walk down linked list until found
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: The connecion-list element or NULL.
 *------------------------------------------------------------------------
 */
p3_connection_p
P3C_GetConnection(
	P3Handle_p	P3handle)		/* P3Handle to look for */
{
    p3_connection_p p3_conn = P3ConnectionList;

    while ((p3_conn!= NULL) && (! STREQ(p3_conn->p3cn_p3_handle, P3handle))) {
	p3_conn = p3_conn->p3cn_next;
    }
    return p3_conn;
} /* P3C_GetConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_DelConnection
 *  Purpose	: Delete all datastructures corresponding to a connection
 *  Method	:
 *  1- Walk down linked list until found
 *  2- Unlink from list
 *  3- call FreeConnection()
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: FALSE if element not found in list, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
P3C_DelConnection(
	p3_connection_p	p3_conn)	/* Connection to delete */
{
    p3_connection_p	*conn_pp = &P3ConnectionList;

    while ((*conn_pp != NULL) && (*conn_pp != p3_conn)) {
	conn_pp = &((*conn_pp)->p3cn_next);
    }
    if (*conn_pp == NULL) {
	return FALSE;
    }
    (*conn_pp) = p3_conn->p3cn_next;
    FreeConnection(p3_conn);
    return TRUE;
} /* P3C_DelConnection() */


/*------------------------------------------------------------------------
 *  Function	: P3C_AddToConnection
 *  Purpose	: Store an element into the connection ADB
 *  Method	:
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3C_AddToConnection(
	P3Handle_p	p3_handle,
	const char	*attrib,
	const char	*value)
{
    CS_LOCAL(	variable, ADB_KEYLENGTH_MAX);

    CSPrintf(variable, "connection[id=%s].%s", p3_handle, attrib);

    return ADBNewVar(conn_adb, CSContents(variable), value);
} /* P3C_AddToConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_GetFromConnection
 *  Purpose	: Get information from the connection ADB
 *  Method	:
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: pointer to value or NULL
 *------------------------------------------------------------------------
 */
const char *
P3C_GetFromConnection(
	P3Handle_p	p3_handle,
	const char	*attrib)
{
    CS_LOCAL(	variable, ADB_KEYLENGTH_MAX);

    CSPrintf(variable, "connection[id=%s].%s", p3_handle, attrib);

    return ADBGet(conn_adb, CSContents(variable));
} /* P3LGetFromConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3XLWM_ADBInit
 *  Purpose	: Initialise the connection ADB
 *  Method	: obvious.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Notes	: Should not fail
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
static bool
P3XLWM_ADBInit(void)
{
    if (XLWMconn_adb != NULL) return TRUE;

    XLWMconn_adb = ADBCreate(ADBT_FULLMEM);
    if (XLWMconn_adb == NULL) {
	P3errno = P3E_NOMEM;
	(void) ERWrite(ADBerrno, "P3XLWM_ADBInit", "creating connection ADB");
    }
    return (P3errno == P3E_NOERROR);
} /* P3XLWM_ADBInit() */

/*------------------------------------------------------------------------
 *  Function	: FreeXLWMConnection
 *  Purpose	: Deallocate a connection
 *  Method	: Destroy the context,
 *		  free the P3Handle and free the element from the list.
 *
 *  History:
 *  13-12-2001	: evd      : creation
 *------------------------------------------------------------------------
 */
static void
XLWMFreeConnection(
	p3_XLWMconnection_p	p3_conn)	/* connection to deallocate */
{
    CS_LOCAL(	key, ADB_KEYLENGTH_MAX);

    WMDeleteContext(p3_conn->p3XLWMcn_wm_context);
    (void) CSPrintf(key, "connection[id=%s]", p3_conn->p3XLWMcn_p3_handle);

    /* Ignore any error messages  from ADBDelVar() */
    (void) ADBDelVar(XLWMconn_adb, CSContents(key));

    /* Cast to get rid of const */
    free((voidp) p3_conn->p3XLWMcn_p3_handle);
    free(p3_conn);
} /* XLWMFreeConnection() */

/*------------------------------------------------------------------------
 *  Function	: XLWMNewConnection
 *  Purpose	: Allocate datastructures for a new connection
 *  Method	:
 *  1- Allocate connection element for linked list
 *  2- Create new P3-handle
 *  3- Create new WM context
 *  4- Create new entry in ADB
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Notes	:
 *  When something fails, everything up to that point will be undone.
 *
 *  Returns	: pointer to new connection-list element (or NULL).
 *------------------------------------------------------------------------
 */
static p3_XLWMconnection_p
XLWMNewConnection(
	NIOHandle_t	nio_handle)	/* niohandle of the new connection */
{
    p3_XLWMconnection_p	tmp_conn;
    const char		*p3_handle;
    char		*tmp_p3_handle = NULL;

    if (! P3XLWM_ADBInit()) {
	return NULL;
    }
    tmp_conn = malloc(sizeof(* tmp_conn));
    if (tmp_conn != NULL) {
	tmp_conn->p3XLWMcn_nio_handle = nio_handle;
	tmp_conn->p3XLWMcn_next = NULL;

	p3_handle = PM_gensym("P3XLWM_");
	tmp_p3_handle = malloc(strlen(p3_handle) + 1);
	if (tmp_p3_handle == NULL) {
	    free(tmp_conn);
	    tmp_conn = NULL;
	} else {
	    (void) strcpy(tmp_p3_handle, p3_handle);
	    tmp_conn->p3XLWMcn_p3_handle = tmp_p3_handle;
	}
    }
    if (tmp_conn != NULL) {
	tmp_conn->p3XLWMcn_wm_context = WMCreateContext(CLIENT);
	if (tmp_conn->p3XLWMcn_wm_context == WM_NOCONTEXT) {
	    /* Cast to get rid of const */
	    free((voidp) tmp_conn->p3XLWMcn_p3_handle);
	    tmp_conn->p3XLWMcn_p3_handle = NULL;
	    free(tmp_conn);
	    tmp_conn = NULL;
	}
    }
    if (tmp_conn != NULL) {
	if ((! ADBNewVar(XLWMconn_adb, "connection", NULL))
	    || (! ADBNewVar(XLWMconn_adb, "connection[:last].id",
					tmp_conn->p3XLWMcn_p3_handle))) {
	    XLWMFreeConnection(tmp_conn);
	    tmp_conn = NULL;
	}
    }
    if (tmp_conn == NULL) {
	P3errno = P3E_NOMEM;
    }
    return tmp_conn;
} /* XLWMNewConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_XLWMAddConnection
 *  Purpose	: Add a new connection to the existing list.
 *  Method	: Create connection and link it before the rest.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: pointer to new connection-list element (or NULL).
 *------------------------------------------------------------------------
 */
p3_XLWMconnection_p
P3C_XLWMAddConnection(
	NIOHandle_t	nio_handle)	/* NIOhandle of the new connection */
{
    p3_XLWMconnection_p p3_conn = XLWMNewConnection(nio_handle);

    if (p3_conn != NULL) {
	p3_conn->p3XLWMcn_next = P3XLWMConnectionList;
	P3XLWMConnectionList = p3_conn;
    }
    return p3_conn;
} /* P3C_XLWMAddConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_XLWMGetConnection
 *  Purpose	: Get connection-list element using the P3Handle.
 *  Method	: Walk down linked list until found
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: The connecion-list element or NULL.
 *------------------------------------------------------------------------
 */
p3_XLWMconnection_p
P3C_XLWMGetConnection(
	P3Handle_p	P3handle)		/* P3Handle to look for */
{
    p3_XLWMconnection_p p3_conn = P3XLWMConnectionList;

    while ((p3_conn!= NULL) && (! STREQ(p3_conn->p3XLWMcn_p3_handle, P3handle))) {
	p3_conn = p3_conn->p3XLWMcn_next;
    }
    return p3_conn;
} /* P3C_XLWMGetConnection() */

/*------------------------------------------------------------------------
 *  Function	: P3C_XLWMDelConnection
 *  Purpose	: Delete all datastructures corresponding to a connection
 *  Method	:
 *  1- Walk down linked list until found
 *  2- Unlink from list
 *  3- call XLWMFreeConnection()
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: FALSE if element not found in list, TRUE otherwise
 *------------------------------------------------------------------------
 */
bool
P3C_XLWMDelConnection(
	p3_XLWMconnection_p	p3_conn)	/* Connection to delete */
{
    p3_XLWMconnection_p	*conn_pp = &P3XLWMConnectionList;

    while ((*conn_pp != NULL) && (*conn_pp != p3_conn)) {
	conn_pp = &((*conn_pp)->p3XLWMcn_next);
    }
    if (*conn_pp == NULL) {
	return FALSE;
    }
    (*conn_pp) = p3_conn->p3XLWMcn_next;
    XLWMFreeConnection(p3_conn);
    return TRUE;
} /* P3C_XLWMDelConnection() */

