/*========================================================================
 *  Copyright (c) 1997 by Pica
 *
 *  File	: p3fetchtitle.c
 *  Description	: Fetch a (part of a) title.
 *
 *  History:
 *  24-10-1997	: HSi	: creation
 *  12-01-1999  : EvD   : Added "\\bib <iln>" option
 *  05-11-2001  : els   : Added "lvl option l = local, m = main.
 *  20-11-2001  : evd   : Added P3XLWMFetchTitle
 *  12-03-2002  : evd   : P3errno = P3E_NOHITS at status P3_NOHITS
 *  12-10-2006  : evd   : charset=utf8 option added to P3XLWMFetchTitle
 *
 *========================================================================
 */

#pragma ident "$Id: p3fetchtitle.c 335412 2016-05-27 11:01:38Z heesakkm $"

#include <stdio.h>
#include <string.h>

#include <pica.h>
#include <picatypes.h>
#include <picamisc.h>
#include <p3.h>
#include <cs.h>

#include "p3local.h"

#define OPTVAL_SIZ	(32)

#define OPT_COPY	"copy"
#define OPT_SYNTAX	"syntax"
#define OPT_EXPANSION	"expansion"
#define OPT_ILN		"iln"
#define OPT_LVL         "lvl"
#define OPT_CHARSET     "charset"


typedef struct {
    char		 copy[8];
    char		 syntax[OPTVAL_SIZ];
    char		 expansion[OPTVAL_SIZ];
    char		 iln[PIC_ILN_DECIMAL_MAX + 1 + 6];
    char                 charset[20];
} commandArgs_t;

/*------------------------------------------------------------------------
 *  Function	: convExpansionAdbToP3
 *  Purpose	: Convert expansion from adb format to P3-protocol format.
 *  Method	: obvious
 *
 *  History:
 *  27-10-1997	: HSi	: creation
 *  02-07-2002	: clemens  : Default to adbExpansion if not in table.
 *
 *  Returns	:
 *------------------------------------------------------------------------
 */
static const char *
convExpansionAdbToP3 (
	const char	 *adbExpansion
	)
{
    static struct {
	const char	*adb;
	const char	*p3;
    } expansionTable[] = {
	{ "pica3",		"X3" },
	{ "pica+",		"XP" },
    };

    int			 i;
    const char		*p3Expansion = adbExpansion;

    if ( adbExpansion != NULL ) {
	for ( i = 0;
	      i < sizeof(expansionTable) / sizeof(expansionTable[0]);
	      i++ ) {
	    if ( STREQ ( expansionTable[i].adb, adbExpansion ) ) {
		p3Expansion = expansionTable[i].p3;
		break;
	    }
	}
    }
    return p3Expansion;
} /* convExpansionAdbToP3 () */

/*------------------------------------------------------------------------
 *  Function	: DetermineArgs
 *  Purpose	: Determine syntax and expansion from options
 *  Method	: obvious
 *
 *  History:
 *  27-10-1997	: HSi	: creation
 *
 *  Returns	:
 *------------------------------------------------------------------------
 */
static bool
DetermineArgs(
    const char		*options,
    commandArgs_t	*args
	)
{
    bool		 ok 		  	     = TRUE;
    char		 copyOption[OPTVAL_SIZ]      = "";
    char		 syntaxOption[OPTVAL_SIZ]    = "";
    char		 expansionOption[OPTVAL_SIZ] = "";
    char		 ilnOption[OPTVAL_SIZ]	     = "";
    char		 lvlOption[OPTVAL_SIZ]	     = "";
    char                 charsetOption[OPTVAL_SIZ]   = "";

    options = PM_opt_findstart ( options );
    while ( ( options != NULL ) && ( *options != '\0' ) ) {
	if ( PM_opt_equal ( options, OPT_COPY ) ) {
	    options = PM_opt_findval ( options );
	    PM_opt_copyval ( options,
			     copyOption, sizeof copyOption );
	} else {
	    if ( PM_opt_equal ( options, OPT_SYNTAX ) ) {
		options = PM_opt_findval ( options );
		PM_opt_copyval ( options,
				 syntaxOption, sizeof syntaxOption );
	    } else {
		if ( PM_opt_equal ( options, OPT_EXPANSION ) ) {
		    options = PM_opt_findval ( options );
		    PM_opt_copyval ( options,
				     expansionOption, sizeof expansionOption );
		} else {
		    if ( PM_opt_equal ( options, OPT_ILN ) ) {
			options = PM_opt_findval ( options );
			PM_opt_copyval ( options,
					 ilnOption, sizeof ilnOption );
		    } else {
			if ( PM_opt_equal ( options, OPT_LVL ) ) {
			    options = PM_opt_findval ( options );
			    PM_opt_copyval ( options,
					     lvlOption, sizeof lvlOption );
			} else {
			    if ( PM_opt_equal ( options, OPT_CHARSET ) ) {
				options = PM_opt_findval ( options );
				PM_opt_copyval ( options,
						 charsetOption, 
						 sizeof charsetOption );
			    }
			}
		    }
		}
	    }
	}
	options = PM_opt_findnext(options);
    }

    if ( copyOption[0] == '\0' ) {
	if ( lvlOption[0] == '\0' ) {
	    args->copy[0] = '\0';
	} else {
	    (void) sprintf(args->copy, "%s", lvlOption);
	}
    } else {
	(void) sprintf(args->copy, "E%02d", atoi (copyOption));
    }

    if ( ilnOption[0] == '\0' ) {
	args->iln[0] = '\0';
    } else {
	(void) sprintf ( args->iln, "\\bib %d", atoi (ilnOption) );
    }

    if (charsetOption[0] == '\0') {
	args->charset[0] ='\0';
    } else {
	strcpy(args->charset, charsetOption);
    }

    /* Convert syntax and expansion from option format to P3 format */
    PM_strncpy(args->syntax, P3L_SyntaxAdbToP3(syntaxOption), sizeof(args->syntax));
    PM_strncpy(args->expansion, convExpansionAdbToP3(expansionOption), sizeof(args->expansion));

    return ok;
} /* DetermineArgs() */

/*------------------------------------------------------------------------
 *  Function	: FillReturnedInfo
 *  Purpose	: Fill counted buffer with returned info.
 *  Method	: obvious
 *
 *  History:
 *  27-10-1997	: HSi	: creation
 *
 *  Returns	: number of fields stored.
 *------------------------------------------------------------------------
 */
static int
FillReturnedInfo(
    CBBuf_t	*cbBuf,
    size_t	 cbSize
    )
{
    const char	*location   = "P3FetchTitle.FillReturnedInfo";
    int		 fieldCount = 0;

    /* If field MO is present not all copies are returned. This is
     * not what we want. (TBD fetch other copies as well?)
     */
    fieldCount += P3L_WMelt2CB ( "MO", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID | FLG_OPTIONAL,
				 location );

    fieldCount += P3L_WMelt2CB ( "PP", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID, location );
    fieldCount += P3L_WMelt2CB ( "TX", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID | FLG_SPLITLINE,
				 location);
    return fieldCount;
} /* FillReturnedInfo() */

/*------------------------------------------------------------------------
 *  Function	: P3FetchTitle
 *  Purpose	: Fetch a (part of a) title.
 *  Method	:
 *
 *  History:
 *  24-10-1997	: HSi	: creation
 *
 *  Notes	: Supported options: "copy", "syntax" and "expansion"
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3FetchTitle (
	P3Handle_p	 p3handle,	/* Handle to Pica3-server  */
	const char	*options,	/* Options		   */
	const char	*ppn,		/* PPN of title to fetch   */
	CBBuf_t		*cbBuf,		/* Buffer to contain title */
	size_t		 cbSize		/* Size of buffer 	   */
	)
{
    static char		 fName [] = "P3FetchTitle";
    p3_connection_p	 p3_conn = P3C_GetConnection ( p3handle );
    WMContext_t		 prev_context;
    CS_LOCAL(		 command, 160);
    commandArgs_t	 commandArgs;

    P3errno = P3E_NOERROR;

    if (p3_conn != NULL) {
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3cn_wm_context);
	if (! DetermineArgs ( options, &commandArgs ) ) {
	    P3errno = P3E_PROTOCOL;
	    (void) ERWrite ( P3errno, fName, "options=%s", options );
	    return FALSE;
	}
	CSPrintf ( command, "\\fet \\ppn %s %s %s %s %s ;", 
		   ppn, commandArgs.iln,
		   commandArgs.copy, commandArgs.syntax, commandArgs.expansion);
	WMPutCmd("US", CSContents(command));
	if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
	    switch (P3status) {
	    case P3_NOTNOW:
		    P3errno = P3E_PROTOCOL;
		    break;
	    case P3_NOAUTH:
		    P3errno = P3E_AUTHORITY;
		    break;
	    case P3_ERRTTL:
		    P3errno = P3E_TTLWRONG;
		    break;
	    case P3_ERROR:
		    P3errno = P3E_SERVER;
		    break;
	    case P3_NOHITS:
		    P3errno = P3E_NOHITS;
		    break;
	    case P3_OK:
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
	    }

	    if (P3errno == P3E_NOERROR) {
		(void) FillReturnedInfo(cbBuf, cbSize);
	    }
	}
	(void) WMSwitchContext(prev_context);
    } else {
	/* The handle is not valid */
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, fName, "invalid handle");
    }

    return (P3errno == P3E_NOERROR);
} /* P3FetchTitle () */

/*------------------------------------------------------------------------
 *  Function	: XLWMFillReturnedInfo
 *  Purpose	: Fill counted buffer with returned info.
 *  Method	: obvious
 *
 *  History:
 *  20-12-2001	: evd	: copied from FillReturnedInfo
 *  09-11-2004	: michael  : Added "LX" handling (links)
 *  08-02-2005	: michael  : Added "EX" handling (99+ copies thingie)
 *
 *  Returns	: number of fields stored.
 *------------------------------------------------------------------------
 */
static int
XLWMFillReturnedInfo(
    CBBuf_t	*cbBuf,
    size_t	 cbSize
    )
{
    const char	*location   = "P3XLWMFetchTitle.XLWMFillReturnedInfo";
    int		 fieldCount = 0;

    /* If field MO is present not all copies are returned. This is
     * not what we want. (TBD fetch other copies as well?)
     */
    fieldCount += P3L_XLWMelt2CB ( "MO", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID | FLG_OPTIONAL,
				 location );

    fieldCount += P3L_XLWMelt2CB ( "PP", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID, location );
    fieldCount += P3L_XLWMelt2CB ( "TX", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID | FLG_SPLITLINE,
				 location);

    fieldCount += P3L_XLWMelt2CB ( "LX", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID | FLG_OPTIONAL,
				 location);

    fieldCount += P3L_XLWMelt2CB ( "EX", cbBuf, cbSize,
				 FLG_FIELD | FLG_ADDID | FLG_OPTIONAL,
				 location);
    fieldCount += P3L_XLWMelt2CB ( "#C", cbBuf, cbSize,
				 FLG_VAR | FLG_ADDID | FLG_OPTIONAL,
				 location);
    return fieldCount;
} /* XLWMFillReturnedInfo() */

/*------------------------------------------------------------------------
 *  Function	: P3XLWMFetchTitle
 *  Purpose	: Fetch a (part of a) title.
 *  Method	:
 *
 *  History:
 *  20-12-2001  :HSi	: creation
 *
 *  Notes	: Supported options are: "copy", "syntax" and "expansion"
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3XLWMFetchTitle (
	P3Handle_p	 p3handle,	/* Handle to Pica3-server  */
	const char	*options,	/* Options		   */
	const char	*ppn,		/* PPN of title to fetch   */
	CBBuf_t		*cbBuf,		/* Buffer to contain title */
	size_t		 cbSize		/* Size of buffer 	   */
	)
{
    static char		 fName [] = "P3XLWMFetchTitle";
    p3_XLWMconnection_p	 p3_conn = P3C_XLWMGetConnection ( p3handle );
    WMContext_t		 prev_context;
    CS_LOCAL(		 command, 160);
    commandArgs_t	 commandArgs;

    P3errno = P3E_NOERROR;

    if (p3_conn != NULL) {
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3XLWMcn_wm_context);
	if (! DetermineArgs ( options, &commandArgs ) ) {
	    P3errno = P3E_PROTOCOL;
	    (void) ERWrite ( P3errno, fName, "options=%s", options );
	    return FALSE;
	}
	CSPrintf ( command, "\\fet \\ppn %s %s %s %s %s ;", 
		   ppn, commandArgs.iln,
		   commandArgs.copy, commandArgs.syntax, commandArgs.expansion);
	WMPutCmd("US", CSContents(command));
        if (commandArgs.charset[0] != '\0') {
	    WMRegVar(P3_VAR_CLIENT_CHARSET);
	    WMPutVar(P3_VAR_CLIENT_CHARSET, commandArgs.charset);
	}
	if (P3L_XLWMSendRecv(p3_conn->p3XLWMcn_nio_handle)) {
	    switch (P3status) {
	    case P3_NOTNOW:
		    P3errno = P3E_PROTOCOL;
		    break;
	    case P3_NOAUTH:
		    P3errno = P3E_AUTHORITY;
		    break;
	    case P3_ERRTTL:
		    P3errno = P3E_TTLWRONG;
		    break;
	    case P3_ERROR:
		    P3errno = P3E_SERVER;
		    break;
	    case P3_NOHITS:
		    P3errno = P3E_NOHITS;
		    break;
	    case P3_OK:
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
	    }

	    if (P3errno == P3E_NOERROR) {
		(void) XLWMFillReturnedInfo(cbBuf, cbSize);
	    }
	}
	(void) WMSwitchContext(prev_context);
    } else {
	/* The handle is not valid */
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, fName, "invalid handle");
    }

    return (P3errno == P3E_NOERROR);
} /* P3XLWMFetchTitle () */
