/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3info.c
 *  Description	: Ask information from a Pica3-server.
 *  Author(s)	: Hans van Eck
 *  Creation	: 16-09-1994
 *
 *  History:
 *  16-09-1994	: eck      : creation
 *  10-10-2013	: martinz  : Copied documentation from P3LIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: p3info.c 335828 2016-06-10 11:33:48Z heesakkm $"

#include <stddef.h>
#include <string.h>

#include <pica.h>
#include <cb.h>
#include <p3.h>

#include "p3local.h"

/*------------------------------------------------------------------------
 *  Function	: P3Info
 *  Purpose	: Retrieve information about a handle
 *  Method	: Call of P3C_GetFromConnection().
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Description	:
 *  - P3handle must be a valid P3handle_p previously returned by P3Open() (and
 *    not P3Close()-d).
 *  - DBattr is a pointer to a string containing (';'-terminated) attributes.
 *    Valid attributes are described in the ADB library.
 *  - CBbuf and CBsize are the counted buffer arguments used to store the data
 *    in.
 *
 *  The routine will look in its ADB if the required information is stored; if
 *  so, CBbuf is filled with that information otherwise an empty counted buffer
 *  is returned.
 *
 *  Routine P3Info() doesn't communicate with the server; P3Explain() does.
 *  If something goes wrong, P3Info() will return FALSE and P3errno should be
 *  examined.
 *
 *  Notes	:
 *  This function may only be used after a successful call to P3Select().
 *
 *  Error values:
 *
 *	P3E_BUFFER	(buffer overflow)
 *	P3E_NOERROR	(ok)
 *	P3E_NOMEM	(not enough memory)
 *
 *  Returns	: TRUE if handle ok, FALSE otherwise
 *------------------------------------------------------------------------
 */
bool
P3Info(
	P3Handle_p	P3handle,	/* Handle to get info about */
	const char	*DBattr,	/* attribute wanted */
	CBBuf_t		*CBbuf,		/* CB to store info */
	size_t		CBsize)		/* size of the CB */
{
    p3_connection_p	p3_conn = P3C_GetConnection(P3handle);
    const char		*tmp;

    P3errno = P3E_NOERROR;
    if (p3_conn == NULL) {
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3Info", "invalid handle");
    } else {
	tmp = P3C_GetFromConnection(P3handle, DBattr);
	if (tmp != NULL) {
	    CBAppElement(CBbuf, CBsize, (voidp) tmp, strlen(tmp) + 1);
	}
    }
    return (P3errno == P3E_NOERROR);
} /* P3Info() */
