/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3list.c
 *  Description	: Get a list of systems/databases from a Pica3-server
 *  Author(s)	: Hans van Eck
 *  Creation	: 12-09-1994
 *
 *  History:
 *  12-09-1994	: eck      : creation
 *  21-06-1995	: eck      : server now returns whole id. instead of part
 *			   : (except from server-part).
 *  10-10-2013	: martinz  : Copied documentation from P3LIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: p3list.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <cb.h>
#include <p3.h>

#include "p3local.h"

/*------------------------------------------------------------------------
 *  Function	: InList
 *  Purpose	: Test if the name has already been expanded
 *  Method	: Test the existence of map with P3C_GetFromConnection().
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *
 *  Returns	: indication of presence
 *------------------------------------------------------------------------
 */
static bool
InList(
	P3Handle_p	P3handle,	/* Handle to Pica3-server */
	const char	*DBname)	/* name to test */
{
    CS_LOCAL(	key,	ADB_KEYLENGTH_MAX);
    (void) CSPrintf(key,"map[from=%s].from", DBname);

    return (P3C_GetFromConnection(P3handle, CSContents(key)) != NULL);
} /* InList() */

/*------------------------------------------------------------------------
 *  Function	: ServerName
 *  Purpose	: Get the server name from an id.
 *  Method	: If there is no '@' in the id. return the whole id
 *		: otherwise, return the part after the '@'.
 *
 *  History:
 *  21-06-1995	: eck      : creation
 *
 *  Returns	: Server part of the id.
 *------------------------------------------------------------------------
 */
static const char *
ServerName(
	const char	*db_id)
{
    const char	*ptr;

    ptr = strchr(db_id, '@');
    if (ptr == NULL) {
	return db_id;
    } else {
	return ptr + 1;
    }
} /* ServerName() */

/*------------------------------------------------------------------------
 *  Function	: SplitListLine
 *  Purpose	: Split a line of systems/databases into its parts
 *  Method	:
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *
 *  Notes	:
 *  A list-line has the following syntax:
 *
 *	['+'] <id> '=' <name> '\r'
 *
 *  The optional '+' indicates that the id. is NOT a database (and thus can
 *  be expanded).
 *
 *  Returns	: pointer to the next line in the buffer.
 *------------------------------------------------------------------------
 */
static const char *
SplitListLine(
	const char	*dblist,	/* ptr to start of line */
	bool		*expandable,	/* FALSE when name is database */
	CSBuf_p 	db_id,		/* id. of system/database */
	CSBuf_p 	db_name)	/* expanded name of id. */
{
    const char	*line_end;
    const char	*id_end;
    int		length;

    CSClear(db_id);
    CSClear(db_name);
    *expandable = FALSE;

    if ((dblist == NULL) || (*dblist == '\0')) return NULL;

    line_end = strchr(dblist, '\r');
    if (line_end == NULL) return NULL;

    id_end = strchr(dblist, '=');
    if ((id_end == NULL) || (id_end > line_end)) {
	id_end = line_end;
    }

    if (*dblist == '+') {
	*expandable = TRUE;
  	dblist++;
    }
    length = id_end - dblist;
    CSPrintf(db_id, "%.*s", length, dblist);

    if (id_end != line_end) {
	id_end++;
	length = line_end - id_end;
	CSPrintf(db_name, "%.*s", length, id_end);
    }
    return (line_end + 1);
} /* SplitListLine() */

/*------------------------------------------------------------------------
 *  Function	: CombineIds
 *  Purpose	: Combine a mapped id. with a returned result.
 *  Method	:
 *  Append the DBname to the db_id, separated by a '@' if DBname is a
 *  server or '.' otherwise.
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *  21-06-1995	: eck      : server now returns the whole id. (minus
 *			   : server-part).
 *
 *  Notes	: No checking is done.
 *------------------------------------------------------------------------
 */
static void
CombineIds(
	const char	*DBname,	/* The name mapped from */
	CSBuf_p 	db_id)		/* Name returned by server */
{
    CSAppendC(db_id, '@');
    CSAppend(db_id, ServerName(DBname));
} /* CombineIds() */

/*------------------------------------------------------------------------
 *  Function	: SaveList
 *  Purpose	: Save the list of names returned by the server.
 *  Method	: Parse the buffer and store through P3C_AddToConnection().
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *
 *  Notes	:
 *  This function creates something like the following structure in the ADB:
 *
 *  map {
 *      from = Dbname;		: a server
 *	to = name1;		: a database
 *      to = name2;		: a system
 *  }
 *  id-type {
 *	id = name1;
 *	type = "database";
 *  }
 *  id-name {
 *	id = name1;
 *	name = "Some printable database name";
 *  }
 *  id-name {
 *	id = name1;
 *	name = "Some printable system name";
 *  }
 *------------------------------------------------------------------------
 */
static void
SaveList(
	P3Handle_p	P3handle,	/* Handle to the server */
	const char	*DBname,	/* The name to expand */
	const char	*dblist)	/* List returned by the server */
{
    CS_LOCAL(	db_id, 100);
    CS_LOCAL(	db_name, 100);
    bool	expandable;
    int		count = 0;

    while ((dblist != NULL) && (*dblist != '\0')) {
	dblist = SplitListLine(dblist, &expandable, db_id, db_name);
	if (dblist == NULL) break;
	CombineIds(DBname, db_id);
	if (count == 0) {
	    (void) P3C_AddToConnection(P3handle, "map", NULL);
	    (void) P3C_AddToConnection(P3handle, "map[:last].from", DBname);
	}
	(void) P3C_AddToConnection(P3handle, "map[:last].to",
					    CSContents(db_id));
	count++;
	(void) P3C_AddToConnection(P3handle, "id-type", NULL);
	(void) P3C_AddToConnection(P3handle, "id-type[:last].id",
					    CSContents(db_id));
	if (! expandable) {
	    (void) P3C_AddToConnection(P3handle, "id-type[:last].type",
						"database");
	}
	if (CSLength(db_name) != 0) {
	    (void) P3C_AddToConnection(P3handle, "id-name", NULL);
	    (void) P3C_AddToConnection(P3handle, "id-name[:last].id",
						CSContents(db_id));
	    (void) P3C_AddToConnection(P3handle, "id-name[:last].name",
						CSContents(db_name));
	}
    }
} /* SaveList() */

/*------------------------------------------------------------------------
 *  Function	: GenerateList
 *  Purpose	: Send a list-command to the server and store the result
 *  Method	: obvious
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *------------------------------------------------------------------------
 */
static void
GenerateList(
	p3_connection_p	p3_conn,	/* the server-connection */
	const char	*DBname,	/* name to generate list from */
	const char	*options)	/* for "language=.." */
{
    WMContext_t		prev_context;
    CS_LOCAL(		command, 160);
    CS_LOCAL(		Pica3Name, 64);
    char		varbuf[MAXLEN];	/* ??? too large! */

    prev_context = P3L_SwitchAndClearContext(p3_conn->p3cn_wm_context);
    P3L_InternToPica3Name(DBname, Pica3Name);

    (void) CSPrintf(command, "\\lis \\cat %s ;", CSContents(Pica3Name));
    P3L_InsertLanguageCommand(command, options);
    WMPutCmd("US", CSContents(command));
    if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
	if ((P3status != P3_OK) && (P3errno == P3E_NOERROR)) {
	    P3errno = P3E_SERVER;
	}
    }
    if (P3errno == P3E_NOERROR) {
	if (! WMGetField("LI", varbuf, sizeof(varbuf))) {
	    P3errno = P3E_BUFFER;
	    (void) ERWrite(WMerrno, "P3List.GenerateList", "variable LI");
	} else {
	    SaveList(p3_conn->p3cn_p3_handle, DBname, varbuf);
	}
    }

    (void) WMSwitchContext(prev_context);
} /* GenerateList() */

/*
 * The list-algorithm is copied and adapted from DBList(). Changes here
 * might reflect changes necessary there too and vice versa.
 */

/*
 * The variable do_list is initialised on every call to DoList(). It can
 * be changed by must_list().
 */
static bool do_list;

/*------------------------------------------------------------------------
 *  Function	: must_list
 *  Purpose	: indicate if an entry should be listed
 *  Method	: the variable do_list (initialised by DBList()) can switch
 *		: to on if name equal to list in previous DBList() call
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *
 *  Notes	: May change variable do_list
 *
 *  Returns	: Indication of whether name should be listed.
 *------------------------------------------------------------------------
 */
static bool
must_list(
	const char	*n_found,	/* The name just found in the map */
	const char	*n_last)	/* last name in previous list */
{
    /* If we were listing, we keep on listing */
    if (do_list) return TRUE;

    /*
     * If the names are equal, we have already listed it, but the next one
     * must be listed.
     */
    if (STREQ(n_found, n_last)) { do_list = TRUE; }

    return FALSE;
} /* must_list() */

/*------------------------------------------------------------------------
 *  Function	: DoList
 *  Purpose	: Generate a list of systems and databases.
 *  Method	: Use the maps generated by SaveList()
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *
 *  Returns	: The number of id's found.
 *------------------------------------------------------------------------
 */
static int
DoList(
	P3Handle_p	P3handle,	/* Handle to the Pica3-server */
	const char	*DBname,	/* the name to expand */
	const char	*DBlast,	/* last of previous expansion */
	int		DBcount,	/* number of names to return */
	CBBuf_t		*CBbuf,		/* CB to return expansions in */
	size_t		CBsize)		/* size of the CB */
{
    int		count = 0;
    int		idx;
    int		subidx;
    CS_LOCAL(	variable, ADB_KEYLENGTH_MAX);
    const char	*name_p;

    do_list = (DBlast == NULL);
    for (idx = 0; count < DBcount && P3errno == P3E_NOERROR; idx++) {
	(void) CSPrintf(variable,"map[%d].from", idx);

	name_p = P3C_GetFromConnection(P3handle, CSContents(variable));
	if (name_p == NULL) {
	    if ((count == 0)
		    && (ADBerrno != ADBE_NOTPRESENT)) {
		ERWrite(ADBerrno, "P3List.DoList.ADBGet", "%s", CSContents(variable));
		/* ignore the error */
	    }
	    return count;
	}
	if ((P3errno == P3E_NOERROR) && STREQ(name_p, DBname)) {
	    for (subidx = 0;
		    count < DBcount && P3errno == P3E_NOERROR; subidx++) {

		(void) CSPrintf(variable,"map[%d].to[%d]", idx, subidx);

		name_p = P3C_GetFromConnection(P3handle, CSContents(variable));

		if (name_p != NULL) {
		    if (must_list(name_p, DBlast)) {
			if (! CBAppElement(CBbuf, CBsize,
				    (voidp) name_p, strlen(name_p) + 1)) {
			    P3errno = P3E_BUFFER;
			} else count ++;
		    }
		} else {
		    break;	/* for (subidx ...) */
		}
	    } /* for (subidx = 0; count < DBcount && P3errno ... ; ...) */
	} /* if (name_to_map == map[...].from) */
    } /* for (idx = 0; count < DBcount && P3errno  ...; ...) */
    if ((P3errno == P3E_BUFFER) && (count > 0)) {
	P3errno = P3E_NOERROR;
    }
    return count;
} /* DoList() */

/*------------------------------------------------------------------------
 *  Function	: P3List
 *  Purpose	: Perform the Pica3-part of DBList().
 *  Method	: - Check if there is already a list.
 *		  - If not generate it
 *		  - Store (a part of) the list into ADB
 *  Author(s)	: Hans van Eck
 *  Creation	: 05-10-1994
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Description	:
 *  - P3Handle must be a valid P3Handle_p previously returned by P3Open() (and
 *    not already closed).
 *  - DBname must be either an empty string or a (pointer to a) name previously
 *    returned by a call to P3List().
 *  - DBlast must be NULL or a name returned by a previous call to P3List() with
 *    the same DBname argument. A NULL indicates that it is the first call.
 *  - Options must be NULL or a pointer to a string containing (';'-separated)
 *    options for the P3List(). At the moment no options are defined yet.
 *  - DBcount is the maximum number of names put in the buffer.
 *  - CBbuf is a counted buffer into which the expansions are to be returned.
 *  - CBsize is the size of the counted buffer.
 *
 *  This function will expand a name to more specific names. This means that
 *  domain-names are expanded to subdomain-, server-, (pica3) system- and
 *  database-names. The list of names is returned in a counted buffer
 *  (consisting of '\0'-terminated strings). The return value is the number of
 *  names in the buffer. The number of names returned will never exceed
 *  DBcount. When the buffer is not large enough to hold DBcount expansions,
 *  P3List() should be called again, with DBlast pointing to the last expanded
 *  name of the previous expansion. The order in which the names are returned
 *  is not determined.
 *
 *  When mapping was unsuccessful, the function will return 0, and DBerrno
 *  will contain the error number.
 *
 *  Error values:
 *
 *	P3E_BUFFER	(buffer overflow)
 *	P3E_DISCONNECT	(server disconnected)
 *	P3E_NOERROR	(ok)
 *	P3E_NOMEM	(not enough memory)
 *	P3E_NOSUPPORT	(function not supported)
 *	P3E_PROTOCOL	(protocol error)
 *	P3E_SERVER	(server error)
 *
 *  Notes	:
 *  The real work is done elsewhere. The algorithm is analogous to the one
 *  in DBList(). Generation of the list is done on a per-connection basis.
 *  This function may only be used after a successful P3Logon() (because the
 *  retrieved list may be user dependent).
 *
 *  Returns	: number of elements stored into the list
 *------------------------------------------------------------------------
 */
int
P3List(
	P3Handle_p	P3handle,	/* Handle to the connection */
	const char	*DBname,	/* the name to expand */
	const char	*DBlast,	/* last of previous expansion */
	const char	*Options,	/* options for the list */
	int		DBcount,	/* number of names to return */
	CBBuf_t		*CBbuf,		/* CB to return expansions in */
	size_t		CBsize)		/* size of the CB */
{
    p3_connection_p	p3_conn = P3C_GetConnection(P3handle);
    int			count = 0;

    P3errno = P3E_NOERROR;
    if (p3_conn == NULL) {
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3List", "invalid handle");
	return 0;
    }
    if (! InList(P3handle, DBname)) {
	GenerateList(p3_conn, DBname, Options);
    }
    if (InList(P3handle, DBname)) {
	count = DoList(P3handle, DBname, DBlast, DBcount, CBbuf, CBsize);
    }
    return count;
} /* P3List() */
