/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3local.c
 *  Description	: Some local stuff for the Pica3-library
 *  Author(s)	: Hans van Eck
 *  Creation	: 06-09-1994
 *
 *  History:
 *  06-09-1994	: eck      : creation
 *  08-09-1995	: eck      : Added P3L_WMelt2CB()
 *  30-08-1996	: eck      : Added specials for editscript-fields
 *  30-10-1997	: HSi	   : Added P3L_SyntaxAdbToP3
 *  08-09-1998	: eck      : Added FLG_UPDATE support in P3L_WMelt2CB()
 *  04-11-1999  : EvD      : Added UNX and USX to syntax table at
 *                           P3L_SyntaxAdbToP3
 *  24-01-2000  : MT       : Added GD to syntax table at P3L_SyntaxAdbToP3
 *  19-06-2000  : MT       : Added P3GetCatDiagnostic
 *  08-09-2000  : MT       : Added GA to syntax table at P3L_SyntaxAdbToP3
 *  11-03-2002  : evd      : bug at P3L_XLWMSendRecv NIORecv till '\003'
 *                           and No check on a full recvbuf
 *  14-03-2002  : evd      : P3errno initialized after a wrong NIORecv at
 *                           P3L_XLWMSendRecv
 *  20-06-2002  : MT       : Added DA to syntax table at P3L_SyntaxAdbToP3
 *  06-08-2002  : MT       : Added P3L_cnvZeroStrChar
 *  19-10-2006  : EvD      : Added pica$ to syntax table at P3L_SyntaxAdbToP3
 *
 *========================================================================
 */

#pragma ident "$Id: p3local.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <string.h>
#include <stdlib.h>

#include <pica.h>
#include <picamisc.h>
#include <cs.h>
#include <p3.h>
#include <er.h>

#include "p3local.h"

P3Status_t P3status = P3_OK;

static struct stat_t {
    const char	*serverstatus;
    P3Status_t	p3status;
} status_tab[] = {
	{ "OK",		P3_OK },	/* request successfully handled */
	{ "NOTSUP",	P3_NOTSUP },	/* not supported */
	{ "NOAUTH",	P3_NOAUTH },	/* no authority */
	{ "NOTNOW",	P3_NOTNOW },	/* request not allowed yet */
	{ "REJECT",	P3_REJECT },	/* logon rejected */
	{ "NOINFO",	P3_NOINFO },	/* no information available */
	{ "NONODE",	P3_NONODE },	/* no node specified */
	{ "NOTDBS",	P3_NOTDBS },	/* no real database specified */
	{ "MAXSET",	P3_MAXSET },	/* set overflow */
	{ "NOHITS",	P3_NOHITS },	/* nothing found */
	{ "NOLIBS",	P3_NOLIBS },	/* no supplying libraries */
	{ "OWNLIB",	P3_OWNLIB },	/* title part is in own library */
	{ "ERRSET",	P3_ERRSET },	/* invalid set */
	{ "ERRTTL",	P3_ERRTTL },	/* invalid title(s) */
	{ "ERRCMD",	P3_ERRCMD },	/* invalid command */
	{ "ERRFLD",	P3_ERRFLD },	/* invalid field */
	{ "EMAIL",	P3_EMAIL },	/* invalid email address */
	{ "ERROR",	P3_ERROR },	/* server error */
	{ "MAXCHN",	P3_MAXCHN },	/* maximum nr. of channels exceeded */
	{ "MRINFO",	P3_MRINFO },	/* more information needed */
	{ "STPWRD",	P3_STPWRD },	/* search on stopword(s) only */
	{ "LIBDEP",	P3_LIBDEP },	/* library deposit exhausted */
	{ "USRDEP",	P3_USRDEP },	/* user deposit exhausted */
	{ "ERRNUM",	P3_ERRNUM },	/* invalid number */

	{ NULL,		P3_UNKNOWN }	/* unknown status returned by server */
    };

/*------------------------------------------------------------------------
 *  Function	: GetP3Status
 *  Purpose	: Set the global Pica3-status variabl
 *  Method	: Scan the table for the status received.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: The status found (or P3_UNKNOWN).
 *------------------------------------------------------------------------
 */
static P3Status_t
GetP3Status(void)
{
    char		varbuf[80];
    struct stat_t	*statusp;

    if (! WMGetVar(P3_VAR_ERROR, varbuf, sizeof(varbuf))) {
	P3errno = P3E_PROTOCOL;
	(void) ERWrite(P3errno, __func__, "Did not receive Pica3-status");
	return P3_UNKNOWN;
    }
    (void) WMDelVar(P3_VAR_ERROR);	/* So it's gone next time around */

    statusp = status_tab;

    while ((statusp->serverstatus != NULL)
	    && (! STREQ(varbuf, statusp->serverstatus))) {
	statusp++;
    }
    if (statusp->serverstatus == NULL) {
	CS_LOCAL(	message, 80);

	P3errno = P3E_PROTOCOL;

	CSPrintf(message, "Unknown status \"%s\"", varbuf);
	(void) ERWrite(P3errno, __func__, "%s", CSContents(message));
    }
    return statusp->p3status;
} /* GetP3Status() */

/*------------------------------------------------------------------------
 *  Function	: P3GetCatDiagnostic
 *  Purpose	: Get the global Pica3 cataloguing diagnostic variable
 *  Method	: Scan the table for the cataloguing diagnostic received.
 *
 *  History:
 *  19-06-2000	: martin   : creation
 *
 *  Returns	: The diagnostic found.
 *------------------------------------------------------------------------
 */
int
P3GetCatDiagnostic(void)
{
    char		varbuf[80];

    if (! WMGetVar(P3_VAR_CATSTATUS, varbuf, sizeof(varbuf))) {
	return 0;
    }
    (void) WMDelVar(P3_VAR_CATSTATUS);	/* So it's gone next time around */

    return atoi(varbuf);
} /* P3GetCatDiagnostic() */

static void
LogMessages(void)
{
    char	message[200];
    int		i = 0;
    WMIdent_t	id;

    while (WMGetMsg(i, id, message, sizeof(message))) {
	if (STREQ(id, "01")) {
	    (void) ERWrite(id, "P3L_SendRecv.LogMessages", "%s", message);
	} else if (STREQ(id, "02")) {
	    (void) ERWarning(id, "P3L_SendRecv.LogMessages", "%s", message);
	} else if (STREQ(id, "03")) {
	    (void) ERTrace(id, "P3L_SendRecv.LogMessages", "%s", message);
	}
	i++;
    }
} /* LogMessages() */

/*------------------------------------------------------------------------
 *  Function	: P3L_SendRecv
 *  Purpose	: Send-to and receive-from Pica3-server.
 *  Method	:
 *  Build worm from current context and pass it to FESend().
 *  Store incoming worm into context.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3L_SendRecv(FEHandle_t	fe_handle)
{
    char	worm[MAXLEN];

    if (! WMBuild(worm)) {
	/* Don't know what else */
	P3errno = P3E_NOMEM;
    }
    if (P3errno == P3E_NOERROR) {
	if ((FESend(fe_handle, worm, strlen(worm)) == FEDisconnected)
	    || (FEReceive(fe_handle, worm, sizeof(worm)) == FEDisconnected)) {
	    P3errno = P3E_DISCONNECT;
	}
    }
    if (P3errno == P3E_NOERROR) {
	WMClearCurrentContext();
	WMDissect(worm);
	P3status = GetP3Status();
	LogMessages();
    }
    return (P3errno == P3E_NOERROR);
} /* P3L_SendRecv() */

/*------------------------------------------------------------------------
 *  Function	: P3L_InternToPica3Name
 *  Purpose	: Convert a name used internally to one that the server
 *		  understands.
 *  Method	:
 *  Delete the server part from an id. If there is no server part, the
 *  whole id. is taken to be a server.
 *
 *  History:
 *  05-10-1994	: eck      : creation
 *
 *  Notes	:
 *
 *	"fil.sys@server.domain"	=>	"fil.sys"
 *	"server.domain"		=>	""
 *------------------------------------------------------------------------
 */
void
P3L_InternToPica3Name(
	const char	*Iname,		/* Internal name */
	CSBuf_p 	Oname)		/* Resulting name for server */
{
    const char	*tmp;
    int		len;

    if (CSSetString(Oname, Iname)) {
	tmp = strchr(CSContents(Oname), '@');
	if (tmp != NULL) {
	    len = tmp - CSContents(Oname);
	    CSDeletePart(Oname, len, CSLength(Oname) - len);
	} else {
	    CSClear(Oname);
	}
    }
} /* P3L_InternToPica3Name() */

static struct cacheelm_t {
    char		P3handle[PM_SYMLEN_MIN];
    P3SetIdent_t	SetIdent;
    long		Teltno;
    char		recidt[16];
    char		rectyp[16];
    char		reltyp[16];
} cacheelm = { "", "", 0, "", "" , "" };

void
P3L_ToCache(
	P3Handle_p		P3handle,
	const P3SetIdent_t	SetIdent,
	long			Teltno,
	const char		*option,
	const char		*value)
{

    if ((! STREQ(P3handle, cacheelm.P3handle))
	|| (! STREQ(SetIdent, cacheelm.SetIdent))
	|| (Teltno != cacheelm.Teltno)) {

	(void) strcpy(cacheelm.P3handle, P3handle);
	(void) strcpy(cacheelm.SetIdent, SetIdent);
	cacheelm.Teltno = Teltno;
	cacheelm.recidt[0] = '\0';
	cacheelm.rectyp[0] = '\0';
	cacheelm.reltyp[0] = '\0';
    }
    if (STREQ(option, "recidt")) {
	(void) strcpy(cacheelm.recidt, value);
    } else if (STREQ(option, "rectyp")) {
	(void) strcpy(cacheelm.rectyp, value);
    } else if (STREQ(option, "reltyp")) {
	(void) strcpy(cacheelm.reltyp, value);
    }
} /* P3L_ToCache() */

bool
P3L_FromCache(
	P3Handle_p		P3handle,
	const P3SetIdent_t	SetIdent,
	long			Teltno,
	const char		*option,
	CBBuf_t			*CBbuf,
	size_t			CBsize)
{
    if ((! STREQ(P3handle, cacheelm.P3handle))
	|| (! STREQ(SetIdent, cacheelm.SetIdent))
	|| (Teltno != cacheelm.Teltno)) {
	return FALSE;
    }
    if (STREQ(option, "recidt")) {
	if (cacheelm.recidt[0] != '\0') {
	    if (! CBAppElement(CBbuf, CBsize,
			    cacheelm.recidt, strlen(cacheelm.recidt) + 1)) {
		P3errno = P3E_BUFFER;
	    }
	}
    } else if (STREQ(option, "rectyp")) {
	if (cacheelm.rectyp[0] != '\0') {
	    if (! CBAppElement(CBbuf, CBsize,
			    cacheelm.rectyp, strlen(cacheelm.rectyp) + 1)) {
		P3errno = P3E_BUFFER;
	    }
	}
    } else if (STREQ(option, "reltyp")) {
	/* Allow empty reltyp. */
	if (! CBAppElement(CBbuf, CBsize,
			cacheelm.reltyp, strlen(cacheelm.reltyp) + 1)) {
	    P3errno = P3E_BUFFER;
	}
    } else {
	P3errno = P3E_NOINFO;
	(void) ERWrite(P3errno, "P3L_FromCache", "option = \"%s\"", option);
	return FALSE;
    }
    if (P3errno == P3E_BUFFER) {
	(void) ERWrite(P3errno, "P3L_FromCache", "option = \"%s\"", option);
    }
    return TRUE;
} /* P3L_FromCache() */

/*------------------------------------------------------------------------
 *  Function	: P3L_InsertLanguageCommand
 *  Purpose	: Insert a language-command into the communication-buffer
 *  Method	: obvious
 *
 *  History:
 *  17-05-1995	: eck      : creation
 *------------------------------------------------------------------------
 */
void
P3L_InsertLanguageCommand(
	CSBuf_p 	command,
	const char	*options)
{
    char	lang_buff[10];
    int		i;
    struct {
	const char	*lang;
	const char	*comm;
    }		languages[] = { { "NE", "\\ned ;" },
				{ "EN", "\\eng ;" },
				{ "DU", "\\deu ;" },
				{ "FR", "\\fra ;" },
			    };

    lang_buff[0] = '\0';

    options = PM_opt_findstart(options);
    while ((options != NULL) && (*options != '\0')) {
	if (PM_opt_equal(options, STR_LANGUAGE)) {
	    options = PM_opt_findval(options);
	    PM_opt_copyval(options, lang_buff, sizeof(lang_buff));
	}
	options = PM_opt_findnext(options);
    }
    if (lang_buff[0] != '\0') {
	for (i = 0; i < (sizeof(languages)/sizeof(languages[0])); i++) {
	    if (STREQ(languages[i].lang, lang_buff)) {
		(void) CSInsert(command, 0, languages[i].comm);
		break;
	    }
	}
    }
} /* P3L_InsertLanguageCommand() */

/*------------------------------------------------------------------------
 *  Function	: P3L_SwitchAndClearContext
 *  Purpose	: Switch to context and clear it (so worm is clean)
 *  Method	: obvious
 *
 *  History:
 *  21-07-1995	: eck      : creation
 *
 *  Returns	: Handle to previous context.
 *------------------------------------------------------------------------
 */
WMContext_t
P3L_SwitchAndClearContext(
	WMContext_t	handle)
{
    WMContext_t	ret_handle = WMSwitchContext(handle);

    if ((ret_handle != WM_NOCONTEXT) || (WMerrno != WME_ILLEGALCONTEXT)) {
	WMClearCurrentContext();
    }
    return ret_handle;
} /* P3L_SwitchAndClearContext() */


/* A #define that gives the start of an editscript. */
#define SCRIPTSTART	"\033H\033P\033%\033V"

/*------------------------------------------------------------------------
 *  Function	: GetLine
 *  Purpose	: Dissect info from a <cr> terminated string
 *  Method	: obvious
 *
 *  History:
 *  08-09-1995	: eck      : creation
 *  27-09-1996	: mespel   : Fixed OBI-WAN error for multiple edit-scripts
 *
 *  Notes	: This routine modifies the buffer
 *
 *  Returns	: A pointer to the start of the next line.
 *------------------------------------------------------------------------
 */
static const char *
GetLine(
	char	**bufp,		/* (part of) buffer sent by server */
	bool	editscript)	/* If we are dealing with an editscript */
{
    const char	*linep;		/* line to return (== start of buffer) */
    char	*endp;		/* ptr to en of line */

    if (**bufp == '\0') return NULL;

    if (! editscript) {
	endp = strchr(*bufp, '\r');
    } else {
	/* Look for the start of the next script. */
	endp = strstr(*bufp+1, SCRIPTSTART);
	if (endp != NULL) {
	    endp -= 1;		/* Go to the '\r' */
	} else {
	    /* There is no next part. Assume we have the last one. */
	    endp = *bufp + strlen(*bufp) - 1;
	}
    }
    if (endp == NULL) return NULL;

    *endp = '\0';
    linep = *bufp;
    *bufp = endp + 1;

    return linep;
} /* GetLine() */

static bool
IsEditScript(
	const char	*bufp)
{
    /* An editscript-field begins with SCRIPSTART. Compare over its
     * length (== sizeof() - 1).
     */
    return strncmp(bufp, SCRIPTSTART, sizeof(SCRIPTSTART) - 1) == 0;
} /* IsEditScript() */
#undef SCRIPTSTART

/*------------------------------------------------------------------------
 *  Function	: P3L_WMelt2CB
 *  Purpose	: Get an element from the WM-context and store it in a CB
 *  Method	: Basically WMGet*() followed by CBApp*()
 *
 *  History:
 *  08-09-1995	: eck      : creation
 *  08-09-1998	: eck      : Added support for FLG_UPDATE
 *
 *  Notes	:
 *  The options parameter contains flags which determine the action to take.
 *  Possibble value are:
 *
 *	FLG_FIELD	- retrieve a field
 *	FLG_VAR		- retrieve a var
 *	FLG_OPTIONAL	- the field/var need not be present
 *	FLG_SPLTLINE	- store lines of the field/var separately
 *	FLG_ADDID	- use CBAppAttribute() instead of CAppElement()
 *	FLG_UPDATE	- append to (first) element in the buffer
 *
 *  FLG_FIELD or FLG_VAR must be present.
 *
 *  Returns	: The number of elements stored
 *------------------------------------------------------------------------
 */
int
P3L_WMelt2CB(
	const WMIdent_t	id,			/* Ident to get */
	CBBuf_t		*CBbuf,			/* result buffer */
	size_t		CBsize,			/* size of buffer */
	uint16		options,		/* option-flags */
	const		char *location)		/* in case of errors */
{
#define TITLEBUF_SIZE (32 * 1024)     /* 32K-bytes for a title-buffer */

    char		hlpbuf[TITLEBUF_SIZE];
    char		varbuf[MAXLEN];
    char		*bufp = varbuf;
    char		*CBelm;
    const char		*value;
    const char		*err_comment = NULL;
    bool		ok = FALSE;
    int			count = 0;

    if (options & FLG_FIELD) {
	ok = WMGetField(id, varbuf, sizeof(varbuf));
    } else if (options & FLG_VAR) {
	ok = WMGetVar(id, varbuf, sizeof(varbuf));
    } else {
	P3errno = P3E_FLDWRONG;	/* after all, a bit-field is wrong */
	(void) ERAbort(P3errno, "P3L_WMelt2CB",
			"retrieving unknown type (from %s)", location);
	/* we never get here! */
    }
    if (! ok) {
	if (options & FLG_OPTIONAL) {
	    return 0;
	} else {
	    P3errno = P3E_PROTOCOL;
	    if (options & FLG_FIELD) err_comment = "retrieving field %s";
	    else err_comment = "retrieving variable %s";
	}
    } else {
	bool	editscript_field;

	if (options & FLG_FIELD) {
	    editscript_field = IsEditScript(bufp);
	} else {
	    editscript_field = FALSE;
	}

	/* The loop is only repeated when the split-line option is set.
	 * This is enforced by the initialisation of linep.
	 */
	do {
	    if (options & FLG_SPLITLINE) {
		value = GetLine(&bufp, editscript_field);
	    } else {
		value = varbuf;
	    }
	    if (value != NULL) {
		if (options & FLG_ADDID) {
		    if (options & FLG_UPDATE) {
			if (! (ok = CBAppToAttribute(CBbuf, CBsize, id, value))) {
			    err_comment = "appending attribute %s";
			} else count++;
		    } else {
			if (! (ok = CBAppAttribute(CBbuf, CBsize, id, value))) {
			    err_comment = "appending attribute %s";
			} else count++;
		    }
		} else {
		    CBelm = CBGetElement(CBbuf, CBsize, 0);
		    if ((options & FLG_UPDATE) && (CBelm != NULL)) {
			size_t	len = CBElementLength(CBelm);
			size_t	vallen = strlen(value);

			if (len >= sizeof(hlpbuf)) {
			    (void) ERTrace(P3E_BUFFER, location,
					"internal buffer length exceeded");
			    return 0;
			}
			(void) strncpy(hlpbuf, CBelm, len);
			if (vallen + len > sizeof(hlpbuf)) {
			    vallen = sizeof(hlpbuf) - len;
			}
			(void) strncpy(hlpbuf + len - 1, value, vallen);
			hlpbuf[len + vallen - 1] = '\0';

			if (! (ok = CBUpdElement(CBbuf, CBsize, CBelm,
					(voidp) hlpbuf, len + vallen))) {
			    err_comment = "updating for element %s";
			} else count++;
		    } else {
			if (! (ok = CBAppElement(CBbuf, CBsize,
				    (voidp) value, strlen(value) + 1))) {
			    err_comment = "appending for element %s";
			} else count++;
		    }
		}
	    } else {
		/* We now have an empty part of an element. Don't write
		 * an error if this is the first part and the element is
		 * optional.
		 */
		if ((count != 0) || ((options & FLG_OPTIONAL) == 0)) {
		    (void) ERWrite(P3E_PROTOCOL, "P3L_WMelt2CB",
				    "element %s (%s) not properly terminated!",
				    id, location);
		}
		break;	/* while */
	    }
	} while ((options & FLG_SPLITLINE) && (*bufp != '\0'));
	if (! ok) P3errno = P3E_BUFFER;
    } /* else (ok == TRUE) */

    if (! ok) {
	(void) ERWrite(P3errno, location, err_comment, id);
    } else if (count == 0) {
	/* Hmm, GetLine() must have gone wrong */
	if (options & FLG_OPTIONAL) return 0;
	P3errno = P3E_BUFFER;
	(void) ERWrite(P3errno, location, "empty element %s", id);
    }
    return count;
} /* P3L_WMelt2CB() */

/*------------------------------------------------------------------------
 *  Function	: P3L_SyntaxAdbToP3
 *  Purpose	: Convert syntax from adb format to P3-protocol format.
 *  Method	: obvious
 *
 *  History:
 *  30-10-1997	: HSi	: creation
 *  04-11-1999  : EvD	: syntaxTable extended with UNX and USX
 *  24-01-2000  : MT	: syntaxTable extended with GD (GatewayDiagnostic)
 *  08-09-2000  : MT	: syntaxTable extended with GA (GatewayAll)
 *  20-06-2002  : MT	: syntaxTable extended with DA (GatewayDiagnosticAll)
 *  19-10-2006  : EvD	: syntaxTable extended with pica$
 *  23-08-2007	: clemens  : use adbSyntax as default
 *
 *  Returns	:
 *------------------------------------------------------------------------
 */
const char *
P3L_SyntaxAdbToP3 (
	const char	 *adbSyntax
	)
{
    static struct {
	const char	*adb;
	const char	*p3;
    } syntaxTable[] = {
	{ "pica$",	"P$" },
	{ "pica+",	"P" },
	{ "pica2",	"D" },
	{ "isbd",	"I" },
	{ "gateway",	"G" },
	{ "gatewaya",	"GA" },
	{ "gatewayd",	"GD" },
	{ "gatewayda",	"GDA" },
	{ "download",	"J" },
	{ "labelled",	"U" },
	{ "unimarc",	"UNM" },
	{ "usmarc",	"USM" },
	{ "mab",	"MAB" },
	{ "unimarc_exchange",	"UNX" },
	{ "usmarc_exchange",	"USX" }
    };
    int			 i;
    const char		*p3Syntax = "";

    if ( adbSyntax != NULL ) {
	for ( i = 0;
	      i < sizeof(syntaxTable) / sizeof(syntaxTable[0]);
	      i++ ) {
	    if ( STREQ ( syntaxTable[i].adb, adbSyntax ) ) {
		p3Syntax = syntaxTable[i].p3;
		break;
	    }
	}
	if (*p3Syntax == '\0') {
	    p3Syntax = adbSyntax;
	    ERWarning(APPLE_PARAM, "P3L_SyntaxAdbToP3", "Could not convert syntax \"%s\"", adbSyntax);
	}
    }

    return p3Syntax;
} /* P3L_SyntaxAdbToP3 () */

/*------------------------------------------------------------------------
 *  Function	: P3L_XLWMelt2CB
 *  Purpose	: Get an element from the WM-context and store it in a CB
 *  Method	: Basically WMGet*() followed by CBApp*()
 *
 *  History:
 *  13-12-2001	: evd      : Copied from P3L_WMelt2CB 
 *
 *  Notes	: This routine handles worms longer then 4k
 *  The options parameter contains flags which determine the action to take.
 *  Possibble value are:
 *
 *	FLG_FIELD	- retrieve a field
 *	FLG_VAR		- retrieve a var
 *	FLG_OPTIONAL	- the field/var need not be present
 *	FLG_SPLTLINE	- store lines of the field/var separately
 *	FLG_ADDID	- use CBAppAttribute() instead of CAppElement()
 *	FLG_UPDATE	- append to (first) element in the buffer
 *
 *  FLG_FIELD or FLG_VAR must be present.
 *
 *  Returns	: The number of elements stored
 *------------------------------------------------------------------------
 */
int
P3L_XLWMelt2CB(
	const WMIdent_t	id,			/* Ident to get */
	CBBuf_t		*CBbuf,			/* result buffer */
	size_t		CBsize,			/* size of buffer */
	uint16		options,		/* option-flags */
	const		char *location)		/* in case of errors */
{
#define TITLEBUF_SIZE (32 * 1024)     /* 32K-bytes for a title-buffer */

    char		*hlpbuf;
    CS_DYNAMIC(csvarbuf);
    char		*bufp;
    char		*CBelm;
    const char		*value;
    const char		*err_comment = NULL;
    bool		ok = FALSE;
    int			count = 0;

    if (options & FLG_FIELD) {
	ok = WMGetFieldCS(id, csvarbuf);
    } else if (options & FLG_VAR) {
	ok = WMGetVarCS(id, csvarbuf);
    } else {
	P3errno = P3E_FLDWRONG;	/* after all, a bit-field is wrong */
	(void) ERAbort(P3errno, "P3L_XLWMelt2CB",
			"retrieving unknown type (from %s)", location);
	/* we never get here! */
    }
    bufp = (char *)CSContents(csvarbuf);
    if (! ok) {
	if (options & FLG_OPTIONAL) {
	    CSDestroy(csvarbuf);
	    return 0;
	} else {
	    P3errno = P3E_PROTOCOL;
	    if (options & FLG_FIELD) err_comment = "retrieving field %s";
	    else err_comment = "retrieving variable %s";
	}
    } else {
	bool	editscript_field;

	if (options & FLG_FIELD) {
	    editscript_field = IsEditScript(CSContents(csvarbuf));
	} else {
	    editscript_field = FALSE;
	}

	/* The loop is only repeated when the split-line option is set.
	 * This is enforced by the initialisation of linep.
	 */
	do {
	    if (options & FLG_SPLITLINE) {
		value = GetLine(&bufp, editscript_field);
	    } else {
		value = CSContents(csvarbuf);
	    }
	    if ((value != NULL) && (strlen(value) > 0)){
		if (options & FLG_ADDID) {
		    if (options & FLG_UPDATE) {
			if (! (ok = CBAppToAttribute(CBbuf, CBsize, id, value))) {
			    err_comment = "appending attribute %s";
			} else count++;
		    } else {
			if (! (ok = CBAppAttribute(CBbuf, CBsize, id, value))) {
			    err_comment = "appending attribute %s";
			} else count++;
		    }
		} else {
		    CBelm = CBGetElement(CBbuf, CBsize, 0);
		    if ((options & FLG_UPDATE) && (CBelm != NULL)) {
			size_t	len = CBElementLength(CBelm);
			size_t	vallen = strlen(value);
			hlpbuf = (char *)malloc(vallen + len + 20);
			if (hlpbuf == NULL) {
			    ERWrite(P3E_BUFFER, location,
			    "malloc  internal buffer length exceeded");
			    CSDestroy(csvarbuf);
			    return 0;
			}
			(void) strncpy(hlpbuf, CBelm, len);
			(void) strncpy(hlpbuf + len - 1, value, vallen);
			hlpbuf[len + vallen - 1] = '\0';

			if (! (ok = CBUpdElement(CBbuf, CBsize, CBelm,
					(voidp) hlpbuf, len + vallen))) {
			    err_comment = "updating for element %s";
			} else count++;
			free(hlpbuf);
		    } else {
			if (! (ok = CBAppElement(CBbuf, CBsize,
				    (voidp) value, strlen(value) + 1))) {
			    err_comment = "appending for element %s";
			} else count++;
		    }
		}
	    } else {
		/* We now have an empty part of an element. Don't write
		 * an error if this is the first part and the element is
		 * optional.
		 */
		if ((count != 0) || ((options & FLG_OPTIONAL) == 0)) {
		    (void) ERWrite(P3E_PROTOCOL, "P3L_XLWMelt2CB",
				    "element %s (%s) not properly terminated!",
				    id, location);
		}
		break;	/* while */
	    }
	} while ((options & FLG_SPLITLINE) && (*bufp != '\0'));
	if (! ok) P3errno = P3E_BUFFER;
    } /* else (ok == TRUE) */

    if (! ok) {
	(void) ERWrite(P3errno, location, err_comment, id);
    } else if (count == 0) {
	/* Hmm, GetLine() must have gone wrong */
	if (options & FLG_OPTIONAL){
	    CSDestroy(csvarbuf);
	    return 0;
	}
	P3errno = P3E_BUFFER;
	(void) ERWrite(P3errno, location, "empty element %s", id);
    }
    CSDestroy(csvarbuf);
    return count;
} /* P3L_XLWMelt2CB() */

/*------------------------------------------------------------------------
 *  Function	: XLWMGetP3Status
 *  Purpose	: Set the global Pica3-status variabl
 *  Method	: Scan the table for the status received.
 *
 *  History:
 *  17-12-2001	: evd      : copied from GetP3Status
 *
 *  Returns	: The status found (or P3_UNKNOWN).
 *------------------------------------------------------------------------
 */
static P3Status_t
XLWMGetP3Status(void)
{
    CS_DYNAMIC(varbuf);
    struct stat_t	*statusp;

    if (! WMGetVarCS(P3_VAR_ERROR, varbuf)) {
	P3errno = P3E_PROTOCOL;
	(void) ERWrite(P3errno, __func__, "Did not receive Pica3-status");
	CSDestroy(varbuf);
	return P3_UNKNOWN;
    }
    (void) WMDelVar(P3_VAR_ERROR);	/* So it's gone next time around */

    statusp = status_tab;

    while ((statusp->serverstatus != NULL)
	    && (! STREQ(CSContents(varbuf), statusp->serverstatus))) {
	statusp++;
    }
    if (statusp->serverstatus == NULL) {
	CS_LOCAL(	message, 80);

	P3errno = P3E_PROTOCOL;

	CSPrintf(message, "Unknown status \"%s\"", CSContents(varbuf));
	(void) ERWrite(P3errno, __func__, "%s", CSContents(message));
    }
    CSDestroy(varbuf);
    return statusp->p3status;
} /* XLWMGetP3Status() */

static void
XLWMLogMessages(void)
{
    CS_DYNAMIC(message);
    int		i = 0;
    WMIdent_t	id;

    (void)CSClear(message);
    while (WMGetMsgNCS(i, id, message)) {
	if (STREQ(id, "01")) {
	    (void) ERWrite(id, "P3L_XLWMSendRecv.XLWMLogMessages", "%s", CSContents(message));
	} else if (STREQ(id, "02")) {
	    (void) ERWarning(id, "P3L_XLWMSendRecv.XLWMLogMessages", "%s", CSContents(message));
	} else if (STREQ(id, "03")) {
	    (void) ERTrace(id, "P3L_XLWMSendRecv.XLWMLogMessages", "%s", CSContents(message));
	}
	i++;
	(void)CSClear(message);
    }
    (void)CSDestroy(message);
} /* LogMessages() */

/*------------------------------------------------------------------------
 *  Function	: P3L_XLWMSendRecv
 *  Purpose	: Send-to and receive-from Pica3-server.
 *  Method	:
 *  Build worm from current context and pass it to NIOSend().
 *  Store incoming worm into context.
 *
 *  History:
 *  17-12-2001  : evd     : Copied from P3L_SendRecv
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3L_XLWMSendRecv(NIOHandle_t	nio_handle)
{
    CS_DYNAMIC(worm);
    char recvbuf[4096];
    int ir, tr = -2;

    if (! WMBuildCS(worm)) {
	/* Don't know what else */
	P3errno = P3E_NOMEM;
    }
    if (P3errno == P3E_NOERROR) {
        if (!NIOSend(nio_handle, (char *)CSContents(worm), CSLength(worm))){
            ERWrite(NIOerrno, "p3local.c/P3L_XLWMSendRecv/NIOSend",
	    "%s", CSContents(worm));  
	    P3errno = P3E_DISCONNECT;
	}
    }
    if (P3errno == P3E_NOERROR) {
	CSClear(worm);
	WMDelBuf();
	WMClearCurrentContext();
	tr = 0;
	do {
	    ir = NIORecv(nio_handle, recvbuf, 
	                  sizeof(recvbuf));
	    tr += ir;
	    if (ir < 0) {
		ERWrite(NIOerrno, "p3local.c/P3L_XLWMSendRecv/NIORecv",
	                "%d", tr);
		CSDestroy(worm);
		P3errno = P3E_SERVER;
		return FALSE;
	    } else if (ir > 0) {
		CSAppendN(worm, recvbuf, ir);
	    } else if (ir == 0) {
		ERWrite(NIOerrno, "p3local.c/P3L_XLWMSendRecv/NIORecv",
		"Number of received characters = 0. Total received %d", tr);
		CSDestroy(worm);
		P3errno = P3E_SERVER;
		return FALSE;
	    }
	} while (recvbuf[ir - 1] != '\003');
	WMClearCurrentContext();
	WMDissect((char *)CSContents(worm));
	P3status = XLWMGetP3Status();
	XLWMLogMessages();
    }
    CSDestroy(worm);
    return (P3errno == P3E_NOERROR);
} /* P3L_XLWMSendRecv() */

/*------------------------------------------------------------------------
 *  Function    : P3L_cnvZeroStrChar
 *  Purpose     : convert all occurrences of a character in a zero-terminated
 *                string to another character up to a given length.
 *  Method      :
 *  Author(s)   : MT
 *  Creation    : 06-08-2002
 *
 *  History:
 *  06-08-2002  : MT    : creation
 *
 *  Returns     : void
 *------------------------------------------------------------------------
 */
void
P3L_cnvZeroStrChar(
    char                                *string,
    size_t                              length,
    char                                searchChar,
    char                                replaceChar)
{
    char                                *stringEnd;

    stringEnd = string + length;

    if ( string != NULL ) {
        for ( ; *string != '\0' && string < stringEnd; string++ ) {
	    if ( *string == searchChar) {
		*string = replaceChar;
 	    }
	}
    }
} /* P3L_cnvZeroStrChar */
