/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3logon.c
 *  Description	: Log-on to a Pica3 server
 *  Author(s)	: Hans van Eck
 *  Creation	: 08-09-1994
 *
 *  History:
 *  08-09-1994	: eck      : creation
 *  18-12-2001  : evd      : Added P3XLWMLogon
 *  26-08-2004	: huug	   : Added characterset negotiation to P3XLWMLogon
 *  26-08-2004	: huug	   : Added characterset negotiation to P3XLWMLogon
 *  07-10-2005	: v_ouf    : removed include of si.h
 *  10-10-2013	: martinz  : Copied documentation from P3LIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: p3logon.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <env.h>
#include <p3.h>
#include <cs.h>

#include "p3local.h"

#define IDLEN_MAX	32
#define SEEDLEN_MAX	32

#define STR_USER	"user"
#define STR_PASSWORD	"password"
#define STR_CHARSET	"charset"

/*------------------------------------------------------------------------
 *  Function	: DetermineUserArgs
 *  Purpose	: Determine usernumber and password from options
 *  Method	: obvious
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *  14-01-2005	: eck      : Fixed charset assignment.
 *
 *  Returns	: TRUE if a uswer as specified, FALSE otherwise.
 *------------------------------------------------------------------------
 */
static bool
DetermineUserArgs(
	const char	*options,
	char		user[IDLEN_MAX + 1],
	char		password[IDLEN_MAX + 1],
	char		charset[IDLEN_MAX + 1])
{
    user[0] = '\0';
    password[0] = '\0';
    charset[0] = '\0';

    options = PM_opt_findstart(options);
    while ((options != NULL) && (*options != '\0')) {
	if (PM_opt_equal(options, STR_USER)) {
	    options = PM_opt_findval(options);
	    PM_opt_copyval(options, user, IDLEN_MAX + 1);
	} else if (PM_opt_equal(options, STR_PASSWORD)) {
	    options = PM_opt_findval(options);
	    PM_opt_copyval(options, password, IDLEN_MAX + 1);
	} else if (PM_opt_equal(options, STR_CHARSET)) {
	    options = PM_opt_findval(options);
	    PM_opt_copyval(options, charset, IDLEN_MAX + 1);
	}
	options = PM_opt_findnext(options);
    }
    return (user[0] != '\0');
} /* DetermineUserArgs() */

/*------------------------------------------------------------------------
 *  Function	: LoginScrambled
 *  Purpose	: Create the scranbled login in the WM-context.
 *  Method	:
 *  Set user number, scrambled password and scrambled client-type.
 *
 *  History:
 *  20-09-1996	: eck      : creation
 *------------------------------------------------------------------------
 */
static void
LoginScrambled(
	const char	*options,	/* To get client-type */
	const char	*user,		/* User to login with */
	const char	*password,	/* Password */
	const char	*seed)		/* Seed to scramble with */
{
    int		scr_val;		/* scrambled value */
    char	tmp_buffer[IDLEN_MAX + 1];

    WMPutCmd("PC", "\\log ;");
    WMPutBufid("00");
    WMPutField("UN", user);

    scr_val = PM_scramble(password, seed);
    WMPutField("WW", PM_itoa(NULL, scr_val));

    /* The PM_opt_fetch() will succeed, or we wouldn't be here. */
    tmp_buffer[0] = '\0';
    (void) PM_opt_fetch(options, "client-type", tmp_buffer, sizeof(tmp_buffer));

    scr_val = PM_scramble(tmp_buffer, seed);
    WMPutField("ID", PM_itoa(NULL, scr_val));
} /* DeterminScrambled() */

/*------------------------------------------------------------------------
 *  Function	: DetermineSeed
 *  Purpose	: Determion the seed for scrambling.
 *  Method	:
 *  1. Determine if the applicationt supplies its client-type.
 *  2. If not, do not scramble
 *  3. Otherwise send an empty log command.
 *  4. If a seed is returned, use it
 *  5. Otherwise don't scramble
 *
 *  History:
 *  20-09-1996	: eck      : creation
 *
 *  Returns	: Whether to scramle or not.
 *------------------------------------------------------------------------
 */
static bool
DetermineSeed(
	p3_connection_p	p3_conn,
	const char	*options,	/* To determine client-type */
	char		*seed)		/* pointer to store seed */
{
    /* If we have no client-type option, we won't use scrambling. */
    if (! PM_opt_fetch(options, "client-type",
			seed, SEEDLEN_MAX)) {
	return FALSE;
    }
    WMPutCmd("US", "\\LOG ;");
    if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
	if (WMGetVar(P3_VAR_SCRAMBLE_SEED, seed, SEEDLEN_MAX)) {
	    WMClearCurrentContext();
	    return TRUE;
	}
    }
    /* The next line is an LBS-hack (it does not send a status variable). */
    P3errno = P3E_NOERROR;

    WMClearCurrentContext();
    return FALSE;
} /* DetermineSeed() */

/*------------------------------------------------------------------------
 *  Function	: P3Logon
 *  Purpose	: Send logon command to Pica3-server
 *  Method	:
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *
 *  Description	:
 *  - P3Handle must be a valid P3Handle_p previously returned by P3Open() (and
 *    not already closed).
 *  - Options is a pointer to a string containing (';'-terminated) options.
 *    Options must (at least) contain:
 *    - "user=...."
 *    - "password=...."
 *
 *    Options can also contain:
 *    - "client-type=...."
 *
 *  The function will log-on to the Pica3 server (at application level). After
 *  establishing a Pica3 connection (P3Open()), any number of log-on's is
 *  allowed. In erroneous situations FALSE is returned and P3errno should be
 *  examined.
 *
 *  The client-type option will trigger password-scrambling. This means that
 *  the server is asked for a seed to be used in the password scrambling. If
 *  no seed is given, a normal login will take place. Otherwise, both the
 *  password and the value of the client-type option are scrambled using the
 *  seed and sent to the server.
 *
 *  Error values:
 *
 *	P3E_DISCONNECT	(server disconnected)
 *	P3E_IDPASSWRONG	(identification failure)
 *	P3E_LIBDEPOSIT	(deposit of library exhausted)
 *	P3E_MAXCHANNEL	(all channels occupied)
 *	P3E_NOERROR	(ok)
 *	P3E_PROTOCOL	(protocol error)
 *	P3E_SERVER	(server error)
 *	P3E_USRDEPOSIT	(deposit of user exhausted)
 *
 *  Notes	:
 *  The user must be specified, but the password need not be. Remark that most
 *  Pica3 servers automatically disconnect if the number of authorization
 *  failures exceed a certain maximum.
 *
 *  P3E_MAXCHANNEL is replied when the maximum number of channels is reached
 *  for the user who wants to log-on; the logon is rejected.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3Logon(
	P3Handle_p	P3handle,	/* Handle to Pica3-server */
	const char	*options)	/* options for logon */
{
    p3_connection_p	p3_conn = P3C_GetConnection(P3handle);
    char		user[IDLEN_MAX + 1];
    char		password[IDLEN_MAX + 1];
    char		charset[IDLEN_MAX + 1];
    CS_LOCAL(		command, 160);
    WMContext_t		prev_context;
    char		seed[SEEDLEN_MAX];

    P3errno = P3E_NOERROR;

    if (p3_conn != NULL) {
	if (! DetermineUserArgs(options, user, password, charset)) {
	    P3errno = P3E_IDPASSWRONG;
	    return FALSE;
	}
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3cn_wm_context);
	if (DetermineSeed(p3_conn, options, seed)) {
	    LoginScrambled(options, user, password, seed);
	} else {
	    CSPrintf(command, "\\log %s %s ;", user, password);
	    WMPutCmd("US", CSContents(command));
	}
	if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
	    switch (P3status) {
	    case P3_REJECT:
		    P3errno = P3E_IDPASSWRONG;
		    break;
	    case P3_MAXCHN:
		    P3errno = P3E_MAXCHANNEL;
		    break;
	    case P3_LIBDEP:
		    P3errno = P3E_LIBDEPOSIT;
		    break;
	    case P3_USRDEP:
		    P3errno = P3E_USRDEPOSIT;
		    break;
	    case P3_OK:
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
	    }
	}
	(void) WMSwitchContext(prev_context);
    } else {
	/* The handle is not valid */
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3Logon", "invalid handle");
    }

    return (P3errno == P3E_NOERROR);
} /* P3Logon() */

/*------------------------------------------------------------------------
 *  Function	: XLWMDetermineSeed
 *  Purpose	: Determion the seed for scrambling.
 *  Method	:
 *  1. Determine if the applicationt supplies its client-type.
 *  2. If not, do not scramble
 *  3. Otherwise send an empty log command.
 *  4. If a seed is returned, use it
 *  5. Otherwise don't scramble
 *
 *  History:
 *  18-12-2001  : evd     : Copied from DetermineSeed
 *
 *  Returns	: Whether to scramle or not.
 *------------------------------------------------------------------------
 */
static bool
XLWMDetermineSeed(
	p3_XLWMconnection_p	p3_conn,
	const char	*options,	/* To determine client-type */
	CSBuf_p seed)
{
    /* If we have no client-type option, we won't use scrambling. */
    if (! PM_opt_fetch(options, "client-type",
			(char *)CSContents(seed), CSSize(seed))) {
	return FALSE;
    }
    WMPutCmd("US", "\\LOG ;");
    if (P3L_XLWMSendRecv(p3_conn->p3XLWMcn_nio_handle)) {
	if (WMGetVarCS(P3_VAR_SCRAMBLE_SEED, seed)) {
	    WMClearCurrentContext();
	    return TRUE;
	}
    }
    /* The next line is an LBS-hack (it does not send a status variable). */
    P3errno = P3E_NOERROR;

    WMClearCurrentContext();
    return FALSE;
} /* XLWMDetermineSeed() */

/*------------------------------------------------------------------------
 *  Function	: P3XLWMLogon
 *  Purpose	: Send logon command to Pica3-server
 *  Method	:
 *
 *  History:
 *  18-12-2001  : evd     : Copied from P3Logon
 *  26-08-2004	: huug	  : Added characterset negotiation
 *
 *  Notes	: The user must be specified, but the password need not be.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3XLWMLogon(
	P3Handle_p	P3handle,	/* Handle to Pica3-server */
	const char	*options)	/* options for logon */
{
    p3_XLWMconnection_p	p3_conn = P3C_XLWMGetConnection(P3handle);
    char		user[IDLEN_MAX + 1];
    char		password[IDLEN_MAX + 1];
    char		charset[IDLEN_MAX + 1];
    char		srv_charset[IDLEN_MAX + 1];
    CS_LOCAL(		command, 160);
    WMContext_t		prev_context;
    CS_LOCAL(seed, SEEDLEN_MAX); 

    P3errno = P3E_NOERROR;

    if (p3_conn != NULL) {
	if (! DetermineUserArgs(options, user, password, charset)) {
	    P3errno = P3E_IDPASSWRONG;
	    return FALSE;
	}
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3XLWMcn_wm_context);
	if (XLWMDetermineSeed(p3_conn, options, seed)) {
	    LoginScrambled(options, user, password, CSContents(seed));
	} else {
	    CSPrintf(command, "\\log %s %s ;", user, password);
	    WMPutCmd("US", CSContents(command));
	}
	if (strlen(charset) > 0 && P3errno == P3E_NOERROR) {
	    WMRegVar(P3_VAR_CLIENT_CHARSET);
	    WMPutVar(P3_VAR_CLIENT_CHARSET, charset);
	}
	if (P3L_XLWMSendRecv(p3_conn->p3XLWMcn_nio_handle)) {
	    switch (P3status) {
	    case P3_REJECT:
		    P3errno = P3E_IDPASSWRONG;
		    break;
	    case P3_MAXCHN:
		    P3errno = P3E_MAXCHANNEL;
		    break;
	    case P3_LIBDEP:
		    P3errno = P3E_LIBDEPOSIT;
		    break;
	    case P3_USRDEP:
		    P3errno = P3E_USRDEPOSIT;
		    break;
	    case P3_OK:
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
	    }
	    if (strlen(charset) > 0) {
		if (!WMGetVar(P3_VAR_RESPONSE_CHARSET,
			      srv_charset, sizeof(srv_charset)))
		    srv_charset[0] = '\0';
		if (PM_strcmpi(charset, srv_charset) != 0) {
		    P3errno = P3E_NOSUPPORT;
		}
	    }
	}
	(void) WMSwitchContext(prev_context);
    } else {
	/* The handle is not valid */
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3XLWMLogon", "invalid handle");
    }

    return (P3errno == P3E_NOERROR);
} /* P3XLWMLogon() */
