/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3open.c
 *  Description	: The open-routine for a Pica3-server
 *  Author(s)	: Hans van Eck
 *  Creation	: 05-09-1994
 *
 *  History:
 *  05-09-1994	: eck      : creation
 *  14-12-2001  : evd      : Added P3XLWMOpen
 *  10-10-2013	: martinz  : Copied documentation from P3LIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: p3open.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <pica.h>
#include <picamisc.h>
#include <p3.h>
#include <fep.h>
#include <nio.h>
#include <er.h>
#include <cs.h>

#include "p3local.h"

/*------------------------------------------------------------------------
 *  Function	: P3Open
 *  Purpose	: Open a connection to a Pica3-server
 *  Method	: Open an active FEP.
 *
 *  History:
 *  05-09-1994	: eck      : creation
 *
 *  Description	:
 *  - Options is a pointer to a string containing (';'-terminated) options.
 *    Options must at least contain the following information:
 *    - type of netwerk protocol (TCP/IP, X25, etc)
 *    - network address
 *
 *  For more information see specification of the NIOOpen() routine.
 *
 *  The function will connect to the Pica3 server and returns a handle that
 *  must be used in subsequent calls to P3 functions.
 *
 *  In erroneous situations NULL is returned and P3errno should be examined.
 *  After a successful P3Open(), a log-on to the Pica3 server should take
 *  place (see P3Logon()).
 *
 *  Error values:
 *
 *	P3E_DISCONNECT	(server disconnected)
 *	P3E_NOERROR	(ok)
 *	P3E_NOMEM	(not enough memory)
 *	P3E_PROTOCOL	(protocol error)
 *	P3E_SERVER	(server error)
 *
 *  Returns	: The handle to the server opened (or NULL).
 *------------------------------------------------------------------------
 */
P3Handle_p
P3Open(
	const char	*options)	/* options */
{
    FEHandle_t		FEhandle;
    p3_connection_p	p3_conn = NULL;

    P3errno = P3E_NOERROR;

    FEhandle = FEStart(FEActive, options);

    if (FEhandle == -1) {
	P3errno = P3E_DISCONNECT;
	(void) ERWrite(P3errno, __func__, "%s", options);
	return NULL;
    }

    p3_conn = P3C_AddConnection(FEhandle);
    if (p3_conn == NULL) {
	(void) ERWrite(P3errno, __func__, "Adding connection");
	FESendAbort(FEhandle);
	return NULL;
    }
    return p3_conn->p3cn_p3_handle;
} /* P3Open() */

/*------------------------------------------------------------------------
 *  Function	: P3XLWMOpen
 *  Purpose	: Open a connection to a Pica3-server
 *  Method	: Open an active NIOHandle.
 *
 *  History:
 *  14-12-2001	: evd      : copied from P3Open
 *
 *  Notes	:
 *
 *  Returns	: The handle to the server opened (or NULL).
 *------------------------------------------------------------------------
 */
P3Handle_p
P3XLWMOpen(const char	*options)
{
    NIOHandle_t		NIOhandle;
    p3_XLWMconnection_p	p3_conn = NULL;
    CS_DYNAMIC(activeoptions);

    P3errno = P3E_NOERROR;
    CSSetString(activeoptions, "mode=active;");
    CSAppend(activeoptions, options);
    NIOhandle = NIOOpen(CSContents(activeoptions));
    CSDestroy(activeoptions);

    if (NIOhandle == NIO_NOHANDLE) {
	P3errno = P3E_DISCONNECT;
	(void) ERWrite(P3errno, __func__, "%s", options);
	return NULL;
    }

    p3_conn = P3C_XLWMAddConnection(NIOhandle);
    if (p3_conn == NULL) {
	(void) ERWrite(P3errno, __func__, "Adding connection");
	return NULL;
    }
    return p3_conn->p3XLWMcn_p3_handle;
} /* P3XLWMOpen() */
