/*========================================================================
 *  Copyright (c) 1995 by Pica
 *
 *  File	: p3option.c
 *  Description	: Set an option
 *  Author(s)	: Hans van Eck
 *  Creation	: 04-10-1995
 *
 *  History:
 *  04-10-1995	: eck      : creation
 *  10-10-2013	: martinz  : Copied documentation from P3LIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: p3option.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <pica.h>
#include <wm.h>
#include <p3.h>
#include <cs.h>
#include <er.h>

#include "p3local.h"

/*------------------------------------------------------------------------
 *  Function	: P3Option
 *  Purpose	: Set an option
 *  Method	:
 *
 *  History:
 *  04-10-1995	: eck      : creation
 *
 *  Description	:
 *  - P3handle must be a valid P3handle_p previously returned by P3Open() (and
 *    not P3Close()-d).
 *  - Option must be a pointer to a string containing a valid option.
 *    Valid options are:
 *    - "language=<language-code>" (NE|EN|DU|FR)
 *
 *  If something goes wrong, P3Option() will return FALSE and P3errno should be
 *  examined.
 *
 *  Notes	:
 *  The language option is used to set a specific language at the server.
 *
 *  Error values:
 *
 *	P3E_DISCONNECT	(server disconnected)
 *	P3E_NOERROR	(ok)
 *	P3E_NOSUPPORT	(function not supported)
 *	P3E_PROTOCOL	(protocol error)
 *	P3E_SERVER	(server error)
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3Option(
	P3Handle_p	P3handle,	/* Handle to Pica3-server */
	const char	*Option)	/* Option to set */
{
    p3_connection_p	p3_conn = P3C_GetConnection(P3handle);
    CS_LOCAL(		command, 80);
    WMContext_t		prev_context;

    P3errno = P3E_NOERROR;
    if (p3_conn == NULL) {
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3Option", "Invalid handle");
    } else {
	P3L_InsertLanguageCommand(command, Option);
	if (CSLength(command) == 0) {
	    P3errno = P3E_NOSUPPORT;
	    (void) ERTrace(P3errno, "P3Option", "option: \"%s\"", Option);
	} else {
	    prev_context = P3L_SwitchAndClearContext(p3_conn->p3cn_wm_context);
	    (void) WMPutCmd("US", CSContents(command));
	    if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
		switch (P3status) {
		case P3_OK:
			break;
		case P3_NOTSUP:
			P3errno = P3E_NOSUPPORT;
			break;
		default:
			if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
		}
		if (P3errno != P3E_NOERROR) {
		    (void) ERWrite(P3errno, "P3Option",
					    "option: \"%s\"", Option);
		}
	    }
	    (void) WMSwitchContext(prev_context);
	}
    }
    return (P3errno == P3E_NOERROR);
} /* P3Option() */
