/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3scandbsupdates.c
 *  Description	: Scan database updates
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 25-07-1996
 *
 *  History:
 *  25-07-1996	: PvdM      : creation
 *  13-12-2001  : evd       : added P3XLWMScanDbsUpdates
 *
 *========================================================================
 */

#pragma ident "$Id: p3scandbsupdates.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <p3.h>
#include <cs.h>

#include "p3local.h"

#define IDLEN_MAX	 32

#define STR_LIBRARY	"library"

/*------------------------------------------------------------------------
 *  Function	: DetermineLibraryArgs
 *  Purpose	: Determine library number from options
 *  Method	: obvious
 *
 *  History:
 *  25-07-1996	: PvdM      : creation
 *
 *  Returns	: TRUE if a library was specified, FALSE otherwise.
 *------------------------------------------------------------------------
 */
static bool
DetermineLibraryArgs(
	const char	*options,
	char		library[IDLEN_MAX + 1])
{
    library[0] = '\0';

    options = PM_opt_findstart(options);
    while ((options != NULL) && (*options != '\0')) {
	if (PM_opt_equal(options, STR_LIBRARY)) {
	    options = PM_opt_findval(options);
	    PM_opt_copyval(options, library, IDLEN_MAX + 1);
	}
	options = PM_opt_findnext(options);
    }
    return (library[0] != '\0');
} /* DetermineLibraryArgs() */

static void
FillReturnedInfo(
	char*	Last,
	CBBuf_t	*CBbuf,
	size_t	CBsize)
{
    char	varbuf[MAXLEN];
    const char	*location = "P3ScanDbsUpdates.FillReturnedInfo";

    if (!WMGetField("ET", varbuf, sizeof(varbuf))) {
        P3errno = P3E_BUFFER;
        (void) ERWrite(P3errno, location, "getting field ET");
        return;
    }
    else {
	(void) strcpy(Last, varbuf);
    }

    (void) P3L_WMelt2CB("IL", CBbuf, CBsize,
		FLG_FIELD | FLG_ADDID, location);
    (void) P3L_WMelt2CB("AL", CBbuf, CBsize,
		FLG_FIELD | FLG_ADDID, location);
    (void) P3L_WMelt2CB("TX", CBbuf, CBsize,
		FLG_FIELD | FLG_ADDID | FLG_SPLITLINE | FLG_OPTIONAL, location);
} /* FillReturnedInfo */

/*------------------------------------------------------------------------
 *  Function	: P3ScanDbsUpdates
 *  Purpose	: Scan database updates
 *  Method	:
 *
 *  History:
 *  25-07-1996	: PvdM      : creation
 *
 *  Notes	: Library need not be specified.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3ScanDbsUpdates(
	P3Handle_p	P3handle,	/* Handle to Pica3-server */
	const char	*Offset,	/* valid date/time */
	char		Last[P3DATETIME_SIZ],
					/* date/time of last element */
	int		Count,		/* maximum number to retrieve */
	const char	*Options,	/* options for scan */
	CBBuf_t		*CBbuf,		/* Buffer to contain updates */
	size_t		CBsize)		/* Size of buffer */
{
    p3_connection_p	p3_conn = P3C_GetConnection(P3handle);
    char		library[IDLEN_MAX + 1];
    CS_LOCAL(		command, 160);
    WMContext_t		prev_context;

    P3errno = P3E_NOERROR;

    if (p3_conn != NULL) {
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3cn_wm_context);
	if (! DetermineLibraryArgs(Options, library)) {
	    CSPrintf(command, "\\too %d \\upd \\sed %s ;", Count, Offset);
	}
	else {
	    CSPrintf(command, "\\too %d \\upd \\sed %s \\doo %s ;", 
		Count, Offset, library);
	}
	WMPutCmd("US", CSContents(command));
	if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
	    switch (P3status) {
	    case P3_NOTNOW:
		    P3errno = P3E_PROTOCOL;
		    break;
	    case P3_NOTSUP:
		    P3errno = P3E_NOSUPPORT;
		    break;
	    case P3_NOAUTH:
		    P3errno = P3E_AUTHORITY;
		    break;
	    case P3_ERROR:
		    P3errno = P3E_SERVER;
		    break;
	    case P3_OK:
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
	    }

	    if (P3errno == P3E_NOERROR) {
		FillReturnedInfo(Last, CBbuf, CBsize);
	    }
	}
	(void) WMSwitchContext(prev_context);
    } else {
	/* The handle is not valid */
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3ScanDbsUpdates", "invalid handle");
    }

    return (P3errno == P3E_NOERROR);
} /* P3ScanDbsUpdates() */

static void
XLWMFillReturnedInfo(
	char*	Last,
	CBBuf_t	*CBbuf,
	size_t	CBsize)
{
    CS_DYNAMIC(varbuf);
    const char	*location = "P3XLWMScanDbsUpdates.XLWMFillReturnedInfo";

    if (!WMGetFieldCS("ET", varbuf)) {
        P3errno = P3E_BUFFER;
        (void) ERWrite(P3errno, location, "getting field ET");
	CSDestroy(varbuf);
        return;
    }
    else {
	(void) strcpy(Last, CSContents(varbuf));
    }

    (void) P3L_XLWMelt2CB("IL", CBbuf, CBsize,
		FLG_FIELD | FLG_ADDID, location);
    (void) P3L_XLWMelt2CB("AL", CBbuf, CBsize,
		FLG_FIELD | FLG_ADDID, location);
    (void) P3L_XLWMelt2CB("TX", CBbuf, CBsize,
		FLG_FIELD | FLG_ADDID | FLG_SPLITLINE | FLG_OPTIONAL, location);
    CSDestroy(varbuf);
} /* FillReturnedInfo */

/*------------------------------------------------------------------------
 *  Function	: P3XLWMScanDbsUpdates
 *  Purpose	: Scan database updates from worm longer than 4k
 *  Method	:
 *
 *  History:
 *  13-12-2001	: evd      : copied from P3ScanDbsUpdates
 *
 *  Notes	: Library need not be specified.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3XLWMScanDbsUpdates(
	P3Handle_p	P3handle,	/* Handle to Pica3-server */
	const char	*Offset,	/* valid date/time */
	char		Last[P3DATETIME_SIZ],
					/* date/time of last element */
	int		Count,		/* maximum number to retrieve */
	const char	*Options,	/* options for scan */
	CBBuf_t		*CBbuf,		/* Buffer to contain updates */
	size_t		CBsize)		/* Size of buffer */
{
    p3_XLWMconnection_p	p3_conn = P3C_XLWMGetConnection(P3handle);
    char		library[IDLEN_MAX + 1];
    CS_LOCAL(		command, 160);
    WMContext_t		prev_context;

    P3errno = P3E_NOERROR;

    if (p3_conn != NULL) {
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3XLWMcn_wm_context);
	if (! DetermineLibraryArgs(Options, library)) {
	    CSPrintf(command, "\\too %d \\upd \\sed %s ;", Count, Offset);
	}
	else {
	    CSPrintf(command, "\\too %d \\upd \\sed %s \\doo %s ;", 
		Count, Offset, library);
	}
	WMPutCmd("US", CSContents(command));
	if (P3L_XLWMSendRecv(p3_conn->p3XLWMcn_nio_handle)) {
	    switch (P3status) {
	    case P3_NOTNOW:
		    P3errno = P3E_PROTOCOL;
		    break;
	    case P3_NOTSUP:
		    P3errno = P3E_NOSUPPORT;
		    break;
	    case P3_NOAUTH:
		    P3errno = P3E_AUTHORITY;
		    break;
	    case P3_ERROR:
		    P3errno = P3E_SERVER;
		    break;
	    case P3_OK:
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
	    }

	    if (P3errno == P3E_NOERROR) {
		XLWMFillReturnedInfo(Last, CBbuf, CBsize);
	    }
	}
	(void) WMSwitchContext(prev_context);
    } else {
	/* The handle is not valid */
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3XLWMScanDbsUpdates", "invalid handle");
    }

    return (P3errno == P3E_NOERROR);
} /* P3XLWMScanDbsUpdates() */
