/*========================================================================
 *  Copyright (c) 1994 by Pica
 *
 *  File	: p3select.c
 *  Description	: Select a database on a Pica3 server
 *  Author(s)	: Hans van Eck
 *  Creation	: 14-09-1994
 *
 *  History:
 *  14-09-1994	: eck      : creation
 *  18-04-1996	: eck      : Added handling of P3_NOAUTH.
 *  13-05-1998	: eck      : Clear scan-cache on database select.
 *  18-12-2001  : evd      : Added P3Select
 *  10-10-2013	: martinz  : Copied documentation from P3LIB.DOC.
 *
 *========================================================================
 */

#pragma ident "$Id: p3select.c 305171 2013-11-01 13:41:52Z sieburgw $"

#include <string.h>

#include <pica.h>
#include <cs.h>
#include <p3.h>
#include <wm.h>
#include <er.h>

#include "p3local.h"

/*------------------------------------------------------------------------
 *  Function	: P3Select
 *  Purpose	: Select a database on a Pica3-server as the current one.
 *  Method	: Send command. If all ok, determine the profile.
 *
 *  History:
 *  06-10-1994	: eck      : creation
 *  08-02-1996	: jos      : Accept NULL DBname argument.
 *
 *  Description	:
 *  - P3Handle must be a valid P3Handle_p previously returned by P3Open() (and
 *    not already closed).
 *  - DbsName must be a valid database name.
 *  - Options must be a pointer to a string containing (';'-terminated) options.
 *    Valid options are:
 *    - "language=<language-code>" (NE|EN|DU|FR)
 *
 *  When the database selection fails, FALSE is returned and P3errno should be
 *  examined. Otherwise TRUE is returned.
 *
 *  Error values:
 *
 *	P3E_DISCONNECT	(server disconnected)
 *	P3E_LIBDEPOSIT	(deposit of library exhausted)
 *	P3E_MAXCHANNEL	(all channels occupied)
 *	P3E_NOERROR	(ok)
 *	P3E_PROTOCOL	(protocol error)
 *	P3E_SERVER	(server error)
 *	P3E_USRDEPOSIT	(deposit of user exhausted)
 *
 *  The function will first convert the expression-stack to a query that the
 *  server that holds the database will understand. Then it will send the
 *  query to the server; the server will perform the search but will not
 *  create a set. The size of the would-be set is returned. If anything went
 *  wrong, zero is returned. The variable QSerrno will contain the error
 *  number.
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3Select(
	P3Handle_p	P3handle,	/* Handle to the Pica3-server */
	const char	*DBname,	/* id. of the database to select */
	const char	*Options)	/* eg. "language=" */
{
    p3_connection_p	p3_conn = P3C_GetConnection(P3handle);
    CS_LOCAL(		command, 160);
    WMContext_t		prev_context;
    CS_LOCAL(		ext_name, 64);

    P3errno = P3E_NOERROR;
    if (p3_conn == NULL) {
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3Select", "invalid handle");
    } else {
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3cn_wm_context);
	if (DBname != NULL) {
	    P3L_InternToPica3Name(DBname, ext_name);
	}
	CSPrintf(command, "\\sel \\cat %s ;", CSContents(ext_name));
	P3L_InsertLanguageCommand(command, Options);
	WMPutCmd("US", CSContents(command));
	if (P3L_SendRecv(p3_conn->p3cn_fe_handle)) {
	    switch (P3status) {
	    case P3_OK:
		    break;
	    case P3_NOTNOW:
	    case P3_NOTDBS:
		    P3errno = P3E_PROTOCOL;
		    (void) ERWrite(P3errno, "P3Select",
					    "internal protocol error");
		    break;
	    case P3_MAXCHN:
                    P3errno = P3E_MAXCHANNEL;
		    break;
	    case P3_LIBDEP:
		    P3errno = P3E_LIBDEPOSIT;
		    break;
	    case P3_NOAUTH:
		    P3errno = P3E_AUTHORITY;
		    break;
	    case P3_USRDEP:
		    P3errno = P3E_USRDEPOSIT;
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
		    break;
	    }
	}
	(void) WMSwitchContext(prev_context);
    }

    return (P3errno == P3E_NOERROR);
} /* P3Select() */

/*------------------------------------------------------------------------
 *  Function	: P3XLWMSelect
 *  Purpose	: Select a database on a Pica3-server as the current one.
 *  Method	: Send command. If all ok, determine the profile.
 *
 *  History:
 *  08-02-1996	: jos      : Accept NULL DBname argument.
 *  18-12-2001  : evd      : Copied from P3Select
 *
 *  Returns	: Indication of success
 *------------------------------------------------------------------------
 */
bool
P3XLWMSelect(
	P3Handle_p	P3handle,	/* Handle to the Pica3-server */
	const char	*DBname,	/* id. of the database to select */
	const char	*Options)	/* eg. "language=" */
{
    p3_XLWMconnection_p	p3_conn = P3C_XLWMGetConnection(P3handle);
    CS_LOCAL(		command, 160);
    WMContext_t		prev_context;
    CS_LOCAL(		ext_name, 64);

    P3errno = P3E_NOERROR;
    if (p3_conn == NULL) {
	P3errno = P3E_SERVER;
	(void) ERWrite(P3errno, "P3Select", "invalid handle");
    } else {
	prev_context = P3L_SwitchAndClearContext(p3_conn->p3XLWMcn_wm_context);
	if (DBname != NULL) {
	    P3L_InternToPica3Name(DBname, ext_name);
	}
	CSPrintf(command, "\\sel \\cat %s ;", CSContents(ext_name));
	P3L_InsertLanguageCommand(command, Options);
	WMPutCmd("US", CSContents(command));
	if (P3L_XLWMSendRecv(p3_conn->p3XLWMcn_nio_handle)) {
	    switch (P3status) {
	    case P3_OK:
		    break;
	    case P3_NOTNOW:
	    case P3_NOTDBS:
		    P3errno = P3E_PROTOCOL;
		    (void) ERWrite(P3errno, "P3Select",
					    "internal protocol error");
		    break;
	    case P3_MAXCHN:
                    P3errno = P3E_MAXCHANNEL;
		    break;
	    case P3_LIBDEP:
		    P3errno = P3E_LIBDEPOSIT;
		    break;
	    case P3_NOAUTH:
		    P3errno = P3E_AUTHORITY;
		    break;
	    case P3_USRDEP:
		    P3errno = P3E_USRDEPOSIT;
		    break;
	    default:
		    if (P3errno == P3E_NOERROR) P3errno = P3E_SERVER;
		    break;
	    }
	}
	(void) WMSwitchContext(prev_context);
    }

    return (P3errno == P3E_NOERROR);
} /* P3XLWMSelect() */
