/*========================================================================
 *  Copyright (c) 2006 by OCLC PICA
 *
 *  File	: pa.c
 *  Description	: pointer array library
 *  Author(s)	: Johan List
 *  Creation	: 14-06-2006
 *
 *  History:
 *  14-06-2006	: listj    : creation
 *  23-06-2006  : listj    : GetLastIndex() renamed to GetNrEntries()
 *  24-07-2006  : listj    : initialize pa_entries in PA_EnsureCapacity to NULL on
 *                           first alloc
 *  24-07-2006  : listj    : simplified logic in PA_EnsureCapacity; any init
 *                           needed moved to PA_Create
 *  07-08-2006  : listj    : added extra entry checks in several functions
 *  31-05-2007	: heesakkm : PA_indirect_compare(): removed stable-sorting code.
 *  23-06-2009	: paul     : Use PA_NOFUNC instead of NULL
 *  10-02-2010	: listj    : added dump element functionality (PA_Dump and
 *			     element callback function)
 *  17-05-2010	: martinz  : Added PA_SortU() and PA_Search().
 *  13-09-2011	: paul     : CBS-3179 Added PA_DelN()
 *  28-08-2018	: zwaalm   : CBS-8356 Added PA_SearchDirect().
 *
 *  Notes	:
 *
 *========================================================================
 */
#pragma ident "$Id: pa.c 359038 2018-08-29 16:05:28Z zwaalm $"

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <pica.h>
#include <picamisc.h>
#include <pa.h>
#include <er.h>

#define PA_DEFAULT_SIZE			(128)

ERROR_DEFINE(PAE_NOERROR);
ERROR_DEFINE(PAE_NOMEM);

PAErrno_t       PAerrno = PAE_NOERROR;

struct PAHandle_t {
    size_t		pa_nrentries;	/* number of entries */
    size_t		pa_capacity;	/* capacity of the pointer array */
    cvoidp		*pa_entries;	/* entry array */
    PA_NewOldCallback_f	pa_newold;
};

static PA_SortCompare_f	PA_userCompare = NULL;
static PA_Compare_f	PA_compare = NULL;
static cvoidp		PA_userdata = NULL;

/*------------------------------------------------------------------------
 *  Function	: PA_NewOldString
 *  Purpose	: NewOld allocation/deallocation function for string elements
 *  Method	:
 *
 *  History:
 *  15-06-2006	: listj    : creation
 *
 *  Notes	:
 *
 *  Returns	: pointer to the allocated/reallocated element
 *------------------------------------------------------------------------
 */
cvoidp
PA_NewOldString(
	cvoidp          newp,   /* The value given to PA_Add()/PA_SetN */
	cvoidp          oldp)   /* The value currently in the array */
{
    const char  *news 	= newp;           /* newp as string */
    char        *tmp;

    /* If newp == NULL, we should deallocate */
    if (newp == NULL) {
	free((voidp) oldp);
	return NULL;
    }

    tmp = realloc((voidp) oldp, strlen((char *)newp) + 1);
    if (tmp != NULL) {
	(void) strcpy(tmp, news);
    } else {
	PAerrno = PAE_NOMEM;
	(void)ERWrite(PAerrno, "PA_NewOldString", "no room to realloc");
	free((voidp) oldp);             /* Just delete the old one */
    }
    return tmp;
} /* PANewOldString() */

/*------------------------------------------------------------------------
 *  Function	: PA_EnsureCapacity
 *  Purpose	: Ensures the array has enough capacity to hold new element
 *  Method	: realloc / initialize additional elements
 *
 *  History:
 *  15-06-2006	: listj    : creation
 *  24-07-2006  : listj    : simplified logic; any init needed moved to PA_Create
 *
 *  Notes	: internal function, so no checks (done by calling
 *                functions)
 *
 *  Returns	: TRUE on success, FALSE on failure
 *------------------------------------------------------------------------
 */
static bool
PA_EnsureCapacity(
	PAHandle_p	paHandle, /* Pointer array handle */
	size_t		idx)
{
    const char	fname[]	= "PA_EnsureCapacity";
    cvoidp	*tmp	= NULL;
    size_t	i;
    size_t	newCapacity;

    /* Enough space? Return */
    if (idx < paHandle->pa_capacity) {
	return TRUE;
    }

    if (paHandle->pa_capacity > 0) {
	/* Not the first time, increase cap until we comply */
	newCapacity = paHandle->pa_capacity;
	while (newCapacity <= idx) {
	    newCapacity *= 2;
	}
    } else {
	newCapacity = idx;
    }

    tmp = (cvoidp *)realloc(paHandle->pa_entries, newCapacity * sizeof(cvoidp));
    if (tmp == NULL) {
        PAerrno = PAE_NOMEM;
	ERWrite(PAerrno, fname, "realloc(paHandle->pa_entries) failed");
	return FALSE;
    }

    paHandle->pa_entries = tmp;
    paHandle->pa_capacity = newCapacity;
    for (i = paHandle->pa_nrentries; i < paHandle->pa_capacity; i++) {
	paHandle->pa_entries[i] = NULL;
    }

    return TRUE;
} /* PA_EnsureCapacity() */

/*------------------------------------------------------------------------
 *  Function	: PA_Create
 *  Purpose	: Create an empty pointer array
 *  Method	: malloc and initialize datastructures
 *
 *  History:
 *  14-06-2006	: listj    : creation
 *
 *  Notes	: this function recognizes the following options:
 *
 *                - size: the initial size (# elements) of the pointer array.
 *                  Default size is 128 elements.
 *
 *  Returns	: - (PAHandle_p) handle to the pointer array created
 *		  - NULL in case of failure, PAerrno will be set
 *------------------------------------------------------------------------
 */
PAHandle_p
PA_Create(
	const char			*options, /* valid options string */
	PA_NewOldCallback_f		NewOld)   /* alloc/dealloc function */
{
    PAHandle_p	paHandle		= NULL; /* Pointer array handle */
    char	valbuf[PM_INTLEN_MAX];
    long	arraySize;

    PAerrno = PAE_NOERROR;

    if (PM_opt_fetch(options, "size", valbuf, sizeof(valbuf))) {
	arraySize = PM_atosize(valbuf);
	if (arraySize <= 0) {
	    arraySize = PA_DEFAULT_SIZE;
	}
    } else {
	arraySize = PA_DEFAULT_SIZE;
    }

    paHandle = (PAHandle_p)malloc(sizeof(*paHandle));
    if (paHandle != NULL) {
	paHandle->pa_capacity = 0;
	paHandle->pa_nrentries = 0;
	paHandle->pa_entries = NULL;
	paHandle->pa_newold = PA_NOFUNC;

	if (!PA_EnsureCapacity(paHandle, arraySize)) {
	    /* PAerrno has been set by PA_EnsureCapacity() */
	    free(paHandle);
	    paHandle = NULL;
	} else {
	    paHandle->pa_newold = NewOld;
	    /* Other members set in PA_EnsureCapacity() */
	}
    }

    return paHandle;
} /* PA_Create() */

/*------------------------------------------------------------------------
 *  Function	: PA_Clear
 *  Purpose	: Clear the pointer array
 *  Method	: If the NewOld function is not PA_NOFUNC, the NewOld
 *		  function is used to dealllocate the pointers otherwise
 *		  the pointers are just set to NULL
 *
 *  History:
 *  14-06-2006	: listj    : creation
 *  07-08-2006  : listj    : NULL check on paHandle
 *
 *  Notes	: elements [ 0..pa_maxentry > are cleared
 *------------------------------------------------------------------------
 */
void
PA_Clear(
	PAHandle_p	paHandle) /* Pointer array handle */
{
    size_t	i;

    if (paHandle != NULL) {
	for (i = 0; i < paHandle->pa_nrentries; i++) {
	    if (paHandle->pa_newold != PA_NOFUNC) {
		while (paHandle->pa_entries[i] != NULL) {
		    paHandle->pa_entries[i] =
			paHandle->pa_newold(NULL, paHandle->pa_entries[i]);
		}
	    } else {
		paHandle->pa_entries[i] = NULL;
	    }
	}

	paHandle->pa_nrentries = 0;
    }
} /* PA_Clear() */

/*------------------------------------------------------------------------
 *  Function	: PA_Destroy
 *  Purpose	: Frees all memory associated with the pointer array
 *  Method	: PA_Clear / free pointer array / free paHandle
 *
 *  History:
 *  14-06-2006	: listj    : creation
 *
 *  Notes	:
 *------------------------------------------------------------------------
 */
void
PA_Destroy(
	PAHandle_p	paHandle) /* Pointer array handle */
{
    if (paHandle != NULL) {
	PA_Clear(paHandle);
	free(paHandle->pa_entries);
	free(paHandle);
    }
} /* PA_Destroy() */

/*------------------------------------------------------------------------
 *  Function	: PA_GetN
 *  Purpose	: retrieves the Nth pointer array element
 *  Method	:
 *
 *  History:
 *  14-06-2006	: listj    : creation
 *  07-08-2006  : listj    : return NULL on NULL paHandle
 *
 *  Notes	: this function uses physical array indexing ([0,..]), not
 *                logical ([1,..])
 *
 *  Returns	: - the chosen pointer array element
 *		  - NULL when paHandle == NULL
 *		  - NULL when pointer does not exist
 *------------------------------------------------------------------------
 */
cvoidp
PA_GetN(
	PAHandle_p	paHandle, /* Pointer array handle */
	int		idx)      /* Index in array, starts from 0 (zero) */
{
    if (paHandle == NULL) {
	return NULL;
    }

    if ((idx < 0) || (paHandle->pa_capacity <= (size_t)idx)) {
	return NULL;
    }

    return paHandle->pa_entries[idx];
} /* PA_GetN() */

/*------------------------------------------------------------------------
 *  Function	: PA_SetN
 *  Purpose	: Set the element with index `N` of the pointer array
 *  Method	:
 *
 *  History:
 *  15-06-2006	: listj    : creation
 *  07-08-2006  : listj    : return NULL on NULL paHandle
 *
 *  Notes	: this function uses physical array indexing ([0,..]),
 *                not logical ([1,..])
 *
 *  Returns	: - the newly added pointer
 *		  - NULL in case of error
 *------------------------------------------------------------------------
 */
cvoidp
PA_SetN(
	PAHandle_p	paHandle, /* Pointer array handle */
	int		idx,      /* Index in array, starts from 0 (zero) */
	cvoidp		ptr)	  /* The newly added pointer */
{
    PAerrno = PAE_NOERROR;

    if (paHandle == NULL) {
	return NULL;
    }

    if (idx < 0) {
	return NULL;
    }

    if (paHandle->pa_capacity <= (size_t)idx) {
	if (!PA_EnsureCapacity(paHandle, idx)) {
	    /* PAerrno has been set by PA_EnsureCapacity */
	    return NULL;
	}
    }

    if (paHandle->pa_newold != PA_NOFUNC) {
	paHandle->pa_entries[idx] =
	    paHandle->pa_newold(ptr, paHandle->pa_entries[idx]);
    } else {
	paHandle->pa_entries[idx] = ptr;
    }

    if (idx >= paHandle->pa_nrentries) {
	paHandle->pa_nrentries = idx + 1;
    }

    return paHandle->pa_entries[idx];
} /* PA_SetN() */

/*------------------------------------------------------------------------
 *  Function	: PA_GetCapacity
 *  Purpose	: get the capacity of the pointer collection
 *  Method	:
 *
 *  History:
 *  20-06-2006	: listj    : creation
 *  07-08-2006  : listj    : return NULL on 0 capacity
 *
 *  Notes	:
 *
 *  Returns	: - the capacity of the pointer collection
 *		  - 0 when paHandle == NULL
 *------------------------------------------------------------------------
 */
size_t
PA_GetCapacity(
	PAHandle_p	paHandle) /* Pointer array handle */
{
    if (paHandle == NULL) {
	return 0;
    }
    return paHandle->pa_capacity;
} /* PA_GetCapacity() */

/*------------------------------------------------------------------------
 *  Function	: PA_GetNrEntries
 *  Purpose	: Get the number of entries
 *  Method	:
 *
 *  History:
 *  21-06-2006	: listj    : creation
 *  07-08-2006  : listj    : return 0 on NULL paHandle
 *
 *  Notes	:
 *
 *  Returns	: - the number of entries
 *		  - 0 when paHandle == NULL
 *------------------------------------------------------------------------
 */
size_t
PA_GetNrEntries(
	PAHandle_p	paHandle) /* Pointer array handle */
{
    if (paHandle == NULL) {
	return 0;
    }
    return paHandle->pa_nrentries;
} /* PA_GetNrEntries() */

/*------------------------------------------------------------------------
 *  Function	: PA_Add
 *  Purpose	: Add an element to the pointer array
 *  Method	: The element just behind the last is set using PA_SetN()
 *
 *  History:
 *  21-06-2006	: listj    : creation
 *  07-08-2006  : listj    : return NULL on NULL paHandle
 *
 *  Notes	: element is simply appended to the array, and makes
 *                no use of existing 'empty' space (ie., NULL entries)
 *
 *  Returns	: a pointer to the newly added element, or NULL in case
 *                of error
 *------------------------------------------------------------------------
 */
cvoidp
PA_Add(
	PAHandle_p	paHandle, /* Pointer array handle */
	cvoidp		ptr)	  /* Element to add to Pointer Array */
{
    if (paHandle == NULL) {
	return NULL;
    }
    return PA_SetN(paHandle, PA_GetNrEntries(paHandle), ptr);
} /* PA_Add() */

/*------------------------------------------------------------------------
 *  Function	: PA_AddAtN
 *  Purpose	: Add an element in the middle of a pointer array
 *                To add elements at the end, use PA_Add() instead.
 *
 *  History:
 *  17-10-2017	: smitp    : CBS-7731 creation
 *
 *  Notes	:
 *  PA_Add(paHandle, ptr)
 *
 *  is identical to:
 *
 *  PA_AddAtN(paHandle, PA_GetNrEntries(paHandle), ptr)
 *
 *  Returns	:
 *  - Pointer to the added element
 *  - or NULL if an error occurred
 *
 *------------------------------------------------------------------------
 */
cvoidp
PA_AddAtN(
	PAHandle_p	paHandle, /* Pointer array handle */
	int		idx,	  /* Index of element to be added */
	cvoidp		ptr)	  /* The newly added pointer */
{
    PAerrno = PAE_NOERROR;

    if (paHandle == NULL || idx < 0) {
	return NULL;
    }

    // Make room for the new entry if position already occupied
    if ((size_t)idx < paHandle->pa_nrentries) {
	if (!PA_EnsureCapacity(paHandle, paHandle->pa_nrentries + 1)) {
	    // PAerrno has been set by PA_EnsureCapacity()
	    return NULL;
	}

	// Make room for the new entry
	for (int i = paHandle->pa_nrentries; i > idx; i--) {
	    paHandle->pa_entries[i] = paHandle->pa_entries[i - 1];
	}
	paHandle->pa_entries[idx] = NULL;
	paHandle->pa_nrentries++;
    }
    
    return PA_SetN(paHandle, idx, ptr); 
} // PA_AddAtN()

/*------------------------------------------------------------------------
 *  Function	: PA_DelN
 *  Purpose	: Delete an element from the pointer arraythe pointer array
 *  Method	: PA_SetN(NULL) and move all trailing elements backwards
 *
 *  History:
 *  13-09-2011	: paul     : creation
 *
 *  Notes	:
 *
 *------------------------------------------------------------------------
 */
void
PA_DelN(
	PAHandle_p	paHandle, /* Pointer array handle */
	int		idx)	  /* Index of element to be deleted */
{
    PAerrno = PAE_NOERROR;

    if (paHandle == NULL) {
	return;
    }

    if ((idx < 0) || (paHandle->pa_nrentries <= (size_t)idx)) {
	return;
    }

    (void)PA_SetN(paHandle, idx, NULL);

    while (idx + 1 < paHandle->pa_nrentries) {
	paHandle->pa_entries[idx] = paHandle->pa_entries[idx + 1];
	idx++;
    }
    paHandle->pa_entries[idx] = NULL;
    paHandle->pa_nrentries--;
} /* PA_DelN() */

/*------------------------------------------------------------------------
 *  Function	: PA_indirect_compare
 *  Purpose	: PA_Sort intermediate compare function
 *  Method	:
 *
 *  History:
 *  23-06-2006	: listj    : creation
 *  31-05-2007	: heesakkm : Removed code that tried to implement stable
 *            	             sorting. This function is not the proper place,
 *            	             since it can never work with qsort().
 *
 *  Notes	:
 *  This function performs a dereference so the application does not need to.
 *  Moreover, the application cannot accidentally destroy the internal data
 *  structure by setting pointer entries.
 *
 *  Returns	: -1 when *arg1 < *arg2
 *                0 when *arg1 equal to *arg2
 *                1 when *arg1 > *arg2
 *------------------------------------------------------------------------
 */
static int
PA_indirectCompare(
	cvoidp arg1,
	cvoidp arg2)
{
    return PA_userCompare(*(cvoidp *)arg1, *(cvoidp *)arg2);
} /* PA_indirectCompare() */

/*------------------------------------------------------------------------
 *  Function	: PA_indirectCompare2
 *  Purpose	: PA_SortU/PA_Search intermediate compare function
 *  Method	: 
 *
 *  History:
 *  21-01-2010	: martinz  : creation
 *
 *  Notes	:
 *
 *  Returns	: -1 when *arg1 < *arg2
 *                0 when *arg1 equal to *arg2
 *                1 when *arg1 > *arg2
 *------------------------------------------------------------------------
 */
static int
PA_indirectCompare2(
	cvoidp arg1,
	cvoidp arg2)
{
    return PA_compare(*(cvoidp *)arg1, *(cvoidp *)arg2, PA_userdata);
} /* PA_indirectCompare2() */

/*------------------------------------------------------------------------
 *  Function	: PA_Sort
 *  Purpose	: sorts the pointer array
 *  Method	: qsort
 *
 *  History:
 *  21-06-2006	: listj    : creation
 *
 *  Notes	:
 *------------------------------------------------------------------------
 */
void
PA_Sort(
	PAHandle_p		paHandle, /* Pointer array handle */
	PA_SortCompare_f	compare)  /* Compare function */
{
    if ((paHandle == NULL) || (compare == NULL)) {
	return;
    }

    PA_userCompare = compare;
    qsort(paHandle->pa_entries,
	  paHandle->pa_nrentries,
	  sizeof(cvoidp),
	  PA_indirectCompare);
    PA_userCompare = NULL;
} /* PA_Sort() */


/*------------------------------------------------------------------------
 *  Function	: PA_SortU
 *  Purpose	: Sorts all elements with user-data.
 *  Method	: qsort
 *
 *  History:
 *  21-01-2010	: martinz  : creation
 *
 *  Notes	:
 *
 *  Returns	: Nothing.
 *
 *------------------------------------------------------------------------
 */
void
PA_SortU(
	PAHandle_p	paHandle, /* Pointer array handle */
	PA_Compare_f	compare,  /* Compare function */
	cvoidp		userdata) /* User data */
{
    if ((paHandle == NULL) || (compare == NULL)) {
	return;
    }

    PA_compare = compare;
    PA_userdata = userdata;
    qsort(paHandle->pa_entries,
	  paHandle->pa_nrentries,
	  sizeof(cvoidp),
	  PA_indirectCompare2);
    PA_compare = NULL;
    PA_userdata = NULL;
} /* PA_SortU() */


/*------------------------------------------------------------------------
 *  Function	: PA_Search
 *  Purpose	: Searches the pointer array.
 *  Method	: Binary search.
 *
 *  History:
 *  12-01-2010	: martinz  : creation
 *
 *  Notes	: This function's interface is like the generic bsearch(),
 *  		  which means the key should contain the address of the
 *  		  pointer to find. If this is too cumbersome, use
 *  		  PA_SearchDirect().
 *
 *  Returns	: - Index of element found on success
 *		  - -1 otherwise.
 *
 *------------------------------------------------------------------------
 */
int
PA_Search(
	PAHandle_p	paHandle, /* Pointer array handle */
	cvoidp		keyptr,	  /* Pointer to key of searched element */
	PA_Compare_f	compare,  /* Compare function */
	cvoidp		userdata) /* Userdata */
{
    cvoidp *element;	/* element found */

    if ((paHandle == NULL) || (compare == NULL)) {
	return -1;
    }

    PA_compare = compare;
    PA_userdata = userdata;
    element = bsearch(keyptr,
	              paHandle->pa_entries,
		      paHandle->pa_nrentries,
		      sizeof(cvoidp),
	  	      PA_indirectCompare2);
    PA_compare = NULL;
    PA_userdata = NULL;
    return element ? (element - paHandle->pa_entries) : -1;
} /* PA_Search() */

/*------------------------------------------------------------------------
 *  Function	: PA_SearchDirect
 *  Purpose	: Search without dereferencing the key.
 *  Method	: Call PA_Search() that dereferences its key.
 *
 *  History:
 *  27-08-2018	: zwaalm   : creation
 *
 *  Returns	: The result of PA_Search().
 *
 *------------------------------------------------------------------------
 */
int
PA_SearchDirect(
	PAHandle_p	paHandle, /* Pointer array handle */
	cvoidp		key,	  /* Key of searched element */
	PA_Compare_f	compare,  /* Compare function */
	cvoidp		userdata) /* Userdata */
{
    return PA_Search(paHandle, &key, compare, userdata);
} /* PA_SearchDirect() */

/*------------------------------------------------------------------------
 *  Function	: PA_Dump
 *  Purpose	: performs a dump of a pointer array using a user-supplied
 *		  element dump function
 *  Method	: loop along elements / call callback function
 *
 *  History:
 *  10-02-2010	: listj    : creation
 *
 *  Notes	: the callback function implementation is responsible for 
 *		  proper casting of element pointers in the callback function!
 *		  
 *
 *  Returns	: nothing
 *
 *------------------------------------------------------------------------
 */
void
PA_Dump(
	PAHandle_p		paHandle,    /* Pointer array handle */
	PA_DumpElement_f	dumpElement) /* Element dump function */
{
    size_t i = 0;

    if ((paHandle == NULL) || (dumpElement == PA_DUMPELEMENT_NOFUNC)) {
	return;
    }

    for (; i < PA_GetNrEntries(paHandle); i++) {
	dumpElement(i, PA_GetN(paHandle, i));
    }
} /* PA_Dump() */

/*------------------------------------------------------------------------
 *  Function	: PA_Walk
 *  Purpose	: Walk along the elements of a Pointer Array
 *  Method	: Obvious.
 *
 *  History:
 *  19-07-2012	: eck      : creation
 *
 *  Notes	:
 *  This function works like PA_Dump(), except that the interface conforms to
 *  other *Walk() interfaces in the library (ie. boolean return value and
 *  xtraparam parameter).
 *  This function stops if the callback returns FALSE.
 *
 *  Returns	: Whether a callback returned FALSE.
 *------------------------------------------------------------------------
 */
bool
PA_Walk(
	PAHandle_p		paHandle, /* Pointer array handle */
	bool			(*HandleElt)(size_t idx, cvoidp eltp, voidp xtraparam), /* Callback function */
	voidp			xtraparam) /* Extra data for Callbackfunction */
{
    if (paHandle == NULL) return TRUE;

    for (size_t i = 0; i < PA_GetNrEntries(paHandle); i++) {
	cvoidp	eltp = PA_GetN(paHandle, i);

	if (eltp != NULL) {
	    if (! HandleElt(i, PA_GetN(paHandle, i), xtraparam)) {
		return FALSE;
	    }
	}
    }
    return TRUE;
} /* PA_Walk() */
