/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ppalloc.c
 *  Description	: Allocation functions for the pp-library
 *  Author(s)	: Hans van Eck
 *  Creation	: 22-01-1996
 *
 *  History:
 *  22-01-1996	: eck      : creation
 *  22-02-1996	: eck      : Made changes to linked-list manipulation
 *  11-06-1999	: BdB	: Add more const's to pointer arguments
 *  06-03-2002	: eck      : Added ALLOC_NOWRAP support
 *  27-03-2002	: paul     : if defined(ALLOC_NOWRAP): initialise
 *                           pp_eltnext and pp_eltprev in PPLEltAlloc()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes	:
 *  This file contains the general allocation routines for title-elements.
 *  All title-elements are contaioned in the same struct, with pp_elttyp as
 *  the indication of what part of the union to use.
 *  The elements are never free()-d, but instead put on a free-list, from
 *  which they can be allocated again.  This reduces the malloc() and
 *  free() overhead, at the expense of extra memory usage. Another
 *  disadvantage is that purify is not good at finding leaks of PPElements
 *  this way. Using -DALLOC_NOWRAP when compiling will cause the wrapper to
 *  be disabled.
 *  The contents of subfields are malloc()-ed and free()-d.
 *========================================================================
 */

#pragma ident "$Id: ppalloc.c 331081 2015-11-12 14:23:40Z heesakkm $"

#include <stdlib.h>
#include <string.h>
#include <pica.h>
#include <picatypes.h>
#include <pp.h>
#include <er.h>
#include "pplocal.h"

#if ! defined(ALLOC_NOWRAP)
#define NR_ELT_ALLOC	16
static PPElement_p PPFreeEltList = NULL;
#endif

/*------------------------------------------------------------------------
 *  Function	: PPLEltAlloc
 *  Purpose	: Allocate a title element
 *  Method	:
 *  Look if free-list is empty.  If it is, allocate an array with a number
 *  of elements and initialise them to form a linked list.
 *  Take the first element from the linked list.
 *
 *  History:
 *  22-01-1996	: eck      : creation
 *  27-03-2002	: paul     : if defined(ALLOC_NOWRAP): initialise
 *                           pp_eltnext and pp_eltprev
 *
 *  Returns	: A pointer to an allocated element or NULL.
 *------------------------------------------------------------------------
 */
static PPElement_p
PPLEltAlloc(void)
{
    PPElement_p	tmp;

#if defined(ALLOC_NOWRAP)
    tmp = (PPElement_p)malloc(sizeof(PPElement_t));
#else
    int			i;

    if (PPFreeEltList == NULL) {
	PPFreeEltList = malloc(NR_ELT_ALLOC * sizeof(*PPFreeEltList));

	if (PPFreeEltList != NULL) {
	    for (i = 0; i < NR_ELT_ALLOC - 1; i++) {
		PPFreeEltList[i].pp_eltnext = &PPFreeEltList[i+1];
	    }
	    PPFreeEltList[NR_ELT_ALLOC - 1].pp_eltnext = NULL;
	}
    }
    tmp = PPFreeEltList;
    if (PPFreeEltList != NULL) {
	PPFreeEltList = PPFreeEltList->pp_eltnext;
    }
#endif

    if (tmp != NULL) {
	tmp->pp_eltnext = NULL;
	tmp->pp_eltprev = NULL;
	tmp->pp_elttyp = PPT_NONE;
    } else {
	PPerrno = PPE_NOMEM;
	(void) ERWrite(PPerrno, "PPLEltAlloc", "out of memory?");
    }
    return tmp;
} /* PPELltAlloc() */

/*------------------------------------------------------------------------
 *  Function	: PPLEltFree
 *  Purpose	: Free a title-element
 *  Method	: Put element back on free list
 *
 *  History:
 *  22-01-1996	: eck      : creation
 *------------------------------------------------------------------------
 */
static void
PPLEltFree(
	PPElement_p	elt)		/* element to free */
{
#if defined(ALLOC_NOWRAP)
    free(elt);
#else
    elt->pp_eltnext = PPFreeEltList;
    elt->pp_eltprev = NULL;
    PPFreeEltList = elt;
#endif
} /* PPLEltFree() */

/*------------------------------------------------------------------------
 *  Function	: PPTtlCreate
 *  Purpose	: Allocate a title-element
 *  Method	: Call of PPLEltAlloc().
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *
 *  Returns	: Pointer to element allocated or NULL.
 *------------------------------------------------------------------------
 */
PPTtl_p
PPTtlCreate(void)
{
    PPTtl_p	tmp;

    PPerrno = PPE_NOERROR;

    if ((tmp = PPLEltAlloc()) != NULL) {
	tmp->pp_elttyp = PPT_TITLE;
	tmp->PPTTL_firsttag = NULL;
    }
    return tmp;
} /* PPTtlCreate() */

/*------------------------------------------------------------------------
 *  Function	: PPTtlDestroy
 *  Purpose	: Deallocate a title (including tags and subfields).
 *  Method	:
 *  Repeated calls of PPTagDestroy() followed by PPLEltFree(title).
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *
 *  Notes	: Also deletes tags and subfields contained in the title.
 *------------------------------------------------------------------------
 */
void
PPTtlDestroy(
	PPTtl_p		title)		/* title to delete */
{
    PPerrno = PPE_NOERROR;

    if (title == NULL) return;
    else if (title->pp_elttyp != PPT_TITLE) {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPTagDestroy", "element is not title");
    }

    while (title->PPTTL_firsttag != NULL) {
	/* PPTagDestroy() changes title->PPTTL_firsttag */
	PPTagDestroy(title->PPTTL_firsttag);
    }
    PPLEltFree(title);
} /* PPTtlDestroy() */

/*------------------------------------------------------------------------
 *  Function	: PPTagCreate
 *  Purpose	: Allocate and initialise tag title-element.
 *  Method	: PPLEltAlloc() + initialisation.
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Returns	: pointer to tag or NULL
 *------------------------------------------------------------------------
 */
PPTag_p
PPTagCreate(
	PP_TagType_t	typ,		/* type of the tag */
	PP_Level_t	lev,		/* level */
	char		ind,		/* indicator */
	PP_Occurrence_t	occ)		/* occurrence */
{
    PPTag_p	tmp;

    PPerrno = PPE_NOERROR;

    if ((tmp = PPLEltAlloc()) != NULL) {
	tmp->pp_elttyp = PPT_TAG;
	tmp->PPTAG_firstsbf = NULL;
	tmp->PPTAG_typ = typ;
	tmp->PPTAG_lev = lev;
	tmp->PPTAG_ind = ind;
	tmp->PPTAG_occ = occ;
    }
    return tmp;
} /* PPTagCreate() */

/*------------------------------------------------------------------------
 *  Function	: PPTagDestroy
 *  Purpose	: Destroy a tag (including subfields) from a title
 *  Method	:
 *  Unlink tag form double-linked list, then use repeated calls of
 *  PPSbfDestroy() followed by PPLEltFree(title).
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *
 *  Notes	: Uses the address of the pointer to this tag,
 *------------------------------------------------------------------------
 */
void
PPTagDestroy(
	PPTag_p		tag)		/* tag to delete */
{
    PPerrno = PPE_NOERROR;

    if (tag == NULL) return;
    else if (tag->pp_elttyp != PPT_TAG) {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPTagDestroy", "element is not tag");
    }

    /* Unlink the tag from the doubly linked list.  */
    PPEltUnlink(tag);

    /* Now destroy all the subfields. */
    while (tag->PPTAG_firstsbf != NULL) {
	/* PPSbfDestroy() changes tag->PPTAG_firstsbf */
	PPSbfDestroy(tag->PPTAG_firstsbf);
    }
    PPLEltFree(tag);
} /* PPTagDestroy() */

/*------------------------------------------------------------------------
 *  Function	: PPSbfCreate
 *  Purpose	: Allocate and initialise subfield title-element.
 *  Method	: PPLEltAlloc() + initialisation.
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  31-01-1996	: eck      : Use PPLEltFree() instead of free()!
 *
 *  Notes	: Uses malloc().
 *
 *  Returns	: pointer to subfield or NULL
 *------------------------------------------------------------------------
 */
PPSbf_p
PPSbfCreate(
	char	typ,		/* type of the subfield */
	cvoidp	contents,	/* the contents */
	int	length)		/* the length (-1 if unknown) */
{
    PPSbf_p	tmp;
    voidp	new_contents;

    PPerrno = PPE_NOERROR;

    if ((tmp = PPLEltAlloc()) != NULL) {
	if (length < 0) {
	    length = strlen((char *)contents);
	}

	/* Allocate extra space for the '\0'-byte. */
	if ((new_contents = malloc(length + 1)) == NULL) {
	    PPLEltFree(tmp);
	    PPerrno = PPE_NOMEM;
	    tmp = NULL;
	} else {
	    tmp->pp_elttyp = PPT_SUBFIELD;
	    tmp->PPSBF_typ = typ;
	    tmp->PPSBF_length = length;
	    (void) memcpy(new_contents, contents, length);
	    *((char *) new_contents + length) = '\0';
	    tmp->PPSBF_contents = new_contents;
	}
     }
     return tmp;
} /* PPSbfCreate() */

/*------------------------------------------------------------------------
 *  Function	: PPSbfDestroy
 *  Purpose	: Destroy a subfield.
 *  Method	: Remove the subfield from the linked list and deallocate.
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *
 *  Notes	: Uses the address of the pointer to this subfield,
 *------------------------------------------------------------------------
 */
void
PPSbfDestroy(
	PPSbf_p		subfield)		/* subfield to delete */
{
    PPerrno = PPE_NOERROR;

    if (subfield == NULL) return;
    else if (subfield->pp_elttyp != PPT_SUBFIELD) {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPSbfDestroy", "element is not subfield");
    }

    /* Unlink the subfield from the doubly linked list. */
    PPEltUnlink(subfield);

    /* Deallocate the contents */
    free(subfield->PPSBF_contents);
    subfield->PPSBF_contents = NULL;

    PPLEltFree(subfield);
} /* PPSbfDestroy() */
