/*=====================================================================
 *  Copyright (c) 2004 by OCLC PICA
 *
 *  File        : ppcmp.c
 *  Description : functionality for comparing two PP titles
 *  Author(s)   : Guido van den Heuvel
 *  Creation    : 26-11-2004
 *
 *  History:
 *  26-11-2004  : Guy     : creation
 *
 *  Notes       :
 *=====================================================================
 */
#pragma ident "$Id: ppcmp.c 302219 2013-08-07 10:04:43Z sieburgw $"

#include <string.h>

#include <picatypes.h>
#include <pp.h>


/*---------------------------------------------------------------------
 *  Function    : compare_sbfs
 *  Purpose     : compares two subfields
 *  Method      : obvious
 *  
 *  History:
 *  26-11-2004  : Guy     : creation
 *
 *  Notes       :
 *
 *  Returns     : whether the subfields are equal (i.e., have the same
 *                type and contents)
 *---------------------------------------------------------------------
 */
static bool
compare_sbfs(
	     PPSbf_p sbf1,
	     PPSbf_p sbf2
	     )
{
    if (sbf1->PPSBF_typ != sbf2->PPSBF_typ ||
	sbf1->PPSBF_length != sbf2->PPSBF_length) {
	return FALSE;
    }
    return memcmp(sbf1->PPSBF_contents, sbf2->PPSBF_contents, 
		  sbf1->PPSBF_length) == 0;
}  /* compare_sbfs */


/*---------------------------------------------------------------------
 *  Function    : compare_tags
 *  Purpose     : compares two tags
 *  Method      : obvious
 *  
 *  History:
 *  26-11-2004  : Guy     : creation
 *
 *  Notes       :
 *
 *  Returns     : whether the tags are equal (i.e., tag headers are
 *                equal, and contain the same subfields)
 *---------------------------------------------------------------------
 */
static bool
compare_tags(
	     PPTag_p tag1,
	     PPTag_p tag2
	     )
{
    PPSbf_p sbf1;
    PPSbf_p sbf2;

    if (tag1->PPTAG_typ != tag2->PPTAG_typ ||
	tag1->PPTAG_ind != tag2->PPTAG_ind ||
	tag1->PPTAG_lev != tag2->PPTAG_lev ||
	tag1->PPTAG_occ != tag2->PPTAG_occ) {
	return FALSE;
    }
    sbf1 = tag1->PPTAG_firstsbf;
    sbf2 = tag2->PPTAG_firstsbf;
    while(sbf1 && sbf2) {
	if (!compare_sbfs(sbf1, sbf2)) {
	    return FALSE;
	}
	sbf1 = sbf1->pp_eltnext;
	sbf2 = sbf2->pp_eltnext;
    }
    return !sbf1 && !sbf2;
}  /* compare_tags */


/*---------------------------------------------------------------------
 *  Function    : PPCompare
 *  Purpose     : compares two lists of PPTag's
 *  Method      : obvious
 *  
 *  History:
 *  26-11-2004  : Guy     : creation
 *
 *  Notes       :
 *
 *  Returns     : whether the two lists are equal (i.e., both contain
 *                exactly the same tags).
 *---------------------------------------------------------------------
 */
bool
PPCompare(
	  PPTag_p first1,
	  PPTag_p last1,
	  PPTag_p first2,
	  PPTag_p last2
	  )
{
    PPTag_p tag1 = first1;
    PPTag_p tag2 = first2;

    while (tag1 && tag2 && tag1 != last1 && tag2 != last2) {
	if (!compare_tags(tag1, tag2)) {
	    return FALSE;
	}
	tag1 = tag1->pp_eltnext;
	tag2 = tag2->pp_eltnext;
    } 
    if (!tag1 && !tag2) {
	return TRUE;
    }
    if (!tag1 || !tag2) {
	return FALSE;
    }
    return tag1 == last1 && tag2 == last2 && compare_tags(tag1, tag2);
}  /* PPCompare */
