/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ppelement.c
 *  Description	: Double-linked list handling
 *  Author(s)	: Hans van Eck
 *  Creation	: 21-02-1996
 *
 *  History:
 *  21-02-1996	: eck      : creation
 *  17-10-2001  : jelle    : Added PPEltClone()
 *  16-11-2001  : paul     : Added PPEltAppend()
 *  26-02-2002	: paul     : PPEltUnlink(): previous may be NULL
 *  08-05-2002	: paul     : Added PPEltInsert()
 *  10-05-2002	: paul     : PPEltAppend() changed into PPEltListAppend()
 *                         : PPEltInsert() changed into PPEltListInsert()
 *  03-10-2003	: paul     : improved performance of PPEltListInsertBetween()
 *  01-09-2009	: sieburgw : Added PPEltDbg.
 *  01-04-2010	: paul     : Changed PPEltListClone() from static to extern
 *  20-05-2011	: paul     : CBS-2906 added PPEltParent()
 *                         : Fixed lint warnings
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ppelement.c 352214 2018-01-17 16:52:17Z vaneckh $"

#include <stdlib.h>

#include <pica.h>
#include <pp.h>
#include <er.h>
#include "pplocal.h"


/*------------------------------------------------------------------------
 *  Function	: PPEltParent
 *  Purpose	: Find the parent of an element
 *  Method	: 
 *
 *  History:
 *  20-05-2011	: paul     : creation
 *
 *  Notes	:
 *
 *  Returns	: The parent of the element
 *                or NULL if no parent present
 *
 *------------------------------------------------------------------------
 */
PPElement_p
PPEltParent(
	PPElement_p	child)
{
    PPElement_p	parent;

    for (parent = child; parent != NULL; parent = parent->pp_eltprev) {
	if (parent->pp_elttyp != child->pp_elttyp) {
	    return parent;
	}
    }

    return NULL;
} /* PPEltParent() */

/*------------------------------------------------------------------------
 *  Function	: PPEltUnlink
 *  Purpose	: Unlink an element from the double-linked list
 *  Method	: Let the previous and next point to each other.
 *
 *  History:
 *  22-02-1996	: eck      : creation
 *  26-02-2002	: paul     : previous may be NULL
 *
 *  Notes	:
 *  - Type checking is minimal.
 *  - There need not be a previous, nor a next.
 *------------------------------------------------------------------------
 */
void
PPEltUnlink(
	PPElement_p	elt)		/* The element to unlink */
{
    PPElement_p		prev;
    PPElement_p		next;

    prev = elt->pp_eltprev;
    next = elt->pp_eltnext;

    elt->pp_eltnext = NULL;
    elt->pp_eltprev = NULL;

    if (prev != NULL) {
	if (prev->pp_elttyp != elt->pp_elttyp) {
	    if (prev->pp_elttyp == PPT_TITLE) {
		prev->PPTTL_firsttag = next;
	    } else {
		prev->PPTAG_firstsbf = next;
	    }
	} else {
	    prev->pp_eltnext = next;
	}
    }
    if (next != NULL) {
	next->pp_eltprev = prev;
    }
} /* PPEltUnlink() */

/*------------------------------------------------------------------------
 *  Function	: PPEltInsertBetween
 *  Purpose	: Insert a single element between two other elements.
 *  Method	: 
 *
 *  History:
 *  22-02-1996	: eck      : creation
 *
 *  Notes	:
 *  This routine depend on the 'normal' values for prev, new and next.
 *  Especially, the field in 'prev' that needs to be changed (depending on
 *  the types) should point to 'next'.
 *  If 'next' is not NULL, its pp_eltprev should point to 'prev'.
 *------------------------------------------------------------------------
 */
void
PPEltInsertBetween(
	PPElement_p	prev,		/* The 'previous element */
	PPElement_p	new,		/* The element to insert */
	PPElement_p	next)		/* The 'next' element */
{
    if (prev->pp_elttyp != new->pp_elttyp) {
	if (prev->pp_elttyp == PPT_TITLE) {
	    prev->PPTTL_firsttag = new;
	} else {
	    prev->PPTAG_firstsbf = new;
	}
    } else {
	prev->pp_eltnext = new;
    }
    new->pp_eltprev = prev;
    new->pp_eltnext = next;
    if (next != NULL) {
	next->pp_eltprev = new;
    }
} /* PPEltInsertBetween() */

/*------------------------------------------------------------------------
 *  Function	: PPEltListInsertBetween
 *  Purpose	: Insert a list of element between to other element.
 *  Method	: 
 *
 *  History:
 *  10-05-2002	: paul     : creation
 *  03-10-2003	: paul     : improved performance if next is equal to NULL
 *
 *  Notes	:
 *  This routine depend on the 'normal' values for prev and next.
 *  Especially, the field in 'prev' that needs to be changed (depending on
 *  the types) should point to 'next'.
 *  If 'next' is not NULL, its pp_eltprev should point to 'prev'.
 *
 *  newlist is a list of elements of the same type; if newlist is preceded by
 *  an element, it is unlinked from this element
 *------------------------------------------------------------------------
 */
static void
PPEltListInsertBetween(
	PPElement_p	prev,		/* The 'previous element */
	PPElement_p	newlist,	/* The list of elements to insert */
	PPElement_p	next)		/* The 'next' element */
{
    if (prev->pp_elttyp != newlist->pp_elttyp) {
	if (prev->pp_elttyp == PPT_TITLE) {
	    prev->PPTTL_firsttag = newlist;
	} else {
	    prev->PPTAG_firstsbf = newlist;
	}
    } else {
	prev->pp_eltnext = newlist;
    }
    if (newlist->pp_eltprev != NULL) {
	/* unlink the list from the element preceding it */
	if (newlist->pp_eltprev->pp_elttyp == newlist->pp_elttyp) {
	    newlist->pp_eltprev->pp_eltnext = NULL;
	} else {
	    switch(newlist->pp_eltprev->pp_elttyp) {
	    case PPT_TITLE:
		newlist->pp_eltprev->PPTTL_firsttag = NULL;
		break;
	    case PPT_TAG: 
		newlist->pp_eltprev->PPTAG_firstsbf = NULL;
		break;
	    default:
    		PPerrno = PPE_TYPE;
		ERWrite(PPerrno, "PPEltListInsertBetween",
			"subfield precedes a title or a tag");
		break;
	    }
	}
    }
    
    newlist->pp_eltprev = prev;
    if (next != NULL) {
	while (newlist->pp_eltnext != NULL) newlist = newlist->pp_eltnext;
	newlist->pp_eltnext = next;
	next->pp_eltprev = newlist;
    }
} /* PPEltListInsertBetween() */

/*------------------------------------------------------------------------
 *  Function	: PPEltDestroy
 *  Purpose	: Destroy a title element
 *  Method	: Call the appropriate destruction routine
 *
 *  History:
 *  01-02-1996	: eck      : creation
 *  01-04-1999	: BdB	: No-op on NULL argument (like free())
 *
 *  Notes	: Does not use a switch (beware of possible compiler bug)
 *------------------------------------------------------------------------
 */
void
PPEltDestroy(
	PPElement_p	PPelt)		/* element to destroy */
{
    if (PPelt == NULL) return;
    if (PPelt->pp_elttyp == PPT_TITLE)		PPTtlDestroy(PPelt);
    else if (PPelt->pp_elttyp == PPT_TAG)	PPTagDestroy(PPelt);
    else					PPSbfDestroy(PPelt);
} /* PPEltDestroy() */

/*------------------------------------------------------------------------
 *  Function    : PPEltListClone
 *  Purpose     : Clone a linked list of title elements
 *  Method      : Call 'PPEltClone' for each element and link in a list
 *
 *  History:
 *  16-10-2001  : jelle    : creation
 *
 *  Notes       : 
 *
 *  Returns	: A clone of the element list or NULL (== error).
 *------------------------------------------------------------------------
 */
PPElement_p 
PPEltListClone(
	PPElement_p	PPelt)	    /* head of elementlist to clone */
{
    PPElement_p  *elt_target_next;
    PPElement_p   elt_source;
    PPElement_p   elt_target;
    PPElement_p   result = NULL;
    bool          stop = FALSE;

    elt_source = PPelt;      
    elt_target = NULL;
    elt_target_next = &result;
    while (!stop) {
        if (((*elt_target_next) = PPEltClone(elt_source)) != NULL) {
	    (*elt_target_next)->pp_eltprev = elt_target;
	    
	    elt_target = (*elt_target_next);
	    elt_target_next = &(elt_target->pp_eltnext);
	    elt_source = elt_source->pp_eltnext;
	} else 
	    stop = TRUE;
    }
    return (result);
} /* PPEltListClone() */

/*------------------------------------------------------------------------
 *  Function    : PPEltClone
 *  Purpose     : Clone a title element and the elements belonging to it.
 *  Method      : make a copy of either a title, a tag or a subfield
 *
 *  History:
 *  16-10-2001  : jd      : creation
 *
 *  Notes       : 
 *  If a title is cloned, all corresponding tags and subfields are cloned
 *  too. If a tag is cloned, all of it's subfields are cloned. If a 
 *  subfield is cloned, a copy of its data is made.
 *------------------------------------------------------------------------
 */
PPElement_p 
PPEltClone(
	PPElement_p	PPelt)		/* element to clone */
{
    PPElement_p clone = NULL;    
    if (PPelt == NULL) return (NULL);
    switch (PPelt->pp_elttyp) {
    case PPT_TITLE:
        clone = PPTtlCreate();
        if ((clone->PPTTL_firsttag =
	     PPEltListClone(PPelt->PPTTL_firsttag)) != NULL) {
	    clone->PPTTL_firsttag->pp_eltprev = clone;
	}
        break;
    case PPT_TAG:
        clone = PPTagCreate(PPelt->PPTAG_typ,
			    PPelt->PPTAG_lev,
			    PPelt->PPTAG_ind,
			    PPelt->PPTAG_occ);
	if ((clone->PPTAG_firstsbf =
	     PPEltListClone(PPelt->PPTAG_firstsbf)) != NULL) {
	    clone->PPTAG_firstsbf->pp_eltprev = clone;
	}
        break;
    case PPT_SUBFIELD:
        clone = PPSbfCreate(PPelt->PPSBF_typ,
			    PPelt->PPSBF_contents,
			    PPelt->PPSBF_length);
        break;
    case PPT_NONE:
        break;
    }
    return (clone);
} /* PPEltClone() */

/*------------------------------------------------------------------------
 *  Function	: PPEltListAppend
 *  Purpose	: Append an element to an other element
 *  Method	: 
 *
 *  History:
 *  16-11-2001	: paul     : creation
 *  10-05-2002	: paul     : elt may be a list of elements
 *
 *  Notes	: 
 *  It is possible to append a subfield to a title. The subfield will then be
 *  appended to the last subfield of the last tag in the title.
 *
 *  Returns	: The element appended or NULL (== error).
 *------------------------------------------------------------------------
 */
PPElement_p
PPEltListAppend(
	PPElement_p	oldelt,		/* Element to add to */
	PPElement_p	newelt)		/* Element to add */
{
    PPElement_p	next_elt;
    PPElement_p	prev_elt;
    PPElement_p	first_elt = NULL;
    bool        same_level;

    PPerrno = PPE_NOERROR;

    if (newelt == NULL) return NULL;
    if (oldelt == NULL) {
	PPerrno = PPE_TYPE;
	ERAbort(PPerrno, "PPEltListAppend",
		"cannot append to NULL pointer");
	return NULL;
    }
    
    same_level = oldelt->pp_elttyp == newelt->pp_elttyp;
    switch (newelt->pp_elttyp) {
    case PPT_SUBFIELD:
	if (!same_level
            && oldelt->pp_elttyp != PPT_TAG
	    && oldelt->pp_elttyp != PPT_TITLE) {
	    PPerrno = PPE_TYPE;
	    ERAbort(PPerrno, "PPEltListAppend",
		    "attempt to append subfield to "
		    "neither subfield nor tag nor title!");
	}
	if (oldelt->pp_elttyp == PPT_TAG) {
	    first_elt = oldelt->PPTAG_firstsbf;
	} else if (oldelt->pp_elttyp == PPT_TITLE) {
	    if (oldelt->PPTTL_firsttag == NULL) {
    		PPerrno = PPE_TYPE;
		ERWrite(PPerrno, "PPEltListAppend",
			"attempt to append subfield to a title without a tag");
		return NULL;
	    }
/*
 *          Append the subfield to the last subfield of the last tag in the
 *          title
 */
	    next_elt = oldelt->PPTTL_firsttag;
	    while (next_elt != NULL) {
		prev_elt = next_elt;
		next_elt = next_elt->pp_eltnext;
	    }
	    first_elt = prev_elt->PPTAG_firstsbf;
	}
	break;
    case PPT_TAG:
	if (!same_level
	    && oldelt->pp_elttyp != PPT_TITLE) {
	    PPerrno = PPE_TYPE;
	    (void) ERAbort(PPerrno, "PPEltListAppend",
			   "attempt to append tag to neither tag nor title!");
	}
	if (oldelt->pp_elttyp == PPT_TITLE) {
	    first_elt = oldelt->PPTTL_firsttag;
	}
	break;
    case PPT_TITLE:
	if (!same_level) {
	    PPerrno = PPE_TYPE;
	    ERAbort(PPerrno, "PPEltListAppend",
		    "attempt to append title to non-title!");
	}
	break;
    default:
	(void) ERAbort(PPerrno, "PPEltListAppend",
		       "attempt to append neither title nor tag nor subfield!");
	break;
    }
    
    prev_elt = oldelt;
    if (!same_level) {
	next_elt = first_elt;
	while (next_elt != NULL) {
	    prev_elt = next_elt;
	    next_elt = next_elt->pp_eltnext;
	}
    } else {
	next_elt = oldelt->pp_eltnext;
    }
    PPEltListInsertBetween(prev_elt, newelt, next_elt);

    return newelt;
} /* PPEltListAppend() */

/*------------------------------------------------------------------------
 *  Function	: PPEltListInsert
 *  Purpose	: Insert an element into an other element or list of elements
 *  Method	: 
 *
 *  History:
 *  08-05-2002	: paul     : creation
 *  10-05-2002	: paul     : elt may be a list of elements
 *
 *  Notes	: 
 *  It is possible to insert a subfield into a title. The subfield will then be
 *  inserted in front of the first subfield of the first tag in the title.
 *
 *  This function cannot insert in front of the first element of a list.
 *
 *  Returns	: The element appended or NULL (== error).
 *------------------------------------------------------------------------
 */
PPElement_p
PPEltListInsert(
	PPElement_p	oldelt,	/* Element in front of which to insert */
	PPElement_p	newelt)	/* Element to add */
{
    PPElement_p	next_elt;
    PPElement_p	prev_elt;
    bool        same_level;

    PPerrno = PPE_NOERROR;

    if (newelt == NULL) return NULL;
    if (oldelt == NULL) {
	PPerrno = PPE_TYPE;
	ERAbort(PPerrno, "PPEltListInsert",
		"cannot insert into NULL pointer");
	return NULL;
    }
    
    same_level = oldelt->pp_elttyp == newelt->pp_elttyp;

    prev_elt = oldelt->pp_eltprev;
    next_elt = oldelt;
    switch (newelt->pp_elttyp) {
    case PPT_SUBFIELD:
	if (!same_level
            && oldelt->pp_elttyp != PPT_TAG
	    && oldelt->pp_elttyp != PPT_TITLE) {
	    PPerrno = PPE_TYPE;
	    ERAbort(PPerrno, "PPEltListInsert",
		    "attempt to insert subfield into "
		    "neither subfield nor tag nor title!");
	}
	if (oldelt->pp_elttyp == PPT_TAG) {
	    next_elt = oldelt->PPTAG_firstsbf;
	    prev_elt = oldelt;
	} else if (oldelt->pp_elttyp == PPT_TITLE) {
	    if (oldelt->PPTTL_firsttag == NULL) {
    		PPerrno = PPE_TYPE;
		ERWrite(PPerrno, "PPEltListInsert",
			"attempt to insert subfield in front of a title without a tag");
		return NULL;
	    }
/*
 *          Insert the subfield in front of the first subfield of the first tag
 *          in the title
 */
	    next_elt = oldelt->PPTTL_firsttag->PPTAG_firstsbf;
	    prev_elt = oldelt->PPTTL_firsttag;
	}
	break;
    case PPT_TAG:
	if (!same_level
	    && oldelt->pp_elttyp != PPT_TITLE) {
	    PPerrno = PPE_TYPE;
	    (void) ERAbort(PPerrno, "PPEltListInsert",
			   "attempt to insert tag into neither tag nor title!");
	}
	if (oldelt->pp_elttyp == PPT_TITLE) {
	    next_elt = oldelt->PPTTL_firsttag;
	    prev_elt = oldelt;
	}
	break;
    case PPT_TITLE:
	if (!same_level) {
	    PPerrno = PPE_TYPE;
	    ERAbort(PPerrno, "PPEltListInsert",
		    "attempt to insert title in front of non-title!");
	}
	break;
    default:
	(void) ERAbort(PPerrno, "PPEltListInsert",
		       "attempt to insert neither title nor tag nor subfield!");
	break;
    }
    
    if (prev_elt == NULL) {
	PPerrno = PPE_TYPE;
	ERAbort(PPerrno, "PPEltListInsert",
		"cannot insert in front of list");
	return NULL;
    }
    
    PPEltListInsertBetween(prev_elt, newelt, next_elt);	

    return newelt;
} /* PPEltListInsert() */

/*------------------------------------------------------------------------
 *  Function	: PPEltDbg
 *  Purpose	: This function may be useful if you want to look into
 *		: the structure of a PPElement
 *  Method	: 
 *
 *  History:
 *  01-09-2009	: sieburgw : creation
 *
 *  Notes	: The initial call should pass '0' for the level
 *		: parameter and NULL for the 'prev' parameter
 *		: The 'buf' parameter should be a created, but
 *		: empty CS structure handle.
 *		: And 'elt' should be the element to be debugged.
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
void
PPEltDbg(CSBuf_p buf, PPElement_p elt, int level, PPElement_p prev)
{
    CS_DYNAMIC	(tmp);
    int i;

    if (NULL == buf){
	return;
    }
    for (i = 0; i < level; i++){
	CSAppend(buf, "  ");
    }
    if (NULL == elt){
	CSAppend(buf, "<NULL>\n");
	return;
    }
    if (elt->pp_eltprev != prev){
	CSPrintf(tmp, "<problem with previous pointer: %p != %p>\n",
		 (voidp)elt->pp_eltprev, (voidp)prev );
	CSAppend(buf, CSContents(tmp));
	for (i = 0; i < level; i++){
	    CSAppend(buf, "  ");
	}
    }
    CSClear(tmp);

    switch (elt->pp_elttyp) {
	case  PPT_TITLE:
	    CSPrintf(tmp, "TTL: \n");
	    CSAppend(buf, CSContents(tmp));
	    PPEltDbg(buf, elt->PPTTL_firsttag, level + 1, elt);
	    break;
	case  PPT_TAG:
	    CSPrintf(tmp, "TAG: %1d%02d%c", elt->PPTAG_lev, abs(elt->PPTAG_typ), elt->PPTAG_ind);
	    CSAppend(buf, CSContents(tmp));
	    if (elt->PPTAG_occ){
		CSPrintf(tmp, "/%03d\n", elt->PPTAG_occ);
	    } else {
		CSPrintf(tmp, "\n");
	    }
	    CSAppend(buf, CSContents(tmp));
	    PPEltDbg(buf, elt->PPTAG_firstsbf, level + 1, elt);
	    break;
	case  PPT_SUBFIELD:
	    CSPrintf(
		    tmp,
		    "SBF: $%c, len: %03d: \"%s\"\n",
		    elt->PPSBF_typ,
		    elt->PPSBF_length,
		    (char *)elt->PPSBF_contents);
	    CSAppend(buf, CSContents(tmp));
	    break;
	default:
	    break;
    }
    PPEltDbg(buf, elt->pp_eltnext, level, elt);
    CSDestroy(tmp);
}
