/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File        : ppextern.c
 *  Description : Convert to and from external pica+ format
 *  Author(s)   : Hans van Eck
 *  Creation    : 23-01-1996
 *
 *  History:
 *  23-01-1996  : eck      : creation
 *  22-02-1996  : eck      : More flexibility in external format.
 *  31-07-1997  : eck      : Be more careful with PPerrno.
 *  12-10-2000  : paul     : PPToNormal() introduced
 *  18-10-2000  : paul     : PPFromNormal() introduced,
 *                           PM_range_from_bits() used in PPToExtern()
 *                           PM_range_to_bits() used in PPFromExtern()
 *  01-12-2000  : paul     : PPToExtern() also converts individual tags
 *  11-03-2002  : paul     : PPBitsToExtern() made extern
 *  07-05-2002  : paul     : PPToExtern() and PPToNormal() suppress control
 *                           characters
 *  01-10-2002	: paul     : Added PPToExternCreate(), PPToExternAppend(),
 *                           PPToNormalCreate() and PPToNormalAppend()
 *  28-01-2003  : paul     : Fixed bug in GetTag()
 *  07-04-2003	: paul     : Added PPToExternAppendCS(), PPToNormalAppendCS(),
 *                           PPToExternCreateCS() and PPToNormalCreateCS()
 *  12-10-2004  : guy      : Optimized TagToExternAppend()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  11-02-2009	: martinz  : Added PPFromExtern2(), PPFromNormal2(), PPFromX().
 *  20-05-2011	: paul     : Fixed lint warning
 *
 *  Notes       :
 *  In the external pica+ format, a title has the following syntax:
 *
 *  title       = <tag> [ <tag> ]* \'0' ;
 *  tag         = <tagheader> [ ' ' <subfield>  [ <subfield> ]* ] <tag_end> ;
 *  tagheader   = <tag-type> <ind> ['/' <occ>] ;
 *  tag_end     = single character (default 30)
 *  subfield    = <sbf_marker> <sbf-type> <contents> ;
 *  sbf_marker  = single character (default 31)
 *  tag-type    = three digits
 *  ind         = single letter of @, A-Z
 *  occ         = two or three digits
 *  sbf-type    = either single alphanumeric character or '@'
 *
 *========================================================================
 */

#pragma ident "$Id: ppextern.c 363677 2019-01-10 15:10:10Z heesakkm $"

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <pica.h>
#include <picamisc.h>
#include <picatypes.h>
#include <cs.h>
#include <pp.h>
#include <er.h>
#include <fcv.h>
#include "pplocal.h"

char PP_ext_tagend = (char) 30;         /* record-separator */
char PP_ext_sbfmark = (char) 31;        /* unit-separator */

#define SEP_OCC         '/'             /* occurrence separator */

#define USERBITS(tag, sbf_typ) \
               (   ((tag)->PPTAG_typ == 1  ) \
                && ((tag)->PPTAG_lev == 0  ) \
                && ((tag)->PPTAG_ind == '@') \
                && ((sbf_typ) == '0'))


/*------------------------------------------------------------------------
 *  Function    : GetTag
 *  Purpose     : Retrieve the tag from extrnal format (not the subfields).
 *  Method      : obvious
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *  28-01-2003  : paul     : always set PPerrno when returning FALSE;
 *                         : empty tag with space processed as empty tag
 *                           withoutace
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  25-09-2017	: smitp    : CBS-7643 Use PP_OCCURRENCE_ANY_NORMALIZED
 *
 *  Notes       : Is very strict in the format it accepts.
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
static bool
GetTag(
        char               *tag,           /* title in external pica+ */
        PP_TagType_t       *typ_p,         /* place to store the type */
        PP_Level_t         *lev_p,         /* place to store the level */
        char               *ind_p,         /* place to store indicator */
        PP_Occurrence_t    *occ_p)         /* place to store occurrence */
{
    /* Test for three digits. */
    if (! isdigit(tag[0])
        || ! isdigit(tag[1])
        || ! isdigit(tag[2])) {
        PPerrno = PPE_FORMAT;
        return FALSE;
    }
    /* Get the level (one digit) */
    *lev_p = tag[0] -'0';
    tag++;

    /* Get the type (two digits) */
    *typ_p = 10 * (tag[0] -'0') + tag[1] - '0';
    tag += 2;           /* Now points to the indicator */

    if (! PP_INDICATOR_OK(tag[0])) {
        PPerrno = PPE_FORMAT;
        return FALSE;
    }
    *ind_p = *tag++;
    if (tag[0] == SEP_OCC) {
        /* The occurrence consists of a '*': meaning any
	 * occurrence, or one or more digits, for a specific
	 * occurrence
	 */
	PP_Occurrence_t occ_tmp;

        tag++;
	if (*tag == PP_OCCURRENCE_ANY_NORMALIZED){
	    occ_tmp = PP_ANYOCC;
	    tag++;
	} else {
	    if (sscanf(tag, "%hd", &occ_tmp) != 1) {
		PPerrno = PPE_FORMAT;
		return FALSE;
	    }
	    while (isdigit(tag[0])) tag++;
	}
	*occ_p = occ_tmp;
    } else {
        *occ_p = 0;
    }
    /* Be as strict as possible: the tag must be empty, or followed by a
     * ' ', or followed by a ' ' and a subfield.
     */
    if ((tag[0] == '\0')
	|| ((tag[0] == ' ') && (tag[1] == '\0'))
	|| ((tag[0] == ' ') && (tag[1] == PP_ext_sbfmark))) {
	return TRUE;
    }

    PPerrno = PPE_FORMAT;
    return FALSE;
} /* GetTag() */

/*------------------------------------------------------------------------
 *  Function    : PPBitsAppend
 *  Purpose     : Append a subfield containing a bit-array.
 *  Method      : PPSbfAppend() and PP_range_to_bits().
 *
 *  History:
 *  18-10-2000  : paul     : Creation
 *  21-10-2004  : eric     : Enlarged bitarr to be able to handle records
 *                           with ilns above PIC_ILN_MAX.
 *
 *  Notes:
 *  This function is used to append a $0 to tag 001@.
 *  If the number of bytes in the bit-array is 0, two 0-byte will be appended
 *  to prevent that the subfield is empty.
 *  PIC_USERBITS_NONE is a string representing userbits that are all set to 0.
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
static PPSbf_p
PPBitsAppend(
        PPElement_p     elt,            /* element to append to */
        char            typ,            /* type of the subfield */
        const char      *contents)      /* contents */
{
    byte        bitarr[PIC_USERBITS_BUF_SIZE];
    size_t      bitlen;

    if (STREQ(contents, PIC_USERBITS_NONE)) {
        bitlen = 2;
        (void) memset(bitarr, '\0', bitlen);
    } else if (   PM_range_to_bits(bitarr, sizeof bitarr, &bitlen,
                                   contents, 1)
               == PM_RANGE_ERROR) {
        if (bitlen > sizeof bitarr) PPerrno = PPE_OVERFLOW;
        else                        PPerrno = PPE_FORMAT;
        return NULL;
    }
    if (bitlen == 0) bitlen = 2;
    return PPSbfAppend(elt, typ, bitarr, (int) bitlen);
} /* PPBitsAppend() */

/*------------------------------------------------------------------------
 *  Function    : PPInTag
 *  Purpose     : Retrieve a tag from the input and return the handle.
 *  Method      :
 *  First get type-indicator-occurrence, then loop over the subfields.
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *  18-10-2000  : paul     : Use PPBitsAppend() to append 001@$0
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  09-11-2018	: zwaalm   : CBS-8437 Fixed invalid read error.
 *
 *  Notes       :
 *  The tag is '\0'-terminated (the 'xd' is overwritten).
 *
 *  Returns     : A pointer to the tag (handle) or NULL.
 *------------------------------------------------------------------------
 */
static PPTag_p
PPInTag(
        PPElement_p     elt,            /* title or tag to append to */
        char            *tag_start)     /* start of the tag */
{
    PPTag_p           tag;
    PP_TagType_t      typ;
    PP_Level_t        lev;
    char              ind;
    PP_Occurrence_t   occ;
    char              *sbf_start;
    char              *sbf_end;

    if (! GetTag(tag_start, &typ, &lev, &ind, &occ)) {
        return NULL;
    }
    tag = PPTagAppend(elt, typ, lev, ind, occ);
    if (tag == NULL) { return NULL; }

    sbf_start = strchr(tag_start, PP_ext_sbfmark);
    while (sbf_start != NULL) {
        /* Now loop over the subfields */
        sbf_start++;
	if (*sbf_start == '\0') return NULL;	/* missing data */
        sbf_end = strchr(sbf_start + 1, PP_ext_sbfmark);

        if (sbf_end != NULL) {
            *sbf_end = '\0';
        }
        if (USERBITS(tag, *sbf_start)) {
            if (PPBitsAppend(tag, *sbf_start, sbf_start + 1) == NULL) {
                break;
            }
        } else if (PPSbfAppend(tag, *sbf_start, sbf_start + 1, -1) == NULL) {
            break;
        }

        sbf_start = sbf_end;
    }
    if (sbf_start != NULL) {
        PPErrno_t       tmp = PPerrno;

        PPTagDestroy(tag);
        PPerrno = tmp;

        return NULL;
    }
    return tag;
} /* PPInTag() */

/*------------------------------------------------------------------------
 *  Function	: PPTagFromExtern
 *  Purpose	: Convert an external Pica+ tag to PP-handle
 *  Method	: 
 *
 *  History:
 *  11-03-2016	: paul     : CBS-6572 creation
 *  03-05-2016	: paul     : CBS-6572 test parameter for NULL, not 'tag'
 *
 *  Returns	:
 *  - The created PP-handle
 *  - or NULL if an error occurred
 *
 *------------------------------------------------------------------------
 */
PPTag_p
PPTagFromExtern(
        const char	*tag_start)	// tag in external pica+
{
    static PPTtl_p	title = NULL;
    static char		*tag_buffer = NULL;
    static size_t	tag_bufsiz = 0;

    PPTag_p		tag;
    const char		*tag_end;

    PPerrno = PPE_NOERROR;

    if (tag_start == NULL) {
	return NULL;
    }

    // Create title to append to, unless already done
    if (title == NULL) {
	title = PPTtlCreate();
	if (title == NULL) {
	    ERWrite(PPerrno, __func__, "Can't create title");
	    return NULL;
	}
    }

    // Find the end of the input tag
    tag_end = strchr(tag_start, PP_ext_tagend);
    if (tag_end == NULL) {
	tag_end = tag_start + strlen(tag_start);
    }

    // Make sure the tag fits into our buffer
    if (tag_bufsiz < tag_end - tag_start + 1) {
	size_t	new_tag_bufsiz = tag_end - tag_start + 1;
	voidp	tmpp = realloc(tag_buffer, new_tag_bufsiz);

	if (tmpp == NULL) {
	    PPerrno = PPE_NOMEM;
	    ERWrite(PPerrno, __func__, "Can't realloc tag_buffer");
	    return NULL;
	}

	tag_buffer = (char *)tmpp;
	tag_bufsiz = new_tag_bufsiz;
    }

    // Copy tag into buffer
    (void)memcpy(tag_buffer, tag_start, tag_end - tag_start);
    tag_buffer[tag_end - tag_start] = '\0';

    // Convert tag to PP-handle, and append to title
    tag = PPInTag(title, tag_buffer);
    if (tag == NULL) {
	// Restore tag_buffer; PPInTag() may have corrupted it
	(void)memcpy(tag_buffer, tag_start, tag_end - tag_start);
	tag_buffer[tag_end - tag_start] = '\0';

	ERWrite(PPerrno, __func__, "Can't convert external Pica+ tag %s",
		tag_buffer);
	return NULL;
    }

    // Unlink tag from title and return it
    PPEltUnlink(tag);
    return tag;
} /* PPTagFromExtern() */

/*------------------------------------------------------------------------
 *  Function    : PPFromExtern
 *  Purpose     : Convert a title from external pica+ to a handle.
 *  Method      :
 *  1- Create title handle
 *  2- Convert a tag
 *  3- Add it to the title
 *  4- If there is another tag, go to step 2
 *  5- Return title-handle
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *  19-10-2000  : paul     : External Pica+ need not be terminated with a
 *                           tag-separator
 *
 *  Notes       : Modifies the title-buffer.
 *
 *  Returns     : The title-handle or NULL.
 *------------------------------------------------------------------------
 */
PPTtl_p
PPFromExtern(
        PPTtl_p title,                  /* Title-handle to add to or NULL */
        char    *ttlbuf)                /* Buffer with the title */
{
    char                *tag_start;
    PPElement_p         elt;
    char                *tag_end;
    bool                tagend_found = TRUE;

    PPerrno = PPE_NOERROR;

    if (title == NULL) title = PPTtlCreate();
    if (title == NULL) { return NULL; }

    tag_start = ttlbuf;
    elt = title;

    while ((tag_start != NULL) && (*tag_start != '\0')) {
        tag_end = strchr(tag_start, PP_ext_tagend);
        if (tag_end == NULL) {
            tag_end = tag_start + strlen(tag_start);
            tagend_found = FALSE;
        }

        *tag_end = '\0';
        elt = PPInTag(elt, tag_start);
        if (elt == NULL) { break; }

        if (tagend_found) tag_start = tag_end + 1;
        else              tag_start = tag_end;
    }
    if (PPerrno != PPE_NOERROR) {
        PPErrno_t       tmp = PPerrno;

        (void) PPTtlDestroy(title);
        title = NULL;
        PPerrno = tmp;
    }
    return title;
} /* PPFromExtern() */

/*------------------------------------------------------------------------
 *  Function    : PPOutC
 *  Purpose     : Output a character to buffer.
 *  Method      :
 *  Put the character using ttlsiz as the offset for the character.
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *
 *  Notes       : Also increases ttlsiz if there is buffer overflow.
 *
 *------------------------------------------------------------------------
 */
static void
PPOutC(
        char    *ttlbuf,                /* Output buffer */
        size_t  bufsiz,                 /* Size of the buffer */
        size_t  *ttlsiz,                /* pointer to current title size */
        char    c)                      /* character to output */
{
    if (*ttlsiz < bufsiz) {
        ttlbuf[*ttlsiz] = c;
    }
    (*ttlsiz)++;
} /* PPOutC() */

/*------------------------------------------------------------------------
 *  Function    : PPOutS
 *  Purpose     : Output a tag-code or the contents of a subfield.
 *  Method      : Repeated calls of PPOutC()
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *  20-10-2000  : paul     : len parameter added
 *  07-05-2002  : paul     : suppress control characters
 *
 *  Notes       : ttlsiz is modified (by PPOutC()).
 *  This function will suppress any control character, since these are allowed
 *  in neither external tag-codes nor in the contents of external subfields. 
 *------------------------------------------------------------------------
 */
static void
PPOutS(
        char            *ttlbuf,        /* Output buffer */
        size_t          bufsiz,         /* Size of the buffer */
        size_t          *ttlsiz,        /* pointer to current title size */
        const char      *s,             /* string to output */
        int		len)            /* length of string to output or -1 */
{
    if (len < 0) {
        while (*s) {
	    if (!iscntrl(*(byte *)s)) PPOutC(ttlbuf, bufsiz, ttlsiz, *s);
            s++;
        }
    } else {
        while (len > 0) {
	    if (!iscntrl(*(byte *)s)) {
		PPOutC(ttlbuf, bufsiz, ttlsiz, *s);
            }
            s++;
            len--;
        }
    }
} /* PPOutS() */

/*------------------------------------------------------------------------
 *  Function    : SbfToExternAppend
 *  Purpose     : Output the contents of a subfield.
 *  Method      : Repeated calls of CSAppendC()
 *
 *  History:
 *  29-01-1996  : eck      : created as PPOutS()
 *  20-10-2000  : paul     : len parameter added
 *  07-05-2002  : paul     : suppress control characters
 *  07-04-2003	: paul     : Renamed and adapted for counted strings
 *  11-10-2004  : MJL      : Changed CSAppendC-calls to use CSAppend(N).
 *                           The code runs a lot faster now.
 *
 *  Notes       : 
 *  This function will suppress any control character, since these are allowed
 *  in neither external tag-codes nor in the contents of external subfields. 
 *------------------------------------------------------------------------
 */
static bool
SbfToExternAppend(
    CSBuf_p		ttl,		/* buffer to append to */
    const char		*sbf)		/* subfield contents to output */
{  
    char	c;
    const char  *start = sbf;

    while ((c = *sbf) != '\0') {
  	if (iscntrl((byte)c)) {
  	    if ( !CSAppendN( ttl, start, sbf - start ) ) {
  		PPerrno = PPE_NOMEM;
  		return FALSE;
  	    } else {
  		start = sbf + 1;
  	    }
  	}
  	sbf++;
    }

    if ( sbf > start ) {
  	if ( !CSAppend( ttl, start ) ) {
  	    PPerrno = PPE_NOMEM;
  	    return FALSE; 
  	} 
    } 

    return TRUE; 
} /* SbfToExternAppend() */

/*------------------------------------------------------------------------
 *  Function    : PPBitsToExtern
 *  Purpose     : Output a bit-array
 *  Method      : PM_range_from_bits()
 *
 *  History:
 *  12-10-2000  : paul     : creation
 *  11-03-2002  : paul     : made extern 
 *
 *  Notes       : ttlsiz is modified (by PPOutC()).
 *     
 *  Returns     : Whether no overflow occurred.
 *------------------------------------------------------------------------
 */
bool
PPBitsToExtern(
    char		*ttlbuf,	/* Output buffer */
    size_t		bufsiz,		/* Size of the buffer */
    size_t		*ttlsiz,	/* pointer to current title size */
    const byte		*bitbuf,	/* buffer containing bits */
    size_t		bitlen)		/* size of buffer containing bits */
{
    unsigned long nr_ilns;

    if (*ttlsiz > bufsiz) {
        *ttlsiz = bufsiz;   /* Better safe than sorry */
    }
    nr_ilns = PM_range_from_bits(ttlbuf + *ttlsiz, bufsiz - *ttlsiz,
                                 bitbuf, bitlen, 1);
    if (nr_ilns == 0) {
        PPOutS(ttlbuf, bufsiz, ttlsiz, PIC_USERBITS_NONE, -1);
    } else {
        *ttlsiz += strlen(ttlbuf + *ttlsiz);

        if (nr_ilns == PM_RANGE_ERROR) {
	    return FALSE;
        }
    }
    return TRUE;
} /* PPBitsToExtern() */

/*------------------------------------------------------------------------
 *  Function    : PPBitsToExternAppend
 *  Purpose     : Output a bit-array as a list of ranges
 *  Method      : PM_range_from_bits()
 *
 *  History:
 *  07-04-2003  : paul     : creation
 *
 *  Notes       :
 *     
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
bool
PPBitsToExternAppend(
    CSBuf_p		ttl,		/* buffer to append to */
    const byte		*bitbuf,	/* buffer containing bits */
    size_t		bitlen)		/* size of buffer containing bits */
{
    char		ranges[PIC_ILN_RANGEBUF_SIZE];
    unsigned long	nr_ilns;

    nr_ilns = PM_range_from_bits(ranges, sizeof ranges, bitbuf, bitlen, 1);
    if (nr_ilns == 0) {
        if (!CSAppend(ttl, PIC_USERBITS_NONE)) {
	    PPerrno = PPE_NOMEM;
	    return FALSE;
	}
    } else if (nr_ilns == PM_RANGE_ERROR) {
	PPerrno = PPE_OVERFLOW;
	return FALSE;
    } else if (!CSAppend(ttl, ranges)) {
	PPerrno = PPE_NOMEM;
	return FALSE;
    }
    return TRUE;
} /* PPBitsToExternAppend() */

/*------------------------------------------------------------------------
 *  Function    : PPExtOutTag
 *  Purpose     : Output the tagheader and subfields.
 *  Method      : obvious
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *  14-02-2005  : MJL      : Replaced CSPrintf for better performance. 
 *  09-03-2005  : michiel  : extremely small fix for "XXXX/5" instead
 *                           of "XXXX/05"
 *------------------------------------------------------------------------
 */
static void
PPExtOutTag(
        PPTag_p		tag,            /* The tag to output */
        char            *ttlbuf,        /* output buffer */
        size_t          bufsiz,         /* size of the output buffer */
        size_t          *ttlsiz)        /* pointer to current title size */
{
    PPTag_p     sbf;

    if (tag->pp_elttyp != PPT_TAG) {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPToExtern.PPExtOutTag", "Not a tag!");
    }

    /* Skip hidden fields. */
    if (tag->PPTAG_typ < 0) return;

    /* Decode the tag by hand instead of using CSPrintf for
       performance reasons. */ 
    PPOutC(ttlbuf, bufsiz, ttlsiz, tag->PPTAG_lev + '0' );
    PPOutC(ttlbuf, bufsiz, ttlsiz, tag->PPTAG_typ / 10 + '0' );
    PPOutC(ttlbuf, bufsiz, ttlsiz, tag->PPTAG_typ % 10 + '0' );
    PPOutC(ttlbuf, bufsiz, ttlsiz, tag->PPTAG_ind );

    if ( tag->PPTAG_occ != 0 ) {
	char occ_buf[ PIC_OCCURRENCE_LEN + 1 ];
	int i = PIC_OCCURRENCE_LEN;
	int occ = tag->PPTAG_occ;

	occ_buf[ i-- ] = '\0';
	while ( occ >= 10 ) {
	    occ_buf[ i-- ] = occ % 10 + '0';
	    occ = occ / 10;
	}
	occ_buf[ i-- ] = occ + '0';

        /* use "XXXX/05" instead of "XXXX/5" */
	if (tag->PPTAG_occ < 10) {
	    occ_buf[ i-- ] = '0';
	}

	PPOutC(ttlbuf, bufsiz, ttlsiz, SEP_OCC );
	PPOutS(ttlbuf, bufsiz, ttlsiz, occ_buf + i + 1, -1 );
    }

    sbf = tag->PPTAG_firstsbf;
    if (sbf != NULL) {
        /* Output separating ' '. */
        PPOutC(ttlbuf, bufsiz, ttlsiz, ' ');

        /* Output the subfields. */
        while (sbf != NULL) {
            PPOutC(ttlbuf, bufsiz, ttlsiz, PP_ext_sbfmark);
            PPOutC(ttlbuf, bufsiz, ttlsiz, sbf->PPSBF_typ);
            if (USERBITS(tag, sbf->PPSBF_typ)) {
                if (!PPBitsToExtern(ttlbuf, bufsiz, ttlsiz,
				    sbf->PPSBF_contents,
				    sbf->PPSBF_length)) {
/*
 *                  Range doesn't fit into the buffer. Fill the rest of the
 *                  buffer with spaces, to indicate buffer overflow.
 */
		    (void) memset(ttlbuf + *ttlsiz, ' ', bufsiz - *ttlsiz);
		    *ttlsiz = bufsiz;
		}
            } else {
                PPOutS(ttlbuf, bufsiz, ttlsiz, sbf->PPSBF_contents,
                       sbf->PPSBF_length);
            }
            sbf = sbf->pp_eltnext;
        }
    }
    /* And the terminating tag-separator. */
    PPOutC(ttlbuf, bufsiz, ttlsiz, PP_ext_tagend);
} /* PPExtOutTag() */

/*------------------------------------------------------------------------
 *  Function    : PPExtOutTagCS
 *  Purpose     : Output the tagheader and subfields into a counted string.
 *  Method      : obvious
 *
 *  History:
 *  29-01-1996  : eck      : creation as PPExtOutTag()
 *  07-04-2003	: paul     : Renamed and adapted for counted strings
 *  12-10-2004  : guy      : speedup
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  19-11-2004  : guy      : removed bug from occurrence change
 *  25-06-2012	: eck      : Renamed to PPExtOutTagCS() and made external
 *  25-09-2017	: smitp    : CBS-7643 Use PP_OCCURRENCE_ANY_NORMALIZED
 *------------------------------------------------------------------------
 */
bool
PPExtOutTagCS(
    CSBuf_p		ttl,		/* buffer to append to */
    PPTag_p		tag)		/* The tag to output */
{
    char        tmp[32];      /* like 003@/05 $0 */
    char        *s = tmp;
    PPTag_p     sbf;

    if (tag->pp_elttyp != PPT_TAG) {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPToExternAppendCS.PPExtOutTagCS", "Not a tag!");
    }

    /* Skip hidden fields. */
    if (tag->PPTAG_typ < 0) return TRUE;

    s[0] = tag->PPTAG_lev + '0';
    s[1] = tag->PPTAG_typ / 10 + '0';
    s[2] = tag->PPTAG_typ % 10 + '0';
    s[3] = tag->PPTAG_ind;
    s += 4;

    if (tag->PPTAG_occ != 0) {
	char tmp2[16];
	if (tag->PPTAG_occ == PP_ANYOCC){
	    sprintf(tmp2, "%c%c", SEP_OCC, PP_OCCURRENCE_ANY_NORMALIZED);
	} else {
	    sprintf(tmp2, "%c%02d", SEP_OCC, (int) tag->PPTAG_occ);
	}
	strcpy(s, tmp2);
	s += strlen(tmp2);
    }

    sbf = tag->PPTAG_firstsbf;
    if (sbf == NULL) {
	s[0] = '\0';
	if (!CSAppend(ttl, tmp)) {
	    PPerrno = PPE_NOMEM;
	    return FALSE;
	}
    } else {
        /* Output separating ' '. */
	*s++ = ' ';

        /* Output the subfields. */
        while (sbf != NULL) {
	    s[0] = PP_ext_sbfmark;
	    s[1] = sbf->PPSBF_typ;
	    s[2] = '\0';
	    CSAppend(ttl, tmp);
	    s = tmp;

            if (USERBITS(tag, sbf->PPSBF_typ)) {
		if (!PPBitsToExternAppend(ttl, sbf->PPSBF_contents,
					  sbf->PPSBF_length)) {
		    PPerrno = PPE_NOMEM;
		    return FALSE;
		}
            } else if (!SbfToExternAppend(ttl, sbf->PPSBF_contents)) {
		PPerrno = PPE_NOMEM;
		return FALSE;
            }
            sbf = sbf->pp_eltnext;
        }
    }

    /* And the terminating tag-separator. */
    if (!CSAppendC(ttl, PP_ext_tagend)) {
	PPerrno = PPE_NOMEM;
	return FALSE;
    }

    return TRUE;
} /* PPExtOutTagCS() */

/*------------------------------------------------------------------------
 *  Function	: PPTagToNormal
 *  Purpose	: Convert a single tag to normalized Pica+
 *
 *  History:
 *  05-12-2017	: smitp    : creation
 *
 *  Returns	:
 *  - (pointer to) buffer that contains the normalized Pica+
 *  - or NULL if an error occurred
 *
 *------------------------------------------------------------------------
 */
CSBuf_p
PPTagToNormal(
	CSBuf_p		csbuf,		// output buffer or NULL: create it
	PPTag_p		tag)		// the tag to print
{
    CSBuf_p		created_csbuf = NULL;
    char		PP_ext_tagend_old  = PP_ext_tagend;
    char		PP_ext_sbfmark_old = PP_ext_sbfmark;
    bool		ok = TRUE;

    PP_ext_tagend  = (char) FCV_SC_TAG;
    PP_ext_sbfmark = (char) FCV_SC_SUBFIELD;

    // Create buffer if not supplied
    if (csbuf == NULL) {
	ok = ok && (created_csbuf = CSCreate()) != NULL;
	csbuf = created_csbuf;
    }

    // Fill buffer with tag-indicator and external Pica+ tag
    ok = ok && CSSetString(csbuf, FCV_SC_TAG_STR);
    ok = ok && PPExtOutTagCS(csbuf, tag);

    // Remove tag-indicator from end of the buffer
    if (ok
	&& CSLength(csbuf) != 0
	&& *(CSContents(csbuf) + CSLength(csbuf) - 1) == FCV_SC_TAG) {
	(void) CSDeletePart(csbuf, CSLength(csbuf) - 1, 1);
    }

    // Restore tag- and sbf-indicator parameters
    PP_ext_tagend  = PP_ext_tagend_old;
    PP_ext_sbfmark = PP_ext_sbfmark_old;

    if (!ok) {
	CSDestroy(created_csbuf);
	return NULL;
    }
    
    return csbuf;
} // PPTagToNormal()

/*------------------------------------------------------------------------
 *  Function    : PPToExtern
 *  Purpose     : Convert a title top external pica+ format
 *  Method      : Output the tags one by one.
 *
 *  History:
 *  29-01-1996  : eck      : creation
 *  01-12-2000  : paul     : PPToExtern() also converts individual tags
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       :
 *  Checks for buffer overflow.  When returning, ttlsiz will contain the
 *  size of the title, even in case of overflow.
 *
 *  Returns     : Indication of success.  Failure means buffer overflow.
 *------------------------------------------------------------------------
 */
bool
PPToExtern(
        PPTtl_p         title,          /* Title to output */
        PP_Level_t      lev,            /* level to output or PP_ANYLEV */
        PP_Occurrence_t occ,            /* occurrence to output or PP_ANYOCC */
        char            *ttlbuf,        /* output buffer */
        size_t          bufsiz,         /* size of the output buffer */
        size_t          *ttlsiz)        /* pointer to current title size */
{
    PPTag_p     tag = NULL;

    PPerrno = PPE_NOERROR;

    *ttlsiz = 0;

    if (title->pp_elttyp == PPT_TITLE) {
        tag = title->PPTTL_firsttag;
    } else if (title->pp_elttyp == PPT_TAG) {
        tag = title;
    } else {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPToExtern", "Neither a title nor a tag!");
    }
    while (tag != NULL) {
        if (   ((lev == PP_ANYLEV) || (tag->PPTAG_lev == lev))
            && ((occ == PP_ANYOCC) || (tag->PPTAG_occ == occ)) ) {
            PPExtOutTag(tag, ttlbuf, bufsiz, ttlsiz);
        }
        tag = tag->pp_eltnext;
    }
    if (bufsiz > *ttlsiz) ttlbuf[*ttlsiz] = '\0';
    else PPerrno = PPE_OVERFLOW;
    return (PPerrno == PPE_NOERROR);
} /* PPToExtern() */

/*------------------------------------------------------------------------
 *  Function    : PPToNormal
 *  Purpose     : Convert a title to normalized pica format
 *  Method      : PPToExtern()
 *
 *  History:
 *  12-10-2000  : paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       :
 *  Normalized pica format is identical to external Pica+ format. The only
 *  differences are:
 *  1. Instead of a terminating tag-separator, there is a tag-indicator that
 *     introduces each tag.
 *  2. The tag-indicator has to be FCV_SC_TAG
 *  3. The subfield-indicator has to be FCV_SC_SUBFIELD
 *
 *  Returns     : Indication of success.  Failure means buffer overflow.
 *------------------------------------------------------------------------
 */
bool
PPToNormal(
        PPTtl_p         PPtitle,        /* Title to output */
        PP_Level_t      lev,            /* level to output or PP_ANYLEV */
        PP_Occurrence_t occ,            /* occurrence to output or PP_ANYOCC */
        char            *ttlbuf,        /* output buffer */
        size_t          bufsiz,         /* size of the output buffer */
        size_t          *ttlsiz)        /* pointer to current title size */
{
    char PP_ext_tagend_old  = PP_ext_tagend;
    char PP_ext_sbfmark_old = PP_ext_sbfmark;

    if (bufsiz == 0) {
        PPerrno = PPE_OVERFLOW;
        return (PPerrno == PPE_NOERROR);
    }
    PP_ext_tagend  = (char) FCV_SC_TAG;
    PP_ext_sbfmark = (char) FCV_SC_SUBFIELD;
    if (PPToExtern(PPtitle, lev, occ, ttlbuf + 1, bufsiz - 1, ttlsiz)
        && *ttlsiz != 0) {
/*
 *      Insert the tag-indicator of the first tag.
 */
        ttlbuf[0] = (char) FCV_SC_TAG;
        ++*ttlsiz;
/*
 *      Delete the (optional) tag-separator at the end of the last tag.
 */
        if (ttlbuf[*ttlsiz - 1] == (char) FCV_SC_TAG) {
            --*ttlsiz;
        }
        ttlbuf[*ttlsiz] = '\0';
    }
    PP_ext_tagend  = PP_ext_tagend_old;
    PP_ext_sbfmark = PP_ext_sbfmark_old;
    return (PPerrno == PPE_NOERROR);
} /* PPToNormal() */

/*------------------------------------------------------------------------
 *  Function	: PPToExternAppendCS
 *  Purpose	: Convert a title or a tag to external Pica+ format and append
 *                the result to a counted buffer.
 *  Method	: 
 *
 *  History:
 *  07-04-2003	: paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       :
 *  If the destination ttl is equal to NULL it will be created.
 *  If the source title is NULL nothing will be appended and TRUE is returned.
 *
 *  Returns	: The buffer with the title in external Pica+ format or NULL
 *------------------------------------------------------------------------
 */
bool
PPToExternAppendCS(
    CSBuf_p		ttl,		/* buffer to append to */
    PPTtl_p		title,		/* handle to title to convert */
    PP_Level_t		lev,		/* level to output or PP_ANYLEV */
    PP_Occurrence_t	occ)		/* occurrence to output or PP_ANYOCC */
{
    PPTag_p     tag = NULL;

    PPerrno = PPE_NOERROR;

    if (title == NULL) return TRUE;

    if (title->pp_elttyp == PPT_TITLE) {
        tag = title->PPTTL_firsttag;
    } else if (title->pp_elttyp == PPT_TAG) {
        tag = title;
    } else {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPToExtern", "Neither a title nor a tag!");
    }

    while (tag != NULL) {
        if (   ((lev == PP_ANYLEV) || (tag->PPTAG_lev == lev))
            && ((occ == PP_ANYOCC) || (tag->PPTAG_occ == occ)) ) {
	    if (!PPExtOutTagCS(ttl, tag)) break;
        }
        tag = tag->pp_eltnext;
    }

    return PPerrno == PPE_NOERROR;
} /* PPToExternAppendCS() */

/*------------------------------------------------------------------------
 *  Function	: PPToNormalAppendCS
 *  Purpose     : Convert a title to normalized pica format and append the
 *                result to an existing buffer.
 *  Method	: Create a buffer with the tag-separator; append the external
 *                Pica+ title and remove the last tag-separator.
 *
 *  History:
 *  07-04-2003	: paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       :
 *  If the destination buffer is equal to NULL it will be created.
 *
 *  Normalized pica format is identical to external Pica+ format. The only
 *  differences are:
 *  1. Instead of a terminating tag-separator, there is a tag-indicator that
 *     introduces each tag.
 *  2. The tag-indicator has to be FCV_SC_TAG
 *  3. The subfield-indicator has to be FCV_SC_SUBFIELD
 *
 *  Returns	: The buffer with the title in normalized pica format or NULL
 *------------------------------------------------------------------------
 */
bool
PPToNormalAppendCS(
    CSBuf_p		ttl,		/* buffer to append to */
    PPTtl_p		title,		/* handle to title to convert */
    PP_Level_t		lev,		/* level to output or PP_ANYLEV */
    PP_Occurrence_t	occ)		/* occurrence to output or PP_ANYOCC */
{
    char		PP_ext_tagend_old  = PP_ext_tagend;
    char		PP_ext_sbfmark_old = PP_ext_sbfmark;

    if (title == NULL) return TRUE;

/*
 *  Append the tag-indicator of the first tag.
 */
    if (!CSAppendC(ttl, FCV_SC_TAG)) {
	PPerrno = PPE_NOMEM;
	return FALSE;
    }

    PP_ext_tagend  = (char) FCV_SC_TAG;
    PP_ext_sbfmark = (char) FCV_SC_SUBFIELD;

    if (PPToExternAppendCS(ttl, title, lev, occ)
	&& CSLength(ttl) != 0
	&& *(CSContents(ttl) + CSLength(ttl) - 1) == FCV_SC_TAG) {
/*
 *	Delete the tag-separator at the end of the last tag.
 */
	(void) CSDeletePart(ttl, CSLength(ttl) - 1, 1);
    }

    PP_ext_tagend  = PP_ext_tagend_old;
    PP_ext_sbfmark = PP_ext_sbfmark_old;

    return TRUE;
} /* PPToNormalAppendCS() */

/*------------------------------------------------------------------------
 *  Function	: PPToExternCreateCS
 *  Purpose	: Convert a title or a tag to external Pica+ format.
 *  Method	: PPToExternAppendCS()
 *
 *  History:
 *  07-04-2003	: paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Returns	: The buffer with the title in external Pica+ format or NULL
 *------------------------------------------------------------------------
 */
CSBuf_p
PPToExternCreateCS(
    PPTtl_p		title,		/* handle to title to convert */
    PP_Level_t		lev,		/* level to output or PP_ANYLEV */
    PP_Occurrence_t	occ)		/* occurrence to output or PP_ANYOCC */
{
    CSBuf_p	ttl = NULL;

    ttl = CSCreate();
    if (ttl == NULL) {
	PPerrno = PPE_NOMEM;
	return NULL;
    }

    if (!PPToExternAppendCS(ttl, title, lev, occ)) {
	CSDestroy(ttl);
	return NULL;
    }

    return ttl;
} /* PPToExternCreateCS() */

/*------------------------------------------------------------------------
 *  Function	: PPToNormalCreateCS
 *  Purpose     : Convert a title to normalized pica format
 *  Method	: PPToNormalAppendCS()
 *
 *  History:
 *  07-04-2003	: paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       :
 *
 *  Returns	: The buffer with the title in normalized pica format or NULL
 *------------------------------------------------------------------------
 */
CSBuf_p
PPToNormalCreateCS(
    PPTtl_p		title,		/* handle to title to convert */
    PP_Level_t		lev,		/* level to output or PP_ANYLEV */
    PP_Occurrence_t	occ)		/* occurrence to output or PP_ANYOCC */
{
    CSBuf_p	ttl = NULL;

    ttl = CSCreate();
    if (ttl == NULL) {
	PPerrno = PPE_NOMEM;
	return NULL;
    }

    if (!PPToNormalAppendCS(ttl, title, lev, occ)) {
	CSDestroy(ttl);
	return NULL;
    }

    return ttl;
} /* PPToNormalCreateCS() */

/*------------------------------------------------------------------------
 *  Function    : PPFromNormal
 *  Purpose     : Convert a title in normalized pica format to PP
 *  Method      : PPFromExtern()
 *
 *  History:
 *  18-10-2000  : paul     : creation
 *
 *  Notes       :
 *  Normalize pica format is identical to external Pica+ format. The only
 *  differences are:
 *  1. Instead of a terminating tag-separator, there is a tag-indicator that
 *     introduces each tag.
 *  2. The tag-indicator has to be FCV_SC_TAG
 *  3. The subfield-indicator has to be FCV_SC_SUBFIELD
 *
 *  Notes       : Modifies the title-buffer (because PPFromExtern() does).
 *
 *  Returns     : The title-handle or NULL.
 *------------------------------------------------------------------------
 */
PPTtl_p
PPFromNormal(
        PPTtl_p title,                  /* Title-handle to add to or NULL */
        char    *ttlbuf)                /* Buffer with the title */
{
    char PP_ext_tagend_old  = PP_ext_tagend;
    char PP_ext_sbfmark_old = PP_ext_sbfmark;

    PP_ext_tagend  = (char) FCV_SC_TAG;
    PP_ext_sbfmark = (char) FCV_SC_SUBFIELD;

    if (ttlbuf[0] == FCV_SC_TAG) ttlbuf++;
    title = PPFromExtern(title, ttlbuf);

    PP_ext_tagend  = PP_ext_tagend_old;
    PP_ext_sbfmark = PP_ext_sbfmark_old;

    return title;
} /* PPFromNormal() */

/*------------------------------------------------------------------------
 *  Function	: PPFromX
 *  Purpose	: Wrapper for PPFromExtern() and PPFromNormal(). 
 *  Method	: 
 *
 *  History:
 *  11-02-2009	: martinz  : creation
 *
 *  Notes	:
 *
 *  Returns	: PPFromExtern() / PPFromNormal() result on success, 
 *		  NULL otherwise.
 *
 *------------------------------------------------------------------------
 */
static PPTtl_p
PPFromX(
        PPTtl_p (*foo)(PPTtl_p, char *),  /* PPFromExtern()/PPFromNormal() */
        PPTtl_p title,			  /* Title-handle to add to or NULL */
        const char *ttlbuf)		  /* Buffer with the title */
{
    static char *buf = NULL;	/* output buffer */
    static size_t size = 0;	/* output buffer size */
    size_t len;			/* output buffer length */
    char *tmpbuf;		/* temporary buffer */

    if (ttlbuf == NULL) {
        return foo(title, NULL);
    }
    if ((len = strlen(ttlbuf) + 1) > size) {
	if ((tmpbuf = realloc(buf, len)) == NULL) {
	    PPerrno = PPE_NOMEM;
	    ERWrite(PPerrno, __func__, "Failed realloc(size=%lu)",
		    (unsigned long)len);
	    return NULL;
	}
	buf = tmpbuf;
	size = len;
    }
    memcpy(buf, ttlbuf, len);
    return foo(title, buf);
} /* PPFromX() */

/*------------------------------------------------------------------------
 *  Function	: PPFromExtern2
 *  Purpose	: PPFromExtern() without overwriting the input buffer.
 *  Method	: 
 *
 *  History:
 *  11-02-2009	: martinz  : creation
 *
 *  Notes	:
 *
 *  Returns	: PPFromExtern() result on success, NULL otherwise.
 *
 *------------------------------------------------------------------------
 */
PPTtl_p
PPFromExtern2(
        PPTtl_p title,		/* Title-handle to add to or NULL */
        const char *ttlbuf)	/* Buffer with the title */
{
    return PPFromX(PPFromExtern, title, ttlbuf);
} /* PPFromExtern2() */

/*------------------------------------------------------------------------
 *  Function	: PPFromNormal2
 *  Purpose	: PPFromNormal() without overwriting the input buffer.
 *  Method	: 
 *
 *  History:
 *  11-02-2009	: martinz  : creation
 *
 *  Notes	:
 *
 *  Returns	: PPFromNormal() result on success, NULL otherwise.
 *
 *------------------------------------------------------------------------
 */
PPTtl_p
PPFromNormal2(
        PPTtl_p title,		/* Title-handle to add to or NULL */
        const char *ttlbuf)	/* Buffer with the title */
{
    return PPFromX(PPFromNormal, title, ttlbuf);
} /* PPFromNormal2() */
