/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ppintern.c
 *  Description	: Conversion to and from internal pica+
 *  Author(s)	: Hans van Eck
 *  Creation	: 22-01-1996
 *
 *  History:
 *  22-01-1996	: eck      : creation
 *  22-02-1996	: eck      : Added arbitrary endian-ness and level suppport.
 *			   : Format now corresponds to database format,
 *  08-07-1997	: eck      : Improved detection of errors in internal format
 *  31-07-1997	: eck      : Be more careful with PPerrno.
 *  04-09-1998	: BdB	   : PPFromIntern takes const byte * instead of byte *
 *  23-03-2001  : evd      : adjusted for tags bigger than 32k
 *  27-04-2001  : evd      : added continuation tag (typ = 255)
 *                         : Intern structure changed. Tag bigger then
 *                           64k always cutted at 64k. No matter in the
 *                           middle of a subfield.
 *  21-11-2001  : evd      : int defs in AddTagFromIntern and TagToIntern
 *                           changed to int32 for TANDEM use.
 *  23-11-2001	: paul     : AddTagFromIntern(): subfield lengths may never be
 *                           less than SBFHEADER_SIZ;
 *                           subfields may never exceed tag boundaries
 *  23-01-2002  : HvdW     : PPToString added as simple interface for
 *                           diagnostic uses.
 *  25-03-2002	: paul     : Added possibility of converting a single tag in
 *                           PPToIntern()
 *  01-10-2002	: paul     : Added PPToInternCreate() and PPToInternAppend()
 *  07-01-2003	: paul     : Fixed Array bounds write in PPToInternAppend()
 *  07_01_2003  : paul     : Check that there is room for pad byte
 *  07-04-2003  : paul     : Added PPToInternCreateCS() + PPToInternAppendCS()
 *                         : Fixed bug for continuation tags in TagToIntern()
 *                         : Replaced "USHRT_MAX" by "UINT16MAX"
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  22-11-2004  : michiel  : bug fix in multi-byte occurrence code
 *  24-11-2004  : michiel  : yet another bug fix to multi-byte occurrence
 *                           hack to work around signed-ness of a tag-type
 *  05-03-2009	: paul     : Allow tags of level 3 (social)
 *
 *  Notes	:
 *  The internal pica+ format is not printable.  The structure is as
 *  follows:
 *
 *  title	= tag [ tag ] * ;
 *  tag		= <tag-type><tag-size><lev+ind><occ> [ subfield ]* [ PAD ] ;
 *  tag-type	= 2-bytes integer
 *  tag-size	= 2-bytes integer with size of tag (including tagheader-size).
 *  lev+ind	= (level << 6) | (ind - '@')
 *  level	= 0..3
 *  ind		= 1 character from [@A-Z]
 *  occ		= encoded 1, 2 or 3-bytes integer (0 .. 2^16) containing occurrence
 *  subfield	= <sbf-type><sbf-size><contents>
 *  sbf-type	= either 1 alphanumeric character or '@'
 *  sbf-size	= 1 byte unsigned int subfield-size (including sbfheader size)
 *  contents	= string of characters (length is sbf-size - 2)
 *  PAD		= '\0'
 *
 *  The 2-byte integers can have arbitrary endianness.  This is necessary,
 *  since titles are stored big-endian on Tandem and little-endian in
 *  Sybase.  The Sybase endianness for titles is the VAX endiannes.
 *
 *  To modify the behaviour of the conversion routines, the variable
 *  PP_int_ttlendian can be changed.
 *  This variable depends on where the titles come from, not where the
 *  program runs.  Thus, on a SPARC (which is a big-endian machine), the
 *  titles are assumed little-endian (they will usually come from Sybase).
 *
 *  Occurrences in the range 0127 are coded as a simple unsigned byte. For
 *  values higher than 127 an extra encoding schema will be introduced: the
 *  high order bit of each occurrence byte will be interpreted as signal that
 *  this field has an adjacent byte; this principle is repeated for successive
 *  bytes although in common practice this is not expected to be used (with
 *  two bytes 2 x 2^7 can be encoded).
 *
 *  The PAD-byte is used when the size of the tag is odd.
 *========================================================================
 */

#pragma ident "$Id: ppintern.c 350772 2017-12-01 11:16:27Z heesakkm $"

#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <pica.h>
#include <cs.h>
#include <pp.h>
#include <er.h>
#include <stdlib.h>
#include "pplocal.h"

/* A Tag-header is at least 6 bytes long and consists of:
 * type (2 bytes), length (2 bytes), indicator (1 byte), occurrence (1+ bytes)
 * The occurrence must fit inside a 2-byte PP_Occurrence_t, therefore the coded
 * occurrence can be at most 3 bytes long.
 */
#define TAGHEADER_SIZ_WITHOUT_OCC 5
#define INTERNAL_OCC_SZ_MIN       1
#define INTERNAL_OCC_SZ_MAX       3
#define TAGHEADER_SIZ_MIN         (TAGHEADER_SIZ_WITHOUT_OCC + \
                                   INTERNAL_OCC_SZ_MIN)
#define TAGHEADER_SIZ_MAX         (TAGHEADER_SIZ_WITHOUT_OCC + \
                                   INTERNAL_OCC_SZ_MAX)

/* A subfield-header is 2 bytes long and consists of:
 * type (1 byte), length (1 byte)
 */
#define SBFHEADER_SIZ	2

/* The continuation subfield. */
#define SBF_CONTINUE	((char) 25)

/* The continuation tag. */
#define TAG_CONTINUE	255

#define ODD(val)	((val % 2) == 1)
#define PAD_VALUE	'\0'

/* capacity of print buffer */
#define INITIAL_PRINT_CAPACITY 100

#if defined(__TANDEM)
PPEndianness_t PP_int_ttlendian = PPEND_BIG;	/* Tandem titles */
#else
PPEndianness_t PP_int_ttlendian = PPEND_LITTLE;	/* Sybase titles */
#endif

/*------------------------------------------------------------------------
 *  Function	: Occurrence2CodedOccurrence
 *  Purpose	: given an occurrence (a 16 bit number), store it
 *                into the "internal format".
 *  Method	: bit manipulation hidden behind simple arithmatic
 *
 *  History:
 *  04-11-2004	: michiel  : creation
 *
 *  Returns     : The number of bytes written, or -1 on error
 *------------------------------------------------------------------------
 */
static int                   
Occurrence2CodedOccurrence(
	PP_Occurrence_t occ, /* the number to code into the internal format */
	byte *buf,           /* where to write the coded occurrence into    */
	size_t sz)           /* the amount of buffer space available in buf */
{
    PP_Occurrence_t k;
    int             i, n, m;

    /* occ = 0 is really an exception, trust me ;-) */
    if (occ == 0) {
	if (sz == 0) {
	    /* no space to store that single byte */
	    return -1;
	}

	buf[0] = (byte) 0;
	return 1;
    }

    /* first, count the number of bytes we actually need */
    k = occ;
    n = 0;
    while (k > 0) {
	k /= 128;
	n++;
    }

    if (n > sz) {
	/* too bad */
	PPerrno = PPE_OVERFLOW;
	(void) ERWrite(PPerrno, __func__, "Not sufficient space in buffer");
	return -1;
    }

    /* 
     * now for real, failure is not an option
     * (we either do it correctly, or do not change "buf" at all)
     */
    k = occ;
    m = 0;
    while (k > 0) {
	buf[n - m - 1] = k % 128;
	k /= 128;
	m++;
    }

    for (i=0 ; i<m-1 ; i++) {
	/* not the final byte --> set continuation bit */
	buf[i] |= 0x80;
    }

    /* these are sanity checks only. If we made a silly mistake somewhere,
     * better to catch the error here
     */
    if (m < INTERNAL_OCC_SZ_MIN || m > INTERNAL_OCC_SZ_MAX) {
	PPerrno = PPE_FORMAT;
	ERWrite(PPerrno, __func__, "Unexpected occurrence size (size=%d)", m);
	return -1;
    }

    return m;
} /* Occurrence2CodedOccurrence() */

/*------------------------------------------------------------------------
 *  Function	: CodedOccurrence2Occurrence
 *  Purpose	: given an coded occurrence (in "internal format"),
 *                retrieve it and pass it back to the caller as a normal
 *                16 bits number.
 *  Method	: bit manipulation hidden behind simple arithmatic
 *
 *  History:
 *  04-11-2004	: michiel  : creation
 *
 *------------------------------------------------------------------------
 */
static int                                       /* the number of bytes read, or -1 on error  */
CodedOccurrence2Occurrence(const byte *buf,      /* where to get the coded occurrence from    */
			 size_t sz,              /* an upper bound on the buffer space in buf */
			 PP_Occurrence_t *occp)  /* out parameter                             */
{
    const byte *p, *q;
    int k;
    int occ_sz;
    PP_Occurrence_t r;

    p = buf;
    while (p - buf < sz && ((*p & 0x80) != 0)) p++;

    if (p - buf == sz) {
	/* i'm so sorry */
	PPerrno = PPE_FORMAT;
	(void) ERWrite(PPerrno, __func__, "End byte not found in occurrence");
	return -1;
    }

    /* now that the end is known, the number can be reconstructed */
    q = p;
    k = 1;
    r = 0;
    while (q >= buf) {
	r += (*q & 0x7f) * k;
	q--;
	k *= 128;
    }

    *occp = r;
    occ_sz = p - buf + 1;

    /* these are sanity checks only. If we made a silly mistake somewhere,
     * better to catch the error here
     */
    if (occ_sz < INTERNAL_OCC_SZ_MIN || occ_sz > INTERNAL_OCC_SZ_MAX) {
	PPerrno = PPE_FORMAT;
	ERWrite(PPerrno, __func__, "Unexpected occurrence size (size=%d)", occ_sz);
	return -1;
    }

    return occ_sz;
} /* CodedOccurrence2Occurrence() */

/*------------------------------------------------------------------------
 *  Function	: Put2BytesInt
 *  Purpose	: Store 2 bytes arbitrary-endian
 *  Method	: obvious
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  22-02-1996	: eck      : Made routine arbitrary-endian
 *
 *  Notes	: Assumes 2's complement.
 *		: CHAR_BIT must be 8 for the 'uint16' to make sense.
 *------------------------------------------------------------------------
 */
static void
Put2BytesInt(
	byte		*bytep,		/* ptr to where length should be set */
	const uint16	length)		/* the length to set */
{
    if (PP_int_ttlendian == PPEND_BIG) {
	/* Store the bytes big-endian.  Assume 2's complement. */
	bytep[0] = (length >> CHAR_BIT) & ((1 << CHAR_BIT) - 1);
	bytep[1] = length & ((1 << CHAR_BIT) - 1);
    } else {
	/* Store the bytes little-endian.  Assume 2's complement. */
	bytep[0] = length & ((1 << CHAR_BIT) - 1);
	bytep[1] = (length >> CHAR_BIT) & ((1 << CHAR_BIT) - 1);
    }
} /* Put2BytesInt() */

/*------------------------------------------------------------------------
 *  Function	: Get2BytesInt
 *  Purpose	: Get an arbitrary-endian 2byte value
 *  Method	: obvious
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  22-02-1996	: eck      : Made routine arbitrary-endian
 *
 *  Returns	: The arbitrary-endian value pointed to.
 *------------------------------------------------------------------------
 */
static uint16
Get2BytesInt(
	const byte	*bytep)		/* ptr to place of int16 */
{
    if (PP_int_ttlendian == PPEND_BIG) {
	return (bytep[0] << CHAR_BIT) + bytep[1];
    }
    return bytep[0] + (bytep[1] << CHAR_BIT);
} /* Get2BytesInt() */

/*------------------------------------------------------------------------
 *  Function	: GetTagHeaderFromIntern
 *  Purpose	: Retrieve a tag header from internal format.
 *  Method	: Digest the input bit by bit.
 *
 *  History:
 *  23-11-2001	: paul     : extracted from AddTagFromIntern()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes	: 
 *
 *  Returns	: whether a correct tag header was present
 *------------------------------------------------------------------------
 */
static bool
GetTagHeaderFromIntern(
    PP_TagType_t    *tag_typ,	 /* type of the tag header */
    uint16	    *tag_siz,	 /* length from the tag header */
    PP_Level_t	    *tag_lev,	 /* level of the tag header */
    char	    *tag_ind,	 /* indicator of the tag header */
    PP_Occurrence_t *tag_occ,	 /* occurrence of the tag header */
    int  	    *tag_occ_sz, /* # of occurrence bytes */
    const byte	    *ttlbuf,	 /* buffer with title */
    int32	    *tag_offs,	 /* offset of current tag */
    size_t	    ttlsiz)	 /* size ot the title */
{
    int occ_sz;

    if (*tag_offs >= ttlsiz) return FALSE;

    if (*tag_offs + TAGHEADER_SIZ_MIN > ttlsiz) {
	PPerrno = PPE_FORMAT;
	(void) ERWrite(PPerrno, "PPFromIntern.GetTagHeaderFromIntern",
		       "Incomplete tag header at end of title");
	return FALSE;
    }

    *tag_typ = Get2BytesInt(ttlbuf + *tag_offs);
    *tag_siz = Get2BytesInt(ttlbuf + *tag_offs + 2);
    *tag_lev = ((*(char *)(ttlbuf + *tag_offs + 4)) >> 6) & 0x3;
    *tag_ind = ((*(char *)(ttlbuf + *tag_offs + 4)) & 0x3f) + '@' ;
    occ_sz = CodedOccurrence2Occurrence(ttlbuf + *tag_offs + 5,
				      ttlsiz - *tag_offs - 5,
				      tag_occ);
    if (occ_sz < 0) {
        return FALSE;
    }
    *tag_occ_sz = occ_sz;

    if (*tag_siz < TAGHEADER_SIZ_MIN) {
	PPerrno = PPE_FORMAT;
	(void) ERWrite(PPerrno, "PPFromIntern.GetTagHeaderFromIntern",
		       "Illegal tag length at end of title");
	return FALSE;
    }
    if (*tag_offs + *tag_siz > ttlsiz) {
	(void) ERWrite(PPerrno, "PPFromIntern.GetTagHeaderFromIntern",
		       "Truncated tag at end of title");
	*tag_siz = ttlsiz - *tag_offs;
    }


    return TRUE;
} /* GetTagHeaderFromIntern() */

/*------------------------------------------------------------------------
 *  Function	: AddTagFromIntern
 *  Purpose	: Retrieve a tag (and subfields) from internal format.
 *  Method	: Digest the input bit by bit.
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  ??_07_2001  : evd      : Structure changed at marks bigger then 64k
 *                           Tag is always cutted at 64k (uint16) even
 *                           in the middle of a subfield. The tag has to
 *                           be build first before reading the subfields.
 *  24-08-2001	: eck      : Determine next tag only when not at end of buffer
 *  23-11-2001	: paul     : extracted GetTagHeaderFromIntern()
 *                           sbf_siz may never be less than SBFHEADER_SIZ
 *                           nor greater than tag_len
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  24-11-2004  : michiel  : hack to work around signed-ness of a tag-type
 *
 *  Notes	: This is rather low-level.
 *
 *  Returns	: tag-handle or NULL.
 *------------------------------------------------------------------------
 */
static PPTag_p
AddTagFromIntern(
	PPElement_p	elt,		/* title/tag to append to */
	const byte	*ttlbuf,	/* buffer with title */
	int32		*tag_offs,	/* offset of current tag */
	size_t		ttlsiz)		/* size ot the title */
{
    PP_TagType_t    tag_typ, tag2_typ;
    PP_Level_t	    tag_lev, tag2_lev;
    uint16	    tag_siz, tag2_siz;
    char	    tag_ind, tag2_ind;
    PP_Occurrence_t tag_occ, tag2_occ;
    int             tag_occ_sz, tag2_occ_sz;
    PPTag_p	    tag;
    byte            *ttltag = NULL;
    int             tagheader_siz, tag2header_siz;

    if (!GetTagHeaderFromIntern(&tag_typ, &tag_siz, &tag_lev, &tag_ind,
				&tag_occ, &tag_occ_sz, ttlbuf, tag_offs, ttlsiz)) {
	return NULL;
    }

    /* Now allocate the tag-handle. */
    tag = PPTagAppend(elt, tag_typ, tag_lev, tag_ind, tag_occ);

    if (tag != NULL) {
	const byte	*tag_buf;	/* points to first subfield of tag */
	int32		tag_len;	/* length of subfields of tag */
	int32		sbf_offs;	/* for current subfield */
	PPSbf_p		sbf = NULL;

	tag_len = 0;
	ttltag = (byte *)malloc((sizeof *ttltag) * (tag_len + tag_siz));
	if (ttltag == NULL) {
	    PPerrno = PPE_NOMEM;
	    (void) ERWrite(PPerrno, "PPFromIntern.AddTagFromIntern",
				    "malloc() failed");
	    return NULL;
	}
	tagheader_siz = TAGHEADER_SIZ_WITHOUT_OCC + tag_occ_sz;
	(void) memcpy(ttltag + tag_len, ttlbuf + *tag_offs + tagheader_siz,
		      tag_siz - tagheader_siz); 
	tag_buf = ttltag;
	tag_len += tag_siz - tagheader_siz;
	*tag_offs += tag_siz + (ODD(tag_siz));

	while (GetTagHeaderFromIntern(&tag2_typ, &tag2_siz, &tag2_lev,
				      &tag2_ind, &tag2_occ, &tag2_occ_sz, ttlbuf, tag_offs,
				      ttlsiz)
	       && (((int)tag2_typ & 0xff) == TAG_CONTINUE)) {

	    ttltag = (byte *)realloc(ttltag, (sizeof *ttltag) * (tag_len +
								 tag2_siz));
	    if (ttltag == NULL) {
		PPerrno = PPE_NOMEM;
		(void) ERWrite(PPerrno, "PPFromIntern.AddTagFromIntern",
					"realloc() failed");
		return NULL;
	    }

	    tag2header_siz = TAGHEADER_SIZ_WITHOUT_OCC + tag2_occ_sz;
	    (void)memcpy(ttltag + tag_len, ttlbuf + *tag_offs +
	                 tag2header_siz, tag2_siz - tag2header_siz); 
	    tag_len += tag2_siz - tag2header_siz;
	    *tag_offs += tag2_siz + (ODD(tag2_siz));
	}
	sbf_offs = 0;
	tag_buf = ttltag;

	while ((sbf_offs < tag_len) && (PPerrno == PPE_NOERROR)) {
	    char	sbf_typ;	/* type of subfield */
	    uint8	sbf_siz;	/* size of subfield */

	    sbf_typ = *(tag_buf + sbf_offs);
	    sbf_siz = *(tag_buf + sbf_offs + 1);
	    if (sbf_siz < SBFHEADER_SIZ) {
		PPerrno = PPE_FORMAT;
		(void) ERWrite(PPerrno, "PPFromIntern.AddTagFromIntern",
			       "Subfield too small"
			       " (tag=%1d%02d%c/%02d)", tag_lev, tag_typ
			       , tag_ind, tag_occ);
		break;
	    }
	    if (sbf_offs + sbf_siz > tag_len) {
		(void) ERWrite(PPerrno, "PPFromIntern.AddTagFromIntern",
			       "Truncated subfield at end of tag");
		sbf_siz = tag_len - sbf_offs;
	    }

	    if (sbf_typ == SBF_CONTINUE) {
		/* Adjust the subfield just read. */
		if (sbf != NULL) {
		    (void) PPSbfAppendContents(sbf,
				(voidp) (tag_buf + sbf_offs + SBFHEADER_SIZ),
				sbf_siz - SBFHEADER_SIZ);
		} else {
		    PPerrno = PPE_FORMAT;
		    (void) ERWrite(PPerrno, "PPFromIntern.AddTagFromIntern",
			    "Unexpected continuation subfield"
			    " (tag=%1d%02d%c/%02d)", tag_lev, tag_typ
						    , tag_ind, tag_occ);
		    break;
		}
	    } else {
		sbf = PPSbfAppend(tag, sbf_typ,
				(voidp) (tag_buf + sbf_offs + SBFHEADER_SIZ),
				sbf_siz - SBFHEADER_SIZ);
	    }
	    sbf_offs += sbf_siz;
	}
	if (PPerrno != PPE_NOERROR) {
	    PPErrno_t	tmp = PPerrno;

	    PPTagDestroy(tag);
	    tag = NULL;
	    PPerrno = tmp;
	}
	free(ttltag);
    }

    return tag;
} /* AddTagFromIntern() */

/*------------------------------------------------------------------------
 *  Function	: PPFromIntern
 *  Purpose	: Convert PicaPlus title from internal- to abstract- format
 *  Method	: Tag-by-tag conversion.
 *
 *  History:
 *  22-01-1996	: eck      : creation
 *  04-09-1998	: BdB	   : PPFromIntern takes const byte * instead of byte *
 *  ??-07-2001  : evd      : Adding padbyte moved to AddTagFromIntern 
 *
 *  Returns	: Handle to abstract title.
 *------------------------------------------------------------------------
 */
PPTtl_p
PPFromIntern(
	PPTtl_p	title,		/* Title-handle to add to or NULL */
	const byte *ttlbuf,	/* Title with internal PicaPlus buffer */
	size_t	ttlsiz)		/* size of the title */
{
    int32		tag_offs = 0;
    PPElement_p		elt;

    PPerrno = PPE_NOERROR;

    if (title == NULL) title = PPTtlCreate();
    if (title == NULL) return NULL;

    elt = title;

    while ((elt = AddTagFromIntern(elt, ttlbuf, &tag_offs, ttlsiz)) != NULL) {
	/* Skip PAD-byte  Moved to AddTagFromIntern */
      elt = elt;
/*	if (ODD(tag_offs)) {
	    tag_offs++;
	}
	*/
    }
    if (PPerrno != PPE_NOERROR) {
	PPErrno_t	tmp = PPerrno;

	PPTtlDestroy(title);
	title = NULL;
	PPerrno = tmp;
    }
    return title;
} /* PPFromIntern() */

/*------------------------------------------------------------------------
 *  Function	: SbfToIntern
 *  Purpose	: Convert a single subfield into internal format
 *  Method	: Output the subfield in parts.
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  16-11-2004  : michiel  : change lenleft from 16 to 32 bits int
 *
 *  Notes	:
 *  Care has to be take with fields > 253 characters: they should be
 *  followed by one or more continuation-subfields.
 *  The total length of the output (including continuation subfields and
 *  overhead) is stored through sbfsiz).
 *
 *  Returns	: indication of success.
 *------------------------------------------------------------------------
 */
static bool
SbfToIntern(
	PPSbf_p		sbf,		/* Subfield to output */
	byte		*ttlbuf,	/* pointer into output buffer */
	size_t		bufsiz,		/* size of output buffer left */
	int32		*sbfsiz)	/* pointer for total subfield size */
{
    int32	lenleft;
    char	typ;
    const byte	*contents;

    *sbfsiz = 0;

    lenleft = sbf->PPSBF_length;
    typ = sbf->PPSBF_typ;
    contents = sbf->PPSBF_contents;

    do {
	uint8	outlen;

	/* Watch out: overhead is included in the length. */
	if (lenleft > (UCHAR_MAX - SBFHEADER_SIZ)) {
	    outlen = (uint8) (UCHAR_MAX - SBFHEADER_SIZ);
	} else {
	    outlen = (uint8) lenleft;
	}
	if (outlen + SBFHEADER_SIZ > bufsiz) {
	    PPerrno = PPE_OVERFLOW;
	    return FALSE;
	}
	*ttlbuf = typ;
	*(ttlbuf + 1) = outlen + SBFHEADER_SIZ;
	(void) memcpy(ttlbuf + SBFHEADER_SIZ, contents, outlen);

	/* Set typ to SBF_CONTINUE, since we are in a loop. */
	typ = SBF_CONTINUE;

	/* The input side should be adjusted with outlen. */
	lenleft -= outlen;
	contents += outlen;

	/* The output side should be adjusted with outlen + SBDHEADER_SIZ. */
	ttlbuf += outlen + SBFHEADER_SIZ;
	bufsiz -= (outlen + SBFHEADER_SIZ);
	*sbfsiz += outlen + SBFHEADER_SIZ;
    } while (lenleft > 0);

    return TRUE;
} /* SbfToIntern() */

/*------------------------------------------------------------------------
 *  Function	: SbfToInternAppend
 *  Purpose	: Convert a single subfield into internal format
 *  Method	: Output the subfield in parts.
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  07-04-2003  : paul     : Adapted from SbfToIntern() for counted strings
 *  16-11-2004  : michiel  : change lenfeft from 16 to 32 bits int
 *
 *  Notes	:
 *  Care has to be take with fields > 253 characters: they should be
 *  followed by one or more continuation-subfields.
 *  The total length of the output (including continuation subfields and
 *  overhead) is stored through sbfsiz).
 *
 *  Returns	: the length of the appended internal subfield or 0 (failure)
 *------------------------------------------------------------------------
 */
static size_t
SbfToInternAppend(
    CSBuf_p		ttl,		/* title to append to */
    PPSbf_p		sbf)		/* Subfield to append */
{
    int32	lenleft;
    char	typ;
    const char	*contents;
    size_t	oldttllen = CSLength(ttl);

    lenleft = sbf->PPSBF_length;
    typ = sbf->PPSBF_typ;
    contents = sbf->PPSBF_contents;

    do {
	uint8	outlen;

	/* Watch out: overhead is included in the length. */
	if (lenleft > (UCHAR_MAX - SBFHEADER_SIZ)) {
	    outlen = (uint8) (UCHAR_MAX - SBFHEADER_SIZ);
	} else {
	    outlen = (uint8) lenleft;
	}
	if (!CSAppendC(ttl, typ)
	    || !CSAppendC(ttl, (char)(outlen + SBFHEADER_SIZ))
	    || !CSAppendN(ttl, contents, outlen)) {
	    PPerrno = PPE_NOMEM;
	    return 0;
	}

	/* Set typ to SBF_CONTINUE, since we are in a loop. */
	typ = SBF_CONTINUE;

	/* Skip the part of the subfield contents that has been converted */
	lenleft -= outlen;
	contents += outlen;
    } while (lenleft > 0);

    return CSLength(ttl) - oldttllen;
} /* SbfToInternAppend() */

/*------------------------------------------------------------------------
 *  Function	: TagToIntern
 *  Purpose	: Convert a tag to internal format.
 *  Method	: 
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  ??_07_2001  : evd      : Structure changed at marks bigger then 64k
 *                           Tag is always cutted at 64k (uint16) even
 *                           in the middle of a subfield. 
 *  07_01_2003  : paul     : Check that there is room for PAD_VALUE
 *  07-04-2003  : paul     : Fixed bug for continuation tags in memmove
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  24-11-2004  : michiel  : hack to work around signed-ness of a tag-type
 *
 *  Notes	:
 *  Not all of the tag-header can be written at once: the length is only
 *  known after the subfields are stored.
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
static bool
TagToIntern(
	PPTag_p		tag,		/* tag to output */
	byte		*ttlbuf,	/* pointer into output buffer */
	size_t		bufsiz,		/* size of bufer left */
	size_t		*tagsiz)	/* pointer to store size of tag */
{
    PPSbf_p		sbf;
    int32		sbfsiz = 0;
    size_t		localtagsiz;
    int32		remainder_len, startmv;
    PP_TagType_t	tagtyp;
    int32       	odd_uint16_max = ODD(UINT16MAX);
    int         	occ_sz;

    if (bufsiz < TAGHEADER_SIZ_MIN) {
	PPerrno = PPE_OVERFLOW;
	return FALSE;
    }
    tagtyp = tag->PPTAG_typ;

    /* Put type, indicator and occurrence, but not with the size. */
    Put2BytesInt(ttlbuf, (uint16) tagtyp & 0xff);
    *(char *)(ttlbuf + 4) = (tag->PPTAG_ind - '@') | (tag->PPTAG_lev << 6);
    occ_sz = Occurrence2CodedOccurrence(tag->PPTAG_occ,
				      ttlbuf + 5,
				      bufsiz - 5);
    if (occ_sz < 0) return FALSE;

    localtagsiz = TAGHEADER_SIZ_WITHOUT_OCC + occ_sz;
    *tagsiz = TAGHEADER_SIZ_WITHOUT_OCC + occ_sz;

    sbf = tag->PPTAG_firstsbf;
    while (sbf != NULL) {
	if (! SbfToIntern(sbf, ttlbuf + *tagsiz, 
			  bufsiz - *tagsiz, &sbfsiz)) {
	    return FALSE;
	}
	*tagsiz += sbfsiz;
	localtagsiz += sbfsiz;
	remainder_len =  localtagsiz - UINT16MAX;
	while (remainder_len > 0){
	    tagtyp = (PP_TagType_t)TAG_CONTINUE;
	    startmv = *tagsiz - remainder_len;
	    localtagsiz = UINT16MAX;
	    Put2BytesInt(ttlbuf + startmv - UINT16MAX + 2, (uint16)localtagsiz);
	    *tagsiz += TAGHEADER_SIZ_WITHOUT_OCC + occ_sz + (odd_uint16_max); 
	    /* move remainder_len  7 positions in ttlbuf and add */
	    /* continuation tag to ttlbuf                        */
	    if (bufsiz < *tagsiz) {
		PPerrno = PPE_OVERFLOW;
		return FALSE;
	    }
	    /* for ( i = remainder_len ; i > 0 ; i-- ) {
		*(ttlbuf + startmv + i + TAGHEADER_SIZ + 1 -1) =
		*(ttlbuf + startmv + i - 1);
	    }
	    */
	    (void)memmove(
	          ttlbuf + startmv + TAGHEADER_SIZ_WITHOUT_OCC + occ_sz + odd_uint16_max,
		  ttlbuf + startmv, remainder_len);
            if (odd_uint16_max) {
		*(ttlbuf + startmv) = PAD_VALUE;
	    }
	    /* Put type, indicator and occurrence, but not with the size. */
	    Put2BytesInt(ttlbuf + startmv + 1, (uint16) tagtyp & 0xff);
	    *(char *)(ttlbuf + startmv + 1 + 4) = 
	    (tag->PPTAG_ind - '@') | (tag->PPTAG_lev << 6);
	    occ_sz = Occurrence2CodedOccurrence(tag->PPTAG_occ,
					      ttlbuf + startmv + 1 + 5,
					      bufsiz - startmv - 1 - 5);
	    if (occ_sz < 0) return FALSE;
	    localtagsiz  = remainder_len + TAGHEADER_SIZ_WITHOUT_OCC + occ_sz;
	    remainder_len = localtagsiz - UINT16MAX;
	}

	sbf = sbf->pp_eltnext;
    }
    /* Now we know the size.  put it into the buffer. */
    Put2BytesInt(ttlbuf + *tagsiz - localtagsiz + 2, (uint16)localtagsiz);
    if (ODD(localtagsiz)) {
	if (bufsiz < *tagsiz + 1) {
	    PPerrno = PPE_OVERFLOW;
	    return FALSE;
	}
	/* Add PAD-byte */
	*(ttlbuf + *tagsiz) = PAD_VALUE;
	*tagsiz += 1;
    }

    return TRUE;
} /* TagToIntern() */

/*------------------------------------------------------------------------
 *  Function	: TagHeaderToIntern
 *  Purpose	: Convert a tag header to internal format.
 *  Method	: 
 *
 *  History:
 *  07-04-2003  : paul     : Extracted from PPIntOutTagCS()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  22-11-2004  : michiel  : bug fix in multi-byte occurrence code
 *  24-11-2004  : michiel  : hack to work around signed-ness of a tag-type
 *
 *  Notes	: 
 *  Put type, indicator and occurrence, but keep the length field empty.
 *
 *------------------------------------------------------------------------
 */
static int                              /* size of header, or -1 on error */
TagHeaderToIntern(
    CSBuf_p		ttl,		/* title to append to */
    size_t		offset,		/* offset of header in ttl buffer */
    PP_TagType_t	tag_typ,	/* type of the tag header */
    PP_Level_t		tag_lev,	/* level of the tag header */
    char		tag_ind,	/* indicator of the tag header */
    PP_Occurrence_t	tag_occ,	/* occurrence of the tag header */
    size_t	  	physicaltagoffset, /* offset of the last header*/
    int			x_bytes)	/* an optional additional byte to
					 * allocate */
{
    static char	  header[1 + TAGHEADER_SIZ_MAX]; /* header + preceding pad byte */
    byte		*bytep;		/* ptr to where header should be put */
    int                 t;
    uint8		occ_tmp[INTERNAL_OCC_SZ_MAX];
    int			tagheader_siz;

    /* first convert the occurrence, then we know its size */
    t = Occurrence2CodedOccurrence(tag_occ, occ_tmp, INTERNAL_OCC_SZ_MAX);
    if (t < 0) {
	/* should not happen */
	return -1;
    }
    tagheader_siz = TAGHEADER_SIZ_WITHOUT_OCC + t;

    /* now that we know the size of the tagheader, we can safely and correctly
     * increase the size of "ttl". We increase the size of "ttl" by appending a
     * dummy array to it, or inserting one at the correct place */
    if (!CSInsertN(ttl, physicaltagoffset, header,
		   tagheader_siz + x_bytes)) {
	PPerrno = PPE_NOMEM;
	return FALSE;
    }

    bytep = (byte *)ttl->cs_bufp + offset;

    /* Put tag type as an int */
    Put2BytesInt(bytep, (uint16) tag_typ & 0xff);

    /* Put indicator and level in a single byte */
    *(char *)(bytep + 4) = (tag_ind - '@') | (tag_lev << 6);

    /* Put occurrence in the last byte(s) */
    memcpy(bytep + 5, occ_tmp, t);

    return tagheader_siz;

} /* TagHeaderToIntern() */

/*------------------------------------------------------------------------
 *  Function	: PPIntOutTagCS
 *  Purpose	: Convert a tag to internal format.
 *  Method	: 
 *
 *  History:
 *  29-01-1996	: eck      : creation
 *  ??_07_2001  : evd      : Structure changed at marks bigger then 64k
 *                           Tag is always cutted at 64k (uint16) even
 *                           in the middle of a subfield. 
 *  07_01_2003  : paul     : Check that there is room for PAD_VALUE
 *  07-04-2003  : paul     : Adapted from TagToIntern() for counted strings
 *                         : Fixed bug for continuation tags
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  22-11-2004  : michiel  : bug fix in multi-byte occurrence code
 *  24-11-2004  : michiel  : yet another bug fix to multi-byte occurrence
 *  03-09-2012	: paul     : made extern and renamed from TagHeaderToIntern()
 *
 *  Notes	:
 *  Not all of the tag-header can be written at once: the length is only
 *  known after the subfields are stored.
 *
 *  A tag consists of a tag header followed by subfields. The tag header
 *  contains the length of the tag in uint16 value. But if a tag is bigger than
 *  UINT16MAX the length doesn't fit. To solve this problem, we distinguish the
 *  logical tag - representing the whole input tag - from the physical tag -
 *  representing a tag with a single tag header. A logical tag consists of one
 *  or more physical tags. All physical tags have length UINT16MAX except the
 *  last one. All physical tags have a special type identifying them as
 *  continuation tag except the first one.
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
PPIntOutTagCS(
    CSBuf_p		ttl,		/* title to append to */
    PPTag_p		tag)		/* tag to output */
{
    PPSbf_p	  sbf;
    size_t	  sbfsiz = 0;
    size_t	  physicaltagsiz;
    size_t	  logicaltagsiz;
    size_t	  logicaltagoffset = CSLength(ttl);/* offset of new tag in ttl */
    size_t	  physicaltagoffset;		 /* offset of the last header*/
    int32	  remainder_len;
    PP_TagType_t  tag_typ;
    int32	  odd_uint16_max = ODD(UINT16MAX);
    int           tagheader_siz;

    /* Put type, indicator and occurrence, but not with the size. */
    tag_typ = tag->PPTAG_typ;
    tagheader_siz = TagHeaderToIntern(ttl, logicaltagoffset, tag_typ,
				      tag->PPTAG_lev, tag->PPTAG_ind, 
				      tag->PPTAG_occ,
				      CSLength(ttl),
				      0);
    if (tagheader_siz < 0) return FALSE;

    physicaltagsiz = tagheader_siz;
    logicaltagsiz = tagheader_siz;

    sbf = tag->PPTAG_firstsbf;
    while (sbf != NULL) {
	sbfsiz = SbfToInternAppend(ttl, sbf);
	if (sbfsiz == 0) return FALSE;

	logicaltagsiz  += sbfsiz;
	physicaltagsiz += sbfsiz;
	remainder_len = physicaltagsiz - UINT16MAX;
	while (remainder_len > 0) {
	    tag_typ = (PP_TagType_t)TAG_CONTINUE;
	    physicaltagoffset = (logicaltagsiz + logicaltagoffset
				 - remainder_len - UINT16MAX);
	    physicaltagsiz = UINT16MAX;

	    /* Now we know the length: put it into the old header */
	    Put2BytesInt((byte *)ttl->cs_bufp + physicaltagoffset + 2,
			 (uint16)physicaltagsiz);

	    /* Jump to the next physical tag */
	    physicaltagoffset += UINT16MAX;

	    tagheader_siz = TagHeaderToIntern(ttl, physicaltagoffset + odd_uint16_max, tag_typ,
					      tag->PPTAG_lev, tag->PPTAG_ind, tag->PPTAG_occ,
					      physicaltagoffset,
					      odd_uint16_max);
	    if (tagheader_siz < 0) return FALSE;

            if (odd_uint16_max) {
		*((byte *)ttl->cs_bufp + physicaltagoffset) = PAD_VALUE;
	    }

	    logicaltagsiz += tagheader_siz + odd_uint16_max; 
	    physicaltagsiz = tagheader_siz + remainder_len;
	    remainder_len = physicaltagsiz - UINT16MAX;
	}

	sbf = sbf->pp_eltnext;
    }

    /* Now we know the length: put it into the buffer. */
    physicaltagoffset = logicaltagoffset + logicaltagsiz - physicaltagsiz;
    Put2BytesInt((byte *)ttl->cs_bufp + physicaltagoffset + 2,
		 (uint16)physicaltagsiz);
    if (ODD(physicaltagsiz)) {
	/* Add PAD-byte */
	if (!CSAppendC(ttl, (char) PAD_VALUE)) {
	    PPerrno = PPE_NOMEM;
	    return FALSE;
	}
	logicaltagsiz += 1;
    }

    return PPerrno == PPE_NOERROR;
} /* PPIntOutTagCS() */

/*------------------------------------------------------------------------
 *  Function	: PPToIntern
 *  Purpose	: Convert a title or a tag to internal format.
 *  Method	: Repeated calls of TagToIntern().
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  25-03-2002	: paul     : Added possibility of converting a single tag
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Returns	: indication of success.
 *------------------------------------------------------------------------
 */
bool
PPToIntern(
	PPTtl_p		title,		/* handle to title to convert */
	PP_Level_t	lev,		/* level to output or PP_ANYLEV */
	PP_Occurrence_t	occ,		/* occurrence to output or PP_ANYOCC */
	byte		*ttlbuf,	/* pointer into output buffer */
	size_t		bufsiz,		/* size of bufer left */
	size_t		*ttlsiz)	/* pointer to store size of title */
{
    PPTag_p	tag = NULL;

    PPerrno = PPE_NOERROR;

    *ttlsiz = 0;

    if (title->pp_elttyp == PPT_TITLE) {
	tag = title->PPTTL_firsttag;
    } else if (title->pp_elttyp == PPT_TAG) {
	tag = title;
    } else {
	 PPerrno = PPE_TYPE;
	 ERAbort(PPerrno, "PPToIntern", "Neither a title nor a tag!");
    }

    while (tag != NULL) {
	size_t	tagsiz;

	/* Make the tag visible again.  It may need to be stored. */
	if (tag->PPTAG_typ < 0) {
	    tag->PPTAG_typ = -tag->PPTAG_typ;
	}
	if (((lev == PP_ANYLEV) || (tag->PPTAG_lev == lev))
	    && (((occ == PP_ANYOCC) || (tag->PPTAG_occ == occ)))) {
	    if (! TagToIntern(tag,
			      ttlbuf + *ttlsiz, bufsiz - *ttlsiz, &tagsiz)) {
		break;
	    }
	    *ttlsiz += tagsiz;
	}
	if (title->pp_elttyp == PPT_TAG) break;
	tag = tag->pp_eltnext;
    }

    return (PPerrno == PPE_NOERROR);
} /* PPToIntern() */

/*------------------------------------------------------------------------
 *  Function	: PPToInternAppendCS
 *  Purpose	: Convert a title or a tag to internal format and append the
 *                result to an existing counted string.
 *  Method	: 
 *
 *  History:
 *  07-04-2003	: paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes	:
 *  If the destination ttl is equal to NULL it will be created.
 *  If the source title is NULL nothing will be appended and TRUE is returned.
 *
 *  Returns	: indication of success
 *------------------------------------------------------------------------
 */
bool
PPToInternAppendCS(
    CSBuf_p		ttl,		/* buffer to append to */
    PPTtl_p		title,		/* handle to title to convert */
    PP_Level_t		lev,		/* level to output or PP_ANYLEV */
    PP_Occurrence_t	occ)		/* occurrence to output or PP_ANYOCC */
{
    PPTag_p	tag = NULL;

    PPerrno = PPE_NOERROR;

    if (title == NULL) return TRUE;

    if (title->pp_elttyp == PPT_TITLE) {
	tag = title->PPTTL_firsttag;
    } else if (title->pp_elttyp == PPT_TAG) {
	tag = title;
    } else {
	 PPerrno = PPE_TYPE;
	 ERAbort(PPerrno, "PPToIntern", "Neither a title nor a tag!");
    }

    while (tag != NULL) {
	/* Make the tag visible again.  It may need to be stored. */
	if (tag->PPTAG_typ < 0) {
	    tag->PPTAG_typ = -tag->PPTAG_typ;
	}
	if (((lev == PP_ANYLEV) || (tag->PPTAG_lev == lev))
	    && (((occ == PP_ANYOCC) || (tag->PPTAG_occ == occ)))) {
	    if (!PPIntOutTagCS(ttl, tag)) break;
	}
	if (title->pp_elttyp == PPT_TAG) break;
	tag = tag->pp_eltnext;
    }

    return PPerrno == PPE_NOERROR;
} /* PPToInternAppendCS() */

/*------------------------------------------------------------------------
 *  Function	: PPToInternCreateCS
 *  Purpose	: Convert a title or a tag to internal format and put it in a
 *                counted string
 *  Method	: 
 *
 *  History:
 *  07-04-2003	: paul     : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes	: If the PP-title is NULL, an empty buffer will be created.
 *
 *  Returns	: The counted string with the title in internal format or NULL
 *------------------------------------------------------------------------
 */
CSBuf_p
PPToInternCreateCS(
    PPTtl_p		title,		/* handle to title to convert */
    PP_Level_t		lev,		/* level to output or PP_ANYLEV */
    PP_Occurrence_t	occ)		/* occurrence to output or PP_ANYOCC */
{
    CSBuf_p	ttl = NULL;

    ttl = CSCreate();
    if (ttl == NULL) {
	PPerrno = PPE_NOMEM;
	return NULL;
    }

    if (!PPToInternAppendCS(ttl, title, lev, occ)) {
	CSDestroy(ttl);
	return NULL;
    }

    return ttl;
} /* PPToInternCreateCS() */

/*------------------------------------------------------------------------
 *  Function	: ensureCapacity
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  23-01-2002	: henk     : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static bool
ensureCapacity(char ** out, size_t minimalCapacity)
{
    static size_t Capacity = 0;
    if (Capacity == 0){
	Capacity = INITIAL_PRINT_CAPACITY;
	if (minimalCapacity > Capacity){
	    Capacity = minimalCapacity;
	}
	*out = (char *) malloc(Capacity);
	if (*out == NULL){
	    PPerrno = PPE_NOMEM;
	    ERWrite(PPerrno, __func__,
		    "malloc(%lu) failed", (unsigned long) Capacity);
	    return FALSE;
	}
    }
    while (Capacity < minimalCapacity){
	void * p;
	Capacity = 2*Capacity + 1;
	p = (char *) realloc((void *)(*out), Capacity);
	if (p == NULL){
	    PPerrno = PPE_NOMEM;
	    ERWrite(PPerrno, __func__,
		    "realloc(%lu) failed", (unsigned long) Capacity);
	    return FALSE;
	}else{
	    *out = (char *) p;
	}
    }
    return TRUE;
}

/*------------------------------------------------------------------------
 *  Function	: subfieldToString
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  22-01-2002	: henk     : creation
 *  11-03-2004	: heesakkm : Cast outlen to int in for-loop (lint ...).
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static size_t 
subfieldToString(char ** out, size_t outlen, PPSbf_p sbf)
{
    int j;
    int p = 0;
    if (ensureCapacity(out, outlen + strlen(" $a(len   )"))){
	outlen += sprintf((*out) + outlen, " $%c(%d)", sbf->PPSBF_typ, 
		sbf->PPSBF_length);
    }
    p = sbf->PPSBF_length;
    if (ensureCapacity(out, outlen + p)){
	(void) strncpy( (*out) + outlen, sbf->PPSBF_contents, p);
	for (j = (int) outlen; j < (int) (outlen + p); j++){
	    if ((*out)[j] < ' '){
		(*out)[j] = '|';
	    }
	}
	outlen += p;
    }
    return outlen;
} /* subfieldToString() */
/*------------------------------------------------------------------------
 *  Function	: tagToString
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  22-01-2002	: henk     : creation
 *  17-09-2014	: paul     : CBS-5827 Don't change the input tag
 *  25-09-2017	: smitp    : CBS-7643 Don't print CCweb copyblock template as:
 *                         "<PP_Title:\n208@/-1 $b(0)\n209A/-1 $a(0) $x(2)00>"
 *                           instead print:
 *                         "<PP_Title:\n208@ ̸* $b(0)\n209A ̸* $a(0) $x(2)00>"
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static size_t 
tagToString(char ** out, size_t outlen, PPTag_p tag)
{
    PPSbf_p	sbf;
    if (ensureCapacity(out, outlen + strlen(" -040/03 "))){
	PICTagType_t	hideundo_typ = tag->PPTAG_typ;	/* Type of the tag
							   after unhiding */
	(*out)[outlen++] = '\n';
	if (hideundo_typ < 0) {
	    (*out)[outlen++] = '-';
	    hideundo_typ = -hideundo_typ;
	}
	outlen += sprintf((*out) + outlen, "%d%02d%c", tag->PPTAG_lev, 
			  hideundo_typ, tag->PPTAG_ind);
	if (tag->PPTAG_occ == PP_OCCURRENCE_ANY){
	    outlen += sprintf((*out) + outlen, "/%c",
			      PP_OCCURRENCE_ANY_NORMALIZED);
	} else if (tag->PPTAG_occ != 0){
	    outlen += sprintf((*out) + outlen, "/%02d", tag->PPTAG_occ);
	}
    }

    sbf = tag->PPTAG_firstsbf;
    while (sbf != NULL){
	outlen = subfieldToString(out, outlen, sbf);
	sbf = sbf->pp_eltnext;
    }
    return outlen;
}
/*------------------------------------------------------------------------
 *  Function	: titleToString
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  22-01-2002	: henk     : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static size_t 
titleToString(char ** out, int outlen, PPTtl_p title)
{
    PPTag_p	tag;
    tag = title->PPTTL_firsttag;
    while (tag != NULL) {
	outlen = tagToString(out, outlen, tag);
	tag = tag->pp_eltnext;
    }
    return outlen;
}
/*------------------------------------------------------------------------
 *  Function	: PPToString
 *  Purpose	: Convert a title to String for diagnostic purposes
 *  Method	: Repeated calls of TagToIntern().
 *
 *  History:
 *  21-01-2002 : HNvdWal
 *
 *  Returns	: Pointer to String
 *  Suggested use: printf("%s\n", PPToString(title));
 *------------------------------------------------------------------------
 */
char *
PPToString(
	PPTtl_p		title)		/* handle to title */
{ 
    size_t outlen = 0;

    static char *out;
    static char err[] = "ERROR in PPToString";

    if (! ensureCapacity(&out, 100)){
	    return err;
    }

    out[0] = '\0';
    (void)strcpy(out, "<PP_");
    if (title != NULL){
	if (title->pp_elttyp == PPT_TITLE) {
	    (void) strcat(out, "Title:");
	    outlen = strlen (out);
	    outlen = titleToString(&out, outlen, title);
	}else if(title->pp_elttyp == PPT_TAG){
	    (void) strcat(out, "Tag:");
	    outlen = strlen (out);
	    outlen = tagToString(&out, outlen, title);
	}else if(title->pp_elttyp == PPT_SUBFIELD){
	    (void) strcat(out, "Subfield:");
	    outlen = strlen (out);
	    outlen = subfieldToString(&out, outlen, title);
	}else if(title->pp_elttyp == PPT_NONE){
	    (void) strcat(out, "NONE");
	    outlen = strlen (out);
	}
    } else {
	return err;
    }
    if (!ensureCapacity(&out, outlen + 2)){
	return err;
    } else {
	out[outlen++] = '>';
	out[outlen] = '\0';
    }
    return out;
} /* PPToString */
