/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ppsort.c
 *  Description	: Sort a title
 *  Author(s)	: Hans van Eck
 *  Creation	: 21-02-1996
 *
 *  History:
 *  21-02-1996	: eck      : creation
 *  03-12-2002	: heesakkm : Support sorting of owner-blocks
 *  28-02-2003	: paul     : PPTtlSort() can now cope with NULL title
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  26-09-2005  : listj    : script ($T/$U) sorting 
 *  17-11-2005  : listj    : script sorting moved to pcccat_sortTransliterations
 *  21-11-2005  : listj    : sorting on PICA+ $T readded (independent from $U)
 *  08-04-2008  : paul     : added PPTtlSortCustom()
 *  13-10-2008	: paul     : fixed bug that caused 201B to be at the end of the
 *                           first copy record if no front_copynr specified
 *  05-03-2009	: paul     : Allow tags of level 3 (social)
 *  12-03-2009	: paul     : Use PIC_LEVEL_ORDER()
 *  05-08-2009	: sutherlr : Introduced owner_block_p object to improve
 *                           performance of building/sorting of PPTtl's.
 *  11-12-2009	: paul     : CBS-607: PPTtlSort should sort level 2 tags on $x
 *                           value
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ppsort.c 373351 2020-01-16 13:16:21Z smitp $"

#include <string.h>
#include <stdlib.h>

#include <pica.h>
#include <picatypes.h>
#include <picamisc.h>
#include <pp.h>
#include <pa.h>
#include <er.h>
#include <env.h>

#include "pplocal.h"

#define ABS(val)	(val >= 0 ? val : -val)

#define IS_OWNER_TAG(tag)					\
			((tag != NULL)				\
			 && (tag->PPTAG_lev == PPLEV_LOCAL)	\
			 && (tag->PPTAG_typ == 1)		\
			 && (tag->PPTAG_ind == '@')		\
			 && (tag->PPTAG_occ == 0) ? TRUE : FALSE)

/*
 * owner_block_p
 */
typedef struct {
    PICIln_t	iln;
    PPTtl_p	owner_title_part;   /* local and copies of a single ILN */
    int		original_index;	    /* for stable sorting */
} *owner_block_p;

// Possible PPTTLSORT_OPTIONS specified in FILEMAP:
typedef enum {
    PPEO_UNINITIALISED = 0,
    PPEO_INITIALISED  = 1 << 0,
    PPEO_LINKAGE_SORT = 1 << 1
} pp_env_opt_t;

// Actual PPTTLSORT_OPTIONS specified in FILEMAP:
static pp_env_opt_t	pp_env_opt_g = PPEO_UNINITIALISED;


/*------------------------------------------------------------------------
 *  Function	: pp_env_opt_initialise
 *  Purpose	: Get title-sort options from FILEMAP
 *
 *  History:
 *  10-02-2016	: paul     : CBS-6257 creation
 *
 *  Example     :
 *  PPTTLSORT_OPTIONS=linkage_sort=no;
 *
 *------------------------------------------------------------------------
 */
static void
pp_env_opt_initialise(void)
{
    const char	*env_opt = ENVGetVar("PPTTLSORT_OPTIONS");

    char	optval[100];

    pp_env_opt_g = PPEO_INITIALISED;

    // Do we have to sort linkage-occs in $T subfields?
    if (PM_opt_fetch(env_opt, "linkage_sort", optval,
		     sizeof(optval)) == 0
	|| !STREQ(optval, "no")) {
	pp_env_opt_g |= PPEO_LINKAGE_SORT;
    }
} // pp_env_opt_initialise()


/*------------------------------------------------------------------------
 *  Function	: pp_linkage_sort_tag
 *  Purpose	: Move parallel tags (with same linkage occ) directly behind
 *                current tag
 *
 *  History:
 *  10-02-2016	: paul     : CBS-6257 creation
 *
 *  Notes	:
 *  Input 'title' is supposed to be sorted with 'tagcmp'.
 *  'tagcmp' is supposed to compare on tag-code; it should return 0 when
 *  called for parallel fields.
 *
 *  Example     :
 *  Current tag is the first of this title:
 *  028C  $T02 $Uea $a丘宗孔 $9343272318
 *  028C  $T01 $Uea $a李廷機 $9149329237
 *  028C  $T00 $Uba $5Zheng $lShihao $9343272350
 *  028C  $T02 $Uba $5Qiu $lZongkong $9343272318
 *  028C  $dPaul $aSmit
 *  028C  $T01 $Uba $5Li $lTingji $9149329237
 *  028C  $T00 $Uea $a鄭世豪 $9343272350
 *  033A  $p[S.l.] $nShuangsongguan
 *
 *  When function returns, title will be modified as follows:
 *  028C  $T02 $Uea $a丘宗孔 $9343272318
 *  028C  $T02 $Uba $5Qiu $lZongkong $9343272318
 *  028C  $T01 $Uea $a李廷機 $9149329237
 *  028C  $T00 $Uba $5Zheng $lShihao $9343272350
 *  028C  $dPaul $aSmit
 *  028C  $T01 $Uba $5Li $lTingji $9149329237
 *  028C  $T00 $Uea $a鄭世豪 $9343272350
 *  033A  $p[S.l.] $nShuangsongguan
 *
 *  And the return value points to the second tag of this title.
 *
 *  Returns	:
 *  - Pointer to the last tag of the linkage occ that has been sorted
 *
 *------------------------------------------------------------------------
 */
static PPTag_p
pp_linkage_sort_tag(
	PPTag_p		tag,		// Last of set of 1 parallel tag
	const char	*linkage,	// Value of $T in 'tag' (not "00")
	PPTagCmp_f	tagcmp)		// Tag compare function ignoring $T
{
    PPTag_p		tag2 = tag;
    
    // Walk thru tags until the tag-code differs from input 'tag'
    while ((tag2 = tag2->pp_eltnext) != NULL
	   && tagcmp(tag, tag2) == 0) {
	PPSbf_p		linkage_sbf2 = PPSbfFind(tag2,
						 PIC_SUBFTYPE_LINKAGE_OCC);
	const char	*linkage2;

	// Skip tags without $T
	if (linkage_sbf2 == NULL) {
	    continue;
	}
	linkage2 = (char *)linkage_sbf2->PPSBF_contents;

	// Skip tags with a different $T
	if (!STREQ(linkage2, linkage)) {
	    continue;
	}

	// If this tag is not immediately behind the last of current parallel
	// tags then move it so it is immediately behind it
	if (tag2 != tag->pp_eltnext) {
	    PPEltUnlink(tag2);
	    PPEltInsertBetween(tag, tag2, tag->pp_eltnext);
	}

	// Keep track of the last of current parallel tags
	tag = tag2;
    }

    // Return the last tag of the linkage occ we've just sorted
    return tag;
} // pp_linkage_sort_tag()


/*------------------------------------------------------------------------
 *  Function	: pp_linkage_sort
 *  Purpose	: Sort tags on linkage-occ in $T
 *
 *  History:
 *  10-02-2016	: paul     : CBS-6257 creation
 *
 *  Notes	:
 *  The 'linkage occ' subfield $T links parallel fields. Fields can only be
 *  parallel if they have identical tag-codes.
 *
 *  Input 'title' is supposed to be sorted with 'tagcmp'.
 *  'tagcmp' is supposed to compare on tag-code; it should return 0 when
 *  called for parallel fields.
 *  As a result, parallel fields cannot be far apart.
 *
 *  $T00 is a special linkage-occ, it indicates that the field doesn't have a
 *  parallel field.
 *
 *  Example     :
 *  Input title:
 *  028C  $T02 $Uea $a丘宗孔 $9343272318
 *  028C  $T00 $Uba $5Zheng $lShihao $9343272350
 *  028C  $T02 $Uba $5Qiu $lZongkong $9343272318
 *  028C  $T01 $Uba $5Li $lTingji $9149329237
 *  028C  $T00 $Uea $a鄭世豪 $9343272350
 *  028C  $dPaul $aSmit
 *  028C  $T01 $Uea $a李廷機 $9149329237
 *  033A  $p[S.l.] $nShuangsongguan
 *
 *  Output title:
 *  028C  $T02 $Uea $a丘宗孔 $9343272318
 *  028C  $T02 $Uba $5Qiu $lZongkong $9343272318
 *  028C  $T00 $Uba $5Zheng $lShihao $9343272350
 *  028C  $T01 $Uba $5Li $lTingji $9149329237
 *  028C  $T01 $Uea $a李廷機 $9149329237
 *  028C  $T00 $Uea $a鄭世豪 $9343272350
 *  028C  $dPaul $aSmit
 *  033A  $p[S.l.] $nShuangsongguan
 *
 *------------------------------------------------------------------------
 */
static void
pp_linkage_sort(
	PPTtl_p		title,		// Sorted title, apart from $T values
	PPTagCmp_f	tagcmp)		// Tag compare function ignoring $T
{
    for (PPTag_p tag = title->PPTTL_firsttag;
	 tag != NULL;
	 tag = tag->pp_eltnext) {
	PPSbf_p		linkage_sbf = PPSbfFind(tag,
						PIC_SUBFTYPE_LINKAGE_OCC);
	const char	*linkage;

	// Skip tags without $T
	if (linkage_sbf == NULL) {
	    continue;
	}
	linkage = (char *)linkage_sbf->PPSBF_contents;

	// Skip tags with $T00
	if (STREQ(linkage, PIC_LINKAGE_OCC_NONE)) {
	    continue;
	}

	// Move parallel tags directly behind current
	tag = pp_linkage_sort_tag(tag, linkage, tagcmp);
    }
} // pp_linkage_sort()


/*------------------------------------------------------------------------
 *  Function	: PPTagCmp
 *  Purpose	: Compare two tags
 *  Method	:
 *  First compare absolute levels (type can be negative when hidden).
 *  If both levels are 2 then
 *	compare in the order occ/typ/ind
 *  Else
 *	compare in the order typ/ind/occ
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  21-02-1996	: eck      : Copied from pptag.c
 *  30-10-2000  : HvdW	   : Interface in pplocal.h
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  26-09-2005  : listj    : script ($T/$U) sorting
 *  17-11-2005  : listj    : script sorting moved to PCCCATSortTransliterations
 *  21-11-2005  : listj    : sorting on PICA+ $T readded (independent from $U)
 *
 *  Returns	:
 *  <0 if 'tag' < 'cmptag'
 *   0 if 'tag' = 'cmptag'
 *  >0 if 'tag' > 'cmptag'
 *------------------------------------------------------------------------
 */
int
PPTagCmp(
	PPTag_p	tag,		/* the new tag */
	PPTag_p	cmptag)		/* tag to compare it to */
{
    PP_TagType_t  typ, cmptyp;
    PP_Level_t	  lev, cmplev;
    int		  cmpval;
    PPSbf_p	  sbf 			= NULL;
    PPSbf_p	  cmpSbf 		= NULL;
    const char	  *sbfCont 		= NULL;
    const char	  *cmpSbfCont 		= NULL;

    // Get options from FILEMAP, unless already done
    if (pp_env_opt_g == PPEO_UNINITIALISED) {
	pp_env_opt_initialise();
    }

    typ = ABS(tag->PPTAG_typ);
    cmptyp = ABS(cmptag->PPTAG_typ);

    lev = tag->PPTAG_lev;
    cmplev = cmptag->PPTAG_lev;

    /* Level 2 tags are sorted specially. */
    if ((lev != cmplev) || (lev != 2)) {
	cmpval = PIC_LEVEL_ORDER(lev) - PIC_LEVEL_ORDER(cmplev);
	if (cmpval == 0) {
	    cmpval = typ - cmptyp;
	}
	if (cmpval == 0) {
	    cmpval = tag->PPTAG_ind - cmptag->PPTAG_ind;
	}
	if (cmpval == 0) {
	    cmpval = tag->PPTAG_occ - cmptag->PPTAG_occ;
	}
    } else {
	cmpval = tag->PPTAG_occ - cmptag->PPTAG_occ;
	if (cmpval == 0) {
	    cmpval = typ - cmptyp;
	}
	if (cmpval == 0) {
	    cmpval = tag->PPTAG_ind - cmptag->PPTAG_ind;
	}

	/* Compare $x occurrence */
	if (cmpval == 0) {
	    PPSbf_p	xoccsbf = PPSbfFind(tag, PIC_XOCC_SBFTYP);
	    PPSbf_p	cmpxoccsbf = PPSbfFind(cmptag, PIC_XOCC_SBFTYP);
	    const char	*xocc = (xoccsbf != NULL
				 ? (const char *)xoccsbf->PPSBF_contents
				 : PIC_XOCC_DEFAULT);
	    const char	*cmpxocc = (cmpxoccsbf != NULL
				    ? (const char *)cmpxoccsbf->PPSBF_contents
				    : PIC_XOCC_DEFAULT);
	    
	    cmpval = strcmp(xocc, cmpxocc);
	}
    }

    // Sort on PICA+ subfield $T, unless "linkage_sort=no"
    if (cmpval == 0 && (pp_env_opt_g & PPEO_LINKAGE_SORT)) {
	sbf = PPSbfFind(tag, PIC_SUBFTYPE_LINKAGE_OCC);
	if (sbf == NULL) {
	    sbfCont = "";
	} else {
	    sbfCont = sbf->PPSBF_contents;
	}

	cmpSbf = PPSbfFind(cmptag, PIC_SUBFTYPE_LINKAGE_OCC);
	if (cmpSbf == NULL) {
	    cmpSbfCont = "";
	} else {
	    cmpSbfCont = cmpSbf->PPSBF_contents;
	}

	cmpval = strcmp(sbfCont, cmpSbfCont);
    }

    return cmpval;
} /* PPTagCmp() */


/*------------------------------------------------------------------------
 *  Function	: MoveFrontCopy
 *  Purpose	: Move the specified copy number to the front of the owner
 *  Method	: Insert in the approprate place or append
 *
 *  History:
 *  08-04-2008	: paul     : creation
 *  13-10-2008	: paul     : fixed test if front_copy is already in front
 *
 *  Notes	: 
 *
 *------------------------------------------------------------------------
 */
static void
MoveFrontCopy(
	PPTtl_p		owner,		/* The owner-block to add */
	int16		front_copynr)	/* Copy nr. to put in front of others*/
{
    PPTag_p	tag, prev;
    PPTag_p	front_copy, first_copy;

    if (front_copynr == PIC_COPYNUM_NONE) {
	/* Delete all copies from this owner block */
	while ((tag = PPTagFind(owner, PP_ANYTYP, PPLEV_COPY, PP_ANYIND,
				PP_ANYOCC))
	       != NULL) {
	    PPEltDestroy(tag);
	}
	return;
    }

    /* Find the first copy record */
    first_copy = PPTagFind(owner, PP_ANYTYP, PPLEV_COPY, PP_ANYIND, PP_ANYOCC);
    if (first_copy == NULL) {
	/* block contains no copy records */
	return;
    }
    if (front_copynr == PP_ANYOCC
	|| front_copynr == first_copy->PPTAG_occ) {
	/* front_copy is already in front */
	return;
    }	

    /* Find the first tag of the front_copy */
    front_copy = PPTagFind(first_copy, PP_ANYTYP, PPLEV_COPY, PP_ANYIND,
			   front_copynr);
    if (front_copy == NULL) {
	/* block doesn't contain front_copy */
	return;
    }

    /* Move all front_copy tags in front of first_copy */
    do {
	prev = front_copy->pp_eltprev;
	PPEltUnlink(front_copy);
	PPEltListInsert(first_copy, front_copy);
    } while ((front_copy = PPTagFind(prev, PP_ANYTYP, PPLEV_COPY, PP_ANYIND,
				     front_copynr))
	     != NULL);
} /* MoveFrontCopy() */


/*------------------------------------------------------------------------
 *  Function	: AddSortedTag
 *  Purpose	: Add a tag to a (partially) sorted tag-list
 *  Method	: Insert in the approprate place
 *
 *  History:
 *  22-02-1996	: eck      : creation
 *  03-12-2002	: heesakkm : Renamed from AddSorted
 *------------------------------------------------------------------------
 */
static void
AddSortedTag(
	PPTtl_p		title,		/* The title to add the tag to */
	PPTag_p		tag,		/* The tag to add */
	PPTagCmp_f	tagcmp)		/* Tag compare function */
{
    PPElement_p	prev_elt;
    PPTag_p	cmptag;

    prev_elt = title;
    cmptag = title->PPTTL_firsttag;

    while ((cmptag != NULL) && (tagcmp(tag, cmptag) >= 0)) {
	prev_elt = cmptag;
	cmptag = cmptag->pp_eltnext;
    }
    PPEltInsertBetween(prev_elt, tag, cmptag);
} /* AddSortedTag() */

/*------------------------------------------------------------------------
 *  Function	: AddSortedTagBackwards
 *  Purpose	: Add a tag to a sorted title, comparing from last to first
 *
 *  History:
 *  06-06-2018	: smitp    : CBS-8221 creation
 *
 *  Notes	:
 *  SortTtlOwners() uses this function instead of AddSortedTag() because
 *  it is better for performance. See Description of SortTtlOwners()
 *
 *  Returns	: The new last tag of the sorted title
 *
 *------------------------------------------------------------------------
 */
static PPTag_p
AddSortedTagBackwards(
	PPElement_p	last,		/* The last tag of the title to add to
					   or the title if it contains no tag*/
	PPTag_p		tag,		// The tag to append
	PPTagCmp_f	tagcmp)		// Tag compare function
{
    PPElement_p	next_tag = NULL;
    PPElement_p	cmpelt = last;

    // Find the last tag that sorts equal to or in front of our current tag
    while ((cmpelt->pp_elttyp == PPT_TAG) && (tagcmp(tag, cmpelt) < 0)) {
	next_tag = cmpelt;
	cmpelt = cmpelt->pp_eltprev;
    }

    PPEltInsertBetween(cmpelt, tag, next_tag);

    return next_tag == NULL ? tag : last;
} // AddSortedTagBackwards()

/*------------------------------------------------------------------------
 *  Function	: ob_create
 *  Purpose	: Create and initialize an owner_block_p
 *  Method	: 
 *
 *  History:
 *  05-08-2009	: sutherlr : creation
 *
 *  Notes	: owner_block_p->original_index initialized here but 
 *                is properly set in AddSortedOwner() when this data i
 *                available
 *
 *  Returns	: a valid owner_block_p or  NULL if error
 *
 *------------------------------------------------------------------------
 */
static owner_block_p
ob_create(PPTtl_p ownerTtl) 
{
    owner_block_p new_owner = malloc(sizeof(*new_owner));

    if (new_owner == NULL) return NULL;

    /* Get the owner iln from 101@$a of ownerTtl */
    if (ownerTtl->PPTTL_firsttag->PPTAG_firstsbf != NULL
	    && ownerTtl->PPTTL_firsttag->PPTAG_firstsbf->PPSBF_typ == 'a'
	    && ownerTtl->PPTTL_firsttag->PPTAG_firstsbf->PPSBF_contents != NULL) {

	new_owner->iln = atoi(ownerTtl->PPTTL_firsttag->PPTAG_firstsbf->PPSBF_contents);
    } else {
	new_owner->iln = PIC_ILN_NONE;
    }
    new_owner->owner_title_part = ownerTtl;

    new_owner->original_index = 0;

    return new_owner;
} /* ob_create() */

/*------------------------------------------------------------------------
 *  Function	: ob_destroy
 *  Purpose	: properly free up the memory used by a owner_block_p
 *  Method	: 
 *
 *  History:
 *  05-08-2009	: sutherlr : creation
 *
 *  Returns	: void
 *
 *------------------------------------------------------------------------
 */
static void
ob_destroy(owner_block_p ob) 
{
    if (ob->owner_title_part != NULL) {
	PPEltDestroy(ob->owner_title_part);
	ob->owner_title_part = NULL;
    }
    free(ob);
}

/*------------------------------------------------------------------------
 *  Function	: ob_compare
 *  Purpose	: Comparison callback function for PA_Sort of 
 *                owner_block_p's
 *  Method	: if both owner_block_p's have the same iln,
 *                then a comparison is done on original_index values
 *
 *  History:
 *  05-08-2009	: sutherlr : creation
 *
 *  Notes	: 
 *
 *  Returns	: <0 if arg2 is larger
 *                >0 if arg2 is larger
 *
 *------------------------------------------------------------------------
 */
static int
ob_compare(
	cvoidp arg1,
	cvoidp arg2)
{
    owner_block_p ob1 = (owner_block_p)arg1;
    owner_block_p ob2 = (owner_block_p)arg2;

    if (ob1->iln == ob2->iln) {
	return ob1->original_index - ob2->original_index;
    }
    if (ob1->iln == PIC_ILN_NONE) {
	return -1;
    }
    if (ob2->iln == PIC_ILN_NONE) {
	return 1;
    }
    return ob1->iln - ob2->iln;
} /* ob_compare() */

/*------------------------------------------------------------------------
 *  Function	: AddSortedOwner
 *  Purpose	: Add a sorted owner-block to a (partially) sorted tag-list
 *  Method	: Insert in the approprate place or append
 *
 *  History:
 *  03-12-2002	: heesakkm : creation
 *  05-08-2009	: sutherlr : now simply adding newly-created owner_block_p
 *                           to a PA
 *
 *  Notes	: this function sets owner_block_p->original_index
 *
 *  Returns	: bool: success/failure indication
 *
 *------------------------------------------------------------------------
 */
static bool
AddSortedOwner(
	PAHandle_p	owner_array,	/* Array to add the owner-block to */
	PPTtl_p		owner)		/* The owner-block to add */
{
    owner_block_p ob = (owner_block_p)PA_Add(owner_array, ob_create(owner));
    
    if (ob != NULL) {
	ob->original_index = PA_GetNrEntries(owner_array);
    }

    return (ob != NULL);
} /* AddSortedOwner() */


/*------------------------------------------------------------------------
 *  Function	: SortTtl
 *  Purpose	: Sort a title that has zero or one owner-block
 *  Method	: Insertion sort
 *
 *  History:
 *  03-12-2002	: heesakkm : creation from PPTtlSort()
 *  16-01-2020	: smitp    : CBS-9041 Don't sort if "inputsort=no"
 *
 *  Notes	:
 *  The tag-list is usually mostly sorted, so the insertion sort is
 *  adequate enough.
 *
 *  Returns	: void
 *
 *------------------------------------------------------------------------
 */
static void
SortTtl(
	PPTtl_p		title,		/* Title to sort */
	PPTagCmp_f	tagcmp,		// Tag compare function
	bool		*linkage,	// out: $T found? or NULL
	bool		inputsort)	// Do we need to sort at all?
{
    PPTag_p	cur_tag;
    PPTag_p	next_tag;

    cur_tag = title->PPTTL_firsttag;
    next_tag = cur_tag->pp_eltnext;

    // Keep track of the presence of linkage-occ subfield $T, if desired
    if (linkage != NULL && !*linkage) {
	*linkage = PPSbfFind(cur_tag, PIC_SUBFTYPE_LINKAGE_OCC) != NULL;
    }
    
    while (next_tag != NULL) {

	// Keep track of the presence of linkage-occ subfield $T, if desired
	if (linkage != NULL && !*linkage) {
	    *linkage = PPSbfFind(next_tag, PIC_SUBFTYPE_LINKAGE_OCC) != NULL;
	}
    
	if (inputsort && tagcmp(cur_tag, next_tag) > 0) {
	    PPEltUnlink(next_tag);
	    AddSortedTag(title, next_tag, tagcmp);
	} else {
	    cur_tag = cur_tag->pp_eltnext;
	}
	next_tag = cur_tag->pp_eltnext;
    }
} /* SortTtl() */

/*------------------------------------------------------------------------
 *  Function	: append_to_sorted
 *  Purpose	: append the taglist from an owner_block_p->owner_title_part
 *                to another PPTtl
 *  Method	: 
 *
 *  History:
 *  05-08-2009	: sutherlr : creation
 *  14-08-2019	: smitp    : CBS-8853 Only append 101@ if the ILN changed
 *
 *  Notes	: because this function returns a pointer to the last tag
 *                be careful not to lose the pointer to the tag itself
 *
 *  The FORMATSPEC may have specified:
 *  <PROCESSING NAME="local-copy-sort" KEY="FCV#pica#localcopysort"/>
 *  <OPTION NAME="prs-inputsort" VALUE="no"/>
 *
 *  In that case there may be more than one 101@ fields from the same ILN!
 *  Goal of this function is to only output a 101@ field if the ILN changes.
 *
 *  Returns	: pointer to the last tag of the PPttl
 *
 *------------------------------------------------------------------------
 */
static PPTag_p
append_to_sorted(
	PPTag_p		sorted_title_last_tag,
	owner_block_p	ob,
	PICIln_t	sorted_title_last_iln)
{
    PPTag_p	owner_firsttag;
    
    // Only append 101@ if the ILN changed
    if (ob->iln == sorted_title_last_iln) {
	PPEltDestroy(ob->owner_title_part->PPTTL_firsttag);
    }

    // Unlink the PP-tags from the PP-title
    owner_firsttag = ob->owner_title_part->PPTTL_firsttag;
    ob->owner_title_part->PPTTL_firsttag = NULL;

    if (owner_firsttag != NULL) {
	owner_firsttag->pp_eltprev = NULL;
	PPEltListAppend(sorted_title_last_tag, owner_firsttag);
    }
    
    // Find the last tag in sorted_title
    while (sorted_title_last_tag->pp_eltnext != NULL) {
	sorted_title_last_tag = sorted_title_last_tag->pp_eltnext;
    }
    return sorted_title_last_tag;
} /* append_to_sorted() */

/*------------------------------------------------------------------------
 *  Function	: handle_front
 *  Purpose	: insert the "front iln" copy record at the front of the
 *                copies list
 *  Method	: 
 *
 *  History:
 *  05-08-2009	: sutherlr : creation
 *
 *  Notes	: NOTE that this function returns a pointer to the end of
 *                the entire PPTtl and NOT the end of the "front iln" copy
 *
 *  Returns	: pointer to last tag of PPttl
 *
 *------------------------------------------------------------------------
 */
static PPTag_p
handle_front(
	PPTag_p		sorted_title_last_tag, 
	PAHandle_p	owner_array, 
	PICIln_t	front_iln, 
	PICCopyNum_t	front_copynr)
{
    int	    i, max;
    PICIln_t	sorted_title_last_iln = PIC_ILN_NONE;

    if (front_iln == PIC_ILN_ANY) {
	return sorted_title_last_tag;
    }

    max = PA_GetNrEntries(owner_array);
    for (i = 0; i < max; i++) {
	owner_block_p ob = (owner_block_p)PA_GetN(owner_array, i);
	if (ob->iln == front_iln) {
	    MoveFrontCopy(ob->owner_title_part, front_copynr);

	    sorted_title_last_tag = append_to_sorted(sorted_title_last_tag, ob, sorted_title_last_iln);
	    sorted_title_last_iln = ob->iln;
	    ob_destroy(ob);
	    PA_SetN(owner_array, i, NULL);
	}
    }

    return sorted_title_last_tag;
} /* handle_front() */

/*------------------------------------------------------------------------
 *  Function	: append_remaining_owners
 *  Purpose	: add all the remaining copies to a PPttl
 *  Method	: linear traversal
 *
 *  History:
 *  05-08-2009	: sutherlr : creation
 *
 *  Notes	: 
 *
 *  Returns	: pointer to last tag of completed PPttl
 *
 *------------------------------------------------------------------------
 */
static PPTag_p
append_remaining_owners(
	PPTag_p		sorted_title_last_tag,
	PAHandle_p	owner_array)
{
    int	    i, max;
    PICIln_t	sorted_title_last_iln = PIC_ILN_NONE;

    max = PA_GetNrEntries(owner_array);
    for (i = 0; i < max; i++) {
	owner_block_p ob = (owner_block_p)PA_GetN(owner_array, i);

	if (ob != NULL) {
	    sorted_title_last_tag = append_to_sorted(sorted_title_last_tag, ob, sorted_title_last_iln);
	    sorted_title_last_iln = ob->iln;
	    ob_destroy(ob);
	    PA_SetN(owner_array, i, NULL);
	}
    }
    return sorted_title_last_tag;
} /* append_remaining_owners() */

/*------------------------------------------------------------------------
 *  Function	: SortTtlOwners
 *  Purpose	: Sort a title that contains multiple owner-blocks
 *  Method	: Linear sort.
 *
 *  History:
 *  03-12-2002	: heesakkm : creation
 *  05-08-2009	: sutherlr : Modified this function to use new 
 *                           owner_block_p structure for level-2 data.
 *                           Copy records are eventually sorted using the
 *                           PA_Sort() functionality
 *  06-06-2018	: smitp    : CBS-8221 Use AddSortedTagBackwards() instead of
 *                           AddSortedTag() to solve performance problem when
 *                           presenting a title with 999 copy blocks
 *  14-08-2019	: smitp    : CBS-8853 Don't sort copy blocks if "inputsort=no"
 *                           but do move the front_iln in front of other ILN's
 *  16-01-2020	: smitp    : CBS-9180 Fixed bug: last sorted_owner wasn't put into sorted_title
 *  16-01-2020	: smitp    : CBS-9180 Fixed bug: artificial tags behind the last owner-block should not remain there
 *
 *  Notes	: the argument 'title' is both the input and output PPTtl
 *
 *                This function will call ERAbort() on serious error.
 *
 *  Description	:
 *  The general structure of the input title is:
 *  - level-0 and level-3 tags
 *  - owner blocks
 *  - artificial tags
 *
 *  A level-0 and level-3 tag can occur inside/behind an owner block, in which case it needs to be sorted to the front
 *  The first artificial tag is a tag with a level greater than 3: such a tag cannot occur in a database record, for example:
 *	999V $aC $b2019-12-23 08:18:35.726 $c263654428 $dcsn_ipusher/626794 $e1602 $f31
 *  No real tag can occur after the first artificial tag if "inputsort=no"
 *
 *  "inputsort=no" means that individual tags must not be sorted, but the owner block of the front_iln must be sorted to the front.
 *
 *  Sort the level-0 and level-3 tags first.
 *  Then for each ILN: sort the level-1 and level-2 tags.
 *  Save the level-1 and level-2 tags of each ILN in an array and sort this
 *  array.
 *
 *  Both sorting level-0/3 tags and sorting an ILN is done by repeating:
 *  - unlink the first tag of the unsorted title
 *  - add it to the correct spot in the sorted title
 *
 *  Most titles are already correctly sorted.
 *
 *  If a title is already correctly sorted, the correct spot in the sorted
 *  title is behind the last tag.
 *  Hence we use AddSortedTagBackwards() rather than AddSortedTag()
 *
 *  Example	: input title of the title-revisions-presentation:
 *      001@ $031
 *      001A $%263652879 $01009:19-08-19
 *      001B $%263652998 $01009:21-08-19 $t09:16:47.000
 *      001D $%263652879 $01009:19-08-19
 *      001U $%263652879 $0utf8
 *      001X $%263652879 $00
 *      001Z $%263652884 $01009
 *      002@ $%263652879 $0Aax
 *      002C $%263652879 $atekst $btxt $2rdacontent/dut
 *      002D $%263652879 $azonder medium $bn $2rdamedia/dut
 *      002E $%263652879 $aband $bnc $2rdacarrier/dut
 *      003@ $%263652879 $041717120X
 *      003O $%263652880 $aOCoLC $0970342943
 *      011@ $%263652879 $a2019
 *      021A $%263652998 $a@Test titel
 *      033A $%263652879 $pOCLC
 *      034I $%263652879 $a15x15cm
 *      045Q/03 $%263652884 $9077611667 $876.65 sportaccomodaties
 *      101@ $a31
 *      101B $%263654428 $023-12-19 $t08:18:35.000
 *      101U $%263654427 $0utf8
 *      103O $%263654428 $aOCoLC $0588688304
 *      147A $%263654427 $ayrdey
 *      201B/02 $%263652888 $019-08-19 $t11:00:16.000
 *      201U/02 $%263652888 $0utf8
 *      203@/02 $%263652888 $01208895753
 *      208@/02 $%263652888 $a19-08-19 $btest
 *      201B/03 $%263652892 $019-08-19 $t11:47:00.000
 *      201U/03 $%263652892 $0utf8
 *      203@/03 $%263652892 $01208895761
 *      208@/03 $%263652892 $a19-08-19 $btest2
 *      999V $aC $b2019-12-23 08:18:35.726 $c263654428 $dcsn_ipusher/626794 $e1602 $f31
 *      999V $aC $b2019-12-23 08:18:30.616 $c263654427 $dcsn_cat/626793 $e1602 $f31
 *      999V $aC $b2019-08-21 09:16:47.406 $c263652998 $dcsn_cat/626602 $e1602
 *      999V $aC $b2019-08-21 09:16:30.943 $c263652997 $dcsn_cat/626602 $e1602
 *      999V $aC $b2019-08-19 11:47:00.326 $c263652892 $dcsn_cat/626581 $e1602 $f31
 *      999V $aC $b2019-08-19 11:46:43.530 $c263652891 $dcsn_ipusher/626582 $e1602
 *      999V $aC $b2019-08-19 11:46:39.770 $c263652890 $dcsn_cat/626581 $e1602 $f31
 *      999V $aC $b2019-08-19 11:00:23.140 $c263652889 $dcsn_ipusher/626582 $e1602
 *      999V $aC $b2019-08-19 11:00:16.560 $c263652888 $dcsn_cat/626581 $e1602 $f31
 *      999V $aC $b2019-08-19 11:00:03.900 $c263652887 $dcsn_cat/626581 $e1602 $f31
 *      999V $aC $b2019-08-19 10:59:40.843 $c263652886 $dcsn_cat/626581 $e1602 $f31
 *      999V $aO $b2019-08-19 10:14:45.723 $c263652885 $dcsn_cat/626581 $e1602 $f31
 *      201B/01 $%A $019-08-19 $t10:14:45.000
 *      201U/01 $%A $0utf8
 *      203@/01 $%A $01208895737
 *      208@/01 $%A $a19-08-19 $bdssdds
 *      999V $aC $b2019-08-19 10:05:18.520 $c263652884 $dcsn_cat/626581 $e1602
 *      999V $aC $b2019-08-19 09:51:00.536 $c263652880 $dcsn_ipusher/626582 $e1602
 *      999V $aC $b2019-08-19 09:50:57.530 $c263652879 $dcsn_cat/626581 $e1602
 *      901U $0utf8
 *
 *  Returns	: void
 *
 *------------------------------------------------------------------------
 */
static void
SortTtlOwners(
	PPTtl_p		title,		/* Title to sort */
	PPTagCmp_f	tagcmp,		/* Tag compare function */
	int16		front_iln,	// ILN to move in front of other ILN's (or PIC_ILN_ANY)
	int16		front_copynr,
	bool		*linkage,	// out: $T found? or NULL
	bool		inputsort)	// inputsort=no in case of local-copy-sort
{
    PPTtl_p	sorted_title; 
    PPTag_p	sorted_title_end;
    PPTtl_p	sorted_owner = NULL;
    PPTag_p	cur_tag, next_tag; 
    PAHandle_p	owner_array = NULL;
    PPElement_p	last = NULL;

    if ((sorted_title = PPTtlCreate()) == NULL) return;
    owner_array = PA_Create(NULL, PA_NOFUNC);
    if (owner_array == NULL) {
	PPTtlDestroy(sorted_title);
	return;
    }
    
    /* Link all level 0 and 3 tags to sorted_title */
    last = sorted_title;
    cur_tag = title->PPTTL_firsttag;
    while (cur_tag != NULL) {

	// There can be no real level-0/3 tag behind the first artificial tag if "inputsort=no"
	if (!inputsort && cur_tag->PPTAG_lev > PPLEV_SOCIAL) {
	    break;
	}

	// Keep track of the presence of linkage-occ subfield $T, if desired
	if (linkage != NULL && !*linkage) {
	    *linkage = PPSbfFind(cur_tag, PIC_SUBFTYPE_LINKAGE_OCC) != NULL;
	}
    
	next_tag = cur_tag->pp_eltnext;
	if (cur_tag->PPTAG_lev == PPLEV_MAIN
	    || cur_tag->PPTAG_lev == PPLEV_SOCIAL) {
	    PPEltUnlink(cur_tag);
	    if (inputsort) {
		last = AddSortedTagBackwards(last, cur_tag, tagcmp);
	    } else {
		last = PPEltListAppend(last, cur_tag);
	    }
	}
	cur_tag = next_tag;
    }

    // title now contains zero or more owner-blocks followed by zero or more artificial tags
    cur_tag = title->PPTTL_firsttag;
    while (cur_tag != NULL) {

	// There can be no real level-1/2 tag behind the first artificial tag if "inputsort=no"
	if (!inputsort && cur_tag->PPTAG_lev > PPLEV_SOCIAL) {
	    break;
	}

	next_tag = cur_tag->pp_eltnext;

	if (sorted_owner == NULL) {
	    if ((sorted_owner = PPTtlCreate()) == NULL) {
		PPTtlDestroy(sorted_title);
		return;
	    }
	    last = sorted_owner;
	}

	/* Link cur_tag to sorted_owner */
	PPEltUnlink(cur_tag);

	if (inputsort) {
	    last = AddSortedTagBackwards(last, cur_tag, tagcmp);
	} else {
	    // FORMATSPEC specifies:
	    // <PROCESSING NAME="local-copy-sort" KEY="FCV#pica#localcopysort"/>
	    // <OPTION NAME="prs-inputsort" VALUE="no"/>
	    //
	    // Don't change the order of copy blocks: suppress sorting
	    last = PPEltListAppend(last, cur_tag);
	}

	if (next_tag == NULL || IS_OWNER_TAG(next_tag)) {
	    /* Reached the end of the title or the start of a new owner */

	    /* Link sorted_owner to sorted_title */
	    if (! AddSortedOwner(owner_array, sorted_owner)) {
		ERAbort(PPerrno, "SortTtlOwners",
			"Cannot store sorted_owner to owner_array:\n%s",
			PPToString(sorted_owner));
	    }
	    sorted_owner = NULL;

	}
	cur_tag = next_tag;
    }

    // Link the last sorted_owner to sorted_title
    if (sorted_owner != NULL) {
	if (! AddSortedOwner(owner_array, sorted_owner)) {
	    ERAbort(PPerrno, "SortTtlOwners",
		    "Cannot store sorted_owner to owner_array:\n%s",
		    PPToString(sorted_owner));
	}
	sorted_owner = NULL;
    }

    sorted_title_end = sorted_title->PPTTL_firsttag;
    if (sorted_title_end != NULL) {
	while (sorted_title_end->pp_eltnext != NULL) {
	    sorted_title_end = sorted_title_end->pp_eltnext;
	}
    }

    PA_Sort(owner_array, ob_compare);
    sorted_title_end = handle_front(sorted_title_end, owner_array, front_iln, front_copynr);
    append_remaining_owners(sorted_title_end, owner_array);

    // Append the artificial tags
    if (!inputsort && cur_tag != NULL) {
	PPEltListAppend(sorted_title, cur_tag);
    }

    /* title should now be empty */
    if (title->PPTTL_firsttag != NULL) {
	ERAbort(PPerrno, "SortTtlOwners",
		"unsorted title is not empty:\n%s",
		PPToString(title));
    }

    /* Link sorted tags to the original title */
    if (PPEltListAppend(title, sorted_title->PPTTL_firsttag) == NULL) {
	ERAbort(PPerrno, "SortTtlOwners",
		"Cannot link sorted_title to title");
    }

    PPEltDestroy(sorted_owner);
    PPEltDestroy(sorted_title);
    PA_Destroy(owner_array);
} /* SortTtlOwners() */


/*------------------------------------------------------------------------
 *  Function	: PPTtlSortCustom
 *  Purpose	: Sort a title
 *  Method	: SortTtlOwners() if the title has multiple owner-blocks.
 *		  Otherwise SortTtl().
 *
 *  History:
 *  21-02-1996	: eck      : creation
 *  03-12-2002	: heesakkm : Added SortTtl() and SortTtlOwners()
 *  28-02-2003	: paul     : Expect title to be equal to NULL
 *  08-04-2008  : paul     : renamed from PPTtlSort()
 *  14-08-2019	: smitp    : CBS-8853 pass "inputsort=no" to SortTtlOwners()
 *
 *  Notes	: An owner block starts with tag 101@
 *                Recognised options:
 *                front_iln	ILN of owner block to put in front
 *                front_copynr	nr. of copy record to put in front
 *
 *------------------------------------------------------------------------
 */
void
PPTtlSortCustom(
	PPTtl_p		title,		/* Title to sort */
	PPTagCmp_f	tagcmp,		/* Tag compare function */
	const char	*options)
{
    PPTag_p	tag = title;
    int		cnt = 0;
    char	optval[20];
    int16	front_iln = PIC_ILN_ANY;
    int16	front_copynr = PIC_COPYNUM_ANY;
    bool	linkage_found = FALSE;		// is there a tag with a $T?
    bool	*linkage = &linkage_found;
    bool	inputsort = TRUE;

    if (title == NULL) return;
    
    if (title->pp_elttyp != PPT_TITLE) {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPTtlSortCustom", "not sorting a title!");
    }
    if (title->PPTTL_firsttag == NULL) return;

    // Get options from FILEMAP, unless already done
    if (pp_env_opt_g == PPEO_UNINITIALISED) {
	pp_env_opt_initialise();
    }

    // Only keep track of presence of $T subfields if "linkage_sort=no"
    if (pp_env_opt_g & PPEO_LINKAGE_SORT) {
	linkage = NULL;
    }
    
    if (PM_opt_fetch(options, "front_iln", optval, sizeof(optval)) > 0) {
	front_iln = atoi(optval);
	if (PM_opt_fetch(options, "front_copynr", optval, sizeof(optval)) > 0) {
	    if (STREQ(optval, "none")) {
		front_copynr = PIC_COPYNUM_NONE;
	    } else {
		front_copynr = atoi(optval);
	    }
	}
    }
    if (PM_opt_fetch(options, "inputsort", optval, sizeof(optval)) > 0
	&& STREQ(optval, "no")) {
	inputsort = FALSE;
    }

    while (cnt < 2 && (tag = PPTagFind(tag, 1, PPLEV_LOCAL, '@', 0)) != NULL) {
	cnt++;
    }

    if (cnt < 1 || (cnt < 2 && front_iln == PIC_ILN_ANY)) {
	SortTtl(title, tagcmp, linkage, inputsort);
    } else {
	SortTtlOwners(title, tagcmp, front_iln, front_copynr, linkage, inputsort);
    }

    // Sort tags on linkage-occ in $T, if not yet done
    if (! (pp_env_opt_g & PPEO_LINKAGE_SORT)
	&& linkage_found) {
	pp_linkage_sort(title, tagcmp);
    }
    
} /* PPTtlSortCustom() */


/*------------------------------------------------------------------------
 *  Function	: PPTtlSort
 *  Purpose	: Wrapper around PPTtlSortCustom()
 *
 *  History:
 *  08-04-2008	: paul     : creation
 *
 *------------------------------------------------------------------------
 */
void
PPTtlSort(
	PPTtl_p		title)		/* Title to sort */
{
    PPTtlSortCustom(title, PPTagCmp, "");
} /* PPTtlSort() */
