/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File        : ppsubfield.c
 *  Description : PicaPlus-manipulation on subfields
 *  Author(s)   : Hans van Eck
 *  Creation    : 23-01-1996
 *
 *  History:
 *  23-01-1996  : eck      : creation
 *  22-02-1996  : eck      : Changed list manipulation, level support.
 *  11-06-1999  : BdB   : Add more const's to pointer arguments
 *  20-07-2000  : paul  : Added PP_ANYSBF
 *  23-11-2001	: paul  : using PPEltListAppend() in PPSbfAppend()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       :
 *
 *========================================================================
 */

#pragma ident "$Id: ppsubfield.c 292315 2012-10-16 15:44:57Z eckh $"

#include <stdlib.h>
#include <string.h>
#include <pica.h>
#include <picatypes.h>
#include <pp.h>
#include <er.h>
#include "pplocal.h"

/*------------------------------------------------------------------------
 *  Function    : PPSbfAppendContentsAt
 *  Purpose     : Append a piece of memory to a subfield contents
 *  Method      : realloc(), memcpy() and adjusting sbf.
 *
 *  History:
 *  30-01-1996  : eck      : creation
 *
 *  Notes       :
 *  Called from both PPSbfReplaceContents() and PPSbfAppendContents().
 *  In the former case, keep_len is 0.
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
static bool
PPSbfAppendContentsAt(
        PPSbf_p         sbf,            /* the subfield */
        int		keep_len,       /* length to keep */
        cvoidp          contents,       /* contents to append */
        int		cont_len)       /* contents-length (-1 if unknown) */
{
    voidp       new_contents;

    if (cont_len < 0) {
        cont_len = (int) strlen((char *)contents);
    }
    new_contents = realloc(sbf->PPSBF_contents, keep_len + cont_len + 1);
    if (new_contents != NULL) {
        (void) memcpy((char *)new_contents + keep_len, contents, cont_len);
        *((char *)new_contents + keep_len + cont_len) = '\0';
        sbf->PPSBF_contents = new_contents;
        sbf->PPSBF_length = keep_len + cont_len;
    } else {
        PPerrno = PPE_NOMEM;
    }
    return (new_contents != NULL);
} /* PPSbfAppendContentsAt() */

/*------------------------------------------------------------------------
 *  Function    : PPSbfReplaceContents
 *  Purpose     : Replace the contents of a subfield with a new one
 *  Method      : Call of PPSbfAppendContentsAt()
 *
 *  History:
 *  30-01-1996  : eck      : creation
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
bool
PPSbfReplaceContents(
        PPSbf_p         sbf,            /* subfield to replace contents of */
        cvoidp          contents,       /* new contents */
        int		length)         /* new length (-1 if unknown) */
{
    PPerrno = PPE_NOERROR;

    if (sbf->pp_elttyp != PPT_SUBFIELD) {
        PPerrno = PPE_TYPE;
        (void) ERAbort(PPerrno, "PPSbfReplaceContents", "not a subfield!");
    }
    /* 'Append' the new contents at position 0. */
    return PPSbfAppendContentsAt(sbf, 0, contents, length);
} /* PPSbfReplaceContents() */

/*------------------------------------------------------------------------
 *  Function    : PPSbfAppendContents
 *  Purpose     : Append to the contents of a subfield
 *  Method      : Call of PPSbfAppendContentsAt()
 *
 *  History:
 *  30-01-1996  : eck      : creation
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
bool
PPSbfAppendContents(
        PPSbf_p         sbf,            /* the subfield */
        cvoidp          contents,       /* the contents */
        int		length)         /* the length (-1 if unknown) */
{
    PPerrno = PPE_NOERROR;

    if (sbf->pp_elttyp != PPT_SUBFIELD) {
        PPerrno = PPE_TYPE;
        (void) ERWrite(PPerrno, "PPSbfAppendContents", "not a subfield!");
        return FALSE;
    }

    return PPSbfAppendContentsAt(sbf, sbf->PPSBF_length, contents, length);
} /* PPSbfAppendContents() */

/*------------------------------------------------------------------------
 *  Function    : PPSbfAppend
 *  Purpose     : Append a subfield to a tag.
 *  Method      : Append it to double linked list.
 *
 *  History:
 *  30-01-1996  : eck      : creation
 *  23-11-2001	: paul     : using PPEltListAppend()
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
PPSbf_p
PPSbfAppend(
        PPElement_p     elt,            /* element to append to */
        char            typ,            /* type of the subfield */
        cvoidp          contents,       /* contents */
        int		len)		/* length ofcontents */
{
    PPSbf_p     sbf;

    PPerrno = PPE_NOERROR;

    sbf = PPSbfCreate(typ, contents, len);
    if (sbf == NULL) return NULL;

    return PPEltListAppend(elt, sbf);
} /* PPSbfAppend() */

/*------------------------------------------------------------------------
 *  Function    : PPSbfInsert
 *  Purpose     : Insert subfield as first element of the tag
 *  Method      : Insert in linked list.
 *
 *  History:
 *  30-01-1996  : eck      : creation
 *
 *  Returns     : indication of success
 *------------------------------------------------------------------------
 */
PPSbf_p
PPSbfInsert(
        PPElement_p     elt,            /* element to insert subfield into */
        char            typ,            /* type of the subfield */
        cvoidp          contents,       /* contents */
        int		len)            /* length ofcontents */
{
    PPSbf_p     sbf;

    PPerrno = PPE_NOERROR;

    if ((elt->pp_elttyp != PPT_TAG)
        && (elt->pp_elttyp != PPT_SUBFIELD)) {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPSbfInsert", "not inserting in tag or subfield!");
    }
    sbf = PPSbfCreate(typ, contents, len);
    if (sbf == NULL) return NULL;

    if (elt->pp_elttyp == PPT_TAG) {
        PPEltInsertBetween(elt, sbf, elt->PPTAG_firstsbf);
    } else {
        PPEltInsertBetween(elt->pp_eltprev, sbf, elt);
    }

    return sbf;
} /* PPSbfInsert() */

/*------------------------------------------------------------------------
 *  Function    : PPSbfFind
 *  Purpose     : Find a subfield within a tag
 *  Method      : obvious
 *
 *  History:
 *  30-01-1996  : eck      : creation
 *  20-07-2000  : paul  : Added PP_ANYSBF
 *
 *  Returns     : pointer to subfield or NULL
 *------------------------------------------------------------------------
 */
PPSbf_p
PPSbfFind(
        PPElement_p     elt,            /* tag to scan */
        char            typ)            /* subfield-typ to scan for */
{
    PPSbf_p     sbf;

    PPerrno = PPE_NOERROR;

    if ((elt->pp_elttyp != PPT_TAG)
        && (elt->pp_elttyp != PPT_SUBFIELD)) {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPSbfFind", "not finding in tag or subfield!");
    }
    if (elt->pp_elttyp == PPT_TAG) {
        sbf = elt->PPTAG_firstsbf;
    } else {
        sbf = elt->pp_eltnext;
    }
    if (typ != PP_ANYSBF) {
        while ((sbf != NULL) && (sbf->PPSBF_typ != typ)) {
            sbf = sbf->pp_eltnext;
        }
    }
    return sbf;
} /* PPSbfFind() */

/*------------------------------------------------------------------------
 *  Function    : PPSbfGet
 *  Purpose     : Get the value of a subfield in title
 *  Method      : First find tag in title, then subfield in tag
 *
 *  History:
 *  01-02-1996  : eck      : creation
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes       : This routine is supplied for convenience.
 *
 *  Returns     : a pointer to the contents or NULL.
 *------------------------------------------------------------------------
 */
voidp
PPSbfGet(
        PPTtl_p         title,          /* Title to look in */
        PP_TagType_t    tagtyp,         /* tag-type to look for */
        PP_Level_t      lev,            /* level ... */
        char            ind,            /* indicator ... */
        PP_Occurrence_t occ,            /* occurrence ... */
        char            sbftyp)         /* subfield to look for */
{
    PPTag_p     tag;
    PPSbf_p     sbf;

    if (title->pp_elttyp != PPT_TITLE) {
        PPerrno = PPE_TYPE;
        ERAbort(PPerrno, "PPSbfGet", "not finding in a title!");
    }
    tag = PPTagFind(title, tagtyp, lev, ind, occ);
    if (tag != NULL) {
        sbf = PPSbfFind(tag, sbftyp);
        if (sbf != NULL) {
            return sbf->PPSBF_contents;
        }
    }
    return NULL;
} /* PPSbfGet() */
