/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: pptag.c
 *  Description	: PicaPlus-routines for tags
 *  Author(s)	: Hans van Eck
 *  Creation	: 23-01-1996
 *
 *  History:
 *  23-01-1996	: eck      : creation
 *  22-02-1996	: eck      : Change PPTagAdd() into PPTagAppend().
 *			   : Changed linked-list manipulation, level support.
 *  23-11-2001	: paul     : using PPEltListAppend() in PPTagAppend()
 *  21-01-2002	: paul     : Added PPTagHideUndo()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  20-05-2011	: paul     : CBS-2906 added PPTagFromSbf()
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: pptag.c 331081 2015-11-12 14:23:40Z heesakkm $"

#include <pica.h>
#include <picatypes.h>
#include <pp.h>
#include <er.h>
#include "pplocal.h"

/*------------------------------------------------------------------------
 *  Function	: PPTagAppend
 *  Purpose	: Append a tag to a title
 *  Method	: Compare tag from the start and append at the right place
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  23-11-2001	: paul     : using PPEltListAppend()
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes	: Appends the tag to the title.
 *
 *  Returns	: The tag added.
 *------------------------------------------------------------------------
 */
PPTag_p
PPTagAppend(
	PPElement_p	elt,		/* Element to add to */
	PP_TagType_t    typ,		/* type of tag to add */
	PP_Level_t	lev,		/* level of ... */
	char		ind,		/* indicator */
	PP_Occurrence_t	occ)		/* occurrence */
{
    PPTag_p	tag;

    PPerrno = PPE_NOERROR;

    tag = PPTagCreate(typ, lev, ind, occ);
    if (tag == NULL) return NULL;

    return PPEltListAppend(elt, tag);
} /* PPTagAppend() */

/*------------------------------------------------------------------------
 *  Function	: PPTagFind
 *  Purpose	: Look for a tag in a title or after a tag
 *  Method	: obvious
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  10-03-2004	: eck      : Allow elt to be NULL
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  08-10-2012	: blankeng : CBS-4023 - write error if PICTagType is 0
 *
 *  Notes	:
 *  The first argument of the routine can be a title as well as a tag.
 *  When it is a tag, it should be skipped.  This way, te routine can be
 *  called in a loop with the previous return-value as argument.
 *
 *  Returns	: pointer to the tag found or NULL
 *------------------------------------------------------------------------
 */
PPTag_p
PPTagFind(
	PPElement_p	elt,		/* title or tag */
	PP_TagType_t	typ,		/* type of tag to find */
	PP_Level_t	lev,		/* level of the tag */
	char		ind,		/* indicator of tag to find */
	PP_Occurrence_t	occ)		/* occurrence of tag to find */
{
    PPTag_p	tag = NULL;

    PPerrno = PPE_NOERROR;

    if (elt == NULL) return NULL;

    if (typ == 0) {
	PPerrno = PPE_TYPE;
	ERWrite(PPerrno, __func__, "CBS-4023 - PICTagType has a 0 value!!!");
    }

    if (elt->pp_elttyp == PPT_TITLE) {
	tag = elt->PPTTL_firsttag;
    } else if (elt->pp_elttyp == PPT_TAG) {
	/* Skip this one. */
	tag = elt->pp_eltnext;
    } else {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPTagFind",
			"element is neither title nor tag!");
    }
    while (tag != NULL) {
	/* Test visibility first */
	if (tag->PPTAG_typ > 0) {
	    if (((typ == PP_ANYTYP) || (typ == tag->PPTAG_typ))
		&& ((lev == PP_ANYLEV) || (lev == tag->PPTAG_lev))
		&& ((ind == PP_ANYIND) || (ind == tag->PPTAG_ind))
		&& ((occ == PP_ANYOCC) || (occ == tag->PPTAG_occ))) {
	        break;
	    }
	}
	tag = tag->pp_eltnext;
    }
    return tag;
} /* PPTagFind() */

/*------------------------------------------------------------------------
 *  Function	: PPTagHide
 *  Purpose	: Hide a tag for external output.
 *  Method	: Make the type negative.
 *
 *  History:
 *  30-01-1996	: eck      : creation
 *  22-02-1996	: eck      : return type now 'void'
 *------------------------------------------------------------------------
 */
void
PPTagHide(
	PPTag_p	tag)			/* Tag to hide */
{
    PPerrno = PPE_NOERROR;

    if (tag->pp_elttyp != PPT_TAG) {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPTagHide", "not hiding a tag!");
    }
    if (tag->PPTAG_typ > 0) tag->PPTAG_typ = -tag->PPTAG_typ;
} /* PPTagHide() */

/*------------------------------------------------------------------------
 *  Function	: PPTagHideUndo
 *  Purpose	: Hide a tag for external output.
 *  Method	: Make the type positive.
 *
 *  History:
 *  21-01-2002	: paul     : creation
 *------------------------------------------------------------------------
 */
void
PPTagHideUndo(
	PPTag_p	tag)			/* Tag to hide */
{
    PPerrno = PPE_NOERROR;

    if (tag->pp_elttyp != PPT_TAG) {
	PPerrno = PPE_TYPE;
	(void) ERAbort(PPerrno, "PPTagHideUndo", "parameter is not a tag!");
    }
    if (tag->PPTAG_typ < 0) tag->PPTAG_typ = -tag->PPTAG_typ;
} /* PPTagHideUndo() */

/*------------------------------------------------------------------------
 *  Function	: PPTagFromSbf
 *  Purpose	: Find the tag of a subfield
 *  Method	: 
 *
 *  History:
 *  20-05-2011	: paul     : creation
 *
 *  Notes	:
 *
 *  Returns	: The tag in which a subfield resides
 *                or NULL if no such tag exists
 *
 *------------------------------------------------------------------------
 */
PPTag_p
PPTagFromSbf(
	PPSbf_p	sbf)
{
    PPElement_p	parent;

    parent = PPEltParent(sbf);
    if (parent == NULL || parent->pp_elttyp != PPT_TAG) {
	return NULL;
    }

    return (PPTag_p)parent;
} /* PPTagFromSbf() */
