/*
 *  Copyright (c) 2000 by Pica
 *
 *  File	: pptitle.c
 *  Description	: Manipulate PP titles
 *  Author(s)	: Bas de Bakker
 *  Creation	: 22-03-2000
 *
 *  History:
 *  22-03-2000	: BdB	: creation
 *  04-12-2000  : RP    : Added PPReadCopy
 *  22-01-2001  : MT    : PPReadCopy: bug solved for elttyp PP_TITLE 
 *  22-01-2003	: eric+eck: PPReadCopy rewritten
 *  28-01-2003	: eric+eck: PPReadCopy bugfix
 *  08-11-2004  : michiel  : multi-byte occurrence
 *  20-05-2011	: paul     : CBS-2906 added PPTtlFromTag()
 *
 *  Notes	:
 *
 */

#pragma ident "$Id: pptitle.c 292315 2012-10-16 15:44:57Z eckh $"

#include <stdlib.h>
#include <pica.h>
#include <picatypes.h>
#include <pp.h>
#include <er.h>

#include "pplocal.h"

/*------------------------------------------------------------------------
 *  Function	: PPTtlConcatenate
 *  Purpose	: Concatenate two titles
 *  Method	: Link src title to dest title
 *  Author(s)	: Bas de Bakker
 *  Creation	: 22-03-2000
 *
 *  History:
 *  22-03-2000	: BdB	: creation
 *  28-07-2009	: eck      : Allow tags as first element.
 *
 *  Returns	: PPTtl_p to concatenated title
 *
 *------------------------------------------------------------------------
 */

/* After calling this routine, the old pointers 'dest' and 'src' should no
   longer be used, only the return value. */

PPTtl_p
PPTtlConcatenate(PPTtl_p dest, PPTtl_p src)
{
    PPTag_p src_firsttag;
    PPTag_p dest_lasttag;
    
    if (src == NULL) return dest;
    if (dest == NULL) return src;

    src_firsttag = src->PPTTL_firsttag;
    src->PPTTL_firsttag = NULL;
    PPTtlDestroy(src);
    if (src_firsttag == NULL) return dest;

    if (dest->pp_elttyp == PPT_TITLE) {
	dest_lasttag = dest->PPTTL_firsttag;
    } else {
	dest_lasttag = dest;
    }
    if (dest_lasttag == NULL) {
	dest->PPTTL_firsttag = src_firsttag;
	src_firsttag->pp_eltprev = dest;
	return dest;
    }
    
    while (dest_lasttag->pp_eltnext != NULL) {
	dest_lasttag = dest_lasttag->pp_eltnext;
    }
    dest_lasttag->pp_eltnext = src_firsttag;
    src_firsttag->pp_eltprev = dest_lasttag;
    return dest;
} /* PPTtlConcatenate() */

PPTag_p
PPTTLTagAddCopy( 
	PPTtl_p		PPTitle,	/* I/O Title to add to */
	PPTag_p		PPTag_src)	/* I   Tag to add	*/
{
    static const char *const fname = "PPTTLTagAddCopy";
    PPTag_p		next_tag;
    PPElement_p		prev_elt;
    PPTag_p		PPTag_new;
    PPSbf_p		sbf;
    
    PPerrno = PPE_NOERROR;

    if (PPTitle->pp_elttyp != PPT_TITLE) {
	PPerrno = PPE_TYPE;
	(void) ERWrite(PPerrno, fname, "adding to unsupported type!");
	return NULL;
    }
    PPTag_new = PPTagCreate(PPTag_src->PPTAG_typ, PPTag_src->PPTAG_lev, 
	    PPTag_src->PPTAG_ind, PPTag_src->PPTAG_occ);
    if (PPTag_new == NULL){
	PPerrno = PPE_NOMEM;
	(void) ERWrite(PPerrno, fname, "Unable to create tag"); 
	return NULL;
    }

    /* copy the subfields */
    sbf = PPTag_src->PPTAG_firstsbf;
    while (sbf != NULL) {
	if( PPSbfAppend(PPTag_new, sbf->PPSBF_typ, sbf->PPSBF_contents,
		sbf->PPSBF_length) == NULL){
	    PPEltDestroy(PPTag_new);
	    return NULL;
	}
	sbf = sbf->pp_eltnext;
    }

    /* Find last tag or first larger tag */
    prev_elt = PPTitle;
    next_tag = PPTitle->PPTTL_firsttag;
    while (next_tag != NULL &&
	    PPTagCmp(PPTag_src, next_tag) >= 0) {
		prev_elt = next_tag;
		next_tag = next_tag->pp_eltnext;
    }

    PPEltInsertBetween(prev_elt, PPTag_new, next_tag);

    return PPTag_new;
} /* PPTTLTagAddCopy() */


/*------------------------------------------------------------------------
 *  Function	: PPReadCopy
 *  Purpose	: Get a single copy (level 2) record + ILN from a title
 *  Method	: 
 *
 *  History:
 *  04-12-2000	: RP      : creation
 *  22-01-2003	: eric+eck: rewritten
 *  28-01-2003  : eric+eck: bugfix for 101@ as last tag
 *  28-08-2003	: heesakkm : Added PPerrno initialization.
 *  19-12-2003	: clemens  : Fixed ILN determination (101@-101@-20* sequence)
 *  08-11-2004  : michiel  : multi-byte occurrence
 *
 *  Notes	: Input "title" is a combination of all level 0/1/2 records
 *                belonging to a PPN/ILN.
 *                The first copy is retrieved by giving the title as a
 *                parameter. For the second and next copies, the parameter
 *                src should be given again as returned by the previous call.
 *
 *                Notice that both parameters are changed by the function! 
 *
 *  Returns	: The newly created tag or NULL in case of failure
 *------------------------------------------------------------------------
 */
PPTtl_p
PPReadCopy(PPTtl_p *src, int16 *iln)
{
    static const char *const fname = "PPReadCopy";
    PPTtl_p    		result;
    PPTag_p    		tag;
    PPTag_p    		ilntag;
    PPSbf_p    		ilnsbf;
    PP_Occurrence_t     occ;

    PPerrno = PPE_NOERROR;

    if (src == NULL || *src == NULL) {
	PPerrno = PPE_TYPE;
        (void) ERWrite(PPerrno, fname, "Can't process NULL pointer as title"); 
        return NULL;
    }
    tag = *src;
    *src = NULL;

    if (tag->pp_elttyp == PPT_TITLE) {
	ilntag = PPTagFind(tag, 1, PPLEV_LOCAL, '@', 0);
    } else if (tag->pp_elttyp == PPT_TAG && tag->PPTAG_lev == PPLEV_LOCAL) {
	ilntag = PPTagFind(tag->pp_eltprev, 1, PPLEV_LOCAL, '@', 0);
    } else if (tag->pp_elttyp == PPT_TAG && tag->PPTAG_lev == PPLEV_COPY) {
	for (ilntag = tag; ilntag != NULL && ilntag->pp_elttyp == PPT_TAG;
	     ilntag = ilntag->pp_eltprev) {
	    if ((ilntag->PPTAG_typ == 1)
		&& (ilntag->PPTAG_lev == PPLEV_LOCAL)
		&& (ilntag->PPTAG_ind == '@')) {
		break;        
	    }
	}
    } else {
        PPerrno = PPE_FORMAT;
        (void) ERWrite(PPerrno, fname, "Title or tag element (level 1 or 2) expected");
        return NULL;
    }
    /* Find and copy the next copy
     */
    if (tag->pp_elttyp == PPT_TITLE) tag = ilntag;
    while (tag != NULL && tag->pp_elttyp == PPT_TAG && 
	   tag->PPTAG_lev != PPLEV_COPY) {
	if ((tag->PPTAG_typ == 1)
	    && (tag->PPTAG_lev == PPLEV_LOCAL)
	    && (tag->PPTAG_ind == '@')) {
	    ilntag = tag;
	}
	tag = tag->pp_eltnext;
    }
    if (tag == NULL) {
	return NULL;
    }
    ilnsbf = NULL;
    if (ilntag != NULL && ilntag->pp_elttyp == PPT_TAG) {
	ilnsbf = PPSbfFind(ilntag, 'a');
    }
    if (ilnsbf == NULL) {
	PPerrno = PPE_FORMAT;
	(void) ERWrite(PPerrno, fname, "101@ $a not found");
	return NULL;
    }
    if (iln != NULL) *iln = atoi((char *)ilnsbf->PPSBF_contents);

    result = PPTtlCreate();
    if (result == NULL) {
        (void) ERWrite(PPerrno, fname, "Can't create new empty title"); 
        return NULL;
    }
    occ = tag->PPTAG_occ;
    while (tag != NULL && tag->pp_elttyp == PPT_TAG &&
	   tag->PPTAG_lev == PPLEV_COPY && tag->PPTAG_occ == occ) {
	if (PPTTLTagAddCopy(result, tag) == NULL) {
	    (void) ERWrite(PPerrno, fname, "Can't copy tag to new title"); 
	    PPEltDestroy(result);
	    return NULL;
	}
	tag = tag->pp_eltnext;
    }
    *src = tag;

    return result;
} /* PPReadCopy() */


/*------------------------------------------------------------------------
 *  Function	: PPTtlFromTag
 *  Purpose	: Find the title of a tag
 *  Method	: 
 *
 *  History:
 *  20-05-2011	: paul     : creation
 *
 *  Notes	:
 *
 *  Returns	: The title in which a tag resides
 *                or NULL if no such title exists
 *
 *------------------------------------------------------------------------
 */
PPTtl_p
PPTtlFromTag(
	PPTag_p	tag)
{
    PPElement_p	parent;

    parent = PPEltParent(tag);
    if (parent == NULL || parent->pp_elttyp != PPT_TITLE) {
	return NULL;
    }

    return (PPTtl_p)parent;
} /* PPTtlFromTag() */
