/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: wmbuffer.c
 *  Description	: The Buffer routines from the WM-library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 16-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: wmbuffer.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>

#include <pica.h>
#include <er.h>
#include <wm.h>
#include "wmlocal.h"

/*------------------------------------------------------------------------
 *  Function	: WMGetBufid
 *  Purpose	: 
 *  Method	: Copy from WMContext
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 16-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *  02-02-1994  : rob      : less fuss
 *
 *  Notes	:
 *  Description	:
 *
 *  bident is a pointer to a buffer identification.
 *
 *  The function will return the identification of the buffer. In erroneous
 *  situations FALSE will be returned and WMerrno should be examined.
 *
 *  Error values	:
 *
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok			|
 *  | WME_IDENT		| invalid identifier	|
 *
 *  Example		:
 *
 *      #include <wm.h>
 *
 *      if (!WMGetBufid("03")) {
 *          ERWrite(WMerrno, jobname, "getting bufferid");
 *      }
 *
 *  Returns	:
 *
 *  | Value	| Meaning		|
 *  | --------: | :-------------------- |
 *  | TRUE	| BufferID is defined	|
 *  | FALSE	| Failure		|
 *
 *------------------------------------------------------------------------
 */
bool
WMGetBufid(
	WMIdent_t bident)
{
    char *id;
    
    id = WMContext[WMCurrentC]->BufferID;
    bident[0] = id[0];
    bident[1] = id[1];
    bident[2] = 0;
    
    WMerrno = WME_NOERROR;
    if (*bident == '\0') { 
	WMerrno = WME_NOIDENT; 
	return FALSE;
    }
    return TRUE;
} /*WMGetBufid*/

/*------------------------------------------------------------------------
 *  Function	: WMPutBufid
 *  Purpose	: 
 *  Method	: Copy to WMContext
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 16-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *  02-02-1994  : rob      : less fuss
 *
 *  Description		:
 *
 *  bident is a pointer to a buffer identification.
 *
 *  The function will store the identification of the buffer.
 *  If there was already an identification, that one is replaced.
 *  If there was no buffer defined yet the buffer is defined (without linked fields).
 * 
 *  Error values	:
 * 
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok			|
 *
 *  Example		:
 *
 *      #include <wm.h>
 *
 *      (void) WMPutBufid("03");
 *
 *  Returns	:
 *
 *  | Value	| Meaning		|
 *  | --------: | :-------------------- |
 *  | TRUE	| Action succeeded	|
 *  | FALSE	| Failure		|
 *
 *------------------------------------------------------------------------
 */
bool 
WMPutBufid(
	const WMIdent_t bident)
{
    char * id;
    
    id = WMContext[WMCurrentC]->BufferID;
    id[0] = bident[0];
    id[1] = bident[1];
    id[2] = 0;
    
    WMerrno = WME_NOERROR;
    return TRUE;
} /* WMPutBufid */

/*------------------------------------------------------------------------
 *  Function	: WMDelBuf
 *  Purpose	: 
 *  Method	: Clear BufferID and all the fields
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 16-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *
 *  Notes	:
 *  Description	:
 *
 *  The function will delete the buffer with all its linked fields.
 *  In case there is no buffer, success is returned.
 *
 *  Error values:
 *
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok			|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      (void) WMPutBufid("03");	 Put buffer id 
 *      (void) WMPutField("TX", "TEKST");
 *
 *      (void) WMDelBuf();		 Delete buffer
 *
 *  Returns	: 
 *
 *  | Value	| Meaning		|
 *  | --------: | :-------------------- |
 *  | TRUE	| Action succeeded	|
 *
 *------------------------------------------------------------------------
 */
bool 
WMDelBuf(void)
{
    WMerrno = WME_NOERROR;
    IDBClose(WMContext[WMCurrentC]->Fields);
    WMContext[WMCurrentC]->Fields = IDBOpen();
    WMContext[WMCurrentC]->BufferID[0] = '\0';
    return TRUE;
} /* WMDelBuf */
