/*==========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: wmcontext.c
 *  Description	: 
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 13-12-1993
 *
 *  History:
 *  13-12-1993	: lolke    : creation
 *  25-04-1994	: eck      : cleaned up a bit && added WMDeleteContext().
 *  07-04-1995	: eck      : Improved WMCreateContext().
 *  21-07-1995	: eck      : Added WMClearCurrentContext().
 *			   : Fixed leak in WMDeleteContext().
 *  23-10-1995	: HSi	   : Added History for server.
 *  10-01-2002  : fb       : Synchronisation with WinIBW version (PvdM).
 *
 *  Notes	:
 *  Still not very clean: indexes and handles are still presumed equal.
 *
 *============================================================================
 */

#pragma ident "$Id: wmcontext.c 313247 2014-06-13 12:57:34Z heesakkm $"

#include <stdlib.h>
#include <pica.h>
#include <er.h>
#include <wm.h>
#include "wmlocal.h"


WMContext_t WMCurrentC = WM_NOCONTEXT;

ContextStruct_p WMContext[MAXCONTEXT];

static int
getfreecontext(void)
{
    int	i;

    for (i = 0; i < MAXCONTEXT; i++) {
	if (WMContext[i] == NULL) break;
    }
    if (i < MAXCONTEXT) return i;
    else return WM_NOCONTEXT;
} /* getfreecontext() */

static int
getcontextidx(WMContext_t handle)
{
    if ((handle < 0)
	|| (handle > MAXCONTEXT)
	|| (WMContext[handle] == NULL)) {
	return WM_NOCONTEXT;
    } else {
	return handle;
    }
} /* getcontextidx() */

/*------------------------------------------------------------------------------------------
 *  Function	: WMCreateContext
 *  Purpose	: Create a new worm context
 *  Method	: Uses malloc and IDBOpen
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 13-12-1993
 *
 *  History:
 *  13-12-1993	: lolke    : creation
 *  22-12-1993	: lolke    : Use of IDB-Library
 *  26-10-1994	: eck      : Do not switch to new context.
 *  07-04-1995	: eck      : Improved initialisation.
 *			   : I don't understand History, though.
 *  15-01-1997  : PvdM     : Added field register.
 *
 *  Notes	:
 *
 *  In case of an error WMCreateContext() implicitly writes to the errorlog 
 *  by means of ERWrite().
 *
 *  Description : 
 * 
 *  csflag indicates whether context is client or server based. 
 *  Its value is either CLIENT or SERVER.
 *
 *  The function will return a handle to a (new) context. In erroneous situations 
 *  WM_NOCONTEXT will be returned and WMerrno should be examined.
 *
 *  Error values	:
 *
 *  | Value		| Description
 *  | ----------------: | :----------------
 *  | WME_NOERROR	| ok
 *  | WME_OVERFLOW	| overflow
 *  | WME_MAXCONTEXT	| too many contexts
 *
 *  Example		:
 *
 *      #include <wm.h>
 *
 *      WMContext_t	appctx;
 *  
 *      appctx = WMCreateContext(SERVER);
 *      if (appctx == WM_NOCONTEXT) {
 *          ERWrite(WMerrno, jobname, "creating context");
 *          exit(EXIT_FAILURE);
 *      }
 *											.
 *  Returns	:
 *
 *  | Value		| Meaning
 *  | --------: 	| :------------
 *  | contexhandle	| Successfully completed
 *  | WM_NOCONTEXT	| Error if it fails
 *
 *-------------------------------------------------------------------------------------------------
 */
WMContext_t
WMCreateContext(
	WMCSflag_t csflag)			/* Client/Server-flag */
{
    WMContext_t	handle;

    WMerrno = WME_NOERROR;
    handle = getfreecontext();

    if (handle != WM_NOCONTEXT) {
	if ((WMContext[handle]
		= (ContextStruct_p) malloc(sizeof(ContextStruct_t))) != NULL) {
	    WMContext[handle]->Fields = IDBOpen();
	    WMContext[handle]->Messages = IDBOpen();
	    WMContext[handle]->Variables = IDBOpen();
	    WMContext[handle]->Commands = IDBOpen();
	    WMContext[handle]->HistoryServer = IDBOpen();
	    if ((WMContext[handle]->Fields != NULL)
		&& (WMContext[handle]->Variables != NULL)
		&& (WMContext[handle]->Messages != NULL)
		&& (WMContext[handle]->Commands != NULL)
		&& (WMContext[handle]->HistoryServer != NULL)) {
		    WMContext[handle]->csflag = csflag;
		    WMContext[handle]->BufferID[0] = '\0';
		    WMContext[handle]->ScreenID[0] = '\0';
		    WMContext[handle]->ScreenDA[0] = '\0';
		    WMContext[handle]->ScreenDest[0] = '\0';

		    /* How about History* ??? */
		    WMContext[handle]->HistoryP = 0;
		    WMContext[handle]->HistoryInit = TRUE;

		    WMContext[handle]->Screen = NULL;
		    WMContext[handle]->Template = NULL;
		    WMContext[handle]->VarRegister = NULL;
		    WMContext[handle]->VarRegisterSize = 0;
                    WMContext[handle]->FieldRegister = NULL;
                    WMContext[handle]->FieldRegisterSize = 0;
		    return handle;
	    }
	    IDBClose(WMContext[handle]->Fields);
	    IDBClose(WMContext[handle]->Messages);
	    IDBClose(WMContext[handle]->Variables);
	    IDBClose(WMContext[handle]->Commands);
	    IDBClose(WMContext[handle]->HistoryServer);
	    free(WMContext[handle]);
	}
	WMerrno = WME_OVERFLOW;
	(void) ERWrite(WMerrno, "WMCreateContext", "Allocation failed");
	return WM_NOCONTEXT;
    } else {
    	WMerrno = WME_MAXCONTEXT;
	(void) ERWrite(WMerrno, "WMCreateContext", "Too many handles");
	return WM_NOCONTEXT;
    }
} /* WMCreateContext() */

/*--------------------------------------------------------------------------------------------------
 *  Function	: WMSwitchContext
 *  Purpose	: Switch between worm contexts
 *  Method	: obvious
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 13-12-1993
 *
 *  History:
 *  13-12-1993	: lolke    : creation
 *
 *  Description		:
 *
 *  handle must be a proper handle to a previously created context.
 *
 *  The function will activate the context at handle and return the handle of the previous context. 
 *  In erroneous situations WM_NOCONTEXT will be returned and WMerrno should be examined.
 *
 *  Notes		:
 *
 *  In case of an error, WMSwitchContext() implicitly writes to the errorlog by means of ERWrite().
 *
 *  Error values	:
 *
 *  | Value                     | Description
 *  | ------------------------:	| :----------------
 *  | WME_NOERROR       	| ok
 *  | WME_ILLEGALCONTEXT    	| illegal context handle
 *
 *  Example		:
 *
 *      #include <wm.h>
 *
 *      WMContext_t	newcontext;
 *      .
 *      newcontext = WMCreateContext(SERVER);
 *      if (WMSwitchContext(newcontext) == WM_NOCONTEXT) {
 *	    ERWrite(WMerrno, jobname, "switching context");
 *	    exit(EXIT_FAILURE);
 *      }
 *                                                                                               .
 *
 *  Returns		:
 *	
 *  | Value		| Meaning
 *  | --------: 	| :------------
 *  | previous handle	| Successfully completed
 *  | WM_NOCONTEXT	| Error if illegal handle is passed
 *-------------------------------------------------------------------------------------------------------
 */
WMContext_t
WMSwitchContext
         (WMContext_t handle) 
{
    WMContext_t	OldContext;

    WMerrno = WME_NOERROR;
    OldContext = WMCurrentC;

    WMCurrentC = getcontextidx(handle);

    if (WMCurrentC != WM_NOCONTEXT) {
	return OldContext;
    } else if (handle == WM_NOCONTEXT) {
	return OldContext;
    } else {
	WMCurrentC = OldContext;
    	WMerrno = WME_ILLEGALCONTEXT;
	(void) ERWrite(WMerrno, "WMSwitchContext", "Handle not defined");
	return WM_NOCONTEXT;
    }
} /* WMSwitchContext() */

/*------------------------------------------------------------------------
 *  Function	: WMDeleteContext
 *  Purpose	: Delete a context
 *  Method	: Free all memory associated with the context.
 *
 *  History:
 *  21-07-1995	: eck      : creation (of comment)
 *  15-01-1997  : PvdM     : Added field register.
 *
 *  Notes	:
 *  Description		:
 *
 *  This function deletes the context at handle. In erroneous situations FALSE will be returned. 
 *  It indicates that the context handle was invalid.
 *
 *  Error values	:
 *
 *  | Value                     | Description
 *  | ------------------------:	| :----------------
 *  | WME_NOERROR       	| ok
 *  | WME_ILLEGALCONTEXT    	| illegal context handle
 *
 *------------------------------------------------------------------------
 */
void
WMDeleteContext(
	WMContext_t	handle) 
{
    int	idx;

    WMerrno = WME_NOERROR;

    idx = getcontextidx(handle);

    if (idx != WM_NOCONTEXT) {
	IDBClose(WMContext[idx]->Fields);
	IDBClose(WMContext[idx]->Messages);
	IDBClose(WMContext[idx]->Variables);
	IDBClose(WMContext[idx]->Commands);
	IDBClose(WMContext[idx]->HistoryServer);
	
	free(WMContext[idx]->Screen);
	free(WMContext[idx]->Template);
	free(WMContext[idx]->VarRegister);
        free(WMContext[idx]->FieldRegister);

	free(WMContext[idx]);
	WMContext[idx] = NULL;		/* Mark as unused */
	if (handle == WMCurrentC) WMCurrentC = WM_NOCONTEXT;
    } else {
	WMerrno = WME_ILLEGALCONTEXT;
	(void) ERWrite(WMerrno, "WMDeleteContext", "Handle not defined");
    }
} /* WMDeleteContext() */

/*------------------------------------------------------------------------
 *  Function	: WMClearCurrentContext
 *  Purpose	: Clear the current context (so worms are cleaner)
 *  Method	: obvious
 *
 *  History:
 *  21-07-1995	: eck      : creation
 *  15-01-1997  : PvdM     : Added field register.
 *
 *  Notes	: There should be better interfaces for the elements of the context
 *  Description	:
 *
 *  This function clears the current context without deleting it.
 *  The function is supposed to succeed in all circumstances.
 *
 *  Error values:
 *
 *  | Value                     | Description
 *  | ------------------------:	| :----------------
 *  | WME_NOERROR       	| ok
 * 
 *------------------------------------------------------------------------
 */
void
WMClearCurrentContext(void)
{
    if (WMCurrentC != WM_NOCONTEXT) {
	WMContext[WMCurrentC]->BufferID[0] = '\0';
	WMContext[WMCurrentC]->ScreenID[0] = '\0';
	WMContext[WMCurrentC]->ScreenDA[0] = '\0';
	WMContext[WMCurrentC]->ScreenDest[0] = '\0';

	IDBEmpty(WMContext[WMCurrentC]->Fields);
	IDBEmpty(WMContext[WMCurrentC]->Messages);
	IDBEmpty(WMContext[WMCurrentC]->Variables);
	IDBEmpty(WMContext[WMCurrentC]->Commands);

	/* How about History* ??? */

	free(WMContext[WMCurrentC]->Screen);
	WMContext[WMCurrentC]->Screen = NULL;
	free(WMContext[WMCurrentC]->Template);
	WMContext[WMCurrentC]->Template = NULL;

	free(WMContext[WMCurrentC]->VarRegister);
        free(WMContext[WMCurrentC]->FieldRegister);
	WMContext[WMCurrentC]->VarRegister = NULL;
	WMContext[WMCurrentC]->VarRegisterSize = 0;
        WMContext[WMCurrentC]->FieldRegister = NULL;
        WMContext[WMCurrentC]->FieldRegisterSize = 0;
    }
} /* WMClearCurrentContext() */
