/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: wmfield.c
 *  Description	: Field routines from WM-library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *  20-11-1997	: lolke    : WMAppToField added
 *  10-01-2002  : fb       : Synchronisation with WinIBW version (PvdM).
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: wmfield.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>
#include <stdlib.h>

#include <pica.h>
#include <wm.h>
#include <idb.h>
#include "wmlocal.h"

#define MUTFLG_ON       '\001'
#define MUTFLG_OFF      ' '


/*------------------------------------------------------------------------
 *  Function    : WMIsRegisteredField
 *  Purpose     :
 *  Method      :
 *  Author(s)   : P.J. van der Mespel
 *  Creation    : 15-01-1997
 *
 *  History:
 *  15-01-1997  : PvdM    : creation
 *
 *  Notes       :
 *
 *  Returns     : TRUE if registered field
 *                FALSE if no registered field
 *
 *------------------------------------------------------------------------
 */
bool
WMIsRegisteredField(char *id)
{
    if (WMContext[WMCurrentC]->FieldRegister == NULL) {
        return FALSE;
    }
    return NULL != strstr(WMContext[WMCurrentC]->FieldRegister, id);
} /* WMIsRegisteredField */

/*------------------------------------------------------------------------
 *  Function    : WMIsMutatedField
 *  Purpose     :
 *  Method      :
 *  Author(s)   : P.J. van der Mespel
 *  Creation    : 15-01-1997
 *
 *  History:
 *  15-01-1997  : PvdM    : creation
 *
 *  Notes       :
 *
 *  Returns     : TRUE if mutated field
 *                FALSE if field is not mutated
 *
 *------------------------------------------------------------------------
 */
bool
WMIsMutatedField(WMIdent_t id)
{
    char *cr;

    WMerrno = WME_NOERROR;
    cr = WMContext[WMCurrentC]->FieldRegister;

    if (cr == NULL) {
        return FALSE;
    }

    id[2] = '\0';
    cr = strstr(cr, id);
    if (cr == NULL) {
        return FALSE;
    }
    return cr[2] == MUTFLG_ON;
} /* WMIsMutatedField */

/*------------------------------------------------------------------------
 *  Function    : SetMutFlag
 *  Purpose     : Sets mutation flag of field
 *  Method      :
 *  Author(s)   : P.J. van der Mespel
 *  Creation    : 15-01-1997
 *
 *  History:
 *  15-01-1997  : PvdM    : creation
 *
 *  Notes       :
 *
 *  Returns     : void
 *
 *------------------------------------------------------------------------
 */
static void
SetMutFlag(const WMIdent_t id)
{
    char *c;

    if (WMContext[WMCurrentC]->FieldRegister == NULL) {
        return;
    }
    c = strstr(WMContext[WMCurrentC]->FieldRegister,id);

    if (c != NULL) {
        c[2] = MUTFLG_ON;
    }
} /* SetMutFlag() */

/*------------------------------------------------------------------------
 *  Function    : Register
 *  Purpose     : Registers field and mutation flag in context
 *  Method      :
 *  Author(s)   : P.J. van der Mespel
 *  Creation    : 15-01-1997
 *
 *  History:
 *  15-01-1997  : PvdM    : creation
 *
 *  Notes       :
 *
 *  Returns     : void
 *
 *------------------------------------------------------------------------
 */
static void
Register(char *id)
{
    char *c;
    char *cr;
    int  l;

    cr = WMContext[WMCurrentC]->FieldRegister;

    if (cr != NULL) {
        c = strstr(cr,id);
        l = strlen(cr);
    } else {
        c = NULL;
        l = 0;
    }

    if (c == NULL) {
        if ((l + 4) > WMContext[WMCurrentC]->FieldRegisterSize) {
            if((cr = realloc(cr, l + 50)) != NULL){
                WMContext[WMCurrentC]->FieldRegisterSize = l + 50;
            }
        }

        if (cr != NULL) {
            /* ignore out of memory for the moment */
            /*WMContext[WMCurrentC]->FieldRegisterSize = l + 50;*/
            cr[l]   = id[0];
            cr[l+1] = id[1];
            cr[l+2] = MUTFLG_OFF;
            cr[l+3] = '\0';
            WMContext[WMCurrentC]->FieldRegister = cr;
        }
    }
} /* Register() */

/*------------------------------------------------------------------------
 *  Function	: WMGetField
 *  Purpose	:
 *  Method	: Uses IBDGet
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *
 *  Notes	:
 *  Description	:
 *
 *  fident is a (returned) pointer to a field identification.
 *  Wild card '?' may be used for any character in the identification
 *  ('??' will retrieve the first field in the buffer).
 *
 *  fbuf is a pointer to a field contents.
 *
 *  fsize is the size of the field contents.
 *
 *  The function will locate the first matching field from the buffer and copy
 *  the field contents to fbuf and the identification to fident. In erroneous situations
 *  FALSE will be returned and WMerrno should be examined.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR	| ok		|
 *  | WME_OVERFLOW	| overflow	|
 *  | WME_NOTFOUND	| not found	|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char		cntnts[10];
 *
 *      if (!WMPutBufid("03")) {
 *          ERWrite(WMerrno, jobname, "putting buffer");
 *      }
 *      if (!WMPutField("TX", "TEKST")) {
 *          ERWrite(WMerrno, jobname, "putting field");
 *      }
 *
 *      if (!WMGetField("TX", cntnts, sizeof(cntnts))) {
 *          ERWrite(WMerrno, jobname, "getting field");
 *      }
 *
 *  Returns	:
 *
 *  | Value	| Meaning	|
 *  | --------: | :------------ |
 *  | TRUE	| IDBGet	|
 *  | FALSE	| WMerrno	|
 *
 *------------------------------------------------------------------------
 */
bool
WMGetField(const WMIdent_t fident, char *fbuf, size_t fsize)
{
    WMerrno = WME_NOERROR;
    if (IDBGet(WMContext[WMCurrentC]->Fields, fident, fbuf, fsize)) {
	return TRUE;
    } else {
	WMerrno = WME_OVERFLOW;
	if (IDBerrno == IDBE_NOTFOUND) { WMerrno = WME_NOTFOUND; }
	return FALSE;
    }
} /* WMGetField */

/*------------------------------------------------------------------------
 *  Function	: WMGetFieldN
 *  Purpose	: Get field N
 *  Method	: Uses IDBGetN
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *
 *  Notes	:
 * 
 *  I guess with this routine it is possible to loop through all the fields of
 *  a worm. The parameter n is the driver. The description and example below are
 *  literairy copied from the WMGetField routine
 *
 *  Description	:
 *
 *  fident is a (returned) pointer to a field identification.
 *  Wild card '?' may be used for any character in the identification
 *  ('??' will retrieve the first field in the buffer).
 *
 *  fbuf is a pointer to a field contents.
 *
 *  fsize is the size of the field contents.
 *
 *  The function will locate the first matching field from the buffer and copy
 *  the field contents to fbuf and the identification to fident. In erroneous situations
 *  FALSE will be returned and WMerrno should be examined.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR	| ok		|
 *  | WME_OVERFLOW	| overflow	|
 *  | WME_NOTFOUND	| not found	|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char		cntnts[10];
 *
 *      if (!WMPutBufid("03")) {
 *          ERWrite(WMerrno, jobname, "putting buffer");
 *      }
 *      if (!WMPutField("TX", "TEKST")) {
 *          ERWrite(WMerrno, jobname, "putting field");
 *      }
 *
 *      if (!WMGetField("TX", cntnts, sizeof(cntnts))) {
 *          ERWrite(WMerrno, jobname, "getting field");
 *      }
 *
 *  Returns	:
 *
 *  | Value	| Meaning	|
 *  | --------: | :------------ |
 *  | TRUE	| IDBGetN	|
 *  | FALSE	| WMerrno	|
 *
 *
 *------------------------------------------------------------------------
 */
bool
WMGetFieldN(int n, WMIdent_t fident, char *fbuf, size_t fsize)
{
    WMerrno = WME_NOERROR;
    if (IDBGetN (WMContext[WMCurrentC]->Fields, n, fident, fbuf, fsize)) {
	return TRUE;
    } else {
	WMerrno = WME_NOTFOUND;
	if (fident[0] == 0) { WMerrno = WME_OVERFLOW; }
	return FALSE;
    }
} /* WMGetFieldN */

/*------------------------------------------------------------------------
 *  Function	: WMPutField
 *  Purpose	:
 *  Method	: Uses IDBPutAppend
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *  13-01-2001  : jh       : Added SetMutFlag
 *
 *  Notes	:
 *  Description	:
 *
 *  fident is a pointer to a field identification.
 *  Wild cards are not allowed.
 *
 *  fbuf is a pointer to a, zero terminated field contents.
 *
 *  The function will try to locate the field in the buffer.
 *  If found, the field contents fbuf is concatenated to the already
 *  defined field preceded by the carriage return character ; 
 *  if not the field is appended to the buffer. In erroneous
 *  situations FALSE will be returned and WMerrno should be examined.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR       | ok		|
 *  | WME_OVERFLOW      | overflow	|
 *
 *  Example	:
 *
 *  See the example at WMGetField().
 *
 *  Returns	:
 *  
 *  | Value	| Meaning	|
 *  | --------: | :------------ |
 *  | TRUE	| IDBPut	|
 *  | FALSE	| WMerrno	|
 *
 *------------------------------------------------------------------------
 */
bool
WMPutField(const WMIdent_t fident, const char *fbuf)
{
    WMerrno = WME_NOERROR;
    if (IDBPut(WMContext[WMCurrentC]->Fields, fident, fbuf)) {
	SetMutFlag(fident);
	return TRUE;
    } else {
	WMerrno = WME_OVERFLOW;
	return FALSE;
    }
} /* WMPutField */

/*------------------------------------------------------------------------
 *  Function	: WMAppField
 *  Purpose	: Append field to internal structure
 *  Method	: Call of IDBPutA()
 *
 *  History:
 *  22-07-1994	: eck      : creation
 *
 *  Notes	:
 *
 *  Returns	: IDBPutA
 *------------------------------------------------------------------------
 */
bool
WMAppField(const WMIdent_t fident, const char *fbuf)
{
    WMerrno = WME_NOERROR;
    if (IDBPutA(WMContext[WMCurrentC]->Fields, fident, fbuf)) {
	return TRUE;
    } else {
	WMerrno = WME_OVERFLOW;
	return FALSE;
    }
} /* WMAppField */

/*------------------------------------------------------------------------
 *  Function	: WMAppToField
 *  Purpose	: Append string to contents of field
 *  Method	: WMGetField and WMPutField
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 19-11-1997
 *
 *  History:
 *  19-11-1997	: lolke    : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
bool
WMAppToField(WMIdent_t fident, const char *fbuf)
{
    char temp[WM_MAXWRMLEN];

    WMerrno = WME_NOERROR;
    (void) WMGetField(fident, temp, sizeof(temp));
    if (WMerrno == WME_NOERROR || WMerrno == WME_NOTFOUND) {
	if (WMerrno == WME_NOTFOUND) {
	    temp[0] = '\0';
	}
	if (strlen(temp) + strlen(fbuf) >= WM_MAXWRMLEN) {
	    WMerrno = WME_OVERFLOW;
	    return FALSE;
	}
	(void) strcat(temp, fbuf);
	return WMPutField(fident, temp);
    }
    return FALSE;
} /* WMAppToField */

/*------------------------------------------------------------------------
 *  Function	: WMDelField
 *  Purpose	:
 *  Method	: Uses IDBDel
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *  13-01-2001  : jh       : Added call to SetMutFlag.
 *
 *  Notes	:
 *  Description	:
 *
 *  fident is a pointer to a field identification.
 *  Wild cards are allowed (see: WMGetField).
 *
 *  The function will delete all fields whose identification match fident.
 *  In case no matching fields were found, success is returned.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR       | ok		|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      (void) WMDelField("TX");
 *
 *  Returns	:
 *  
 *  | Value	| Meaning	|
 *  | --------: | :------------ |
 *  | TRUE	| IDBDel	|
 *
 *------------------------------------------------------------------------
 */
bool
WMDelField(WMIdent_t fident)
{
    WMerrno = WME_NOERROR;

    SetMutFlag(fident);
    (void) IDBDel(WMContext[WMCurrentC]->Fields, fident);
    return TRUE;
} /* WMDelField */

/*------------------------------------------------------------------------
 *  Function    : WMRegField
 *  Purpose     : register field for mutation logging
 *  Author(s)   : P.J. van der Mespel
 *  Creation    : 13-01-1997
 *
 *  History:
 *  13-01-1997  : PvdM creation
 *
 *  Notes       : the names of the fields are put in a string.
 *                2 pos id, 1 pos flag
 *                MUTFLG_ON  = mutated
 *                MUTFLG_OFF = not mutated
 *
 *  Returns     : FALSE if operation fails
 *
 *------------------------------------------------------------------------
 */
bool
WMRegField(WMIdent_t fident)
{
    WMerrno = WME_NOERROR;
    Register(fident);
    return TRUE;
} /* WMRegField */

/*------------------------------------------------------------------------
 *  Function    : WMClearFieldReg
 *  Purpose     : Clears the entire field register in context
 *  Method      :
 *  Author(s)   : P.J. van der Mespel
 *  Creation    : 15-01-1997
 *
 *  History:
 *  15-01-1997  : PvdM    : creation
 *
 *  Notes       :
 *
 *  Returns     : void
 *
 *------------------------------------------------------------------------
 */
void WMClearFieldReg(void)
{
    if (WMContext[WMCurrentC]->FieldRegister)
	{
	    WMContext[WMCurrentC]->FieldRegister[0] = '\0';
	}
}
