/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: wmmessage.c
 *  Description	: Message routines from the WM-library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 16-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: wmmessage.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>

#include <pica.h>
#include <er.h>
#include <wm.h>
#include "wmlocal.h"

/*------------------------------------------------------------------------
 *  Function	: WMGetMsg WMPutMsg WMDelMsg
 *  Purpose	: Get, Put and Del messages
 *  Method	: Uses IDB
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 16-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *  22-12-1993	: lolke    : Use of IDB-Library
 *  12-01-1994  : rob      : made it work
 *
 *  Notes	: 
 *
 *  with get one can step till returns false
 *
 *  WMGetMsg
 *  --------
 *
 *  Description	:
 *
 *  seqnum is the sequence number of the requested message.
 *
 *  mident is a (returned) pointer to a message identification.
 *  Wild cards are allowed (see: WMGetField()).
 *
 *  mbuf is a pointer to a message contents.
 *
 *  msize is the size of the message contents.
 *
 *  The function will copy the first matching message contents to mbuf and the message
 *  identification to mident. In erroneous situations FALSE will be returned and
 *  WMerrno should be examined.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char      msg[80];
 *
 *      if (!WMGetMsg(0, "04", msg, sizeof(msg))) {
 *          ERWrite(WMerrno, jobname, "getting message");
 *      }
 *
 *  WMPutMsg
 *  --------
 *
 *  Description	:
 *
 *  mident is a pointer to a message identification.
 *  Wild cards are not allowed.
 *
 *  mbuf is a pointer to a (zero-terminated) message contents.
 *
 *  The function will store the message contents with the proper identification. In case
 *  there was already a message part defined the contents of the message to store is
 *  concatenated to the contents in the library, preceded by the carriage return character.
 *  In erroneous situations FALSE will be returned.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      if (!WMPutMsg("05", msgtxt)) {
 *          ERWrite(WMerrno, jobname, "putting message");
 *      }
 *							.
 *  Returns	: 
 *
 *  | Value	| Description			|
 *  | --------: | :---------------------------- |
 *  | TRUE      | ok                            |
 *  | FALSE     | out of memory and is serious  |
 *
 *  WMDelMsg
 *  --------
 *
 *  Description	:
 *
 *  The function will delete all messages in the library. In case no messages were found,
 *  success is returned. The value of WMerrno is unaffected.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *      #include <wm.h>
 *	
 *      (void) WMDelMsg();
 *
 *------------------------------------------------------------------------
 */
bool
WMGetMsg(int n, WMIdent_t mident, char *mbuf, size_t msize)
{
    WMerrno = WME_NOERROR;
    return IDBGetN(WMContext[WMCurrentC]->Messages, n, mident, mbuf, msize);
} /* WMGetMsg */


bool
WMPutMsg(const WMIdent_t mident, const char *mbuf)
{
    WMerrno = WME_NOERROR;
    return IDBPutA (WMContext[WMCurrentC]->Messages,mident, mbuf);
} /* WMPutMsg */


bool
WMDelMsg(void)
{
    WMerrno = WME_NOERROR;
    IDBEmpty (WMContext[WMCurrentC]->Messages);
    return TRUE;
} /* WMDelMsg */
