/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: wmscreen.c
 *  Description	: Screen routines from the WM-library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *  20-04-1994	: eck      : Added WMGetForm() (get screen definition).
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: wmscreen.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>
#include <stdlib.h>
#include <pica.h>
#include <er.h>
#include <wm.h>
#include "wmlocal.h"

/*------------------------------------------------------------------------
 *  Function	: WMGet/Put/DelFrmid
 *                WMGet/Put/DelForm
 *
 *  History:
 *  17-12-1993	: lolke    : creation
 *
 *  Notes	:
 *
 *  WMGetFrmid
 *  ----------
 *
 *  Description	:
 *
 *  fident is a (returned) pointer to a screenform identification.
 *
 *  Beware that a screenform part has no contents, only an identification and that there
 *  is never more than one screenform.
 *  The function will copy the identification to fident. In erroneous situations FALSE
 *  will be returned. In this case, a form-id has never been put.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR	| ok            |
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char     fid[3];
 *
 *      (void) WMGetFrmid(fid);
 *
 *  WMPutFrmid
 *  ----------
 *
 *  Description	:
 *
 *  fident is a pointer to a screenform identification.
 *  Wild cards are not allowed.
 *
 *  The function will store the identification fident. In case there was already a
 *  screenform defined, that one is replaced. The function will always return TRUE.
 *  WMerrno will be set to WME_NOERROR.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR	| ok            |
 *
 *  Example	:
 *
 *      #include <wm.h>
 * 
 *      (void) WMPutFrmid("04");
 *
 *  WMDelFrmid
 *  ----------
 *
 *  Description	:
 *
 *  The function will delete the screenform. In case there was no  screenform defined yet,
 *  success is returned. WMerrno will always be set to WME_NOERROR.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR	| ok            |
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      (void) WMDelFrmid();
 *
 *  WmgetForm
 *  ---------
 *
 *  Description	:
 *
 *  sform is pointer to a buffer, used to store the screenform.
 *
 *  vsize is the size of sform.
 *
 *  Beware that there is never more than one screenform.
 *  The function will copy the screenform to sform. In erroneous situations FALSE will
 *  be returned and WMerrno should be examined.
 *
 *  Error values:
 *
 *  | Value		| Description	        |
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok                    |
 *  | WME_OVERFLOW      | overflow              |
 *  | WME_NOTFOUND      | identifier not found	|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char     buffer[1500];
 *
 *      if (!WMGetForm(buffer, sizeof(buffer))) {
 *          ERWrite(WMerrno, jobname, "getting form");
 *      }
 *
 *  WMPutForm
 *  ---------
 *
 *  Description	:
 *
 *  sform is a pointer to a screenform.
 *  Wild cards are not allowed.
 *
 *  The function will store the screenform fident. In case there was already a
 *  screenform defined, that one is replaced. The function return no errors.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char     buffer[1500];
 *
 *      (void) WMPutForm(buffer);
 *
 *  WMDelForm
 *  ---------
 *
 *  Description	:
 *
 *  The function will delete the screenform. No errors are returned.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      (void) WMDelForm();
 *
 *
 *------------------------------------------------------------------------
 */
bool
WMGetFrmid(WMIdent_t fident)
{
    char *p;    

    WMerrno = WME_NOERROR;
    p = (char *)WMContext[WMCurrentC]->ScreenDA;
    fident[0] = p[0];
    fident[1] = p[1];
    fident[2] = 0;

    return p[0] && p[1];
} /*WMGetFrmid */

bool
WMPutFrmid(const WMIdent_t fident)
{
    char *p;    

    WMerrno = WME_NOERROR;
    p = (char *)WMContext[WMCurrentC]->ScreenDA;
    p[0] = fident[0];
    p[1] = fident[1];
    p[2] = 0;
    return TRUE;
} /*WMPutFrmid */

bool
WMDelFrmid(void)
{
    WMerrno = WME_NOERROR;
    WMContext[WMCurrentC]->ScreenDA[0] = '\0';
    return TRUE;
} /*WMDelFrmid */

bool
WMGetForm(char *form, size_t vsize)
{
    size_t	len;
    
    WMerrno = WME_NOERROR;

    if (WMContext[WMCurrentC]->Screen == NULL) {
	WMerrno = WME_NOTFOUND;
    } else {
	len = strlen(WMContext[WMCurrentC]->Screen);

	if (vsize > len) {
	    (void) strcpy(form, WMContext[WMCurrentC]->Screen);
	} else {
	    *form = '\0';
	    WMerrno = WME_OVERFLOW;
	}
    }
    return (WMerrno == WME_NOERROR);
} /* WMGetForm() */

void
WMPutForm(const char *form)
{
    int l;
    
    WMerrno = WME_NOERROR;
    /* copieer het formulier */
    l = strlen(form);
    free(WMContext[WMCurrentC]->Screen);
    WMContext[WMCurrentC]->Screen = (char *) malloc ( l+1 );
    (void) strcpy(WMContext[WMCurrentC]->Screen, form);

    /* remove a possible etx */
    if (l>0){
	if (form[l-1]==3){
	    WMContext[WMCurrentC]->Screen[l-1] = 0;
      	}
    }
} /* WMPutForm */

void
WMDelForm (void)
{
    WMerrno = WME_NOERROR;
    free (WMContext[WMCurrentC]->Screen);
    WMContext[WMCurrentC]->Screen = NULL;    
} /* WMDelForm */
