/*========================================================================
 *  Copyright (c) 1993 by Pica
 *
 *  File	: wmvariable.c
 *  Description	: Variable routines from the WM-library
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *  15-02-2002	: paul     : Renamed SetMutFlag to VarSetMutFlag (for TANDEM)
 *                         : Renamed Register to RegisterVar (for TANDEM)
 *
 *  Notes	: RegVar is not yet implemented
 *                the implementation is not obvious!
 *
 *========================================================================
 */

#pragma ident "$Id: wmvariable.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>
#include <stdlib.h>
#include <pica.h>
#include <er.h>
#include <wm.h>
#include "wmlocal.h"

#define MUTFLG_ON	'\001'
#define MUTFLG_OFF	' '

bool
WMIsRegistered (const WMIdent_t id)
{
    if (WMContext[WMCurrentC]->VarRegister == NULL) { return FALSE; }
    return NULL != strstr(WMContext[WMCurrentC]->VarRegister,id);
} /* WMRegistered */

bool
WMIsMutatedVar(
	WMIdent_t	id)
{
    char *cr;

    WMerrno = WME_NOERROR;
    cr = WMContext[WMCurrentC]->VarRegister;

    if (cr == NULL) {
	return FALSE;
    }

    id[2] = '\0';
    cr = strstr(cr, id);
    if (cr == NULL) {
	return FALSE;
    }
    return cr[2] == MUTFLG_ON;
} /* WMIsMutatedVar() */

static void
VarSetMutFlag (const WMIdent_t id)
{
    char *c;

    if (WMContext[WMCurrentC]->VarRegister == NULL) { return; }
    c = strstr(WMContext[WMCurrentC]->VarRegister,id);

    if (c!=NULL){ c[2] = MUTFLG_ON; }
} /* VarSetMutFlag */

static void
RegisterVar (const WMIdent_t id)
{
    char *c;
    int  l;
    char *cr;

    cr = WMContext[WMCurrentC]->VarRegister;

    if (cr!=NULL) {
        c = strstr(cr,id);
	l = strlen(cr);
    } else {
        c = NULL;
        l = 0;
    }

    if (c==NULL){
	if (l+4>WMContext[WMCurrentC]->VarRegisterSize){
	    cr = realloc(cr,WMContext[WMCurrentC]->VarRegisterSize + 50);
	    if (cr != NULL) {
		WMContext[WMCurrentC]->VarRegisterSize += 50;
		WMContext[WMCurrentC]->VarRegister = cr;
	    }
	}

	/* ignore out of memory for the moment */
	if (cr != NULL) {
	    cr[l]   = id[0];
	    cr[l+1] = id[1];
	    cr[l+2] = MUTFLG_OFF;
	    cr[l+3] = '\0';
	}
    }
}


/*------------------------------------------------------------------------
 *  Function	: WMGetVar, WMDelvar, WMPutVar
 *  Purpose	: Get, Del or Put the contents of variable
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  16-12-1993	: lolke    : creation
 *  22-12-1993	: lolke    : make use of IDB-Library
 *
 *  Notes	: Put fails only when out of memory!
 *
 *  WMGetVar
 *  --------
 *  Description	:
 *
 *  vident is a pointer to a variable identification.
 *  Wild cards are not allowed.
 *
 *  vbuf is a pointer to a variable contents.
 *
 *  vsize is the size of the variable contents.
 *
 *  The function will copy the variable contents to vbuf.In erroneous situations
 *  FALSE will be returned. WMerrno is always set to WME_NOERROR.
 *
 *  Error values:
 *
 *  | Value		| Description	|
 *  | ----------------: | :------------ |
 *  | WME_NOERROR	| ok		|
 *
 *  Example	:
 *
 *       #include <wm.h>
 *       char           lines[10];
 *
 *       (void) WMGetVar("!L", lines, sizeof(lines));
 *
 *  Returns	: FALSE if operation fails
 *
 *  WMGetVarN
 *  ---------
 *
 *  Description	:
 *
 *  seqnum is the sequence number of the variable to be retrieved.
 *
 *  vident is a pointer to a variable identification.
 *  Wild cards are not allowed.
 *
 *  vbuf is a pointer to a variable contents.
 *
 *  vsize is the size of the variable contents.
 *
 *  The function will copy the appropriate variable contents to vbuf. In erroneous
 *  situations FALSE will be returned and WMerrno should be examined.
 *
 *  Error values:
 *
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok			|
 *  | WME_NOTFOUND      | identifier not found	|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      int       i;
 *      char      tmp[255];
 *
 *      for (i = 0; WMGetVarN(i, id, tmp, sizeof(tmp)); i++) { 
 *          ERWrite(Noerrno, "getting all variables", 
 *               "id: %s contents %s", id, tmp);
 *      }
 *
 *  WMPutVar
 *  --------
 *
 *  Description	:
 *
 *  vident is a pointer to a variable identification.
 *  Wild cards are not allowed.
 *
 *  vbuf is a pointer to a (zero-terminated) variable contents.
 *
 *  The function will store the variable contents with the proper identification.
 *  In case there was already a variable with the same identification defined,
 *  that one is replaced. In erroneous situations FALSE will be returned.
 *  WMerrno is always set to WME_NOERROR.
 *
 *  Notes	:
 *
 *  A variable must always be registered (using WMRegVar()) and then set to an initial
 *  value (using WMPutVar()) before it is really used.
 *
 *  Error values:
 *
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok			|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      if (!WMRegVar("TA")) {
 *          ERWrite(WMerrno, jobname, "registering variable");
 *      }
 *
 *      if (!WMPutVar("TA", "NE")) {
 *          ERWrite(WMerrno, jobname, "putting variable");
 *      }
 *
 *  WMDelVar
 *  --------
 *
 *  Description	:
 *
 *  vident is a pointer to a variable identification.
 *  Wild cards are allowed (see: WMGetField()).
 *
 *  The function will delete all variables in the library matching vident. In case
 *  no such variables were found, success is returned. In erroneous situations
 *  FALSE will be returned. WMerrno is always set to WME_NOERROR.
 *
 *  Error values:
 *
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | WME_NOERROR	| ok			|
 *
 *  Example	:
 *
 *      #include <wm.h>
 *
 *      if (!WMDelVar("SE")) {
 *          ERWrite(WMerrno, jobname, "deleting variable");
 *      }
 *
 *------------------------------------------------------------------------
 */
bool
WMGetVar(const WMIdent_t mident, char *mbuf, size_t msize)
{
    WMerrno = WME_NOERROR;
    return IDBGet(WMContext[WMCurrentC]->Variables, mident, mbuf, msize);
} /* WMGetVar */

bool
WMGetVarN(int n, WMIdent_t mident, char *mbuf, size_t msize)
{
    WMerrno = WME_NOERROR;
    if (IDBGetN (WMContext[WMCurrentC]->Variables, n, mident, mbuf, msize)) {
	return TRUE;
    } else {
	WMerrno = WME_NOTFOUND;
	return FALSE;
    }
} /* WMGetVar */


bool
WMPutVar(const WMIdent_t mident, const char *mbuf)
{
    WMerrno = WME_NOERROR;
    VarSetMutFlag(mident);
    return IDBPut(WMContext[WMCurrentC]->Variables, mident, mbuf);
} /* WMPutVar */

bool
WMDelVar(const WMIdent_t mident)
{
    WMerrno = WME_NOERROR;
    VarSetMutFlag(mident);
    return IDBDel(WMContext[WMCurrentC]->Variables, mident);
} /* WMDelVar */

/*------------------------------------------------------------------------
 *  Function	: WMRegVar, WMMutatedVar
 *  Purpose	: register variables for mutation logging
 *  Author(s)	: Lolke Boonstra
 *  Creation	: 17-12-1993
 *
 *  History:
 *  10-02-1994  : rk creation
 *
 *  Notes	: the names of the variables are put in a string.
 *	          2 pos id, 1 pos flag
 *                MUTFLG_ON  = mutated
 *                MUTFLG_OFF = not mutated
 *
 *  WMRegVar
 *  --------
 *
 *  Description	:
 *
 *  vident is a pointer to a variable identification.
 *  Wild cards are allowed (see: WMGetField()).
 *
 *  The function will register all variables in the library, matching vident,
 *  as a context variable. The default type for variables is global. The function
 *  always return TRUE. WMerrno is not affected.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *  See the example at WMPutVar().
 *
 *  | Value		| Description		|
 *  | ----------------: | :-------------------- |
 *  | TRUE              | ok			|
 *
 *  WMMutatedVar
 *  ------------
 *
 *  Description	:
 *
 *  vident is a (returned) pointer to a variable identification.
 *  Wild cards are allowed see: WMGetField().
 *
 *  The function will untag all variables in the library, matching vident, as
 *  being updated. The default type for variables is global. In case no such
 *  variables were found, FALSE is returned. WMerrno is unaffected.
 *
 *  Error values:
 *
 *  (none)
 *
 *  Example	:
 *
 *      #include <wm.h>
 *      char          id[3];
 *
 *      while (WMMutatedVar(id));  Untag all variables
 *
 *------------------------------------------------------------------------
 */
bool
WMRegVar(const WMIdent_t vident)
{
    WMerrno = WME_NOERROR;
    RegisterVar(vident);
    return TRUE;
} /* WMRegVar */

bool
WMMutatedVar(WMIdent_t id)
{
    char *cr;

    WMerrno = WME_NOERROR;
    id[0] = '\0';
    cr = WMContext[WMCurrentC]->VarRegister;
    if (cr ==NULL){ return FALSE; }

    cr = strchr(cr, MUTFLG_ON);
    if (cr == NULL){ return FALSE; }
    id[0] = cr[-2];
    id[1] = cr[-1];
    id[2] = '\0';
    cr[0] = MUTFLG_OFF;

    return TRUE;
} /* WMMutatedVar */
