/*
 *  Copyright (c) 2001 by Pica
 *
 *  File	: wmwormcs.c
 *  Description	: Implementation of WMBuild using the CS library (WMBuildCS).
 *  Author(s)	: Jan van der Steen
 *  Creation	: 22-08-2001
 *
 *  History:
 *  22-08-2001	: steenj : creation
 *  10-01-2002  : fb     : Synchronisation of BufBuildCS with WinIBW version.
 *  15-02-2002  : paul   : Renamed AddStrToWorm to AddStrToWormCS (for TANDEM)
 *                       : Renamed EOOS to EOOS_CS (for TANDEM)
 *                       : Renamed EOVS to EOVS_CS (for TANDEM)
 *                       : Renamed ETXS to ETXS_CS (for TANDEM)
 *
 *  Notes	: This file implements dynamic versions of functions
 *		  which can also be found elsewhere. Instead of fixed
 *		  buffers, counted strings (CS library) are used to
 *		  store values. Most notably, WMBuildCS can build
 *		  arbitrary large worms.
 *		  This work makes part of the project to get rid of
 *		  the internal limits concerning the FEP protocol
 *		  between server and client.
 *		  The WMGetVarCS and friends function equivalents (such
 *		  as WMGetVar) are in other files. But the *CS variants
 *		  are all kept together in this file.
 */

#pragma ident "$Id: wmwormcs.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include <pica.h>
#include <er.h>
#include <cb.h>
#include <cs.h>
#include <idb.h>
#include <wm.h>
#include "wmlocal.h"

/*------------------------------------------------------------------------
 *  Function	: WMerrnoFromIDBerrno
 *  Purpose	: Propagate errors in the IDB layer to the WM layer.
 *  Returns	: WMerrno is set based on IDBerrno.
 *		  Unknown IDBerrno values are mapped to WME_NOERROR.
 *
 *  Notes	: Unfortunately, some of the WMGet-functions do not set
 *		  WMerrno properly so we have to rely on IDBerrno and do
 *		  the propagation ourselves.
 *		  These WMGet-functions should be fixed but this is tedious
 *		  since they are called from various places.
 *  History	:
 *  13-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
static void
WMerrnoFromIDBerrno(void)
{
    /*
     * Unknown IDBerrno values are mapped to WME_NOERROR.
     */
    if (IDBerrno == IDBE_NOTFOUND) WMerrno = WME_NOTFOUND; else
    if (IDBerrno == IDBE_NOMEM   ) WMerrno = WME_OVERFLOW; else
    if (IDBerrno == IDBE_NOERROR ) WMerrno = WME_NOERROR ; else
                                   WMerrno = WME_NOERROR;
} /* WMerrnoFromIDBerrno() */


/*-----------------------------------------------------------------------
 *  Function    : AddStrToWormCS
 *  Purpose     : Add string to worm
 *  Method	: Trivial.
 *  24-08-2001	: steenj : CS implementation.
 *  Notes       : Propagate possible CSerrno values as WME_OVERFLOW
 *-----------------------------------------------------------------------
 */
static bool
AddStrToWormCS(
	CSBuf_p 	Worm,
	const char	*str)
{
    if (CSAppend(Worm, str) == FALSE) {
	WMerrno = WME_OVERFLOW;
	return FALSE;
    }
    return TRUE;
} /* AddStrToWormCS */


/*------------------------------------------------------------------------
 *  Function	: BufBuildCS
 *  Purpose	: Put buffer id and fields in worm
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *  10-01-2002  : fb     : Synchronisation with WinIBW version.
 *------------------------------------------------------------------------
 */
static bool
BufBuildCS(CSBuf_p Worm, const char *type)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;
    WMIdent_t	id;
    int		n;

    if (WMGetBufid(id)) {
	ok = ok && AddStrToWormCS(Worm, type);
	ok = ok && AddStrToWormCS(Worm, id  );
	for (n = 0; ok && WMGetFieldNCS(n, id, storage); n++) {
	    ok = ok && AddStrToWormCS(Worm, "V");
	    ok = ok && AddStrToWormCS(Worm, id);
	    if (type[0] == 'B') {
		if (WMIsMutatedField(id)) {
		    ok = ok && AddStrToWormCS(Worm, "M");
		} else {
		    ok = ok && AddStrToWormCS(Worm, "0");
		}
	    }
	    ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	    ok = ok && AddStrToWormCS(Worm, EOV);
	}
	ok = ok && (WMerrno != WME_OVERFLOW);
	ok = ok && AddStrToWormCS(Worm, EOO);
    }
    CSDestroy(storage);

    return ok;
} /* BufBuildCS */


/*------------------------------------------------------------------------
 *  Function	: ComBuildCS
 *  Purpose	: Put command in worm
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
ComBuildCS(CSBuf_p Worm)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;
    WMIdent_t	id;

    id[0] = id[1] = '?';
    if (WMGetCmdCS(id, storage)) {
	ok = ok && AddStrToWormCS(Worm, "C");
	ok = ok && AddStrToWormCS(Worm, id);
	ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	ok = ok && AddStrToWormCS(Worm, EOO);
    }
    ok = ok && (WMerrno != WME_OVERFLOW);
    CSDestroy(storage);

    return ok;
} /* ComBuildCS */


/*------------------------------------------------------------------------
 *  Function	: DefBuildCS
 *  Purpose	: Put screen definition in worm
 *  Method	: Read definition and copy to worm
 *  Notes       : Since definitions might already be terminated by an EOO
 *                we check this and only append an EOO if necessary.
 *		  Nearly identical to TemplateBuildCS.
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
DefBuildCS(CSBuf_p Worm)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;

    if (WMGetFormCS(storage)) {
	ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	WMDelForm();
	if (WMObjectComplete(CSContents(storage), CSLength(storage)) == FALSE) {
	    ok = ok && AddStrToWormCS(Worm, EOO);
	}
    }
    ok = ok && (WMerrno != WME_OVERFLOW);
    CSDestroy(storage);

    return ok;
} /* DefBuildCS */


/*------------------------------------------------------------------------
 *  Function	: TemplateBuildCS
 *  Purpose	: Put template in worm
 *  Method	: Read template and copy to worm
 *  Notes       : Since templates might already be terminated by an EOO
 *                we check this and only append an EOO if necessary.
 *		  Nearly identical to DefBuildCS.
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
TemplateBuildCS(CSBuf_p Worm)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;

    if (WMGetTemplateCS(storage)) {
	ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	WMDelTemplate();
	if (WMObjectComplete(CSContents(storage), CSLength(storage)) == FALSE) {
	    ok = ok && AddStrToWormCS(Worm, EOO);
	}
    }
    ok = ok && (WMerrno != WME_OVERFLOW);
    CSDestroy(storage);

    return ok;
} /* TemplateBuildCS */


/*------------------------------------------------------------------------
 *  Function	: FrmBuildCS
 *  Purpose	: Put "Sxx" object in worm
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
FrmBuildCS(CSBuf_p Worm)
{
    WMIdent_t	id;
    static char	hlp[] = "SXX" EOO;

    if (WMGetFrmid(id) != TRUE) { return FALSE; }
    hlp[1] = id[0];
    hlp[2] = id[1];
    return AddStrToWormCS(Worm, hlp);
} /* FrmBuildCS */


/*------------------------------------------------------------------------
 *  Function	: HisBuildCS
 *  Purpose	: Put history records in worm
 *  Notes	: Copy from MesBuildCS.
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
HisBuildCS(CSBuf_p Worm)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;
    WMIdent_t	id;
    int		n;

    for (n = 0; ok && WMGetHisNCS(n, id, storage); n++) {
	ok = ok && AddStrToWormCS(Worm, "G");
	ok = ok && AddStrToWormCS(Worm, id);
	ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	ok = ok && AddStrToWormCS(Worm, EOO);
    }
    ok = ok && (WMerrno != WME_OVERFLOW);
    (void) WMDelHis();
    CSDestroy(storage);

    return ok;
} /* HisBuildCS */


/*------------------------------------------------------------------------
 *  Function	: MesBuildCS
 *  Purpose	: Put messages in the worm
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
MesBuildCS(CSBuf_p Worm)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;
    WMIdent_t	id;
    int		n;

    for (n = 0; ok && WMGetMsgNCS(n, id, storage); n++) {
	ok = ok && AddStrToWormCS(Worm, "M");
	ok = ok && AddStrToWormCS(Worm, id);
	ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	ok = ok && AddStrToWormCS(Worm, EOO);
    }
    ok = ok && (WMerrno != WME_OVERFLOW);
    (void) WMDelMsg();
    CSDestroy(storage);

    return ok;
} /* MesBuildCS */


/*------------------------------------------------------------------------
 *  Function	: VarBuildCS
 *  Purpose	: Put variables in worm
 *  Notes	: All variables are put in the worm.
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *------------------------------------------------------------------------
 */
static bool
VarBuildCS(CSBuf_p Worm)
{
    CS_DYNAMIC(storage);
    bool	ok = TRUE;
    WMIdent_t	id;
    int		n;

    for (n = 0; WMGetVarNCS(n, id, storage); n++) {
	if (WMIsMutatedVar(id)) {
	    if (WMContext[WMCurrentC]->csflag == WM_CLIENT) {
		ok = ok && AddStrToWormCS(Worm, "V");
	    } else {
		ok = ok && AddStrToWormCS(Worm, "K");
	    }
	    ok = ok && AddStrToWormCS(Worm, id);
	    ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	    ok = ok && AddStrToWormCS(Worm, EOO);
	} else
	if (WMContext[WMCurrentC]->csflag == WM_SERVER ||
	    WMContext[WMCurrentC]->csflag == WM_BOTH) {
	    if (WMIsRegistered(id)) {
		ok = ok && AddStrToWormCS(Worm, "K");
	    } else {
		ok = ok && AddStrToWormCS(Worm, "V");
	    }
	    ok = ok && AddStrToWormCS(Worm, id);
	    ok = ok && AddStrToWormCS(Worm, CSContents(storage));
	    ok = ok && AddStrToWormCS(Worm, EOO);
	}
    }
    ok = ok && (WMerrno != WME_OVERFLOW);
    CSDestroy(storage);

    return ok;
} /* VarBuildCS */


/*
 *------------------------------------------------------------------------
 *  Function	: WMBuildCS
 *  Purpose	: Put the local datastructure WMContext into the worm
 *
 *  History	:
 *  24-08-2001	: steenj : CS implementation.
 *  13-11-2001	: eck      : Re-ordered Worm (LBS gave problems on login
 *  			   : because /V variable was at the beginning)
 *
 *  Returns	: FALSE if something fails
 *------------------------------------------------------------------------
 */
bool
WMBuildCS(CSBuf_p Worm)
{
    bool ok = TRUE;

    CSClear(Worm);

    if (WMCurrentC == WM_NOCONTEXT) return FALSE;

    if (WMContext[WMCurrentC]->csflag == WM_SERVER ||
	WMContext[WMCurrentC]->csflag == WM_BOTH) {
	ok = ok &&      HisBuildCS(Worm);
	ok = ok &&      MesBuildCS(Worm);
	ok = ok &&      DefBuildCS(Worm);
	ok = ok && TemplateBuildCS(Worm);
	if (FrmBuildCS(Worm)) {
	    ok = ok && BufBuildCS(Worm, "W");
	}
	ok = ok &&      VarBuildCS(Worm);
    } else if (WMContext[WMCurrentC]->csflag == WM_CLIENT) {
	ok = ok && VarBuildCS(Worm     );
	ok = ok && ComBuildCS(Worm     );
	ok = ok && BufBuildCS(Worm, "B");
    } else {
	return FALSE;
    }
    ok = ok && AddStrToWormCS(Worm, ETX);

    return ok;
} /* WMBuildCS */


/*------------------------------------------------------------------------
 *  Function	: increase_buffer
 *  Purpose	: Increase supplied buffer in size.
 *  Method	: realloc
 *
 *  History:
 *  19-09-2001	: steenj   : creation
 *
 *  Notes	: This function avoids memory leaks in case the realloc fails.
 *
 *  Returns	: In case of failure we return FALSE and as side-effects
 *		  WMerrno is set to WME_OVERFLOW and the partly filled
 *		  buffer is free-ed and zero-ed.
 *
 *------------------------------------------------------------------------
 */
static bool
increase_buffer(
	char	**buffer,
	size_t	*capacity)
{
    size_t	bytes = *capacity * 2;
    char	*tmp  = realloc(*buffer, bytes);

    if (tmp == NULL) {
	WMerrno = WME_OVERFLOW;
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    } else {
	*buffer   = tmp;
	*capacity = bytes;
    }
    return (tmp != NULL);
}


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetCmdCS
 *  Purpose	: Get cmd without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetCmdCS(
	WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetCmd(id, message, capacity);

	WMerrnoFromIDBerrno();

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetCmdCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetFieldCS
 *  Purpose	: Get field without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetFieldCS(
	const WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetField(id, message, capacity);

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetFieldCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetFieldNCS
 *  Purpose	: Get N-th field without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  Notes	: We use the IDB layer since the return codes of
 *		  WMGetFieldN are unreliable (WME_OVERFLOW and
 *		  WME_NOTFOUND are exchanged!).
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetFieldNCS(
	int		n,
	WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = IDBGetN(WMContext[WMCurrentC]->Fields, n, id,
			       message, capacity);

	WMerrnoFromIDBerrno();

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetFieldNCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetFormCS
 *  Purpose	: Get screen without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetFormCS(
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetForm(message, capacity);

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }
    return ok;
} /* WMGetFormCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetTemplateCS
 *  Purpose	: Get template without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetTemplateCS(
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetTemplate(message, capacity);

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetTemplateCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetHisCS
 *  Purpose	: Get history item without any restrictions on it's size.
 *  Method	: Use WMGetHis.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  05-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetHisCS(
	WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetHis(id, message, capacity);

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetHisCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetHisNCS
 *  Purpose	: Get N-th history item without any restrictions on it's size.
 *  Method	: Use IDBGetN since WMGetHis seems to be used in client-mode.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  05-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetHisNCS(
	int		n,
	WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = IDBGetN( WMContext[WMCurrentC]->HistoryServer
			      , n
			      , id
			      , message
			      , capacity
			      );

	WMerrnoFromIDBerrno();

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetHisNCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetMsgNCS
 *  Purpose	: Get N-th message without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetMsgNCS(
	int		n,
	WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetMsg(n, id, message, capacity);

	WMerrnoFromIDBerrno();

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetMsgNCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetVarCS
 *  Purpose	: Get variable without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  05-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetVarCS(
	const WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetVar(id, message, capacity);

	WMerrnoFromIDBerrno();

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetVarCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetVarNCS
 *  Purpose	: Get N-th variable without any restrictions on it's size.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetVarNCS(
	int		n,
	WMIdent_t	id,
	CSBuf_p storage)
{
    size_t     capacity = 512;
    char *     message  = malloc(capacity);
    bool       ok       = TRUE;

    CSClear(storage);
    while (message) {
	bool retcode = WMGetVarN(n, id, message, capacity);

	WMerrnoFromIDBerrno();

	if (retcode == FALSE) {
	    if (WMerrno == WME_OVERFLOW) {
		/*
		 * Increase the buffer and try again
		 */
		if (increase_buffer(&message, &capacity)) continue;
	    }
	    ok = FALSE;
	}
	break;
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    if (message) {
	ok = ok && AddStrToWormCS(storage, message);
	free(message);
    }

    return ok;
} /* WMGetVarNCS() */


/*
 *------------------------------------------------------------------------
 *  Function	: WMGetMsgByIDCS
 *  Purpose	: Collect all messages with supplied id.
 *  Returns	: Value stored in supplied CS.
 *		  FALSE if something fails, TRUE otherwise.
 *  Notes	: The WM library has WMGetMsg which actually behaves
 *		  like WMGetMsgN. The CS implementation has no
 *		  WMGetMsgCS, but WMGetMsgByIDCS and WMGetMsgNCS.
 *  History	:
 *  04-09-2001	: steenj : Creation
 *------------------------------------------------------------------------
 */
bool
WMGetMsgByIDCS(
	WMIdent_t	wanted_id,
	CSBuf_p storage)
{
    CS_DYNAMIC(msg);
    WMIdent_t	id;
    bool	ok = TRUE;
    int		n;

    CSClear(storage);
    for (n = 0; WMGetMsgNCS(n, id, msg); n++) {
	if (STREQ(wanted_id, id)) {
	    ok = ok && AddStrToWormCS(storage, CSContents(msg));
	}
    }
    ok = ok && (WMerrno != WME_OVERFLOW);

    CSDestroy(msg);

    return ok;
} /* WMGetMsgByIDCS() */
