/*========================================================================
 *  Copyright (c) 2004 by OCLC PICA
 *
 *  File	: ouf_cbs.c
 *  Description	: 
 *  Author(s)	: Michael Lyons
 *  Creation	: 30-12-2004
 *
 *  History:
 *  30-12-2004	: michael  : creation
 *  23-02-2007	: clemens  : fixed problems with/cleaned-up array sizes
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ouf_cbs.c 376252 2020-04-22 11:25:54Z zwaalm $"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <env.h>
#include <cs.h>
#include <er.h>
#include <p3.h>
#include <nmt.h>
#include <hsht.h>
#include <picatypes.h>
#include <picamisc.h>
#include "ouf_pf.h"
#include "ouf_trace.h"
#include "ouf.h"

/* Initialize CBS data */

/*------------------------------------------------------------------------
 *  Function	: CBS_init 
 *  Purpose	: read CBS -specific config options common to both ouf_fetch
 *                and ouf_update
 *  Method	: PFRead() and PM_strncpy()
 *
 *  History:
 *  12-01-2004	: michael  : creation
 *  21-02-2005	: michael  : reformatted into "house style"
 *  21-02-2005	: michael  : changed usage of strncpy() to PM_strncpy()
 *
 *  Notes	:
 *
 *  Returns	: TRUE, unless UTF8 is specified for a non-CBS4 system
 *------------------------------------------------------------------------
 */
bool
CBS_init(
	FILE *pfp,
	CBS_t *cbsp)
{
    char field[OUF_FIELD_L + 1];

    cbsp->bConnected = FALSE;
    cbsp->bCBS4inuse = TRUE;

    (void) PFRead(pfp, "[CBS]", "Host", field);
    (void) PM_strncpy(cbsp->host, field, OUF_HOST_L+1);

    (void) PFRead(pfp, "[CBS]", "Port", field);
    (void) PM_strncpy(cbsp->port, field, OUF_PORT_L+1);

    (void) PFRead(pfp, "[CBS]", "User", field);
    (void) PM_strncpy(cbsp->user, field, OUF_USER_L+1);

    (void) PFRead(pfp, "[CBS]", "Password", field);
    (void) PM_strncpy(cbsp->password, field, OUF_PASSWORD_L+1);

    (void) PFRead(pfp, "[CBS]", "Database", field);
    (void) PM_strncpy(cbsp->database, field, OUF_DATABASE_L+1);
    
    if (strchr(cbsp->database, '@') == NULL) {
	(void) strcat(cbsp->database, OUF_DUMMY_DOMAIN);
    }

    (void) PFRead(pfp, "[CBS]", "Encrypt Password", field);
    cbsp->bEncryptPwd = (((int) strcmp(field, "YES")) == 0) ? TRUE : FALSE;

    (void) PFRead(pfp, "[CBS]", "CBS4 in use", field);
    cbsp->bCBS4inuse = (strcmp(field, "YES") == 0) ? TRUE : FALSE;

    (void) PFRead(pfp, "[CBS]", "UTF-8 in use", field);
    cbsp->bUseUTF8 = (strcmp(field, "YES") == 0) ? TRUE : FALSE;
    if (!cbsp->bCBS4inuse && cbsp->bUseUTF8) {
	/* OUFSetError(OUF_UTF8_NOT_SUPPORTED); */
	return FALSE;
    }
    return TRUE;
} /* CBS_init() */

/*------------------------------------------------------------------------
 *  Function	: close_connection
 *  Purpose	: Closes connection with CBS
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 07-08-1996
 *
 *  History:
 *  07-08-1996	: peter jan: creation
 *  21-02-2005  : michael : restyled
 *
 *  Notes	: 
 *
 *  Returns	:  nothing, it's void
 *------------------------------------------------------------------------
 */
void
close_connection(
	CBS_t *cbsp)
{
    /* Close connection with server */
    if (cbsp->bConnected) {
	(void) tr_date();
        (void) tr_write("Closing connection...\n");

	/* Ignore result of function, because we might receive
	   FALSE even if the connection was successfully closed.
	   This will happen when a previous P3 function caused
	   P3errno to be set to a value other than P3_NOERROR */
        if (!cbsp->bCBS4inuse) {
	    (void) P3Close(cbsp->P3Handle);
	} else {
	    (void) P3XLWMClose(cbsp->P3Handle);
	}
	(void) tr_date();
        (void) tr_write("Connection closed.\n");
    }
}  /* close_connection() */

/*------------------------------------------------------------------------
 *  Function	: select_database
 *  Purpose	: Selects a CBS database
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 07-08-1996
 *
 *  History:
 *  07-08-1996	: peter jan: creation
 *
 *  Returns	: TRUE if worked, FALSE if it didn't
 *------------------------------------------------------------------------
 */
bool
select_database(
	CBS_t *cbsp)
{
    char options[OUF_OPTIONS_L + 1];
    bool ret;
    
    (void) tr_date();
    (void) tr_write ("Selecting database '%s'...\n", cbsp->database);

    /* First, check that database is non-empty */
    if (cbsp->database[0] == '\0') return TRUE;

    /* Construct options string for database select */
    options[0] = '\0';

    /* Select a database */
    if (!cbsp->bCBS4inuse) {
	ret = P3Select(cbsp->P3Handle, cbsp->database, options);
    } else {
	ret = P3XLWMSelect(cbsp->P3Handle, cbsp->database, options);
    }
    if (!ret) {
	/* OUFSetError(OUF_P3ERROR); */
	return FALSE;
    }
    (void) tr_date();
    (void) tr_write("Database selected.\n");

    return TRUE;
} /* select_database() */

/*------------------------------------------------------------------------
 *  Function	: setup_connection
 *  Purpose	: Sets up a connection with CBS
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 07-08-1996
 *
 *  History:
 *  07-08-1996	: peter jan: creation
 *  02-07-2004	: huug	   : characterset negotiation
 *  21-04-2020	: zwaalm   : CBS-9355 Fixed memory leak.
 *  22-04-2020	: zwaalm   : CBS-9355 Code review changes.
 *
 *  Returns	: TRUE if worked, FALSE if it didn't
 *------------------------------------------------------------------------
 */
bool
setup_connection(
	CBS_t *cbsp)
{
    P3Handle_p P3Handle;
    CS_DYNAMIC(options);
    bool       ok;
    CBBuf_t    cbuf[OUF_CBUF_L];

    CBClear(cbuf, sizeof(cbuf));
    /* Construct options string for connect based on gFSettings */
    (void) CSPrintf(options, "protocol=tcp/ip;address=%s;port=%s",
	    cbsp->host, cbsp->port);

    (void) tr_date();
    (void) tr_write("Connecting to '%s' (port: %s)...\n",cbsp->host,cbsp->port);

    /* Connect to server */
    if (!cbsp->bCBS4inuse) {
	P3Handle = P3Open(CSContents(options));
    } else {
	P3Handle = P3XLWMOpen(CSContents(options));
    }
    ok = (P3Handle != NULL);

    if (ok) {
	(void) tr_date();
	(void) tr_write("Connected.\n");
	
	cbsp->P3Handle   = P3Handle;
	cbsp->bConnected = TRUE;

	/* Construct options string for log-on based on gFSettings */
	(void) CSPrintf(options, "user=%s;password=%s",
			cbsp->user, cbsp->password);
	if (cbsp->bEncryptPwd) CSAppend(options, ";client-type=04OUFT00");
	if (cbsp->bUseUTF8) CSAppend(options, ";charset=UTF-8");

	/* log-on to server */
	(void) tr_date();
	(void) tr_write("Logging on...\n");
	if (!cbsp->bCBS4inuse) {
	    ok = P3Logon(P3Handle, CSContents(options));
	} else {
	    ok = P3XLWMLogon(P3Handle, CSContents(options));
	}
	if (ok) {
	    (void) tr_date();
	    (void) tr_write("Logged on.\n");
	}
    }
    CSDestroy(options);
    return ok;
} /* setup_connection() */

/*------------------------------------------------------------------------
 *  Function	: retrieve_title
 *  Purpose	: Retrieves title from CBS
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 07-08-1996
 *
 *  History:
 *  07-08-1996	: peter jan: creation
 *  31-10-1997	: HSi	: Added expansion. Use P3FetchTitle.
 *  10-11-1997	: HSi	: Back to P3EditTitle for bugfix.
 *  21-02-2005	: michael : restyled
 *
 *  Returns	: OUF_OK, or an error
 *------------------------------------------------------------------------
 */
int
retrieve_title(
	CBBuf_t *CBbuf,
	size_t CBSize,
	const char *PPN,
	const char *copy,
	const char *syntax,
	const char *expansion,
	const char *ILN,
	CBS_t *cbsp)
{
    int  nResult;
    char options[OUF_OPTIONS_L + 1];
    bool ok;

    if (copy != NULL && *copy == '\0' ) copy = NULL;
    
    if (syntax != NULL && *syntax == '\0' ) syntax = NULL;
    
    if (expansion != NULL && *expansion == '\0' ) expansion = NULL;
    
    if (ILN != NULL && *ILN == '\0' ) ILN = NULL;
    
    
    OUF_DEBUG (5, ("Retrieving title from CBS (PPN = %s, copy = %s, "
       "syntax = %s, expansion = %s, ILN = %s)\n", 
       PPN,
       (copy      == NULL) ? "<none>" : copy,
       (syntax    == NULL) ? "<none>" : syntax,
       (expansion == NULL) ? "<none>" : expansion,
       (ILN	     == NULL) ? "<none>" : ILN
       ));

    /* Clear counted buffer */
    if (!CBClear(CBbuf, CBSize)) {
	/* OUFSetError(OUF_BUFFER_NOT_CLEARED); */
	return OUF_ERROR;
    }

    /* Construct options string for fetch title command */
    (void) sprintf ( options, "%s%s%s%s%s%s%s%s",
	 (copy      == NULL) ? "" : "copy=",
	 (copy      == NULL) ? "" : copy,
	 (syntax    == NULL) ? "" : ";syntax=",
	 (syntax    == NULL) ? "" : syntax,
	 (expansion == NULL) ? "" : ";expansion=",
	 (expansion == NULL) ? "" : expansion,
	 (ILN       == NULL) ? "" : ";iln=",
	 (ILN       == NULL) ? "" : ILN
	 );

    if (!cbsp->bCBS4inuse) {
        ok = P3FetchTitle(cbsp->P3Handle, options, PPN, CBbuf, CBSize);
    } else {
	ok = P3XLWMFetchTitle(cbsp->P3Handle, options, PPN, CBbuf, 
	                      CBSize);
    }
    if (!ok) {
        OUF_DEBUG(5, ("%s: Error while retrieving title from CBS.\n",
        (char *) P3errno));

	if (STREQ(P3errno, P3E_AUTHORITY)) {
	    nResult = OUF_AUTHORITY;
	} else if (STREQ(P3errno, P3E_DISCONNECT)) {
	    nResult = OUF_DISCONNECT;
	} else if (STREQ(P3errno, P3E_NOSUPPORT)) {
	    nResult = OUF_NO_SUPPORT;
	} else if (STREQ(P3errno, P3E_PROTOCOL)) {
	    nResult = OUF_PROTOCOL;
	} else if (STREQ(P3errno, P3E_SERVER)) {
	    nResult = OUF_SERVER;
	} else if (STREQ(P3errno, P3E_TTLWRONG)) {
	    nResult = OUF_TTLWRONG;
	} else {
	    nResult = OUF_ERROR;
	}
	return nResult;
    }
    OUF_DEBUG(5, ("Title retrieved from CBS.\n"));
    return OUF_OK;
} /* retrieve_title() */
