/*
 * Copyright (c) 2005 by Pica
 *
 * File	: ouf_fetch.c
 * Description	: routines for new OUF module
 * Author   	: Michael D. Lyons
 * Creation	: 30-12-2004
 *
 * History:
 * 17-01-2005	: michael : Replaced few remaining sprintf calls with snprintf
 *  23-02-2007	: clemens  : fixed problems with/cleaned-up array sizes
 *  27-05-2008	: huug     : get_title_fp: create new file when title table exists for
 *			     this ILN but its fp is NULL.
 * 
 * Notes:
 * This module was originally based on ouf.c, which will presumably still be
 * around somewhere, so I've removed the ouf.c development history from here.
 * 
 * The few functions from 1996-2004 from ouf.c that are more or less intact (a
 * lot of variables have been renamed for consistency's sake) still
 * have their original comment headers.
 *
 * The changes made to ouf are documented in the "OUF User Manual" and the
 * release notes.
 *========================================================================
 */

#pragma ident "$Id: ouf_fetch.c 364642 2019-02-06 11:20:33Z vaneckh $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <env.h>
#include <cs.h>
#include <er.h>
#include <p3.h>
#include <nmt.h>
#include <hsht.h>
#include <picatypes.h>
#include <picamisc.h>
#include "ouf_pf.h"
#include "ouf_trace.h"
#include "ouf.h"

static NMTHandle_p gPPN_XTab; /* global to hold handle to PPN exclusion table */

static HSHTHandle_p gTHTab; /* global to hold handle to Title hash table */

static int glinkTableSize = -1; /* global for size of link match table */

/* gLink_match_table is used for holding the pattern strings used for
 * accepting or rejecting the links found in a title. init_link_matchtab()
 * and match_link() both use this table.
 */

static struct {
    int sign;
    char s[4];
} gLink_match_table [OUF_MAX_LINKTAB];

/* gCBS is a struct for CBS connection handling, broken out of FSettings
 * as part of splitting ouf into the fetch and update modules. General functions
 * (initialization, etc.) using this are now in ouf_cbs.c               
 */

static CBS_t gCBS;

/* gConf is THE ugly global structure holding lots of vital counters, flags,
 * etc. that need to be visible everywhere.
 */

static struct
{
    /* Error stuff */

    int  ouferrno;
    char error[OUF_ERROR_L + 1];
    bool bErrorSet;

    /* Directories */

    char temp_d[OUF_DIR_L + 1];
    char fetch_d[OUF_DIR_L + 1];
    char title_d[OUF_DIR_L + 1];

    /* Fetch Requests */

    char format[OUF_FORMAT_L + 1];
    char expansion[OUF_EXPANSION_L + 1];
    int  nLoopCheckInterval;
    int  nMaxPPNRetrievalFreq;
    int  nFetchRequestPoll;
    int  nMaxFetchRecords;
    int  nMaxFetchTime;
    int  nFetchLines;

    /* Title Files */

    off_t nMaxTitleBytes;
    int nMaxTitleAge;
    int  HeaderSupprLev;
    char titlepatt [OUF_MAX_TITLEPATTERN];
} gConf;

/* The TitleTab_t struct is used to hold the opened FILE pointer, file name
 * and creation time for each open Title being maintained by ouf_fetch. These
 * structs are held in a hash table indexed on the ILN corresponding to the
 * given title file.  -michael
 */

typedef struct {
    FILE *fp;
    char fname[FILENAME_MAX];
    time_t ctime;
    int cnt;
} TitleTab_t, *TitleTab_p;

/* The PPN_XTab_t struct is used to implement the PPN/ILN exclusion table
 * The name table is indexed on the PPN/ILN. timadded is a timestamp for when
 * the entry was added to the table, and cnt is the number of times that a
 * fetch has been done for the particular PPN/ILN pair. retr is a bit of a
 * kludge, used as a toggle to indicate whether a given title has ever been
 * fetched or not.
 */

typedef struct {
    time_t timeadded;
    int cnt;
    bool retr;
} PPN_XTab_t, *PPN_XTab_p;

/* fetch_t is TitleTab_t's dumb cousin, and is used to hold info about the one
 * and only fetch request file that we have open and write to at any one time 
 * 
 * fp is the FILE pointer to the open file, fname is the file name, ctime is
 * the creation time of the file and linecnt is number of the line last read
 * from the file. The typedef is only there so that I can later pass the address
 * of the struct around  instead of accessing the global struct directly all of
 * the time This might make later changes less painful. -michael      
 */

typedef struct {
    FILE *fp;
    char fname[FILENAME_MAX];
    time_t ctime;
    int linecnt;
} fetch_t, *fetch_p;

static fetch_t gFetch;
static fetch_p gFetchp = &gFetch;

/* gLoopcnt is an a non-essential global counter used for stats, and may
 * disappear :-)
 */

static int gLoopcnt;

/* gStopSig is a global set by the signal handler so that other parts of the
 * program know that it is time to shut down. In particular, the main
 * processing loop always checks gStopSig.         
 */

static volatile int gStopSig = 0;

/*------------------------------------------------------------------------
 *  Function	: expire_title
 *  Purpose	: expire a title in the Title table based on age or size
 *
 *  History:
 *  30-11-2004	: michael : creation
 *  22-12-2004	: michael : added special handling for empty files
 *                          we want zero-length files to be cleaned
 *                          up when ouf exits.
 *  23-12-2004	: michael : now, zero-length files "can't happen" so the
 *                          special-case code is gone again :-)
 *
 *  Notes	: 
 *  expire_title() is called by HSHTWalk for each member of the hash table, and
 *  does the following:
 *
 *  0. checks to see whether the associated Title File is old enough or big
 *     enough to be moved to the Title directory (can be forced, as mentioned
 *     above)
 *
 *  1. fclose() of the title file
 *
 *  2. mv of title file from temp to title directory
 *
 *  3. reinitialization of the associated hash table entry.
 *
 *  The idea is for these routines to be called periodically from within the
 *  OUF main processing loop (and possibly via the signal handler during
 *  cleanup if the process gets killed :-))
 *
 *  It is unfortunate that system() is used to move the files,
 *  but the rename() syscall doesn't work across file systems,
 *  and I can't force the location of the user-definable temp
 *  directory. :-(
 *
 *  Returns	: bool
 *------------------------------------------------------------------------
 */
static bool
expire_title(
	HSHTKey_t key,
	cvoidp val,
	voidp nowtime)
{
    char sbuf[BUFSIZ];
    TitleTab_p ttp = (TitleTab_p) val;
    int filesize;

    if (ttp->fp == NULL) {
	/* this struct is currently not (yet) in use */
	OUF_DEBUG(50, ("%s: expire_title(): %03d not in use\n", OUF_FETCH,key));
	return TRUE;
    }
    filesize = ftell(ttp->fp);

    /* A NULL nowtime pointer means, "just do this" */

    if (!nowtime
	|| filesize >= gConf.nMaxTitleBytes
	|| (int) difftime(*(time_t *) nowtime,ttp->ctime) > gConf.nMaxTitleAge){

	(void) fclose(ttp->fp);

	/* Move title file to title directory */
	(void) sprintf(sbuf, "mv %s %s/", ttp->fname, gConf.title_d);
	(void) system(sbuf);
	
	if (gDebug >= 5) {
	    int n = 0;
	    if (nowtime) {
		n = (int) difftime(*(time_t *) nowtime, ttp->ctime);
	    }
	    (void)tr_write(
	       "%s: moving %s, off: %d\tcnt: %d\tdifftime: %dm %ds\n",
		OUF_FETCH, OUFStripFilename(ttp->fname),
		filesize, ttp->cnt, n/60, n%60);
	}
	(void) memset((void *) ttp, '\0', sizeof(ttp));
    }
    return TRUE;
} /* expire_title() */

/*------------------------------------------------------------------------
 * Function	: get_title_fp
 * Purpose	: Get the FILE pointer to the title file for a given ILN
 *
 * History:
 *  26-11-2004	: michael  : creation
 *  27-05-2008	: huug     : create new file when title table exists for
 *			     this ILN but its fp is NULL.
 *
 * Notes	:
 * This does more than the name implies, since it actually
 * accesses a hash table indexed on the ILN and returns
 * the FILE pointer from a structure containing that and
 * other info about a given title (e.g. the file name).
 *
 * If the hash table entry doesn't exist, then the
 * function actually will actually malloc() a structure
 * for a new entry, fopen() a title file and return the
 * FILE pointer (which is in the structure, of course :-)).
 *
 * Returns:
 * a FILE pointer to the opened file, or NULL if something failed
 *------------------------------------------------------------------------
 */
static FILE *
get_title_fp(
	char *ILN)
{
    TitleTab_p ttp;
    int iln = atoi(ILN);
    char *s;

    ttp = (TitleTab_p) HSHTGet(gTHTab, iln);

    if (ttp != NULL && ttp->fp != NULL) {
	ttp->cnt++;
	return ttp->fp;
    }

    if (ttp == NULL) {
	/* need to create a new table entry for this iln */
	ttp = (TitleTab_p) malloc(sizeof(TitleTab_t));

	if (ttp == NULL) {
	    (void) tr_write("%s: can't malloc %d bytes in get_title_fp\n",
			    OUF_FETCH, (int) sizeof(TitleTab_t));
	    return NULL;
	}
	(void) memset((void *) ttp, '\0', sizeof(ttp));

	/* ttp will still be the correct address after the HSHTPut */
	HSHTPut(gTHTab, iln, (void *) ttp);
    }

    s = get_titlefname(gConf.titlepatt, gConf.temp_d, iln);
    
    if (s == NULL) {
	/* no title name? :-( */
	(void) tr_write("get_titlefname() failed in get_title_fp!\n");
	return NULL;
    }
    ttp->fp = fopen(s, "w");
    if (ttp->fp != NULL) {
	(void) strcpy(ttp->fname, s); /* strcpy is okay here */
	ttp->cnt = 1;
	ttp->ctime = time(NULL); /* now */
	HSHTPut(gTHTab, iln, (void *) ttp);
	OUF_DEBUG(2,("%s: title file '%s' created.\n", OUF_FETCH, ttp->fname));
    } else {
	(void) tr_write("%s: can't open title file \"%s\" in get_title_fp\n",
	    OUF_FETCH, ttp->fname);
	OUFSetError(OUF_TTL_FILE_NOT_OPENED);
    }
    return ttp->fp;
} /* get_title_fp() */

/*------------------------------------------------------------------------
 *  Function	: add_ptabentry  
 *  Purpose	: insert an entry into the PPN table, or bump the counters
 *                for an entry which is already in the table.
 *  Method	: 
 *
 *  History:
 *  25-11-2004	: michael : creation
 *
 *  Notes	: 
 *
 *  Returns	: the address of the struct corresponding to the key
 *                or NULL if something breaks  
 *------------------------------------------------------------------------
 */
static PPN_XTab_p
add_ptabentry(
	NMTHandle_p ptab,
	char *pkey)
{
    PPN_XTab_p ppnp = (PPN_XTab_p) NMTGet(ptab, pkey);

    if (ppnp != NULL) {
        /* already in table, bump counter and return */
	OUF_DEBUG(5,("%s is already in the table (%d)\n", pkey, ppnp->cnt));
        ppnp->cnt++;
	return ppnp;
    }
    ppnp = (PPN_XTab_p) malloc(sizeof(PPN_XTab_t));
    if (ppnp != NULL) {
	ppnp->retr = FALSE;
	ppnp->cnt = 1;
	ppnp->timeadded = time(NULL);
	NMTPut(ptab, pkey, (void *) ppnp);
	OUF_DEBUG(5,("%s: pair %s has been added to the table\n",
	    OUF_FETCH, pkey));
    } else {
	OUF_DEBUG(5,("%s: can't allocate PPN entry for %s in add_ptabentry\n",
	    OUF_FETCH, pkey));
    }
    return ppnp;
} /* add_ptabentry() */

/*------------------------------------------------------------------------
 * Function	: expire_ppn
 * Purpose	: detects/releases expired entries in the PPN name table.
 * Method	: 
 *
 * History:
 * 29-11-2004	: michael : creation
 * 17-12-2004	: michael : added timediff variable for readability :-)
 *                          while doing so, found that I was using ppnp
 *                          after I'd done a free() on it. I got lucky.
 *
 * Notes       :
 * This function is a generic handler function intended for
 * use in an NMTWalk(). The caller should initialize a
 * time_t variable with a timestamp (usually the "current"
 * time), and expire_ppn will walk through the name table,
 * deleting entries that were created before the timestamp,
 * and freeing their storage.
 *
 * A built-in assumption here is that this routine will be
 * called frequently enough that the timeouts will be more
 * or less "obeyed". I have seen variations of a maximum
 * of around 30 seconds when processing enormous fetch files
 * and tens of updates per second.
 *
 *  Returns	: Always TRUE; this is an NMTWalk() callback
 *------------------------------------------------------------------------
 */
static bool
expire_ppn(
	const char *name,
	cvoidp value,
	voidp ct)
{
    PPN_XTab_p ppnp = (PPN_XTab_p) value;

    int timediff = (int) difftime(*(time_t *)ct, ppnp->timeadded);

    if (timediff >= gConf.nLoopCheckInterval){
	NMTDel(gPPN_XTab, name);
	free((void *)value);
	OUF_DEBUG(6,("%s: %s expired - +%02d secs\n", OUF_FETCH,
	    (char *) name, timediff - gConf.nLoopCheckInterval));
    }
    return TRUE;
} /* expire_ppn() */

/*------------------------------------------------------------------------
 * Function	: NMTcnt
 * Purpose	: "counts" the entries of a name table.
 * Method	: 
 *
 * History:
 * 29-11-2004	: michael : creation
 *
 * Notes       :
 * This function is a generic handler function intended for
 * use in an NMTWalk(). The caller should initialize an int
 * variable to zero, and then pass its address as the third
 * parameter to NMTCnt. handler. Afterwards, the variable
 * should contain the number of entries in the table.
 * 
 * This is a very simple function; note how the comment is
 * about ten times longer than the actual code. :-)
 *
 * Returns: TRUE, always.
 *------------------------------------------------------------------------
 */
static bool
NMTCnt (
	const char *name,
	cvoidp value,
	voidp cnt)
{
    (void) name;
    (void) value;

    (*(int *) cnt)++;
    return TRUE;
} /* NMTCnt () */

/*------------------------------------------------------------------------
 * Function	: THTcnt
 * Purpose	: "counts" the entries of the title hash table.
 * Method	: 
 *
 * History:
 * 29-11-2004	: michael : creation
 *
 * Notes       : This function is similar to NMTCnt, but is specifically
 *                for the title hash table, as can be seen by the reference
 *                to the embedded FILE pointer  
 *
 * Returns: Always TRUE
 *------------------------------------------------------------------------
 */
static bool
THTcnt(
	HSHTKey_t key,
	cvoidp value,
	voidp cnt)
{
    (void) key;

    if (((TitleTab_p) value)->fp != NULL) (*(int *) cnt)++;
    return TRUE;
} /* THTcnt() */

/*------------------------------------------------------------------------
 * Function	: check_fetch_dirs
 * Purpose	: Checks the existence of required directories
 * Method	: 
 *
 * History:
 * 07-10-1996	: peter jan: creation
 * 28-10-2004	: michael  : Completely changed from revision before
 *                           last. Now based around dir_exists calls
 *
 * Returns	: returns FALSE if one the dir_exists calls returns
 *                FALSE, TRUE otherwise.
 *--------------------------------------------------------------------*/
static bool
check_fetch_dirs(void)
{
    OUF_DEBUG(2,("%s: checking for required directories...\n", OUF_FETCH));

    if (!dir_exists(gConf.fetch_d, OUF_FETCH_DIR_NOT_FOUND,
	OUF_FETCH_DIR_PROTECTED, OUF_FETCH_DIR_NOTDIR)) {
	return FALSE;
    }
    if (!dir_exists(gConf.title_d, OUF_TITLE_DIR_NOT_FOUND,
	OUF_TITLE_DIR_PROTECTED, OUF_TITLE_DIR_NOTDIR)) {
	return FALSE;
    }
    if (!dir_exists(gConf.temp_d, OUF_TEMP_DIR_NOT_FOUND,
	OUF_TEMP_DIR_PROTECTED, OUF_TEMP_DIR_NOTDIR)) {
	return FALSE;
    }
    OUF_DEBUG(2,("%s: required directories checked.\n", OUF_FETCH));
    return TRUE;
} /* check_fetch_dirs() */

/*------------------------------------------------------------------------
 * Function	: expire_linkfetch
 *
 * Purpose	: expires the currently open link fetch file
 *
 * History:
 * 04-01-2005	: michael  : Creation
 *
 * Notes:
 * expire_linkfetch() checks the age- and size-based criteria for expiration
 * of the currently open link fetch file, and moves it to the fetch directory
 * if the file is "ripe" :-)
 *
 * The function also reinitializes the structure holding the counter, timer
 * and FILE pointer to fetch file after moving the file, if it has expired.
 *
 * Returns: Nothing
 *--------------------------------------------------------------------------
 */
static void
expire_linkfetch(
	fetch_p ftcp)
{
    char sbuf[BUFSIZ];

    int timediff = (int) difftime(time(NULL), ftcp->ctime);

    /* if gStopSig is non-zero, then we're shutting down */

    if ((ftcp->fp != NULL) && ((ftcp->linecnt >= gConf.nMaxFetchRecords) ||
	    timediff > gConf.nMaxFetchTime || gStopSig)) {
	(void) fclose(ftcp->fp);
	(void) sprintf(sbuf, "mv %s %s", ftcp->fname, gConf.fetch_d);
	if (system(sbuf) == 0) {
	    (void) tr_write(
		 "   Link fetch file %s moved to fetch d: %d lines, %ds\n",
		OUFStripFilename(ftcp->fname), ftcp->linecnt, timediff);
	} else {
	    (void) tr_write("%s: move of link fetch file %s failed!\n",
		OUF_FETCH, OUFStripFilename(ftcp->fname));
	}
	(void) memset((void *) ftcp, '\0', sizeof(ftcp));
    }
} /* expire_linkfetch() */

/*------------------------------------------------------------------------
 * Function	: expire_title_files
 * Purpose	: processes title table, expiring title files
 * Method	: 
 *
 * History:
 * 29-11-2004	: michael : creation
 *
 * Notes:
 * expire_title_files() is just a wrapper around an HSHTWalk over 
 * the Title Table using expire_title as a call back.      
 * The added value of this function is some extra trace out-
 * put, and a before-and-after count of the entries.
 *
 * Returns: nothing
 *------------------------------------------------------------------------
 */
static void
expire_title_files(
	int force)
{
    int cnt = 0;
    time_t nowtime;

    (void) HSHTWalk(gTHTab, THTcnt, &cnt);
    if (cnt) {
	OUF_DEBUG(6,("%s: before T-expire, %d entries\n", OUF_FETCH, cnt));
	
	if (!force) {
	    (void) time(&nowtime);
	}
	(void) HSHTWalk(gTHTab, expire_title, force ? NULL : &nowtime);

	cnt = 0;
	(void) HSHTWalk(gTHTab, THTcnt, &cnt);
	OUF_DEBUG(6,("%s: after T-expire, %d entries\n", OUF_FETCH, cnt));
    } else {
	OUF_DEBUG(6,("%s: title table is empty\n", OUF_FETCH));
    }
} /* expire_title_files() */

/*------------------------------------------------------------------------
 * Function	: get_linkfetch_fp
 *
 * Purpose	: obtain the (possibly new) open FILE pointer to the current
 *                link fetch file.
 *
 * History	:
 * 04-01-2005  : michael : creation
 *
 * Notes	:
 * If the call to expire_linkfetch() actually cause the link fetch file to be
 * expired, then the FILE pointer ftcp will be NULL, and a new file must
 * be created and opened.
 *
 * The function gets a filename in the temp directory via a function with a
 * verrrrry long name, and then does an fopen(). Some maintenance is done on
 * the counter and timer in the fetch file structure, and the FILE pointer to
 * the new fetch file is returned.
 *
 * Returns: a FILE pointer to the opened link fetch file (or NULL)
 *--------------------------------------------------------------------
 */
static FILE *
get_linkfetch_fp(
	fetch_p ftcp)
{
    expire_linkfetch(ftcp);
    
    if (ftcp->fp == NULL) {
	(void) OUFConstructTimeStampedFilename(
	   ftcp->fname,gConf.temp_d,OUF_FETCH_FILE);

	ftcp->fp = fopen(ftcp->fname, "w");
	if (ftcp->fp != NULL) {
	    ftcp->linecnt = 0;
	    ftcp->ctime = time(NULL); /* now */
	    (void) tr_write("   Link fetch file %s created in temp dir.\n",
		OUFStripFilename(ftcp->fname));
	} else {
	    (void)tr_write("%s: couldn't open fetch file %s in get_linkfetch\n",
	       OUF_FETCH, ftcp->fname);
	    OUFSetError(OUF_TTL_FILE_NOT_OPENED);
	}
    }
    return ftcp->fp;
} /* get_linkfetch_fp() */

/*------------------------------------------------------------------------
 * Function	: match_link()
 * Purpose	: Checks whether a given LinkType token matches a pattern
 *                in the OUF link table
 * Method	: 
 *
 * History:
 * 08-11-2004	: michael : creation
 *
 * Notes	:
 * The string to be matched is in the function parameter
 * char * lt. (link token)
 * 
 * The static global glinkTableSize represents the current
 * size of the link table array gLink_match_table, if any.
 *
 * glinkTableSize is visible to both this function and 
 * init_link_matchtab, which actually changes the counter's
 * value, maybe. :-)
 *
 * If the glinkTableSize has not been changed, then it will
 * still be set to -1.
 *
 * The function is boolean, and returns TRUE on a positive
 * match, and FALSE otherwise
 *
 * See init_link_matchtab() below as well.
 *
 * Returns: TRUE if a match was found, FALSE otherwise
 *------------------------------------------------------------------------
 */
static bool
match_link(
	const char *lt)
{
    int i;

    if (glinkTableSize < 0) return FALSE;

    for (i = 0; i <= glinkTableSize; i++) {
	if (((gLink_match_table[i].s[0] == '*')
	    || (gLink_match_table[i].s[0] == lt[0]))
	    && (((gLink_match_table[i].s[1] == '*')
	    || (gLink_match_table[i].s[1] == lt[1])))) {
		return  (gLink_match_table[i].sign == -1) ? FALSE : TRUE;
	}
    }
    return FALSE;
} /* match_link() */

/*------------------------------------------------------------------------
 * Function	: process_links
 * Purpose	: Parses link variable strings and builds a fetch request
 *                file from the PPNs that it finds.
 * Method	: 
 *
 * History:
 * 19-11-2004	: michael: creation
 * 24-11-2004	: michael: now passing in ILN as well
 *
 * Notes	:
 * this routine takes a string containing a list of link variable tags and
 * their associated PPNs, matches the link types with a set of rules as
 * configured by the user and then write any PPNs corresponding to the matched
 * links into a fetch request file for later processing in the OUF main
 * loop.
 * 
 * The incoming string should be a list of link-type/PPN
 * pairs separated by carriage return characters.
 *
 * process_links parses each pair out of the string using strtok(), and then
 * compares the link type (e.g. "Tp") with a pattern matching table built
 * earlier when OUF started up.
 *
 * If the pattern matches, then this function writes the
 * associated PPN to a fetch request file.
 *
 * The name and location of the fetch request file is hidden from
 * process_links(), and maintained by get_linkfetch_fp().
 *
 * process_links() gets a FILE pointer from get_linkfetch_fp().
 *
 * process_links() can, however force a new fetch request
 * file to be created by sending a parameter to get_linkfetch_fp()
 *
 * In particular, process_links() monitors the number of records it has written
 * to a fetch request file, and forces a new file to be started when the
 * parameterizable maximum number of lines in the fetch request file has
 * been exceeded.
 *
 * Also, this is not a bug, per se, but this function and get_linkfetch_fp
 * share a global counter for the number of lines written to the current fetch
 * request file because get_fetchp() needs to be able to reset its value to
 * zero when creating a new file.
 *
 * An alternative would be to make the file size limits be based on the file
 * size in bytes, and not the number of lines; in this case, an ftell() on the
 * FILE pointer could be done.
 *
 * In the meantime, the fetch file pointer, line counter timer have all been
 * banished to a global structure, and all of the functions access the
 * variables via a pointer to the structure.
 *  
 * Returns	: bool
 *------------------------------------------------------------------------
 */
static bool
process_links(
	const char *iln,
	const char *format,
	const char *str)
{
    char *pos;
    char key [OUF_MAX_KEYLEN];
    char LXbuf [2 * BUFSIZ];
    PPN_XTab_p ppnp;

    (void) memset((void *) LXbuf, '\0', sizeof(LXbuf));
    (void) PM_strncpy(LXbuf, str, sizeof(LXbuf));

    pos = strtok(LXbuf, "\r");

    if (pos != NULL) {
	do {
	    pos[2] = '\0';
	    if (match_link(pos)) {
		(void) sprintf(key, "%s:%0*d",pos+3, PIC_ILN_DECIMAL_MAX, strlen(iln) ? atoi(iln):0);
		OUF_DEBUG(6,("%s: LX: %s Key: %s\n", OUF_FETCH, pos, key));
		
		ppnp = add_ptabentry(gPPN_XTab, key);
		if (ppnp == NULL) {
		    /* something went terribly wrong */
		    (void) tr_write("%s: process_links table add failure.\n",
			OUF_FETCH);
		    return FALSE;
		}
		if (ppnp->cnt <= gConf.nMaxPPNRetrievalFreq) {
		    FILE *fp = get_linkfetch_fp(gFetchp);
		    if (fp == NULL) return FALSE;
		    OUF_DEBUG(1,("%s: writing fetch from process_links\n",
			OUF_FETCH));
		    (void) fprintf(fp, "%s %s %s\n", pos+3, iln, format);
		    gFetchp->linecnt++;
		} else {
		    OUF_DEBUG (6,("%s: link %s was already in PPN table\n",
			OUF_FETCH, key));
		}
	    }
	} while((pos = strtok(NULL, "\r")) != NULL);
    }
    return TRUE;
} /* process_links() */

/*------------------------------------------------------------------------
 * Function	: get_request_tokens
 * Purpose	: Parses the tokens out a fetch request string in the
 *               first parameter req, and write them into the PPN,
 *               ILN and format strings.
 *
 * History	:
 * 19-11-2004   : michael : creation
 *
 * Notes	:
 * This routine is fairly simple-minded, and just uses strtok() to break up the
 * input string.   There's no error-handling as such, but it does return the
 * number of fields that it parsed, which the calling function can then decide
 * to dispose of as it sees fit.
 *------------------------------------------------------------------------
 */
static int
get_request_tokens(
	char *req,
	char *PPN,
	char *ILN,
	char *format)
{
    static char delim2[] = "\n\t\r ";
    char *pos;
    int nfields = 0;

    /* The caller had better make certain that all */
    /* of these pointers are good, or else :-)     */

    if (gDebug >= 7) {
	char *p;
	/* only need to get rid of the newline if */
	/* we're pretty printing some debug messages */
	if ((p = strrchr(req, '\n'))) *p = 0;
	OUF_DEBUG(7,("%s: get_request_tokens (\"%s\")\n", OUF_FETCH, req));
    }
   
    PPN[0] = format[0] = ILN[0] = '\0';
    pos = strtok(req, delim2);
    if (pos != NULL) {
	nfields++;
	(void) PM_strncpy(PPN, pos, OUF_PPN_L+1);
	OUF_DEBUG(7, ("%s: grabbed PPN: \"%s\"\n", OUF_FETCH, PPN));
	pos = strtok(NULL, delim2);
	if (pos != NULL) {
	    nfields++;
	    (void) PM_strncpy(ILN, pos, OUF_ILN_L+1);
	    OUF_DEBUG(7,("%s: grabbed ILN: \"%s\"\n", OUF_FETCH, ILN));
	    pos = strtok(NULL, delim2);
	    if (pos != NULL) {
		nfields++;
		(void) PM_strncpy(format, pos, OUF_FORMAT_L+1);
		OUF_DEBUG(7,("%s: grabbed FMT: \"%s\"\n", OUF_FETCH, format));
	    }
	}
    }
    OUF_DEBUG(7,("%s: get_request_tokens - nfields: %d\n",OUF_FETCH, nfields));
    return nfields;
} /* get_request_tokens() */

/*------------------------------------------------------------------------
 *  Function	: write_tags
 *  Purpose	: Writes title tags to title file
 *
 *  History:
 *  23-08-1996	: peter jan: creation
 *  02-10-1996  : peter jan: SEP will be replaced by '\r'
 *  07-01-2005  : michael: name changed from OUFWriteTitleTags
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
write_tags(
	FILE *tfp,
	CBBuf_t *CBbuf,
	size_t CBSize)
{
    int		 nSrcIndex, nDestIndex;
    CBElm_p	CBElm;
    const char	*AttrValue;
    char	AttrRaw[OUF_BUFFER];
    int		cnt = 0;

    OUF_DEBUG(5,("%s: beginning to write title tags...\n", OUF_FETCH));

    /* Get contents of TX field */
    CBElm = CBGetAttribute(CBbuf, CBSize, "TX", &AttrValue);

    while (CBElm != NULL) {
	/* Check for edit script header */
	if (strncmp(AttrValue, "\033H\033P\033%\033V", 8) == 0) {
            /* Remove edit script header */

            /* Ignore first part: \033H\033P\033%\033V */

            /* Copy until \033H */
            nDestIndex = 0;
            for (nSrcIndex = 8; ((AttrValue[nSrcIndex] != '\033') &&
	         (AttrValue[nSrcIndex+1] != 'H')); nSrcIndex++) {
	        if (gConf.HeaderSupprLev < 2) {
		    AttrRaw[nDestIndex++] = AttrValue[nSrcIndex];    
		}
            }
            /* Copy until: SEP */
            nSrcIndex += 2;
            while (AttrValue[nSrcIndex] != OUF_SEP) {
	        if (gConf.HeaderSupprLev < 2) {
		    AttrRaw[nDestIndex++] = AttrValue[nSrcIndex];
		}
		nSrcIndex++;
            }
	    if (gConf.HeaderSupprLev < 2) AttrRaw[nDestIndex++] = '\r';

            /* Ignore length and \033V\033U\r */
            while (AttrValue[++nSrcIndex] != '\r');

            /* Copy rest of attribute */
            (void) strcpy(&(AttrRaw[nDestIndex]), &(AttrValue[nSrcIndex+1]));

            /* Write contents of TX field to title file */
            (void) fprintf(tfp, "%s\r\r", AttrRaw);
	    cnt++;
        }
        /* Get next TX attribute, if there is any... */
        CBElm = CBNextAttribute(CBbuf, CBSize, CBElm, &AttrValue);
    }
    OUF_DEBUG(5,("%s: %d title tags written.\n", OUF_FETCH, cnt));
    return TRUE;
} /* write_tags() */

/*------------------------------------------------------------------------
 *  Function	: write_body
 *  Purpose	: Writes title body (copies included) to title file
 *
 *  History:
 *  01-10-1996	: peter jan: creation
 *  07-01-2005	: michael : name changed from OUFWriteTitleBody
 *  12-01-2005	: michael : got rid of unnecessary variable cCopy
 *  08-02-2005	: michael : added support for 99+ copies (EX attribute)
 *                          will break this function up later, maybe :-)
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
write_body(
	FILE *tfp,
	CBBuf_t *CBbuf,
	size_t CBSize,
	char *PPN,
	char *syntax,
	char *ILN)
{
    const char *AttrValue;
    CBElm_p    CBElm;
    bool       bOk = TRUE;
    char       attrcopy[OUF_MORE_L + 1];
    char       copy[PM_INTLEN_MAX];
    int        nResult;

    bOk = write_tags(tfp, CBbuf, CBSize);

    CBElm = CBGetAttribute(CBbuf, CBSize, "EX", &AttrValue);

    if (CBElm != NULL) {
	const char *np;
	np = AttrValue;
	OUF_DEBUG(7,("%s: EX attribute present: \"%s\"\n", OUF_FETCH, np));
	while ((np != NULL) && bOk) {
	    unsigned long first;
	    unsigned long last;
	    np = PM_range_get_bounds (np, &first, &last);
	    if (np != NULL) {
		for (unsigned long i = first; i <= last; i++) {
		    (void) sprintf(copy, "%03d", (int) i);
		    nResult = retrieve_title(CBbuf, CBSize,
			    PPN, copy, syntax, gConf.expansion, ILN, &gCBS);
		    if (nResult == OUF_OK) {
			bOk = write_tags (tfp, CBbuf, CBSize);
		    }
		}
	    }
	}
    } else {
	/* Check if an MO field is present */

	CBElm = CBGetAttribute(CBbuf, CBSize, "MO", &AttrValue);
	if (CBElm != NULL) {
	    /* Copy the attribute we just retrieved,
	     * because we are going to use the same counted
	     * buffer several times in a row.
	     */
	    bOk = TRUE;
	    OUF_DEBUG(7,("%s: MO attribute present: \"%s\"\n",
			OUF_FETCH, AttrValue));
	    /* tr_write ("NSA: \"%s\"\n", AttrValue); */
	    (void) PM_strncpy(attrcopy, AttrValue, OUF_MORE_L+1);
	    for (int i = 1; (i <= OUF_MORE_L) && bOk; i++) {
		if (*(attrcopy + i - 1) != ' ') {
		    /* Copy is present */

		    /* Construct title copy number */
		    (void) sprintf(copy, "%03d", i);

		    /* Retrieve title copy from CBS */
		    nResult = retrieve_title(CBbuf, CBSize,
			    PPN, copy, syntax, gConf.expansion, ILN, &gCBS);
		    if (nResult == OUF_OK) {
			bOk = write_tags (tfp, CBbuf, CBSize);
		    }
		}
	    }
	}
    }
    return bOk;
} /* write_body() */

/*------------------------------------------------------------------------
 *  Function	: write_header
 *  Purpose	: Writes title header to title file
 *
 *  History:
 *  23-08-1996	: peter jan: creation
 *
 *  Returns	: TRUE
 *------------------------------------------------------------------------
 */
static bool
write_header(
	FILE *fp,
	char *PPN,
	char *format,
	int retval,
	char *ILN)
{
    if (retval == OUF_OK){
      if (*ILN == '\0'){
	(void) fprintf(fp, "PPN:%s %s\r", PPN, format);
      } else {
        (void) fprintf(fp, "PPN:%s %03d %s\r", PPN, atoi(ILN), format);
      }
    } else {
	if (*ILN == '\0'){
	    (void) fprintf(fp, "PPN:%s %s %d %s\r", PPN, format, retval,
	       OUFGetResultMessage(retval, gConf.error));
	} else {
	    (void) fprintf(fp, "PPN:%s %03d %s %d %s\r",
	       PPN, atoi(ILN), format, retval, 
		       OUFGetResultMessage(retval, gConf.error));
	}
    }
    return TRUE;
} /* write_header() */

/*------------------------------------------------------------------------
 * Function	: fetch_title()
 * Purpose	: fetch_title() takes a PPn, ILN and format strings
 *                as parameters.
 * History	:
 * 19-11-2004   : michael : creation
 *  21-09-2007	: clemens  : allow unrecognized formats
 *
 * Notes:
 * fetch_title()  first performs a few simple tests on the input to make
 * certain that the data is in the correct format, and may also perform a
 * syntax conversion on the format string, if needed.
 *
 * If any one of the checks or the conversion fails, the function returns 1 to
 * the caller to indicate the error.
 *
 * If no error has occurred, then the PPN and ILN are combined into a string
 * (e.g. "200114301:99"), and this value is used as parameter to
 * add_ptabentry() to be used as an index into the "PPN exclusion table"
 * 
 * This table is used to avoid loops in retrievals, and gives each PPN/ILN pair
 * a fixed maximum number of times that the associated title with that PPN and
 * ILN can be retrieved from CBS.
 *
 * If the maximum retrieval frequency for the given PPN/ILN pair has, in fact,
 * not been exceeded, then fetch_title send a title retrieval command to CBS
 * with the function retrieve_title().
 *
 * Finally, if the title was successfully retrieved from CBS, then
 * fetch_title() first calls process_links() to handle any material type links
 * that may have been returned with the title, and then writes the title to the
 * appropriate ILN-specific title file in OUF format using the functions
 * write_header() and write_body()
 *                
 * Returns:
 * TRUE if all of the checks and the fetch itself panned out, and
 * FALSE otherwise.
 *------------------------------------------------------------------------
 */
static bool
fetch_title(
	char *PPN,
	char *ILN,
	char *format)
{
    int nLen;
    int i;
    char    syntax[OUF_SYNTAX_L + 1];
    char    pkey [OUF_MAX_KEYLEN];
    PPN_XTab_p ppnp;
    int     nResult;
    FILE    *tfp;
    CBBuf_t CBbuf[OUF_BUFFER];
    CBElm_p    CBElm;
    const char *attr;
    bool    bOk = TRUE;
    
    OUF_DEBUG(9,("%s: fetch_title (\"%s\", \"%s\", \"%s\")\n", OUF_FETCH,
	PPN, ILN, format));

    /* Check PPN */
    nLen = strlen(PPN);
    
    for (i = 0; (i < nLen) && bOk; i++) {
	bOk = isdigit(PPN[i]) || tolower(PPN[i]) == 'x';
    }

    /* Check ILN */
    nLen = strlen(ILN);
    if (nLen > PIC_ILN_DECIMAL_MAX) {
	(void) tr_write("%s: ILN \"%s\" is too long, discarding request\n",
	    OUF_FETCH, ILN);
	return FALSE;
    }
    for (i = 0; (i < nLen) && bOk; i++) bOk = (isdigit(ILN[i]) != 0);

    if (bOk) {
	/* Convert Pica3 format to ADB format */
	bOk = OUFConvertPica3ToADBFormat(syntax, format);
	/* If format is not recognized use the one specified literally */
	if (! bOk) {
	    PM_strncpy(syntax, format, sizeof(syntax));
	    bOk = TRUE;
	}
    }

    if (!bOk) {
	OUFSetError(OUF_FR_SYNTAX_ERROR);
	(void) tr_write("%s: syntax error in request... discarding\n",
	     OUF_FETCH);
	return FALSE;
    }

    (void) sprintf(pkey, "%s:%0*d", PPN, PIC_ILN_DECIMAL_MAX, atoi(ILN));

    OUF_DEBUG(6,("%s: calling add_ptabentry(\"%s\") from fetch_title()\n",
	OUF_FETCH, pkey));

    ppnp = add_ptabentry(gPPN_XTab, pkey);

    if (ppnp == NULL) {
	OUF_DEBUG(2,("%s: bad name table element pointer in fetch loop.\n",
	    OUF_FETCH));
	return FALSE;
    }

    if (ppnp->cnt <= gConf.nMaxPPNRetrievalFreq || !ppnp->retr) {

	/* Retrieve requested title from CBS */

	/* the retr flag is a simple fix so that we can */
	/* tell whether a given PPN has already been    */
	/* processed for LX fields or not. -Michael     */
	ppnp->retr = TRUE;

	nResult = retrieve_title(CBbuf, sizeof(CBbuf),
		  PPN, NULL, syntax, gConf.expansion, ILN, &gCBS);

	/* If there are any LinkType patterns in the */
	/* table to check against, then we try to    */
	/* find and pull out any link record fields  */
	/* of type LX, and do the right thing... :-) */ 

	if (nResult == OUF_OK) {
	    if (glinkTableSize >= 0) {
		OUF_DEBUG(6,("%s: checking record for LX attribute...\n",
		    OUF_FETCH));
		CBElm = CBGetAttribute(CBbuf, sizeof(CBbuf),"LX",&attr);
		if (CBElm != NULL) {
		    /* the attribute is present, its value is in
		     * attr. Let's try to do useful stuff with it
		     */
		    OUF_DEBUG(6,("%s: %s: LX found...call process_links\n",
			OUF_FETCH, PPN));
		    (void) process_links(ILN, format, attr);
		} else {
		    OUF_DEBUG(6,("%s: %s: NO LX attribute present...\n",
			OUF_FETCH, PPN));
		}
	    }

	    /* we have to do this each time through the loop 
	     * because the ILN might change. It's a low over- 
	     * head call if the file has already been opened.
	     */

	    tfp = get_title_fp(ILN);

	    /* Write title header */
	    if (tfp != NULL) {
		if (gConf.HeaderSupprLev <= 0) {
		    OUF_DEBUG(6,("%s: writing PPN title header.\n", OUF_FETCH));
		    bOk = write_header(tfp, PPN, format, nResult, ILN);
		} else {
		    OUF_DEBUG(6,("%s: skipping PPN title header.\n", OUF_FETCH));
		}
	    } else {
		(void) tr_write("no tfp in fetch title!\n");
		bOk = FALSE;
	    }
	}

	if ((nResult == OUF_OK) && bOk) {
	    /* tfp must be okay if we're here */
	    bOk = write_body(tfp, CBbuf, sizeof(CBbuf),PPN,syntax,ILN);
	    if (!bOk) (void) tr_write("write_body failed in fetch_title\n");
	}
    } else {
	gLoopcnt++;
	OUF_DEBUG(5,("%s: possible loop in retrieval %s:%s (%d %d %d)\n",
	    OUF_FETCH, PPN, ILN, ppnp->retr, ppnp->cnt,
	    gConf.nMaxPPNRetrievalFreq));
    }
    return bOk;
} /* fetch_title() */

/*------------------------------------------------------------------------
 * Function	: process_fetch_request
 * Purpose	: process_fetch_request for a PPN from a fetch request file
 * Method	: 
 *
 * History:
 * 19-11-2004	: michael  : creation
 *
 * Notes:
 * process_fetch_request is called with a line from a fetch
 * request file as a parameter, and first calls
 * get_request_tokens to break up the request into a
 * PPN, ILN and format (the last two fields are optional).
 *
 * The function does some simple sanity checks on the sizes
 * and number of fields in the request, and then calls
 * fetch_title() to do the real work . :-)
 *
 * A return value of 1 is an error, otherwise, the return
 * value of fetch_title() is returned (which may also be
 * an error :-))
 *
 * Returns: TRUE if the processing went well, FALSE if not :-)
 *------------------------------------------------------------------------
 */
static bool
process_fetch_request(
	char *req)
{
    bool    bOk = TRUE;
    char    PPN[OUF_PPN_L + 1];
    char    ILN[OUF_ILN_L + 1];
    char    format[OUF_FORMAT_L + 1];
    int  i;
    int  nFields;
    int  nLen;

    OUF_DEBUG(5,("%s: parsing: \"%s\"\n", OUF_FETCH, req));

    /* Check length of request */

    if (strlen(req) <= (size_t)OUF_REQUEST_L) {

	/* It's the caller's responsibility to make certain      */
	/* that get_request_tokens() receives completely initialized*/
	/* strings for storing the parsed PPN, ILN and format    */

	(void) memset(PPN, '\0', sizeof(PPN));
	(void) memset(ILN, '\0', sizeof(ILN));
	(void) memset(format, '\0', sizeof(format));

	/* Split fetch request into production number and format */

	nFields = get_request_tokens(req, PPN, ILN, format);

	/* Check number of fields */

	if (!((nFields >= 1) && (nFields <= 3))) {
	    OUF_DEBUG(5,
	    ("%s: wrong num of fields (%d) in process_fetch_request.\n",
		OUF_FETCH, nFields));
	    return FALSE;
	}

	/* Check: ILN and format both optional */

	if (nFields == 2) {
	    nLen = strlen(ILN);

	    for (i = 0; (i < nLen) && bOk; i++) bOk = (isdigit(ILN[i]) != 0);

	    if (bOk && (nLen > PIC_ILN_DECIMAL_MAX)) {
		(void) tr_write(
		    "%s: ILN \"%s\" too long, in process_fetch_request.\n",
		    OUF_FETCH,ILN);
		return FALSE;
	    }

	    if (!bOk) {
		(void) strcpy(format, ILN);
		ILN[0] = '\0';
		bOk = TRUE;
	    }
	}

	/* Check if default format should be used */
	if (format[0] == '\0') {
	    (void) strcpy(format, gConf.format);
	}
	if (bOk) {
	    return fetch_title(PPN, ILN, format);
	} else {
	    return FALSE;
	}
    } else {
	(void) tr_write( "%s: request too long, in process_fetch_request.\n",
	    OUF_FETCH);
	return FALSE;
    }
} /* process_fetch_request() */

/*------------------------------------------------------------------------
 * Function	: process_ffile ()
 * Purpose	: Process a fetch request file from the fetch directory
 *                
 * Method	: 
 *
 * History:
 * 22-12-2004	: michael : creation
 *
 * Notes	:
 * process_ffile is called with the full path name of a
 * file from the fetch directory
 *
 * An fopen() is done, and if that succeeds, the function
 * reads the file line by line with fgets(), sending
 * request line that it reads to process_fetch_request().
 *
 * process_ffile() keeps track of the number of lines it
 * reads, and the number of errors in the file (this is
 * returned by process_fetch_request)
 *
 * Finally, at EOF on the fetch file, process_ffile
 * prints out some statistics, and then either removes
 * the fetch file when no errors occurred, or renames the
 * file and places it in the temp directory if at least
 * one error occurred.
 *
 * It's important to note that process_ffiles doesn't
 * quit when the fetch file has errors, nor does it
 * return any values itself.
 *
 * Returns: Absolutely nothing.
 *------------------------------------------------------------------------
 */
static void
process_ffile(
	const char *fname)
{
    char request[OUF_FIELD_L + 1];
    char sbuf[BUFSIZ];
    int linecnt = 0;
    int errcnt = 0;
    FILE *fp;

    fp = fopen(fname, "r");

    if (fp == NULL) {
	(void) tr_write("%s: couldn't open fetch file %s for reading.\n", 
	    OUF_FETCH, fname);
	return;
    }

    (void) tr_write("\n");
    (void) tr_date();
    (void) tr_write("\n%s: fetch request file '%s' opened.\n", 
	OUF_FETCH, OUFStripFilename(fname));

    request[0] = request[OUF_FIELD_L] = '\0';

    while (!gStopSig && fgets(request, OUF_FIELD_L, fp)) {
	linecnt++;
	if (process_fetch_request(request) == FALSE) {
	    errcnt++;
	} else {
	    expire_title_files(0);
	}
    }
    
    /* Check if an error occurred */
    if (!feof(fp) && !gStopSig) {
	OUF_DEBUG(5,("%s: %s: I/O error\n",OUF_FETCH, OUFStripFilename(fname)));
	OUFSetError(OUF_FR_READ_ERROR);
    }

    /* Log fetch request file data */
    
    (void) tr_write("\n");
    (void) tr_date();
    (void) tr_write("\n");

    if (gStopSig)  {
	(void) tr_write("%s: SIGNAL CAUGHT\n", OUF_FETCH);
    } else {
	(void) tr_write("%s: fetch request file '%s' closed.\n", 
	    OUF_FETCH, OUFStripFilename(fname));
    }
    (void) tr_write("- Total # lines  : %d\n", linecnt);
    (void) tr_write("- Total # records: %d\n", linecnt - errcnt);

    if (errcnt) {
	(void) tr_write("- Total # errors : %d\n", errcnt);
    }

    if (gLoopcnt) {
	(void) tr_write("- Total # loops  : %d\n", gLoopcnt);
    }

    /* If we've made it to the EOF, then it makes sense to do this
     * "moving this file around" stuff.
     */

    if (feof(fp)) {
	if (errcnt) {
	    /* Rename file */
	    (void) sprintf(sbuf, "mv %s %s%s", fname, fname, ERR_EXT);
	    if (system(sbuf)) {
		(void) tr_write("%s: system {\"%s\") failed!\n",OUF_FETCH,sbuf);
		perror(OUF_FETCH);
	    } else {
		(void) tr_write("%s: %s renamed to %s%s\n", OUF_FETCH,
		    OUFStripFilename(fname), OUFStripFilename(fname), ERR_EXT);
	    }
	} else {
	    /* Delete fetch request file */
	    OUF_DEBUG(5,("%s: deleting fetch request file %s\n",
		    OUF_FETCH, OUFStripFilename(fname)));
	    (void) remove(fname);
	}
    }
    (void) fclose(fp);
    gLoopcnt = 0;
    (void) fflush(NULL); /* give everybody a friendly flush */
    return;
} /* process_ffile() */

/*------------------------------------------------------------------------
 * Function	: fetch ()
 * Purpose	: Primary work function called in the main loop
 *                
 * Method	: 
 * History:
 * 22-12-2004	: michael : creation
 *
 * Notes	:
 * fetch opens the directory containing the fetch request files with opendir()
 * and then loops over the files with readdir(), processing them
 * one by one via calls ro process_ffile.  char * lt. (link token)
 * 
 * Returns:
 * fetch() returns wth an error if it can't open the fetch directory, and
 * returns 0 there are no more files left to process, or when a signal has
 * been caught and the gStopSig global to a non-zero value.
 *------------------------------------------------------------------------
 */
static int
fetch(
	void)
{
    struct dirent *entryp;
    struct stat   statbuf;
    DIR	*fdirp;
    char fname[BUFSIZ];
    int filecnt = 0;

    OUF_DEBUG(10,("%s: Fetching...\n", OUF_FETCH));

    /* Open fetch request directory */

    fdirp = opendir((const char *) gConf.fetch_d);

    if (fdirp == NULL) {
	(void) tr_write("%s: can't open fetch directory %s\n", OUF_FETCH,
	    gConf.fetch_d);
	return OUF_ERROR;
    }

    /* Begin looping through the directory, trying to */
    /* process every non-directory file we see.       */

    while (!gStopSig && (entryp = readdir(fdirp))) {
	/* following block is the rebirth of the ".err" extension checking */
	int errextlen = strlen(ERR_EXT);
	int d_namelen = strlen(entryp->d_name);
	if (((d_namelen - errextlen)>0) &&
	    STREQ(&entryp->d_name[d_namelen-errextlen],ERR_EXT)){
                /* entryp->d_name ends in ".err", so skip it */
		(void) tr_write("%s: skipping \"%s\"\n",
		     OUF_FETCH, entryp->d_name);
		continue;
	}
	(void) sprintf(fname, "%s/%s", gConf.fetch_d, entryp->d_name);
	if (stat(fname, &statbuf)) {
	    /* output an error message and continue */
	    (void) tr_write("%s: stat failed in fetch, moving on...\n",
		 OUF_FETCH);
	    continue;
        }
        if (S_ISDIR(statbuf.st_mode)) {
	    /* hmmm, there aren't supposed to be any
	     * subdirectories in the fetch dir, but 
	     * we'll just ignore it.
	     */
	    continue;
	}
	filecnt++;
	process_ffile(fname);
    }
    (void) closedir(fdirp);

    if (gStopSig) {
	(void) fprintf(stderr, "%s: last file name was %s\n", OUF_FETCH,fname);
    } else if (!gStopSig && !filecnt && (gDebug >= 8)) {
	(void) tr_write("%s: fetch directory empty\n", OUF_FETCH);
    }

    /* overloading gStopSig this so that we can know what made the while loop
     * upstairs stop when we return
     */
    return gStopSig;
} /* fetch() */

/*------------------------------------------------------------------------
 * Function	: init_link_matchtab
 * Purpose	: Initializes the table for matching LinkType fields
 * Method	: 
 * History:
 * 08-11-2004	: michael : creation
 *
 * Notes	:
 * (a long-winded explanation of something simple follows :-))
 * init_link_matchtab is a very simple-minded routine which parses a LinkTypes
 * configuration lines from OUF.INI.  strtok() is used with a set of possible
 * delimiters which are in the delim[] character string.
 *
 * The function only (blindly) recognizes two-character parameter strings or
 * three-character strings which are preceded by a minus character.
 *
 * The parsed strings get shoved into a static array, and in the case of the
 * three character strings, a "sign" field gets set to -1
 *
 * The static variable glinkTableSize is the size of the array.
 * glinkTableSize is also a sentinel value, if it is -1, then we know that
 * there is nothing in the Link table, and that all links in the requests
 * will be ignored.
 *
 * Returns: nothing
 *------------------------------------------------------------------------
 */
static void
init_link_matchtab(
	char *s)
{
    static char delim[] = "\t;|/!, "; /* strtok delimiters -> overkill :-) */
    char *pos = strtok(s, delim);

    OUF_DEBUG(11,("\n%s: entering init_link_matchtab\n", OUF_FETCH));

    if (pos != NULL) {
	/* a do-while can be used here as we must make at least one pass */
	do {
	    switch(strlen(pos)) {
	    case 2:
		glinkTableSize++;
		gLink_match_table[glinkTableSize].sign = 1;
		(void) strcpy(gLink_match_table[glinkTableSize].s, pos);
		OUF_DEBUG(11,("%s: added LX pattern: \"%s\"\n",OUF_FETCH, pos));
		break;
	    case 3:
		if (*pos == '-') {
		    glinkTableSize++;
		    gLink_match_table[glinkTableSize].sign = -1;
		    (void) strcpy(gLink_match_table[glinkTableSize].s, pos + 1);
		    OUF_DEBUG(11,("%s: added LX pattern: \"%s\"\n", 
			OUF_FETCH, pos));
		} else {
		    OUF_DEBUG(11,("%s: ignoring bad LX pattern: \"%s\"\n",
			OUF_FETCH, pos));
		}
		break;
	    default:
		OUF_DEBUG(11,("%s: ignoring bad LX pattern: \"%s\"\n",
		    OUF_FETCH, pos));
		break;
	    }
	} while (((pos = strtok(NULL,delim)) != NULL) && (glinkTableSize < OUF_MAX_LINKTAB));
    }
    OUF_DEBUG(11,("%s: leaving init_link_matchtab, glinkTableSize = %d\n\n",
	OUF_FETCH, glinkTableSize));
} /* init_link_matchtab() */

/*------------------------------------------------------------------------
 *  Function	: OUFSetError
 *  Purpose	: Sets error code
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 07-10-1996
 *
 *  History:
 *  07-10-1996	: peter jan: creation
 *
 *  Returns	: 
 *  void
 *------------------------------------------------------------------------
 */
void
OUFSetError(
	int ouferrno)
{
    if (!gConf.bErrorSet) {
	gConf.ouferrno = ouferrno;
	gConf.bErrorSet = TRUE;
    }
} /* OUFSetError() */

/*------------------------------------------------------------------------
 * Function	: process_ptable
 * Purpose	: process elements of the PPN table
 * Method	: 
 *
 * History:
 * 29-11-2004	: michael : creation
 *
 * Notes:  NMTCnt is called to set tablecnt to the number of entries the PPN
 *          table. tabelcnt must be reset to zero before the counting begins
 *          :-)
 *      
 *          We used to use that value to decide whether or not we should try
 *          to expire the table or not, but that same to be a waste of
 *          resources to go through the table three times when we  aren't
 *          debugging.
 *      
 *          There should be (and probably is) a quicker way of determining
 *          whether the table is empty or not, but I don't know what it is.
 *
 * Reutrns: It is void         
 *------------------------------------------------------------------------
 */
static void
process_ptable(
	void)
{
    int tablecnt = 0;
    time_t now;

    if (gDebug >= 6) {
	(void) NMTWalk(gPPN_XTab, NMTCnt, &tablecnt);
	if (tablecnt) {
	    (void) tr_write("%s: before PPN expire %d entries\n", OUF_FETCH,
		 tablecnt);
	}
    }

    (void) time(&now);
    (void) NMTWalk(gPPN_XTab, expire_ppn, &now);

    if (gDebug >= 6) {
	tablecnt = 0;
	(void) NMTWalk(gPPN_XTab, NMTCnt, &tablecnt);
	if (gDebug >= 6) {
	    (void) tr_write("%s: after PPN expire %d entries\n", OUF_FETCH,
		 tablecnt);
	}
    }
} /* process_ptable() */

/*------------------------------------------------------------------------
 *  Function	: sighandler
 *  Purpose	: like he name says, this is a signal hanlder 
 *  Method	: 
 *
 *  History:
 *  21-01-2005	: michael  : creation
 *
 *  Notes	:
 *  this function sets the volatile global gStopSig to a non-zero values,
 *  which various other parts of the program logic check for.
 *
 *  Returns	: nothing :-)
 *------------------------------------------------------------------------
 */
static void
sighandler(
	int signo)
{
    /* set to non-zero, which is what we want */
    gStopSig = signo;
}  /* sighandler() */

/*------------------------------------------------------------------------
 *  Function	: init_ttab()
 *  Purpose	: init_ttab() initializes the hash table used for handling
 *                the (per-ILN) title files created, written and moved
 *                around by the fetch process.
 *  Method	: 
 *
 *  History:
 *  30-11-2004	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	: HSHTHandle_p to the newly created hash table
 *------------------------------------------------------------------------
 */
static HSHTHandle_p
init_ttab(
	void)
{
    return gTHTab = HSHTCreate(NULL, NULL);
} /* init_ttab() */

/*------------------------------------------------------------------------
 *  Function	: init_ptab
 *  Purpose	: Initialize the PPN (exclusion) table
 *  Method	: 
 *  Author(s)	: Michael D. Lyons
 *  Creation	: 25-11-2004
 *
 *  History:
 *  25-11-2004	: michael : creation
 *
 *  Notes	: Address of the table is placed in global gPPN_XTab
 *
 *  Returns	: NMTHandle_p to the newly created name table
 *------------------------------------------------------------------------
 */
static NMTHandle_p
init_ptab(
	void)
{
    return gPPN_XTab = NMTCreate(NULL, NMT_NOFUNC);
} /* init_ptab() */

/*------------------------------------------------------------------------
 * Function	: init_conf
 * Purpose	: Initializes OUF gConf
 * Method	: 
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 07-08-1996
 *
 * History:
 *
 * 07-08-1996	: peter jan: creation
 *
 * 28-12-2001  : evd added bCBS4inuse
 *
 * 04-07-2003	: huug	   : SRC option in \SEL \UPD command
 *
 * 27-12-2004	: michael : at various times, added new variables for Title
 *                         file gConf, stdio buffering, etc. etc. Comments
 *                         can be found in the "appropriate" places. 
 *
 * 07-01-2005	: michael : stdio buf options gone again, new sections added
 *                         to conf file, name of function changed from
 *                         OUFInitConf or whatever it was. :-)
 * 18-02-2005	: michael : added support for Title Level Header option
 *
 * Returns	: bool
 *------------------------------------------------------------------------
 */
static bool
init_conf(
	void)
{
    const char *confdir;
    char        field[OUF_FIELD_L + 1];
    FILE        *pfp;
    char	inifile[BUFSIZ];

    OUF_DEBUG(2,("%s: initializing gConf...\n", OUF_FETCH));

    /* Reset gConf */
    (void) memset(&gConf, '\0', sizeof(gConf));
    gConf.ouferrno   = OUF_NO_ERROR;
    gConf.bErrorSet  = FALSE;

    /* Get name of configuration file */
    confdir = ENVGetVar("DOTINIDIR");
    if (confdir == NULL) {
	OUFSetError(OUF_CFG_DIR_NOT_SET);
	return FALSE;
    }
    (void) sprintf(inifile, "%s/%s", confdir, OUF_INI);

    /* Open profile file */
    pfp = PFOpen(inifile, "r");
    if (pfp == NULL) {
	(void) fprintf(stderr, "%s: can't open config \"%s\"\n", OUF_FETCH,
	    inifile);
	OUFSetError(OUF_PROFILE_NOT_OPENED);
	return FALSE;
    }
    if (!CBS_init(pfp, &gCBS)) {
	(void) tr_write("%s: initialization of CBS variables failed.\n",
	     OUF_FETCH);
	return FALSE;
    }

    /* Init trace settings, note that gDebug global is  */
    /* actually set (or not) within this function call. */

    if (!tr_init(pfp, &gDebug)) {
        (void) tr_write("%s: initialization of trace variables failed\n",
	     OUF_FETCH);
        return FALSE;
    }

    /* Initialize Fetch Request Data */
    (void) PFRead(pfp, "[FetchRequest]", "Format", field);
    if (field[0] != '\0') {
        (void) PM_strncpy(gConf.format, field, OUF_FORMAT_L+1);
    } else {
        (void) PM_strncpy(gConf.format, OUF_DEF_FORMAT, OUF_FORMAT_L+1);
    }

    (void) PFRead(pfp, "[FetchRequest]", "Expansion", field);
    if (field[0] != '\0') {
        (void) PM_strncpy(gConf.expansion, field, OUF_EXPANSION_L+1);
    } else {
        (void) PM_strncpy(gConf.expansion, OUF_DEF_EXPANSION, OUF_EXPANSION_L+1);
    }

    (void) PFRead(pfp, "[FetchRequest]", "Loop Check Interval", field);
    gConf.nLoopCheckInterval = atoi(field);
    if (gConf.nLoopCheckInterval == 0) {
	gConf.nLoopCheckInterval = OUF_DEF_LOOP_CHECK_TIMEOUT;
    }
    (void) PFRead(pfp, "[FetchRequest]", "Max PPN Retrieval Freq", field);
    gConf.nMaxPPNRetrievalFreq = atoi(field);
    if (gConf.nMaxPPNRetrievalFreq == 0) {
        gConf.nMaxPPNRetrievalFreq = OUF_DEF_MAX_RETR_FREQ;
    }
    (void) PFRead(pfp, "[FetchRequest]", "Polling Interval", field);
    gConf.nFetchRequestPoll = atoi(field);
    if (gConf.nFetchRequestPoll == 0) {
        gConf.nFetchRequestPoll = OUF_DEF_FR_POLL;
    }

    /* stuff for our linked record processing and request files */

    (void) PFRead(pfp, "[FetchLinks]", "Maximum Time", field);
    gConf.nMaxFetchTime = atoi(field);
    if (gConf.nMaxFetchTime == 0) {
        gConf.nMaxFetchTime = OUF_DEF_MAX_FETCHTIME;
    }

    (void) PFRead(pfp, "[FetchLinks]", "Maximum Records", field);
    gConf.nMaxFetchRecords = atoi(field);
    if (gConf.nMaxFetchRecords == 0) {
        gConf.nMaxFetchRecords = OUF_DEF_MAX_FETCHRECORDS;
    }

    (void) PFRead(pfp, "[FetchLinks]", "LinkTypes", field);
    if (*field) init_link_matchtab(field);

    /* Initialize directories */
    (void) PFRead(pfp, "[Directories]", "Temp Dir", field);

    if (!*field) {
	/* Uh-oh, someone forgot to set the Temp Dir
	 * There is no default. There's also no way
	 * to tell whether it's just "Temp Dir" that's 
	 * missing, or whether the entire [Directories]
	 * section is broken.  -michael
	 */

	(void) tr_write("%s: configuration error: Temp Dir was not set\n",
	    OUF_FETCH);
	return FALSE;
    }

    (void) PM_strncpy(gConf.temp_d, field, OUF_DIR_L+1);

    (void) PFRead(pfp, "[Directories]", "Fetch Request Dir",field);

    if (!*field) {

	/* same thing as above for Temp Dir */
	/* (but [Directories] does exist)   */

	(void) tr_write("%s: config error: Fetch Request Dir was not set\n",
	    OUF_FETCH);
	return FALSE;
    }

    (void) PM_strncpy(gConf.fetch_d, field, OUF_DIR_L+1);

    (void) PFRead(pfp, "[Directories]", "Title Dir", field);
    (void) PM_strncpy(gConf.title_d, field, OUF_DIR_L+1);

    if (!*field) {
	/* same thing as above for Temp Dir */
	(void) tr_write("%s: configuration error: Title Dir was not set\n",
	    OUF_FETCH);
	return FALSE;
    }

    /* Title File Stuff */

    (void) PFRead(pfp, "[TitleFile]", "Maximum Size", field);

    gConf.nMaxTitleBytes = atoi(field);
    if (gConf.nMaxTitleBytes <= 0) {
        gConf.nMaxTitleBytes = OUF_DEF_MAX_TITLEBYTES;
    }

    (void) PFRead(pfp, "[TitleFile]", "Maximum Time", field);

    gConf.nMaxTitleAge = atoi(field);
    if (gConf.nMaxTitleAge <= 0) {
        gConf.nMaxTitleAge = OUF_DEF_MAX_TITLEAGE;
    }
    gConf.nMaxTitleAge *= 60; /* need this in seconds internally :-) */

    (void) PFRead(pfp, "[TitleFile]", "Title Filename Pattern", field);

    (void) PM_strncpy(gConf.titlepatt, field, OUF_MAX_TITLEPATTERN);

    (void) PFRead(pfp, "[TitleFile]", "Header Suppression Level", field);
    gConf.HeaderSupprLev = atoi(field);
    if ((gConf.HeaderSupprLev < 0) || (gConf.HeaderSupprLev > 2)) {
	int temp = gConf.HeaderSupprLev;
	gConf.HeaderSupprLev = ((gConf.HeaderSupprLev < 0) ? 0 : 2);
	(void) tr_write(
	    "%s: unsupported Title Header Level %d changed to %d\n",
	    OUF_FETCH, temp, gConf.HeaderSupprLev);
    }

    /* Close profile file */
    if (!PFClose(pfp)) {
        OUFSetError(OUF_PROFILE_NOT_CLOSED);
        return FALSE;
    }

    OUF_DEBUG(2,("%s: gConf initialized.\n", OUF_FETCH));
    return TRUE;
} /* init_conf() */

/*------------------------------------------------------------------------
 * Function	: fetch_init_all
 * Purpose	: Initializes OUF module
 * Method	: 
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 06-08-1996
 *
 * History:
 * 06-08-1996	: peter jan: creation
 * 17-12-2004	: michael  : renamed from OUFInit, also added check_fetch_dirs()
 *                           init_ttab() and init_ptab().
 *
 * 17-01-2005	: michael  : renamed to fetch_init_all to get rid of tags
 *                           conflict with ouf_upd.c
 *
 * Returns	: FALSE if one the functions that it calls failed,
 * TRUE otherwise.
 *------------------------------------------------------------------------
 */
static bool
fetch_init_all(
	void)
{
    if (!init_conf()) {
	(void) fprintf(stderr, "%s: init_conf() failed!\n", OUF_FETCH);
	return FALSE;
    }
    if (!check_fetch_dirs()) {
	(void) tr_write("%s: check_fetch_dirs() failed!\n", OUF_FETCH);
	return FALSE;
    }
    if (!setup_connection(&gCBS)) {
	(void) tr_write("%s: failed to set up CBS connection!\n", OUF_FETCH);
	return FALSE;
    }
    if (!select_database(&gCBS)) {
	(void) tr_write("%s: select_database() failed!\n", OUF_FETCH);
	return FALSE;
    }
    if (!init_ptab()) {
	(void) tr_write("%s: init_ptab() failed!\n", OUF_FETCH);
	return FALSE;
    }
    if (!init_ttab()) {
	(void) tr_write("%s: init_ttab() failed!\n", OUF_FETCH);
	return FALSE;
    }
    return TRUE;
} /* fetch_init_all() */

/*------------------------------------------------------------------------
 *  Function	: siginit
 *  Purpose	: initializes the signal handle
 *  Method	: 
 *
 *  History:
 *  21-01-2005	: michael  : creation (i.e. theft from Hans' library code)
 *
 *  Notes	:
 *
 *  Returns	: nothing at all.
 *------------------------------------------------------------------------
 */
static void
siginit(
	void) 
{
    sigset_t            mask;
    struct sigaction    act;

    (void) sigemptyset(&mask);
    act.sa_handler = sighandler;
    act.sa_mask = mask;
#if defined(SA_RESTART)
    act.sa_flags = SA_RESTART;
#else
    act.sa_flags = 0;
#endif
    (void) sigaction(SIGINT, &act, NULL);
    (void) sigaction(SIGTERM, &act, NULL);
}  /* siginit() */

/*------------------------------------------------------------------------
 *  Function	: main
 *
 *  History:
 *  28-12-2004	: michael
 *  17-01-2005	: michael : added PID file creation to support start/stop
 *                          scripts
 *
 *  17-01-2005  : michael : introduced OUF_FETCH #define
 *
 *  Notes: This is the main function of The Great Rewrite. main () used 
 *         to live in a different file, and called a function named   
 *         OUFApplication.  Not any more, though :-)                   
 *
 * Returns: EXIT_SUCCESS or EXIT_FAILURE, depending on whether it was
 * successful or failed ;-)
 *------------------------------------------------------------------------
 */
int
main(
	int argc,
	char **argv)
{
    if (!ENVInit(argc, argv, "SVN ident $Id: ouf_fetch.c 364642 2019-02-06 11:20:33Z vaneckh $", NULL, ENV_NOFUNC)) {
        ERWrite(ENVerrno, "main.ENVInit", NULL);
        return EXIT_FAILURE;
    }

    siginit();

    if (!fetch_init_all()) { 
	(void) fprintf(stderr, "%s: failed to init configuration\n", OUF_FETCH);
	close_connection(&gCBS);
        return EXIT_FAILURE;
    }

    /* main loop - if we get stopped by a signal that we catch, then */
    /* gStopSig will actually be set to the value of the signal   */

    while (!gStopSig) {
	if (fetch()) {
	    (void) tr_write("%s: fetch(): returned non-zero.\n", OUF_FETCH);
	} else {
	    process_ptable();
	    expire_title_files(0);
	    (void) fflush(NULL); /* fflush all stdio streams before bedtime */
	    expire_linkfetch(&gFetch);

	    OUF_DEBUG(7,("%s: sleeping %d secs\n",
		OUF_FETCH, gConf.nFetchRequestPoll));
	
	    if (!gStopSig) {
		(void) sleep(gConf.nFetchRequestPoll);
	    }
	}
    }

    /* everything after this point is cleaning up after a signal */

    (void) tr_write("%s: caught signal %d, exiting...\n", OUF_FETCH, gStopSig);
    (void) fflush(NULL);
    expire_linkfetch(&gFetch);
    (void) HSHTWalk(gTHTab, expire_title, NULL);
    close_connection(&gCBS);
    return EXIT_SUCCESS;
} /* main() */
