/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ouf_trace.c
 *  Description	: routines for TRACE module
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 23-04-1996
 *
 *  History:
 *  23-04-1996	: peter jan      : creation
 *  05-08-2013	: sieburgw : removed unused function
 *			    - TRACEGetLastError
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ouf_trace.c 372153 2019-11-21 13:52:14Z heesakkm $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <env.h>
#include <stdarg.h>
#include <cs.h>
#include <er.h>
#include <p3.h>
#include <nmt.h>
#include <hsht.h>
#include <picatypes.h>
#include <picamisc.h>
#include "ouf_pf.h"
#include "ouf_trace.h"
#include "ouf.h"

/* Global variables */
static TRACE_t TRSettings = 
{
    FALSE, FALSE, FALSE, FALSE, "", TRACE_NO_ERROR
};

/*------------------------------------------------------------------------
 *  Function	: tr_date
 *  Purpose	: Writes current date and time to file
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 06-06-1996
 *
 *  History:
 *  06-06-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: void
 *
 *------------------------------------------------------------------------
 */
bool
tr_date(
	void)
{
    char   strDateTime[TRACE_DATETIME_L+1];
    time_t CurTime;

    /* Get current date and time */
    if (time(&CurTime) == -1) {
	return FALSE;
    }

    /* Put time into string */
    if (strftime(strDateTime, TRACE_DATETIME_L, "[%d-%m-%Y, %H:%M:%S] ",
	localtime(&CurTime)) == 0) {
	return FALSE;
    }

    /* Write date and time to trace file */
    return tr_write("%s", strDateTime);
} /* tr_date() */

/*------------------------------------------------------------------------
 *  Function	: tr_write
 *  Purpose	: writes line of text to screen and trace file, 
 *                only if tracing is enabled
 *  Method	: 
 *
 *  History:
 *  23-04-1996	: peter jan: creation
 *  23-12-2004	: michael  : noticed while running a truss on ouf that it
 *                           opening and closing the trace file for every
 *                           single line ouf output. I have no idea why
 *                           this was done, and have now made the appropriate
 *                           changes. :-)  Also changed fprintf to fputs, since
 *                           there is never any format processing to be done.
 *
 *  23-12-2004	: michael  : added a separator line for when the trace file
 *                           is being appended to instead of replaced
 *
 *  14-01-2005	: michael  : last minute change! :-)  tr_write() is now a
 *                           varargs function. It has a large internal buffer
 *                           and does a vsnprintf() into it.
 *                 
 *                           The trace output to the trace file and the screen
 *                           are both just an fputs() of this buffer now.
 *
 *                           The intended output might be truncated because
 *                           of the vsnprintf() call, but anyone trying to
 *                           print a trace message that large deserves to
 *                           be punished, anyway. :-)
 *
 *  14-01-2005	: michael  : Now using CS library functions.
 *
 *  Notes	: 
 *
 *  Returns	: bool
 *
 *------------------------------------------------------------------------
 */

bool
tr_write(
	const char *fmt,
	...)
{
    va_list ap;

    static CSBuf_p csbuf;

    static FILE *fpTraceFile = NULL;

    if (!csbuf) csbuf = CSCreate();

    va_start (ap, fmt);

    (void) CSVPrintf (csbuf, fmt, ap);

    va_end (ap);

    /* Check trace flag in settings */
    if (TRSettings.bTrace) {
        /* Check if tracing module is initialized */
        if (!TRSettings.bInitialized) {
            TRSettings.nError = TRACE_NOT_INITIALIZED;
            return FALSE;
        }

	/* Check screen flag */
	if (TRSettings.bScreen) {
	    /* Write to screen */
	    (void) fputs (CSContents (csbuf), stdout);
	}

	/* Write to trace file */

	/* tr_init() has set the value of bReplace */
	/* to something or other. We do the rest :-) */

	if (!fpTraceFile) {
	    fpTraceFile = fopen(TRSettings.tracefile,
		    TRSettings.bReplace ? "w" : "a+");
	    if (fpTraceFile) {
		/* Tracefile is line buffered */
		(void) setvbuf (fpTraceFile, NULL, _IOLBF, BUFSIZ);
		if (!TRSettings.bReplace) {
		    (void) fputs("\n***********************\n\n", fpTraceFile);
		}
	    }
	}

	if (fpTraceFile) {
	    (void) fputs(CSContents (csbuf), fpTraceFile);
	} else {
	    TRSettings.nError = TRACE_FILE_NOT_OPENED;
	    return FALSE;
	}
    }
    return TRUE;
} /* tr_write() */

/*------------------------------------------------------------------------
 *  Function	: tr_init
 *  Purpose	: Initializes tracing module
 *
 *  Author(s)	: Michael D. Lyons
 *  Creation	: December 2004
 *
 * This is a rewrite of tr_init() incorporating pieces of the 
 * original version plus code surgically removed from ouf.c and
 * its children, plus some new code.   -michael               
 *
 * Brought back the idea of setting the debug level here. Also
 * now overloading the Debug parameter to take ON, OFF or num.
 *
 *  18-01-2005	: michael  : Removed fallback to /tmp if the Trace directory
 *                           is somehow broken. The program just dies now
 *
 *  Returns	: bool
 *
 *------------------------------------------------------------------------
 */

bool
tr_init (
	FILE *pfp,
	int *debuglev)
{
    char sfield[BUFSIZ];
    char sbuf[BUFSIZ];
    char tracebuf[BUFSIZ];

    /* Initialize Trace Data */

    TRSettings.bTrace = TRUE;
    TRSettings.bInitialized = TRUE;
    TRSettings.nError = TRACE_NO_ERROR;

    (void) PFRead(pfp, "[Trace]", "Trace", sfield);
    if (STREQ (sfield, "ON")) {
        (void) PFRead(pfp, "[Trace]", "Dir", sfield);

	if (!dir_exists (sfield, OUF_TRACE_DIR_NOT_FOUND,
	OUF_TRACE_DIR_PROTECTED, OUF_TRACE_DIR_NOTDIR)) {
	    /* the directory didn't exist, and we were unable to */
	    /* create it. There was a fallback to /tmp here, but */
	    /* now we just give up. -michael                     */
	    return FALSE;
	}

	/* Added the Tracefile option while debugging; it seems */
	/* to bo non-lethal to leave it in. :-)  -michael       */

	*sbuf = 0;

        (void) PFRead(pfp, "[Trace]", "Tracefile", sbuf);

	(void) memset ((void *) tracebuf, '\0', sizeof (tracebuf));

	if (*sbuf) {
	    (void) sprintf (tracebuf, "%s/%s", sfield, sbuf);
	} else {
	    (void) OUFConstructTimeStampedFilename(tracebuf,
	       sfield, OUF_TRACE_FILE);
	}

	if ((int) strlen (tracebuf) > TRACE_FILE_L) {
	    TRSettings.nError = TRACE_FILENAME_TOO_LONG;
	    return FALSE;
	}

	(void) PM_strncpy(TRSettings.tracefile, tracebuf, TRACE_FILE_L);
	TRSettings.tracefile[TRACE_FILE_L] = '\0';

        (void) PFRead(pfp, "[Trace]", "Replace", sfield);
        TRSettings.bReplace = STREQ(sfield, "YES") ? TRUE : FALSE;
    
        (void) PFRead(pfp, "[Trace]", "Screen", sfield);
        TRSettings.bScreen = STREQ(sfield, "ON") ? TRUE : FALSE;

        (void) PFRead(pfp, "[Trace]", "Debug", sfield);

	if (STREQ(sfield, "") || STREQ(sfield, "ON")) {
	    *debuglev = TRACE_DEFAULT_DEBUGLEV;
	} else if (STREQ(sfield, "OFF")) {
	    *debuglev = 0;
	} else {
	    *debuglev = atoi (sfield);
	}
    }
    return TRUE;
} /* tr_init () */
