/*
 * Copyright (c) 2005 by Pica
 *
 * File : ouf_upd.c
 * Description  : routines for new OUF Update signal processing module
 * Author       : Michael D. Lyons
 * Creation     : 30-12-2004
 *
 * History:
 * 17-01-2005   : michael : Replaced few remaining sprintf calls with snprintf
 *  23-02-2007	: clemens  : fixed problems with/cleaned-up array sizes
 *
 * Notes:
 * This module was originally based on ouf.c, which will presumably still be
 * around somewhere, so I've removed the ouf.c development history from here.
 *
 * The few functions from 1996-2004 from ouf.c that are more or less intact (a
 * lot of variables) have been renamed for consistency's sake) still
 * have their original comment headers.
 *
 * The changes made to ouf are documented in the "OUF User Manual" and the
 * release notes.
 *========================================================================
 */

#pragma ident "$Id: ouf_upd.c 337014 2016-08-05 16:10:14Z heesakkm $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include <pica.h>
#include <picamisc.h>
#include <env.h>
#include <cs.h>
#include <er.h>
#include <p3.h>
#include <nmt.h>
#include <hsht.h>
#include <picatypes.h>
#include "ouf_pf.h"
#include "ouf_trace.h"
#include "ouf.h"

static CBS_t gCBS;

/* Type definitions */

/* gConf: main global struct for keeping track of very very very
 * important things. :-)
 */
static struct
{
    int  OUFerrno;
    bool bErrorSet;

    /* Directories */
    char temp_d[OUF_DIR_L + 1];
    char upd_d[OUF_DIR_L + 1];

    /* Files */
    char inifile[OUF_FILE_L + 1];

} gConf;

/* gUPD contains most of the important counters, timers, limits, whatever
 * that are share amongst various routines in ouf_upd.
 */
static struct
{
    /* Update Signals */
    char lasttime[OUF_DATETIME_L + 1];  /* Earliest timestamp */
    char newtime[OUF_DATETIME_L + 1];   /* Most recent timestamp */
    long lMaxRecords;
    int  nMaxTime;
    int  nUpdateSignalPoll;
    char source[OUF_SOURCE_L + 1]; /* update source (SRC) e.g. "ONLINE" */
    char library[OUF_LIBRARY_L + 1]; /* ilns update signals (BIB) */
    char bylib[OUF_BYLIB_L + 1]; /* filter on ilns update signals (DOO) */
    char rectype[OUF_RECTYPE_L + 1]; /* recordtype update signals (REC) */
    bool   bNewFile;
    long   lTotalRecords;
    char   fname[OUF_FILE_L + 1];
    time_t timestart;
    FILE   *fp;
} gUPD;

/* gStopSig is a global set by the signal handler  */
/* so that other parts of the program know that it */
/* is time to shut down. In particular, the main   */
/* processing loop always checks gStopSig.         */

static volatile int gStopSig = 0;

/*------------------------------------------------------------------------
 * Function	: check_upd_dirs
 * Purpose	: Checks the existence of required directories
 * Method	: 
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 07-10-1996
 *
 * History:
 * 07-10-1996	: peter jan: creation
 * 28-10-2004	: michael  : Completely changed from revision before
 *                          last. Now based around dir_exists calls
 *
 * Returns	: 
 * bool
 *--------------------------------------------------------------------*/
static bool
check_upd_dirs (
	void)
{
    OUF_DEBUG(1, ("%s: checking for required directories...\n", OUF_UPD));

    if (!dir_exists (gConf.upd_d, OUF_UPDSIG_DIR_NOT_FOUND,
	OUF_UPDSIG_DIR_PROTECTED, OUF_UPDSIG_DIR_NOTDIR)) {
	return FALSE;
    }

    if (!dir_exists (gConf.temp_d, OUF_TEMP_DIR_NOT_FOUND,
	OUF_TEMP_DIR_PROTECTED, OUF_TEMP_DIR_NOTDIR)) {
	return FALSE;
    }

    OUF_DEBUG(1, ("%s: required directories checked.\n", OUF_UPD));
    return TRUE;
} /* check_upd_dirs () */

/*------------------------------------------------------------------------
 * Function	: OUFSetError
 * Purpose	: Sets error code
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 07-10-1996
 *
 * History:
 * 07-10-1996	: peter jan: creation
 *
 * Returns	: 
 * void
 *------------------------------------------------------------------------
 */
void
OUFSetError (
	int errval)
{
    if (!gConf.bErrorSet) {
	gConf.OUFerrno = errval;
	gConf.bErrorSet = TRUE;
    }
} /* OUFSetError () */

/*------------------------------------------------------------------------
 * Function	: write_lasttime
 * Purpose	: Update settings in profile file
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 02-10-1996
 *
 * History:
 * 02-10-1996	: peter jan: creation
 * 13-01-2005	: michael: changed name from OUFUpdateSettings
 *
 * Returns	: 
 * bool
 *------------------------------------------------------------------------
 */
static bool
write_lasttime (
	void)
{
    OUF_DEBUG(10, ("%s writing \"Last Time\" to config file...\n", OUF_UPD));

    /* Initialize Update Signal Data */
    if (!PFUpdate(gConf.inifile, "[UpdateSignal]", "Last Time", gUPD.newtime)){
	OUFSetError(OUF_PROFILE_NOT_UPDATED);
	return FALSE;
    }
    OUF_DEBUG (10, ("%s: \"Last Time\" written.\n", OUF_UPD));
    return TRUE;
} /* write_lasttime () */

/*------------------------------------------------------------------------
 * Function	: expire_update_file
 * Purpose	: Checks update signal criteria
 *               and takes actions if needed
 * Method	: 
 *
 * History:
 * 30-09-1996	: peter jan: creation
 * 11-01-2005	: michael: removed strTimeStamp[], which wasn't being used
 *
 * Returns	: 
 * bool
 *------------------------------------------------------------------------
 */
static bool
expire_update_file(
	bool bOverride)
{
    char   command[BUFSIZ];
    int    nTotalMinutes;
    time_t now;

    /* Check new file flag first */
    if (gUPD.bNewFile) return TRUE;

    /* Check criteria for end of current update signal file */
    now = time(NULL);
    nTotalMinutes = (int) (difftime(now, gUPD.timestart) / 60.0);
    if ((gUPD.lTotalRecords >= gUPD.lMaxRecords)
        || (nTotalMinutes >= gUPD.nMaxTime)
	|| bOverride) {
	    /* Log update signal file data */
	    (void) tr_date();
	    (void) tr_write ("\n%s: update signal file '%s' closed.\n",
	       OUF_UPD, OUFStripFilename(gUPD.fname));
	    (void) tr_write ("- Total # records   : %ld\n", gUPD.lTotalRecords);
	    (void) tr_write ("- Oldest timestamp  : %s\n", gUPD.lasttime);
	    (void) tr_write ("- Youngest timestamp: %s\n", gUPD.newtime);

	    /* Set oldest timestamp to youngest timestamp */
	    (void) strcpy (gUPD.lasttime, gUPD.newtime);

	    /* Move file from temp dir to update signal dir */
	    (void) sprintf(command, "mv %s %s", gUPD.fname, gConf.upd_d);
	    (void) system(command);

	    /* Update last time in profile file */
	    (void) write_lasttime();

	    /* Activate new file flag */
	    gUPD.bNewFile = TRUE;
    }
    return TRUE;
} /* expire_update_file() */

/*------------------------------------------------------------------------
 * Function	: close_update_file
 * Purpose	: Closes update signal file
 * Method	: 
 *
 * History:
 * 30-09-1996	: peter jan: creation
 *
 * Returns     :
 * bool
 *------------------------------------------------------------------------
 */
static bool
close_update_file(
	void)
{
    /* Check new file flag */
    if (gUPD.bNewFile) return TRUE;

    /* Close update signal file */
    if (fclose (gUPD.fp) != 0) {
	OUFSetError (OUF_US_FILE_NOT_CLOSED);
	return FALSE;
    }
    return TRUE;
}  /* close_update_file() */

/*------------------------------------------------------------------------
 * Function	: open_update_file
 * Purpose	: Opens update signal file
 * Method	: 
 *
 * History:
 * 30-09-1996	: peter jan: creation
 * 01-11-1996  : peter jan: timestamped filename extended
 *
 * Returns     :
 * bool
 *------------------------------------------------------------------------*/
static bool
open_update_file(
	void)
{
    FILE *fp;

    /* Reset update signal file pointer */
    gUPD.fp = NULL;

    /* Check if a new update signal file should be created */
    if (gUPD.bNewFile) {
        /* Construct update signals file name */
	(void) OUFConstructTimeStampedFilename(gUPD.fname,
	    gConf.temp_d, OUF_UPDATE_SIGNAL_FILE);

        /* Open update signals file */
        fp = fopen(gUPD.fname, "w");
        if (fp == NULL) {
	    OUFSetError(OUF_US_FILE_NOT_OPENED);
	    return FALSE;
        }

	/* Log update signal file data */
	(void) tr_date();
	(void) tr_write ("%s: update signal file '%s' created.\n",
	    OUF_UPD, OUFStripFilename(gUPD.fname));
	 
        /* Reset new file flag */
	gUPD.bNewFile = FALSE;

	/* Reset variables used in file close criteria */
	gUPD.timestart = time(NULL);
	gUPD.lTotalRecords = 0L;
    } else {
        /* Open update signals file */
        fp = fopen(gUPD.fname, "a");
        if (fp == NULL) {
	    OUFSetError(OUF_US_FILE_NOT_OPENED);
	    return FALSE;
        }
    }
    gUPD.fp = fp;
    return TRUE;
} /* open_update_file() */

/*------------------------------------------------------------------------
 * Function	: produce_updates
 * Purpose	: Produces update signal
 * Method	: 
 *
 * History:
 * 07-08-1996	: peter jan: creation
 * 13-01-2005	: michael :changed logic in part where fprintf is located
 *               also removed unnecessary function write_update()
 *
 * Returns	: 
 * bool
 *------------------------------------------------------------------------
 */
static bool
produce_updates(
	CBBuf_t *CBbuf,
	size_t CBSize)
{
    bool       bOk;
    CBElm_p    CBElm;
    const char *attr;

    OUF_DEBUG(50, ("%s: start producing update signals...\n", OUF_UPD));

    /* Get first database update */
    CBElm = CBGetAttribute(CBbuf, CBSize, "TX", &attr);

    if (CBElm == NULL) return TRUE;

    bOk = open_update_file();

    if (bOk) {
	while (CBElm && bOk) {
	    (void) fprintf (gUPD.fp, "%s\n", attr);
	    gUPD.lTotalRecords++;
	    /* Get next database update */
	    CBElm = CBNextAttribute(CBbuf, CBSize, CBElm, &attr);
	}
	(void) close_update_file();
    }
    return bOk;
} /* produce_updates() */

/*------------------------------------------------------------------------
 * Function	: retrieve_updates
 * Purpose	: Retrieves CBS update signals
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 07-08-1996
 *
 * History:
 * 07-08-1996	: peter jan: creation
 * 04-07-2003	: huug	   : SRC option in \SEL \UPD command
 *
 * Returns	: 
 * integer
 * OUF_ERROR    : error
 * OUF_OK       : no error
 * OUF_NOT_FOUND: nothing retrieved
 *------------------------------------------------------------------------
 */
static int
retrieve_updates(
	CBBuf_t *CBbuf,
	size_t CBSize)
{
    CBElm_p    CBElm;
    char       strLast[OUF_DATETIME_L + 1];
    char       strOptions[OUF_OPTIONS_L + 1];
    const char *strAttrValue;
    char       *s;
    bool       ok;

    OUF_DEBUG(50, ("%s: scanning for database updates...\n", OUF_UPD));

    /* Construct options string for database update sel */
    strOptions[0] = '\0';

    /* 
     * source is optional
     * library is optional
     * bylib & rectype are also both optional, but only
     * have an effect when the library is specified
     */

    s = strOptions;
    if (gUPD.source[0] != '\0') {
	s += sprintf(s,"%s%s","source=",gUPD.source);
    }
    if (gUPD.library[0] != '\0') {
	/* print semicolon if strOptions is not empty */
	if (s != strOptions) s += sprintf(s,"%s",";");
	/* print library and optional bylib and rectype options */
	s += sprintf(s, "%s%s%s%s%s%s",
	    "library=", gUPD.library,
	    gUPD.bylib[0]=='\0'?"":";bylib=",
	    gUPD.bylib[0]=='\0'?"":gUPD.bylib,
	    gUPD.rectype[0]=='\0'?"":";rectype=",
	    gUPD.rectype[0]=='\0'?"":gUPD.rectype);
    }

    /* Clear counted buffer */
    if (!CBClear(CBbuf, CBSize)) {
	OUFSetError(OUF_BUFFER_NOT_CLEARED);
	return OUF_ERROR;
    }

    if (strOptions[0] == '\0') {
	if (!gCBS.bCBS4inuse) {
	    ok = P3ScanDbsUpdates(gCBS.P3Handle, gUPD.newtime,
		  strLast, OUF_MAX_UPDATES, strOptions, CBbuf, CBSize);
	} else {
	    ok = P3XLWMScanDbsUpdates(gCBS.P3Handle, gUPD.newtime,
		  strLast, OUF_MAX_UPDATES, strOptions, CBbuf, CBSize);
	}
	if (!ok) {
	    OUFSetError(OUF_P3ERROR);
	    return OUF_ERROR;
	}
	/* Save date and time of last update */
	(void) PM_strncpy(gUPD.newtime, &(strLast[4]), strlen(strLast)-6);
    } else {
        if (!gCBS.bCBS4inuse) {
	    ok = P3SelDbsUpdates(gCBS.P3Handle, gUPD.newtime,
	      strLast, OUF_MAX_UPDATES, strOptions, CBbuf, CBSize);
	} else {
	    ok = P3XLWMSelDbsUpdates(gCBS.P3Handle, gUPD.newtime,
	      strLast, OUF_MAX_UPDATES, strOptions, CBbuf, CBSize);
	}
	if (!ok) {
	    OUFSetError(OUF_P3ERROR);
	    return OUF_ERROR;
	}
	/* Save date and time of last update */
	(void) strcpy (gUPD.newtime, strLast); 
    }

    /* The trace messages below have been commented out because
     * they can flood the trace output when there is a very large
     * number of updates (e.g. from an off-line process). Plus,  
     * it's doubtful that the messages are useful. :-)   -michael 
     */

    /* okay, they're back now, but note the high gDebug level :-) */

    if (gDebug >= 50) {
	(void) tr_date();
	(void) tr_write ("%s: scanning for updates completed.\n", OUF_UPD);
    }
    CBElm = CBGetAttribute(CBbuf, CBSize, "TX", &strAttrValue);

    return CBElm ? OUF_OK : OUF_NOT_FOUND;
} /* retrieve_updates() */

/*----------------------------- -------------------------------------------
 * Function	: process_updates
 * Purpose	: Processes update signals
 * Method	: 
 *
 * History:
 * 06-08-1996	: peter jan: creation
 *
 * Notes	:
 * michael : This function is misnamed. :-) The function
 * doesn't really "process" anything, but simply retrieves updates from
 * CBS and then writes them out to the current update signal file.
 *
 * Returns	:  OUF_ERROR or the return value of retrieve_updates()
 *------------------------------------------------------------------------
 */
static int
process_updates(
	void)
{
    CBBuf_t CBbuf[OUF_BUFFER];
    int     nRetValue;

    if (!expire_update_file(FALSE)) {
	OUF_DEBUG (1, ("%s: expire_update_file failed!\n", OUF_UPD));
	return OUF_ERROR;
    }

    nRetValue = retrieve_updates(CBbuf, sizeof(CBbuf));

    if (nRetValue == OUF_OK) {
        if (!produce_updates(CBbuf, sizeof(CBbuf))) {
	    OUF_DEBUG (1, ("%s: produce_updates failed!\n", OUF_UPD));
	    return OUF_ERROR;
	}
    }
    return nRetValue;
} /* process_updates() */

/*------------------------------------------------------------------------
 *  Function	: sighandler
 *  Purpose	: signal handler
 *  Method	: sets gStopSig global to number of received signal
 *
 *  History:
 *  16-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	: nothing
 *------------------------------------------------------------------------
 */
static void
sighandler(
	int signo)
{
    /* set to non-zero */
    /* which is what we want */
    gStopSig = signo;
} /* sighandler() */
/*------------------------------------------------------------------------
 *  Function	: siginit 
 *  Purpose	: set up the signal handler
 *  Method	: sigaction(2)
 *
 *  History:
 *  16-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	: nothing
 *------------------------------------------------------------------------
 */
static void
siginit (
	void)
{
    sigset_t            mask;
    struct sigaction    act;

    (void) sigemptyset (&mask);
    act.sa_handler = sighandler;
    act.sa_mask = mask;
#if defined(SA_RESTART)
    act.sa_flags = SA_RESTART;
#else
    act.sa_flags = 0;
#endif
    (void) sigaction(SIGINT, &act, NULL);
    (void) sigaction(SIGTERM, &act, NULL);
} /* siginit () */

/*------------------------------------------------------------------------
 * Function	: init_settings
 * Purpose	: Initializes OUF settings
 * Method	: 
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 07-08-1996
 *
 * History:
 * 07-08-1996	: peter jan: creation
 * 28-12-2001  : evd added bCBS4inuse
 * 04-07-2003	: huug	   : SRC option in \SEL \UPD command
 * 27-12-2004	: michael : at various times, added new variables for Title
 *                         file settings, stdio buffering, etc. etc. Comments
 *                         can be found in the "appropriate" places. 
 *
 * Returns	: 
 * bool
 *------------------------------------------------------------------------
 */
static bool
init_settings(
	void)
{
    const char *confdir;
    char        sfield[OUF_FIELD_L + 1];
    FILE        *pfp;

    OUF_DEBUG(1, ("%s: initializing settings...\n", OUF_UPD));

    /* Reset settings */
    (void) memset(&gConf, '\0', sizeof(gConf));
    gConf.OUFerrno     = OUF_NO_ERROR;
    gConf.bErrorSet  = FALSE;
    gCBS.bConnected = FALSE;
    gCBS.bCBS4inuse = TRUE;

    /* Get name of configuration file */
    confdir = ENVGetVar("DOTINIDIR");
    if (confdir == NULL) {
	OUFSetError(OUF_CFG_DIR_NOT_SET);
	return FALSE;
    }
    /* let's hope that gConf.inifile is big enugh */
    (void) sprintf (gConf.inifile, "%s/%s", confdir, OUF_INI);

    /* Open profile file */

    pfp = PFOpen(gConf.inifile, "r");
    if (pfp == NULL) {
	(void)fprintf(stderr, "%s: PFOpen(%s) failed\n", OUF_UPD,gConf.inifile);
	OUFSetError(OUF_PROFILE_NOT_OPENED);
	return FALSE;
    }

    if (!tr_init (pfp, &gDebug)) {
	(void) fprintf (stderr,"%s: initialization of trace variables failed\n",
	    OUF_UPD);
        return FALSE;
    }

    if (!CBS_init (pfp, &gCBS)) {
        (void)tr_write("%s: initialization of CBS variables failed\n", OUF_UPD);
        return FALSE;
    }

    /* Initialize Update Signal Data */
    (void) PFRead(pfp, "[UpdateSignal]", "Last Time", sfield);
    if (sfield[0] != '\0') {
        (void) PM_strncpy(gUPD.lasttime, sfield, OUF_DATETIME_L+1);
    } else {
        (void) PM_strncpy(gUPD.lasttime, OUF_DEF_LAST_TIME, OUF_DATETIME_L+1);
    }
    (void) strcpy (gUPD.newtime, gUPD.lasttime);

    (void) PFRead(pfp, "[UpdateSignal]", "Maximum Records", sfield);
    gUPD.lMaxRecords = atol(sfield);
    if (gUPD.lMaxRecords == 0) gUPD.lMaxRecords = OUF_DEF_MAX_RECORDS;

    (void) PFRead(pfp, "[UpdateSignal]", "Maximum Time", sfield);
    gUPD.nMaxTime = atoi(sfield);
    if (gUPD.nMaxTime == 0) gUPD.nMaxTime = OUF_DEF_MAX_TIME;
    
    (void) PFRead(pfp, "[UpdateSignal]", "Polling Interval", sfield);
    gUPD.nUpdateSignalPoll = atoi(sfield);
    gUPD.nUpdateSignalPoll *= 60;

    (void) PFRead(pfp, "[UpdateSignal]", "Polling Interval Min", sfield);

    if (atoi (sfield)) {
	gUPD.nUpdateSignalPoll = atoi(sfield);
	gUPD.nUpdateSignalPoll *= 60;
    }

    (void) PFRead(pfp, "[UpdateSignal]", "Polling Interval Sec", sfield);
    gUPD.nUpdateSignalPoll += atoi(sfield);
    if (gUPD.nUpdateSignalPoll == 0) {
        gUPD.nUpdateSignalPoll = OUF_DEF_US_POLL;
	gUPD.nUpdateSignalPoll *= 60;
    }
    if (gUPD.nUpdateSignalPoll <  OUF_MIN_UPDATE_POLL) {
	gUPD.nUpdateSignalPoll = OUF_MIN_UPDATE_POLL;
    }

    (void) PFRead(pfp, "[UpdateSignal]", "Source", sfield);
    (void) PM_strncpy(gUPD.source, sfield, OUF_SOURCE_L+1);

    (void) PFRead(pfp, "[UpdateSignal]", "Library", sfield);
    (void) PM_strncpy(gUPD.library, sfield, OUF_LIBRARY_L+1);

    (void) PFRead(pfp, "[UpdateSignal]", "By", sfield);
    (void) PM_strncpy(gUPD.bylib, sfield, OUF_BYLIB_L+1);

    (void) PFRead(pfp, "[UpdateSignal]", "Recordtype", sfield);
    (void) PM_strncpy(gUPD.rectype, sfield, OUF_RECTYPE_L+1);

    /* Initialize directories */
    (void) PFRead(pfp, "[Directories]", "Temp Dir", sfield);
    (void) PM_strncpy(gConf.temp_d, sfield, OUF_DIR_L+1);

    (void) PFRead(pfp, "[Directories]", "Update Signal Dir",sfield);
    if (!*sfield) {
	(void) tr_write ("%s: no Update Signal Dir set!\n", OUF_UPD);
	return FALSE;
    }
    (void) PM_strncpy(gConf.upd_d, sfield, OUF_DIR_L+1);
    
    /* Close profile file */
    if (!PFClose(pfp)) {
        OUFSetError(OUF_PROFILE_NOT_CLOSED);
        return FALSE;
    }

    /* Log initial timestamp */
    (void) tr_write ("%s: initial timestamp: %s.\n", OUF_UPD, gUPD.lasttime);

    /* Initialize data concerning update signal file status */
    gUPD.bNewFile = TRUE;
    gUPD.lTotalRecords = 0L;
    gUPD.fname[0] = '\0';

    OUF_DEBUG(1, ("%s: settings initialized.\n", OUF_UPD));
    return TRUE;
} /* init_settings() */

/*------------------------------------------------------------------------
 * Function	: upd_init_all
 * Purpose	: Initializes OUF module
 * Method	: 
 *
 * History:
 * 06-08-1996	: peter jan: creation
 * 17-12-2004	: michael  : renamed from OUFInit, also, uh, completely
 *                changed the function. :-)))
 *
 * Returns	: 
 * bool
 *------------------------------------------------------------------------
 */
static bool
upd_init_all(
	void)
{
    if (!init_settings()) {
        (void) fprintf (stderr, "%s: init_settings() failed!\n", OUF_UPD);
        return FALSE;
    }
    if (!check_upd_dirs()) {
        (void) tr_write ("%s: check_upd_dirs() failed!\n", OUF_UPD);
        return FALSE;
    }
    if (!setup_connection(&gCBS)) {
        (void) tr_write ("%s: failed to set up CBS connection!\n", OUF_UPD);
        return FALSE;
    }
    if (!select_database(&gCBS)) {
        (void) tr_write ("%s: select_database() failed!\n", OUF_UPD);
        return FALSE;
    }
    return TRUE;
} /* upd_init_all() */

/*------------------------------------------------------------------------
 * Function	: main
 * Method	: 
 *
 * History:
 * 28-12-2004	: michael : I made it all up
 *
 * Returns	:
 * EXIT_FAILURE or EXIT_SUCCESS
 *------------------------------------------------------------------------
 */
int
main (
	int argc,
	char **argv)
{
    int retval;

    if (!ENVInit(argc, argv, "SVN ident $Id: ouf_upd.c 337014 2016-08-05 16:10:14Z heesakkm $", NULL, ENV_NOFUNC)) {
        ERWrite(ENVerrno, "main.ENVInit", NULL);
        return EXIT_FAILURE;
    }

    siginit ();

    if (!upd_init_all()) { 
	(void) fprintf (stderr, "%s: failed to init configuration\n", OUF_UPD);
	close_connection(&gCBS);
        return EXIT_FAILURE;
    }

    while (!gStopSig) {
	while((retval = process_updates()) == OUF_OK); /* main proc loop */

	switch (retval) {
	    case OUF_NOT_FOUND:
		(void) tr_write ("%s: sleeping %d secs\n", OUF_UPD,
		     gUPD.nUpdateSignalPoll);
		(void) sleep (gUPD.nUpdateSignalPoll);
		break;
	    case OUF_ERROR:
		(void) tr_write ("%s: OUF_ERROR return in main loop, exiting\n",
		    OUF_UPD);
		close_connection(&gCBS);
		return EXIT_FAILURE;
		/* the compiler cries if there's
		 * a break here, so there isn't
		 */
	    default:
		(void) tr_write(
		    "%s: unknown error %d in upd main loop. Exiting.\n",
		    OUF_UPD, retval);
		close_connection(&gCBS);
		return EXIT_FAILURE;
		/* the compiler cries here as well... */
	}
    }
    (void) tr_write ("%s: received signal %d, cleaning up and exiting\n",
	OUF_UPD, gStopSig);
    (void) expire_update_file(TRUE);
    close_connection(&gCBS);
    return EXIT_SUCCESS;
} /* main () */
