/*========================================================================
 *  Copyright (c) 2004 by OCLC PICA
 *
 *  File	: ouf_util.c
 *  Description	: 
 *  Author(s)	: Michael Lyons
 *  Creation	: 30-12-2004
 *
 *  History:
 *  30-12-2004	: michael  : creation
 *  19-10-2006  : evd      : added pica$ P$ to  OUFConvertPica3ToADBFormat
 *  05-08-2013	: sieburgw : removed unused function:
 *			    - OUFGetLastError
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ouf_util.c 360943 2018-10-23 16:48:25Z heesakkm $"

#include <pica.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <env.h>
#include <cs.h>
#include <er.h>
#include <p3.h>
#include <nmt.h>
#include <hsht.h>
#include <picatypes.h>
#include "ouf_pf.h"
#include "ouf_trace.h"
#include "ouf.h"

/*------------------------------------------------------------------------
 *  Function	: dir_exists
 *  Purpose	: Checks for existence of a directory, creating it if needed.
 *  Method	: 
 *  Author(s)	: Michael D. Lyons
 *  Creation	: 22-10-2004
 *
 *  History:
 *  22-10-2004	: michael : creation
 *  26-10-2004	: michael : changed function to Boolean
 *  27-10-2004	: michael : changed name of mydir parameter to directory
 *  28-10-2004	: michael : changed name of function to dir_exists
 *  11-01-2005	: michael : changed function BACK to int again.
 *  21-01-2005	: michael : you won't believe this, but it's bool again ;-)
 *
 *  Notes	: built to alter the behavior of check_directories
 *                dir_exists takes a directory name as a parameter,
 *                followed by a list of directory-specific error codes
 *                for non-existence of the directory, lack of write
 *                access to the directory or the pathname not being
 *                a directory.
 *
 *                "non-existence" of the directory actually means that
 *                the attempt to create the directory has somehow failed.
 *
 *                If an error occurs, then OUFSetError is called with
 *                the parameter passed to dir_exists corresponding to
 *                the error type.
 *
 *  Returns	: TRUE when dir exists, FALSE when it does not. :-)
 *------------------------------------------------------------------------
 */
bool
dir_exists(
	char *directory,
	int no_entry,
	int no_access,
	int no_dir)
{
    struct stat statbuf;
    int ret;

    errno = 0;

    ret = access(directory, R_OK | W_OK);

    /* the following casts are for lint */

    (void) no_entry;
    (void) no_access;
    (void) no_dir;

    if (ret){
        if (errno == ENOENT){

            /* directory doesn't exist, so we're going */
            /* to try to create it ourselves.          */

	    OUF_DEBUG(2, ("dir %s doesn't exist, trying to create it.\n",
		directory));

            ret = mkdir(directory, OUF_DIRMODE);

            if (ret) {
		OUF_DEBUG(2,("cannot create directory \"%s\"\n", directory));
                /* OUFSetError(no_entry); */
                return FALSE;
            } else {
		OUF_DEBUG (2, ("directory \"%s\" created.\n", directory));
                return TRUE;
            }
        } else {
	    OUF_DEBUG(2, ("dir %s/ exists, but is write-protected.\n",
		    directory));
            /* OUFSetError(no_access); */
            return FALSE;
        }
    }

    /* directory exists, and this process has read and write access, */
    /* so we just need to make certain that it is a directory        */

    ret = stat(directory, &statbuf);

    if (ret == 0) {
        if (!S_ISDIR(statbuf.st_mode)) {
            /* directory is not a directory, so we're in */
            /* in trouble. The function returns FALSE    */
            
	    OUF_DEBUG(2, ("\"%s\" is not a directory!\n", directory));
            /* OUFSetError(no_dir); */
            return FALSE;
        } else {
            return TRUE;
        }
    } else {
	/* something is terribly wrong if the stat() fails   */
	/* nothing that we can do but to bail out. Using the */
	/* no_entry stuff for this "cannot happen" case      */

	OUF_DEBUG(2, ("unable to stat \"%s\"!\n", directory));
	/* OUFSetError(no_entry); */
	return FALSE;
    }
} /* dir_exists() */

/*------------------------------------------------------------------------
 *  Function	: OUFConstructTimeStamp
 *  Purpose	: Constructs timestamp string
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 26-08-1996
 *
 *  History:
 *  26-08-1996	: peter jan: creation
 *  22-10-1997	: HSi	   : Chopped format string of non-extended format
 *			     in chunks. Old format contained sequences
 *			     processed by sccs.
 *			     Replace all spaces by zeros in non-extended
 *			     format.
 *  12-01-2004	: michael: now declared static, deleted bExtended stuff
 *
 *  Notes	: 
 *  bUseSavedTimeStamp - if TRUE, a previously saved timestamp will be
 *                       used instead of the current time.
 *
 *------------------------------------------------------------------------
 */
static void
OUFConstructTimeStamp(
	char *strTimeStamp,
	bool bUseSavedTimeStamp)
{
    time_t        TimeStamp;
    static time_t SavedTimeStamp;
    char	  *spacepos;
    
    /* Check if a previously saved timestamp must be used */
    if (bUseSavedTimeStamp) {
	TimeStamp = SavedTimeStamp;
    } else {
        (void) time(&TimeStamp);
	SavedTimeStamp = TimeStamp;
    }

    (void) strftime(strTimeStamp, OUF_DATETIME_L,
	"%Y" "%m" "%d" "%H" "%M" "%S", localtime(&TimeStamp));

    /* Remove spaces. Some formats give leading space in case of */
    /* of single digit                                           */

    spacepos = strTimeStamp;
    while ((spacepos = strchr(spacepos, ' '))) *spacepos = '0';
} /* OUFConstructTimeStamp() */

/*------------------------------------------------------------------------
 *  Function	: OUFConstructTimeStampedFilename
 *  Purpose	: Constructs timestamped filename
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-11-1996
 *
 *  History:
 *  01-11-1996	: peter jan: creation
 *  11-01-2005	: michael: got rid of pointless bFirstTime variable and logic
 *                         surrounding it. :-)
 *
 *  11-01-2005	: michael: extended length of infamous secret running counter
 *                         to 3 digits, changed string buffer sizes to BUFSIZ,
 *                         hacked sprintf to Destfname apart so that it's
 *                         possible to get just the timestamp/counter part
 *                         when Dir and Srcfname are empty.
 *
 *  20-01-2005	: michael: changed return type to char *, which makes more sense
 *
 *  Notes	:
 *  The counters for the constructed file names for a given file type may have
 *  gaps, since several other types of file names may be generated with this
 *  function.  This was the case in the orignial version as well, though,
 *  and since the counters are guaranteed to be monotonically increasing, it
 *  won't mess up and algorithms based on the names being in sequential
 *  order.
 *
 *  Returns: char * to the generated string
 *------------------------------------------------------------------------
 */
char *
OUFConstructTimeStampedFilename(
	char *Destfname,
	const char *Dir,
	const char *Srcfname)
{
    static char PreviousTime[BUFSIZ];
    char        strTime[BUFSIZ];
    static int  cnt = 0;
    char sbuf[BUFSIZ];

    OUFConstructTimeStamp(strTime, FALSE);

    if (STREQ(strTime, PreviousTime)){
	cnt++;
    } else {
	cnt = 0;
    }

    *Destfname = 0;

    if ((Dir != NULL) && (*Dir != '\0')) {
	(void) sprintf(Destfname, "%s/", Dir);
    }

    if ((Srcfname != NULL) && (*Srcfname != '\0')) {
	(void) strcat(Destfname, Srcfname);
	(void) strcat(Destfname, "_");
    }

    (void) sprintf(sbuf, "%s%03d", strTime, cnt);
    (void) strcat(Destfname, sbuf);

    /* Save current timestamp */
    (void) strcpy(PreviousTime, strTime);

    return Destfname;
} /* OUFConstructTimeStampedFilename() */

/*------------------------------------------------------------------------
 *  Function	: OUFConvertPica3ToADBFormat
 *  Purpose	: Converts Pica3 format to ADB format
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-10-1996
 *
 *  History:
 *  01-10-1996	: peter jan: creation
 *  19-10-2006  : evd : Added pica$ to rec_table
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
bool
OUFConvertPica3ToADBFormat(
	char *strDest,
	char *strSrc)
{
    bool bFound;
    int  i;

    static struct
    {
	const char  *adb_syntax;
	const char  *p3_syntax;
    } rec_table[] =
    {
	{ "pica$",            "P$" },
	{ "pica+",            "P" },
	{ "pica2",            "D" },
	{ "isbd",             "I" },
	{ "download",         "J" },
	{ "labelled",         "U" },
	{ "unimarc",          "UNM" },
	{ "usmarc",           "USM" },
	{ "mab",              "MAB" },
	{ "unimarc_exchange", "UNX" },
	{ "usmarc_exchange",  "USX" }
    };
    /* Translate format to adb syntax */
    bFound = FALSE;
    for (i = 0; (i < sizeof(rec_table)/sizeof(rec_table[0])) && !bFound; i++) {
        if (STREQ(rec_table[i].p3_syntax, strSrc)) {
            (void) strcpy(strDest, rec_table[i].adb_syntax);
	    bFound = TRUE;
        }
    }
    /* if (!bFound) OUFSetError(OUF_UNKNOWN_FORMAT); */

    return bFound;
} /* OUFConvertPica3ToADBFormat() */

/*------------------------------------------------------------------------
 *  Function	: OUFGetResultMessage
 *  Purpose	: Retrieves message related to a certain result code
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 25-09-1996
 *
 *  History:
 *  25-09-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  Pointer to a string containing the result message.
 *------------------------------------------------------------------------
 */
const char*
OUFGetResultMessage(
	int nResult,
	char *s)
{
    switch(nResult) {
	case OUF_AUTHORITY:
	    (void) strcpy (s, "Invalid authority.");
	    break;
	case OUF_DISCONNECT:
	    (void) strcpy (s, "Server disconnected.");
	    break;
	case OUF_NO_SUPPORT:
	    (void) strcpy(s, "Function not supported.");
	    break;
	case OUF_PROTOCOL:
	    (void) strcpy(s, "Protocol error.");
	    break;
	case OUF_SERVER:
	    (void) strcpy(s, "Server error.");
	    break;
	case OUF_TTLWRONG:
	    (void) strcpy(s, "Invalid title.");
	    break;
	case OUF_LOOP_DETECTED:
	    (void) strcpy(s, "Possible loop in retrieval.");
	    break;
	default:
	    (void) sprintf(s, "Unknown error: %d.", nResult);
    }
    return s;
} /* OUFGetResultMessage() */

/*------------------------------------------------------------------------
 * Function	: OUFStripFilename
 * Purpose	: Strips filename of path
 * Author(s)	: Peter Jan van der Mespel
 * Creation	: 26-08-1996
 *
 * History:
 * 26-08-1996	: peter jan: creation
 * 21-01-2005	: michael : completely and utterly rewritten
 *
 * Notes: 
 * This function bheaves like a dumb version of basename(1) With a normal
 * pathane, which is what this is expected to get as input, it will return the
 * filename portion. If the argument ends with a '/', then the original
 * argument will be returned.
 *
 * Returns	: 
 * Pointer to filename in absolute filename string
 *------------------------------------------------------------------------
 */
const char*
OUFStripFilename(
	const char *fname)
{
    char *p = strrchr(fname, '/');

    if ((p != NULL) && (strlen(p) > 1)) {
        return p + 1;
    } else {
        return fname;
    }
} /* OUFStripFilename() */

/*------------------------------------------------------------------------
 *  Function	: get_titlefname
 *  Purpose	: generates a title file name from a pattern
 *
 *  Notes       : The fmt parameter should be a "Title Filename Pattern"
 *                as read from the ouf configuration file during startup
 *         
 *                  A pattern string is allowed to consist of runs of
 *		  alphanumerical characters [a-z][A-Z][0-9], and the
 *		  hyphen, underscore and period characters.
 *
 *		  In addition three special tags are recognized in a
 *		  format string:
 *
 *		  %T will be replaced in the output by a current timestamp
 *		     in the form YYYYMMDDHHMMSSCCC. The "CCC" portion of
 *		     the timestamp is a counter maintained by the timestamp
 *		     generation routine, and is meant to provide for
 *		     uniqueness in timestamps created during the same
 *		     clock second.
 *
 *		  %P will be replaced by a five-digit representation of the
 *		     process ID of the current process with leading zeros.
 *
 *		  %I will be replaced by a four-digit representation of the
 *		     numerical iln parameter passed to get_titlefname.
 *
 *		  A tag may appear multiple times within the format string,
 *		  and "illegal" characters and tags are silently discarded.
 *
 *		  If the dir parameter is a non-empty string, then that
 *		  string will prepended to the filename that get_titlefname
 *		  generates (followed by a / character, of course :-))
 *
 *		  If the fmt string is NULL or an empty string, then
 *		  get_titlefname will generate a default file name
 *		  of the form "title_<timestamp>"
 *
 *		  Also, if a file name pattern does not contain a time-
 *		  stamp tag, then a timestamp will be automatically 
 *		  generated and appended to the generated filename
 *		  with a preceding underscore character.
 *
 *		  Some examples:
 *
 *		  (assume that the current PID is 8661, and the iln is 22)
 *
 *		  "T_%I_%T.%P" ->  T_022_20050114134546002.08661
 *
 *		  "%P%P%P%P"   ->  08661086610866108661_20050114134546002
 *
 *		  ""           ->  title_20050114134546002
 *
 *		  "hi there!"  ->  hithere_20050114134546002
 *
 *		  "&^%*%&^*&$" ->  title_20050114134546002
 *
 *		  The function returns a pointer to the internal static buffer
 *		  where the output string is built.  This string must be
 *		  copied to a user buffer if the file name is to be retained
 *		  across calls to get_titlefname.
 *
 *		  If you look at the code, you'll notice that %t, %i and %p
 *		  will also be accepted as tags; it's not certain whether
 *		  that particular feature will be documented or not :-)
 *
 *  History: 
 *  14-01-2005 : michael : creation
 *  24-01-2005 : michael : removed "if (iln)" in %I portion of switch
 *
 *  Returns	: 
 *  Pointer to filename built in an internal static buffer.
 *------------------------------------------------------------------------
 */
char *
get_titlefname(
	char *fmt,
	char *dir,
	int iln)
{

    char *p;
    static char pid[16];
    int stamped = 0;

    static char sbuf[BUFSIZ];
    char timestamp[BUFSIZ];
    char *runner = sbuf;

    if (!fmt || !*fmt) {
        (void) OUFConstructTimeStampedFilename(sbuf, dir, OUF_TITLE_FILE);
        return sbuf;
    }

    /* need to get a timestamp for later use */
    (void) OUFConstructTimeStampedFilename(timestamp, NULL, NULL);

    (void) memset(sbuf, 0, sizeof sbuf);

    if (dir && *dir) {
	runner += sprintf(runner, "%s/", dir);
    }

    if (!*pid) (void) sprintf(pid, "%04ld", (long int) getpid()); /* just this once */

    for (p = fmt; *p; p++) {
	if (*p != '%') {
	    if (isalnum(*p) || strchr("._-", *p)) {
		*runner++ = *p;
	    }
	    continue;
	}

	switch (*++p) {
	    case 'T':
	    case 't':
		stamped++;
		runner += sprintf(runner, "%s", timestamp);
		break;
	    case 'I':
	    case 'i':
		runner += sprintf(runner, "%0*d", PIC_ILN_DECIMAL_MAX, iln);
		break;
	    case 'P':
	    case 'p':
		runner += sprintf(runner, "%s", pid);
		break;
	    default:
		/* ignore */
		break;
	}
    }
    if (!stamped) runner += sprintf(runner, "_%s", timestamp);
    return sbuf;
} /* get_titlefname() */
