/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ouftest.c
 *  Description	: routines for OUFTest module
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan      : creation
 *  22-10-1997	: HSi	: Changed format for timestamp used in producing
 *			  filenames (see OUFConstructTimeStamp).
 *  12-01-1999  : EvD   : ABES other iln than user iln possible.
 *                        If iln and inst. in updatesignal file then iln in fetch file.
 *
 *  04-11-1999  : EvD   : Hard coded format P removed from fetch request
 *                        file. The format in ouf.ini defined is used now
 *  14-09-2007	: sieburgw : renamed sigset to sig_set for portablity
 *  21-04-2020	: zwaalm   : CBS-9161 Code review changes.
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ouftest.c 376284 2020-04-23 10:25:46Z zwaalm $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <env.h>
#include <er.h>
#include <p3.h>
#include <pp.h>
#include <picamisc.h>
#include <cs.h>
#include "ouf_pf.h"
#include "ouf_trace.h"
#include "ouftest_parse.h"
#include "ouftest.h"

#define OUFTEST_DEBUG(expr) if (Settings.bDebug) (void) TRACEWrite(expr)

typedef struct
{
    FILE *fpFile;
    char strFilename[OUFTEST_FILE_L+1];
} FileData;

typedef struct
{
    char strIln[OUFTEST_ILN_L+1];
    char strInstitution[OUFTEST_INSTITUTION_L+1];
    char strDateTime[OUFTEST_DATETIME_L+1];
    char strType[OUFTEST_TYPE_L+1];
    char strPPN[OUFTEST_PPN_L+1];
    int  nLevel;
    int  nCopy;
    char strMaterial[OUFTEST_MAT_L+1];
    char strEPN[OUFTEST_PPN_L+1];
} UpdateSignalRec;

typedef struct PPNEntry_
{
    char             strPPN[OUFTEST_PPN_L+1];
    char	     strIln[OUFTEST_ILN_L+1];
    struct PPNEntry_ *pNext;
} PPNEntry;

typedef struct
{
    PPNEntry* pFirst;
    PPNEntry* pLast;
    int       nTotal;
} PPNList;

/* Functions used in this module only */
static PPNList* OUFTestAddPPNToList(PPNList *pPPNList, const char *strPPN, const char *strIln);
static bool OUFTestCheckDirectories(void);
static bool OUFTestCheckDirForWriteProtection(const char *strDir);
static void OUFTestClearPPNList(PPNList *pPPNList);
static void OUFTestConstructTimeStamp(char *strTimeStamp,
				  bool bUseSavedTimeStamp,
				  bool bExtended);
static void OUFTestConstructTimeStampedFilename(char *strDestFilename,
						const char *strDir,
						const char *strSrcFilename);
static bool OUFTestCopyFile(const char *strSrcFile, const char *strDestFile);
static PPNList* OUFTestCreatePPNList(FileData *pUpdSigFile);
static bool OUFTestEnterCriticalSection(void);
static void OUFTestExit(void);
static bool OUFTestExitCriticalSection(void);
static bool OUFTestExitProcessTitles(FileData *pTitleFile);
static bool OUFTestExitProcessUpdateSignals(FileData *pUpdSigFile);
static int  OUFTestGetFirstFileInDir(const char *strDir, char *strFile);
static int  OUFTestGetNextEvent(void);
static int  OUFTestGetUpdateSignal(FileData *pUpdSigFile,
				   UpdateSignalRec *pUpdateSignal);
static bool OUFTestHandleTitleEvent(void);
static bool OUFTestHandleUpdateEvent(void);
static bool OUFTestInit(void);
static int  OUFTestInitProcessTitles(FileData *pTitleFile);
static int  OUFTestInitProcessUpdateSignals(FileData *pUpdSigFile);
static bool OUFTestInitSettings(void);
static bool OUFTestInitSignalHandler(void);
static bool OUFTestLogFetchRequests(FileData *pFetchReqFile);
static bool OUFTestLogTitles(FileData *pTitleFile);
static bool OUFTestLogUpdateSignals(FileData *pUpdSigFile);
static bool OUFTestMoveFile(const char *strFile, const char *strDir);
static int  OUFTestProcessTitles(void);
static int  OUFTestProcessUpdateSignals(void);
static PPNList* OUFTestScanForReferences(FileData *pTitleFile);
static bool OUFTestSearchPPNInList(PPNList *pPPNList, const char *strPPN, const char *strIln);
static void OUFTestSetError(int nError);
static PPNList* OUFTestSortPPNList(PPNList *pPPNList);
static const char* OUFTestStripFilename(const char *strFilename);
static bool OUFTestWriteFetchRequests(PPNList *pPPNList,
				      FileData *pFetchReqFile);

/* Type definitions */
static struct
{
    int  nError;
    char strError[OUFTEST_ERROR_L+1];
    bool bErrorSet;
    
    /* Trace */
    char strTraceDir[OUFTEST_DIR_L+1];
    bool bDebug;

    /* Directories */
    char strProcessedDir[OUFTEST_DIR_L+1];
    char strUpdSigDir[OUFTEST_DIR_L+1];
    char strFetchDir[OUFTEST_DIR_L+1];
    char strTitleDir[OUFTEST_DIR_L+1];

    /* Processing status */
    bool bProcessed;
} Settings;

static volatile bool bStopProgram = FALSE;
static sigset_t sig_set;

/*------------------------------------------------------------------------
 *  Function	: OUFTestAddPPNToList
 *  Purpose	: Adds a PPN to a PPN list
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 05-11-1996
 *
 *  History:
 *  05-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	:
 *  A pointer to a structure of type PPNList.
 *------------------------------------------------------------------------
 */
static PPNList*
OUFTestAddPPNToList(
	PPNList *pPPNList,
	const char *strPPN,
	const char *strIln)
{
    PPNEntry *pTemp;

    /* Check if PPN already present in list */
    if (OUFTestSearchPPNInList(pPPNList, strPPN, strIln)) {
	return pPPNList;
    }

    if (pPPNList == NULL) {
	pPPNList = (PPNList *)malloc(sizeof(PPNList));
	if (pPPNList == NULL) {
	    OUFTestSetError(OUFTEST_NOT_ENOUGH_MEMORY);
	    return NULL;
	}
	pPPNList->pFirst = NULL;
	pPPNList->pLast = NULL;
	pPPNList->nTotal = 0;
    }

    /* Create new PPN entry */
    pTemp = (PPNEntry *)malloc(sizeof(PPNEntry));
    if (pTemp == NULL) {
        OUFTestSetError(OUFTEST_NOT_ENOUGH_MEMORY);
        return NULL;
    } else {
        /* Initialize PPN entry */
        (void) strcpy(pTemp->strPPN, strPPN);
	if (strIln[0] == '\0') {
	    (void) strcpy(pTemp->strIln, "000");
	} else {
	    (void) strcpy(pTemp->strIln, strIln);
	}
        pTemp->pNext = NULL;
   
        /* Add new PPN entry to PPN list */
        if (pPPNList->pFirst == NULL) {
            pPPNList->pFirst = pTemp;
        } else {
	    pPPNList->pLast->pNext = pTemp;
	}
	pPPNList->pLast = pTemp;
	pPPNList->nTotal++;
    }
    return pPPNList;
} /* OUFTestAddPPNToList */

/*------------------------------------------------------------------------
 *  Function	: OUFTestApplication
 *  Purpose	: main procedure for OUFTest application
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	:
 *  bool
 *------------------------------------------------------------------------
 */
bool
OUFTestApplication(
	void)
{
    bool bOk;
    int  nEvent;

    bOk = OUFTestInit();

    while (bOk && !bStopProgram) {
	nEvent = OUFTestGetNextEvent();
	switch(nEvent) {
	case OUFTEST_UPDATE_EVENT:
	    OUFTEST_DEBUG("Checking for update signal files...\n");
	    if (!OUFTestHandleUpdateEvent()) {
		ERWrite(OUFTEST_ERR, __func__, "%s", OUFTestGetLastError());
		Settings.bErrorSet = FALSE;
	    }
	    break;
	case OUFTEST_TITLE_EVENT:
	    OUFTEST_DEBUG("Checking for title files...\n");
	    if (!OUFTestHandleTitleEvent()) {
		ERWrite(OUFTEST_ERR, __func__, "%s", OUFTestGetLastError());
		Settings.bErrorSet = FALSE;
	    }
	    break;
	case OUFTEST_SLEEP_EVENT:
	    OUFTEST_DEBUG("Sleeping for 5 seconds...\n");
	    (void) sleep(5);
	    OUFTEST_DEBUG("Waking up...\n");
	    break;
	default:
	    break;
	}
    }

    OUFTestExit();

    return bOk;
} /* OUFTestApplication() */

/*------------------------------------------------------------------------
 *  Function	: OUFTestCheckDirectories
 *  Purpose	: Checks the existance of required directories
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestCheckDirectories(
	void)
{
    DIR  *pDir;

    OUFTEST_DEBUG("Checking for required directories...\n");

    /* Attempt to open fetch request directory */
    pDir = opendir(Settings.strFetchDir);
    if (pDir == NULL) {
	OUFTestSetError(OUFTEST_FETCH_DIR_NOT_FOUND);
	return FALSE;
    }
    (void) closedir(pDir);

    /* Attempt to open title directory */
    pDir = opendir(Settings.strTitleDir);
    if (pDir == NULL) {
	OUFTestSetError(OUFTEST_TITLE_DIR_NOT_FOUND);
	return FALSE;
    }
    (void) closedir(pDir);

    /* Attempt to open processed directory */
    pDir = opendir(Settings.strProcessedDir);
    if (pDir == NULL) {
	OUFTestSetError(OUFTEST_PROC_DIR_NOT_FOUND);
	return FALSE;
    }
    (void) closedir(pDir);

    /* Attempt to open update signal directory */
    pDir = opendir(Settings.strUpdSigDir);
    if (pDir == NULL) {
	OUFTestSetError(OUFTEST_UPDSIG_DIR_NOT_FOUND);
	return FALSE;
    }
    (void) closedir(pDir);

    /* Check for write protection */
    if (OUFTestCheckDirForWriteProtection(Settings.strFetchDir)) {
	OUFTestSetError(OUFTEST_FETCH_DIR_PROTECTED);
	return FALSE;
    }

    if (OUFTestCheckDirForWriteProtection(Settings.strProcessedDir)) {
	OUFTestSetError(OUFTEST_PROC_DIR_PROTECTED);
	return FALSE;
    }

    if (OUFTestCheckDirForWriteProtection(Settings.strTitleDir)) {
	OUFTestSetError(OUFTEST_TITLE_DIR_PROTECTED);
	return FALSE;
    }

    if (OUFTestCheckDirForWriteProtection(Settings.strUpdSigDir)) {
	OUFTestSetError(OUFTEST_UPDSIG_DIR_PROTECTED);
	return FALSE;
    }

    OUFTEST_DEBUG("Required directories checked.\n");

    return TRUE;
} /* OUFTestCheckDirectories */

/*------------------------------------------------------------------------
 *  Function	: OUFTestCheckDirForWriteProtection
 *  Purpose	: Checks directory for write protection
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *  21-02-2005	: michael: ThisFunctionHasGotOneOfTheStupidestNamesEver
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *  TRUE: Directory is write-protected
 *------------------------------------------------------------------------
 */
static bool
OUFTestCheckDirForWriteProtection(
	const char *strDir)
{
    FILE *pFile;
    char strFilename[OUFTEST_FILE_L + sizeof(OUFTEST_TEMP_FILE) + 1];

    /* Check for write protection */
    (void) sprintf(strFilename, "%s/%s", strDir, OUFTEST_TEMP_FILE);
    pFile = fopen(strFilename, "w+");
    if (pFile == NULL) {
	return TRUE;
    }
    (void) fclose(pFile);
    (void) remove(strFilename);

    return FALSE;
} /* OUFTestCheckDirForWriteProtection */

/*------------------------------------------------------------------------
 *  Function	: OUFTestClearPPNList
 *  Purpose	: Frees memory used by PPN list
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  void
 *------------------------------------------------------------------------
 */
static void
OUFTestClearPPNList(
	PPNList *pPPNList)
{
    PPNEntry *pEntry;
    PPNEntry *pTemp;

    OUFTEST_DEBUG("Freeing memory used by PPN list...\n");

    if (pPPNList != NULL) {
	pEntry = pPPNList->pFirst;
        while (pEntry) {
	    pTemp = pEntry;
	    pEntry = pEntry->pNext;
	    free(pTemp);
	}
	free(pPPNList);
    }

    OUFTEST_DEBUG("Memory freed.\n");
} /* OUFTestClearPPNList */

/*------------------------------------------------------------------------
 *  Function	: OUFTestConstructTimeStamp
 *  Purpose	: Constructs timestamp string
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *  22-10-1997	: HSi	   : Chopped format string of non-extended format
 *			     in chunks. Old format contained sequences
 *			     processed by sccs.
 *			     Replace all spaces by zeros in non-extended
 *			     format.
 *
 *  Notes	: 
 *  bUseSavedTimeStamp - if TRUE, a previously saved timestamp will be
 *                       used instead of the current time.
 *  bExtended - if TRUE, an extended format will be used for the
 *              timestamp.
 *------------------------------------------------------------------------
 */
static void
OUFTestConstructTimeStamp(
	char *strTimeStamp,
	bool bUseSavedTimeStamp,
	bool bExtended)
{
    time_t        TimeStamp;
    static time_t SavedTimeStamp;
    char	  *spacePosition;

    /* Check if a previously saved timestamp must be used */
    if (bUseSavedTimeStamp) {
	TimeStamp = SavedTimeStamp;
    } else {
        (void) time(&TimeStamp);
	SavedTimeStamp = TimeStamp;
    }

    /* Check if an extended format must be used */
    if (bExtended) {
        (void) strftime(strTimeStamp, OUFTEST_DATETIME_L, "%d-%m-%Y %H:%M:%S",
                 localtime(&TimeStamp));
    } else {
        (void) strftime(strTimeStamp, OUFTEST_DATETIME_L,
			"%Y" "%m" "%d" "%H" "%M" "%S", 
                 localtime(&TimeStamp));
	/* Remove spaces. Some formats give leading space in case of
	 * of single digit
	 */
	spacePosition = strTimeStamp;
	while ( ( spacePosition = strchr(spacePosition, ' ') ) != NULL ) {
	    *spacePosition = '0';
	}
    }
} /* OUFTestConstructTimeStamp */

/*------------------------------------------------------------------------
 *  Function	: OUFTestConstructTimeStampedFilename
 *  Purpose	: Constructs timestamped filename
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *------------------------------------------------------------------------
 */
static void
OUFTestConstructTimeStampedFilename(
	char *strDestFilename,
	const char *strDir,
	const char *strSrcFilename)
{
    static bool bFirstTime = TRUE;
    static char strPreviousTimeStamp[OUFTEST_DATETIME_L+1];
    static int  nCounter = 0;
    char        strTimeStamp[OUFTEST_DATETIME_L+1];

    if (bFirstTime) {
	bFirstTime = FALSE;
	strPreviousTimeStamp[0] = '\0';
    }

    OUFTestConstructTimeStamp(strTimeStamp, FALSE, FALSE);
    if (strcmp(strTimeStamp, strPreviousTimeStamp) == 0) {
	nCounter++;
    } else {
	nCounter = 0;
    }
    (void) sprintf(strDestFilename, "%s/%s_%s%02d",
		   strDir, strSrcFilename, strTimeStamp, nCounter);

    /* Save current timestamp */
    (void) strcpy(strPreviousTimeStamp, strTimeStamp);
} /* OUFTestConstructTimeStampedFilename */

/*------------------------------------------------------------------------
 *  Function	: OUFTestCopyFile
 *  Purpose	: Copies file to another destination
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-11-1996
 *
 *  History:
 *  01-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *------------------------------------------------------------------------
 */
static bool
OUFTestCopyFile(
	const char *strSrcFile,
	const char *strDest)
{
    bool bOk;
    char strCommand[OUFTEST_COMMAND_L+1];

    (void) sprintf(strCommand, "cp %s %s", strSrcFile, strDest);
    OUFTEST_DEBUG("Issuing system command: ");
    OUFTEST_DEBUG(strCommand);
    OUFTEST_DEBUG("\n");
    bOk = (system(strCommand) != -1);

    return bOk;
} /* OUFTestCopyFile */

/*------------------------------------------------------------------------
 *  Function	: OUFTestCreatePPNList
 *  Purpose	: Creates list of unique PPNs based on update signal file
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  A pointer to a structure of type PPNList or NULL if an error occurs.
 *------------------------------------------------------------------------
 */
static PPNList*
OUFTestCreatePPNList(
	FileData *pUpdSigFile)
{
    UpdateSignalRec UpdateSignal;
    PPNList         *pPPNList;
    PPNList         *pResult;
    bool            bOk;
    int             nReturn;

    OUFTEST_DEBUG("Creating PPN list...\n");

    /* Rewind to begin of update signal file */
    rewind(pUpdSigFile->fpFile);

    /* Initialize local variables */
    bOk = TRUE;
    pPPNList = NULL;

    /* Get first update signal */
    nReturn = OUFTestGetUpdateSignal(pUpdSigFile, &UpdateSignal);

    while (bOk && (nReturn == OUFTEST_OK)) {
	/* Check update type */
	if ((strcmp(UpdateSignal.strType, "IN") == 0)
	    || (strcmp(UpdateSignal.strType, "MU") == 0)) {
            pResult = OUFTestAddPPNToList(pPPNList, UpdateSignal.strPPN, 
			UpdateSignal.strIln);
	    bOk = (pResult != NULL);
	    if (bOk) {
	        pPPNList = pResult;
	    }
        }
    	
	if (bOk) {
	    /* Get next update signal, if any... */
            nReturn = OUFTestGetUpdateSignal(pUpdSigFile, &UpdateSignal);
	}
    }
    bOk = bOk && (nReturn != OUFTEST_ERROR);
    
    if (bOk) {
        OUFTEST_DEBUG("PPN list created.\n");
    } else {
	OUFTestClearPPNList(pPPNList);
        OUFTEST_DEBUG("PPN list not created.\n");
    }

    return bOk ? pPPNList : NULL;
} /* OUFTestCreatePPNList */

/*------------------------------------------------------------------------
 *  Function	: OUFTestEnterCriticalSection
 *  Purpose	: Prepares application for entry of critical section
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestEnterCriticalSection(
	void)
{
    /* Block SIGTERM signal, we are entering a critical area... */
    if (sigprocmask(SIG_BLOCK, &sig_set, NULL) == -1) {
        OUFTestSetError(OUFTEST_BLOCK_FAILED);
        return FALSE;
    }
    return TRUE;
} /* OUFTestEnterCriticalSection */

/*------------------------------------------------------------------------
 *  Function	: OUFTestExit
 *  Purpose	: Exists OUF module
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static void
OUFTestExit(
	void)
{
} /* OUFTestExit */

/*------------------------------------------------------------------------
 *  Function	: OUFTestExitCriticalSection
 *  Purpose	: Prepares application for exit of critical section
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestExitCriticalSection(
	void)
{
    /* Remove blocking of SIGTERM signal */
    if (sigprocmask(SIG_UNBLOCK, &sig_set, NULL) == -1) {
        OUFTestSetError(OUFTEST_UNBLOCK_FAILED);
        return FALSE;
    }
    return TRUE;
} /* OUFTestExitCriticalSection */

/*------------------------------------------------------------------------
 *  Function	: OUFTestExitProcessTitles
 *  Purpose	: Ends processing of titles
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestExitProcessTitles(
	FileData *pTitleFile)
{
    char strText[OUFTEST_TEXT_L+1];
    const char *strStrippedFilename;

    strStrippedFilename = OUFTestStripFilename(pTitleFile->strFilename);

    /* Log title file data */
    (void) TRACEDate();
    (void) sprintf(strText, "Title file '%s' closed.\n", strStrippedFilename);
    (void) TRACEWrite(strText);

    /* Move title file to processed directory */
    if (!OUFTestMoveFile(pTitleFile->strFilename, Settings.strProcessedDir)) {
	OUFTestSetError(OUFTEST_TTL_FILE_NOT_MOVED);
	return FALSE;
    }

    /* Log title file data */
    (void) TRACEDate();
    (void) sprintf(strText, "Title file '%s' moved.\n", strStrippedFilename);
    (void) TRACEWrite(strText);

    return TRUE;
} /* OUFTestExitProcessTitles */

/*------------------------------------------------------------------------
 *  Function	: OUFTestExitProcessUpdateSignals
 *  Purpose	: Ends processing of update signals
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestExitProcessUpdateSignals(
	FileData *pUpdSigFile)
{
    char strText[OUFTEST_TEXT_L+1];
    const char *strStrippedFilename;

    /* Close update signal file */
    if (fclose(pUpdSigFile->fpFile) != 0) {
	OUFTestSetError(OUFTEST_US_FILE_NOT_CLOSED);
	return FALSE;
    }

    strStrippedFilename = OUFTestStripFilename(pUpdSigFile->strFilename);

    /* Log update signal file data */
    (void) TRACEDate();
    (void) sprintf(strText, "Update signal file '%s' closed.\n",
		   strStrippedFilename);
    (void) TRACEWrite(strText);

    /* Move update signal file to processed directory */
    if (!OUFTestMoveFile(pUpdSigFile->strFilename, Settings.strProcessedDir)) {
	OUFTestSetError(OUFTEST_US_FILE_NOT_MOVED);
	return FALSE;
    }

    /* Log update signal file data */
    (void) TRACEDate();
    (void) sprintf(strText, "Update signal file '%s' moved.\n",
		   strStrippedFilename);
    (void) TRACEWrite(strText);

    return TRUE;
} /* OUFTestExitProcessUpdateSignals */

/*------------------------------------------------------------------------
 *  Function	: OUFTestGetFirstFileInDir
 *  Purpose	: Retrieves the name of the first file in a specified
 *                directory
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-11-1996
 *
 *  History:
 *  01-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  Integer (OUFTEST_OK, OUFTEST_ERROR and OUFTEST_NOT_FOUND).
 *------------------------------------------------------------------------
 */
static int
OUFTestGetFirstFileInDir(
	const char *strDir,
	char *strFile)
{
    struct dirent *pDirEntry;
    struct stat   EntryStat;
    bool          bFound;
    bool          bOk;
    DIR           *pDir;
    int           nReturn;

    /* Open directory */
    pDir = opendir(strDir);
    if (pDir == NULL) {
	OUFTestSetError(OUFTEST_DIR_NOT_FOUND);
	return OUFTEST_ERROR;
    }

    /* Find first file in directory */
    bOk = TRUE;
    bFound = FALSE;
    pDirEntry = readdir(pDir); 
    while (pDirEntry && !bFound && bOk) {
	(void) sprintf(strFile, "%s/%s", strDir, pDirEntry->d_name);

	nReturn = stat(strFile, &EntryStat);
	if (nReturn == -1) {
	    OUFTestSetError(OUFTEST_FILE_STATUS_NOT_RETR);
	    bOk = FALSE;
        } else {
            bFound = S_ISDIR(EntryStat.st_mode) ? FALSE : TRUE;
	    if (!bFound) {
                pDirEntry = readdir(pDir); 
	    }
	}
    }

    /* Close directory */
    (void) closedir(pDir);

    return (bFound ? OUFTEST_OK : (bOk ? OUFTEST_NOT_FOUND : OUFTEST_ERROR));
} /* OUFTestGetFirstFileInDir */

/*------------------------------------------------------------------------
 *  Function	: OUFTestGetLastError
 *  Purpose	: Retrieves last error string
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: String containing last error
 *------------------------------------------------------------------------
 */
char*
OUFTestGetLastError(
	void)
{
    switch(Settings.nError) {
    case OUFTEST_NO_ERROR:
	(void) sprintf(Settings.strError, "OUFTest: No error.");
	break;
    case OUFTEST_ERROR:
	(void) sprintf(Settings.strError, "OUFTest: Error.");
	break;
    case OUFTEST_PROFILE_NOT_OPENED:
	(void) sprintf(Settings.strError, "OUFTest: Profile not opened.");
	break;
    case OUFTEST_PROFILE_NOT_CLOSED:
	(void) sprintf(Settings.strError, "OUFTest: Profile not closed.");
	break;
    case OUFTEST_TRACE_ERROR:
	/*(void) strcpy(Settings.strError, TRACEGetLastError()); */
	break;
    case OUFTEST_CFG_FILE_NOT_SET:
	(void) sprintf(Settings.strError,
		       "OUFTest: Profile environment variable not set.");
	break;
    case OUFTEST_NOT_ENOUGH_MEMORY:
	(void) sprintf(Settings.strError, "OUFTest: Memory not allocated.");
	break;
    case OUFTEST_US_FILE_NOT_OPENED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Update signal file not opened.");
	break;
    case OUFTEST_US_FILE_NOT_CLOSED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Update signal file not closed.");
	break;
    case OUFTEST_FR_READ_ERROR:
	(void) sprintf(Settings.strError,
		       "OUFTest: Fetch request file read error.");
	break;
    case OUFTEST_FETCH_DIR_NOT_FOUND:
	(void) sprintf(Settings.strError,
		       "OUFTest: Fetch request directory not found.");
	break;
    case OUFTEST_FILE_STATUS_NOT_RETR:
	(void) sprintf(Settings.strError,
	       "OUFTest: Fetch request file status not retrieved.");
	break;
    case OUFTEST_FR_FILE_NOT_OPENED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Fetch request file not opened.");
	break;
    case OUFTEST_FR_FILE_NOT_CLOSED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Fetch request file not closed.");
	break;
    case OUFTEST_TTL_FILE_NOT_OPENED:
	(void) sprintf(Settings.strError, "OUFTest: Title file not opened.");
	break;
    case OUFTEST_TTL_FILE_NOT_CLOSED:
	(void) sprintf(Settings.strError, "OUFTest: Title file not closed.");
	break;
    case OUFTEST_BLOCK_FAILED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Blocking of SIGTERM failed.");
	break;
    case OUFTEST_UNBLOCK_FAILED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Removal of SIGTERM block failed.");
	break;
    case OUFTEST_SIGNAL_SET_NOT_EMPTIED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Signal set was not emptied.");
	break;
    case OUFTEST_SIGNAL_HANDLER_NOT_SET:
	(void) sprintf(Settings.strError,
		       "OUFTest: Signal handler was not set.");
	break;
    case OUFTEST_SIGNAL_NOT_ADDED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Signal mask was not added.");
	break;
    case OUFTEST_TITLE_DIR_NOT_FOUND:
	(void) sprintf(Settings.strError,
		       "OUFTest: Title directory not found.");
	break;
    case OUFTEST_PROC_DIR_NOT_FOUND:
	(void) sprintf(Settings.strError,
		       "OUFTest: Processed directory not found.");
	break;
    case OUFTEST_UPDSIG_DIR_NOT_FOUND:
	(void) sprintf(Settings.strError,
		       "OUFTest: Update signal directory not found.");
	break;
    case OUFTEST_TITLE_DIR_PROTECTED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Title directory is write-protected.");
	break;
    case OUFTEST_PROC_DIR_PROTECTED:
	(void) sprintf(Settings.strError,
	       "OUFTest: Processed directory is write-protected.");
	break;
    case OUFTEST_UPDSIG_DIR_PROTECTED:
	(void) sprintf(Settings.strError,
	       "OUFTest: Update signal directory is write-protected.");
	break;
    case OUFTEST_DIR_NOT_FOUND:
	(void) sprintf(Settings.strError, "OUFTest: Directory not found.");
	break;
    case OUFTEST_TTL_FILE_NOT_MOVED:
	(void) sprintf(Settings.strError, "OUFTest: Title file not moved.");
	break;
    case OUFTEST_US_FILE_NOT_MOVED:
	(void) sprintf(Settings.strError,
		       "OUFTest: Update signal file not moved.");
	break;
    case OUFTEST_US_READ_ERROR:
	(void) sprintf(Settings.strError,
		       "OUFTest: Update signal file read error.");
	break;
    case OUFTEST_PARSE_ERROR:
	(void) strcpy(Settings.strError, ParseGetLastError());
	break;
    case OUFTEST_PPLIB_ERROR:
	(void) sprintf(Settings.strError, "PPLIB Error (%s).", PPerrno);
	break;
    case OUFTEST_FETCH_DIR_PROTECTED:
	(void) sprintf(Settings.strError,
	       "OUFTest: Fetch request directory is write-protected.");
	break;
    default:
	(void) sprintf(Settings.strError, 
		       "OUFTest: Unknown error (%d).", Settings.nError);
    }
    return Settings.strError;
} /* OUFTestGetLastError() */

/*------------------------------------------------------------------------
 *  Function	: OUFTestGetNextEvent
 *  Purpose	: Returns ID of next event
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  An integer that represents the next event to be processed:
 *  OUFTEST_TITLE_EVENT, OUFTEST_UPDATE_EVENT, OUFTEST_SLEEP_EVENT.
 *------------------------------------------------------------------------
 */
static int
OUFTestGetNextEvent(
	void)
{
    static int nPreviousEvent = OUFTEST_NO_EVENT;
    int nNextEvent = 0;

    switch(nPreviousEvent) {
    case OUFTEST_NO_EVENT:
	Settings.bProcessed = FALSE;
	nNextEvent = OUFTEST_UPDATE_EVENT;
	break;
    case OUFTEST_UPDATE_EVENT:
	nNextEvent = OUFTEST_TITLE_EVENT;
	break;
    case OUFTEST_TITLE_EVENT:
	if (!Settings.bProcessed) {
	    nNextEvent = OUFTEST_SLEEP_EVENT;
	} else {
	    Settings.bProcessed = FALSE;
	    nNextEvent = OUFTEST_UPDATE_EVENT;
	}
	break;
    case OUFTEST_SLEEP_EVENT:
	nNextEvent = OUFTEST_UPDATE_EVENT;
	break;
    }

    nPreviousEvent = nNextEvent;

    return nNextEvent;
} /* OUFTestGetNextEvent */

/*------------------------------------------------------------------------
 *  Function	: OUFTestGetUpdateSignal
 *  Purpose	: Retrieves the next update signal from the update signal
 *                file
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *  16-04-2020	: zwaalm   : CBS-9161 Fixed parsing PPN and EPN (unused).
 *
 *  Notes	: 
 *
 *  The input contains output from command "\\too %d \\upd \\sed %s ;".
 *  Example:
 *  dd-mm-jjjj hh:mm:ss.ttt.ttt type PPN______ l ex ma EPN______
 *  09-04-2020 13:18:01.670.000 [IN] 414716345 2 01    1208899775
 *  14-04-2020 15:58:46.180.000 [MU] 417172486 0 00 Aa          
 *  14-04-2020 16:40:08.650.000 [MU] 417172486 0 00 Aa          
 *  14-04-2020 17:03:54.530.000 [IN] 12345678909 0 00 Aa          
 *  15-04-2020 10:31:37.480.000 [IN] 19345678905 0 00 Ao          
 *
 *  PPN's and EPN's can have a flexible length, the others have fixed length.
 *  When [UpdateSignal] Library is defined, OUF adds ILN and library in
 *  front of these lines (length 15). Example:
 *
 *  015 1999       09-04-2020 13:18:01.670.000 [IN] 414716345 2 01    1208899775
 *
 *  Returns	: 
 *  Integer (OUFTEST_OK, OUFTEST_ERROR and OUFTEST_NOT_FOUND).
 *------------------------------------------------------------------------
 */
static int
OUFTestGetUpdateSignal(
	FileData *pUpdSigFile,
	UpdateSignalRec *pUpdateSignal)
{
/*
 * Use compile-time assertion to check defined lengths of ouftest.h.
 * See https://www.pixelbeat.org/programming/gcc/static_assert.html
 */
#define ASSERT_CONCAT_(a, b) a##b
#define ASSERT_CONCAT(a, b) ASSERT_CONCAT_(a, b)
#define ct_assert(e) enum { ASSERT_CONCAT(assert_line_, __LINE__) = 1/(!!(e)) }

/*
 * Determine the minimum of dstlen and srclen, and do compile-time checking as
 * a side-effect. This macro is only to improve readability; you can use:
 *
 * len = OUFTEST_MIN_CHECK(sizeof(pUpdateSignal->X) - 1, UPDSIGREC_X_L);
 *
 * instead of:
 *
 * len = UPDSIGREC_X_L;
 * ct_assert(sizeof(pUpdateSignal->X) - 1 >= len);
 */
#define OUFTEST_MIN_CHECK(dstlen, srclen)	\
    (srclen);					\
    ct_assert((dstlen) >= (srclen));

/*
 * Use runtime check to see if an input string field was completely parsed,
 * if it's not then the destination buffer is too small.
 */
#define PARSE_CHECK(condition, str, off) {				\
    if (!(condition)) {		  		        		\
	ERWrite(APPLE_PARAM, __func__,					\
		"Condition \"%s\" not met at pos %d in \"%s\".",	\
                #condition, (off), (str));				\
        OUFTestSetError(OUFTEST_PARSE_ERROR);				\
	return OUFTEST_ERROR;						\
    }									\
}

/* Lengths of update signal record fixed length string fields. */
#define UPDSIGREC_ILN_L 	3
#define UPDSIGREC_LIBRARY_L 	10
#define UPDSIGREC_DATETIME_L 	27
#define UPDSIGREC_TYPE_L 	2
#define UPDSIGREC_MAT_L 	2

/*
 * Define the maximum buffer size needed to parse an UPDSIG record.
 * The sizeof(dst-element) includes room for a terminating '\0' which
 * correspondes with a space in strText.
 */
#define MAX_SIGNAL_BUFFER_SIZE                               \
    sizeof(pUpdateSignal->strIln) +                          \
    sizeof(pUpdateSignal->strInstitution) +                  \
    sizeof(pUpdateSignal->strDateTime) +                     \
    sizeof(pUpdateSignal->strType) + 2 /* '[' and ']' */ +   \
    sizeof(pUpdateSignal->strPPN) +                          \
    2 /* nLevel: 1 byte + space */ +                         \
    3 /* nCopy: 2 bytes + space */ +                         \
    sizeof(pUpdateSignal->strMaterial) +                     \
    sizeof(pUpdateSignal->strEPN)

    char strText[OUFTEST_TEXT_L+1];
    int offset;
    int len;	/* length of string field */

    /*
     * Prevent possible out of bounds read from source.
     * Note that this is a compile-time check.
     */
    ct_assert(sizeof(strText) >= MAX_SIGNAL_BUFFER_SIZE);

    /* Get next line in update signal file */
    if (fgets(strText, OUFTEST_TEXT_L, pUpdSigFile->fpFile) == NULL) {
        /* Check if an error occurred */
        if (!feof(pUpdSigFile->fpFile)) {
            OUFTestSetError(OUFTEST_US_READ_ERROR);
	    return OUFTEST_ERROR;
	}

	/* End of update signal file has been reached */
	return OUFTEST_NOT_FOUND;
    }

    /* Fill update signal record with data from update signal file */
    /* !!!! DIRTY WORK !!!! */

    /* '\0'-teminate all destination strings at once. */
    memset(pUpdateSignal, '\0', sizeof(*pUpdateSignal));

    /* Iln/Library */
    offset = 0;
    if (strText[2] == '-') {
	(void) strcpy(pUpdateSignal->strIln, "000");
    } else {
	len = OUFTEST_MIN_CHECK(sizeof(pUpdateSignal->strIln) - 1,
		                UPDSIGREC_ILN_L);
        (void) strncpy(pUpdateSignal->strIln, strText, len);
	offset += len;
        PARSE_CHECK(strText[offset] == ' ', strText, offset);
	offset++; /* space */

	len = OUFTEST_MIN_CHECK(sizeof(pUpdateSignal->strInstitution) - 1,
			        UPDSIGREC_LIBRARY_L);
	(void) strncpy(pUpdateSignal->strInstitution, &(strText[offset]), len);
	offset += len;
        PARSE_CHECK(strText[offset] == ' ', strText, offset);
	offset++; /* space */
    }

    /* Datetime */
    len = OUFTEST_MIN_CHECK(sizeof(pUpdateSignal->strDateTime) - 1,
		            UPDSIGREC_DATETIME_L);
    (void) strncpy(pUpdateSignal->strDateTime, &(strText[offset]), len);
    offset += len;
    PARSE_CHECK(strText[offset] == ' ', strText, offset);
    offset++; /* space */

    /* Type */
    PARSE_CHECK(strText[offset] == '[', strText, offset);
    offset++; /* '[' */
    len = OUFTEST_MIN_CHECK(sizeof(pUpdateSignal->strType) - 1,
		            UPDSIGREC_TYPE_L);
    (void) strncpy(pUpdateSignal->strType, &(strText[offset]), len);
    offset += len;
    PARSE_CHECK(strText[offset] == ']', strText, offset);
    offset++; /* ']' */
    PARSE_CHECK(strText[offset] == ' ', strText, offset);
    offset++; /* space */

    /* PPN */
    for (len = 0; len < (int)sizeof(pUpdateSignal->strPPN) - 1; len++) {
        if (strText[offset] == ' ') {
	    break;
	}
        pUpdateSignal->strPPN[len] = strText[offset++];
    }
    PARSE_CHECK(strText[offset] == ' ', strText, offset);
    offset++; /* space */

    /* Title level */
    pUpdateSignal->nLevel = strText[offset++] - '0';
    PARSE_CHECK(strText[offset] == ' ', strText, offset);
    offset++; /* space */

    /* Copy sequence number */
    pUpdateSignal->nCopy = (strText[offset] - '0') * 10 +
			   (strText[offset + 1] - '0');
    offset += 2;
    PARSE_CHECK(strText[offset] == ' ', strText, offset);
    offset++; /* space */

    /* Material type */
    len = OUFTEST_MIN_CHECK(sizeof(pUpdateSignal->strMaterial) - 1,
		            UPDSIGREC_MAT_L);
    (void) strncpy(pUpdateSignal->strMaterial, &(strText[offset]), len);
    offset += len;
    PARSE_CHECK(strText[offset] == ' ', strText, offset);
    offset++; /* space */

    /* EPN */
    for (len = 0; len < (int)sizeof(pUpdateSignal->strEPN) - 1; len++) {
        if (PM_is_ascii_space(strText[offset])) {
	    break;
	}
        pUpdateSignal->strEPN[len] = strText[offset++];
    }
    PARSE_CHECK(PM_is_ascii_space(strText[offset]), strText, offset);

    return OUFTEST_OK;
#undef ASSERT_CONCAT_
#undef ASSERT_CONCAT
#undef ct_assert
#undef OUFTEST_MIN_CHECK
#undef PARSE_CHECK
#undef UPDSIGREC_ILN_L
#undef UPDSIGREC_LIBRARY_L
#undef UPDSIGREC_DATETIME_L
#undef UPDSIGREC_TYPE_L
#undef UPDSIGREC_MAT_L
#undef MAX_SIGNAL_BUFFER_SIZE
} /* OUFTestGetUpdateSignal */

/*------------------------------------------------------------------------
 *  Function	: OUFTestHandleTitleEvent
 *  Purpose	: Handles title event
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestHandleTitleEvent(
	void)
{
    bool bOk;
    int  nReturn = 0;

    bOk = OUFTestEnterCriticalSection();
    if (bOk) {
        nReturn = OUFTestProcessTitles();
        bOk = (nReturn != OUFTEST_ERROR) && OUFTestExitCriticalSection();
    }

    if (nReturn == OUFTEST_OK) {
	Settings.bProcessed = TRUE;
    }

    return bOk;
} /* OUFTestHandleTitleEvent */

/*------------------------------------------------------------------------
 *  Function	: OUFTestHandleUpdateEvent
 *  Purpose	: Handles update signal event
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestHandleUpdateEvent(
	void)
{
    bool bOk;
    int  nReturn = 0;

    bOk = OUFTestEnterCriticalSection();
    if (bOk) {
        nReturn = OUFTestProcessUpdateSignals();
        bOk = (nReturn != OUFTEST_ERROR) && OUFTestExitCriticalSection();
    }

    if (nReturn == OUFTEST_OK) {
	Settings.bProcessed = TRUE;
    }

    return bOk;
} /* OUFTestHandleUpdateEvent */

/*------------------------------------------------------------------------
 *  Function	: OUFTestInit
 *  Purpose	: Initializes OUFTest module
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestInit(
	void)
{
    if (!OUFTestInitSignalHandler()) {
	return FALSE;
    }

    if (!OUFTestInitSettings()) {
	return FALSE;
    }

    if (!OUFTestCheckDirectories()) {
	return FALSE;
    }

    return TRUE;
} /* OUFTestInit */

/*------------------------------------------------------------------------
 *  Function	: OUFInitProcessTitles
 *  Purpose	: Initializes processing of titles
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  Integer (OUFTEST_OK, OUFTEST_ERROR and OUFTEST_NOT_FOUND)
 *------------------------------------------------------------------------
 */
static int
OUFTestInitProcessTitles(
	FileData *pTitleFile)
{
    int  nReturn;
    char strText[OUFTEST_TEXT_L+1];

    OUFTEST_DEBUG("Initializing processing of titles...\n");

    nReturn = OUFTestGetFirstFileInDir(Settings.strTitleDir,
				       pTitleFile->strFilename);
    switch(nReturn) {
    case OUFTEST_ERROR:
	return OUFTEST_ERROR;
    case OUFTEST_NOT_FOUND:
	OUFTEST_DEBUG("Title directory is empty...\n");
	return OUFTEST_NOT_FOUND;
    }

    /* Log title file data */
    (void) TRACEDate();
    (void) sprintf(strText, "Title file '%s' opened.\n",
		   OUFTestStripFilename(pTitleFile->strFilename));
    (void) TRACEWrite(strText);

    OUFTEST_DEBUG("Processing of titles initialized.\n");

    return OUFTEST_OK;
} /* OUFTestInitProcessTitles */

/*------------------------------------------------------------------------
 *  Function	: OUFInitProcessUpdateSignals
 *  Purpose	: Initializes processing of update signals
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-11-1996
 *
 *  History:
 *  01-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  Integer (OUFTEST_OK, OUFTEST_ERROR and OUFTEST_NOT_FOUND)
 *------------------------------------------------------------------------
 */
static int
OUFTestInitProcessUpdateSignals(
	FileData *pUpdSigFile)
{
    int  nReturn;
    char strText[OUFTEST_TEXT_L+1];

    OUFTEST_DEBUG("Initializing processing of update signals...\n");

    nReturn = OUFTestGetFirstFileInDir(Settings.strUpdSigDir,
				       pUpdSigFile->strFilename);
    switch(nReturn) {
    case OUFTEST_ERROR:
	return OUFTEST_ERROR;
    case OUFTEST_NOT_FOUND:
	OUFTEST_DEBUG("Update signal directory is empty...\n");
	return OUFTEST_NOT_FOUND;
    }

    /* Open first update signal file in update signal directory */
    pUpdSigFile->fpFile = fopen(pUpdSigFile->strFilename, "r");
    if (pUpdSigFile->fpFile == NULL) {
	OUFTestSetError(OUFTEST_US_FILE_NOT_OPENED);
	return OUFTEST_ERROR;
    }

    /* Log update signal file data */
    (void) TRACEDate();
    (void) sprintf(strText, "Update signal file '%s' opened.\n",
		   OUFTestStripFilename(pUpdSigFile->strFilename));
    (void) TRACEWrite(strText);

    OUFTEST_DEBUG("Processing of update signals initialized.\n");

    return OUFTEST_OK;
} /* OUFTestInitProcessUpdateSignals */

/*------------------------------------------------------------------------
 *  Function	: OUFTestInitSettings
 *  Purpose	: Initializes OUFTest settings
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestInitSettings(
	void)
{
    const char *strConfDir;
    char        strIniFile[OUFTEST_FILE_L+1];
    char        strField[OUFTEST_FIELD_L+1];
    char        strTraceFile[OUFTEST_FILE_L+1];
    FILE        *fpProfileFile;
    //bool        bReplace;
    //bool        bScreen;

    OUFTEST_DEBUG("Initializing settings...\n");

    /* Reset settings */
    Settings.nError = OUFTEST_NO_ERROR;
    Settings.bErrorSet = FALSE;

    /* Get name of configuration file */
    strConfDir = ENVGetVar("DOTINIDIR");
    if (strConfDir == NULL) {
	OUFTestSetError(OUFTEST_CFG_FILE_NOT_SET);
	return FALSE;
    }
    (void) sprintf(strIniFile, "%s/ouftest.ini", strConfDir);

    /* Open profile file */
    fpProfileFile = PFOpen(strIniFile, "r");
    if (fpProfileFile == NULL) {
	OUFTestSetError(OUFTEST_PROFILE_NOT_OPENED);
	return FALSE;
    }

    /* Initialize Trace Data */
    (void) PFRead(fpProfileFile, "[Trace]", "Trace", strField);
    if (PM_strcmpi(strField, "ON") == 0) {
        strTraceFile[0] = '\0';
        (void) PFRead(fpProfileFile, "[Trace]", "Dir", strField);
	OUFTestConstructTimeStampedFilename(strTraceFile, strField,
					OUFTEST_TRACE_FILE);

        //(void) PFRead(fpProfileFile, "[Trace]", "Replace", strField);
        //bReplace = (PM_strcmpi(strField, "YES") == 0) ? TRUE : FALSE;
    
        //(void) PFRead(fpProfileFile, "[Trace]", "Screen", strField);
        //bScreen = (PM_strcmpi(strField, "ON") == 0) ? TRUE : FALSE;

        (void) PFRead(fpProfileFile, "[Trace]", "Debug", strField);
        Settings.bDebug = (PM_strcmpi(strField, "ON") == 0) ? TRUE : FALSE;

        /* Initialize tracing module */

	if (!tr_init(fpProfileFile, &gDebug)) {
	    (void) tr_write("%s: initialization of trace variables failed\n",
		 OUFTEST_NAME);
	    return FALSE;
	}
    }

    /* Initialize directories */
    (void) PFRead(fpProfileFile, "[Directories]", "Processed Dir", strField);
    (void) strncpy(Settings.strProcessedDir, strField, OUFTEST_DIR_L);
    Settings.strProcessedDir[OUFTEST_DIR_L] = '\0';

    (void) PFRead(fpProfileFile, "[Directories]", 
	            "Update Signal Dir", strField);
    (void) strncpy(Settings.strUpdSigDir, strField, OUFTEST_DIR_L);
    Settings.strUpdSigDir[OUFTEST_DIR_L] = '\0';
    
    (void) PFRead(fpProfileFile, "[Directories]", 
	        "Fetch Request Dir", strField);
    (void) strncpy(Settings.strFetchDir, strField, OUFTEST_DIR_L);
    Settings.strFetchDir[OUFTEST_DIR_L] = '\0';
    
    (void) PFRead(fpProfileFile, "[Directories]", "Title Dir", strField);
    (void) strncpy(Settings.strTitleDir, strField, OUFTEST_DIR_L);
    Settings.strTitleDir[OUFTEST_DIR_L] = '\0';

    /* Close profile file */
    if (!PFClose(fpProfileFile)) {
        OUFTestSetError(OUFTEST_PROFILE_NOT_CLOSED);
        return FALSE;
    }

    OUFTEST_DEBUG("Settings initialized.\n");

    return TRUE;
} /* OUFTestInitSettings */

/*------------------------------------------------------------------------
 *  Function	: OUFTestSignalHandler
 *  Purpose	: Catches SIGTERM signal
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  void
 *------------------------------------------------------------------------
 */
static void
OUFTestSignalHandler(
	int nSignal)
{
    switch(nSignal) {
	case SIGTERM:
	    bStopProgram = TRUE;
	    break;
	default:
	    break;
    }
} /* OUFTestSignalHandler */

/*------------------------------------------------------------------------
 *  Function	: OUFTestInitSignalHandler
 *  Purpose	: Initializes signal handler
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestInitSignalHandler(
	void)
{
    struct sigaction action;

    action.sa_flags = 0;
    if (sigemptyset(&action.sa_mask) == -1) {
	OUFTestSetError(OUFTEST_SIGNAL_SET_NOT_EMPTIED);
	return FALSE;
    }

    action.sa_handler = OUFTestSignalHandler;
    if (sigaction(SIGTERM, &action, NULL) == -1) {
	OUFTestSetError(OUFTEST_SIGNAL_HANDLER_NOT_SET);
	return FALSE;
    }

    if (sigemptyset(&sig_set) == -1) {
	OUFTestSetError(OUFTEST_SIGNAL_SET_NOT_EMPTIED);
	return FALSE;
    }

    if (sigaddset(&sig_set, SIGTERM) == -1) {
	OUFTestSetError(OUFTEST_SIGNAL_NOT_ADDED);
	return FALSE;
    }

    return TRUE;
} /* OUFTestInitSignalHandler */

/*------------------------------------------------------------------------
 *  Function	: OUFTestLogFetchRequests
 *  Purpose	: Adds timestamp, filename and contents of
 *                fetch request file to the log file
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestLogFetchRequests(
	FileData *pFetchReqFile)
{
    char strText[OUFTEST_TEXT_L+1];

    (void) sprintf(strText, "Contents of file '%s':\n",
		   OUFTestStripFilename(pFetchReqFile->strFilename));
    (void) TRACEWrite(strText);
    
    /* Open fetch request file */
    pFetchReqFile->fpFile = fopen(pFetchReqFile->strFilename, "r");
    if (pFetchReqFile->fpFile == NULL) {
	OUFTestSetError(OUFTEST_FR_FILE_NOT_OPENED);
	return FALSE;
    }

    /* Copy contents of fetch request file to log file */
    while (fgets(strText, OUFTEST_TEXT_L, pFetchReqFile->fpFile) != NULL) {
	TRACEWrite(strText);
    }
    TRACEWrite("---- end of file ----\n");

    /* Check if an error occurred */
    if (!feof(pFetchReqFile->fpFile)) {
        OUFTestSetError(OUFTEST_FR_READ_ERROR);
        return FALSE;
    }

    /* Close fetch request file */
    if (fclose(pFetchReqFile->fpFile) != 0) {
	OUFTestSetError(OUFTEST_FR_FILE_NOT_CLOSED);
	return FALSE;
    }
     
     return TRUE;
} /* OUFTestLogFetchRequests */

/*------------------------------------------------------------------------
 *  Function	: OUFTestLogFiles
 *  Purpose	: Adds a summary of the contents of the title file to
 *                the log file
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 05-11-1996
 *
 *  History:
 *  05-11-1996	: peter jan: creation
 *  09-04-2020	: zwaalm   : CBS-9337 Fixed usage of uninitialised value.
 *  22-04-2020	: zwaalm   : CBS-9337 Code review changes.
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestLogTitles(
	FileData *pTitleFile)
{
    const TitleRec *pTitle;
    char           strPrevPPN[OUFTEST_PPN_L+1] = "";
    char           strText[OUFTEST_TEXT_L+1];

    if (!OUFT_ParseInit(pTitleFile->strFilename)) {
	OUFTestSetError(OUFTEST_PARSE_ERROR);
	return FALSE;
    }

    TRACEWrite("Summary of contents:\n");

    while ((pTitle = ParseGetNextRecord())) {
	if (PM_strcmpi(pTitle->strPPN, strPrevPPN) != 0) {
	    (void) strcpy(strPrevPPN, pTitle->strPPN);
	    (void) sprintf(strText, "PPN:%s %s %c -> %s\n",
	       pTitle->strPPN, pTitle->strIln, pTitle->cFormat,
	       (pTitle->nResultCode == 0) ? "OK" :
	       pTitle->strMessage);
            TRACEWrite(strText);
	}
    }

    TRACEWrite("---- end of summary ----\n");

    if (!ParseExit()) {
	OUFTestSetError(OUFTEST_PARSE_ERROR);
	return FALSE;
    }

    return TRUE;
} /* OUFTestLogTitles */

/*------------------------------------------------------------------------
 *  Function	: OUFTestLogUpdateSignals
 *  Purpose	: Adds timestamp, filename and contents of
 *                update signal file to the log file
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestLogUpdateSignals(
	FileData *pUpdSigFile)
{
    char strText[OUFTEST_TEXT_L+1];

    (void) sprintf(strText, "Contents of file '%s':\n",
		   OUFTestStripFilename(pUpdSigFile->strFilename));
    (void) TRACEWrite(strText);
    
    /* Copy contents of update signal file to log file */
    rewind(pUpdSigFile->fpFile);
    while (fgets(strText, OUFTEST_TEXT_L, pUpdSigFile->fpFile) != NULL) {
	TRACEWrite(strText);
    }
    TRACEWrite("---- end of file ----\n");

    /* Check if an error occurred */
    if (!feof(pUpdSigFile->fpFile)) {
        OUFTestSetError(OUFTEST_US_READ_ERROR);
        return FALSE;
    }

    return TRUE;
} /* OUFTestLogUpdateSignals */

/*------------------------------------------------------------------------
 *  Function	: OUFTestMoveFile
 *  Purpose	: Moves file to another directory
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-11-1996
 *
 *  History:
 *  01-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestMoveFile(
	const char *strFile,
	const char *strDir)
{
    bool bOk;
    char strCommand[OUFTEST_COMMAND_L+1];

    (void) sprintf(strCommand, "mv %s %s", strFile, strDir);
    OUFTEST_DEBUG("Issuing system command: ");
    OUFTEST_DEBUG(strCommand);
    OUFTEST_DEBUG("\n");
    bOk = (system(strCommand) != -1);

    return bOk;
} /* OUFTestMoveFile */

/*------------------------------------------------------------------------
 *  Function	: OUFTestProcessTitles
 *  Purpose	: Processes titles
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  Integer (OUFTEST_OK, OUFTEST_ERROR and OUFTEST_NOT_FOUND).
 *------------------------------------------------------------------------
 */
static int
OUFTestProcessTitles(
	void)
{
    FileData FetchReqFile;
    FileData TitleFile;
    PPNList  *pPPNList = NULL;
    bool     bOk;
    int      nReturn;

    OUFTEST_DEBUG("Start processing titles...\n");

    /* Initialize processing of titles */
    nReturn = OUFTestInitProcessTitles(&TitleFile);
    if (nReturn != OUFTEST_OK) {
	return nReturn;
    }

    /* Log titles */
    bOk = OUFTestLogTitles(&TitleFile);

    /* Create PPN list based on references in title file */
    if (bOk) {
        pPPNList = OUFTestScanForReferences(&TitleFile);
	bOk = !((pPPNList == NULL) && (Settings.nError != OUFTEST_NO_ERROR));
    }

    if (pPPNList) {
        /* Sort PPN list */
        if (bOk) {
            pPPNList = OUFTestSortPPNList(pPPNList);
        }

        /* Write fetch request file based on PPN list */
        if (bOk) {
            bOk = OUFTestWriteFetchRequests(pPPNList, &FetchReqFile);
        }

        /* Clear PPN list */
        OUFTestClearPPNList(pPPNList);

        /* Log fetch requests */
        if (bOk) {
	    bOk = OUFTestLogFetchRequests(&FetchReqFile);
        }

        /* Copy fetch requests file */
        if (bOk) {
	    bOk = OUFTestCopyFile(FetchReqFile.strFilename, 
			          Settings.strFetchDir);
        }
    }

    /* Exit processing of titles */
    bOk = bOk & OUFTestExitProcessTitles(&TitleFile);

    OUFTEST_DEBUG("Finished processing titles.\n");

    return bOk ? OUFTEST_OK : OUFTEST_ERROR;
} /* OUFTestProcessTitles */

/*------------------------------------------------------------------------
 *  Function	: OUFTestProcessUpdateSignals
 *  Purpose	: Processes update signals
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static int
OUFTestProcessUpdateSignals(
	void)
{
    FileData UpdSigFile;
    FileData FetchReqFile;
    PPNList  *pPPNList = NULL;
    bool     bOk;
    int      nReturn;

    OUFTEST_DEBUG("Start processing update signals...\n");

    /* Initialize processing of update signals */
    nReturn = OUFTestInitProcessUpdateSignals(&UpdSigFile);
    if (nReturn != OUFTEST_OK) {
	return nReturn;
    }

    /* Log update signals */
    bOk = OUFTestLogUpdateSignals(&UpdSigFile);

    /* Create PPN list based on update signal file */
    if (bOk) {
        bOk = ((pPPNList = OUFTestCreatePPNList(&UpdSigFile)) != NULL);
    }

    /* Sort PPN list */
    if (bOk) {
        pPPNList = OUFTestSortPPNList(pPPNList);
    }

    /* Write fetch request file based on PPN list */
    if (bOk) {
        bOk = OUFTestWriteFetchRequests(pPPNList, &FetchReqFile);
    }

    /* Clear PPN list */
    OUFTestClearPPNList(pPPNList);

    /* Log fetch requests */
    if (bOk) {
	bOk = OUFTestLogFetchRequests(&FetchReqFile);
    }

    /* Copy fetch requests file */
    if (bOk) {
	bOk = OUFTestCopyFile(FetchReqFile.strFilename, 
			      Settings.strFetchDir);
    }

    /* Exit processing of update signals */
    bOk = bOk && OUFTestExitProcessUpdateSignals(&UpdSigFile);

    OUFTEST_DEBUG("Finished processing update signals.\n");

    return bOk ? OUFTEST_OK : OUFTEST_ERROR;
} /* OUFTestProcessUpdateSignals */

/*------------------------------------------------------------------------
 *  Function	: OUFTestScanForReferences
 *  Purpose	: Scans title file for references to other records
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 05-11-1996
 *
 *  History:
 *  05-11-1996	: peter jan: creation
 *  21-04-2020	: zwaalm   : CBS-9355 Fixed memory leak.
 *  23-04-2020	: zwaalm   : CBS-9352 Used counted string title buffer.
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  A pointer to a structure of type PPNList.
 *------------------------------------------------------------------------
 */
static PPNList*
OUFTestScanForReferences(
	FileData *pTitleFile)
{
    bool     bStop;
    bool     bOk;
    TitleRec *pTitle;
    char     strPPN[OUFTEST_PPN_L+1];
    PPTtl_p  PPTitle;
    PPSbf_p  PPSubField;
    PPTag_p  PPTag;
    PPNList  *pPPNList;
    PPNList  *pResult;

    OUFTEST_DEBUG("Start scanning title file for references...\n");

    PP_ext_tagend = '\r';
    PP_ext_sbfmark = (char)159;

    if (!OUFT_ParseInit(pTitleFile->strFilename)) {
	OUFTestSetError(OUFTEST_PARSE_ERROR);
	return NULL;
    }

    bOk = TRUE;
    bStop = FALSE;
    pPPNList = NULL;
    while (((pTitle = ParseGetNextRecord()) != NULL) && bOk && !bStop) {
	if ((pTitle->nResultCode == 0) &&
	    (strcmp(pTitle->strDatabaseLevel, "BIB") == 0)) {
	    PPTitle = PPFromExtern2(NULL, CSContents(pTitle->pchTitleComponent));
	    if (PPTitle == NULL) {
		OUFTestSetError(OUFTEST_PPLIB_ERROR);
		bOk = FALSE;
	    } else {
		bStop = FALSE;
		PPTag = PPTitle->PPTTL_firsttag;
		while ((PPTag != NULL) && bOk && !bStop) {
                    PPSubField = PPSbfFind(PPTag, '9');
		    if (PPSubField != NULL) {
		        (void)
			strncpy(strPPN,(char *)(PPSubField->PPSBF_contents), 
			        OUFTEST_PPN_L);
		        strPPN[OUFTEST_PPN_L] = '\0';
		        pResult = OUFTestAddPPNToList(pPPNList, strPPN, 
						pTitle->strIln);
                        bOk = (pResult != NULL);
		        if (bOk) {
			    pPPNList = pResult;
			    bStop = (pPPNList->nTotal == OUFTEST_MAX_REFS);
			}
		    }
		    PPTag = PPTag->pp_eltnext;
		}
		PPTtlDestroy(PPTitle);
	    }
	}
    }

    if (!ParseExit()) {
	OUFTestSetError(OUFTEST_PARSE_ERROR);
	return NULL;
    }

    OUFTEST_DEBUG("Title file scanned for references.\n");

    return bOk ? pPPNList : NULL;
} /* OUFTestScanForReferences */

/*------------------------------------------------------------------------
 *  Function	: OUFTestSearchPPNInList
 *  Purpose	: Searches PPN in PPN list
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 05-11-1996
 *
 *  History:
 *  05-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestSearchPPNInList(
	PPNList *pPPNList,
	const char *strPPN,
	const char *strIln)
{
    bool     bFound;
    PPNEntry *pPPNEntry;

    bFound = FALSE;
    if (pPPNList) {
	pPPNEntry = pPPNList->pFirst;
	while (pPPNEntry && !bFound) {
	    bFound = ((PM_strcmpi(pPPNEntry->strPPN, strPPN) == 0) &&
	    (PM_strcmpi(pPPNEntry->strIln, strIln) == 0));
	    pPPNEntry = pPPNEntry->pNext;
	}
    }
    return bFound;
} /* OUFTestSearchPPNInList */

/*------------------------------------------------------------------------
 *  Function	: OUFTestSetError
 *  Purpose	: Sets error code
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 31-10-1996
 *
 *  History:
 *  31-10-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  void
 *------------------------------------------------------------------------
 */
static void
OUFTestSetError(
	int nError)
{
    if (!Settings.bErrorSet) {
	Settings.nError = nError;
	Settings.bErrorSet = TRUE;
    }
} /* OUFTestSetError */

/*------------------------------------------------------------------------
 *  Function	: OUFTestSortPPNList
 *  Purpose	: Sorts PPN list
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  A pointer to a structure of type PPNList
 *------------------------------------------------------------------------
 */
static PPNList*
OUFTestSortPPNList(
	PPNList *pPPNList)
{
    PPNEntry *pPPNA;
    PPNEntry *pPPNB;
    char     strTemp[OUFTEST_PPN_L+1];

    OUFTEST_DEBUG("Sorting PPN list...\n");

    pPPNA = pPPNList->pFirst;
    while (pPPNA) {
	pPPNB = pPPNA->pNext;
	while (pPPNB) {
	    if (PM_strcmpi(pPPNA->strPPN, pPPNB->strPPN) > 0) {
		(void) strcpy(strTemp, pPPNA->strPPN);
		(void) strcpy(pPPNA->strPPN, pPPNB->strPPN);
		(void) strcpy(pPPNB->strPPN, strTemp);
		(void) strcpy(strTemp, pPPNA->strIln);
		(void) strcpy(pPPNA->strIln, pPPNB->strIln);
		(void) strcpy(pPPNB->strIln, strTemp);
	    }
	    pPPNB = pPPNB->pNext;
	}
	pPPNA = pPPNA->pNext;
    }

    OUFTEST_DEBUG("PPN list sorted.\n");

    return pPPNList;
} /* OUFTestSortPPNList */

/*------------------------------------------------------------------------
 *  Function	: OUFTestStripFilename
 *  Purpose	: Strips filename of path
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 01-11-1996
 *
 *  History:
 *  01-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  Pointer to filename in absolute filename string
 *------------------------------------------------------------------------
 */
static const char*
OUFTestStripFilename(
	const char *strFilename)
{
    int nIndex;

    /* Start at the back of the filename,
       walk back until we encounter a slash... */
    for (nIndex = strlen(strFilename); (nIndex > 0) && 
	 (strFilename[nIndex-1] != '/'); nIndex--);

    return &(strFilename[nIndex]);
} /* OUFTestStripFilename */

/*------------------------------------------------------------------------
 *  Function	: OUFTestWriteFetchRequests
 *  Purpose	: Writes fetch requests based on PPNs in list
 *  Method	: 
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan: creation
 *
 *  Notes	: 
 *
 *  Returns	: 
 *  bool
 *------------------------------------------------------------------------
 */
static bool
OUFTestWriteFetchRequests(
	PPNList *pPPNList,
	FileData *pFetchReqFile)
{
    const char *strStrippedFilename;
    PPNEntry   *pTemp;
    char       strText[OUFTEST_TEXT_L+1];

    OUFTEST_DEBUG("Start writing fetch requests...\n");

    /* Construct name of new fetch request file */
    OUFTestConstructTimeStampedFilename(pFetchReqFile->strFilename,
					Settings.strProcessedDir,
					OUFTEST_FETCH_REQUEST_FILE);

    /* Create new fetch request file */
    pFetchReqFile->fpFile = fopen(pFetchReqFile->strFilename, "w");
    if (pFetchReqFile->fpFile == NULL) {
	OUFTestSetError(OUFTEST_FR_FILE_NOT_OPENED);
	return FALSE;
    }

    strStrippedFilename = OUFTestStripFilename(pFetchReqFile->strFilename);

    (void) TRACEDate();
    (void) sprintf(strText, "Fetch request file '%s' created.\n",
		   strStrippedFilename);
    (void) TRACEWrite(strText);

    /* Write fetch requests for all PPNs in list */
    pTemp = pPPNList->pFirst;
    while (pTemp) {
        /* Construct fetch request */
	if (strcmp(pTemp->strIln, "000") == 0) {
	    (void) fprintf(pFetchReqFile->fpFile, "%s\n", pTemp->strPPN);
	} else {
	    (void) fprintf(pFetchReqFile->fpFile, "%s %s\n", pTemp->strPPN,
		    pTemp->strIln);
	}
	pTemp = pTemp->pNext;
    }

    /* Close fetch request file */
    if (fclose(pFetchReqFile->fpFile) != 0) {
	OUFTestSetError(OUFTEST_FR_FILE_NOT_CLOSED);
	return FALSE;
    }
    
    (void) TRACEDate();
    (void) sprintf(strText, "Fetch request file '%s' closed.\n",
		   strStrippedFilename);
    (void) TRACEWrite(strText);

    OUFTEST_DEBUG("Fetch requests written.\n");

    return TRUE;
} /* OUFTestWriteFetchRequests */
