/*========================================================================
 *  Copyright (c) 1996 by Pica
 *
 *  File	: ouftest_parse.c
 *  Description	: routines for title parsing module
 *  Author(s)	: Peter Jan van der Mespel
 *  Creation	: 04-11-1996
 *
 *  History:
 *  04-11-1996	: peter jan      : creation
 *  05-02-2004	: huug		: read titles bigger than 8K, independent
 *		: 		: from length field
 *  23-04-2020	: zwaalm   : CBS-9352 Used counted string title buffer.
 *  23-04-2020	: zwaalm   : CBS-9352 Code review changes.
 *
 *  Notes	:
 *
 *========================================================================
 */

#pragma ident "$Id: ouftest_parse.c 376297 2020-04-23 12:56:42Z zwaalm $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <cs.h>
#include "ouftest_parse.h"

/* Global variables */
typedef struct
{
    FILE *fpTitleFile;
    char strTitleFile[PARSE_FILE_L+1];

    bool bHeaderRead;

    bool bErrorSet;
    int  nError;
    char strError[PARSE_ERROR_L+1];
} ParseStatus;

/* variables used in this module only */
static ParseStatus Status;
static TitleRec    Title;
static char        pchBuffer[PARSE_LARGE_BUFFER_L+1];

static bool        ParseGetBody(void);
static bool        ParseGetHeader(void);
static char*       ParseGetLine(bool bRetainCR);
static int	   ParseGetTitle(CSBuf_p buffer);
static void        ParseSetError(int nError);

/*------------------------------------------------------------------------
 *  Function	: ParseExit
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  21-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
bool
ParseExit(void)
{
    if (fclose(Status.fpTitleFile) != 0) {
	ParseSetError(PARSE_FILE_NOT_CLOSED);
	return FALSE;
    }
    return TRUE;
} /* ParseExit */

/*------------------------------------------------------------------------
 *  Function	: ParseGetBody
 *  Purpose	: retrieve title data from body
 *  Method	: 
 *
 *  History:
 *  05-02-2004	: huug     : header added
 *  05-02-2004	: 	   : read titles bigger than 8K, independent
 *		: 	   : from length field
 *  09-02-2005	:michael   : added support for three digit copy field
 *                           (for 99+ copies)
 *
 *  Notes	:
 *
 *  Returns	: TRUE when successful, FALSE when not
 *
 *------------------------------------------------------------------------
 */
static bool
ParseGetBody(
	void)
{
    char *pchLine;
    char strLength[PARSE_LENGTH_L+1];

    /* Clear body fields */
    Title.strDatabaseLevel[0] = '\0';
    Title.nLength = 0;
    Title.strTime[0] = '\0';
    if (!CSClear(Title.pchTitleComponent)) {
	return FALSE;
    }

    /* Get next line */
    pchLine = ParseGetLine(FALSE);
    if (pchLine == NULL) {
	return FALSE;
    }

    /* Check database level field */

    /* First character of a copy can be a digit now /michael */

    if (!((strncmp(pchLine, "BIB", 3) == 0) ||
        (strncmp(pchLine, "LOK", 3) == 0) ||
        (((pchLine[0] == 'E') || isdigit(pchLine[0])) &&
	isdigit(pchLine[1]) &&
	isdigit(pchLine[2]))))
    {
	ParseSetError(PARSE_SYNTAX_ERROR);
	return FALSE;
    }
    (void) strncpy(Title.strDatabaseLevel, pchLine, PARSE_DBSLVL_L);
    Title.strDatabaseLevel[PARSE_DBSLVL_L] = '\0';

    /* Length of title component */
    (void) strncpy(strLength, &(pchLine[3]), PARSE_LENGTH_L);
    strLength[PARSE_LENGTH_L] = '\0';
    Title.nLength = atoi(strLength);
    if (Title.nLength == 0) {
	ParseSetError(PARSE_SYNTAX_ERROR);
	return FALSE;
    }
    /* 04-02-2004 huug : Length overflow removed */

    /* Time */
    (void) strncpy(Title.strTime, &(pchLine[7]), PARSE_TIME_L);
    Title.strTime[PARSE_TIME_L] = '\0';

    /*
     * Read title component delimited by PARSE_CR_CHAR PARSE_CR_CHAR
     * 05-02-2004 : huug : do not use Title.nLength
     */
    if (ParseGetTitle(Title.pchTitleComponent) == 0) {
	return FALSE;
    }

    return TRUE;
} /* ParseGetBody */

/*------------------------------------------------------------------------
 *  Function	: ParseGetHeader
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  21-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static bool
ParseGetHeader(
	void)
{
    char *pchLine;
    char *pchField;

    /* Clear header fields */
    Title.strPPN[0] = '\0';
    Title.strIln[0] = '\0';
    Title.cFormat = '\0';
    Title.nResultCode = 0;
    Title.strMessage[0] = '\0';

    /* Get next line */
    pchLine = ParseGetLine(TRUE);
    if (pchLine == NULL) {
	return FALSE;
    }

    /* Check header field */
    if (strncmp(pchLine, "PPN:", 4) != 0) {
	ParseSetError(PARSE_SYNTAX_ERROR);
	return FALSE;
    }

    pchField = strtok(&(pchLine[4]), PARSE_SPACE);
    if (pchField == NULL) {
	ParseSetError(PARSE_SYNTAX_ERROR);
	return FALSE;
    }

    (void) strncpy(Title.strPPN, pchField, PARSE_PPN_L);
    Title.strPPN[PARSE_PPN_L] = '\0';

    pchField = strtok(NULL, PARSE_SPACE_CR);
    if (pchField == NULL) {
	ParseSetError(PARSE_SYNTAX_ERROR);
	return FALSE;
    } else {
	if (isalpha(pchField[0])) {
	    Title.cFormat = pchField[0];
	} else {
	    (void) sprintf(Title.strIln, "%03d", atoi(pchField));
	    pchField = strtok(NULL, PARSE_SPACE_CR);
	    if (pchField == NULL) {
		ParseSetError(PARSE_SYNTAX_ERROR);
		return FALSE;
	    }
	    Title.cFormat = pchField[0];
	}
	
    }

    /* Check if a result code is present */
    pchField = strtok(NULL, PARSE_SPACE);
    if (pchField == NULL) {
	/* No result code present */
	Title.nResultCode = 0;
	Title.strMessage[0] = '\0';
    } else {
	Title.nResultCode = atoi(pchField);

        pchField = strtok(NULL, PARSE_CR_STR);
        if (pchField == NULL) {
	    ParseSetError(PARSE_SYNTAX_ERROR);
	    return FALSE;
	}

	(void) strncpy(Title.strMessage, pchField, PARSE_MSG_L);
	Title.strMessage[PARSE_MSG_L] = '\0';
    }

    return TRUE;
} /* ParseGetHeader */

/*------------------------------------------------------------------------
 *  Function	: ParseGetLastError
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  21-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
const char*
ParseGetLastError(void)
{
    return Status.strError;
} /* ParseGetLastError */

/*------------------------------------------------------------------------
 *  Function	: ParseGetTitle
 *  Purpose	: read title from file, tite is terminated 
 *		: by a double PARSE_CR_CHAR
 *  Method	: fgetc
 *
 *  History:
 *  05-02-2004	: huug     : creation
 *
 *  Notes	: only one of the two terminating PARSE_CR_CHARs 
 *		: is included in the title
 *
 *  Returns	: length of the title
 *
 *------------------------------------------------------------------------
 */
static int
ParseGetTitle(
	CSBuf_p buffer)
{
    int 	nChar;
    int 	nLastChar = 0;
    bool	ok = TRUE;	/* whether to proceed */

    /* Read line into buffer up to the next two consecutive PARSE_CR character */
    ok = ok && CSClear(buffer);
    while (ok
           && ((nChar = getc(Status.fpTitleFile)) != EOF)
           && (nChar != PARSE_CR_CHAR || nLastChar != PARSE_CR_CHAR)) {
	ok = ok && CSAppendC(buffer, (char)nChar);
	nLastChar = nChar;
    }

    /* Check if a read error occurred */
    if ((nChar == EOF) && !feof(Status.fpTitleFile)) {
	ParseSetError(PARSE_READ_ERROR);
	return 0;
    }

    return ok ? CSLength(buffer) : 0;
} /* ParseGetTitle */

/*------------------------------------------------------------------------
 *  Function	: ParseGetLine
 *  Purpose	: Parse a line which we got :-)
 *  Method	: 
 *
 *  History:
 *  09-02-2005	: michael  : there was no function header
 *  09-02-2005	: michael  : added support for three digit copy field
 *                           (for 99+ copies)
 *
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static char*
ParseGetLine(
	bool bRetainCR)
{
    int nChar;
    int nIndex;

    /* Read line into buffer up to the next PARSE_CR character */
    nIndex = 0;
    while (((nChar = getc(Status.fpTitleFile)) != EOF) &&
	(nChar != PARSE_CR_CHAR) && (nIndex < PARSE_LARGE_BUFFER_L)) {
	pchBuffer[nIndex++] = (char)nChar;
    }

    /* Check if CR character should be retained */
    if ((nChar == PARSE_CR_CHAR) && bRetainCR &&
	(nIndex < PARSE_LARGE_BUFFER_L)) {
        pchBuffer[nIndex++] = (char)nChar;
    }

    /* End buffer properly */
    pchBuffer[nIndex] = '\0';

    /* Check if a read error occurred */
    if ((nChar == EOF) && !feof(Status.fpTitleFile)) {
	ParseSetError(PARSE_READ_ERROR);
	return NULL;
    }

    return pchBuffer;
} /* ParseGetLine */

/*------------------------------------------------------------------------
 *  Function	: ParseGetNextRecord
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  09-02-2005	: michael  : creation (of function header :-))
 *  09-02-2005	: michael  : added support for three digit copy field
 *                           (for 99+ copies)
 *
 *
 *  Notes	:
 *
 *  Returns	:
 *------------------------------------------------------------------------
 */
TitleRec*
ParseGetNextRecord(
	void)
{
    long lCurPos;
    char strPeek[4];

    /* Take a peek at the first 3 characters */
    lCurPos = ftell(Status.fpTitleFile);
    if (fread(strPeek, 1, 3, Status.fpTitleFile) == 0) {
	if (!feof(Status.fpTitleFile)) {
	    ParseSetError(PARSE_READ_ERROR);
	    return NULL;
	}
	return NULL;
    }
    strPeek[3] = '\0';
    (void) fseek(Status.fpTitleFile, lCurPos, SEEK_SET);
    
    if (strcmp(strPeek, "PPN") == 0) {
	Status.bHeaderRead = FALSE;
        if (!ParseGetHeader()) {
	    return NULL;
        }
	Status.bHeaderRead = TRUE;
    } else {
	/* First character of a copy can be a digit now /michael */

        if (
	    !(
		(
			(strncmp(strPeek, "BIB", 3) == 0)
		    ||
			(strncmp(strPeek, "LOK", 3) == 0)
		    ||
		    (
			(
			    ((strPeek[0] == 'E') || isdigit(strPeek[0]))
			    &&
			    isdigit(strPeek[1])
			    &&
			    isdigit(strPeek[2])
			)
			&& Status.bHeaderRead
		    )
		)
	    )
	) {
	    ParseSetError(PARSE_SYNTAX_ERROR);
	    return NULL;
	}
    }

    /* Check result code */
    if (Title.nResultCode == 0) {
        if (!ParseGetBody()) {
	    return NULL;
        }
    }

    return &Title;
} /* ParseGetNextRecord */

/*------------------------------------------------------------------------
 *  Function	: OUFT_ParseInit
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  21-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
bool
OUFT_ParseInit(
	char *strTitleFile)
{
    FILE *fpTmp;

    Status.bErrorSet = FALSE;
    Status.nError = PARSE_NO_ERROR;
    Status.bHeaderRead = FALSE;

    if (Title.pchTitleComponent == NULL
        && (Title.pchTitleComponent = CSCreate()) == NULL) {
        return FALSE;
    }
    fpTmp = fopen(strTitleFile, "r");
    if (fpTmp == NULL) {
	ParseSetError(PARSE_FILE_NOT_OPENED);
	return FALSE;
    }

    (void) strcpy(Status.strTitleFile, strTitleFile);
    Status.fpTitleFile = fpTmp;

    return TRUE;
} /* OUFT_ParseInit */

/*------------------------------------------------------------------------
 *  Function	: ParseSetError
 *  Purpose	: 
 *  Method	: 
 *
 *  History:
 *  21-02-2005	: michael  : creation
 *
 *  Notes	:
 *
 *  Returns	:
 *
 *------------------------------------------------------------------------
 */
static void
ParseSetError(int nError)
{
    if (!Status.bErrorSet) {
	Status.nError = nError;
	Status.bErrorSet = TRUE;
    }
} /* ParseSetError() */
