#!/usr/bin/perl

# First attempt to decoding Pica data into XML
# Reads from file or files, outputs to rundir/xml directory
use strict;
use warnings;
use XML::LibXML;
use File::Copy;
use utf8;
use Config::Tiny;
use LWP::UserAgent;
use File::Path qw(make_path);
use Log::Log4perl qw(:easy);

binmode(STDOUT, 'utf8');

my $confdir = 'confdir';
my $conf = Config::Tiny->read("$confdir/other.ini");
my $ouf_conf = Config::Tiny->read("$confdir/ouf.ini");

my $donedir = $conf->{dir}->{done};
my $xmldir = $conf->{dir}->{xml};
my $wait = $ENV{OUF_SLEEP} || $conf->{settings}->{sleep};
my $trace = $conf->{log}->{decode};
my $lib = $ouf_conf->{UpdateSignal}->{Library};
my $tempdir = $ouf_conf->{Directories}->{"Temp Dir"};
my $hurl = $conf->{harvester}->{url};
if (exists $ENV{HARVESTER_URL}) {
  $hurl = $ENV{HARVESTER_URL};
}

my $migrate = 0;
my $mdirs = 1;
if ($ENV{OUF_MODE} && $ENV{OUF_MODE} =~ /migrate/i) {
  $migrate = 1;
  $wait = $ENV{OUF_SLEEP} || $conf->{migrate}->{sleep};
  $mdirs = $ENV{OUF_DIR_COUNT} || $conf->{migrate}->{dir_count};
  $hurl = '';
  for (my $x = 0; $x < $mdirs; $x++) {
    mkdir "$xmldir/m$x"
  }
}

my $busy_jobs = {};

if ($ENV{OUF_DEBUG}) {
  Log::Log4perl->easy_init({level => $DEBUG, });
} else {
  Log::Log4perl->easy_init({file => "> $trace", level => $DEBUG, });
}

# Set the line separator to CR, as is in Pica files
$/ = "\r";

# Print results to rundir/xml with date/time as filename
my $titledir = $ouf_conf->{Directories}->{"Title Dir"};
my $ending = 0;
$SIG{TERM} = sub { DEBUG("Gracefully stopping $0"); $ending = 1 };

DEBUG("Starting $0");
DEBUG("Running in 'migrate' mode...") if $migrate;

my $file_count = 0;
while (1) {
  my @ts = gmtime();
  my $y = $ts[5] + 1900;
  my $m = $ts[4] + 1;
  my $d = $ts[3];
  my $subdir = sprintf("%04d-%02d-%02d", $y, $m, $d);
  my $prevsub = '';
  my $filename = $ts[5] + 1900 . sprintf("%02d", $ts[4] + 1) . sprintf("%02d", $ts[3]) . sprintf("%02d", $ts[2]) . sprintf("%02d", $ts[1]) . sprintf("%02d", $ts[0]);
  my $xout;
  if ($migrate) {
    my $part = $file_count % $mdirs;
    $xout = "$xmldir/m$part/$filename.xml";
  } else {
    $xout = "$xmldir/$filename.xml";
  }

  for (keys %{ $busy_jobs }) {
    if ($busy_jobs->{$_} > 0) {
      DEBUG("Retrying skipped jobs for $_");
      run_job($hurl, $_);
    }
  }

  my $doc = XML::LibXML::Document->new('1.0', 'utf-8');

  my $root = $doc->createElement('collection');
  $doc->setDocumentElement($root);
  my $deferred = $doc->createElement('deferred_collection');

  my $count = 0;
  foreach (<$titledir/title*>) {
    my $infile = $_;
    
    DEBUG("Converting $_ to XML") unless $migrate;

    my $rawrecord;
    my $section;
    my $metadata;
    my $sectiontype;
    my $ppn;
    my $epn;
 
    open PICA, "<:encoding(utf8)", $infile or DEBUG("WARN There was a problem opening input file!");
    my $record = $doc->createElement('record');
    my $status = 'upsert';
    my $new_sect = 0;
    my $item;
    my $is_item = 0;
    while (<PICA>) {
      chomp();
      $rawrecord .= $_ . "\r";

      if ( ! $_ ) {
        $new_sect = 1;
        $is_item = 0;
      } else {
        if (!$metadata) {
          $metadata = $doc->createElement('metadata');
          $record->appendChild($metadata);
        }
        if (/^([^\/ ]+)(\/(\d+))? (ƒ.*)$/) {
          my $tag = $1;
          if ($new_sect) {
            if ($tag =~ /^2/) {
              $item = $doc->createElement('item');
              $metadata->appendChild($item);
              $is_item = 1;
            }
            $new_sect = 0;
          }
          my $occurrence = $3;
          my $fulltag = $tag;
          if ($occurrence) { $fulltag .= "/" . $occurrence; }
          my $fielddata = $4;
          my $fld = $doc->createElement('datafield');
          $fld->setAttribute('tag' => $tag);
          if ($occurrence) {
            $fld->setAttribute('occurrence' => $occurrence);
          }
          $fld->setAttribute('fulltag' => $fulltag);
          if ($is_item) {
            $item->appendChild($fld);
          } else {
            $metadata->appendChild($fld);
          }
          my @subs = ($fielddata =~ /ƒ(.)([^ƒ]+)/g);
          if ($tag eq '003@') {
            $ppn = $subs[1];
          } elsif ($tag eq '203@') {
            $epn = $subs[1];
            $item->setAttribute('epn' => $epn);
          }
          my $sfi = 0;
          while (defined($subs[$sfi])) {
            my $subfieldcode = $subs[$sfi++];
            my $txt = $subs[$sfi++];
            $txt =~ s/\$\w/ /g if $subfieldcode eq '8';
            if ($txt !~ /^\s*$/) {
              my $sf = $doc->createElement('subfield');
              $sf->setAttribute('code' => $subfieldcode);
              $fld->appendChild($sf);
              $sf->appendTextNode($txt);
            }
          }
        } elsif (/^STATUS:delete/) {
          $status = 'deleted';
        }
      }
    }
    # Store the raw record in a CDATA section, to avoid entity-encoding things
    # The record looks ugly, as the \r makes the text overwrite it. You can
    # see it better with piping it through   sed 's/\r/XX\n/g'
    my $raw = $doc->createElement('rawrecord');
    my $cdata = XML::LibXML::CDATASection->new($rawrecord);
    $raw->appendChild($cdata);
    $record->appendChild($raw);
    my $id_node = $doc->createElement('hrid');
    $id_node->appendTextNode($ppn);
    $record->insertBefore($id_node, $metadata);
    my $st_node = $doc->createElement('status');
    $st_node->appendTextNode($status);
    $record->insertBefore($st_node, $id_node);

    if ($migrate == 0 && -e "$tempdir/deferred_$ppn") {
      DEBUG("Adding PPN $ppn to deferred collection");
      $deferred->appendChild($record);
      unlink "$tempdir/deferred_$ppn";
    } elsif ($status eq 'deleted') {
      DEBUG("Adding deleted PPN $ppn to deferred collection");
      $deferred->appendChild($record);
    } else {
      DEBUG("Adding PPN $ppn to collection");
      $root->appendChild($record);
    }

    # If in migrate mode, then delete title file, otherwise move it to done directory
    if ($migrate) {
      unlink $infile;
    } else {
      make_path("$donedir/$subdir") if $subdir ne $prevsub;
      move($infile, "$donedir/$subdir") or DEBUG("ERROR Couldn't move $_ to the $donedir");
      $prevsub = $subdir;
    }

    $count++;
  }

  if ($deferred->hasChildNodes()) {
    my $def_count = 0;
    foreach my $rec ($deferred->childNodes()) {
      $def_count++;
      $root->appendChild($rec);
    }
    DEBUG("$def_count records added from deferred collection");
  }

  if ($count > 0) {
    $root->setAttribute('count' => $count);
    DEBUG("Writing $count records to $xout");
    open XMLOUT, ">$xout" or DEBUG("ERROR Can't write to $xout");
    print XMLOUT $doc->toString(2);
    close XMLOUT;
    $file_count++;
    if ($hurl =~ /^http/) {
     run_job($hurl, $lib);
    }
  }

  DEBUG("Waiting $wait seconds...");
  sleep $wait;
  last if $ending;
}


sub run_job {
  my $url = shift;
  my $iln= shift;
  my $path = shift;

  my $brief = get_brief_job($url, $iln);
  my $jid = $brief->{id};
  if ($jid) {
    if ($brief->{status} eq 'RUNNING') {
      my $name = $brief->{name};
      $busy_jobs->{$iln}++;
      DEBUG("WARN Job $jid -- $name is already running, skipping for now...");
    } else {
      my $jobdom = get_job($url, $jid);
      my $putdom = set_job_elements($jobdom, $path);
      put_job($url, $jid, $putdom);
      $busy_jobs->{$iln} = 0;
    }
  }
}

sub set_job_elements {
  my $dom = shift;
  my $burl = shift;
  DEBUG("Updating payload");
  my ($immediate) = $dom->findnodes('//harvestImmediately/text()');
  my $imtext = XML::LibXML::Text->new('true');
  $immediate->setData($imtext);
  if ($dom->exists('//scheduleString/text()')) {
    my ($schedule) = $dom->findnodes('//scheduleString/text()');
    my $sctext = XML::LibXML::Text->new('');
    $schedule->setData($sctext);
  }
  my ($update) = $dom->findnodes('//lastUpdated/text()');
  my $uptext = XML::LibXML::Text->new(&new_date);
  $update->setData($uptext);
  foreach (@{ $dom->findnodes('//script/text()') }) {
    my $empty_script = XML::LibXML::Text->new('');
    $_->setData($empty_script);
  }
  if ($burl) {
    my ($bulkurl) = $dom->findnodes('//xmlBulk/url/text()');
    my $urtext = XML::LibXML::Text->new($burl);
    $bulkurl->setData($urtext);
  }
  return $dom;
}

# put complete harvester job settings-- immediate start.
sub put_job {
  my $url = shift;
  my $jid = shift;
  my $dom = shift;
  my $content = $dom->toString(2);
  my @call = caller(0);
  my $ua = LWP::UserAgent->new();

  DEBUG("Putting updated payload to $url/harvester/records/harvestables/$jid");
  my $res = $ua->put("$url/harvester/records/harvestables/$jid", "Content-type" => "application/xml", Content => $content);
  
  if ($res->is_success) {
      my $dummy = $res->decoded_content;
  }
  else {
    DEBUG("WARN $call[3]: " . $res->status_line . "\n");
  }
  return $dom;
}

# get complete harvester job settings by job id.
sub get_job {
  my $url = shift;
  my $jid = shift;
  my $dom = {};
  my @call = caller(0);
  my $ua = LWP::UserAgent->new();

  my $furl = "$url/harvester/records/harvestables/$jid";
  DEBUG("Getting harvest job settings from $furl");
  my $res = $ua->get($furl);
  
  if ($res->is_success) {
      my $job = $res->decoded_content;
      $dom = XML::LibXML->load_xml(string=>$job);
  }
  else {
    DEBUG("WARN $call[3]: " . $res->status_line);
  }
  return $dom;
}

# get harvester job id using library code.  The code must be wrapped in parenthesis in the harvest job name 
sub get_brief_job {
  my $url = shift;
  my $lib = shift;
  my $jid = '';
  my $status = '';
  my $name = '';
  my @call = caller(0);
  my $ua = LWP::UserAgent->new();

  my $furl = "$url/harvester/records/harvestables/?query=(usedBy=$lib)";
  DEBUG("Finding harvest job at $furl");
  my $res = $ua->get("$furl");
  
  if ($res->is_success) {
      my $brief_recs = $res->decoded_content;
      my $dom = XML::LibXML->load_xml(string=>$brief_recs);
      foreach ($dom->findnodes("//id")) {
        $jid = $_->to_literal();
      }
      foreach ($dom->findnodes("//currentStatus")) {
        $status = $_->to_literal();
      }
      foreach ($dom->findnodes("//name")) {
        $name = $_->to_literal();
      }
      if (!$jid) {
        DEBUG("WARN Can't find harvest job with library code $lib-- aborting harvester process!");
      }
  }
  else {
    DEBUG("WARN $call[3]: " . $res->status_line);
  }
  return { id=>$jid, status=>$status, name=>$name };
}

sub new_date {
  my @t = gmtime();
  my $y = $t[5] + 1900;
  my $m = sprintf("%02d", $t[4] + 1);
  my $d = sprintf("%02d", $t[3]);
  my $h = sprintf("%02d", $t[2]);
  my $min = sprintf("%02d", $t[1]);
  my $sec = sprintf("%02d", $t[0]);
  my $date = "$y-$m-${d}T$h:$min:${sec}Z";
  return $date;
}
