#!/usr/bin/perl

# picapull.pl  - script to pull updated records from GBV's system
#
# Overview:
# - ouf_update created updsig files for records that have been changed
# - This script reads them, deduplicates PPNs, and produces fetch files
# - This causes ouf_fetch to retrieve the records
#
# We produce one fetch file for each PPN we see. That way, we can see if
# ouf_fetch renamed it to .err to indicate an error, and see which PPN
# we had problems with. Also, this is an extra safeguard for deduplicating,
# if the file already exists, perhaps from a previous run.
#
# TODO: Check for .err files in the fetch directory, move them away to
# somewhere else.

# Run from the main directory
use strict;
use warnings;

use File::Copy;
use File::Path qw(make_path);
use Config::Tiny;
use Log::Log4perl qw(:easy);

my $confdir = 'confdir';
my $conf = Config::Tiny->read("$confdir/other.ini");
my $ouf_conf = Config::Tiny->read("$confdir/ouf.ini");

my $updsigdir = $ouf_conf->{Directories}->{"Update Signal Dir"};
my $fetchdir = $ouf_conf->{Directories}->{"Fetch Request Dir"};
my $titledir = $ouf_conf->{Directories}->{"Title Dir"};
my $tempdir = $ouf_conf->{Directories}->{"Temp Dir"};
my $donedir = $conf->{dir}->{processed};
my $wait = $ENV{OUF_SLEEP} || $conf->{settings}->{sleep};
my $trace = $conf->{log}->{picapull};
my $migrate = 0;

if ($ENV{OUF_MODE} && $ENV{OUF_MODE} =~ /migrate/i) {
  $migrate = 1;
}

if ($ENV{OUF_DEBUG}) {
  Log::Log4perl->easy_init({level => $DEBUG, });
} else {
  Log::Log4perl->easy_init({file => "> $trace", level => $DEBUG, });
}

my $ending = 0;
$SIG{TERM} = sub { DEBUG("Gracefully stopping $0"); $ending = 1 };

DEBUG("Starting $0");

my $prevsub = '';
while (1) {
  my $nfiles = 0;
  my $nlines = 0;
  my $seen = {};
  my @t = localtime();
  my $y = $t[5] + 1900;
  my $m = $t[4] + 1;
  my $d = $t[3];
  my $subdir = sprintf("%04d-%02d-%02d", $y, $m, $d);
  for my $fn ( <"$updsigdir/updsig*">) {
    my $deleted_ppn = {};
    my $deleted_epn = {};
    DEBUG("Processing updsig file: $fn");
    $nfiles++;
    open my $F, "<$fn" or ERROR("Could not open $fn");
    while(<$F>) {
      chomp();
      my ( $user, $library, $date, $time, $type, $ppn, $level, $copy, $mat, $epn ) =  split( / +/, $_);
      next if $deleted_ppn->{$ppn};
      if (!$ppn) {
        ERROR("OOPS, no PPN in line $_");
      } elsif ($mat =~ /^T/) {
        DEBUG("Looks like an authority record -- skipping ($_)");
      } elsif ($level eq '0' && $type =~ /\[(DE|OU)\]/) { # looks like a deleted record
        if ($migrate == 0) {
          my $dfile = "$titledir/title_Z${ppn}"; # prepend Z to ppn in the filename so that decode.pl processes these last.
          DEBUG("Delete line detected-- writing to $dfile");
          my $drec = "PPN:$ppn PXB\rSTATUS:delete\r003@ ƒ0$ppn";
          open DEL, ">", $dfile or ERROR("Can't open delete file!");
          print DEL $drec;
          close DEL;
          $deleted_ppn->{$ppn} = 1;
        }
      } else {
        $nlines++;
        $seen->{$ppn} .= "$level:$type ";
      }
    }
    if ($migrate) {
      unlink $fn;
    } else {
      make_path("$donedir/$subdir") if $subdir ne $prevsub;
      move($fn, "$donedir/$subdir");
      $prevsub = $subdir;
    }
  }

  # Write the fetch files
  my $uniq = scalar( keys(%{ $seen }) );
  DEBUG("$nlines lines from $nfiles files resulted in $uniq unique ppns") if $uniq;
  for my $ppn (keys(%{ $seen })) {
    if ($seen->{$ppn} =~ /2:\[DE\]/ && $migrate == 0) { # Detect if record has deleted items
      open DEFILE, ">$tempdir/deferred_$ppn";
      print DEFILE "1\n";
    }
    my $fetchfilename = "$fetchdir/fetch_$ppn";
    DEBUG("Saving fetch file as $fetchfilename");
    open my $F, ">$fetchfilename" or ERROR("Could not open fetch file $fetchfilename");
    print $F "$ppn\n";
    close $F;
  }

  DEBUG("Waiting $wait seconds...");
  last if $ending;
  sleep $wait;
}
